/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine;

import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.util.CompilationUnitSet;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.TransformationState;


/**
 * Esta classe  representa o engine de Transformao do JaTS e  responsvel
 * por realizar as tranformaes propriamente ditas. Trata-se basicamente de um
 * componente que isola as operaes  relacionadas a tranformaes
 *
 * @author Adeline de Sousa Silva
 * @author Fernando Castor
 */
public class JatsEngine {

    // realiza os casamentos
    private JatsMatcher matcher;
    // realiza as substituies
    private JatsReplacer replacer;
    // clona a rvore
    private JatsClonner clonner;

    private static JatsEngine instance;


    /**
     * Retorna a nica instncia da classe JatsEngine
     *
     * @return  instncia nica da classe JatsEngine
     */
    public static JatsEngine getInstance() {
        if (instance == null) {
            instance = new JatsEngine();
        }
        return instance;
    }


    /**
     * Construtor da classe
     */
    private JatsEngine() {

        this.matcher = new JatsMatcher();
        this.replacer = new JatsReplacer();
        this.clonner = new JatsClonner();
    }


    /**
     * Faz a aplicao de uma transformao, ou seja, realiza o casamento entre
     * ladoEsquerdo e fonteJava e depois realiza a substituicao em ladoDireito
     * com o ResultSet  resultado do casamento entre os dois primeiros
     */
    public INode apply(INode source, TransformationState transState)
			throws IllegalArgumentException, NodesNotMatchedException,
			InconsistentNodeException, ExecutionException {

        if (source == null || transState == null) {
            throw new IllegalArgumentException();
        }

        
        //se a transformacao nao tiver o mesmo tipo do no'
        //retorna sem processamento
        if (!transState.getTransformation(0).getType().isInstance(source)) {
        	return source;
        }
        
        INode retorno = null;
        
        
        for (int i = 0; i < transState.getTransformation().size(); i++) {
	        // pega cada lado da transformao
	        INode left = transState.getSource(i);
	        INode right = transState.getTarget(i);
	        ResultSet rset = transState.getResultSetClone();
	
	        // clona o lado direito
	        INode targetClone = this.clonner.clone(right);
	
	        // realiza o casamento entre o lado esquerdo da transformacao e o
	        // programa java
	        try {
				this.matcher.match(left, source, rset);
			} catch (NodesNotMatchedException e) {
				continue;
			}
	
	        // faz a substituicao no programa destino, das variaveis pelos valores
	        // a elas mapeados no ResultSet
	        this.replacer.replace(targetClone, rset);

	        if (transState.getTransformation(0).getPrecondition() != null)
	        	this.replacer.replace(transState.getTransformation(0).getPrecondition(), rset);
	        
	        if (!transState.getTransformation(0).getPreconditionValue(rset)) {
	        	break;
	        }
	        
	        // realiza o processamento dos ns executveis
	        targetClone.process(rset);
	        retorno = targetClone;
	        break;
        }
        return retorno;
    }

    
    /**
     * Faz a aplicao de uma transformao, ou seja, realiza o casamento entre
     * ladoEsquerdo e fonteJava e depois realiza a substituicao em ladoDireito
     * com o ResultSet  resultado do casamento entre os dois primeiros
     */
    public INode apply(INode ladoEsquerdo,
                       INode fonteJava, 
                       INode ladoDireito,
                       ResultSet resultados) 
    		throws IllegalArgumentException, NodesNotMatchedException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, ExecutionException {

        if (ladoDireito == null) {
            throw new IllegalArgumentException("Direita da transformacao eh null");
        }

        INode retorno = null;
//        if (ladoDireito.getPreconditionValue(resultados)) {
            INode cloneDestino = (INode) this.clonner.clone(ladoDireito);

            // realiza o casamento entre o lado esquerdo da transformacao e o
            // programa java
            this.matcher.match(ladoEsquerdo, fonteJava, resultados);

            // faz a substituicao no programa ladoDireito, das variaveis pelos valores
            // a elas mapeados no ResultSet
            this.replacer.replace(cloneDestino, resultados);

            // realiza o processamento dos ns executveis
            cloneDestino = (INode) cloneDestino.process(resultados);
            
            if (cloneDestino instanceof ICompilationUnit)
            	PrePrettyPrinter.preprocess((ICompilationUnit)cloneDestino);
//            if (cloneDestino.getPreconditionValue(resultados)) {
                retorno = cloneDestino;
//            }
//        }
        return retorno;
    }
    
    

    /**
     * Faz a aplicao de uma transformao, ou seja, realiza o casamento entre
     * ladoEsquerdo e fonteJava e depois realiza a substituicao em ladoDireito
     * com o ResultSet  resultado do casamento entre os dois primeiros
     */
    public ICompilationUnit apply(ICompilationUnit ladoEsquerdo,
                                  ICompilationUnit fonteJava, ICompilationUnit ladoDireito,
                                  ResultSet resultados) throws IllegalArgumentException,
            NodesNotMatchedException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            ExecutionException {

        if (ladoDireito == null) {
            throw new IllegalArgumentException("Direita da transformacao eh null");
        }

        ICompilationUnit retorno = null;
        if (ladoDireito.getPreconditionValue(resultados)) {
            ICompilationUnit cloneDestino = (ICompilationUnit) this.clonner.clone(ladoDireito);

            // realiza o casamento entre o lado esquerdo da transformacao e o
            // programa java
            this.matcher.match(ladoEsquerdo, fonteJava, resultados);

            // faz a substituicao no programa ladoDireito, das variaveis pelos valores
            // a elas mapeados no ResultSet
            this.replacer.replace(cloneDestino, resultados);

            // realiza o processamento dos ns executveis
            cloneDestino.process(resultados);
            PrePrettyPrinter.preprocess(cloneDestino);
            if (cloneDestino.getPreconditionValue(resultados)) {
                retorno = cloneDestino;
            }
        }
        return retorno;
    }


    /**
     *  Faz a aplicao de uma transformao, ou seja, realiza o casamento entre
     *  ladosEsquerdos e fontesJava, executa todas as estruturas executveis do
     *  lado direito da transformao e depois realiza a substituicao do ResultSet
     *  resultado do casamento.
     *  @return O conjunto de rvores sintticas do lado direito da <br>
     *		  transformao, devidamente executadas e substitudas.
     */
    public CompilationUnitSet apply(CompilationUnitSet ladosEsquerdos,
                                    CompilationUnitSet fontesJava, CompilationUnitSet ladosDireitos,
                                    ResultSet resultados) throws IllegalArgumentException,
            NodesNotMatchedException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            ExecutionException {

        if (ladosDireitos == null) {
            throw new IllegalArgumentException("Direita da transformacao eh null");
        }
        CompilationUnitSet clonesDestino = new CompilationUnitSet();
        CompilationUnitSet retorno = new CompilationUnitSet();
        int i = 0;
        while (i < ladosDireitos.size()) {
            ICompilationUnit tmp = ladosDireitos.getCompilationUnit(i);
            if (tmp.getPreconditionValue(resultados)) {
                ICompilationUnit compUnit =
                        (ICompilationUnit) (this.clonner.clone(ladosDireitos.getCompilationUnit(i)));
                clonesDestino.addCompilationUnit(compUnit);
            }
            i++;
        }

        this.match(ladosEsquerdos, fontesJava, resultados);

        this.replace(clonesDestino, resultados);
        i = 0;

        ICompilationUnit tmpCompUnit;
        while (i < clonesDestino.size()) {
            tmpCompUnit = clonesDestino.getCompilationUnit(i);
            tmpCompUnit.process(resultados);
            if (tmpCompUnit.getPreconditionValue(resultados)) {
                retorno.addCompilationUnit(tmpCompUnit);
            }
            i++;
        }
        return retorno;
    }
    
    public NodeList apply(NodeList ladosEsquerdos,
			NodeList fontesJava, NodeList ladosDireitos, ResultSet resultados)
			throws IllegalArgumentException, NodesNotMatchedException,
			cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
			ExecutionException {

		if (ladosDireitos == null) {
			throw new IllegalArgumentException(
					"Direita da transformacao eh null");
		}
		NodeList clonesDestino = new NodeList();
		NodeList retorno = new NodeList();
		int i = 0;
		while (i < ladosDireitos.size()) {
			INode tmp = ladosDireitos.elementAt(i);
			
			if ((tmp instanceof ICompilationUnit && ((ICompilationUnit)tmp).getPreconditionValue(resultados))
					|| !(tmp instanceof ICompilationUnit)) {
				INode compUnit = this.clonner
						.clone(ladosDireitos.elementAt(i));
				clonesDestino.addElement(compUnit);
			}
			i++;
		}

		this.match(ladosEsquerdos, fontesJava, resultados);

        i = 0;
        while (i < clonesDestino.size()) {
            this.replace(clonesDestino.elementAt(i), resultados);
            i++;
        }
		
		i = 0;
		INode tmpCompUnit;
		while (i < clonesDestino.size()) {
			tmpCompUnit = clonesDestino.elementAt(i);
			tmpCompUnit = (INode) tmpCompUnit.process(resultados);
			if ((tmpCompUnit instanceof ICompilationUnit && ((ICompilationUnit) tmpCompUnit)
					.getPreconditionValue(resultados))
					|| !(tmpCompUnit instanceof ICompilationUnit)) {
				retorno.addElement(tmpCompUnit);
			}
			i++;
		}
		return retorno;
	}

    /**
     * Realiza as substituies e o processamento nas rvores passadas como
     * parametro.
     */
    public CompilationUnitSet apply(CompilationUnitSet ladosDireitos, ResultSet resultados)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            ExecutionException {

        if (ladosDireitos == null) {
            throw new IllegalArgumentException("Direita da transformacao eh null");
        }
        CompilationUnitSet clonesDestino = new CompilationUnitSet();
        CompilationUnitSet retorno = new CompilationUnitSet();
        int i = 0;
        while (i < ladosDireitos.size()) {
            ICompilationUnit tmp = ladosDireitos.getCompilationUnit(i);

            if (tmp.getPreconditionValue(resultados)) {
                ICompilationUnit compUnit =
                        (ICompilationUnit) (this.clonner.clone(ladosDireitos.getCompilationUnit(i)));
                clonesDestino.addCompilationUnit(compUnit);
            }
            i++;
        }

        this.replace(clonesDestino, resultados);
        i = 0;
        ICompilationUnit tmpCompUnit = null;

       while (i < clonesDestino.size()) {
            tmpCompUnit = clonesDestino.getCompilationUnit(i);
            System.out.println("Cloning");
            tmpCompUnit.process(resultados);
            if (tmpCompUnit instanceof JCompilationUnit)
            	PrePrettyPrinter.preprocess(tmpCompUnit);                 	
            	
            if (tmpCompUnit.getPreconditionValue(resultados)) {
                retorno.addCompilationUnit(tmpCompUnit);
            }
            i++;
        }

        return retorno;
    }


    /**
     * Realiza o matching entre n JaTS e o n Java
     */
    public void match(INode left, INode source, ResultSet results)
            throws IllegalArgumentException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, NodesNotMatchedException {

        this.matcher.match(left, source, results);
    }

    /**
     * Realiza o matching entre o programa JaTS  ladoEsquerdo e o  programa Java
     * fonteJava. O ResultSet retornado corresponde ao resultado  do matching,
     * se houver o casamento.
     */
    public void match(ICompilationUnit ladoEsquerdo, ICompilationUnit fonteJava,
                      ResultSet resultados) throws IllegalArgumentException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, NodesNotMatchedException {

        this.matcher.match(ladoEsquerdo, fonteJava, resultados);
    }


    /**
     * Mtodo match que realiza o casamento entre um nmero n lados esquerdos de
     * transformao e um  nmero n de fontes em Java e guarda todos os resultados
     * dos casamentos em um conjunto resultado.
     */
    public void match(CompilationUnitSet ladosEsquerdos,
                      CompilationUnitSet fontesJava, ResultSet resultados)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {
        if (ladosEsquerdos == null) {
            throw new IllegalArgumentException("O conjunto de programas JaTS " +
                    "correspondente ao lado esquerdo" +
                    " da transformao no  vlido.");
        } else if (fontesJava == null) {
            throw new NodesNotMatchedException("O conjunto de programas fonte em" +
                    " Java no  vlido.");
        }

        if (ladosEsquerdos.size() != fontesJava.size()) {
            throw new IllegalArgumentException("Os dois conjuntos de programas" +
                    " passados no podem ser casados." +
                    " leftSize "+ ladosEsquerdos.size() +" != "+ fontesJava.size() +" rightSize" );
        }
        int i = 0;
        while (i < ladosEsquerdos.size()) {
            System.out.println("Matching");
            this.match(ladosEsquerdos.getCompilationUnit(i),
                    fontesJava.getCompilationUnit(i), resultados);
            i++;
        }
    }
    
    public void match(NodeList sourcePatterns,
    		NodeList sourceFiles, ResultSet results)
			throws IllegalArgumentException,
			cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
			NodesNotMatchedException {
        int i = 0;
        while (i < sourcePatterns.size()) {
            System.out.println("Matching");
            this.match(sourcePatterns.elementAt(i),
                    sourceFiles.elementAt(i), results);
            i++;
        }

	}


    /**
	 * 
	 */
    public void replace(INode right, ResultSet results)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        this.replacer.replace(right, results);
    }

    /**
     * Realiza as substituicoes no ladoDireito da transformacao, a partir do
     * ResultSet passado como parametro. Este mtodo pode ser utilizado em
     * situaes onde haver apenas gerao de cdigo.
     */
    public void replace(ICompilationUnit ladoDireito, ResultSet resultados)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        this.replacer.replace(ladoDireito, resultados);
    }


    /**
     * Mtodo replace para um conjunto de rvores
     */
    public void replace(CompilationUnitSet ladosDireitos, ResultSet resultados)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (ladosDireitos == null) {
            throw new IllegalArgumentException("O conjunto de lados direitos de" +
                    " transformao passado no  " +
                    "vlido.");
        }
        int i = 0;
        while (i < ladosDireitos.size()) {
            this.replace(ladosDireitos.getCompilationUnit(i), resultados);
            i++;
        }
    }


}
