/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine;

import java.io.*;
import java.util.Properties;
import java.util.Vector;

import cin.jats.engine.parser.ParseException;
import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.metajats.JaTSProgram;
import cin.jats.engine.util.CompilationUnitSet;
import cin.jats.engine.util.NodeTable;
import cin.jats.engine.util.Util;

/**
 * This class accomplishes the JaTS I/O operations. It translates Strings 
 * (programs in text format) to ASTs and ASTs to Strings.
 *
 * Program -> AST (parsing)
 * AST -> Program (unparsing or pretty-printing)
 * 
 * @author Adeline de Sousa Silva
 * @author Fernando Castor
 * @author Gustavo Santos
 * 
 * @version 14/10/2005
 */
public class JatsIO {

    // unique instance of JatsIO (singleton)
    private static JatsIO instance;

    // parser reference used in all methods
    private IParser parser;

    // parser for patterns (templates)
    private IParser patternLangParser;

    // parser for base language fragments
    private IParser baseLangParser;
    
    // AST to text translator (prints the program)
    private PrettyPrinter prettyPrinter;

    // node clonner
    private JatsClonner clonner;

    // table to keep the AST nodes
    private NodeTable nodeTable;

    /**
     * Constructor for JatsIO objects
     */
    private JatsIO() {
        this.baseLangParser = ParserFactory.createParser(ParserFactory.JAVA_PARSER);
        this.patternLangParser = ParserFactory.createParser(ParserFactory.JATS_PARSER);
        this.prettyPrinter = PrettyPrinter.getInstance();
        this.clonner = new JatsClonner();
        this.nodeTable = new NodeTable();
    }

    /**
     * Constructor for JatsIO objects
     * 
     * @param parserType    Parser type to be used as default
     */
    private JatsIO(int parserType) {
        this.parser = ParserFactory.createParser(parserType);
        this.prettyPrinter = PrettyPrinter.getInstance();
        this.clonner = new JatsClonner();
        this.nodeTable = new NodeTable();
    }
    
    /**
     * Retrieves the unique instance of JatsIO
     * 
     * @return The JatsIO instance
     */
    public static JatsIO getInstance() {
        if (instance == null) {
            instance = new JatsIO();
        }
        return instance;
    }

    /**
     * Creates a new instance (releasing the old one) and returns it
     * 
     * @param parserType    Parser type
     * @return The new instance
     */
    public static JatsIO getNewInstance(int parserType) {       
        instance = new JatsIO(parserType);
       
        return instance;
    }
    
    /**
     * Changes the parser to be used in the class
     * 
     * @param parserType
     */
    public void changeParserType(int parserType) {
        if (parserType == ParserFactory.JATS_PARSER) {
        	this.parser = this.patternLangParser;
        } else if (parserType == ParserFactory.JAVA_PARSER) {
        	this.parser = this.baseLangParser;
        }
        	
    }
    
    
    /*********               Mtodos de Pretty-Printing             **********/


    public void setPrettyPrinterProperties(Properties prop) {
        this.prettyPrinter.setPrettyPrinterProperties(prop);
    }

    /**
     * Realiza o pretty printing da sub-rvore que tem node como  raiz no Writer
     * output. O Writer vai ser passado como parmetro e no retornado como um
     * novo objeto por que assim o cliente pode escolher qual o tipo do Writer
     * aonde ele quer que o pretty printing seja realizado (e nao fica limitado
     * apenas ao tipo que for fixado aqui).
     */
    public void prettyPrint(INode node, Writer output)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException, IOException,
            ExecutionException {

        this.prettyPrinter.prettyPrint(node, output);
    }

    public void prettyPrint(INode node, Writer output, boolean jatsTypesEnabled)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException, IOException,
            ExecutionException {
        this.prettyPrinter.prettyPrint(node, output, jatsTypesEnabled);
    }



    /*********                  Mtodos de Parsing                  **********/

    /**
     * Realiza o parser de um conjunto de arquivos, retornando um
     * CompilationUnitSet, onde cada CompilationUnit do conjunto corresponde 
     * raiz da rvore sinttica gerada. Os argumentos, podem, tanto ser um
     * array de Readers quanto de Strings
     */
    public CompilationUnitSet parse(Reader[] entrada) throws ParseException,
            IllegalArgumentException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, ExecutionException {

        int tam = entrada.length;
        CompilationUnitSet resultado = new CompilationUnitSet();

        for (int i = 0; i < tam; i++) {
            ICompilationUnit compUnit =
                    (ICompilationUnit) this.clonner.clone(this.parse(entrada[i]));
            this.nodeTable = this.parser.getNodeTable();
            resultado.addCompilationUnit(compUnit);
        }
        return resultado;

    }

    public CompilationUnitSet parse(String[] entrada) throws ParseException,
            IllegalArgumentException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, ExecutionException {

        int tam = entrada.length;
        CompilationUnitSet resultado = new CompilationUnitSet();
        for (int i = 0; i < tam; i++) {
            ICompilationUnit compUnit =
                    (ICompilationUnit) this.clonner.clone(this.parse(entrada[i]));
            this.nodeTable = this.parser.getNodeTable();

            resultado.addCompilationUnit(compUnit);
        }
        return resultado;
    }

    public CompilationUnitSet parseFiles(String[] files) throws ParseException,
            FileNotFoundException, IllegalArgumentException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, ExecutionException {

        CompilationUnitSet resultado;
        if (files != null) {
            int tam = files.length;
            resultado = new CompilationUnitSet();
            for (int i = 0; i < tam; i++) {
                ICompilationUnit compUnit =
                        (ICompilationUnit) this.clonner.clone(this.parseFile(files[i]));
                this.nodeTable = this.parser.getNodeTable();

                resultado.addCompilationUnit(compUnit);
            }
        } else {
            throw new IllegalArgumentException();
        }
        return resultado;
    }


    public Pattern parsePatternFile(Reader entrada) throws ParseException,
			IllegalArgumentException {
		Pattern retorno = null;
		this.parser.reInitTable();
		retorno = ((JatsParser)this.parser).parsePatternFile(entrada);
		return retorno;
	}
    
    public JaTSProgram parseMetaProgram(Reader entrada) throws ParseException,
			IllegalArgumentException {
    	JaTSProgram retorno = null;
		this.parser.reInitTable();
		retorno = ((JatsParser) this.parser).parseMetaProgram(entrada);
		return retorno;
	}
    
    /**
     * Realiza o parsing propriamente dito e retorna um n do tipo
     * IType, correspondente  raiz da rvore sinttica
     * gerada. O argumento passado pode ser o String correspondente
     * ao nome do arquivo a ser aberto ou o Reader correspondente
     */
    public ICompilationUnit parse(Reader entrada) throws ParseException,
            IllegalArgumentException {
        ICompilationUnit retorno = null;
        this.parser.reInitTable();
        retorno = this.parser.parse(entrada);
        this.nodeTable = this.parser.getNodeTable();
        return retorno;

    }

    public ICompilationUnit parse(String entrada) throws ParseException,
            IllegalArgumentException {
        this.parser.reInitTable();
        return this.parser.parse(entrada);
    }

    public ICompilationUnit parse(InputStream entrada) throws ParseException,
            IllegalArgumentException {
        this.parser.reInitTable();
        InputStreamReader reader = new InputStreamReader(entrada);
        return this.parse(reader);
    }

    public ICompilationUnit parseFile(String fileName) throws ParseException,
            FileNotFoundException, IllegalArgumentException {
        this.parser.reInitTable();
        return this.parser.parseFile(fileName);
    }

    /**
     * Realiza o parsing de uma declarao de mtodo. A entrada pode tanto ser
     * uma String quanto um Reader
     */
    public JMethodDeclaration parseMethod(Reader entrada) throws ParseException,
            IllegalArgumentException {

        return this.parser.parseMethod(entrada);
    }

    public JMethodDeclaration parseMethod(String entrada) throws ParseException,
            IllegalArgumentException {

        return this.parser.parseMethod(entrada);
    }

    /**
     * Realiza o parsing de uma declarao de atributo. A entrada pode tanto ser
     * uma String quanto um Reader
     */
    public JVariableDeclaration parseField(Reader entrada) throws ParseException,
            IllegalArgumentException {

        return this.parser.parseField(entrada);
    }

    public JVariableDeclaration parseField(String entrada) throws ParseException,
            IllegalArgumentException {
        return this.parser.parseField(entrada);
    }

    /**
     * Realiza o parsing de uma declarao de construtor. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JConstructorDeclaration parseConstructor(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return this.parser.parseConstructor(entrada);
    }

    public JConstructorDeclaration parseConstructor(String entrada)
            throws ParseException, IllegalArgumentException {
        return this.parser.parseConstructor(entrada);
    }

    /**
     * Realiza o parsing de uma declarao de import. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JImportDeclaration parseImport(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseImport(entrada);
    }

    public JImportDeclaration parseImport(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseImport(entrada);
    }


    /**
     * Realiza o parsing de uma declarao de pacote. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JPackageDeclaration parsePackage(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parsePackage(entrada);
    }

    public JPackageDeclaration parsePackage(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parsePackage(entrada);
    }


    /**
     * Realiza o parsing de um identificador. A entrada pode ser uma String ou
     * um reader
     */
    public JIdentifier parseIdentifier(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseIdentifier(entrada);
    }

    public JIdentifier parseIdentifier(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseIdentifier(entrada);
    }


    /**
     * Realiza o parsing de um name. A entrada pode ser uma String ou
     * um reader
     */
    public JName parseName(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseName(entrada);
    }

    public JName parseName(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseName(entrada);
    }


    /**
     * Realiza o parsing de um name. A entrada pode ser uma String ou
     * um reader
     */
    public JLiteral parseLiteral(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseLiteral(entrada);
    }

    public JLiteral parseLiteral(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseLiteral(entrada);
    }


    /**
     * Realiza o parsing de um tipo. A entrada pode ser uma String ou
     * um reader
     */
    public JType parseType(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseType(entrada);
    }

    public JType parseType(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseType(entrada);
    }
    
    public JType parseResultType(Reader entrada) throws ParseException,
			IllegalArgumentException {
		return parser.parseResultType(entrada);
	}

	public JType parseResultType(String entrada) throws ParseException,
			IllegalArgumentException {
		return parser.parseResultType(entrada);
	}

    /**
     * Realiza o parsing de um tipo. A entrada pode ser uma String ou
     * um reader
     */
    public JBlock parseBlock(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseBlock(entrada);
    }

    public JBlock parseBlock(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseBlock(entrada);
    }

    /**
     * Realiza o parsing de uma expresso. A entrada pode ser uma String ou
     * um reader
     */
    public JExpression parseExpression(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseExpression(entrada);
    }

    public JExpression parseExpression(String entrada)
            throws ParseException, IllegalArgumentException {

        return parser.parseExpression(entrada);
    }

    /**
     * Realiza o parsing de um parmetro. A entrada pode ser uma String ou
     * um reader
     */
    public JFormalParameter parseParameter(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseParameter(entrada);
    }

    public JFormalParameter parseParameter(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseParameter(entrada);
    }


    /**
     * Realiza o parsing de uma lista de parmetros. A entrada pode ser uma
     * String ou  um reader
     */
    public JParameterList parseParameterList(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseParameterList(entrada);
    }

    public JParameterList parseParameterList(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseParameterList(entrada);

    }


    /**
     * Realiza o parsing de uma lista de modificadores
     */
    public JModifierList parseModifierList(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseModifierList(entrada);
    }

    public JModifierList parseModifierList(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseModifierList(entrada);
    }

    /**
     *  Realiza o parsing de uma lista de argumentos
     */
    public JExpressionList parseArgumentList(Reader entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseArgumentList(entrada);
    }

    public JExpressionList parseArgumentList(String entrada)
            throws ParseException, IllegalArgumentException {
        return parser.parseArgumentList(entrada);
    }

	public JStatementList parseStatementList(Reader strValue)
        throws ParseException, IllegalArgumentException {
            return parser.parseStatementList(strValue);
	}

	public JStatementList parseStatementList(String strValue)
			throws ParseException, IllegalArgumentException {
		return parser.parseStatementList(strValue);
	}



    /**
     * Retorna a tabela de ns que foi construda durante o processo de parsing
     * do arquivo passado como parametro
     */
    public NodeTable getNodeTable(String entrada) throws ParseException,
            IllegalArgumentException {

        this.parse(entrada);
        return this.parser.getNodeTable();
    }

    /**
     * Retorna a tabela de ns que foi construda durante o processo de parsing
     * do arquivo passado como parametro
     */
    public NodeTable getNodeTable(Reader entrada) throws ParseException,
            IllegalArgumentException {
        this.parse(entrada);
        this.nodeTable = this.parser.getNodeTable();
        return this.nodeTable;
    }


    /**
     * Retorna a tabela de ns, referente ao ltimo arquivo que foi analisado.
     * Ou seja, do ltimo que sofreu parsing
     */
    public NodeTable getNodeTable() {
        return this.nodeTable;
    }


    /**
     * This method parses the String <code>input</code> into appropriate
     * INode. The choice of the correct parsing method is based on
     * the variable's type code. All nodes in JaTS have a integer code that
     * identifies the type of the correspondent class, instanceof test
     * is not required.
     *
     * Constraint: Only declarations and some others simple types (like
     * JIdentifier, JName, etc) are allowed.
     *
     * @param  code     Variable JType
     * @param  input    String to be parsed
     */
    public static INode parseString(int code, String input) throws ParseException {
        INode retorno = null;
        JatsIO jatsIO = JatsIO.getInstance();
        switch (code) {
            case JaTSNode.IDENTIFIER:
                retorno = jatsIO.parseIdentifier(input);
                break;
            case JaTSNode.IMPORT_DECLARATION:
                retorno = jatsIO.parseImport(input);
                break;
            case JaTSNode.FIELD_DECLARATION:
                retorno = jatsIO.parseField(input);
                break;
            case JaTSNode.CONSTRUCTOR_DECLARATION:
                retorno = jatsIO.parseConstructor(input);
                break;
            case JaTSNode.METHOD_DECLARATION:
                retorno = jatsIO.parseMethod(input);
                break;
            case JaTSNode.NAME:
                retorno = jatsIO.parseName(input);
                break;
            case JaTSNode.PACKAGE_DECLARATION:
                retorno = jatsIO.parsePackage(input);
                break;
            case JaTSNode.LITERAL:
                retorno = jatsIO.parseLiteral(input);
                break;
            case JaTSNode.EXPRESSION:
                retorno = jatsIO.parseExpression(input);
                break;
            case JaTSNode.TYPE:
                retorno = jatsIO.parseType(input);
                break;

            case JaTSNode.FORMAL_PARAMETER:
                retorno = jatsIO.parseParameter(input);
                break;
            case JaTSNode.PARAMETER_LIST:
                retorno = jatsIO.parseParameterList(input);
                break;
            case JaTSNode.MODIFIER_LIST:
                retorno = jatsIO.parseModifierList(input);
                break;

            case JaTSNode.COMMENT:
                retorno = new JComment(input);
                break;

            case JaTSNode.EXPRESSION_LIST:
                String newInput = "(" + input + ")";
                retorno = jatsIO.parseArgumentList(newInput);
                break;
                
            case JaTSNode.STATEMENT_LIST:
                retorno = jatsIO.parseStatementList(input);
                break;

            default:
                throw new IllegalArgumentException("Tipo invlido para varivel JaTS " + input);

        }
        return retorno;
    }


    /**
     * Realiza o parsing de uma String representando uma varivel e ento
     * retorna o n correpondente.
     *
     * @param varName String correspondente ao nome da varivel
     * @param type tipo da varivel
     *
     * @return Um n do tipo especificado no parametro, setado como varivel e
     * com o nome passado como parmetro.
     */
    public static INode parseVariable(String varName, int type) {
        // na realidade, no  feito o parsing propriamente dito. Apenas 
        // criada a varivel correspondente
        INode retorno = null;
        if (Util.checkString(varName)) {
            switch (type) {
	            case JaTSNode.BLOCK:
					retorno = new JBlock();
					retorno.setVariableType(JaTSNode.BLOCK);
					retorno.setVariableName(varName);
					break;
				case JaTSNode.IDENTIFIER:
					retorno = new JIdentifier();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.IMPORT_DECLARATION:
                    retorno = new JImportDeclaration();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.IMPORT_DECLARATION_SET:
                    retorno = new JImportDeclarationSet();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.FIELD_DECLARATION:
                    retorno = new JFieldDeclaration();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.FIELD_DECLARATION_SET:
                    retorno = new JFieldDeclarationSet();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.CONSTRUCTOR_DECLARATION:
                    retorno = new JConstructorDeclaration();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.CONSTRUCTOR_DECLARATION_SET:
                    retorno = new JConstructorDeclarationSet();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.METHOD_DECLARATION:
                    retorno = new JMethodDeclaration();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.METHOD_DECLARATION_SET:
                    retorno = new JMethodDeclarationSet();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.MODIFIER_LIST:
                    retorno = new JModifierList();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.NAME:
                    retorno = new JName();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.PACKAGE_DECLARATION:
                    retorno = new JPackageDeclaration();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.LITERAL:
                    retorno = new JLiteral();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.TYPE:
                    retorno = new JType();
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.FORMAL_PARAMETER:
                    retorno = new JFormalParameter();
                    retorno.setVariableName(varName);
                    break;

                case JaTSNode.COMMENT:
                    retorno = new JComment();
                    retorno.setVariableType(JaTSNode.COMMENT);
                    retorno.setVariableName(varName);
                    break;

                case JaTSNode.EXPRESSION_LIST:
                    retorno = new JExpressionList();
                    retorno.setVariableType(JaTSNode.EXPRESSION_LIST);
                    retorno.setVariableName(varName);
                    break;

                case JaTSNode.INITIALIZER_SET:
                    retorno = new JInitializerSet();
                    retorno.setVariableType(JaTSNode.INITIALIZER_SET);
                    retorno.setVariableName(varName);
                    break;

                case JaTSNode.MAP_TABLE:
                    retorno = new MapTable();
                    retorno.setVariableType(JaTSNode.MAP_TABLE);
                    retorno.setVariableName(varName);
                    break;
                case JaTSNode.STATEMENT_LIST:
                    retorno = new JStatementList();
                    retorno.setVariableName(varName);
                    break;

                default:
                    System.out.println("varName = " + varName);
                    throw new IllegalArgumentException("Tipo invlido para varivel JaTS " + type);
            }
        } else {
            throw new IllegalArgumentException("Argumento varName nulo em  JatsIO.parseVariable");
        }
        return retorno;
    }

    public  JFieldDeclaration[] getFields(JCompilationUnit compUnit, boolean var) {
        NodeTable table;
        if (compUnit != null) {
            table = compUnit.getTable();
        } else {
            throw new IllegalArgumentException("Argumento compUnit nulo");
        }
        return parser.getFields(table, var);
    }


    public JMethodDeclaration[] getMethods(JCompilationUnit compUnit, boolean var) {
        NodeTable table;
        if (compUnit != null) {
            table = compUnit.getTable();
        } else {
            throw new IllegalArgumentException("Argumento compUnit nulo");
        }
        return parser.getMethods(table, var);
    }

	public INode parseMetaExpression(Reader in) throws ParseException {
		return parser.parseMetaExpression(in);
	}


}
