/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine;

import java.io.Reader;
import java.io.StringReader;
import java.io.FileReader;
import java.io.FileNotFoundException;

import cin.jats.engine.parser.ParseException;
import cin.jats.engine.parser.nodes.*;
import cin.jats.engine.parser.java.*;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.NodeTable;
import cin.jats.engine.util.Util;


/**
 * Esta classe define uma interface de acesso ao parser propriamente dito.
 * Atravs dela  possivel fazer uso dos diversos mtodos de parsing providos
 * pelo parser que o suporta. Alm de mtodos para fazer parsing de um arquivo
 * java, h mtodos para fazer parsing de strings que representam declaraes de
 * mtodo, de construtor, de atributo, etc.
 * H tambm mtodos para extrair informaes da rvore que acabou de ser gerada
 * pelo processo de parsing, por exemplo, os identificadores presentes, ou as
 * declaraoes de mtodo.
 *
 * @author Adeline de Sousa
 * @author Fernando Castor
 *
 * @version  12/08/2002
 * @since JDK 1.3
 */
public class JavaParser implements IParser {


    private cin.jats.engine.parser.java.JavaParser parser;

    private static JavaParser instancia;


    /**
     * Construtor
     */
    private JavaParser() {
        parser = null;
    }

    public synchronized static JavaParser getInstance() {
        if (instancia == null) {
            instancia = new JavaParser();
        }
        return instancia;
    }

    /**
     *  Este mtodo serve para (re)inicializar o parser desta classe.
     *  @param entrada Consiste em um objeto do tipo Reader que contm a entrada
     */
    private void initJavaParser(Reader entrada) throws IllegalArgumentException {

        if (entrada == null) {
            throw new IllegalArgumentException("O programa fornecido como " +
                    "entrada no  vlido.");
        }
        if (parser == null) {
            // Se o parser no foi inicializado ainda
            parser = new cin.jats.engine.parser.java.JavaParser(entrada);
        } else {
        	cin.jats.engine.parser.java.JavaParser.ReInit(entrada);

        }
    }

    /**
     * Realiza o parsing propriamente dito e retorna um n do tipo <code>
     * IType</code>, correspondente  root da rvore sinttica gerada.
     * O argumento passado pode ser o String correspondente ao nome do arquivo
     * a ser aberto ou o Reader correspondente.
     */
    public ICompilationUnit parse(Reader entrada) throws ParseException,
            IllegalArgumentException {

        this.initJavaParser(entrada);
        JCompilationUnit root = cin.jats.engine.parser.java.JavaParser.CompilationUnit();
        return root;
    }

    public ICompilationUnit parse(String entrada) throws ParseException,
            IllegalArgumentException {

        if (!Util.checkString(entrada)) {
            throw new IllegalArgumentException();
        }
        Reader entrada2 = new StringReader(entrada);
        return this.parse(entrada2);
    }

    public ICompilationUnit parseFile(String fileName) throws ParseException,
            IllegalArgumentException, FileNotFoundException {

        if (!Util.checkString(fileName)) {
            throw new IllegalArgumentException();
        }
        Reader entrada2 = new FileReader(fileName);
        return this.parse(entrada2);
    }


    /**
     * Realiza o parsing de uma declarao de mtodo. A entrada pode tanto ser
     * uma String quanto um Reader
     */
    public JMethodDeclaration parseMethod(Reader entrada) throws ParseException,
            IllegalArgumentException {

        this.initJavaParser(entrada);
        JMethodDeclaration root = cin.jats.engine.parser.java.JavaParser.MethodDeclaration();
        return root;
    }

    public JMethodDeclaration parseMethod(String entrada) throws ParseException,
            IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseMethod(entrada2);
    }

    /**
     * Realiza o parsing de uma declarao de atributo. A entrada pode tanto ser
     * uma String quanto um Reader
     */
    public JVariableDeclaration parseField(Reader entrada) throws ParseException,
            IllegalArgumentException {

        this.initJavaParser(entrada);
        JVariableDeclaration root = cin.jats.engine.parser.java.JavaParser.FieldDeclaration();
        return root;
    }

    public JVariableDeclaration parseField(String entrada) throws ParseException,
            IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseField(entrada2);
    }

    /**
     * Realiza o parsing de uma declarao de construtor. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JConstructorDeclaration parseConstructor(Reader entrada)
            throws ParseException, IllegalArgumentException {

        this.initJavaParser(entrada);
        JConstructorDeclaration root = cin.jats.engine.parser.java.JavaParser.ConstructorDeclaration();
        return root;
    }

    public JConstructorDeclaration parseConstructor(String entrada)
            throws ParseException, IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseConstructor(entrada2);
    }


    /**
     * Realiza o parsing de uma declarao de import. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JImportDeclaration parseImport(Reader entrada)
            throws ParseException, IllegalArgumentException {

        this.initJavaParser(entrada);
        JImportDeclaration root = cin.jats.engine.parser.java.JavaParser.ImportDeclaration();
        return root;
    }

    public JImportDeclaration parseImport(String entrada)
            throws ParseException, IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseImport(entrada2);
    }


    /**
     * Realiza o parsing de uma declarao de pacote. A entrada pode tanto
     * ser uma String quanto um Reader
     */
    public JPackageDeclaration parsePackage(Reader entrada)
            throws ParseException, IllegalArgumentException {

    	cin.jats.engine.parser.java.JavaParser.ReInit(entrada);
        JPackageDeclaration root = cin.jats.engine.parser.java.JavaParser.PackageDeclaration();
        return root;
    }

    public JPackageDeclaration parsePackage(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parsePackage(entrada2);
    }


    /**
     * Realiza o parsing de um identificador. A entrada pode ser uma String ou
     * um reader
     */
    public JIdentifier parseIdentifier(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JIdentifier id = cin.jats.engine.parser.java.JavaParser.Identifier();
        return id;
    }

    public JIdentifier parseIdentifier(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseIdentifier(entrada2);
    }


    /**
     * Realiza o parsing de um name. A entrada pode ser uma String ou
     * um reader
     */
    public JName parseName(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JName id = cin.jats.engine.parser.java.JavaParser.Name();
        return id;
    }

    public JName parseName(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseName(entrada2);
    }

    /**
     * Realiza o parsing de um literal. A entrada pode ser uma String ou
     * um reader
     */
    public JLiteral parseLiteral(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JLiteral id = (JLiteral) cin.jats.engine.parser.java.JavaParser.Literal();
        return id;
    }

    public JLiteral parseLiteral(String entrada)
            throws ParseException, IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseLiteral(entrada2);

    }


    /**
     * Realiza o parsing de um tipo. A entrada pode ser uma String ou
     * um reader
     */
    public JType parseType(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JType type = cin.jats.engine.parser.java.JavaParser.Type();
        return type;
    }

    public JType parseType(String entrada)
            throws ParseException, IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseType(entrada2);
    }

    public JType parseResultType(Reader entrada) throws ParseException,
			IllegalArgumentException {
		this.initJavaParser(entrada);
		JType type = cin.jats.engine.parser.java.JavaParser.ResultType();
		return type;
	}

	public JType parseResultType(String entrada) throws ParseException,
			IllegalArgumentException {

		Reader entrada2 = new StringReader(entrada);
		return this.parseResultType(entrada2);
	}
    
    /**
     * Realiza o parsing de um tipo. A entrada pode ser uma String ou
     * um reader
     */
    public JBlock parseBlock(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JBlock retorno = cin.jats.engine.parser.java.JavaParser.Block();
        return retorno;
    }

    public JBlock parseBlock(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseBlock(entrada2);

    }

    /**
     * Realiza o parsing de uma expresso. A entrada pode ser uma String ou
     * um reader
     */
    public JExpression parseExpression(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JExpression exp = cin.jats.engine.parser.java.JavaParser.Expression();
        return exp;
    }

    public JExpression parseExpression(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseExpression(entrada2);

    }


    /**
     * Realiza o parsing de um parmetro. A entrada pode ser uma String ou
     * um reader
     */
    public JFormalParameter parseParameter(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JFormalParameter param = cin.jats.engine.parser.java.JavaParser.FormalParameter();
        return param;
    }

    public JFormalParameter parseParameter(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseParameter(entrada2);
    }


    /**
     * Realiza o parsing de uma lista de parmetros. A entrada pode ser uma
     * String ou  um reader
     */
    public JParameterList parseParameterList(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JParameterList param = cin.jats.engine.parser.java.JavaParser.FormalParameters();
        return param;
    }

    public JParameterList parseParameterList(String entrada)
            throws ParseException, IllegalArgumentException {

        Reader entrada2 = new StringReader(entrada);
        return this.parseParameterList(entrada2);

    }

    /**
     * Realiza o parsing de uma lista de modificadores
     * 
     * 
     * VER SE EH NESSESSARIO NO PARSER JAVA MESMO!!!!
     */
    public JModifierList parseModifierList(Reader entrada)
            throws ParseException, IllegalArgumentException {
//        this.initJavaParser(entrada);
//        JModifierList modList = cin.jats.engine.parser.java.JavaParser.Modifiers();
//        return modList;
        return null;
    }

    public JModifierList parseModifierList(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseModifierList(entrada2);
    }

    /**
     *  Realiza o parsing de uma lista de argumentos
     */
    public JExpressionList parseArgumentList(Reader entrada)
            throws ParseException, IllegalArgumentException {
        this.initJavaParser(entrada);
        JExpressionList param = cin.jats.engine.parser.java.JavaParser.Arguments();
        return param;
    }

    public JExpressionList parseArgumentList(String entrada)
            throws ParseException, IllegalArgumentException {
        Reader entrada2 = new StringReader(entrada);
        return this.parseArgumentList(entrada2);

    }




    /**
     * Retorna a tabela de ns que foi construda durante o processo de parsing
     * do arquivo fonte
     */
    public NodeTable getNodeTable() {
        return cin.jats.engine.parser.java.JavaParser.getNodeTable();
    }


    /**
     * Reinicializa a tabela
     */
    public void reInitTable() {
        cin.jats.engine.parser.java.JavaParser.reInitTable();
    }

    /**
     * Retorna um array contendo apenas os ns do tipo especificado pelo
     * parmetro <code>typeIdentifier</code>. Os ns retornados ou so todos
     * variveis ou no. Para tanto, basta especificar o valor do parmetro
     * <code>var</code>.
     * Os valores possveis para o parmetro typeIdentifier so os valores
     * possveis para o atributo typeIdentifier de cada n. Estes esto listados
     * na interface INode
     *
     * @param  typeIdentifier Tipo dos ns a serem retornados
     * @param  var  true, para variveis <br>
     *              false, caso contrrio
     *
     * @see cin.jats.engine.parser.nodes.INode
     */
    private static INode[] getNodes(int typeIdentifier, boolean var, NodeTable table) {

        INode[] retorno;
        if (table != null) {
            NodeList nodes = null;
            retorno = null;
            if (var) {
                nodes = table.getNodeVarLine(typeIdentifier);
            } else {
                nodes = table.getNodeLine(typeIdentifier);
            }
            retorno = getNodeArray(nodes);
        } else {
            throw new IllegalArgumentException("Argumento table nulo em " +
                    "JavaParser.getNodes");
        }
        return retorno;
    }


    public JFieldDeclaration[] getFields(NodeTable table, boolean var) {
        JFieldDeclaration[] retorno = null;
        INode[] tmpRetorno = getNodes(JaTSNode.FIELD_DECLARATION, var, table);
        if (tmpRetorno != null) {
            retorno = new JFieldDeclaration[tmpRetorno.length];
            for (int i = 0; i < tmpRetorno.length; i++) {
                // o instanceof no  necessrio, pois em tmpRetorno s existem
                // ns do tipo JFieldDeclaration, devido ao parmetro passado.
                retorno[i] = (JFieldDeclaration) tmpRetorno[i];
            }
        } else {
            throw new IllegalArgumentException("Argumento interno nulo em JavaParser.getFields");
        }

        return retorno;
    }




    public  JMethodDeclaration[] getMethods(NodeTable table, boolean var) {
        JMethodDeclaration[] retorno = null;
        INode[] tmpRetorno = getNodes(JaTSNode.METHOD_DECLARATION, var, table);
        if (tmpRetorno != null) {
            retorno = new JMethodDeclaration[tmpRetorno.length];
            for (int i = 0; i < tmpRetorno.length; i++) {
                // o instanceof no  necessrio, pois em tmpRetorno s existem
                // ns do tipo JMethodDeclaration, devido ao parmetro passado.
                retorno[i] = (JMethodDeclaration) tmpRetorno[i];
            }
        } else {
            throw new IllegalArgumentException("Argumento interno nulo em JavaParser.getMethods");
        }

        return retorno;
    }


    /**
     * Converte um objeto do tipo NodeList em um array de ns.
     *
     * @param nodes NodeList a ser convertido em array
     *
     * @return Array contendo os objetos presentes na NodeList passada como
     *         parametro
     */
    private static INode[] getNodeArray(NodeList nodes) {
        INode[] retorno;
        int size = nodes.size();
        retorno = new INode[size];
        INode tmpNode = null;
        for (int i = 0; i < size; i++) {
            tmpNode = nodes.elementAt(i);
            retorno[i] = tmpNode;
        }
        return retorno;
    }

	public JStatementList parseStatementList(Reader sr) {
		throw new RuntimeException("parseStatementList should not be called in java parser");
	}
	
	public JStatementList parseStatementList(String str) throws ParseException {
		throw new RuntimeException("parseStatementList should not be called in java parser");
	}

	public INode parseMetaExpression(Reader in) throws ParseException {
		throw new RuntimeException("parseMetaExpression should not be called in java parser");
	}
	

} // fim da classe
