/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.JatsClonner;
import cin.jats.engine.JatsIO;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;

import java.io.StringWriter;


/**
 * Esta classe implementa a parte comum a todos os ns JaTS. Tais como tratamento
 * de variveis, declaraes executveis  e etc.
 *
 *
 * @author Fernando Castor
 * @version  14/01/2003
 * @since JDK 1.3
 * @see JaTSNode
 */
public abstract class AbstractNode implements JaTSNode {


    // Constante que define o caractere que identifica um identificador
    // qualquer como uma varivel.
    public static final char VARIABLECHAR = '#';

    // Armazena o flag que determina se este n  ou no uma varivel.
    private boolean variable;

    // Armazena um String correspondente ao nome da varivel que este n
    // representa.
    private String variableName;

    // Armazena o tipo de varivel ao qual este n corresponde, caso ele seja
    //  uma varivel.
    private int variableType;

    // Armazena o flag que determina se o n atual  executvel ou no, ou seja,
    //  se existe um n contendo uma expresso executvel associada a este n.
    private boolean executable;

    // Armazena a declarao executvel associada a este n, caso ele seja
    // executvel.
    private INode executableDeclaration;

    // Diz se esse n  opcional ou no no que concerne a casamento e
    //  substituio.
    private boolean optional;

    // Guarda o tipo deste n
    private int typeIdentifier;

    // Seta o nvel de aninhamento - essa informao ser utilizada para fazer
    // o pretty-printing do objeto
//    private int nestingLevel;

    // diz a linha em que o elemento se encontra dentro do arquivo
    private int line;

    // armazena a declarao "or", associada a este n.
    private INode orDeclaration;


    //this node's parent
    private INode parent;
    
    
    /**
     * Construtor vazio da classe
     */
    protected AbstractNode() {
        this.variable = false;
//        this.nestingLevel = 0;
        this.variableName = "";
        this.variableType = 0;
        this.executable = false;
        this.executableDeclaration = null;
        this.optional = false;
        this.setTypeIdentifier(NO_TYPE);
    }


    /**
     * Construtor da classe que recebe os parmetros necessrios  gerao
     * de um n consistente.
     *
     * @param varBool Um boolean que determina se o n corresponde ou no a
     *                uma varivel.
     * @param varName Um String correspondendo ao nome da varivel que este
     *                n representa (caso ele seja uma varivel).
     * @param varType Um inteiro correspondendo ao tipo de varivel ao qual
     *                este n corresponde.
     * @param opt     Um boolean que diz se um n  ou no executvel.
     */
    protected AbstractNode(boolean varBool, String varName, int varType,
                      boolean varExec, JExecutableDeclaration execDec,
                      boolean opt) {
        this();
        this.variable = varBool;
        this.variableName = varName;
        this.variableType = varType;
        this.executable = varExec;
        this.executableDeclaration = execDec;
        this.optional = opt;
        this.setTypeIdentifier(NO_TYPE);
    }


    /**
     * Retorna o nvel de indentao desse n, para que este possa ser usado
     * pelo prettyprinter no momento da impresso do n
     *
     * @return o nvel de aninhamento desse n
     */
//    public int getIndentLevel() {
//        return nestingLevel;
//    }

    /**
     * Seta o nvel de indentao desse n, para que este possa ser usado
     * pelo prettyprinter no momento da impresso do n
     *
     * @param nLevel o nvel de aninhamento desse n
     */
//    public void setIndentLevel(int nLevel) {
//        this.nestingLevel = nLevel;
//    }


    public int getLine() {
        return line;
    }

    public void setLine(int line) {
        this.line = line;
    }

    /**
     * Faz com que os valores dos atributos deste n, no que concerne a
     * variveis, fiquem iguais aos do n fornecido como argumento.
     *
     * @param node O n a ser copiado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          argumento <code>node</code> tiver valor <code>null</code>.
     */
    public void cloneVariable(INode node) throws IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        this.setVariable(node.isVariable());
        this.setVariableName(node.getVariableName());
        this.setVariableType(node.getVariableType());
        this.setExecutable(node.isExecutable());
        this.setOptional(node.isOptional());


    }

    /**
     * Diz se no se trata de um n varivel ou n executvel. (Talvez eu deva
     * mudar seu nome, pois mesmo o n no sendo, ele mesmo, um n executvel,
     * ele pode conter estruturas executveis em um de seus filhos, o que o
     * descaracterizaria como n pertencente a um programa fonte)
     *
     * @return true, se o n no for nem executvel nem varivel
     *         false, em caso contrrio
     */
    public boolean isASourceNode() {
        return (!this.isVariable() && !this.isExecutable() &&
                !this.isOrDeclaration());
    }

    /**
     * Compara dois ns JaTS, que correspondam a variveis ou expresses
     * executveis e diz se estes so iguais. Se os dois ns forem apenas ns
     * Java, o retorno  sempre <code>false</code>.
     *
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica que
     *         o valor do objeto com o qual <code>this</code> est sendo
     *         comparado  igual ao de <code>this</code>.
     */
    public boolean compareJaTSNode(Object objeto) {
        boolean retorno = false;
        if (objeto instanceof INode) {
            JaTSNode node = (JaTSNode) objeto;
            // se os dois ns forem variveis, compara os nomes e os tipos das
            // variveis
            if (this.isVariable() && node.isVariable()) {
                retorno = (this.getVariableType() == node.getVariableType()) &&
                        AbstractNode.equals(this.getVariableName(), node.getVariableName());
            }
            // se os dois forem executveis, compara as expresses executveis
            // associadas
            else if (this.isExecutable() && node.isExecutable()) {
                try {
                    retorno = AbstractNode.equals(this.getExecutableDeclaration(), node.getExecutableDeclaration());
                } catch (InconsistentNodeException e) {
                    // no faz nada, j que , teoricamente no cai nesse catch,
                    // uma vez que j foi verificado que os ns so executveis
                    // apenas deixa retornar false
                }
            }
            // se os ns so do tipo OrDeclaration, ento, verifica se
            // as declaraes associadas so iguais. Se sim, ento so ns
            // equivalentes
            else if (this.isOrDeclaration() && node.isOrDeclaration()) {
                retorno = AbstractNode.equals(this.getOrDeclaration(), node.getOrDeclaration());
            }

            // MUDAR ESSE COMENTARIO
            // se os dois no forem nem executveis nem variveis retorna false
            // Note que este mtodo serve apenas para comparar ns executveis
            // ou ns variveis. Se for aplicado a um n completo, vai retornar
            //false. Entretanto, se apenas um n for um n Java e o outro for um
            // n JaTS, ento o retorno  false, como era de se esperar.
        }

        return retorno;
    }


    /**
     * Retorna a declarao executvel correspondente ao n atual, caso ele seja
     * executvel.
     *
     * @return Um n do tipo <code>JExecutableDeclaration</code>. Esse n
     *         encapsula cdigo Java a ser executado durante o processamento dos
     *         ns.
     * @throws InconsistentNodeException Levantada caso o n seja executvel
     *                                   mas no possua um n executvel associado.
     */
    public INode getExecutableDeclaration() throws
            InconsistentNodeException {

        INode retorno;

        if (this.isExecutable() && this.executableDeclaration == null) {
            throw new InconsistentNodeException(this);
        }
        retorno = this.executableDeclaration;
        return retorno;
    }

    /**
     * Retorna a declarao
     *
     * @return
     */
    public INode getOrDeclaration() {
        return orDeclaration;
    }


    /**
     * Seta a declarao
     *
     * @param orDeclaration
     */
    public void setOrDeclaration(INode orDeclaration) {
        this.orDeclaration = orDeclaration;
    }


    public boolean isOrDeclaration() {
        return this.orDeclaration != null;
    }

    /**
     * Retorna o nome da varivel que este n representa. Se o n no for uma
     * varivel, o mtodo pode retornar <code>null</code>.
     *
     * @return Um String correspondendo ao nome da varivel que este n
     *         representa.
     */
    public String getVariableName() {
        return this.variableName;
    }


    /**
     * Retorna o tipo da varivel.
     *
     * @return Um inteiro correspondendo ao tipo da varivel. O tipo de uma
     *         varivel JaTS corresponde ao tipo de construo sinttica de Java
     *         que a varivel representa. Exemplos possveis so um tipo, uma
     *         lista de nomes, um identificador, uma declarao de mtodo, entre
     *         outros.
     */
    public int getVariableType() {
        return this.variableType;
    }


    /**
     * Retorna um inteiro que representa o tipo do n. Os valores retornados
     * aqui so os das constantes contidas em JaTSNode.
     *
     * @return Um inteiro correspondente ao tipo do n
     */
    public int getTypeIdentifier() {

        return this.typeIdentifier;
    }

    /**
     * Seta o valor do inteiro que representa o tipo do n. Os valores que o
     * tipo de um n podem assumir so os das constantes contidas em JaTSNode.
     *
     * @param ti Um inteiro correspondente ao tipo do n
     */
    public void setTypeIdentifier(int ti) {

        this.typeIdentifier = ti;
        this.setVariableType(ti);

    }

    /**
     * Mtodo responsvel por invocar o mtodo accept em ns do tipo
     * <code>JExecutableDeclaration</code>. Este mtodo  necessrio porque,
     * embora todos os ns possam, potencialmente, ter filhos do tipo
     * JExecutableDeclaration, eles normalmente no tratam com ns desse tipo
     * explicitamente.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    protected void invokeAcceptOnExecutableDeclaration(IVisitor visitor,
                                                       Object data)
            throws IllegalArgumentException, ExecutionException,
            InconsistentNodeException {

        if (visitor == null) {
            throw new IllegalArgumentException();
        }

        if (this.isExecutable()) {
            INode exec = this.getExecutableDeclaration();
            exec.accept(visitor, data);
        }
    }

    /**
     * Diz se um n  ou no executvel onde, por executvel, entende-se que um
     * dos filhos do n  uma expresso sujeita a execuo como se fosse um
     * trecho de um programa em Java.
     *
     * @return Um valor booleano que, caso seja <code>true</code> indica que o
     *         n  executvel.
     */
    public boolean isExecutable() {
        return (this.executableDeclaration != null);
    }


    /**
     * Diz se um no e ou nao opcional.
     *
     * @return Um boolean que diz se o no e ou nao opcional para fins de
     *         casamento e substituicao.
     */
    public boolean isOptional() {
        return this.optional;
    }


    /**
     * Este mtodo diz se um n  ou no uma varivel JaTS.
     *
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica
     *         que o n  uma varivel JaTS.
     */
    public boolean isVariable() {
        return this.variable;
    }


    /**
     * Este mtodo verifica se <code>this</code>  uma varivel e, em caso
     * positivo, se  possvel cas-lo com o n <code>node</code>. Se for, o
     * par  adicionado ao conjunto resultado do casamento e o mtodo retorna
     * <code>true</code>. Caso contrrio, nada  feito e o mtodo simplesmente
     * retorna <code>false</code>.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @param varType O tipo de varivel que se espera que <code>this</code> seja.
     *                Caso <code>this</code> seja uma varivel mas seja de um tipo
     *                diferente de <code>varType</code>, o casamento no pode ser
     *                realizado. Se varType tiver valor JaTSNode.NO_TYPE, o tipo dessa
     *                varivel no importa.
     * @return Um <code>boolean</code> que tem valor </code>true</code> se este
     *         n  uma varivel e pde ser casado com <code>node</code>. <code>false</code>
     *         caso contrrio.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     */
    public boolean matchesAsAVariable(INode node, ResultSet results,
                                         int varType)
            throws NodesNotMatchedException, InconsistentNodeException,
            IllegalArgumentException {

        boolean retorno = false;

        if (node == null || results == null) {
            throw new IllegalArgumentException();
        }

        if (!this.isOrDeclaration() && this.isVariable() && !node.isVariable() &&
                (varType == NO_TYPE
                || this.getTypeIdentifier() == node.getTypeIdentifier()
                || this.getVariableType() == varType
                || Util.compatibleVariableTypes(this.getTypeIdentifier(),
                        node.getTypeIdentifier()))) {

            if (results.alreadyMatched(this)) {
                JaTSNode tmpNode = (JaTSNode) results.get(this);
                node.match(tmpNode, results);
                retorno = true;
            } else {
                results.put(this, node);
                retorno = true;
            }
        }
        // para o caso de ORDeclarations...
        else if (this.isOrDeclaration()) {
            OrDeclaration or = (OrDeclaration) this.getOrDeclaration();
            INode fstNode = or.getChoice1();
            ResultSet tmpRs = (ResultSet) results.clone();
            try {
                fstNode.match(node, tmpRs);
                results.put(this, node);
                retorno = true;
            } catch (NodesNotMatchedException e) {
                INode sndNode = or.getChoice2();
                sndNode.match(node, tmpRs);
                results.put(this, node);
                retorno = true;
            }
        } else if (node.isVariable()) {
            // node corresponde a um n na rvore sinttica de um programa Java e
            // programas Java no podem ter variveis. Caso uma seja encontrada,
            // as duas rvores no casam.
            throw new NodesNotMatchedException("Unfortunately, remained a JaTS " +
                    "variable in the tree. Replacement or process didn't work",
                    this, node);
        }

        return retorno;
    }


    /**
     * Caso o n a partir do qual este mtodo tenha sido invocado seja uma
     * varivel, este mtodo devolve um <code>String</code> correspondendo ao
     * nome da varivel que o n representa seguido pelo tipo da varivel, caso
     * este tenha sido declarado. Caso contrrio,  retornado um
     * <code>String</code> vazio.
     *
     * @return Um <code>String</code> consistindo do nome da varivel. Caso o
     *         tipo dessa varivel tenha sido declarado tambm, o nome dela 
     *         seguido por ":" e por um <code>String</code> correspondendo a esse
     *         tipo.
     */
    public String printVariable() {
        String retorno = "";
        if (this.isVariable()) {
            retorno = this.getVariableName();
            // isso  pra no imprimir com o Tipo, no caso de
            // variveis de tipos simples
            if (this.getVariableType() != NO_TYPE &&
                    this.getVariableType() != JaTSNode.IDENTIFIER &&
                    this.getVariableType() != JaTSNode.NAME &&
                    this.getVariableType() != JaTSNode.TYPE &&
                    this.getVariableType() != JaTSNode.MAP_TABLE) {
                retorno = this.printVariableType() + ":" + retorno;
            }
        }

        return retorno;
    }

    /**
     * Caso o n a partir do qual este mtodo tenha sido invocado seja
     * executvel, este mtodo devolve uma <code>String</code> correspondendo a
     * expresso executvel associada ao n.
     *
     * @return String correspondente a Expresso executvel associada ao n.
     *         Caso o n no seja executvel, retorna a String vazia
     */
    public String printExecutableDeclaration() {
        String retorno = "";
        if (this.isExecutable()) {
            try {
                retorno = this.getExecutableDeclaration().toString();
            } catch (InconsistentNodeException e) {

            }
        }

        return retorno;
    }

    /**
     * Imprime o String correspondente ao tipo desse n, caso ele seja uma
     * varivel declarada como tendo um tipo definido (!= NO_TYPE).
     */
    private String printVariableType() {
        String retorno = "";
        if (this.isVariable() && this.getVariableType() != NO_TYPE) {

            if (this.getVariableType() == JaTSNode.BLOCK) {
                retorno = "Block";
            } else if (this.getVariableType() == JaTSNode.CLASS_DECLARATION) {
                retorno = "ClassDeclaration";
            } else if (this.getVariableType() == JaTSNode.CONSTRUCTOR_DECLARATION) {
                retorno = "ConstructorDeclaration";
            } else if (this.getVariableType() ==
                    JaTSNode.CONSTRUCTOR_DECLARATION_SET) {
                retorno = "ConstructorDeclarationSet";
            } else if (this.getVariableType() == JaTSNode.FIELD_DECLARATION) {
                retorno = "FieldDeclaration";
            } else if (this.getVariableType() == JaTSNode.FIELD_DECLARATION_SET) {
                retorno = "FieldDeclarationSet";
            } else if (this.getVariableType() == JaTSNode.IDENTIFIER) {
                retorno = "Identifier";
            } else if (this.getVariableType() == JaTSNode.IMPORT_DECLARATION) {
                retorno = "ImportDeclaration";
            } else if (this.getVariableType() ==
                    JaTSNode.IMPORT_DECLARATION_SET) {
                retorno = "ImportDeclarationSet";
            } else if (this.getVariableType() == JaTSNode.INITIALIZER) {
                retorno = "Initializer";
            } else if (this.getVariableType() == JaTSNode.INITIALIZER_SET) {
                retorno = "InitializerSet";
            } else if (this.getVariableType() == JaTSNode.INTERFACE_DECLARATION) {
                retorno = "InterfaceDeclaration";
            } else if (this.getVariableType() == JaTSNode.LITERAL) {
                retorno = "Literal";
            } else if (this.getVariableType() == JaTSNode.METHOD_DECLARATION) {
                retorno = "MethodDeclaration";
            } else if (this.getVariableType() == JaTSNode.METHOD_DECLARATION_SET) {
                retorno = "MethodDeclarationSet";
            } else if (this.getVariableType() == JaTSNode.MODIFIER_LIST) {
                retorno = "ModifierList";
            } else if (this.getVariableType() == JaTSNode.NAME) {
                retorno = "Name";
            } else if (this.getVariableType() == JaTSNode.NAME_LIST) {
                retorno = "NameList";
            } else if (this.getVariableType() == JaTSNode.PACKAGE_DECLARATION) {
                retorno = "PackageDeclaration";
            } else if (this.getVariableType() == JaTSNode.FORMAL_PARAMETER) {
                retorno = "Parameter";
            } else if (this.getVariableType() == JaTSNode.PARAMETER_LIST) {
                retorno = "ParameterList";
            } else if (this.getVariableType() == JaTSNode.STATEMENT) {
                retorno = "Statement";
            } else if (this.getVariableType() == JaTSNode.TYPE) {
                retorno = "Type";
            } else if (this.getVariableType() == JaTSNode.VARIABLE_DECLARATOR) {
                retorno = "VariableDeclarator";
            } else if (this.getVariableType() == JaTSNode.MAP_TABLE) {
                retorno = "MapTable";
            } else if (this.getVariableType() == JaTSNode.PROPERTY_DECLARATION_SET) {
                retorno = "PropertyDeclarationSet";
            } else if (this.getVariableType() == JaTSNode.INDEXER_DECLARATION_SET) {
                retorno = "IndexerDeclarationSet";
            }
        }

        return retorno;
    }


    /**
     * Imprime o String correspondente ao tipo passado como parmetro
     */
    public static String printType(int type) {
        String retorno;

        switch (type) {
            case JaTSNode.BLOCK:
                retorno = "Block";
                break;
            case JaTSNode.CONSTRUCTOR_DECLARATION:
                retorno = "ConstructorDeclaration";
                break;
            case JaTSNode.CONSTRUCTOR_DECLARATION_SET:
                retorno = "ConstructorDeclarationSet";
                break;
            case JaTSNode.EXPRESSION:
                retorno = "Expression";
                break;
            case JaTSNode.FIELD_DECLARATION:
                retorno = "FieldDeclaration";
                break;
            case JaTSNode.FIELD_DECLARATION_SET:
                retorno = "FieldDeclarationSet";
                break;
            case JaTSNode.FORMAL_PARAMETER:
                retorno = "Parameter";
                break;
            case JaTSNode.IDENTIFIER:
                retorno = "Identifier";
                break;
            case JaTSNode.IMPORT_DECLARATION:
                retorno = "ImportDeclaration";
                break;
            case JaTSNode.IMPORT_DECLARATION_SET:
                retorno = "ImportDeclarationSet";
                break;
            case JaTSNode.LITERAL:
                retorno = "Literal";
                break;
            case JaTSNode.METHOD_DECLARATION:
                retorno = "MethodDeclaration";
                break;
            case JaTSNode.METHOD_DECLARATION_SET:
                retorno = "MethodDeclarationSet";
                break;
            case JaTSNode.MODIFIER_LIST:
                retorno = "ModifierList";
                break;
            case JaTSNode.NAME:
                retorno = "Name";
                break;
            case JaTSNode.NAME_LIST:
                retorno = "NameList";
                break;
            case JaTSNode.PACKAGE_DECLARATION:
                retorno = "PackageDeclaration";
                break;
            case JaTSNode.PARAMETER_LIST:
                retorno = "ParameterList";
                break;
            case JaTSNode.STATEMENT:
                retorno = "Statement";
                break;
            case JaTSNode.TYPE:
                retorno = "Type";
                break;
            case JaTSNode.VARIABLE_DECLARATOR:
                retorno = "VariableDeclarator";
                break;
            default:
                retorno = "NoType";
        }
        return retorno;
    }


    public static int getTypeCode(String str) {
        int retorno = 0;

        if (Util.checkString(str)) {

            if (str.equalsIgnoreCase("PackageDeclaration")) {
                retorno = JaTSNode.PACKAGE_DECLARATION;
            }
            if (str.equalsIgnoreCase("ImportDeclaration")) {
                retorno = JaTSNode.IMPORT_DECLARATION;
            } else if (str.equalsIgnoreCase("ImportDeclarationSet")) {
                retorno = JaTSNode.IMPORT_DECLARATION_SET;
            } else if (str.equalsIgnoreCase("FieldDeclaration")) {
                retorno = JaTSNode.FIELD_DECLARATION;
            } else if (str.equalsIgnoreCase("FieldDeclarationSet")) {
                retorno = JaTSNode.FIELD_DECLARATION_SET;
            } else if (str.equalsIgnoreCase("ConstructorDeclaration")) {
                retorno = JaTSNode.CONSTRUCTOR_DECLARATION;
            } else if (str.equalsIgnoreCase("ConstructorDeclarationSet")) {
                retorno = JaTSNode.CONSTRUCTOR_DECLARATION_SET;
            } else if (str.equalsIgnoreCase("MethodDeclaration")) {
                retorno = JaTSNode.METHOD_DECLARATION;
            } else if (str.equalsIgnoreCase("MethodDeclarationSet")) {
                retorno = JaTSNode.METHOD_DECLARATION_SET;
            } else if (str.equalsIgnoreCase("Identifier")) {
                retorno = JaTSNode.IDENTIFIER;
            } else if (str.equalsIgnoreCase("Name")) {
                retorno = JaTSNode.NAME;
            } else if (str.equalsIgnoreCase("NameList")) {
                retorno = JaTSNode.NAME_LIST;
            } else if (str.equalsIgnoreCase("Parameter")) {
                retorno = JaTSNode.FORMAL_PARAMETER;
            } else if (str.equalsIgnoreCase("ParameterList")) {
                retorno = JaTSNode.PARAMETER_LIST;
            } else if (str.equalsIgnoreCase("ModifierList")) {
                retorno = JaTSNode.MODIFIER_LIST;
            } else if (str.equalsIgnoreCase("Expression")) {
                retorno = JaTSNode.EXPRESSION;
            } else if (str.equalsIgnoreCase("Literal")) {
                retorno = JaTSNode.LITERAL;
            } else if (str.equalsIgnoreCase("Block")) {
                retorno = JaTSNode.BLOCK;
            } else if (str.equalsIgnoreCase("Type")) {
                retorno = JaTSNode.TYPE;
            } else if (str.equalsIgnoreCase("Statement")) {
                retorno = JaTSNode.STATEMENT;
            } else if (str.equalsIgnoreCase("JExpressionList")) {
                retorno = JaTSNode.EXPRESSION_LIST;
            }
        } else {
            retorno = NO_TYPE;
        }


        return retorno;
    }


    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {
        return this;
    }


    /**
     * Outro mtodo auxiliar para diminuir a repetio de cdigo que ocorre nos
     * mtodos process. Este mtodo  responsvel por realizar o processamento
     * deste n apenas quando ele  executvel e s deve ser invocado nessas
     * circunstncias.
     */
    protected Object processExecutableNode(Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (!this.isExecutable()) {
            throw new ExecutionException(this);
        }
        Object retorno;
        INode exec = this.getExecutableDeclaration();
        retorno = exec.process(data);

        return retorno;
    }


    protected Object evaluateExecutableNode(Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (!this.isExecutable()) {
            throw new ExecutionException(this);
        }
        Object retorno;
        JExecutableDeclaration exec = (JExecutableDeclaration)
                this.getExecutableDeclaration();
        retorno = exec.process(data);

        return retorno;
    }

    /**
     * Mtodo auxiliar criado para evitar uma parte da repetio de cdigo
     * nos mtodos process.
     */
    protected Object processNode(INode node, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new InconsistentNodeException("Node recebido  nulo", this);
        }
        Object ret = node.process(data);
        if (node.isExecutable() && ret == null) {
            throw new InconsistentNodeException(this);
        }

        return ret;
    }


    /**
     * Determina que um n  executvel ou no.
     *
     * @param exec Um boolean que, caso tenho tenha o valor <code>true</code>,
     *             determina que o n executvel.
     */
    public void setExecutable(boolean exec) {
        this.executable = exec;
        if (!this.executable) {
            this.executableDeclaration = null;
        }
    }


    /**
     * Seta o n executvel associado a este n. Se o n no era executvel antes
     * da invocao deste mtodo (ou seja, se o flag <code>executable</code>
     * tinha valor <code>false</code>), o mtodo torna o valor desse flag
     * <code>true</code>.
     *
     * @param dec A expresso executvel a ser associada a este n.
     */
    public void setExecutableDeclaration(INode dec) {
        this.executableDeclaration = dec;
        if (dec == null) {
            this.setExecutable(false);
        } else {
            this.setExecutable(true);
        }
    }


    /**
     * Faz de um no opcional, para fins de casamento e substituicao.
     *
     * @param opt Um boolean que diz se o no e opcional ou nao.
     */
    public void setOptional(boolean opt) {
        this.optional = opt;
    }


    /**
     * Este mtodo determina se um n  ou no uma varivel.
     *
     * @param tf Um boolean que, caso tenha o valor <code>true</code>, determina
     *           que o n  uma varivel.
     */
    public void setVariable(boolean tf) {
        this.variable = tf;
    }


    /**
     * Seta o nome da varivel que este n representa.
     *
     * @param varName Um String, correspondente ao nome da varivel.
     */
    public void setVariableName(String varName) {
        if (Util.isAValidVariableName(varName)) {
            this.setVariable(true);
            this.variableName = varName;
        } else if (!Util.checkString(this.variableName)) {
            this.setVariable(false);
        }
    }


    /**
     * Seta o tipo da varivel.
     *
     * @param type Um inteiro correspondendo a um dos possveis tipos de
     *             variveis. Os tipos correspondem s construes sintticas de
     *             Java, embora algumas construes no estejam diretamente
     *             representadas entre os tipos possveis (ou seja, no  possvel
     *             que uma varivel tenha o tipo correspondendo a qualquer construo
     *             Java, embora isso seja possvel com a maioria dos tipos).
     */
    public void setVariableType(int type) {
        this.variableType = type;
    }

    /**
     * Este metodo faz o mesmo papel do toString. No entanto ele lana excecoes
     * e no permite a impresso de tipos JaTS. Portanto, deve ser usado para
     * impresso da rvore Java.
     *
     * @return
     * @throws InconsistentNodeException
     * @throws ExecutionException
     */
    public String print() throws InconsistentNodeException, ExecutionException {
        JatsIO io = JatsIO.getInstance();
        StringWriter sw = new StringWriter();
        String retorno = "";

        try {

            if (this.isASourceNode()) {
                // imprime com tipos JaTS desabilitados
                io.prettyPrint(this, sw, false);
                retorno = sw.toString();
            } else {
                throw new InconsistentNodeException(this,
                        InconsistentNodeException.INVALID_CONSTRUCTION);
            }

        } catch (java.io.IOException ioe) {
            throw new RuntimeException("IOException");
        }
        return retorno;
    }

    /**
     * Retorna a String que representa o n
     *
     * @return
     */
    public String toString() {

        JatsIO io = JatsIO.getInstance();
        StringWriter sw = new StringWriter();
        String retorno = "";
        try {
            if (this.isASourceNode()) {
                io.prettyPrint(this, sw, true);
                retorno = sw.toString();
            } else if (this.isVariable()) {
                retorno = printVariable();
            } else if (this.isExecutable()) {
                retorno = printExecutableDeclaration();
            } else if (this.isOrDeclaration()) {
                retorno = printOrDeclaration();

            }
        } catch (ExecutionException ee) {
            throw new RuntimeException("ExecutionException");
        } catch (cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException ine) {
            throw new RuntimeException("Inconsistent node\n" + ine.getNode().getTypeIdentifier());
        } catch (java.io.IOException ioe) {
            throw new RuntimeException("IOException");
        }
        return retorno;
    }

    public String printOrDeclaration() {
        String retorno = "";
        if (this.isOrDeclaration()) {
            OrDeclaration or = (OrDeclaration) this.getOrDeclaration();
            // nenhuma das opes pode ser nula, seno, trata-se de um n
            // inconsistente!
            String choice1 = or.getChoice1().toString();
            String choice2 = or.getChoice2().toString();
            retorno = "(" + choice1 + " | " + choice2 + ")";

        }
        return retorno;
    }

    /**
     * Verifica se o numero passado como parametro  uma constante JaTSNode
     * vlida
     *
     * @param num Numero que se deseja verificar a validade
     * @return true, caso seja uma constante vlida,
     *         false caso contrrio.
     */
    public static boolean isValidType(int num) {

        boolean retorno = false;

        if ((num == NO_TYPE) ||
                (num == JaTSNode.CONSTRUCTOR_DECLARATION) ||
                (num == JaTSNode.CONSTRUCTOR_DECLARATION_SET) ||
                (num == JaTSNode.FIELD_DECLARATION) ||
                (num == JaTSNode.FIELD_DECLARATION_SET) ||
                (num == JaTSNode.METHOD_DECLARATION) ||
                (num == JaTSNode.METHOD_DECLARATION_SET) ||
                (num == JaTSNode.NAME) ||
                (num == JaTSNode.NAME_LIST) ||
                (num == JaTSNode.MODIFIER_LIST) ||
                (num == JaTSNode.TYPE) ||
                (num == JaTSNode.IDENTIFIER) ||
                (num == JaTSNode.EXPRESSION) ||
                (num == JaTSNode.VARIABLE_DECLARATOR) ||
                (num == JaTSNode.FORMAL_PARAMETER) ||
                (num == JaTSNode.PARAMETER_LIST) ||
                (num == JaTSNode.IMPORT_DECLARATION) ||
                (num == JaTSNode.IMPORT_DECLARATION_SET) ||
                (num == JaTSNode.PACKAGE_DECLARATION) ||
                (num == JaTSNode.TYPE_DECLARATION) ||
                (num == JaTSNode.INTERFACE_DECLARATION) ||
                (num == JaTSNode.BLOCK) ||
                (num == JaTSNode.EXPRESSION_LIST) ||
                (num == JaTSNode.LITERAL) ||
                (num == JaTSNode.STATEMENT) ||
                (num == JaTSNode.INITIALIZER) ||
                (num == JaTSNode.INITIALIZER_SET) ||
                (num == JaTSNode.MAP_TABLE) ||
                (num == JaTSNode.PROPERTY_DECLARATION) ||
                (num == JaTSNode.PROPERTY_DECLARATION_SET) ||
                (num == JaTSNode.INDEXER_DECLARATION) ||
                (num == JaTSNode.INDEXER_DECLARATION_SET) ||
				(num == JaTSNode.ENUM_DECLARATION) ||
				(num == JaTSNode.USING_DECLARATION) ||
				(num == JaTSNode.ATTRIBUTE_SECTION_LIST) ||
//				(num == JaTSNode.STRUCT_DECLARATION) ||
//				(num == JaTSNode.STRUCT_DECLARATION_SET) ||
//				(num == JaTSNode.DELEGATE_DECLARATION) ||
//				(num == JaTSNode.DELEGATE_DECLARATION_SET) ||
//				(num == JaTSNode.NAMESPACE_MEMBER_DECLARATION) ||
//				(num == JaTSNode.NAMESPACE_MEMBER_DECLARATION_SET)
				(num == JaTSNode.STATEMENT_LIST)
        ) {

            retorno = true;
        }

        return retorno;
    }


    /**
     * Returns a hash code value for the object. This method is
     * supported for the benefit of hashtables such as those provided by
     * <code>java.util.Hashtable</code>.
     * <p/>
     * The general contract of <code>hashCode</code> is:
     * <ul>
     * <li>Whenever it is invoked on the same object more than once during
     * an execution of a Java application, the <tt>hashCode</tt> method
     * must consistently return the same integer, provided no information
     * used in <tt>equals</tt> comparisons on the object is modified.
     * This integer need not remain consistent from one execution of an
     * application to another execution of the same application.
     * <li>If two objects are equal according to the <tt>equals(Object)</tt>
     * method, then calling the <code>hashCode</code> method on each of
     * the two objects must produce the same integer result.
     * <li>It is <em>not</em> required that if two objects are unequal
     * according to the {@link java.lang.Object#equals(java.lang.Object)}
     * method, then calling the <tt>hashCode</tt> method on each of the
     * two objects must produce distinct integer results.  However, the
     * programmer should be aware that producing distinct integer results
     * for unequal objects may improve the performance of hashtables.
     * </ul>
     * <p/>
     * As much as is reasonably practical, the hashCode method defined by
     * class <tt>Object</tt> does return distinct integers for distinct
     * objects. (This is typically implemented by converting the internal
     * address of the object into an integer, but this implementation
     * technique is not required by the
     * Java<font size="-2"><sup>TM</sup></font> programming language.)
     *
     * @return a hash code value for this object.
     * @see java.lang.Object#equals(java.lang.Object)
     * @see java.util.Hashtable
     */
    public int hashCode() {
        int tipo;
        if (this.isVariable()) {
            tipo = this.variableType;
        } else {
            tipo = this.typeIdentifier;
        }
        return tipo;
    }

    /**
     * Creates and returns a copy of this object.  The precise meaning
     * of "copy" may depend on the class of the object. The general
     * intent is that, for any object <tt>x</tt>, the expression:
     * <blockquote>
     * <pre>
     * x.clone() != x</pre></blockquote>
     * will be true, and that the expression:
     * <blockquote>
     * <pre>
     * x.clone().getClass() == x.getClass()</pre></blockquote>
     * will be <tt>true</tt>, but these are not absolute requirements.
     * While it is typically the case that:
     * <blockquote>
     * <pre>
     * x.clone().equals(x)</pre></blockquote>
     * will be <tt>true</tt>, this is not an absolute requirement.
     * Copying an object will typically entail creating a new instance of
     * its class, but it also may require copying of internal data
     * structures as well.  No constructors are called.
     * <p/>
     * The method <tt>clone</tt> for class <tt>Object</tt> performs a
     * specific cloning operation. First, if the class of this object does
     * not implement the interface <tt>Cloneable</tt>, then a
     * <tt>CloneNotSupportedException</tt> is thrown. Note that all arrays
     * are considered to implement the interface <tt>Cloneable</tT>.
     * Otherwise, this method creates a new instance of the class of this
     * object and initializes all its fields with exactly the contents of
     * the corresponding fields of this object, as if by assignment; the
     * contents of the fields are not themselves cloned. Thus, this method
     * performs a "shallow copy" of this object, not a "deep copy" operation.
     * <p/>
     * The class <tt>Object</tt> does not itself implement the interface
     * <tt>Cloneable</tt>, so calling the <tt>clone</tt> method on an object
     * whose class is <tt>Object</tt> will result in throwing an
     * exception at run time. The <tt>clone</tt> method is implemented by
     * the class <tt>Object</tt> as a convenient, general utility for
     * subclasses that implement the interface <tt>Cloneable</tt>, possibly
     * also overriding the <tt>clone</tt> method, in which case the
     * overriding definition can refer to this utility definition by the
     * call:
     * <blockquote>
     * <pre>
     * super.clone()</pre></blockquote>
     *
     * @return a clone of this instance.
     * @throws OutOfMemoryError if there is not enough memory.
     * @see Cloneable
     */
    public Object clone() {
        INode retorno = null;
        try {
            JatsClonner clonner = new JatsClonner();
            retorno = clonner.clone(this);
        } catch (InconsistentNodeException e) {
            e.printStackTrace();
            throw new IllegalArgumentException();
        } catch (ExecutionException e) {
            throw new IllegalArgumentException();
        }
        return retorno;
    }

    /**
     * Este mtodo verifica se dois objetos so iguais.
     * Os objetos obj1 e obj2 devem ser do mesmo tipo
     *
     * @param obj1 Objeto a ser comparado
     * @param obj2 Objeto a ser comparado
     * @return true, se <code>obj1</code> for igual a <code>obj2</code> <br>
     *         false, caso contrario
     */
    public static boolean equals(Object obj1, Object obj2) {
        boolean retorno = false;
        if (obj1 != null && obj2 != null) {
            retorno = obj1.equals(obj2);
        } else if (obj1 == null && obj2 == null) {
            retorno = true;
        }

        return retorno;
    }

    /**
     * Sets the 'parent' attribute of this node
     * @param parent
     */
    public void setParent(INode parent) {
    	this.parent = parent;
    }
    
    /**
     * Retrieves this node's parent
     * 
     * @return
     */
    public INode getParent() {
    	return this.parent;
    }    

}
