/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

/**
 * Esta classe representa uma declarao de tipo abstrata. A qual pode descender
 * para um tipo Java (classe ou interface, enumerao(jdk1.5)) ou
 * CSharp (os mesmos de java) + structs.
 */
public abstract class AbstractTypeDeclaration extends AbstractNode
        implements Commentable, ITypeDeclaration {

    // guarda o comentario referente ao tipo declarado
    protected JComment comment;
    // modificadores do tipo (opcional)
    protected JModifierList modifiers;
    // nome do tipo
    protected JIdentifier name;

    protected AbstractTypeDeclaration() {
    }

    /**
     * Retorna true, caso o n, possua comment.
     *
     * @return true, se  o n possui um comment
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return (this.comment != null && this.comment.getText() != null);
    }

    /**
     * Retorna o Javadoc associado a esta declarao de Tipo.
     *
     * @return comment associado a esta declarao de Tipo
     */
    public JComment getComment() {
        return this.comment;
    }

    /**
     * Retorna a lista de modificadores do tipo.
     *
     * @return Modificadores do tipo
     */
    public JModifierList getModifiers() {
        return this.modifiers;
    }

    /**
     * Retorna o nome identificador (nome) do tipo.
     *
     * @return Nome do tipo
     */
    public JIdentifier getName() {
        return this.name;

    }

    /**
     * Seta o comment referente a esta declarao de Tipo.
     *
     * @param javadoc Comentrio comment referente a esta declarao de Tipo
     */
    public void setComment(JComment javadoc) {
        if (javadoc != null) {
            this.comment = javadoc;
        }
    }

    /**
     * Seta a lista de modificadores do mtodo.
     *
     * @param modList Lista de modificadores deste mtodo
     * @throws java.lang.IllegalArgumentException
     *          - se o modList passado for null.
     */
    public void setModifiers(JModifierList modList)
            throws IllegalArgumentException {

        if (modList != null) {
            this.modifiers = modList;
        } else {
            throw new IllegalArgumentException("Lista de modificadores nula");
        }
    }

    /**
     * Seta o identificador (nome) do tipo.
     *
     * @param id Nome  classe
     * @throws java.lang.IllegalArgumentException
     *          - se o id passado for null.
     */
    public void setName(JIdentifier id) throws IllegalArgumentException {

        if (id != null) {
            this.name = id;
        } else {
            throw new IllegalArgumentException("Nome do tipo nulo");
        }

    }

    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        JIdentifier result = null;
        if (this.isASourceNode()) {
            result = this.getName();
        }
        return result;
    }

    /**
     * Retorna true se o tipo for publico.
     *
     * @return true, se o tipo for publico,
     *         false, caso contrrio
     */
    public boolean isPublic() {
        return this.modifiers.hasModifier(JModifier.PUBLIC);
    }
}
