/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;

/**
 * Classe que   guarda as informacoes a respeito das declaracoes
 * de  atributo
 * Essa classe  formada basicamente por tres conjuntos : um pra
 * atributos, outro para declaracoes do tipo <b>JFieldDeclaration</b> e outra para
 * declaracoes do tipo <b>JFieldDeclarationSet</b>
 *
 * @author Adeline de Sousa
 * @author Fernando Castor
 */
public class BodyFieldDeclarations extends AbstractNode {

    // atributos do programa java
    private JFieldDeclarationSet fieldsSet;

    // declaracao no programa jats que casa com um unico atributo do programa
    // java
    private NodeList fieldVarSet;

    // declaracao no programa jats que casa com um conjunto de  atributos
    // no do programa java
    private NodeList fieldsVarSet;


    public BodyFieldDeclarations() {
        this.fieldsSet = new JFieldDeclarationSet();
        this.fieldVarSet = new NodeList();
        this.fieldsVarSet = new NodeList();
    }


    /**
     * Adiciona uma declaracao de atributo ao conjunto fieldsSet
     *
     * @param fieldDec Declaracao de atributo a ser adicionada
     */
    public void addFieldDeclaration(JVariableDeclaration fieldDec)
            throws IllegalArgumentException {

        if (fieldDec != null) {
            this.fieldsSet.add(fieldDec);
        } else {
            throw new IllegalArgumentException("fieldDec nulo");
        }
    }

    /**
     * Adiciona uma declaracao 'FieldDeclarationSet' ao conjunto fieldVarSet
     *
     * @param atts Declaracao 'FieldDeclarationSet' a ser adicionada
     */
    public void addFieldsVarDeclaration(JFieldDeclarationSet atts)
            throws IllegalArgumentException {

        if (atts != null && atts.isVariable() &&
                atts.getVariableType() == JaTSNode.FIELD_DECLARATION_SET) {
            this.fieldsVarSet.addElement(atts);
        } else {
            throw new IllegalArgumentException("atts nulo");
        }
    }

    /**
     * Adiciona uma declaracao do tipo "#f:JFieldDeclaration" ao conjunto
     * fieldVarSet
     *
     * @param field Declaracao a ser adicionada
     */
    public void addFieldVarDeclaration(JFieldDeclaration field)
            throws IllegalArgumentException {

        if (field != null && field.isVariable() &&
                field.getVariableType() == JaTSNode.FIELD_DECLARATION) {
            this.fieldVarSet.addElement(field);
        } else {
            throw new IllegalArgumentException("field nulo");
        }
    }

    /**
     * Retorna o conjunto de construcoes do tipo FieldDeclaration do programa jats
     *
     * @return Conjunto de construcoes do tipo FieldDeclaration do programa jats
     */
    public NodeList getFieldVarSet() {
        return this.fieldVarSet;
    }

    /**
     * Retorna o conjunto de construcoes do tipo FieldDeclarationSet do programa jats
     *
     * @return Conjunto de construcoes do tipo FieldDeclarationSet do programa jats
     */
    public NodeList getFieldsVarSet() {
        return this.fieldsVarSet;
    }

    /**
     * Retorna o conjunto de atributos do programa java
     *
     * @return Conjunto de atributos do programa java
     */
    public JFieldDeclarationSet getFieldsSet() {
        return this.fieldsSet;
    }


    /**
     * Seta o conjunto de construcoes do tipo attribute do programa jats
     *
     * @param attSet Construcoes do tipo attribute do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setFieldVarSet(NodeList attSet) {
        if (attSet != null) {
            Iterator it = attSet.iterator();
            while (it.hasNext()) {
                if (!((it.next()) instanceof JVariableDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.fieldVarSet = attSet;
        } else {
            throw new IllegalArgumentException("Conjunto attribute vazio");
        }
    }


    /**
     * Seta o conjunto de construcoes do tipo attributes do programa jats
     *
     * @param fieldsSet Construcoes do tipo attributes do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setFieldsVarSet(NodeList fieldsSet) {

        if (fieldsSet != null) {
            Iterator it = fieldsSet.iterator();
            while (it.hasNext()) {
                if (!((it.next()) instanceof JFieldDeclarationSet)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.fieldsVarSet = fieldsSet;
        } else {

            throw new IllegalArgumentException("Conjunto attributes vazio");
        }

    }

    /**
     * Seta o conjunto de atributos do programa java
     *
     * @param fields Conjunto de atributos do programa java
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setFieldsSet(JFieldDeclarationSet fields) {

        if (fields != null) {
            JaTSIterator jit = fields.getIterator();
            while (jit.hasNext()) {
                if (!(jit.next() instanceof JVariableDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.fieldsSet = fields;
        } else {
            throw new IllegalArgumentException("Conjunto de atributos vazio");
        }

    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.fieldsSet</code>, <code>this.fieldVarSet</code> e
     * <code>this.fieldsVarSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno;

        // Variveis que casam com vrias declaraes de atributos
        if (this.fieldsVarSet != null) {
            int size = fieldsVarSet.size() - 1;

            while (size >= 0) {
                this.fieldsVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Variveis que casam com uma nica declarao de atributo
        if (this.fieldVarSet != null) {
            int size = fieldVarSet.size() - 1;

            while (size >= 0) {
                this.fieldVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Declaraes de atributo
        if (this.fieldsSet != null) {
            this.fieldsSet.accept(visitor, data);
        } else {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;

    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (node == null || !(node instanceof BodyFieldDeclarations)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        BodyFieldDeclarations fieldDecSet = (BodyFieldDeclarations) node;
        JFieldDeclarationSet nodeFieldDecs = (JFieldDeclarationSet) fieldDecSet.getFieldsSet().clone();
        if (nodeFieldDecs == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(fieldDecSet);
        }
        int nodeFieldDecsSize = nodeFieldDecs.size();
        boolean achou = false;

        // Casamento das declaraes do tipo varivel:JFieldDeclaration com os
        // atributos que ainda esto em nodeFieldDecs. Primeiro so casadas as
        // variveis que J FORAM mapeadas a algum valor. Isso  necessrio para
        // garantir que nenhuma declarao no fonte Java  mapeada a mais de uma
        // varivel.
        int k = 0;
        int max3 = this.fieldVarSet.size();

        // Lista de variveis do programa JaTS que ainda no foram casados
        NodeList unmatchedFieldDecVars = new NodeList();

        // Lista de variveis do tipo JFieldDeclarationSet do programa JaTS que
        // ainda no foram casadas.
        NodeList unmatchedFieldDecSetVars = new NodeList();

        // Com essa implementao padro,  necessrio verificar essa con-
        // dio para garantir que nenhuma varivel em this.fieldVarSet
        // ficar sem ser casada.
        if (max3 > nodeFieldDecs.size()) {
            throw new NodesNotMatchedException("There is a different number of " +
                    "fields in each declaration (" + max3 + ", " +
                    nodeFieldDecs.size() + ")", this, fieldDecSet);
        } else {
            while (k < max3) {
                JaTSNode matchesSingleField = (JaTSNode) this.fieldVarSet.elementAt(k);
                if (!matchesSingleField.isVariable() ||
                        matchesSingleField.getVariableType() !=
                        JaTSNode.FIELD_DECLARATION ||
                        !(matchesSingleField instanceof JVariableDeclaration)) {
                    throw new InconsistentNodeException(this);
                }
                // se a varivel j est casada, ento varre a lista de ns do
                // tipo JFieldDeclaration at encontrar o valor que lhe foi
                // mapeado e o remove da lista, para que este no possa ser
                // mapeado a outra varivel
                if (results.alreadyMatched(matchesSingleField)) {
                    int l = 0;
                    nodeFieldDecsSize = nodeFieldDecs.size();
                    achou = false;
                    Object tempNode = results.get(matchesSingleField);

                    if (!(tempNode instanceof JVariableDeclaration)) {
                        throw new IllegalMatchingException(matchesSingleField);
                    }
                    while ((l < nodeFieldDecsSize) && !achou) {
                        // Em nodeFieldDecs s existem ns do tipo
                        // JFieldDeclaration, por isso o cast sem instanceof
                        JFieldDeclaration tmp =
                                (JFieldDeclaration) nodeFieldDecs.elementAt(l);
                        try {
                            // Se encontrou o valor, remove o n da lista de
                            // declaraes de atributo do programa Java
                            tmp.match((JVariableDeclaration) tempNode, results);
                            nodeFieldDecs.removeElementAt(l);
                            achou = true;
                        } catch (NodesNotMatchedException e) {
                            l++;
                        }
                    }
                }
                // a varivel ainda no foi casada, adiciona-a  lista de
                // variveis que ainda no foram casadas
                else {
                    unmatchedFieldDecVars.addElement(matchesSingleField);
                }

                k++;
            }
        }

        // Casamento de variveis do tipo JFieldDeclarationSet. Similar ao
        // anterior: apenas para remover da lista de declaraes de atributo
        // do tipo fonte Java as declaraes que j esto casadas a variveis.
        // Nesse caso, a variveis do tipo JFieldDeclarationSet.
        int fieldsVarSetSize = this.fieldsVarSet.size();
        int n = 0;
        INode tempFieldsVar = null;

        while (n < fieldsVarSetSize) {
            tempFieldsVar = this.fieldsVarSet.elementAt(n);
            if (!(tempFieldsVar instanceof JFieldDeclarationSet)) {
                throw new IllegalMatchingException(tempFieldsVar);
            }
            JFieldDeclarationSet fieldSetVar =
                    (JFieldDeclarationSet) tempFieldsVar;

            if (results.alreadyMatched(fieldSetVar)) {
                Object tempFds = results.get(fieldSetVar);

                if (!(tempFds instanceof JFieldDeclarationSet)) {
                    throw new IllegalMatchingException();
                }

                JFieldDeclarationSet fds = (JFieldDeclarationSet) tempFds;
                JaTSIterator ji = fds.getIterator();

                while (ji.hasNext()) {
                    INode fieldsVarDecSetTemp = ji.next();

                    int c1 = 0;
                    int maxFDS = nodeFieldDecs.size();
                    boolean achou2 = false;

                    while (c1 < maxFDS && !achou2) {
                        JaTSNode fDecTemp = (JaTSNode) nodeFieldDecs.elementAt(c1);

                        try {
                            fieldsVarDecSetTemp.match(fDecTemp, results);
                            nodeFieldDecs.removeElementAt(c1);
                            achou2 = true;
                        } catch (NodesNotMatchedException nnme) {
                            c1++;
                        }
                    }
                }
            } else {
                unmatchedFieldDecSetVars.addElement(fieldSetVar);
            }
            n++;
        }


        // Casamento padro de declaraes de atributo no lado esquerdo da
        // transformao (this) com declaraes de atributo no programa fonte
        // (contidas em nodeFieldDecs).
        this.fieldsSet.match(nodeFieldDecs, results);

        // Casamento das variveis do tipo JFieldDeclaration que ainda no esto
        // mapeadas a nenhum valor.
        int m = 0;
        int max5 = unmatchedFieldDecVars.size();
        if (max5 <= nodeFieldDecs.size()) {
            while (m < max5) {
                JaTSNode tempVar = (JaTSNode) unmatchedFieldDecVars.elementAt(m);
                if (!(tempVar instanceof JVariableDeclaration)) {
                    throw new IllegalMatchingException(tempVar);
                }
                JaTSNode tempDec = (JaTSNode) nodeFieldDecs.elementAt(0);
                // tenta casar os ns restantes
                tempVar.matchesAsAVariable(tempDec, results,
                        JaTSNode.FIELD_DECLARATION);
                nodeFieldDecs.removeElementAt(0);
                m++;
            }
        } else {
            // no pode restar variveis do tipo "FieldDeclaration" sem estar
            // casadas
            throw new NodesNotMatchedException("Some fields couldn't be matched " +
                    "in BodyFieldDeclararions", this, node);
        }


        // Casamento entre as declaraes do tipo varivel:FieldDeclarationSet
        // e os atributos restantes em nodeFieldDecs.
        int fieldsVarSize = unmatchedFieldDecSetVars.size();
        if (fieldsVarSize > 0) {
            // Se houver apenas uma varivel do tipo FieldDeclarationSet, ento
            // esta ser casada com as declaraes de atributo restante
            if (fieldsVarSize == 1) {
                INode matchesTheRest = unmatchedFieldDecSetVars.elementAt(0);

                if (matchesTheRest == null || !matchesTheRest.isVariable() ||
                        matchesTheRest.getVariableType() !=
                        JaTSNode.FIELD_DECLARATION_SET ||
                        !(matchesTheRest instanceof JFieldDeclarationSet))
                    throw new InconsistentNodeException(this);

                JFieldDeclarationSet fieldDecsVar = (JFieldDeclarationSet) matchesTheRest;

                fieldDecsVar.matchesAsAVariable(nodeFieldDecs, results,
                        NO_TYPE);
            } else {
               // Se houver mais que uma varivel do tipo JFieldDeclarationSet
                // no casada, o sistema no  capaz de decidir como proceder e
                // o casamento falha.
                JFieldDeclarationSet unmatchedvarDecs = new JFieldDeclarationSet(unmatchedFieldDecSetVars);
                throw new NodesNotMatchedException("The system could't decide " +
                        "how to perform the matching. (" + fieldsVarSize + " VarSets)",
                        unmatchedvarDecs, nodeFieldDecs);
            }
        } else if (nodeFieldDecs.size() > 0) {
            throw new NodesNotMatchedException("Some fields could not be matched " +
                    "(number of felds: " + nodeFieldDecs.size() + ")", this, fieldDecSet);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * BodyFieldDeclarations se resume a invocar o mtodo <code>process</code> no
     * n correspondente  lista de declaraes de atributos filha do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Ns do tipo BodyFieldDeclarations nunca podem ser executveis.

        INode fDecList = this.getFieldsSet();
        if (fDecList != null) {
            this.processNode(fDecList, data);
        }

        return retorno;
    }

}
