/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;

/**
 * Classe que  guarda as informacoes a respeito das declaracoes  de  metodo.
 * Essa classe  formada basicamente por tres conjuntos : um pra
 * metodos, outro para declaracoes do tipo <b>MethodDeclaration</b> e outra
 * para declaracoes do tipo <b>MethodDeclarationSet</b>
 *
 * @author Adeline de Sousa Silva
 * @version 12/06/2001
 */
public class BodyMethodDeclarations extends AbstractNode {

    // metodos do programa java
    private JMethodDeclarationSet methodsSet;

    // declaracao no programa jats que casa com um unico metodo do programa
    // java Ex: #mtd : JMethodDeclaration
    private NodeList methodVarSet;

    // declaracao no programa jats que casa com um conjunto de  metodos no do
    // programa java #mtds : MethodDeclarationSet
    private NodeList methodsVarSet;


    public BodyMethodDeclarations() {
        
        this.methodsSet = new JMethodDeclarationSet();
        this.methodVarSet = new NodeList();
        this.methodsVarSet = new NodeList();
    }

    /**
     * Adiciona uma declaracao de metodo ao conjunto methodsSet
     *
     * @param methDec Declaracao de metodo a ser adicionada
     */
    public void addMethodDeclaration(JMethodDeclaration methDec)
            throws IllegalArgumentException {

        if (methDec != null) {

            this.methodsSet.add(methDec);
        } else {
            throw new IllegalArgumentException("methDec nulo");
        }
    }

    /**
     * Adiciona uma declaracao '#mds:JMethodDeclarationSet' ao conjunto methodVarSet
     *
     * @param mtds Declaracao '#mds:JMethodDeclarationSet' a ser adicionada
     */
    public void addMethodsVarDeclaration(JMethodDeclarationSet mtds)
            throws IllegalArgumentException {

        if ((mtds != null) && mtds.isVariable() &&
                mtds.getVariableType() == JaTSNode.METHOD_DECLARATION_SET) {

            this.methodsVarSet.addElement(mtds);
        } else {
            throw new IllegalArgumentException("mtds nulo");
        }
    }

    /**
     * Adiciona uma declaracao '#mtd:JMethodDeclaration' ao conjunto methodVarSet
     *
     * @param mtd Declaracao '#mtd:JMethodDeclaration' a ser adicionada
     */
    public void addMethodVarDeclaration(JMethodDeclaration mtd)
            throws IllegalArgumentException {

        if ((mtd != null) && mtd.isVariable() &&
                mtd.getVariableType() == JaTSNode.METHOD_DECLARATION) {

            this.methodVarSet.addElement(mtd);
        } else {
            throw new IllegalArgumentException("mtd nulo");
        }
    }

    /**
     * Retorna o conjunto de metodos do programa java
     *
     * @return Conjunto de metodos do programa java
     */
    public JMethodDeclarationSet getMethodsSet() {
        return this.methodsSet;
    }

    /**
     * Retorna o conjunto de construcoes do tipo '#mds:JMethodDeclarationSet' do programa jats
     *
     * @return Conjunto de construcoes do tipo '#mds:JMethodDeclarationSet' do programa jats
     */
    public NodeList getMethodsVarSet() {

        return this.methodsVarSet;

    }

    /**
     * Retorna o conjunto de construcoes do tipo '#md:JMethodDeclaration' do programa jats
     *
     * @return Conjunto de construcoes do tipo '#md:JMethodDeclaration' do programa jats
     */
    public NodeList getMethodVarSet() {

        return this.methodVarSet;

    }


    /**
     * Seta o conjunto de metodos do programa java
     *
     * @param methods Conjunto de metodos do programa java
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setMethodsSet(JMethodDeclarationSet methods) {

        if (methods != null) {

            int i = 0;
            int size = methods.size();
            while (i < size) {
                if (!(methods.elementAt(i) instanceof JMethodDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
                i++;
            }
            this.methodsSet = methods;
        } else {
            throw new IllegalArgumentException("Conjunto de metodos nulo");
        }

    }

    /**
     * Seta o conjunto de construcoes do tipo 'BodyMethodDeclarations' do programa
     * jats
     *
     * @param methodsSet Construcoes do tipo 'BodyMethodDeclarations' do
     *                   programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setMethodsVarSet(NodeList methodsSet) {

        if (methodsSet != null) {
            Iterator it = methodsSet.iterator();
            while (it.hasNext()) {
                if (!(it.next() instanceof JMethodDeclarationSet)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }

            this.methodsVarSet = methodsSet;
        } else {
            throw new IllegalArgumentException("Conjunto metodos vazio");
        }

    }

    /**
     * Seta o conjunto de construcoes do tipo JMethodDeclaration do programa jats
     *
     * @param methodSet Construcoes do tipo JMethodDeclaration do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setMethodVarSet(NodeList methodSet) {

        if (methodSet != null) {
            Iterator it = methodSet.iterator();
            while (it.hasNext()) {
                if (!(it.next() instanceof JMethodDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.methodVarSet = methodSet;
        } else {
            throw new IllegalArgumentException("Conjunto de metodos nulo");
        }

    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.methodsSet</code>, <code>this.methodVarSet</code> e
     * <code>this.methodsVarSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Variveis que casam com vrias declaraes de metodos
        if (this.methodsVarSet != null) {

            int num = methodsVarSet.size() - 1;

            while (num >= 0) {
                this.methodsVarSet.elementAt(num).accept(visitor, data);
                num--;
            }
        }

        // Variveis que casam com uma nica declarao de metodo
        if (this.methodVarSet != null) {
            int num = methodVarSet.size() - 1;

            while (num >= 0) {
                this.methodVarSet.elementAt(num).accept(visitor, data);
                num--;
            }
        }

        // Declaraes de metodo
        if (this.methodsSet != null) {
            this.methodsSet.accept(visitor, data);
        } else {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (node == null || !(node instanceof BodyMethodDeclarations)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        BodyMethodDeclarations methodDecSet = (BodyMethodDeclarations) node;
        JMethodDeclarationSet nodeMethodDecs =
                (JMethodDeclarationSet) methodDecSet.getMethodsSet().clone();

        if (nodeMethodDecs == null) {
            throw new InconsistentNodeException(methodDecSet);
        }
        // Lista de variveis do tipo MethodDeclaration que ainda no foram casados
        NodeList unmatchedMethodDecVars = new NodeList();

        // Lista de variveis do tipo JMethodDeclarationSet que ainda no foram casadas.
        NodeList unmatchedMethodDecSetVars = new NodeList();


        unmatchedMethodDecVars = getUnmatchedMethodVariables(nodeMethodDecs, methodDecSet, results);


        unmatchedMethodDecSetVars = getUnmatchedDecSetVariables(results, nodeMethodDecs);


        // Casamento padro de declaraes de metodo no lado esquerdo da
        // transformao (this) com declaraes de metodo no programa fonte
        // (contidas em nodeMethodDecs).
        this.methodsSet.match(nodeMethodDecs, results);

        // Casamento das variveis do tipo JMethodDeclaration que ainda no esto
        // mapeadas a nenhum valor.
        matchMethodDecVariables(unmatchedMethodDecVars, nodeMethodDecs, results, node);


        // Casamento entre as declaraes do tipo varivel:JMethodDeclarationSet
        // e os metodos restantes em nodeMethodDecs.
        matchMethodDecSetVariables(unmatchedMethodDecSetVars, nodeMethodDecs, results, methodDecSet);
    }

    private void matchMethodDecSetVariables(NodeList decSetVars,
                                            JMethodDeclarationSet nodeMethodDecs,
                                            ResultSet results,
                                            BodyMethodDeclarations methodDecSet)
            throws InconsistentNodeException, NodesNotMatchedException {

        int methodsVarSize = decSetVars.size();
        if (methodsVarSize > 0) {
            // Se houver apenas uma varivel do tipo JMethodDeclarationSet, ento
            // esta ser casada com as declaraes de metodo restante
            if (methodsVarSize == 1) {
                INode matchesTheRest = decSetVars.elementAt(0);

                if (matchesTheRest == null || !matchesTheRest.isVariable() ||
                        matchesTheRest.getVariableType() !=
                        JaTSNode.METHOD_DECLARATION_SET ||
                        !(matchesTheRest instanceof JMethodDeclarationSet))
                    throw new InconsistentNodeException(this);

                JMethodDeclarationSet methodDecsVar = (JMethodDeclarationSet) matchesTheRest;

                methodDecsVar.matchesAsAVariable(nodeMethodDecs, results,
                        NO_TYPE);
            } else {
                throw new NodesNotMatchedException("The system could't decide how to perform the matching.");
            }
        }
        else if (nodeMethodDecs.size() > 0) {
            throw new NodesNotMatchedException("Some methods could not be matched " +
                    "(methods not matched: " + nodeMethodDecs.toString() + ")",
                    this, methodDecSet);
        }
    }

    private void matchMethodDecVariables(NodeList methodDecVars,
                                         JMethodDeclarationSet nodeMethodDecs,
                                         ResultSet results, INode node)
            throws NodesNotMatchedException, InconsistentNodeException {
        int m = 0;
        int max = methodDecVars.size();
        if (max <= nodeMethodDecs.size()) {
            while (m < max) {
                JaTSNode tempVar = (JaTSNode) methodDecVars.elementAt(m);
                if (!(tempVar instanceof JMethodDeclaration)) {
                    throw new IllegalMatchingException(tempVar);
                }
                JaTSNode tempDec = (JaTSNode) nodeMethodDecs.elementAt(0);
                // tenta casar os ns restantes
                tempVar.matchesAsAVariable(tempDec, results,
                        JaTSNode.METHOD_DECLARATION);
                nodeMethodDecs.removeElementAt(0);
                m++;
            }
        } else {
            // no pode restar variveis do tipo "MethodDeclaration" sem estar
            // casadas
            throw new NodesNotMatchedException("Some methods coudn't be matched " +
                    "in BodyMethodDeclararions", this, node);
        }
    }


    /**
     * Retorna as variveis do tipo MethodDeclarationSet que ainda no foram
     * mapeadas a algum valor
     *
     * @param results
     * @param nodeMethodDecs
     * @return
     * @throws InconsistentNodeException
     */
    private NodeList getUnmatchedDecSetVariables(ResultSet results,
                                                 JMethodDeclarationSet nodeMethodDecs)
            throws InconsistentNodeException {

        NodeList retorno = new NodeList();
        int i = 0;
        int max = this.methodsVarSet.size();
        INode tempMethodsVar = null;

        while (i < max) {

            tempMethodsVar = this.methodsVarSet.elementAt(i);

            if (!(tempMethodsVar instanceof JMethodDeclarationSet)) {
                throw new IllegalMatchingException(tempMethodsVar);
            }
            JMethodDeclarationSet methodSetVar =
                    (JMethodDeclarationSet) tempMethodsVar;

            if (results.alreadyMatched(methodSetVar)) {
                Object tempMds = null;
                tempMds = results.get(methodSetVar);

                if (!(tempMds instanceof JMethodDeclarationSet)) {
                    throw new IllegalMatchingException();
                }

                JMethodDeclarationSet mds = (JMethodDeclarationSet) tempMds;
                JaTSIterator ji = mds.getIterator();

                while (ji.hasNext()) {
                    INode methodsVarDecSetTemp = ji.next();

                    int c1 = 0;
                    int maxMDS = nodeMethodDecs.size();
                    boolean achou = false;

                    while (c1 < maxMDS && !achou) {
                        JaTSNode fDecTemp = (JaTSNode) nodeMethodDecs.elementAt(c1);

                        try {
                            methodsVarDecSetTemp.match(fDecTemp, results);
                            nodeMethodDecs.removeElementAt(c1);
                            achou = true;
                        } catch (NodesNotMatchedException nnme) {
                            c1++;
                        }
                    }
                }
            } else {
                retorno.addElement(methodSetVar);
            }
            i++;
        }
        return retorno;
    }


    /**
     * Remove as  variveis do tipo MethodDeclaration que j se encontram mapeadas a algum valor no ResultSet.
     *
     * @param nodeMethodDecs
     * @param methodDecSet
     * @param results
     * @return
     * @throws NodesNotMatchedException
     * @throws InconsistentNodeException
     */
    private NodeList getUnmatchedMethodVariables(JMethodDeclarationSet nodeMethodDecs,
                                                 BodyMethodDeclarations methodDecSet,
                                                 ResultSet results)
            throws NodesNotMatchedException, InconsistentNodeException {

        NodeList retorno = new NodeList();
        int nodeMethodDecsSize;
        boolean achou = false;
        int k = 0;
        int max = this.methodVarSet.size();

        // Com essa implementao padro,  necessrio verificar essa condio
        // para garantir que nenhuma varivel em this.methodVarSet
        // ficar sem ser casada.
        if (max > nodeMethodDecs.size()) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_SIZES_MSG,
                    this, methodDecSet);
        } else {
            while (k < max) {
                JaTSNode matchesSingleMethod = (JaTSNode) this.methodVarSet.elementAt(k);
                if (!matchesSingleMethod.isVariable() ||
                        !(matchesSingleMethod instanceof JMethodDeclaration)) {
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
                }
                // se a varivel j est casada, ento varre a lista de ns do
                // tipo JMethodDeclaration at encontrar o valor que lhe foi
                // mapeado e o remove da lista, para que este no possa ser
                // mapeado a outra varivel
                if (results.alreadyMatched(matchesSingleMethod)) {
                    int l = 0;
                    nodeMethodDecsSize = nodeMethodDecs.size();
                    achou = false;
                    Object tempNode = null;

                    tempNode = results.get(matchesSingleMethod);

                    if (!(tempNode instanceof JMethodDeclaration)) {
                        throw new IllegalMatchingException(matchesSingleMethod);
                    }
                    while ((l < nodeMethodDecsSize) && !achou) {
                        // Em nodeMethodDecs s existem ns do tipo
                        // JMethodDeclaration, por isso o cast sem instanceof
                        JMethodDeclaration tmp =
                                (JMethodDeclaration) nodeMethodDecs.elementAt(l);
                        try {
                            // Se encontrou o valor, remove o n da lista de
                            // declaraes de metodo do programa Java
                            tmp.match((JMethodDeclaration) tempNode, results);
                            nodeMethodDecs.removeElementAt(l);
                            achou = true;
                        } catch (NodesNotMatchedException e) {
                            l++;
                        }
                    }
                }
                // a varivel ainda no foi casada, adiciona-a  lista de
                // variveis que ainda no foram casadas
                else {
                    retorno.addElement(matchesSingleMethod);
                }

                k++;
            }
        }
        return retorno;
    }

    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * BodyMethodDeclarations se resume a invocar o mtodo <code>process</code> no
     * n correspondente  lista de declaraes de metodos filha do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Ns do tipo BodyMethodDeclarations nunca podem ser executveis.

        INode mDecList = this.getMethodsSet();
        if (mDecList != null) {
            this.processNode(mDecList, data);
        }

        return retorno;
    }


}
