/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa casts.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class CastExpression extends JExpression {

    public static final int DEFAULT_CAST = 0;
    public static final int AS_CAST = 1;

    /**
     * Armazena a expresso que recebe o <i>cast</i>.
     */
    private JExpression operand;

    /**
     * Armazena o tipo para o qual a expresso est sendo <i>casted</i>.
     */
    private JType type;


    private int castType;

    /**
     * Construtor vazio da classe
     */
    public CastExpression() {
        this.operand = null;
        this.type = null;
        this.castType = DEFAULT_CAST;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetros o operando que receber o
     * <i>cast</i> e o tipo para o qual ele ser <i>casted</i>.
     *
     * @param theOperand O operando que ter seu tipo <i>casted</i> para um outro tipo.
     * @param theType    O tipo para o qual o tipo do operando ser casted.
     */
    public CastExpression(JType theType, JExpression theOperand) {
        this.operand = theOperand;
        this.type = theType;
        this.castType = DEFAULT_CAST;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);

    }

    /**
     * Construtor da classe que recebe como parmetros o operando que receber o
     * <i>cast</i> e o tipo para o qual ele ser <i>casted</i> e o tipo do cast.
     * <br>
     * Se  o default, ou seja:
     *  (type) operand
     * Ou se  um cast de C#:
     *  operand "as" type
     * </br>
     * @param theOperand O operando que ter seu tipo <i>casted</i> para um outro tipo.
     * @param theType    O tipo para o qual o tipo do operando ser casted.
     */
    public CastExpression(JType theType, JExpression theOperand, int castType) {
        this.operand = theOperand;
        this.type = theType;
        this.castType = castType;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    public boolean isDefaultCast(){
        return this.castType == DEFAULT_CAST;
    }


    /**
     * Mtodo que retorna o operando da expresso.
     *
     * @return Uma objeto do tipo JExpression correspondendo  expresso que
     *         receber o <i>cast</i>.
     */
    public JExpression getOperand() {
        return this.operand;
    }

    /**
     * Mtodo que retorna o tipo para o qual o operando ser <i>casted</i>.
     *
     * @return Uma objeto do tipo <code>JType</code>, representando o tipo para o
     *         qual o operando ser <i>casted</i>.
     */
    public JType getType() {
        return this.type;
    }


    /**
     * Mtodo que seta o operando da expresso.
     *
     * @param newOperand Um objeto do tipo <code>JExpression</code> correspondendo
     *                   ao operando desta expresso.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>newOperand</code> for <code>null</code>.
     */
    public void setOperand(JExpression newOperand)
            throws IllegalArgumentException {

        if (newOperand == null) {
            throw new IllegalArgumentException();
        }
        this.operand = newOperand;
    }

    /**
     * Mtodo que seta o tipo para o qual o operando ser <i>casted</i>.
     *
     * @param newType Uma objeto do tipo <code>JType</code>, representando o tipo para o
     *                qual o operando ser <i>casted</i>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>newType</code> for <code>null</code>.
     */
    public void setType(JType newType)
            throws IllegalArgumentException {

        if (newType == null) {
            throw new IllegalArgumentException();
        }
        this.type = newType;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.type</code> e o de <code>this.operand</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = this;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            if (this.type == null || this.operand == null) {
                throw new InconsistentNodeException(this);
            }

            // envia o visitor pela sub-rvore do operando
            this.operand.accept(visitor, data);

            // envia o visitor pela sub-rvore do tipo
            this.type.accept(visitor, data);

        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException("A node received is null (nodes:" +
                    node + ", " +
                    this.getClass().toString() + ")", this, node);
        }

        if (!(node instanceof CastExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            CastExpression exp = (CastExpression) node;

            if (this.getOperand() == null || this.getType() == null) {
                throw new InconsistentNodeException(this);
            }

            (this.getType()).match(exp.getType(), results);
            (this.getOperand()).match(exp.getOperand(), results);
        }

    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do tipo para o qual o operando ser casted.
            INode cType = this.getType();
            if (cType != null) {
                Object ret1 = this.processNode(cType, data);
                if (cType.isExecutable()) {
                    try {
                        this.setType(JType.unwrapType(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento do operando  direita do operador.
            INode oper = this.getOperand();
            if (oper != null) {
                Object ret2 = this.processNode(oper, data);
                if (oper.isExecutable()) {
                    if (!(ret2 instanceof JExpression)) {
                        throw new ExecutionException(this);
                    }
                    this.setOperand((JExpression) ret2);
                }
            }
        }

        return retorno;
    }


    public Object evaluate(Object data) throws ExecutionException {
        
        return super.evaluate(data);
    }


}
