/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;


/**
 * Classe que  guarda as informacoes a respeito das declaracoes de  construtor
 * presentes na classe Java.
 *
 * @author Adeline de Sousa
 * @author Fernando Castor
 */
public class ClassBodyConstructorDeclarations extends AbstractNode {

    // construtores do programa java
    private JConstructorDeclarationSet constructorsSet;

    // declaracao no programa jats que casa com um unico construtor do programa
    // java
    private NodeList constructorVarSet;

    // declaracao no programa jats que casa com um conjunto de  construtores
    // do programa java
    private NodeList constructorsVarSet;


    public ClassBodyConstructorDeclarations() {
        this.constructorsSet = new JConstructorDeclarationSet();
        this.constructorVarSet = new NodeList();
        this.constructorsVarSet = new NodeList();
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.constructorsSet</code>, <code>this.constructorVarSet</code> e
     * <code>this.constructorsVarSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Variveis que casam com vrias declaraes de construtors
        if (this.constructorsVarSet != null) {
            int size = constructorsVarSet.size() - 1;

            while (size >= 0) {
                this.constructorsVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Variveis que casam com uma nica declarao de construtor
        if (this.constructorVarSet != null) {
            int size = constructorVarSet.size() - 1;

            while (size >= 0) {
                this.constructorVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Declaraes de construtor
        if (this.constructorsSet != null) {
            this.constructorsSet.accept(visitor, data);
        } else {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }

    /**
     * Adiciona uma declaracao de construtor ao conjunto constructorsSet
     *
     * @param constructorDec Declaracao de construtor a ser adicionada
     */
    public void addConstructorDeclaration(JConstructorDeclaration constructorDec)
            throws IllegalArgumentException {

        if (constructorDec != null) {

            this.constructorsSet.add(constructorDec);
        } else {
            throw new IllegalArgumentException("constructorDec nulo");
        }
    }

    /**
     * Adiciona uma declaracao '#cds: JConstructorDeclarationSet' ao conjunto constructorVarSet
     *
     * @param atts Declaracao '#cds: JConstructorDeclarationSet' a ser adicionada
     */
    public void addConstructorsVarDeclaration(JConstructorDeclarationSet atts)
            throws IllegalArgumentException {

        if (atts != null && atts.isVariable() &&
                atts.getVariableType() == JaTSNode.CONSTRUCTOR_DECLARATION_SET) {

            this.constructorsVarSet.addElement(atts);
        } else {
            throw new IllegalArgumentException("atts nulo");
        }
    }

    /**
     * Adiciona uma declaracao do tipo "#c:JConstructorDeclaration" ao conjunto
     * constructorVarSet
     *
     * @param constructor Declaracao a ser adicionada
     */
    public void addConstructorVarDeclaration(JConstructorDeclaration constructor)
            throws IllegalArgumentException {

        if (constructor != null && constructor.isVariable() &&
                constructor.getVariableType() == JaTSNode.CONSTRUCTOR_DECLARATION) {

            this.constructorVarSet.addElement(constructor);
        } else {
            throw new IllegalArgumentException("constructor nulo");
        }
    }

    /**
     * Retorna o conjunto de construcoes do tipo #cd: JConstructorDeclaration do programa jats
     *
     * @return Conjunto de construcoes do tipo #cd: JConstructorDeclaration do programa jats
     */
    public NodeList getConstructorVarSet() {
        return this.constructorVarSet;

    }

    /**
     * Retorna o conjunto de construcoes do tipo #cds: JConstructorDeclarationSet do programa jats
     *
     * @return Conjunto de construcoes do tipo #cds: JConstructorDeclarationSet do programa jats
     */
    public NodeList getConstructorsVarSet() {
        return this.constructorsVarSet;

    }

    /**
     * Retorna o conjunto de construtors do programa java
     *
     * @return Conjunto de construtors do programa java
     */
    public JConstructorDeclarationSet getConstructorsSet() {
        return this.constructorsSet;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (node == null || !(node instanceof ClassBodyConstructorDeclarations)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        ClassBodyConstructorDeclarations constructorDecSet =
                (ClassBodyConstructorDeclarations) node;

        JConstructorDeclarationSet nodeConstructorDecs =
                (JConstructorDeclarationSet) constructorDecSet.getConstructorsSet().clone();
        if (nodeConstructorDecs == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(constructorDecSet);
        }
        int nodeConstructorDecsSize = nodeConstructorDecs.size();
        boolean achou = false;

        // Casamento das declaraes do tipo varivel:JConstructorDeclaration com os
        // construtors que ainda esto em nodeConstructorDecs. Primeiro so casadas as
        // variveis que J FORAM mapeadas a algum valor. Isso  necessrio para
        // garantir que nenhuma declarao no fonte Java  mapeada a mais de uma
        // varivel.
        int k = 0;
        int max3 = this.constructorVarSet.size();

        // Lista de variveis do programa JaTS que ainda no foram casados
        NodeList unmatchedConstructorDecVars = new NodeList();

        // Lista de variveis do tipo JConstructorDeclarationSet do programa JaTS que
        // ainda no foram casadas.
        NodeList unmatchedConstructorDecSetVars = new NodeList();

        // Com essa implementao padro,  necessrio verificar essa condio
        // para garantir que nenhuma varivel em this.constructorVarSet
        // ficar sem ser casada.
        if (max3 > nodeConstructorDecs.size()) {
            throw new NodesNotMatchedException("There is a different number of " +
                    "constructor in each declaration (" + max3 + ", " +
                    nodeConstructorDecs.size() + ")", this, constructorDecSet);
        } else {
            while (k < max3) {
                JaTSNode matchesSingleConstructor = (JaTSNode) this.constructorVarSet.elementAt(k);
                if (!matchesSingleConstructor.isVariable() ||
                        matchesSingleConstructor.getVariableType() !=
                        JaTSNode.CONSTRUCTOR_DECLARATION ||
                        !(matchesSingleConstructor instanceof JConstructorDeclaration)) {
                    throw new InconsistentNodeException(this);
                }
                // se a varivel j est casada, ento varre a lista de ns do
                // tipo JConstructorDeclaration at encontrar o valor que lhe foi
                // mapeado e o remove da lista, para que este no possa ser
                // mapeado a outra varivel
                if (results.alreadyMatched(matchesSingleConstructor)) {
                    int l = 0;
                    nodeConstructorDecsSize = nodeConstructorDecs.size();
                    achou = false;
                    Object tempNode = null;
                    tempNode = results.get(matchesSingleConstructor);

                    if (!(tempNode instanceof JConstructorDeclaration)) {
                        throw new IllegalMatchingException(matchesSingleConstructor);
                    }
                    while ((l < nodeConstructorDecsSize) && !achou) {
                        // Em nodeConstructorDecs s existem ns do tipo
                        // JConstructorDeclaration, por isso o cast sem instanceof
                        JConstructorDeclaration tmp =
                                (JConstructorDeclaration) nodeConstructorDecs.elementAt(l);
                        try {
                            // Se encontrou o valor, remove o n da lista de
                            // declaraes de construtor do programa Java
                            tmp.match((JConstructorDeclaration) tempNode, results);
                            nodeConstructorDecs.removeElementAt(l);
                            achou = true;
                        } catch (NodesNotMatchedException e) {
                            l++;
                        }
                    }
                }
                // a varivel ainda no foi casada, adiciona-a  lista de
                // variveis que ainda no foram casadas
                else {
                    unmatchedConstructorDecVars.addElement(matchesSingleConstructor);
                }

                k++;
            }
        }

        // Casamento de variveis do tipo JConstructorDeclarationSet. Similar ao
        // anterior: apenas para remover da lista de declaraes de construtor
        // do tipo fonte Java as declaraes que j esto casadas a variveis.
        // Nesse caso, a variveis do tipo JConstructorDeclarationSet.
        int constructorsVarSetSize = this.constructorsVarSet.size();
        int n = 0;
        INode tempConstructorsVar = null;

        while (n < constructorsVarSetSize) {
            tempConstructorsVar = this.constructorsVarSet.elementAt(n);
            if (!(tempConstructorsVar instanceof JConstructorDeclarationSet)) {
                throw new IllegalMatchingException(tempConstructorsVar);
            }
            JConstructorDeclarationSet constructorSetVar =
                    (JConstructorDeclarationSet) tempConstructorsVar;

            if (results.alreadyMatched(constructorSetVar)) {
                Object tempCds = results.get(constructorSetVar);

                if (!(tempCds instanceof JConstructorDeclarationSet)) {
                    throw new IllegalMatchingException();
                }

                JConstructorDeclarationSet cds = (JConstructorDeclarationSet) tempCds;
                JaTSIterator ji = cds.getIterator();

                while (ji.hasNext()) {
                    INode constructorsVarDecSetTemp = ji.next();

                    int c1 = 0;
                    int maxCDS = nodeConstructorDecs.size();
                    boolean achou2 = false;

                    while (c1 < maxCDS && !achou2) {
                        JaTSNode fDecTemp = (JaTSNode) nodeConstructorDecs.elementAt(c1);

                        try {
                            constructorsVarDecSetTemp.match(fDecTemp, results);
                            nodeConstructorDecs.removeElementAt(c1);
                            achou2 = true;
                        } catch (NodesNotMatchedException nnme) {
                            c1++;
                        }
                    }
                }
            } else {
                unmatchedConstructorDecSetVars.addElement(constructorSetVar);
            }
            n++;
        }


        // Casamento padro de declaraes de construtor no lado esquerdo da
        // transformao (this) com declaraes de construtor no programa fonte
        // (contidas em nodeConstructorDecs).
        this.constructorsSet.match(nodeConstructorDecs, results);

        // Casamento das variveis do tipo JConstructorDeclaration que ainda no esto
        // mapeadas a nenhum valor.
        int m = 0;
        int max5 = unmatchedConstructorDecVars.size();
        if (max5 <= nodeConstructorDecs.size()) {
            while (m < max5) {
                JaTSNode tempVar = (JaTSNode) unmatchedConstructorDecVars.elementAt(m);
                if (!(tempVar instanceof JConstructorDeclaration)) {
                    throw new IllegalMatchingException(tempVar);
                }
                JaTSNode tempDec = (JaTSNode) nodeConstructorDecs.elementAt(0);
                // tenta casar os ns restantes
                tempVar.matchesAsAVariable(tempDec, results,
                        JaTSNode.CONSTRUCTOR_DECLARATION);
                nodeConstructorDecs.removeElementAt(0);
                m++;
            }
        } else {
            // no pode restar variveis do tipo "JConstructorDeclaration" sem estar
            // casadas
            throw new NodesNotMatchedException("Some constructors couldn't be matched " +
                    "in ClassBodyConstructorDeclararions", this, node);
        }


        // Casamento entre as declaraes do tipo varivel:JConstructorDeclarationSet
        // e os construtors restantes em nodeConstructorDecs.
        int constructorsVarSize = unmatchedConstructorDecSetVars.size();
        if (constructorsVarSize > 0) {
            // Se houver apenas uma varivel do tipo JConstructorDeclarationSet, ento
            // esta ser casada com as declaraes de construtor restante
            if (constructorsVarSize == 1) {
                INode matchesTheRest = unmatchedConstructorDecSetVars.elementAt(0);

                if (matchesTheRest == null || !matchesTheRest.isVariable() ||
                        matchesTheRest.getVariableType() !=
                        JaTSNode.CONSTRUCTOR_DECLARATION_SET ||
                        !(matchesTheRest instanceof JConstructorDeclarationSet))
                    throw new InconsistentNodeException(this);

                JConstructorDeclarationSet constructorDecsVar = (JConstructorDeclarationSet) matchesTheRest;

                constructorDecsVar.matchesAsAVariable(nodeConstructorDecs, results,
                        NO_TYPE);
            } else {
                // Se houver mais que uma varivel do tipo JConstructorDeclarationSet
                // no casada, o sistema no  capaz de decidir como proceder e
                // o casamento falha.
                JConstructorDeclarationSet unmatchedvarDecs =
                        new JConstructorDeclarationSet(unmatchedConstructorDecSetVars);
                throw new NodesNotMatchedException("The system could't decide " +
                        "how to perform the matching. (" + constructorsVarSize + " VarSets)",
                        unmatchedvarDecs, nodeConstructorDecs);
            }
        } else if (nodeConstructorDecs.size() > 0) {
            throw new NodesNotMatchedException("Some constructors could not be matched " +
                    "(number of felds: " + nodeConstructorDecs.size() + ")", this,
                    constructorDecSet);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * ClassBodyConstructorDeclarations se resume a invocar o mtodo <code>process</code> no
     * n correspondente  lista de declaraes de construtors filha do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Ns do tipo ClassBodyConstructorDeclarations nunca podem ser
        // executveis. Por isso, temos apenas o encaminhamento.

        INode cDecList = this.getConstructorsSet();
        if (cDecList != null) {
            this.processNode(cDecList, data);
        }

        return retorno;
    }


    /**
     * Seta o conjunto de construcoes do tipo attribute do programa jats
     *
     * @param constructorVarSet Construcoes do tipo attribute do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setConstructorVarSet(NodeList constructorVarSet) {

        if (constructorVarSet == null) {
            throw new IllegalArgumentException("Conjunto attribute vazio");
        }
        Iterator it = constructorVarSet.iterator();
        while (it.hasNext()) {
            if (!((it.next()) instanceof JConstructorDeclaration)) {
                throw new IllegalArgumentException("Elemento invalido");
            }
        }
        this.constructorVarSet = constructorVarSet;

    }


    /**
     * Seta o conjunto de construcoes do tipo attributes do programa jats
     *
     * @param constructorsSet Construcoes do tipo attributes do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setConstructorsVarSet(NodeList constructorsSet) {
        if (constructorsSet == null) {
            throw new IllegalArgumentException("Conjunto attributes vazio");
        }
        Iterator it = constructorsSet.iterator();
        while (it.hasNext()) {
            if (!((it.next()) instanceof JConstructorDeclarationSet)) {
                throw new IllegalArgumentException("Elemento invalido");
            }
        }
        this.constructorsVarSet = constructorsSet;
    }

    /**
     * Seta o conjunto de construtors do programa java
     *
     * @param constructors Conjunto de construtors do programa java
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setConstructorsSet(JConstructorDeclarationSet constructors) {

        if (constructors == null) {
            throw new IllegalArgumentException("Conjunto de construtors vazio");
        }
        JaTSIterator jit = constructors.getIterator();
        while (jit.hasNext()) {
            if (!(jit.next() instanceof JConstructorDeclaration)) {
                throw new IllegalArgumentException("Elemento invalido");
            }
        }
        this.constructorsSet = constructors;
    }

}
