/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;

/**
 * Classe que   guarda as informacoes a respeito das declaracoes
 * de  inicializador
 *
 * @author Adeline de Sousa
 */
public class ClassBodyInitializers extends AbstractNode {

    // inicializadores do programa java
    private JInitializerSet initsSet;

    // declaracao no programa jats que casa com um unico inicializador do programa
    // java
    private NodeList initVarSet;

    // declaracao no programa jats que casa com um conjunto de  inicializadores
    // no do programa java
    private NodeList initsVarSet;


    public ClassBodyInitializers() {
        this.initsSet = new JInitializerSet();
        this.initVarSet = new NodeList();
        this.initsVarSet = new NodeList();
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.initsSet</code>, <code>this.initVarSet</code> e
     * <code>this.initsVarSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        Object retorno = this;

        // Variveis que casam com vrias declaraes de inicializadores
        if (this.initsVarSet != null) {
            int size = initsVarSet.size() - 1;

            while (size >= 0) {
                this.initsVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Variveis que casam com uma nica declarao de inicializador
        if (this.initVarSet != null) {
            int size = initVarSet.size() - 1;

            while (size >= 0) {
                this.initVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Declaraes de inicializador
        if (this.initsSet != null) {
            this.initsSet.accept(visitor, data);
        } else {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }

    /**
     * Adiciona uma declaracao de inicializador ao conjunto initsSet
     *
     * @param initDec Declaracao de inicializador a ser adicionada
     */
    public void addInitializer(JInitializer initDec)
            throws IllegalArgumentException {

        if (initDec != null) {
            this.initsSet.add(initDec);
        } else {
            throw new IllegalArgumentException("initDec nulo");
        }
    }

    /**
     * Adiciona uma declaracao '#init:InitializerSet' ao conjunto initVarSet
     *
     * @param atts Declaracao InitializerSet a ser adicionada
     */
    public void addInitializersSetVar(JInitializerSet atts)
            throws IllegalArgumentException {

        if (atts != null && atts.isVariable() &&
                atts.getVariableType() == JaTSNode.INITIALIZER_SET) {

            this.initsVarSet.addElement(atts);
        } else {
            throw new IllegalArgumentException("atts nulo");
        }
    }

    /**
     * Adiciona uma declaracao do tipo "#init:Initializer" ao conjunto
     * initVarSet
     *
     * @param init Declaracao a ser adicionada
     */
    public void addInitializerVar(JInitializer init)
            throws IllegalArgumentException {

        if (init != null && init.isVariable() &&
                init.getVariableType() == JaTSNode.INITIALIZER) {

            this.initVarSet.addElement(init);
        } else {
            throw new IllegalArgumentException("init nulo");
        }
    }

    /**
     * Retorna o conjunto de construcoes do tipo Initializer do programa jats
     *
     * @return Conjunto de construcoes do tipo Initializer do programa jats
     */
    public NodeList getInitializerVarSet() {
        return this.initVarSet;
    }

    /**
     * Retorna o conjunto de construcoes do tipo InitializerSet do programa jats
     *
     * @return Conjunto de construcoes do tipo InitializerSet do programa jats
     */
    public NodeList getInitializersVarSet() {
        return this.initsVarSet;
    }

    /**
     * Retorna o conjunto de inicializadores do programa java
     *
     * @return Conjunto de inicializadores do programa java
     */
    public JInitializerSet getInitializersSet() {
        return this.initsSet;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (node == null || !(node instanceof ClassBodyInitializers)) {
            String nodeTypeStr = "null";
            if (node != null) nodeTypeStr = node.getClass().toString();
            throw new NodesNotMatchedException("Nodes have different types (" +
                    nodeTypeStr + ", " + this.getClass().toString() + ")",
                    this, node);
        }

        ClassBodyInitializers initDecSet = (ClassBodyInitializers) node;
        JInitializerSet nodeInitializerDecs =
                (JInitializerSet) initDecSet.getInitializersSet().clone();
        if (nodeInitializerDecs == null) {
            throw new InconsistentNodeException(initDecSet);
        }
        int nodeInitializerDecsSize = nodeInitializerDecs.size();
        boolean achou = false;

        // Casamento das declaraes do tipo varivel:JInitializer com os
        // inicializadores que ainda esto em nodeInitializerDecs. Primeiro so casadas as
        // variveis que J FORAM mapeadas a algum valor. Isso  necessrio para
        // garantir que nenhuma declarao no fonte Java  mapeada a mais de uma
        // varivel.
        int k = 0;
        int max3 = this.initVarSet.size();

        // Lista de variveis do programa JaTS que ainda no foram casados
        NodeList unmatchedInitializerDecVars = new NodeList();

        // Lista de variveis do tipo JInitializerSet do programa JaTS que
        // ainda no foram casadas.
        NodeList unmatchedInitializerDecSetVars = new NodeList();

        // Com essa implementao padro,  necessrio verificar essa con-
        // dio para garantir que nenhuma varivel em this.initVarSet
        // ficar sem ser casada.
        if (max3 > nodeInitializerDecs.size()) {
            throw new NodesNotMatchedException("There is a different number of " +
                    "initializers in each declaration (" + max3 + ", " +
                    nodeInitializerDecs.size() + ")", this, initDecSet);
        } else {
            while (k < max3) {
                JaTSNode matchesSingleInitializer = (JaTSNode) this.initVarSet.elementAt(k);
                if (!matchesSingleInitializer.isVariable() ||
                        matchesSingleInitializer.getVariableType() !=
                        JaTSNode.INITIALIZER ||
                        !(matchesSingleInitializer instanceof JInitializer)) {
                    throw new InconsistentNodeException(this);
                }
                // se a varivel j est casada, ento varre a lista de ns do
                // tipo JInitializer at encontrar o valor que lhe foi
                // mapeado e o remove da lista, para que este no possa ser
                // mapeado a outra varivel
                if (results.alreadyMatched(matchesSingleInitializer)) {
                    int l = 0;
                    nodeInitializerDecsSize = nodeInitializerDecs.size();
                    achou = false;
                    Object tempNode = results.get(matchesSingleInitializer);

                    if (!(tempNode instanceof JInitializer)) {
                        throw new IllegalMatchingException(matchesSingleInitializer);
                    }
                    while ((l < nodeInitializerDecsSize) && !achou) {
                        // Em nodeInitializerDecs s existem ns do tipo
                        // JInitializer, por isso o cast sem instanceof
                        JInitializer tmp =
                                (JInitializer) nodeInitializerDecs.elementAt(l);
                        try {
                            // Se encontrou o valor, remove o n da lista de
                            // declaraes de inicializador do programa Java
                            tmp.match((JInitializer) tempNode, results);
                            nodeInitializerDecs.removeElementAt(l);
                            achou = true;
                        } catch (NodesNotMatchedException e) {
                            l++;
                        }
                    }
                }
                // a varivel ainda no foi casada, adiciona-a  lista de
                // variveis que ainda no foram casadas
                else {
                    unmatchedInitializerDecVars.addElement(matchesSingleInitializer);
                }

                k++;
            }
        }

        // Casamento de variveis do tipo JInitializerSet. Similar ao
        // anterior: apenas para remover da lista de declaraes de inicializador
        // do tipo fonte Java as declaraes que j esto casadas a variveis.
        // Nesse caso, a variveis do tipo JInitializerSet.
        int initsVarSetSize = this.initsVarSet.size();
        int n = 0;
        INode tempInitializersVar = null;

        while (n < initsVarSetSize) {
            tempInitializersVar = this.initsVarSet.elementAt(n);
            if (!(tempInitializersVar instanceof JInitializerSet)) {
                throw new IllegalMatchingException(tempInitializersVar);
            }
            JInitializerSet initSetVar =
                    (JInitializerSet) tempInitializersVar;

            if (results.alreadyMatched(initSetVar)) {
                Object tempInitSet = results.get(initSetVar);

                if (!(tempInitSet instanceof JInitializerSet)) {
                    throw new IllegalMatchingException();
                }

                JInitializerSet initSet = (JInitializerSet) tempInitSet;
                JaTSIterator jit = initSet.getIterator();

                while (jit.hasNext()) {
                    INode initsVarDecSetTemp = jit.next();
                    int c1 = 0;
                    int maxIS = nodeInitializerDecs.size();
                    boolean achou2 = false;
                    while (c1 < maxIS && !achou2) {
                        JaTSNode iTemp = (JaTSNode) nodeInitializerDecs.elementAt(c1);
                        try {
                            initsVarDecSetTemp.match(iTemp, results);
                            nodeInitializerDecs.removeElementAt(c1);
                            achou2 = true;
                        } catch (NodesNotMatchedException nnme) {
                            c1++;
                        }
                    }
                }
            } else {
                unmatchedInitializerDecSetVars.addElement(initSetVar);
            }
            n++;
        }

        // Casamento padro de declaraes de inicializador no lado esquerdo da
        // transformao (this) com declaraes de inicializador no programa fonte
        // (contidas em nodeInitializerDecs).
        this.initsSet.match(nodeInitializerDecs, results);

        // Casamento das variveis do tipo JInitializer que ainda no esto
        // mapeadas a nenhum valor.
        int m = 0;
        int max5 = unmatchedInitializerDecVars.size();
        if (max5 <= nodeInitializerDecs.size()) {
            while (m < max5) {
                JaTSNode tempVar = (JaTSNode) unmatchedInitializerDecVars.elementAt(m);
                if (!(tempVar instanceof JInitializer)) {
                    throw new IllegalMatchingException(tempVar);
                }
                JaTSNode tempDec = (JaTSNode) nodeInitializerDecs.elementAt(0);
                // tenta casar os ns restantes
                tempVar.matchesAsAVariable(tempDec, results,
                        JaTSNode.INITIALIZER);
                nodeInitializerDecs.removeElementAt(0);
                m++;
            }
        } else {
            // no pode restar variveis do tipo "Initializer" sem estar
            // casadas
            throw new NodesNotMatchedException("Some initializers couldn't be matched " +
                    "in ClassBodyInitializers", this, node);
        }


        // Casamento entre as declaraes do tipo varivel:JInitializerSet
        // e os inicializadores restantes em nodeInitializerDecs.
        int initsVarSize = unmatchedInitializerDecSetVars.size();
        if (initsVarSize > 0) {
            // Se houver apenas uma varivel do tipo JInitializerSet, ento
            // esta ser casada com as declaraes de inicializador restante
            if (initsVarSize == 1) {
                INode matchesTheRest = unmatchedInitializerDecSetVars.elementAt(0);

                if (matchesTheRest == null || !matchesTheRest.isVariable() ||
                        matchesTheRest.getVariableType() !=
                        JaTSNode.INITIALIZER_SET ||
                        !(matchesTheRest instanceof JInitializerSet))
                    throw new InconsistentNodeException(this);

                JInitializerSet initDecsVar = (JInitializerSet) matchesTheRest;

                initDecsVar.matchesAsAVariable(nodeInitializerDecs, results,
                        NO_TYPE);
            } else {
                // Se houver mais que uma varivel do tipo JInitializerSet
                // no casada, o sistema no  capaz de decidir como proceder e
                // o casamento falha.
                JInitializerSet unmatchedvarDecs = new JInitializerSet(unmatchedInitializerDecSetVars);
                throw new NodesNotMatchedException("The system could't decide " +
                        "how to perform the matching. (" + initsVarSize + " VarSets)",
                        unmatchedvarDecs, nodeInitializerDecs);
            }
        } else if (nodeInitializerDecs.size() > 0) {
            throw new NodesNotMatchedException("Some fields could not be matched " +
                    "(number of felds: " + nodeInitializerDecs.size() + ")", this, initDecSet);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * BodyInitializers se resume a invocar o mtodo <code>process</code> no
     * n correspondente  lista de declaraes de inicializadores filha do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Levantada
     *                                   caso o parmetro <code>data</code> seja <code>null</code> numa
     *                                   situao em que ele  necessrio.
     * @throws InconsistentNodeException Levantada se um n inconsistente (incompleto) for encontrado.
     *                                   Se algum dos componentes (filhos) essenciais ao correto
     *                                   funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Ns do tipo BodyInitializers nunca podem ser executveis.
        // a funo deste mtodo  a de encaminhamento
        INode initList = this.getInitializersSet();
        if (initList != null) {
            this.processNode(initList, data);
        }

        return retorno;
    }


    /**
     * Seta o conjunto de construcoes do tipo attribute do programa jats
     *
     * @param attSet Construcoes do tipo attribute do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setInitializerVarSet(NodeList attSet) {

        if (attSet != null) {
            Iterator it = attSet.iterator();
            while (it.hasNext()) {
                if (!((it.next()) instanceof JInitializer)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.initVarSet = attSet;
        } else {
            throw new IllegalArgumentException("Conjunto attribute vazio");
        }

    }


    /**
     * Seta o conjunto de construcoes do tipo attributes do programa jats
     *
     * @param initsSet Construcoes do tipo attributes do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setInitializersVarSet(NodeList initsSet) {

        if (initsSet != null) {
            Iterator it = initsSet.iterator();
            while (it.hasNext()) {
                if (!((it.next()) instanceof JInitializerSet)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.initsVarSet = initsSet;
        } else {
            throw new IllegalArgumentException("Conjunto attributes vazio");
        }

    }

    /**
     * Seta o conjunto de inicializadores do programa java
     *
     * @param inits Conjunto de inicializadores do programa java
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setInitializersSet(JInitializerSet inits) {

        if (inits != null) {
            int i = 0;
            int size = inits.size();
            while (i < size) {
                if (!(inits.elementAt(i) instanceof JInitializer)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
                i++;
            }
            this.initsSet = inits;
        } else {
            throw new IllegalArgumentException("Conjunto de inicializadores vazio");
        }

    }

}
