/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.Position;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.SelectionProcessor;
import cin.jats.engine.visitors.IVisitor;


/**
 * This class represents a context variable declaration. This kind of
 * declaration is compound by two parts and appears in the pattern code as 
 * follows:
 * <br><br>
 * ContextDeclaration:#cont{#code}
 * <br><br>
 * where #code represents a portion of valid code and #cont the context code
 * that wraps #code.
 * 
 * @author Gustavo Santos
 * @version 31/08/05
 */
public class ContextDeclaration extends JStatementList {
	
	/**
	 * The interest code position relative to the context (represents a 
	 * "role" in the context).
	 * Note this is different from the value inside interestCode
	 * since the later is relative to the hole program code.
	 */
//	private Position codePosition;
	
	/**
	 * The context code without the portion pointed by interestCode field
	 */
	//private JStatementList context;
	
	/**
	 * Code of interest
	 */
	private JStatementList codeOfInterest;

	
	/**
	 * Constructor for ContextDeclaration objects
	 */
	public ContextDeclaration() {
//		this.context = new JStatementList(new NodeList());
		super(new NodeList());
		this.codeOfInterest = new JStatementList(new NodeList());
	}
	
	/**
	 * Constructor for ContextDeclaration objects
	 * 
	 * @param context Context code
	 * @param code Code of interest
	 */
	public ContextDeclaration(JStatementList context, JStatementList code) {
		//this.context = context;
		super(context.getStatements());
		this.setVariableName(context.getVariableName());
		this.setVariableType(context.getVariableType());
		this.codeOfInterest = code;
	}

	/**
	 * Standard accept method for visitors
	 * 
	 * @see cin.jats.engine.parser.nodes.INode#accept(cin.jats.engine.visitors.IVisitor, java.lang.Object)
	 */
	public Object accept(IVisitor visitor, Object data)
			throws InconsistentNodeException, IllegalArgumentException,
			ExecutionException {		
		Object retorno;
		
		//this.context.accept(visitor, data);
		
		if (this.codeOfInterest != null)
		this.codeOfInterest.accept(visitor, data);
		
        // ... invoke the accept of the executable declaration associated to 
		//this node, if any...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // then visit this node.
        retorno = visitor.visit(this, data);

		return retorno;
	}

	/**
	 * @see cin.jats.engine.parser.nodes.INode#match(cin.jats.engine.parser.nodes.INode, cin.jats.engine.util.ResultSet)
	 */
	public void match(INode node, ResultSet results)
			throws IllegalArgumentException, InconsistentNodeException,
			NodesNotMatchedException {
		JStatementList nodeList = (JStatementList)node;
		
		//JExpression is the root node for all nodes inside methods
		//because this, any concrete code must be a JExpression
		JExpression concreteCode;
		
		//replaces the jats variable by the equivalent abstract node
		//in the context declaration node
		this.setCodeOfInterest((JStatementList) results.get(
				this.getCodeOfInterest()));

		//the position if dinamically changed during the process
		this.setCodePosition(this.getCodeOfInterest().getPosition());
		
		//retrieves the concrete refrences to the code of interest
		concreteCode = (JExpression) SelectionProcessor.extractCode(nodeList, this);
		
		//the position is needed for introducing the code back
		concreteCode.setPosition(this.getCodePosition());
		
		//replaces the mapping to the code of interest 
		results.put(this.getCodeOfInterest(), concreteCode);

		//keep the code position value
		nodeList.setPosition(this.getCodeOfInterest().getPosition());

		//maps the context (already without the internal portion)
		results.put(this.getContext(), nodeList);
	}
	

	/**
	 * @return Returns the codePosition.
	 */
	public Position getCodePosition() {
		return this.getPosition();
	}

	/**
	 * @param codePosition The codePosition to set.
	 */
	public void setCodePosition(Position codePosition) {
		this.setPosition(codePosition);
	}

	/**
	 * @return Returns the context.
	 */
	public JStatementList getContext() {
		return this;
	}

	/**
	 * @param context The context to set.
	 */
	public void setContext(JStatementList context) {
		this.setStatements(context.getStatements());
		this.setPosition(context.getPosition());
	}

	/**
	 * @return Returns the interestCode.
	 */
	public JStatementList getCodeOfInterest() {
		return codeOfInterest;
	}

	/**
	 * @param interestCode The interestCode to set.
	 */
	public void setCodeOfInterest(JStatementList interestCode) {
		this.codeOfInterest = interestCode;
	}

}
