/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe abstrata representa um conjunto de declaraes, que pode ser de
 * atributos, de  mtodos, de contrutores ou de importaes. Ela concentra a
 * maior parte dos mtodos de <code>cin.jats.engine.parser.nodes.JaTSCollection</code>.
 * E serve, basicamente, para dar um tratamento uniforme aos conjuntos de
 * declaraes.
 * 
 * @author Adeline de Sousa
 * @version 27/12/2002
 */
public class DeclarationSet extends AbstractNode implements JaTSCollection {

    protected NodeList declarations = new NodeList();

    /**
     * Construtor vazio da classe
     */
    public DeclarationSet() {
        this.declarations = new NodeList();
    }


    public NodeList getDeclarationSet() {
        return this.declarations;
    }

    /**
     * Diz se a lista es vazia.
     *
     * @return Um boolean que, caso tenho o valor <code>true</code>, indica
     *         que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.declarations.isEmpty();
    }

    /**
     * Retorna a capacidade desta lista.
     *
     * @return Um inteiro correspondendo  capacidade da lista.
     */
    public int capacity() {
        return this.declarations.capacity();
    }

    /**
     * Retorna um objeto do tipo <code>JaTSIterator</code>
     *
     * @return um objeto do tipo JaTSIterator para que se possa percorrer esta
     *         lista de atributos
     */
    public JaTSIterator getIterator() {

        return this.declarations.getIterator();
    }


    /**
     * Remove o elemento que se encontra na  posio <code>pos</code> desta lista.
     *
     * @param pos Um inteiro correspondendo  posio do n que deve ser<br>
     *            removido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void removeElementAt(int pos) throws IllegalArgumentException {
        if (pos < 0) {
            throw new IllegalArgumentException();
        }
        this.declarations.removeElementAt(pos);
    }

    /**
     * Adiciona o elemento na posio <code>pos</code> desta lista e d um shift
     * para a direita de todos os elementos que se encontram a direita a partir
     * desta posio.
     *
     * @param elem Elemento a ser adicionado
     * @param pos  Um inteiro correspondendo  posio do n que deve ser
     *             removido.
     * @throws IllegalArgumentException Levantada quando o
     *                                  argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode elem, int pos) throws IllegalArgumentException {
        this.declarations.insertElementAt(elem, pos);
    }

    /**
     * Retorna o nmero de elementos da lista.
     *
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        return this.declarations.size();
    }

    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     *
     * @param pos Um inteiro correspondendo  posio desejada.
     * @return Um objeto do tipo <code>JFieldDeclaration</code> que se encontra na posio <br>
     *         desejada.
     */
    public INode elementAt(int pos) {
        return this.declarations.elementAt(pos);
    }

    /**
     * Retorna o ndice do elemento passado como parametro, ou -1 se este
     * elemento no pertencer ao conjunto
     *
     * @param node Elemento a ser pesquisado
     * @return O ndice do elemento, se ele existir no conjunto  ou  -1
     */
    public int indexOf(INode node) {
        return declarations.indexOf(node);
    }


    /**
     * Compara dois objetos do tipo DeclarationSet e diz se estes so idnticos,
     * ou seja, ou se se trata de uma varivel ou todos os elementos (em ordem)
     * so idnticos
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof DeclarationSet) {
            DeclarationSet decSet = (DeclarationSet) obj;
            if (decSet.isASourceNode()) {
                retorno = AbstractNode.equals(this.declarations, decSet.declarations);
            } else {
                retorno = compareJaTSNode(decSet);
            }
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        throw new UnsupportedOperationException("Accept not allowed");
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
    }

    /**
     * Adiciona um elemento  lista.
     *
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>node</code> for <code>null</code>.
     */
    public void add(INode node) throws IllegalArgumentException {
        if (node instanceof Declaration) {
            this.declarations.addElement(node);
        } else {
            throw new IllegalArgumentException("node isn't a Declaration");
        }
    }


}
