/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import java.util.Vector;

import cin.jats.engine.JatsTransformer;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.transformation.MicroTransformation;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.TransformationState;
import cin.jats.engine.visitors.IVisitor;

/**
 * @author Gustavo Santos
 * @version 19/10/2005
 *
 */
public class ExploreDeclaration extends AbstractNode {

	private INode baseNode;
	private NodeList whereClauses;
	
	public ExploreDeclaration() {
		this.whereClauses = new NodeList();
	}
	
	/**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
	public Object accept(IVisitor visitor, Object data)
			throws InconsistentNodeException, IllegalArgumentException,
			ExecutionException {
        Object retorno = this.acceptExploreDeclaration(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
	}

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
	private Object acceptExploreDeclaration(IVisitor visitor, Object data)
			throws InconsistentNodeException, IllegalArgumentException,
			ExecutionException {

		Object retorno = null;
		
		if (this.baseNode == null) {
			throw new InconsistentNodeException(this);
		}
		
		this.baseNode.accept(visitor, data);
		
		if (this.whereClauses == null) {
			throw new InconsistentNodeException(this);			
		}
		
        for (int i = this.whereClauses.size() - 1; i >= 0; i--) {
            INode node = this.whereClauses.elementAt(i);
            node.accept(visitor, data);
        }

		return retorno;
	}

	/* (non-Javadoc)
	 * @see cin.jats.engine.parser.nodes.INode#match(cin.jats.engine.parser.nodes.INode, cin.jats.engine.util.ResultSet)
	 */
	public void match(INode node, ResultSet results)
			throws IllegalArgumentException, InconsistentNodeException,
			NodesNotMatchedException {
        throw new NodesNotMatchedException("An ExploreDeclaration should " +
                "not be matched", this, node);

	}

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

    	WhereClause wc = null;
    	Vector transf = null;
    	JatsTransformer transformer = null;
    	TransformationState tState = null;
    	ResultSet rs = (ResultSet)data;
    	MicroTransformation microTemp = null;
    	for (int i = 0; i < this.whereClauses.size(); i++) {
    		wc = (WhereClause)this.whereClauses.elementAt(i);
			transf = new Vector();
    		for (int j = 0; j < wc.getTransformationNodes().size(); j++) {
    			microTemp = ((MicroTransformationNode)wc.getTransformationNodes().elementAt(j)).getTransformation();
    			microTemp.setType(wc.getClauseType());
    			transf.add(microTemp);
    		}
    		transformer = new JatsTransformer();
    		tState = new TransformationState(transf, (ResultSet)rs.clone());
    		transformer.transform(this.baseNode, tState);    		
    	}
    	
    	return this.baseNode;
    }
	
	public INode getBaseNode() {
		return baseNode;
	}

	public void setBaseNode(INode baseNode) {
		this.baseNode = baseNode;
	}

	public NodeList getWhereClauses() {
		return whereClauses;
	}

	public void setWhereClauses(NodeList whereClauses) {
		this.whereClauses = whereClauses;
	}
	
	public void addWhereClause(INode wc) {
	    this.whereClauses.addElement(wc);
	}

}
