/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * @author Adeline de Sousa
 */
public interface INode {

    // Tipo no definido previamente.
    int NO_TYPE = 0;

    /**
     * Retorna a declarao executvel correspondente ao n atual, caso ele seja
     * executvel.
     *
     * @return Um n do tipo <code>JExecutableDeclaration</code>. Esse n
     *         encapsula cdigo Java a ser executado durante o processamento dos
     *         ns.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada caso o n seja executvel
     *                                   mas no possua um n executvel associado.
     */
    INode getExecutableDeclaration() throws
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;

    /**
     * Retorna o nome da varivel que este n representa. Se o n no for uma
     * varivel, o mtodo pode retornar <code>null</code>.
     *
     * @return Um String correspondendo ao nome da varivel que este n
     *         representa.
     */
    String getVariableName();

    /**
     * Retorna o tipo da varivel.
     *
     * @return Um inteiro correspondendo ao tipo da varivel. O tipo de uma
     *         varivel JaTS corresponde ao tipo de construo sinttica de Java
     *         que a varivel representa. Exemplos possveis so um tipo, uma
     *         lista de nomes, um identificador, uma declarao de mtodo, entre
     *         outros.
     */
    int getVariableType();

    /**
     * Diz se um n  ou no executvel onde, por executvel, entende-se que um
     * dos filhos do n  uma expresso sujeita a execuo como se fosse um
     * trecho de um programa em Java.
     *
     * @return Um valor booleano que, caso seja <code>true</code> indica que o
     *         n  executvel.
     */
    boolean isExecutable();

    /**
     * Diz se um no e ou nao opcional.
     *
     * @return Um boolean que diz se o no e ou nao opcional para fins de
     *         casamento e substituicao.
     */
    boolean isOptional();

    /**
     * Este mtodo diz se um n  ou no uma varivel JaTS.
     *
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica
     *         que o n  uma varivel JaTS.
     */
    boolean isVariable();

    /**
     * Retorna um inteiro que representa o tipo do n. Os valores retornados
     * aqui so os das constantes contidas em JaTSNode.
     *
     * @return Um inteiro correspondente ao tipo do n
     */
    int getTypeIdentifier();

    /**
     * Seta o valor do inteiro que representa o tipo do n. Os valores que o
     * tipo de um n podem assumir so os das constantes contidas em JaTSNode.
     *
     * @param ti Um inteiro correspondente ao tipo do n
     */
    void setTypeIdentifier(int ti);

    /**
     * Determina que um n  executvel ou no.
     *
     * @param exec Um boolean que, caso tenho tenha o valor <code>true</code>,
     *             determina que o n executvel.
     */
    void setExecutable(boolean exec);

    /**
     * Seta o n executvel associado a este n. Se o n no era executvel antes
     * da invocao deste mtodo (ou seja, se o flag <code>executable</code>
     * tinha valor <code>false</code>), o mtodo torna o valor desse flag
     * <code>true</code>.
     *
     * @param dec A expresso executvel a ser associada a este n.
     */
    void setExecutableDeclaration(INode dec);

    /**
     * Faz de um no opcional, para fins de casamento e substituicao.
     *
     * @param opt Um boolean que diz se o no e opcional ou nao.
     */
    void setOptional(boolean opt);

    /**
     * Este mtodo determina se um n  ou no uma varivel.
     *
     * @param tf Um boolean que, caso tenha o valor <code>true</code>, determina
     *           que o n  uma varivel.
     */
    void setVariable(boolean tf);

    /**
     * Seta o nome da varivel que este n representa.
     *
     * @param varName Um String, correspondente ao nome da varivel.
     */
    void setVariableName(String varName);

    /**
     * Seta o tipo da varivel.
     *
     * @param type Um inteiro correspondendo a um dos possveis tipos de
     *             variveis. Os tipos correspondem s construes sintticas de
     *             Java, embora algumas construes no estejam diretamente
     *             representadas entre os tipos possveis (ou seja, no  possvel
     *             que uma varivel tenha o tipo correspondendo a qualquer construo
     *             Java, embora isso seja possvel com a maioria dos tipos).
     */
    void setVariableType(int type);

    /**
     * Retorna o nvel de indentao desse n, para que este possa ser usado
     * pelo prettyprinter no momento da impresso do n
     *
     * @return o nvel de aninhamento desse n
     */
//    boolean getIndentLevel();

    /**
     * Seta o nvel de indentao desse n, para que este possa ser usado
     * pelo prettyprinter no momento da impresso do n
     *
     * @param nLevel o nvel de aninhamento desse n
     */
//    void setIndentLevel(int nLevel);

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException;

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException;

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException;


    /**
     * Sets the 'parent' attribute of this node
     * @param parent
     */
    public void setParent(INode parent);
    
    /**
     * Retrieves this node's parent
     * 
     * @return
     */
    public INode getParent();
    
    
    
//*******************[ typechecking ]************************
//    public IType typeCheck(Environment env);

}
