/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;

/**
 * Classe que   guarda as informacoes a respeito das declaracoes
 * de  importao
 * Essa classe  formada basicamente por tres conjuntos : um pra
 * importaos, outro para declaracoes do tipo <b>importVar</b> e outra para
 * declaracoes do tipo <b>importVar</b>
 * 
 * @author Adeline de Sousa
 * @author Fernando Castor
 */
public class ImportDeclarations extends AbstractNode {

    // importaos do programa java
    private JImportDeclarationSet importsSet;

    // declaracao no programa jats que casa com um nica declarao de import
    // no programa java
    private NodeList importVarSet;

    // declaracao no programa jats que casa com um conjunto declaraes de
    // import no programa java
    private NodeList importsVarSet;


    public ImportDeclarations() {
        this.importsSet = new JImportDeclarationSet();
        this.importVarSet = new NodeList();
        this.importsVarSet = new NodeList();
    }

    /**
     * Adiciona uma declaracao de importao ao conjunto importsSet
     *
     * @param importDec Declaracao de importao a ser adicionada
     */
    public void addImportDeclaration(JImportDeclaration importDec)
            throws IllegalArgumentException {

        if (importDec == null) {
            throw new IllegalArgumentException("importDec nulo");
        }
        this.importsSet.add(importDec);
    }

    /**
     * Adiciona uma declaracao 'importsVar' ao conjunto importVarSet
     *
     * @param imports Declaracao 'importsVar' a ser adicionada
     */
    public void addImportsVarDeclaration(JImportDeclarationSet imports)
            throws IllegalArgumentException {

        if (imports != null && imports.isVariable() &&
                imports.getVariableType() == JaTSNode.IMPORT_DECLARATION_SET) {
            this.importsVarSet.addElement(imports);
        } else {
            throw new IllegalArgumentException("imports nulo");
        }
    }

    /**
     * Adiciona uma declaracao 'importVar' ao conjunto importVarSet
     *
     * @param idImport Declaracao 'importVar' a ser adicionada
     */
    public void addImportVarDeclaration(JImportDeclaration idImport)
            throws IllegalArgumentException {

        if (idImport != null && idImport.isVariable() &&
                idImport.getVariableType() == JaTSNode.IMPORT_DECLARATION) {

            this.importVarSet.addElement(idImport);
        } else {
            throw new IllegalArgumentException("idImport nulo");
        }
    }

    /**
     * Retorna o conjunto de construcoes do tipo importVar do programa jats
     *
     * @return Conjunto de construcoes do tipo importVar do programa jats
     */
    public NodeList getImportVarSet() {
        return this.importVarSet;

    }

    /**
     * Retorna o conjunto de construcoes do tipo importsVar do programa jats
     *
     * @return Conjunto de construcoes do tipo importsVar do programa jats
     */
    public NodeList getImportsVarSet() {
        return this.importsVarSet;

    }

    /**
     * Retorna o conjunto de importaos do programa java
     *
     * @return Conjunto de importaos do programa java
     */
    public JImportDeclarationSet getImportsSet() {
        return this.importsSet;
    }


    /**
     * Seta o conjunto de construcoes do tipo importVar do programa jats
     *
     * @param idImportSet Construcoes do tipo importVar do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setImportVarSet(NodeList idImportSet) {
        if (idImportSet != null) {
            Iterator it = idImportSet.iterator();
            while (it.hasNext()) {
                if (!(it.next() instanceof JImportDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }

            this.importVarSet = idImportSet;
        } else {
            throw new IllegalArgumentException("Conjunto importVar vazio");
        }

    }

    /**
     * Seta o conjunto de construcoes do tipo importsVar do programa jats
     *
     * @param importsSet Construcoes do tipo importsVar do programa jats
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setImportsVarSet(NodeList importsSet) {
        if (importsSet != null) {
            Iterator it = importsSet.iterator();
            while (it.hasNext()) {
                if (!(it.next() instanceof JImportDeclarationSet)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
            }
            this.importsVarSet = importsSet;
        } else {
            throw new IllegalArgumentException("Conjunto importsVar vazio");
        }

    }

    /**
     * Seta o conjunto de importaos do programa java
     *
     * @param imports Conjunto de importaos do programa java
     * @throws IllegalArgumentException - Levantada quando o conjunto passado
     *                                  for vazio  ou se houver algum elemento invlido no conjunto.
     */
    public void setImportsSet(JImportDeclarationSet imports) {
        if (imports != null) {
            int i = 0;
            int size = imports.size();
            while (i < size) {
                if (!(imports.elementAt(i) instanceof JImportDeclaration)) {
                    throw new IllegalArgumentException("Elemento invalido");
                }
                i++;
            }
            this.importsSet = imports;
        } else {
            throw new IllegalArgumentException("Conjunto de importaos vazio");
        }

    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (node == null || !(node instanceof ImportDeclarations)) {
            String nodeTypeStr = "null";
            if (node != null) nodeTypeStr = node.getClass().toString();
            throw new NodesNotMatchedException("Nodes have different types (" +
                    nodeTypeStr + ", " + this.getClass().toString() + ")",
                    this, node);
        }

        ImportDeclarations importDecSet = (ImportDeclarations) node;
        JImportDeclarationSet nodeImportDecs = (JImportDeclarationSet) importDecSet.getImportsSet().clone();
        if (nodeImportDecs == null) {
            throw new InconsistentNodeException(importDecSet);
        }
        int nodeImportDecsSize = nodeImportDecs.size();
        boolean achou = false;

        // Casamento das declaraes do tipo varivel:JImportDeclaration com os
        // importaos que ainda esto em nodeImportDecs. Primeiro so casadas as
        // variveis que J FORAM mapeadas a algum valor. Isso  necessrio para
        // garantir que nenhuma declarao no fonte Java  mapeada a mais de uma
        // varivel.
        int k = 0;
        int max3 = this.importVarSet.size();

        // Lista de variveis do programa JaTS que ainda no foram casados
        NodeList unmatchedImportDecVars = new NodeList();

        // Lista de variveis do tipo JImportDeclarationSet do programa JaTS que
        // ainda no foram casadas.
        NodeList unmatchedImportDecSetVars = new NodeList();

        // Com essa implementao padro,  necessrio verificar essa con-
        // dio para garantir que nenhuma varivel em this.importVarSet
        // ficar sem ser casada.
        if (max3 > nodeImportDecs.size()) {
            throw new NodesNotMatchedException("There is a different number of " +
                    "imports in each declaration (" + max3 + ", " +
                    nodeImportDecs.size() + ")", this, importDecSet);
        } else {
            while (k < max3) {
                JaTSNode matchesSingleImport = (JaTSNode) this.importVarSet.elementAt(k);
                if (!matchesSingleImport.isVariable() ||
                        matchesSingleImport.getVariableType() !=
                        JaTSNode.IMPORT_DECLARATION ||
                        !(matchesSingleImport instanceof JImportDeclaration)) {
                    throw new InconsistentNodeException(this);
                }
                // se a varivel j est casada, ento varre a lista de ns do
                // tipo JImportDeclaration at encontrar o valor que lhe foi
                // mapeado e o remove da lista, para que este no possa ser
                // mapeado a outra varivel
                if (results.alreadyMatched(matchesSingleImport)) {
                    int l = 0;
                    nodeImportDecsSize = nodeImportDecs.size();
                    achou = false;
                    Object tempNode = null;
                    tempNode = results.get(matchesSingleImport);


                    if (!(tempNode instanceof JImportDeclaration)) {
                        throw new IllegalMatchingException(matchesSingleImport);
                    }
                    while ((l < nodeImportDecsSize) && !achou) {
                        // Em nodeImportDecs s existem ns do tipo
                        // JImportDeclaration, por isso o cast sem instanceof
                        JImportDeclaration tmp =
                                (JImportDeclaration) nodeImportDecs.elementAt(l);
                        try {
                            // Se encontrou o valor, remove o n da lista de
                            // declaraes de importao do programa Java
                            tmp.match((JImportDeclaration) tempNode, results);
                            nodeImportDecs.removeElementAt(l);
                            achou = true;
                        } catch (NodesNotMatchedException e) {
                            l++;
                        }
                    }
                }
                // a varivel ainda no foi casada, adiciona-a  lista de
                // variveis que ainda no foram casadas
                else {
                    unmatchedImportDecVars.addElement(matchesSingleImport);
                }

                k++;
            }
        }

        // Casamento de variveis do tipo JImportDeclarationSet. Similar ao
        // anterior: apenas para remover da lista de declaraes de importao
        // do tipo fonte Java as declaraes que j esto casadas a variveis.
        // Nesse caso, a variveis do tipo JImportDeclarationSet.
        int importsVarSetSize = this.importsVarSet.size();
        int n = 0;
        INode tempImportsVar = null;

        while (n < importsVarSetSize) {
            tempImportsVar = this.importsVarSet.elementAt(n);
            if (!(tempImportsVar instanceof JImportDeclarationSet)) {
                throw new IllegalMatchingException(tempImportsVar);
            }
            JImportDeclarationSet importSetVar =
                    (JImportDeclarationSet) tempImportsVar;

            if (results.alreadyMatched(importSetVar)) {
                Object tempIds = null;
                tempIds = results.get(importSetVar);

                if (!(tempIds instanceof JImportDeclarationSet)) {
                    throw new IllegalMatchingException();
                }

                JImportDeclarationSet ids = (JImportDeclarationSet) tempIds;
                JaTSIterator ji = ids.getIterator();

                while (ji.hasNext()) {
                    INode importsVarDecSetTemp = ji.next();

                    int c1 = 0;
                    int maxIDS = nodeImportDecs.size();
                    boolean achou2 = false;

                    while (c1 < maxIDS && !achou2) {
                        JaTSNode fDecTemp = (JaTSNode) nodeImportDecs.elementAt(c1);
                        try {
                            importsVarDecSetTemp.match(fDecTemp, results);
                            nodeImportDecs.removeElementAt(c1);
                            achou2 = true;
                        } catch (NodesNotMatchedException nnme) {
                            c1++;
                        }
                    }
                }
            } else {
                unmatchedImportDecSetVars.addElement(importSetVar);
            }
            n++;
        }


        // Casamento padro de declaraes de importao no lado esquerdo da
        // transformao (this) com declaraes de importao no programa fonte
        // (contidas em nodeImportDecs).
        this.importsSet.match(nodeImportDecs, results);

        // Casamento das variveis do tipo JImportDeclaration que ainda no esto
        // mapeadas a nenhum valor.
        int m = 0;
        int max5 = unmatchedImportDecVars.size();
        if (max5 <= nodeImportDecs.size()) {
            while (m < max5) {
                JaTSNode tempVar = (JaTSNode) unmatchedImportDecVars.elementAt(m);
                if (!(tempVar instanceof JImportDeclaration)) {
                    throw new IllegalMatchingException(tempVar);
                }
                JaTSNode tempDec = (JaTSNode) nodeImportDecs.elementAt(0);
                // tenta casar os ns restantes
                tempVar.matchesAsAVariable(tempDec, results,
                        JaTSNode.IMPORT_DECLARATION);
                nodeImportDecs.removeElementAt(0);
                m++;
            }
        } else {
            // no pode restar variveis do tipo "JImportDeclaration" sem estar
            // casadas
            throw new NodesNotMatchedException("Some fields couldn't be matched " +
                    "in ImportDeclararions", this, node);
        }


        // Casamento entre as declaraes do tipo varivel:JImportDeclarationSet
        // e os importaos restantes em nodeImportDecs.
        int importsVarSize = unmatchedImportDecSetVars.size();
        if (importsVarSize > 0) {
            // Se houver apenas uma varivel do tipo JImportDeclarationSet, ento
            // esta ser casada com as declaraes de importao restante
            if (importsVarSize == 1) {
                INode matchesTheRest = unmatchedImportDecSetVars.elementAt(0);

                if (matchesTheRest == null || !matchesTheRest.isVariable() ||
                        matchesTheRest.getVariableType() !=
                        JaTSNode.IMPORT_DECLARATION_SET ||
                        !(matchesTheRest instanceof JImportDeclarationSet))
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);

                JImportDeclarationSet importDecsVar = (JImportDeclarationSet) matchesTheRest;

                importDecsVar.matchesAsAVariable(nodeImportDecs, results,
                        NO_TYPE);
            } else {
                // Se houver mais que uma varivel do tipo JImportDeclarationSet
                // no casada, o sistema no  capaz de decidir como proceder e
                // o casamento falha.
                JImportDeclarationSet unmatchedvarDecs = new JImportDeclarationSet(unmatchedImportDecSetVars);
                throw new NodesNotMatchedException("The system could't decide " +
                        "how to perform the matching. (" + importsVarSize + " VarSets)",
                        unmatchedvarDecs, nodeImportDecs);
            }
        } else if (nodeImportDecs.size() > 0) {
            throw new NodesNotMatchedException("Some imports could not be matched " +
                    "(number of imports: " + nodeImportDecs.size() + ")", this, importDecSet);
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.importsSet</code>, <code>this.importVarSet</code> e
     * <code>this.importsVarSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Variveis que casam com vrias declaraes de importaos
        if (this.importsVarSet != null) {
            int size = importsVarSet.size() - 1;

            while (size >= 0) {
                this.importsVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Variveis que casam com uma nica declarao de importao
        if (this.importVarSet != null) {
            int size = importVarSet.size() - 1;

            while (size >= 0) {
                this.importVarSet.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        // Declaraes de importao
        if (this.importsSet != null) {
            this.importsSet.accept(visitor, data);
        } else {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }

    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * ImportDeclarations se resume a invocar o mtodo <code>process</code> no
     * n correspondente  lista de declaraes de importaos filha do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Ns do tipo ImportDeclarations nunca podem ser executveis.

        INode iDecList = this.getImportsSet();
        if (iDecList != null) {
            this.processNode(iDecList, data);
        }

        return retorno;
    }
}
