/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa expresses do tipo <code>instanceof</code>.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class InstanceOfExpression extends JExpression {


    // Armazena o operando da expresso.
    private INode operand;

    //  Armazena o tipo com o qual o tipo do operando ser comparado.
    private JType type;


    /**
     * Construtor vazio da classe
     */
    public InstanceOfExpression() {
        this.operand = null;
        this.type = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetros o operando cujo tipo
     * ser verificado e o tipo propriamente dito.
     *
     * @param theOperand O operando cujo tipo ser comparado com o tipo
     *                   fornecido como argumento.
     * @param theType    O tipo com o qual o tipo do operando ser comparado.
     */
    public InstanceOfExpression(INode theOperand, JType theType) {
        this.operand = theOperand;
        this.type = theType;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.operand</code> e de <code>this.type</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Envie o visitor atravs da rvore ...

        if (this.operand == null || this.type == null)
            throw new InconsistentNodeException(this);

        this.operand.accept(visitor, data);
        this.type.accept(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Mtodo que retorna o operando da expresso.
     *
     * @return Uma objeto do tipo <code>INode</code> correspondendo ao
     *         operando da expresso.
     */
    public INode getOperand() {
        return this.operand;
    }

    /**
     * Mtodo que retorna o tipo com o qual o tipo do operando deveria ser
     * comparado.
     *
     * @return Uma objeto do tipo <code>JType</code>, correspondendo ao tipo que
     *         ser comparado ao tipo do operando.
     */
    public JType getType() {
        return this.type;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof InstanceOfExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            InstanceOfExpression exp = (InstanceOfExpression) node;

            if (this.getOperand() == null || this.getType() == null) {
                throw new InconsistentNodeException(this);
            }

            (this.getOperand()).match(exp.getOperand(), results);
            (this.getType()).match(exp.getType(), results);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do tipo para o qual o operando ser casted.
            INode cType = this.getType();
            if (cType != null) {
                Object ret1 = this.processNode(cType, data);

                if (cType.isExecutable()) {

                    try {
                        this.setType(JType.unwrapType(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
            // Processamento do operando  direita do operador.
            INode oper = this.getOperand();
            if (oper != null) {
                Object ret2 = this.processNode(oper, data);

                if (oper.isExecutable()) {

                    if (!(ret2 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setOperand((JExpression) ret2);
                }
            }
        }

        return retorno;
    }

    /**
     * Mtodo que seta o operando da expresso.
     *
     * @param newOperand Um objeto do tipo <code>INode</code> correspondendo
     *                   ao operando desta expresso.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>newOperand</code> for <code>null</code>.
     */
    public void setOperand(INode newOperand)
            throws IllegalArgumentException {

        if (newOperand == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.operand = newOperand;
    }

    /**
     * Mtodo que seta o tipo a ser comparado com o tipo do operando.
     *
     * @param newType Um objeto do tipo <code>JType</code>, correspondendo ao
     *                tipo que ser comparado com o tipo do operando.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>newType</code> for <code>null</code>.
     */
    public void setType(JType newType)
            throws IllegalArgumentException {

        if (newType == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.type = newType;
    }

    /**
     * Compara dois objetos do tipo <code>InstanceOfExpression</code> e diz se
     * estes so idnticos
     *
     * @param obj Objeto a ser comparado
     * @return true, se os objetos forem idnticos <br>
     *         false, caso contrrio
     */
    public boolean equals(Object obj) {

        boolean retorno = false;
        if (obj instanceof InstanceOfExpression) {
            InstanceOfExpression instExp = (InstanceOfExpression) obj;
            if (!instExp.isVariable()) {
                retorno = super.equals(instExp) &&
                        AbstractNode.equals(this.operand, instExp.getOperand()) &&
                        AbstractNode.equals(this.type, instExp.getType());
            } else {
                retorno = compareJaTSNode(instExp);
            }
        }
        return retorno;
    }

    public Object evaluate(Object data) throws ExecutionException {

        Boolean result = null;
        JType tmpType = this.getType();
        Object oper = this.getOperand();
        Class tipo = null;
        try {
            tipo = Class.forName(tmpType.getTypeName().toString());
        } catch (ClassNotFoundException e) {
            throw new ExecutionException(e.getMessage());
        }

        // seta o resultado com o valor da expresso isAssignableFrom
        result = new Boolean(tipo.isAssignableFrom(oper.getClass()));

        return result;
    }

}
