/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.CloningVisitor;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.visitors.ReplacementVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.IllegalMatchingException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe implementa declaraes iterativas em JaTS. Declaraes iterativas
 * so teis quando se deseja gerar colees de declaraes que diferem nas
 * informaes obtidas a partir da etapa de casamento. Assim como declaraes
 * executveis, declaraes iterativas tambm tm que ser processadas e,
 * consequentemente, tambm podem ser consideradas "executveis".
 *
 * @author Fernando Castor
 * @since JDK 1.3
 * @version 14/01/2003
 * @see INode
 */
public class IterativeDeclaration extends AbstractNode
        implements NestableExecutableDeclaration {

    // O n pai deste. Uma coleo (lista de nomes, lista de parmetros, lista
    // de declaraes no corpo de uma classe, etc.
    private INode parentNode;


    //  A varivel sobre a qual a iterao acontecer.  usada como placeholder
    // para os valores no iteration set.
    private INode placeholder;

    //  Lista de declaraes a partir da qual o resultado do processamento da
    //  declarao iterativa  gerado.
    private NodeList declarationSet;

    // Lista de declaraes baseada na qual a lista de declaraes  modificada e o
    //  resultado do processamento da declarao iterativa  gerado.
    private INode iterationSet;


    /**
     * Construtor sem parmetros.
     */
    public IterativeDeclaration() {
        this.parentNode = null;
        this.iterationSet = null;
        this.placeholder = null;
        this.declarationSet = new NodeList();
    }


    /**
     * Construtor que recebe como argumento os valores necessrios  criao
     * de uma declarao iterativa consistente.
     *
     * @param parNode Um <code>INode</code> correspondente ao n pai da
     *                declarao iterativa.
     * @param ph      Uma varivel JaTS a ser usada como placeholder para os
     *                valores do declaration set.
     * @param decSet  Uma lista de declaraes que serviro de esqueleto para o
     *                resultado gerado pelo processamento da declarao iterativa.
     * @param iterSet Uma lista de declaraes que ter suas informaes usadas
     *                na gerao do resultado do processamento da declarao iterativa.
     */
    public IterativeDeclaration(INode parNode, INode ph,
                                NodeList decSet, INode iterSet) {

        this.parentNode = parNode;
        this.placeholder = ph;
        this.declarationSet = decSet;
        this.iterationSet = iterSet;
    }


    /**
     * Retorna a lista de declaraes do "corpo" da declarao iterativa.
     *
     * @return Um conjunto de declaraes. Os tipos dos elementos do conjunto
     *         dependem do tipo da declarao iterativa.
     */
    public NodeList getDeclarationSet() {
        return this.declarationSet;
    }


    /**
     * Retorna o elemento que se encontra na posio <code>i</code> da lista de
     * declaraes do <code>iteration set</code> da declarao iterativa atual.
     *
     * @param i O ndice do elemento a ser retornado.
     * @return A declarao que se encontra na posio <code>i</code> no
     *         <i>iteration set</i>.
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>i</code> seja invlido.
     */
    private INode getElementOfIterationSet(int i)
            throws IllegalArgumentException {

        INode retorno = null;

        if (this.iterationSet != null
                && this.iterationSet instanceof JaTSCollection) {
            JaTSCollection col = (JaTSCollection) this.iterationSet;
            boolean validPosition = i >= 0 && i <= col.size();
            if (!validPosition) {
                throw new IllegalArgumentException();
            }
            retorno = col.elementAt(i);
        }

        return retorno;
    }


    /**
     * Retorna a lista de declaraes na qual o processamento da declarao
     * iterativa  baseado.
     *
     * @return Um conjunto de declaraes. O tipo dos elementos do conjunto
     *         depende do tipo de declarao iterativa.
     */
    public INode getIterationSet() {
        return this.iterationSet;
    }


    /**
     * Retorna o n pai da declarao iterativa.
     *
     * @return O pai deste n.
     */
    public INode getParentNode() {
        return this.parentNode;
    }


    /**
     * Retorna o placeholder da declarao iterativa.
     *
     * @return O placeholder da declarao iterativa.
     */
    public INode getPlaceholder() {
        return this.placeholder;
    }


    /**
     * Retorna o nmero de elementos no <i>iteration set</i> da declarao
     * iterativa.
     *
     * @return O nmero de elementos do <i>iteration set</i> da declarao
     *         iterativa.
     */
    private int getSizeOfIterationSet() {
        int retorno = 0;
        if ((this.iterationSet != null)
                && (this.iterationSet instanceof JaTSCollection)) {

            retorno = ((JaTSCollection) this.iterationSet).size();
        }
        return retorno;
    }


    /**
     * Seta a lista de declaraes do "corpo" da declarao iterativa.
     *
     * @param decSet Um conjunto de declaraes. Os tipos dos elementos do conjunto
     *               dependem do tipo da declarao iterativa.
     * @throws IllegalArgumentException Levantada se <code>decSet</code> for
     *                                  <code>null</code>.
     */
    public void setDeclarationSet(NodeList decSet) throws IllegalArgumentException {

        if (decSet == null) {
            throw new IllegalArgumentException();
        }
        this.declarationSet = decSet;
    }


    /**
     * Seta a lista de declaraes na qual o processamento da declarao
     * iterativa  baseado.
     *
     * @param iterSet Um conjunto de declaraes. O tipo dos elementos do conjunto
     *                depende do tipo da declarao iterativa.  importante notar que
     *                todos os elementos do iteration set tm que ser do mesmo tipo.
     * @throws IllegalArgumentException Levantada se <code>iterSet</code> for
     *                                  <code>null</code>.
     */
    public void setIterationSet(INode iterSet) throws IllegalArgumentException {

        if (iterSet == null) {
            throw new IllegalArgumentException();
        }
        this.iterationSet = iterSet;
    }


    /**
     * Seta o n pai da declaraoiterativa.
     *
     * @param parNode O pai deste n.
     * @throws IllegalArgumentException Levantada se <code>parNode</code> for
     *                                  <code>null</code>.
     */
    public void setParentNode(INode parNode) throws IllegalArgumentException {

        if (parNode == null) {
            throw new IllegalArgumentException();
        }
        this.parentNode = parNode;
    }


    /**
     * Seta o placeholder da declarao iterativa.
     *
     * @param ph O novo placeholder da declarao iterativa.
     * @throws IllegalArgumentException Levantada se <code>ph</code> for
     *                                  <code>null</code>.
     */
    public void setPlaceholder(INode ph) throws IllegalArgumentException {

        if (ph == null) {
            throw new IllegalArgumentException();
        }
        this.placeholder = ph;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JName</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JName</code>.
     *                Por consequncia, objetos do tipo<code>JName</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException("An IterativeDeclaration should " +
                    "not be matched", this, node);
        }

    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = this.acceptIterativeDeclaration(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    private Object acceptIterativeDeclaration(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.placeholder == null) {
            throw new InconsistentNodeException(this);
        }

        this.placeholder.accept(visitor, data);

        if (this.declarationSet == null) {
            throw new InconsistentNodeException(this);
        }

        // invoca o accept sobre cada uma das declaraes presentes no forall
        for (int i = this.declarationSet.size() - 1; i >= 0; i--) {
            INode node = this.declarationSet.elementAt(i);
            node.accept(visitor, data);
        }
        if (this.iterationSet == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
        }
        this.iterationSet.accept(visitor, data);
        return retorno;
    }


    /**
     * @param data Um objeto que pode conter informaes necessrias ao
     *             processamento.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        if (data == null || !(data instanceof ResultSet)) {
            throw new IllegalArgumentException();
        }
        ResultSet rset = (ResultSet) data;

        if (this.getIterationSet() == null || this.getPlaceholder() == null
                || this.getParentNode() == null || this.getDeclarationSet() == null) {
            throw new InconsistentNodeException(this);
        }
        if (rset.alreadyMatched(this.getPlaceholder())) {
            throw new IllegalMatchingException();
        }
        ReplacementVisitor rvis = new ReplacementVisitor(rset);

        // verifica se o iterationSet  executvel, se for, pega-se o resultado
        // de seu processamento, o qual deve ser um DeclarationSet e ento
        // realiza o processamento do n como um todo.
        if (iterationSet instanceof JExecutableDeclaration) {
            Object ret = this.processNode(iterationSet, data);
            if (ret instanceof JaTSCollection) {
                this.iterationSet = (JaTSCollection) ret;
            }
        }

        int max = this.getSizeOfIterationSet();
        int i = 0;
        int decSetSize = this.declarationSet.size();
        INode ph = this.getPlaceholder();
        NodeList declarations = new NodeList();

        while (i < max) {
            int j = 0;
            INode currentElementOfDecSet = null;
            INode valueOfPlaceholder = this.getElementOfIterationSet(i);
            rset.put(ph, valueOfPlaceholder);

            while (j < decSetSize) {
                CloningVisitor cvis = new CloningVisitor();
                INode tempDecSetElement = this.declarationSet.elementAt(j);
                INode result = null;
                Object tempResult = null;
                tempDecSetElement.accept(cvis, null);
                currentElementOfDecSet = cvis.getClone();
                rvis.setResults(rset);

                if (currentElementOfDecSet instanceof NestableExecutableDeclaration) {
                    ((NestableExecutableDeclaration) currentElementOfDecSet).setParentNode(this);
                }
                /**************/

                WrapperNode wnode = new WrapperNode();
                wnode.setNode(currentElementOfDecSet);
                wnode.accept(rvis, null);
                currentElementOfDecSet = wnode.getNode();

                /**************/

                tempResult = currentElementOfDecSet.process(rset);
                if (tempResult instanceof INode) {
                    result = (INode) tempResult;
                    if (!(result instanceof NestableExecutableDeclaration)) {

                        // NestableExecutableDeclarations no so passadas diretamente
                        // para o n pai. Apenas o resultado do seu processamento .
                        declarations.addElement(result);
                    }
                } else if (tempResult instanceof NodeList) {
                    declarations.addAll((NodeList) tempResult);
                }
                j++;
            }
            rset.removeResult(ph);
            i++;
        }

        this.parentNode.setExecutable(false);

        return declarations;
    }
}
