/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.EvaluateFunctions;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.lang.reflect.Constructor;


/**
 * Classe que implementa as expresses que envolvem o operador de alocao
 * dinmica <code>new</code>.
 *
 * @author Fernando Castor
 * @see JExpression
 * @see JName
 * @see JExpressionList
 * @see JArrayDimsAndInits
 */
public class JAllocationExpression extends JExpression {

    // objeto referenciado. Usado nos casos de instanciao de inner classes
    private INode referencedObject;

    // Responsvel por armazenar o nome do tipo do objeto que est sendo criado.
    private JName typeName;

    // Armazena o tipo de alocao que est sendo realizada. Dois tipos de
    // alocao so possveis: alocao de um array de um tipo primitivo ou
    // um objeto/array de objetos. No segundo caso, tambm  possvel estar
    // criando um objeto que seja uma instncia de uma classe interna.
    // Conseqentemente,  necessrio guardar o corpo dessa classe.
    private int allocationType;

    // Armazena a lista de argumentos que  passada para o construtor da classe
    // que est sendo instanciada.
    private JExpressionList arguments;

    // Caso um array esteja sendo alocado, este atributo  responsvel por
    // guardar o nmero de dimenses do array e a sua inicializao.
    private JArrayDimsAndInits arrayDimsAndInits;

    // Se o objeto que est sendo criado for uma instncia de uma classe
    // interna, o corpo dessa classe precisa ser guardado. Isso  feito por
    // meio deste atributo.
    private TypeBody internalClassBody;


    // Constantes inteiras correspondendo aos tipos de alocao possveis.

    // Alocao de uma matriz com elementos pertencendo a um tipo primitivo.
    public final static int PRIMITIVETYPEARRAYALLOCATION = 1;

    // Alocao de um objeto.
    public final static int OBJECTALLOCATION = 2;

    // Alocao de uma matriz de objetos.
    public final static int OBJECTARRAYALLOCATION = 3;


    /**
     * Construtor vazio da classe
     */
    public JAllocationExpression() {
        this.typeName = null;
        this.allocationType = 0;
        this.arguments = null;
        this.arrayDimsAndInits = null;
        this.internalClassBody = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro o nome do tipo de objeto
     * que est sendo inicializado, o tipo de alocao, os argumentos passados
     * para o construtor (caso a alocao seja de um objeto) e o nmero de
     * dimenses, caso uma matriz esteja sendo alocada.
     *
     * @param tName O um objeto do tipo <code>JName</code>, correspondendo
     *              ao nome do tipo.
     * @param aType O tipo da alocao que est sendo realizada.
     * @param args  A lista de argumentos que  passada para o construtor da
     *              classe, caso esteja sendo realizada a alocao de um objeto.
     * @param adis  Quando uma matriz est sendo alocada (seja de um tipo
     *              primitivo ou objeto), este objeto contm informaes como as
     *              dimenses da matriz, o tamanho de cada dimenso e a
     *              inicializao da matriz.
     */
    public JAllocationExpression(JName tName, int aType, JExpressionList args,
                                 JArrayDimsAndInits adis) {
        this.typeName = tName;
        this.allocationType = aType;
        this.arguments = args;
        this.arrayDimsAndInits = adis;
        this.internalClassBody = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de seus
     * filhos.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.referencedObject != null) {
            this.referencedObject.accept(visitor, data);
        }

        // Envie o visitor atravs da rvore ...
        if (this.typeName == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);

        this.typeName.accept(visitor, data);

        if (this.getAllocationType() == OBJECTALLOCATION) {
            if (this.arguments == null)
                throw new InconsistentNodeException(this);

            this.arguments.accept(visitor, data);

            if (this.getInternalClassBody() != null) {
                this.getInternalClassBody().accept(visitor, data);

            }
        } else if (this.getAllocationType() == PRIMITIVETYPEARRAYALLOCATION
                || this.getAllocationType() == OBJECTARRAYALLOCATION) {

            if (this.arrayDimsAndInits == null)
                throw new InconsistentNodeException(this);

            this.arrayDimsAndInits.accept(visitor, data);


        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Retorna o tipo de alocao que est sendo realizado.
     *
     * @return Um inteiro correspondendo a um dos tipos de alocao possveis.
     */
    public int getAllocationType() {
        return this.allocationType;
    }

    /**
     * Retorna a lista de argumentos passados para o construtor da classe na
     * alocao de objetos.
     *
     * @return A lista de argumentos.
     */
    public JExpressionList getArguments() {
        return this.arguments;
    }

    /**
     * Este mtodo retorna um objeto contendo informaes que concernem 
     * inicializao de uma matriz. Entre essas informaes incluem-se o nmero
     * de dimenses da matriz, o tamanho de cada dimenso e a inicializao da
     * matriz.
     *
     * @return Um objeto do tipo <code>ArrayDimsAndInitis</code> contendo as
     *         informaes descritas acima.
     */
    public JArrayDimsAndInits getArrayDimsAndInits() {
        return this.arrayDimsAndInits;
    }

    /**
     * Retorna o nome do tipo de objeto que est sendo alocado.
     *
     * @return Um objeto do tipo <code>JName</code>, correspondente ao nome do
     *         tipo de objeto que est sendo alocado.
     */
    public JName getTypeName() {
        return this.typeName;
    }


    public INode getReferencedObject() {
        return referencedObject;
    }

    public void setReferencedObject(INode refObject) {

        this.referencedObject = refObject;
    }


    public TypeBody getInternalClassBody() {
        return internalClassBody;
    }

    public void setInternalClassBody(TypeBody internalClassBody) {
        this.internalClassBody = internalClassBody;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JAllocationExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JAllocationExpression exp = (JAllocationExpression) node;


            if (this.getReferencedObject() != null) {
                this.getReferencedObject().match(exp.getReferencedObject(), results);
            }

            if (this.getTypeName() == null)
                throw new InconsistentNodeException(this);

            this.getTypeName().match(exp.getTypeName(), results);

            if (this.getAllocationType() == OBJECTALLOCATION) {

                if (this.getArguments() == null)
                    throw new InconsistentNodeException(this);

                (this.getArguments()).match(exp.getArguments(), results);
            } else if (this.getAllocationType() == PRIMITIVETYPEARRAYALLOCATION
                    || this.getAllocationType() == OBJECTARRAYALLOCATION) {

                if (this.getArrayDimsAndInits() == null)
                    throw new InconsistentNodeException(this);

                (this.getArrayDimsAndInits()).match(exp.getArrayDimsAndInits(),
                        results);

                if (this.internalClassBody != null) {
                    this.getInternalClassBody().match(exp.getInternalClassBody(), results);

                }
            }
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento da lista de argumentos do n.
            if (this.getAllocationType() == OBJECTALLOCATION) {
                INode args = this.getArguments();
                if (args != null) {
                    Object ret1 = this.processNode(args, data);
                    if (args.isExecutable()) {
                        if (!(ret1 instanceof JExpressionList)) {
                            throw new ExecutionException(this);
                        }
                        this.setArguments((JExpressionList) ret1);
                    }
                }
            }

            // Processamento do nome do tipo de objeto a ser criado.
            INode tName = this.getTypeName();
            if (tName != null) {
                Object ret2 = this.processNode(tName, data);
                if (tName.isExecutable()) {
                    try {
                        this.setName(JName.unwrapName(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento da inicializao do array e das expresses
            // correspondentes ao nmero de dimenses e ao tamanho destas.
            if (this.getAllocationType() == PRIMITIVETYPEARRAYALLOCATION
                    || this.getAllocationType() == OBJECTARRAYALLOCATION) {
                INode di = this.getArrayDimsAndInits();
                if (di != null) {
                    Object ret3 = this.processNode(di, data);

                    if (di.isExecutable()) {
                        if (!(ret3 instanceof JArrayDimsAndInits)) {
                            throw new ExecutionException(this);
                        }
                        this.setArrayDimsAndInits((JArrayDimsAndInits) ret3);
                    }
                }
                INode body = this.getInternalClassBody();
                if (body != null) {
                    Object ret4 = this.processNode(body, data);
                    if (body.isExecutable()) {
                        if (!(ret4 instanceof TypeBody)) {
                            throw new ExecutionException(this);
                        }
                        this.setInternalClassBody((TypeBody) ret4);
                    }
                }
            }
        }

        return retorno;
    }

    /**
     * Retorna o tipo de alocao que est sendo realizado.
     *
     * @param aType Um inteiro correspondendo a um dos tipos de alocao
     *              possveis.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          parmetro <code>aType</code> assumir um valor menor ou igual
     *          a zero.
     */
    public void setAllocationType(int aType) throws IllegalArgumentException {
        if (aType <= 0) {
            throw new IllegalArgumentException();
        }

        this.allocationType = aType;
    }

    /**
     * Seta a lista de argumentos que ser passada para o construtor da classe
     * na alocao de um objeto.
     *
     * @param args A lista de argumentos.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          parmetro <code>args</code> for <code>null</code>.
     */
    public void setArguments(JExpressionList args)
            throws IllegalArgumentException {

        if (args == null) {
            throw new IllegalArgumentException();
        }
        this.arguments = args;
    }

    /**
     * Seta o objeto que contm as informaes necessrias  inicializao de
     * uma matriz.
     *
     * @param adi Um objeto do tipo <code>ArrayDimsAndInitis</code> contendo as
     *            informaes descritas no mtodo <code>getArrayDimsAndInits</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          parmetro <code>adi</code> for <code>null</code>.
     */
    public void setArrayDimsAndInits(JArrayDimsAndInits adi)
            throws IllegalArgumentException {

        if (adi == null) {
            throw new IllegalArgumentException();
        }
        this.arrayDimsAndInits = adi;
    }

    /**
     * Retorna o nome do tipo de objeto que est sendo alocado.
     *
     * @param nome Um objeto do tipo <code>JName</code>, correspondente ao nome do
     *             tipo de objeto sendo alocado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          parmetro <code>nome</code> for <code>null</code>.
     */
    public void setName(JName nome) throws IllegalArgumentException {
        if (nome == null) {
            throw new IllegalArgumentException();
        }

        this.typeName = nome;
    }

    /**
     * Compara dois objetos do tipo JAllocationExpression e diz se estes so iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JAllocationExpression) {
            JAllocationExpression allocExp = (JAllocationExpression) obj;
            retorno = this.allocationType == allocExp.allocationType &&
                    AbstractNode.equals(this.typeName, allocExp.getTypeName()) &&
                    AbstractNode.equals(this.arguments, allocExp.getArguments()) &&
                    AbstractNode.equals(this.arrayDimsAndInits, allocExp.getArrayDimsAndInits()) &&
                    AbstractNode.equals(this.internalClassBody, allocExp.getInternalClassBody());


        }
        return retorno;

    }


    public Object evaluate(Object data) throws ExecutionException {
        Object result = this;
        if (this.getAllocationType() == JAllocationExpression.OBJECTALLOCATION) {

            // Falta a parte do corpo da classe, para classes anonimas!!
            Object[] argList = this.arguments.toObjectArray();

            String typeNameString = this.typeName.toString();

            // Criao do novo objeto propriamente dito.
            Class classe = null;
            try {
                classe = Class.forName(typeNameString);
            } catch (ClassNotFoundException e) {
                throw new ExecutionException(e.getMessage(), this);
            }

            Class[] argTypes = new Class[argList.length];
            int i = 0;
            while (i < argTypes.length) {
                argTypes[i] = (argList[i]).getClass();
                i++;
            }

            Constructor[] constructors = classe.getConstructors();
            int j = 0;
            boolean foundIt = false;
            while (j < constructors.length && !foundIt) {
                if ((constructors[j].getParameterTypes()).length ==
                        argTypes.length) {
                    Class[] parTypes = constructors[j].getParameterTypes();
                    int k = 0;
                    boolean ok = true;
                    while (k < argTypes.length && ok) {
                        if (!EvaluateFunctions.typesAreCompatible(parTypes[k],
                                argTypes[k])) {
                            ok = false;
                        }
                        k++;
                    }
                    if (ok) {
                        foundIt = true;
                    }
                }
                if (!foundIt) {
                    j++;
                }
            }
            if (!foundIt) {
                throw new ExecutionException(this);
            }
            try {
                result = constructors[j].newInstance(argList);
            } catch (Exception e) {
                throw new ExecutionException(e.getMessage(), this);
            }
        } else if (this.getAllocationType() ==
                JAllocationExpression.PRIMITIVETYPEARRAYALLOCATION
                || this.getAllocationType() ==
                JAllocationExpression.OBJECTARRAYALLOCATION) {

            // A SER FEITO NO FUTURO
        }

        return result;
    }
}
