/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

public class JAnonymousClass extends AbstractNode {

    private JName name;

    private JExpressionList arguments;

    private TypeBody body;


    /**
     * Construtor vazio da classe
     */
    public JAnonymousClass() {
    }

    public JName getName() {
        return name;
    }

    public void setName(JName name) {
        this.name = name;
    }

    public JExpressionList getArguments() {
        return arguments;
    }

    public void setArguments(JExpressionList arguments) {
        if (arguments != null) {
            JaTSIterator it = arguments.getIterator();
            while (it.hasNext()) {
                if (!(it.next() instanceof JExpression)) {
                    throw new IllegalArgumentException("");
                }
            }
            this.arguments = arguments;
        }
    }

    public TypeBody getBody() {
        return body;
    }

    public void setBody(TypeBody body) {
        this.body = body;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        Object retorno = null;

        if (this.isASourceNode()) {
            if (this.getBody() != null) {
                this.getBody().accept(visitor, data);
            }
            if (this.getArguments() != null) {
                this.getArguments().accept(visitor, data);
            }
            if (this.getName() != null) {
                this.getName().accept(visitor, data);
            }
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (!(node instanceof JAnonymousClass)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {
            JAnonymousClass anonymous = (JAnonymousClass) node;
            if (this.name != null) {
                // casa os nomes
                this.name.match(anonymous.getName(), results);
            }
            if (this.arguments != null) {
                this.arguments.match(anonymous.getArguments(), results);
            }
            if (this.body != null) {
                this.body.match(anonymous.getBody(), results);
            }
        }
    }

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do nome do tipo.
            INode tName = this.getName();
            if (tName != null) {
                Object ret = this.processNode(tName, data);
                if (tName.isExecutable()) {
                    try {
                        this.setName(JName.unwrapName(ret));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento da lista de argumentos
            JExpressionList tmpArguments = this.getArguments();
            int j = 0;
            if (tmpArguments != null) {
                int max1 = tmpArguments.size();
                while (j < max1) {
                    INode currExpr = tmpArguments.getExpressionAt(j);
                    if (currExpr != null) {
                        Object ret = this.processNode(currExpr, data);

                        if (currExpr.isExecutable()) {
                            if (!(ret instanceof JExpression))
                                throw new ExecutionException(this);

                            this.arguments.setExpressionAt((JExpression)ret, j);
                        }
                    }
                    j++;
                }
            }

            // Processamento do corpo da classe ou interface.
            INode body = this.getBody();
            if (body != null) {
                Object ret = this.processNode(body, data);
                if (body.isExecutable()) {
                    if (!(ret instanceof TypeBody)) {
                        throw new ExecutionException(this);
                    }
                    this.setBody((TypeBody) ret);
                }
            }
        }
        return retorno;
    }
}
