/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe representa um acesso a elementos de um array.  composta pelo
 * nome do array e pela lista de expresses de acesso(caso o array seja
 * multi-dimensional).
 *
 * @autor Adeline Sousa
 *
 * @version 01/07/2004
 */
public class JArrayAccess extends JExpression {

    // lista de expresses de referncia. Indicam qual o elemento do array ser
    // acessado
    private JExpressionList indexers;

    // O nome do array sendo acessado.
    private INode arrayName;


    /**
     * Construtor vazio da classe
     */
    public JArrayAccess() {
        this.indexers = null;
        this.arrayName = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }


    /**
     * Construtor vazio da classe
     */
    public JArrayAccess(JaTSNode arrayName, JExpressionList expList) {
        this.setArrayName(arrayName);
        this.setIndexers(expList);
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }


    /**
     * Retorna a expressao que representa o array sendo acessado
     *
     * @return Expressao que representa o array sendo acessado
     */
    public INode getArrayName() {
        return this.arrayName;
    }

    /**
     * Retorna o objeto a partir do qual o atributo ser acessado.
     *
     * @return Um <code>JaTSNode</code> correspondente ao objeto a partir
     *         do qual o atributo ser acessado ou  expresso que, uma vez
     *         avaliada, produzir esse objeto como resultado.
     */
    public JExpressionList getIndexers() {
        return this.indexers;
    }


    /**
     * Seta o nome do atributo sendo acessado.
     *
     * @param name Expressao que representa o array sendo acessado
     */
    public void setArrayName(INode name) {
        this.arrayName = name;
    }

    /**
     * Seta o objeto a partir do qual o atributo ser acessado.
     *
     * @param expressions Um <code>JaTSNode</code> correspondente ao objeto a partir
     *                    do qual o atributo ser acessado. <code>expressions</code> pode
     *                    corresponder tambm a uma expresso que ser avaliada. Nesse caso
     *                    o atributo  invocado a partir do objeto resultante dessa avaliao.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>expressions</code> for <code>null</code>.
     */
    public void setIndexers(JExpressionList expressions)
            throws IllegalArgumentException {
        if (expressions == null) {
            throw new IllegalArgumentException();
        }
        this.indexers = expressions;
    }

    /**
     * Compara dois objetos do tipo JFieldAccess e diz se os dois so iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JArrayAccess) {
            JArrayAccess arrayAcc = (JArrayAccess) obj;
            if (this.isASourceNode()) {
                retorno =
                        AbstractNode.equals(this.indexers, arrayAcc.getIndexers()) &&
                        AbstractNode.equals(this.arrayName, arrayAcc.getArrayName());

            } else {
                retorno = compareJaTSNode(arrayAcc);
            }
        }
        return retorno;

    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de seus
     * filhos.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {

            if (this.indexers != null) {
                this.indexers.accept(visitor, data);
            }

            if (this.arrayName == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }
            this.arrayName.accept(visitor, data);
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);
        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada quando algum componente
     *                                   essencial do n tiver um valor invlido, tornando o n, como um
     *                                   todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JArrayAccess)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JArrayAccess array = (JArrayAccess) node;

            if (this.getArrayName() == null) {
                throw new InconsistentNodeException(this);
            }

            if (this.getIndexers() != null) {
                this.indexers.match(array.getIndexers(), results);
            } else if (array.getIndexers() != null) {
                throw new NodesNotMatchedException("A reference expression is null",
                        this, node);
            }

            this.getArrayName().match(array.getArrayName(), results);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Levantada
     *                                   caso o parmetro <code>data</code> seja <code>null</code> numa
     *                                   situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada se um n inconsistente (incompleto) for encontrado.
     *                                   Se algum dos componentes (filhos) essenciais ao correto
     *                                   funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do nome do mtodo a ser invocado.
            INode fName = this.getArrayName();
            if (fName != null) {
                Object ret1 = this.processNode(fName, data);
                if (fName.isExecutable()) {
                    try {
                        this.setArrayName(JName.unwrapName(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }


            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            if (this.indexers != null) {
                int max = this.indexers.size();

                while (i < max) {
                    INode currExpr = this.indexers.getExpressionAt(i);
                    if (currExpr != null) {
                        Object ret = this.processNode(currExpr, data);
                        if (currExpr.isExecutable() && ret instanceof JExpression) {
                            this.indexers.setExpressionAt((JExpression) ret, i);
                        }
                    }
                    i++;
                }
            }


        }

        return retorno;
    }

}
