/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe responsvel por armazenar as dimenses de inicializaes na alocao
 * de uma matriz.
 *
 * @author Fernando Castor
 * @see JAllocationExpression
 */
public class JArrayDimsAndInits extends AbstractNode {

    /**
     * Responsvel por armazenar o nmero de dimenses da matriz.
     */
    private int dimensions;

    /**
     * Armazena uma lista de expresses que correspondem aos tamanhos das
     * dimenses. Por exemplo, para uma matriz da forma <code>int[3+5][4][]</code>, este
     * vetor teria armazenadas as expresses <code>3+5</code> e <code>4</code>.
     */
    private NodeList dimensionSizes;

    /**
     * Armazena a expresso de inicializao da matriz.
     */
    private JArrayInitializer init;


    /**
     * Construtor vazio da classe
     */
    public JArrayDimsAndInits() {
        this.dimensions = 0;
        this.dimensionSizes = new NodeList();
        this.init = null;
    }

    /**
     * Construtor da classe que recebe como parmetro o nome do tipo de objeto
     * que est sendo inicializado e o tipo de alocao.
     *
     * @param dims      Inteiro indicando o nmero de dimenses da matriz.
     * @param dimSizes  Um <code>NodeList</code> contendo os tamanhos de algumas
     *                  das dimenses.
     * @param arrayInit A inicializao da matriz.
     */
    public JArrayDimsAndInits(int dims, NodeList dimSizes,
                              JArrayInitializer arrayInit)
            throws IllegalArgumentException {

        this.dimensions = dims;

        if (dimSizes != null) {
            for (int i = 0; i < dimSizes.size(); i++) {
                // Os tamanhos das dimenses devem todos ser expresses.
                if (!(dimSizes.elementAt(i) instanceof JExpression))
                    throw new IllegalArgumentException();
            }
        }

        this.dimensionSizes = dimSizes;
        this.init = arrayInit;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * cada elemento de <code>this.dimensionSizes</code> e o de <code>this.init</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        int i = 0;

        // Envie o visitor atravs da rvore ...
        while (i < this.dimensionSizes.size()) {
            (this.dimensionSizes.elementAt(i)).accept(visitor, data);
            i++;
        }

        if (this.init != null)
            this.init.accept(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Adiciona uma nova expresso correspondendo ao tamanho de uma das
     * dimenses da matriz  lista de tamanhos de dimenses <code>dimensionSizes</code>.
     *
     * @param size Uma expresso indicando o tamanho que a dimenso ter.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          <code>dims</code> corresponder a uma dimenso invlida ou o parmetro .
     *          <code>size</code> for <code>null</code>.
     */
    public void addSizeOfDimension(JExpression size)
            throws IllegalArgumentException {

        if (size == null) {
            throw new IllegalArgumentException();
        }
        this.dimensionSizes.addElement(size);
    }

    /**
     * Retorna a(s) expresso(es) de inicializao da matriz.
     *
     * @return Um objeto do tipo <code>JArrayInitializer</code>, responsvel
     *         pela inicializao da matriz.
     */
    public JArrayInitializer getArrayInitializer() {
        return init;
    }

    /**
     * Retorna o nmero de dimenses com o qual o array foi declarado.
     *
     * @return Um inteiro correspondendo ao nmero de dimenses com o qual a
     *         matriz foi declarada.
     */
    public int getDimensions() {
        return this.dimensions;
    }

    /**
     * Retorna o nmero de dimenses que tiverem o tamanho especificado.
     *
     * @return Um inteiro correspondendo ao nmero de dimenses cujo tamanho foi
     *         explicitamente especificado.
     */
    public int getNumberOfDimensionSizesSpecified() {
        return dimensionSizes.size();
    }

    /**
     * Retorna o tamanho associado  dimenso pedida.
     *
     * @param dim Dimenso da qual se deseja saber o tamanho. Como matrizes
     *            tm pelo menos uma dimenso, este parmetro tem que receber valor
     *            pelo menos 1.
     * @return Uma expresso, correspondendo ao tamanho da dimenso <code>dim</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>dim</code> for um inteiro menor ou igual a zero.
     */
    public JExpression getSizeOfDimension(int dim)
            throws IllegalArgumentException, InconsistentNodeException {

        if (dim <= 0 || dim > dimensionSizes.size()) {
            throw new IllegalArgumentException();
        }
        if (!(dimensionSizes.elementAt(dim - 1) instanceof JExpression)) {
            throw new InconsistentNodeException(this);
        }
        return (JExpression) dimensionSizes.elementAt(dim - 1);
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JArrayDimsAndInits)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JArrayDimsAndInits adi = (JArrayDimsAndInits) node;
            int i = 1;
            int max = adi.getNumberOfDimensionSizesSpecified();

            if (this.getDimensions() != adi.getDimensions() ||
                    this.getNumberOfDimensionSizesSpecified() != max) {
                throw new NodesNotMatchedException("Arrays have different dimensions (" +
                        this.getDimensions() + ", " + adi.getDimensions() + ")",
                        this, node);
            }
            while (i <= max) {
                (this.getSizeOfDimension(i)).match(adi.getSizeOfDimension(i),
                        results);
                i++;
            }
            if (this.getArrayInitializer() != null) {
                (this.getArrayInitializer()).match(adi.getArrayInitializer(),
                        results);
            }
        }
    }

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {
        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento de cada uma das expresses de dimenso.
            int i = 0;
            int max = this.dimensionSizes.size();

            while (i < max) {
                INode currDimSize = this.getSizeOfDimension(i + 1);
                if (currDimSize != null) {
                    Object ret = this.processNode(currDimSize, data);
                    if (currDimSize.isExecutable()) {
                        if (!(ret instanceof JExpression))
                            throw new ExecutionException(this);

                        this.setSizeOfDimension((JExpression) ret, i + 1);
                    }
                }
                i++;
            }

            // Processamento da inicializao do array.
            INode initializer = this.getArrayInitializer();
            if (initializer != null) {
                Object ret = this.processNode(initializer, data);
                if (initializer.isExecutable()) {
                    if (!(ret instanceof JArrayInitializer))
                        throw new ExecutionException(this);

                    this.setArrayInitializer((JArrayInitializer) ret);
                }
            }
        }

        return retorno;
    }

    /**
     * Seta a inicializao da matriz.
     *
     * @param initializer Um objeto correspondendo  inicializao da matriz.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          <code>initializer</code> for <code>null</code>.
     */
    public void setArrayInitializer(JArrayInitializer initializer)
            throws IllegalArgumentException {
        if (initializer == null) throw new IllegalArgumentException(); //Texto da Exceo

        this.init = initializer;
    }

    /**
     * Seta o nmero de dimenses da matriz.
     *
     * @param dims Um inteiro correspondendo ao nmero de dimenses da matriz.
     */
    public void setDimensions(int dims) {
        this.dimensions = dims;
    }

    /**
     * Seta o tamanho de uma das dimenses da matriz.
     *
     * @param size Uma expresso indicando o novo tamanho que a dimenso ter.
     * @param dim  A dimenso que ter seu tamanho modificado para <code>size</code>.
     *             Assim como no mtodo <code>getSizeOfDimension</code>, o menor valor
     *             para esse parmetro  1.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          <code>dims</code> corresponder a uma dimenso invlida ou o parmetro .
     *          <code>size</code> for <code>null</code>.
     */
    public void setSizeOfDimension(JExpression size, int dim)
            throws IllegalArgumentException {

        if (dim <= 0 || dim > dimensions || size == null)
            throw new IllegalArgumentException(); // Texto da exceo.
        if (dim == this.dimensionSizes.size() + 1) {
            this.dimensionSizes.addElement(size);
        } else {
            this.dimensionSizes.setElementAt(size, dim - 1);
        }
    }
}
