/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe responsvel por armazenar a inicializao de uma matriz.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class JArrayInitializer extends AbstractNode {

    /**
     * Armazena as expresses de inicializao da matriz. Dois tipos so
     * possveis: expresses propriamente ditas e inicializadores contendo
     * vrias expressoes ou inicializadores. Essa estrutura recursiva 
     * necessria para que seja possvel representar a multidimensionalidade
     * de uma matriz.
     */
    private NodeList init;


    /**
     * Construtor vazio da classe.
     */
    public JArrayInitializer() {
        super();
        init = new NodeList();
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada
     * elemento de <code>this.init</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        int i = 0;

        // Envie o visitor atravs da rvore ...
        while (i < this.init.size()) {
            (this.init.elementAt(i)).accept(visitor, data);
            i++;
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Adiciona um novo inicializador a esta lista de inicializadores.
     *
     * @param initializer Um objeto correspondendo a um inicializador. Pode ser
     *                    uma instncia de <code>JExpression</code> ou <code>JArrayInitializer</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          argumento <code>JaTSNode</code> for <code>null</code> ou de um
     *          tipo considerado invlido. .
     */
    public void addInitializer(INode initializer)
            throws IllegalArgumentException {
        if (initializer == null || !(initializer instanceof JExpression ||
                initializer instanceof JArrayInitializer))
            throw new IllegalArgumentException();

        this.init.addElement(initializer);
    }


    /**
     * Retorna um objeto correspondente  inicializao de uma parte da matriz,
     * podendo ser uma expresso ou um outro objeto do tipo <code>JArrayInitializer</code>.
     *
     * @param dim A dimenso  qual o n desejado corresponde  inicializao.
     *            O valor mnimo para esse parmetro  1.
     * @return Um n que pode ser tanto do tipo <code>JExpression</code> quanto do tipo
     *         <code>JArrayInitializer</code>.
     */
    public JaTSNode getInitializerAt(int dim) throws IllegalArgumentException {
        if (dim <= 0 || dim > init.size()) {
            throw new IllegalArgumentException();
        }
        return (JaTSNode) this.init.elementAt(dim - 1);
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JArrayInitializer)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JArrayInitializer aInit = (JArrayInitializer) node;
            this.init.match(aInit.init, results);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {
            // Processamento de cada uma das expresses de inicializao,
            // respectivamente.
            int i = 0;
            int max = this.size();

            while (i < max) {
                INode currInit = this.getInitializerAt(i + 1);
                if (currInit != null) {
                    Object ret = this.processNode(currInit, data);
                    if (currInit.isExecutable()) {
                        if (!(ret instanceof INode))
                            throw new ExecutionException(this);

                        this.setInitializerAt((INode) ret, i);
                    }
                }
                i++;
            }
        }

        return retorno;
    }


    /**
     * Remove inicializador desta lista de inicializadores.
     *
     * @param dim A dimenso da qual deseja-se remover o inicializador.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          valor do argumento <code>dim</code> for invlido.
     */
    public void removeInitializerAt(int dim) throws IllegalArgumentException {
        if (dim <= 0 || dim > init.size()) {
            throw new IllegalArgumentException();
        }
        this.init.removeElementAt(dim - 1);
    }


    /**
     * Adiciona um novo inicializador a esta lista de inicializadores na posio
     * desejada.
     *
     * @param initializer Um objeto correspondendo a um inicializador. Pode ser
     *                    uma instncia de <code>JExpression</code> ou <code>JArrayInitializer</code>.
     * @param dim         A dimenso  qual o n desejado corresponde  inicializao.
     *                    O valor mnimo para esse parmetro  1.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          argumento <code>JaTSNode</code> for <code>null</code> ou de um
     *          tipo considerado invlido. Tambm pode ser levantada se o
     *          valor do argumento <code>dim</code> for invlido.
     */
    public void setInitializerAt(INode initializer, int dim)
            throws IllegalArgumentException {
        if (initializer == null || !(initializer instanceof JExpression ||
                initializer instanceof JArrayInitializer)
                || dim <= 0 || dim - 1 > init.size()) {
            throw new IllegalArgumentException();
        }

        this.init.setElementAt(initializer, dim - 1);
    }


    /**
     * Retorna o nmero de inicializaes presentes nesta lista.
     *
     * @return Um inteiro correspondendo ao nmero de inicializaes da lista.
     */
    public int size() {
        int retorno = 0;
        if (this.init != null) {
            retorno = this.init.size();
        }

        return retorno;
    }

}
