/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.EvaluateFunctions;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.ExpressionEvaluationException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa expresses binrias.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class JBinaryExpression extends JExpression {

	
    // Armazena o operador desta expresso.
    private int binaryOperator;


    // Armazena o operando que fica  esquerda do operador.
    private JExpression leftHandOperand;


    // Armazena o operando que fica  direita do operador.
    private JExpression rightHandOperand;


    /* Constantes inteiras correspondendo aos operadores possveis. */

    // OU lgico ("!!").
    public final static int CONDITIONAL_OR_OPERATOR = 1;

    // AND lgico ("&&").
    public final static int CONDITIONAL_AND_OPERATOR = 2;

    // OU bit a bit inclusivo ("|").
    public final static int INCLUSIVE_OR_OPERATOR = 3;

    // OU bit a bit exclusivo ("^").
    public final static int EXCLUSIVE_OR_OPERATOR = 4;

    // AND bit a bit ("&").
    public final static int AND_OPERATOR = 5;

    // Igualdade ("==").
    public final static int EQUALITY_OPERATOR = 6;

    // Desigualdade ("!=").
    public final static int INEQUALITY_OPERATOR = 7;

    // Inferioridade nmerica ("<").
    public final static int LESSER_THAN_OPERATOR = 8;

    // Superioridade nmerica (">").
    public final static int GREATER_THAN_OPERATOR = 9;

    // Inferioridade ou igualdade numrica ("<=").
    public final static int LESSER_OR_EQUAL_TO_OPERATOR = 10;

    // Superioridade ou igualdade numrica (">=").
    public final static int GREATER_OR_EQUAL_TO_OPERATOR = 11;

    // Operador de deslocamento para a esquerda ("<<").
    public final static int LEFT_SHIFT_OPERATOR = 12;

    // Operador de deslocamento para a direita (">>").
    public final static int RIGHT_SHIFT_OPERATOR = 13;

    // Operador de deslocamento para a direita sem sinal (">>>".
    public final static int UNSIGNED_RIGHT_SHIFT_OPERATOR = 14;

    // Adio ("+").
    public final static int ADDITION_OPERATOR = 15;

    // Subtrao ("-").
    public final static int SUBTRACTION_OPERATOR = 16;

    // Multiplicao ("*").
    public final static int MULTIPLICATION_OPERATOR = 17;

    // Diviso ("/").
    public final static int DIVISION_OPERATOR = 18;

    // Resto da diviso inteira ("%").
    public final static int REMAINDER_OPERATOR = 19;

    // atribuio simples
    public static final int ASSIGN = 20;

    // atribuio e soma "+="
    public static final int PLUS_ASSIGN = 21;

    // atribuio e subtrao "-="
    public static final int MINUS_ASSIGN = 22;

    // atribuio e multiplicao "*="
    public static final int TIMES_ASSIGN = 23;

    // atribuio e diviso "/="
    public static final int DIVIDE_ASSIGN = 24;

    // atribuio e shift a esquerda "<<="
    public static final int LSHIFT_ASSIGN = 25;

    // atribuio e shift  direita ">>="
    // leva o sinal em considerao
    public static final int RSIGNED_SHIFT_ASSIGN = 26;

    // atribuio e shift lgico  direita ">>>="
    // no leva o sinal em considerao
    public static final int RUNSIGNED_SHIFT_ASSIGN = 27;

    // atribuio e resto da diviso "%="
    public static final int REM_ASSIGN = 28;

    // atribuio e and lgico ( feito bit a bit) "&="
    public static final int AND_ASSIGN = 29;

    // atribuio e or lgico (bit a bit) "|="
    public static final int OR_ASSIGN = 30;

    // atribuio e xor lgico  "^="
    public static final int XOR_ASSIGN = 31;


    /**
     * Construtor vazio da classe
     */
    public JBinaryExpression() {
        super();
        this.binaryOperator = 0;
        this.leftHandOperand = null;
        this.rightHandOperand = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro o operador binrio
     * e os dois operandos sobre os quais ele ser aplicado.
     *
     * @param binOperator O operador binrio da operao. Pode ser qualquer um
     *                    dos operadores binrios de Java.
     * @param leftHand    O operando que se encontra do lado esquerdo do operador.
     * @param rightHand   O Operando que se encontra do lado direito do operador.
     */
    public JBinaryExpression(int binOperator, JExpression leftHand,
                             JExpression rightHand) {
        super();
        this.binaryOperator = binOperator;
        this.leftHandOperand = leftHand;
        this.rightHandOperand = rightHand;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * de <code>this.leftHandOperator</code> e o de <code>this.rightHandOperator</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            if (this.leftHandOperand == null || this.rightHandOperand == null) {
                throw new InconsistentNodeException(this);
            }

            this.leftHandOperand.accept(visitor, data);
            this.rightHandOperand.accept(visitor, data);

            // ... invoque o accept da declarao executvel associada a este n,
            // caso haja alguma...
            this.invokeAcceptOnExecutableDeclaration(visitor, data);

            // e ento visite este n.
            retorno = visitor.visit(this, data);
        }

        return retorno;
    }

    /**
     * Retorna o string correspondente ao operador binrio desta expresso.
     *
     * @return Um <code>String</code> correspondendo ao operador binrio desta expresso.
     */
    public String getBinaryOperatorString() {
        String retorno = "";
        int op = this.getOperator();

        if (op == CONDITIONAL_OR_OPERATOR) {
            retorno = "||";
        } else if (op == CONDITIONAL_AND_OPERATOR) {
            retorno = "&&";
        } else if (op == INCLUSIVE_OR_OPERATOR) {
            retorno = "|";
        } else if (op == EXCLUSIVE_OR_OPERATOR) {
            retorno = "^";
        } else if (op == AND_OPERATOR) {
            retorno = "&";
        } else if (op == EQUALITY_OPERATOR) {
            retorno = "==";
        } else if (op == INEQUALITY_OPERATOR) {
            retorno = "!=";
        } else if (op == LESSER_THAN_OPERATOR) {
            retorno = "<";
        } else if (op == GREATER_THAN_OPERATOR) {
            retorno = ">";
        } else if (op == LESSER_OR_EQUAL_TO_OPERATOR) {
            retorno = "<=";
        } else if (op == GREATER_OR_EQUAL_TO_OPERATOR) {
            retorno = ">=";
        } else if (op == LEFT_SHIFT_OPERATOR) {
            retorno = "<<";
        } else if (op == RIGHT_SHIFT_OPERATOR) {
            retorno = ">>";
        } else if (op == UNSIGNED_RIGHT_SHIFT_OPERATOR) {
            retorno = ">>>";
        } else if (op == ADDITION_OPERATOR) {
            retorno = "+";
        } else if (op == SUBTRACTION_OPERATOR) {
            retorno = "-";
        } else if (op == MULTIPLICATION_OPERATOR) {
            retorno = "*";
        } else if (op == DIVISION_OPERATOR) {
            retorno = "/";
        } else if (op == REMAINDER_OPERATOR) {
            retorno = "%";
        } else if (op == ASSIGN) {
            retorno = "=";
        } else if (op == PLUS_ASSIGN) {
            retorno = "+=";
        } else if (op == MINUS_ASSIGN) {
            retorno = "-=";
        } else if (op == TIMES_ASSIGN) {
            retorno = "*=";
        } else if (op == DIVIDE_ASSIGN) {
            retorno = "/=";
        } else if (op == REM_ASSIGN) {
            retorno = "%=";
        } else if (op == AND_ASSIGN) {
            retorno = "&=";
        } else if (op == OR_ASSIGN) {
            retorno = "|=";
        } else if (op == LSHIFT_ASSIGN) {
            retorno = "<<=";
        } else if (op == RSIGNED_SHIFT_ASSIGN) {
            retorno = ">>=";
        } else if (op == RUNSIGNED_SHIFT_ASSIGN) {
            retorno = ">>>=";
        }
        return retorno;
    }

    /**
     * Mtodo que retorna o operando  esquerda do operador.
     *
     * @return Uma objeto do tipo JExpression correspondendo ao operando 
     *         esquerda do operador.
     */
    public JExpression getLeftHandOperand() {
        return this.leftHandOperand;
    }

    /**
     * Mtodo que retorna o operador binrio propriamente dito.
     *
     * @return Um inteiro correspondendo a um operador binrio.
     */
    public int getOperator() {
        return this.binaryOperator;
    }

    /**
     * Mtodo que retorna o operando  direita do operador.
     *
     * @return Uma objeto do tipo JExpression correspondendo ao operando 
     *         direita do operador.
     */
    public JExpression getRightHandOperand() {
        return this.rightHandOperand;
    }

    /**
     * Verifica o operador  um operador de atribuio
     *
     * @param op Operador que se deseja verificar se  ou no um operador de
     *           atribuio
     * @return true, se for verdade <br>
     *         false, caso contrrio
     */
    public static boolean isAssignOperator(int op) {
        boolean retorno = false;

        retorno = (op == ASSIGN ||
                op == PLUS_ASSIGN ||
                op == MINUS_ASSIGN ||
                op == TIMES_ASSIGN ||
                op == DIVIDE_ASSIGN ||
                op == REM_ASSIGN ||
                op == LSHIFT_ASSIGN ||
                op == RUNSIGNED_SHIFT_ASSIGN ||
                op == RSIGNED_SHIFT_ASSIGN ||
                op == AND_ASSIGN ||
                op == OR_ASSIGN);
        return retorno;

    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JBinaryExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JBinaryExpression exp = (JBinaryExpression) node;

            if (this.getOperator() != exp.getOperator()) {
                throw new NodesNotMatchedException("Binary expressions have different " +
                        "operators", this, node);
            }

            if (this.getLeftHandOperand() == null ||
                    this.getRightHandOperand() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }

            (this.getLeftHandOperand()).match(exp.getLeftHandOperand(), results);
            (this.getRightHandOperand()).match(exp.getRightHandOperand(), results);
        }

    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do operando  esquerda do operador.
            INode lhso = this.getLeftHandOperand();
            if (lhso != null) {
                Object ret1 = this.processNode(lhso, data);
                if (lhso.isExecutable()) {

                    if (!(ret1 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setLeftHandOperand((JExpression) ret1);
                }
            }

            // Processamento do operando  direita do operador.
            INode rhso = this.getRightHandOperand();
            if (rhso != null) {
                Object ret2 = this.processNode(rhso, data);
                if (rhso.isExecutable()) {

                    if (!(ret2 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setRightHandOperand((JExpression) ret2);
                }
            }
        }

        return retorno;
    }

    /**
     * Mtodo que seta o operando  esquerda do operador.
     *
     * @param leftHand A expresso que ficar  esquerda do operador binrio.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>leftHand</code> for <code>null</code>.
     */
    public void setLeftHandOperand(JExpression leftHand)
            throws IllegalArgumentException {

        if (leftHand == null) {
            throw new IllegalArgumentException();
        }
        this.leftHandOperand = leftHand;
    }

    /**
     * Mtodo que seta o operador binrio da expresso.
     *
     * @param op Um inteiro correspondendo ao operador binrio.
     */
    public void setOperator(int op) {
        this.binaryOperator = op;
    }

    /**
     * Mtodo que seta o operando  direita do operador.
     *
     * @param rightHand A expresso que ficar  direita do operador binrio.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>rightHand</code> for <code>null</code>.
     */
    public void setRightHandOperand(JExpression rightHand)
            throws IllegalArgumentException {

        if (rightHand == null) {
            throw new IllegalArgumentException();
        }
        this.rightHandOperand = rightHand;
    }

    /**
     * Compara dois objetos do tipo JBinaryExpression e diz se estes dois so
     * iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false;
        if (obj instanceof JBinaryExpression) {
            JBinaryExpression bexp = (JBinaryExpression) obj;
            if (this.isASourceNode()) {
                retorno = (super.equals(bexp)) &&
                        (this.binaryOperator == bexp.binaryOperator) &&
                        (AbstractNode.equals(this.leftHandOperand, bexp.leftHandOperand)) &&
                        (AbstractNode.equals(this.rightHandOperand, bexp.rightHandOperand));
            } else {
                retorno = compareJaTSNode(bexp);
            }
        }
        return retorno;
    }


    public Object evaluate(Object data) throws ExecutionException {

        Object rightHand = this.getRightHandOperand().evaluate(data);
        Object leftHand = this.getLeftHandOperand().evaluate(data);

        int oper = this.getOperator();
        Object result = null;

        try {
            result = EvaluateFunctions.applyBinaryOperator(oper, leftHand, rightHand);
        } catch (IllegalArgumentException e) {
            throw new ExpressionEvaluationException(e.getMessage());
        }

        return result;
    }

}
