/*
 *  JaTS Project - Java Transformation System
 *  http://www.cin.ufpe.br/~jats
 *  
 *  Copyright  2005 by Informatics Center - Federal University of Pernambuco
 *     
 *   This document can only be used for non comercial purposes and without 
 *   *any warranty*! The information contained here can not be published, or 
 *   disclosed, or disseminated without the authors' consent. Reverse 
 *   engineering of any or all information in this document is prohibited.
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;

/**
 * Esta classe guarda as informacoes a respeito de um Bloco, o qual 
 * representado por uma lista de Statements
 *
 * @author Adeline de Sousa
 * @author Gustavo Santos
 * @version 12/01/2003
 * @since JDK 1.3
 */
public class JBlock extends JStatement implements JaTSCollection {


    // armazena os statements presentes no corpo do bloco, de forma sequencial
    private JStatementList statements;


    /**
     * Construtor vazio da classe
     */
    public JBlock() {
        this.statements = new JStatementList();
        this.setTypeIdentifier(JaTSNode.BLOCK);
    }

    /**
     * Gets the statements attribute of the JBlock object
     *
     * @return The statements value
     */
    public JStatementList getStatements() {
        return this.statements;
    }

    /**
     * Sets the statements attribute of the JBlock object
     *
     * @param nos The new statements value
     */
    public void setStatements(JStatementList nos) {
        this.statements = nos;
    }

    /**
     * Sets the statements attribute of the JBlock object
     *
     * @param nos The new statements value
     */
    public void setStatements(NodeList nos) {
        this.statements = new JStatementList();
    	this.statements.setStatements(nos);
    }


    /**
     * Gets the iterator attribute of the JBlock object
     *
     * @return The iterator value
     */
    public JaTSIterator getIterator() {
        return this.statements.getIterator();
    }


    /**
     * Verifies if the JBlock object is empty
     *
     * @return The empty value
     */
    public boolean isEmpty() {
        return this.statements.isEmpty();
    }


    /**
     * Gets the block size (number of statements)
     *
     * @return The number of statements
     */
    public int size() {
        return this.statements.size();
    }


    /**
     * Remove the element from the position indicated by index 
     *
     * @param index The position index
     */
    public void removeElementAt(int index) {
        this.statements.removeElementAt(index);
    }


    /**
     * Gets the specified statement
     *
     * @param index Statement index
     * @return Desired statement
     */
    public INode elementAt(int index) {
        return (INode) this.statements.elementAt(index);
    }

    
    /**
     * Adiciona o elemento na posio <code>pos</code> desta lista e d um shift
     * para a direita de todos os elementos que se encontram a direita a partir
     * desta posio.
     *
     * @param elem Elemento a ser adicionado
     * @param pos  Um inteiro correspondendo  posio em que o elemento ser
     *             adicionado
     * @throws IllegalArgumentException Levantada quando o
     *                                  argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode elem, int pos)
            throws IllegalArgumentException {
        this.statements.insertElementAt(elem, pos);
    }

    /**
     * Seta o elemento que se encontra na posio <code>pos</code> desta lista
     * com o valor passado como parmetro.
     *
     * @param elem Elemento a ser adicionado
     * @param pos  Um inteiro correspondendo  posio do n que deve ser
     *             removido.
     * @throws IllegalArgumentException Levantada quando o
     *                                  argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode elem, int pos)
            throws IllegalArgumentException {
        this.statements.getStatements().setElementAt(elem, pos);
    }

    /**
     * Description of the Method
     *
     * @param node Description of the Parameter
     */
    public void add(INode node) {
        this.statements.getStatements().addElement(node);

    }

    /**
     * Removes the specified element from the block
     * 
     * @param elem
     */
    public void remove(INode elem) {
        int index = this.statements.getStatements().indexOf(elem);
        this.removeElementAt(index);

    }


    /**
	 * Este mtodo realiza o casamento entre este n, que faz parte da rvore
	 * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
	 * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
	 * casam em duas situaes: quando eles so completamente idnticos e quando
	 * o primeiro deles  uma varivel JaTS e, ou essa varivel ainda no foi
	 * mapeada a nenhum valor no conjunto resultado do casamento (
	 * <code>results</code>) ou o valor ao qual ela foi mapeada  idntico
	 * quele ao qual ela est sendo mapeada agora (<code>node</code>).
	 * "primeiro", no contexto acima, se refere ao n que faz parte do programa
	 * JaTS que est sendo casado.
	 * 
	 * @param node
	 *            O objeto do tipo <code>JaTSNode</code> com o qual 
	 *            <code>this</code> ser casado.
	 * @param results
	 *            O conjunto resultado do casamento.  usado para armazenar os
	 *            pares <code>(varivel,cdigo)</code> que so gerados pelo
	 *            casamento.
	 * @throws IllegalArgumentException
	 *             Levantada quando o parmetro qualquer um dos argumentos tiver
	 *             valor <code>null</code>.
	 * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
	 *             Levantada quando algum componente essencial do n tiver um
	 *             valor, tornando o n, como um todo, inconsistente.
	 * @throws NodesNotMatchedException
	 *             Lanada sempre que o casamento falhar porque
	 *             <code>node</code> no  um n compatvel com
	 *             <code>this</code> , ou seja, nem <code>this</code>  uma
	 *             varivel no casada nem <code>this</code>  idntico a
	 *             <code>this</code>.
	 */
    public void match(INode node, ResultSet results)
			throws IllegalArgumentException, InconsistentNodeException,
			NodesNotMatchedException {

		if (node == null) {
			throw new NodesNotMatchedException(
					NodesNotMatchedException.NODES_IS_NULL_MSG, this, node);
		}
		if (!(node instanceof JBlock)) {
			throw new NodesNotMatchedException(
					NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
					this, node);
		}
		
		//try to match as block variable or as context variable
		if (!matchesAsAVariable(node, results, JaTSNode.BLOCK)
//				&& !this.matchesAsContext((JBlock)node, results)
				) {
			
			//if it is not a variable nor an executable declaration
			if (this.isASourceNode()) {
				this.statements.match(((JBlock)node).getStatements(), results);
			}
		}
	}


    /**
     * This method tries to do a matching involving a context variable
     * declared inside a block. <br>
     * Context declarations have the following structure: <br>
     * <code> ContextDeclaration:#cont{#code} </code><br>
     * The behavior is: identify the code of interest (#code) and extract
     * him from the context (#cont). Then, map #code to the portion of 
     * code extracted and #cont to the remaining code.
     * 
     * @param concreteContex The hole block
     * @param results The result set
     * @return true if succeed
     * @throws NodesNotMatchedException
     */
//    private boolean matchesAsContext(JBlock concreteContex, ResultSet results)
//			throws NodesNotMatchedException {
//    	boolean success = false;
//    	ContextDeclaration cd = null;
//    	
//    	if (!this.statements.isEmpty()) {
//    		INode stm = this.statements.elementAt(0);
//    		if (stm instanceof ContextDeclaration) {
//    			cd = (ContextDeclaration)stm;
//
//    			JStatementList concreteCode;
//    			
//    			//replaces the jats variable by the equivalent abstract node
//    			//in the context declaration node
//    			cd.setCodeOfInterest((JStatementList) results.get(
//    					cd.getCodeOfInterest()));
//
//    			//the position if dinamically changed during the process
//    			cd.setCodePosition(cd.getCodeOfInterest().getPosition());
//    			
//    			//retrieves the concrete refrences to the code of interest
//    			concreteCode = SelectionProcessor.processStatement(
//						concreteContex, cd);
//    			
//    			//the position is needed for introducing the code back
//    			concreteCode.setPosition(cd.getCodePosition());
//    			
//    			//replaces the mapping to the code of interest 
//    			results.put(cd.getCodeOfInterest(), concreteCode);
//
//    			//keep the code position value
//    			concreteContex.getStatements().setPosition(cd.getCodeOfInterest().getPosition());
//
//    			//maps the context (already without the internal portion)
//    			results.put(cd.getContext(), concreteContex.getStatements());
//    			
//    			success = true;
//    			
//    		}
//    	}
//    	
//		return success;
//	}

	
    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br>
     * <br>
     * 1 - <i>O n a ser processado  executvel</i> . Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code>
     * associado)  retornado pelo seu mtodo process e corresponde ao
     * resultado de seu processamento. <br>
     * <br>
     * 2 - <i>O n  pai de um n executvel</i> . Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de filhos
     * pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo <code>process</code>
     * em cada um de seus filhos. <br>
     * <br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i> . Aqui, o
     * processamento do n se resume a invocar o mtodo <code>process</code>
     * para cada um dos filhos do n. <br>
     * <br>
     * 4 - <i>O n a ser processado  do tipo </i> <code>Executable Declaration</code>
     * . O processamento de ns do tipo <code>Executable Declaration</code> 
     * completamente diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores sintticas
     * correspondentes s expresses a ser executadas so processadas de modo
     * que o resultado gerado seja equivalente ao que seria se o cdigo Java
     * dessas expresses fosse executado. Por exemplo, o processamento da
     * rvore sinttica de uma invocao de mtodo fornece como resultado o
     * mesmo resultado que seria fornecido se esse mtodo estivesse sendo
     * invocado dentro de um programa em Java.
     *
     * @param data Um objeto que pode conter
     *             informaes necessrias  execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException        Se algum erro ocorrer durante o
     *                                   processamento das estruturas executveis do programa.
     * @throws IllegalArgumentException  Caso o parmetro <code>data</code>
     *                                   seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n
     *                                   inconsistente (incompleto) for encontrado. Por exemplo, se uma
     *                                   declarao de atributo no possuir um filho correspondendo ao tipo
     *                                   da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;
        if (this.isExecutable()) {
            retorno = processExecutableNode(data);
        } else {
        	this.statements.process(data);
        }

        return retorno;
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>
     *                , funcionando de acordo com o padro de projeto homnimo, ou seja,
     *                implementa uma operao sobre vrios tipos diferentes de objetos
     *                fora do cdigo desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>, que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser
     *                utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a
     *         ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando algum componente
     *                                   essencial do n tiver um valor, tornando o n, como um todo,
     *                                   inconsistente.
     * @throws ExecutionException        Levantada sempre que a execuo de
     *                                   um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        
        this.statements.accept(visitor, data);
        
        // chama o visitor para a declarao executavel associadas ao JBlock
        invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Compara dois objetos do tipo <code>JBlock<code> e diz se estes so iguais
     *
     * @param obj Description of the Parameter
     * @return true, caso os dois objetos sejam iguais false, caso contrrio
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JBlock) {
            JBlock block = (JBlock) obj;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.getStatements(), block.getStatements());
            } else {
                retorno = compareJaTSNode(block);
            }
        }
        return retorno;
    }


}
