/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe representa uma clausula catch, a qual  composta por um parametro
 * e um bloco.
 * <p/>
 * Ex:
 * <p/>
 * catch(Exception ex){
 * // bloco catch
 * }
 * 
 * @author Adeline de Sousa
 * @version 13/11/2002
 */
public class JCatchClause extends AbstractNode {


    // parametro da clausula catch
    private JFormalParameter parameter;

    // bloco da clausula catch
    private JBlock block;

    // comentrio referente ao catch
    private String comment;


    /**
     * Construtor vazio da classe
     */
    public JCatchClause() {
        super();
    }


    /**
     * Construtor da classe
     *
     * @param parameter Parametro da clausula catch
     * @param block     Bloco associado a esta clusula catch
     */
    public JCatchClause(JFormalParameter parameter, JBlock block) {
        this();
        this.setParameter(parameter);
        this.setBlock(block);
    }


    /**
     * Retorna o parametro desta clausula Catch
     *
     * @return Parametro desta clausula  Catch
     */
    public JFormalParameter getParameter() {
        return parameter;
    }


    /**
     * Seta o parametro da clausula catch
     *
     * @param param Parametro da clausula catch
     */
    public void setParameter(JFormalParameter param) {
        this.parameter = param;
    }


    /**
     * Retorna o bloco associado a esta clusula catch
     *
     * @return Bloco associado a esta clusula catch
     */
    public JBlock getBlock() {
        return block;
    }


    /**
     * Seta o bloco associado a esta clusula catch
     *
     * @param block Bloco associado a esta clusula catch
     */
    public void setBlock(JBlock block) {
        this.block = block;
    }


    /**
     * Retorna o comentrio associado a esta clusula catch
     *
     * @return Comentrio associado a esta clusula catch
     */
    public String getComment() {
        return comment;
    }

    /**
     * Seta o comentrio associado a esta clusula catch
     *
     * @param comment Comentrio associado a esta clusula catch
     */
    public void setComment(String comment) {
        if (Util.checkString(comment)) {
            this.comment = comment;
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        Object retorno = null;

        if (this.isASourceNode()) {

            if (this.block != null) {
                this.block.accept(visitor, data);
            }

            if (this.parameter != null) {
                this.parameter.accept(visitor, data);
            }
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JCatchClause)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JCatchClause catchClause = (JCatchClause) node;
            if (this.getParameter() != null && catchClause.getParameter() != null) {
                this.getParameter().match(catchClause.getParameter(), results);
            }

            if (this.getBlock() != null && catchClause.getBlock() != null) {
                this.getBlock().match(catchClause.getBlock(), results);
            }
        }
    }

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);

        } else {
            INode param = this.getParameter();
            if (param != null) {
                Object ret1 = this.processNode(param, data);
                if (param.isExecutable()) {
                    try {
                        this.setParameter((JFormalParameter) ret1);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
            INode cBlock = this.getBlock();
            if (cBlock != null) {
                Object ret2 = this.processNode(cBlock, data);
                if (cBlock.isExecutable()) {
                    try {
                        this.setBlock((JBlock) ret2);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
        }
        return retorno;
    }

}
