/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe guarda as informacoes a respeito de uma declaracao de classe.
 * Uma declarao de classe  constituda por uma lista de modificadores, um
 * identificador da classe, possivelmente uma superclasse e uma lista de
 * interfaces implementadas e o corpo da classe.
 * 
 * @author Adeline de Sousa
 * @version 14/01/2003
 * @since JDK 1.3
 */
public class JClassDeclaration extends JTypeDeclaration {


    // nome da super classe (opcional)
    private JName superClass;

    /**
     * Construtor vazio da classe
     */
    public JClassDeclaration() {
        super();
        this.setTypeIdentifier(JaTSNode.CLASS_DECLARATION);
        this.superClass = null;
    }


    /**
     * Construtor da classe que recebe todos os parametros, referentes aos
     * atributos desta classe
     *
     * @param mods lista de modificadores
     * @param id   identificador da classe
     * @param sc   nome da superclasse
     * @param ifs  lista de interfaces implementadas
     * @param body corpo da classe
     * @throws IllegalArgumentException - se o id passado for null.
     *                                  Os outros podem ser null, prque sao opcionais
     */
    public JClassDeclaration(JModifierList mods, JIdentifier id, JName sc,
                             JNameList ifs, TypeBody body) {
        super(mods, id, ifs);
        this.superClass = sc;
        this.setBody(body);
        this.setTypeIdentifier(JaTSNode.CLASS_DECLARATION);
    }


    /**
     * Retorna o nome da superclasse
     *
     * @return Superclasse
     */
    public JName getSuperClass() {
        return this.superClass;

    }


    /**
     * Seta a superclasse  desta classe
     *
     * @param sc Classe que  herdada pela classe atual
     */
    public void setSuperClass(JName sc) {
        this.superClass = sc;
    }


    /**
     * Retorna o conjunto de construtores da classe
     *
     * @return Conjunto de construtores da classe
     */
    public JConstructorDeclarationSet getConstructors() {
        return this.getBody().getConstructorDeclarationSet().getConstructorsSet();
    }

    /**
     * Seta o conjunto de construtores da classe
     *
     * @param constructors Conjunto de contrutores da classe
     */
    public void setConstructors(JConstructorDeclarationSet constructors) {

        this.getBody().getConstructorDeclarationSet().setConstructorsSet(constructors);

    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }

        if (!(node instanceof JTypeDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JClassDeclaration classDec = null;
            try {
                classDec = (JClassDeclaration) node;
                super.match(classDec, results);
            } catch (ClassCastException exc) {
                throw new NodesNotMatchedException("The node is not a class (" +
                        "node type: " + node.getClass().toString() + ")", this, node);
            }

            if (this.superClass != null) {
                // casa a superclasse
                this.superClass.match(classDec.getSuperClass(), results);
            } else if (classDec.getSuperClass() != null) {
                throw new NodesNotMatchedException("Superclasses couldn't be " +
                        "matched", this, classDec);
            }
            if (this.getBody() != null) {
                //casa o corpo da classe
                this.getBody().match(classDec.getBody(), results);
            }
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.parameters</code> e <code>this.identifier</code>
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo
     *                desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>,que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.getBody() != null) {
            this.getBody().accept(visitor, data);
        }

        if (this.getInterfaces() != null) {
            this.getInterfaces().accept(visitor, data);
        }

        if (this.superClass != null) {
            this.superClass.accept(visitor, data);
        }

        if (this.getName() != null) {
            this.getName().accept(visitor, data);
        }

        if (this.getModifiers() != null) {
            this.getModifiers().accept(visitor, data);
        }

        // Envie o visitor pela sub-arvore da lista de atributos
//        if (this.getAttributes() != null) {
//        	this.getAttributes().accept(visitor, data);
//        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {
            // processa os filhos - modifiers, name, interfaces e corpo
            super.process(data);

            // Processamento do nome da superclasse.
            INode scName = this.getSuperClass();
            if (scName != null) {
                Object ret1 = this.processNode(scName, data);
                if (scName.isExecutable()) {
                    try {
                        this.setSuperClass(JName.unwrapName(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
                this.getSuperClass().setOptional(false);
            }
        }

        return retorno;
    }


} // fim da classe
