/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa um comentrio. Pode ser Javadoc ou no. E est
 * associada a alguns ns da rvore sinttica. Os ns referentes a declaraes
 * recebem comentrios tipo Javadoc e os demais, apenas comentrios de
 * documentao.
 *
 * @author Adeline de Sousa
 * @version 28/10/2002
 * @see Commentable
 */
public class JComment extends AbstractNode {

    // texto relativo ao comentrio
    private String text;

    // tipo do comentrio: linha, bloco, doc ou nenhum (string vazia)
    private int commentType;

    // ausencia de comentrio
    public static int NONE = 0;
    // comentrio de documentao - javadoc
    public static int JAVADOC = 1;
    // comentrio de linha
    public static int SINGLE_LINE = 2;
    // comentrio de bloco
    public static int MULTI_LINE = 4;
    // variavel de comentario
    public static int VARIABLE = 8;

    public static final String MULTI = "/*";
    public static final String SINGLE = "//";
    public static final String DOC = "/**";
    public static final String VAR = "#";

    /**
     * Construtor da classe.
     */
    public JComment() {
        this.commentType = NONE;
        this.setTypeIdentifier(JaTSNode.COMMENT);
    }

    /**
     * Construtor da classe, que recebe uma string representando o comentrio
     * como parametro.
     *
     * @param text Texto que representa o comentrio.
     */
    public JComment(String text) {
        this.setText(text);
        this.commentType = getCommentCode(text);
        if (this.commentType == VARIABLE) {
            this.setVariableName(text);
            this.setVariable(true);
        }
    }

    /**
     * Retorna o texto referente ao comentrio (incluindo delimitadores)
     *
     * @return Texto referente ao comentrio (incluindo delimitadores)
     */
    public String getText() {
        return text;
    }

    /**
     * Retorna o tipo do comentrio. Se este  Javadoc, se  comentrio de linha
     * ou de bloco.
     *
     * @return Tipo do comentrio
     */
    public int getCommentType() {
        return commentType;
    }

    /**
     * Seta o texto referente ao comentrio
     *
     * @param text Texto do comentrio
     */
    public void setText(String text) {
        if (Util.checkString(text)) {
            this.text = text;
        }
    }

    /**
     * Retorna o tipo de comentrio. Se  javadoc, se  multi_line ou se 
     * single_line
     */
    public static int getCommentCode(String comment) {
        int result = 0;
        if (Util.checkString(comment)) {
            String tmpComment = comment.trim();
            if (tmpComment.startsWith(DOC)) {
                result = JAVADOC;
            } else if (
                    tmpComment.startsWith(MULTI) && tmpComment.charAt(2) != '*') {
                result = MULTI_LINE;
            } else if (tmpComment.startsWith(SINGLE)) {
                result = SINGLE_LINE;
            } else if (tmpComment.startsWith(VAR)) {
                result = VARIABLE;
            } else {
                result = NONE;
            }
        }
        return result;
    }

    /**
     * Retorna o tipo de comentrio. Se  javadoc, se  multi_line ou se 
     * single_line
     */
    public int getCommentCode() {
        int result = 0;
        if (Util.checkString(this.text)) {
            String tmpComment = this.text.trim();
            if (tmpComment.startsWith(DOC)) {
                result = JAVADOC;
            } else if (
                    tmpComment.startsWith(MULTI) && tmpComment.charAt(2) != '*') {
                result = MULTI_LINE;
            } else if (tmpComment.startsWith(SINGLE)) {
                result = SINGLE_LINE;
            } else if (tmpComment.startsWith(VAR)) {
                result = VARIABLE;
            } else {
                result = NONE;
            }
        }
        return result;
    }

    /**
     * Retorna a String que representa este comentrio.
     *
     * @return String que representa o comentrio
     */
    public String toString() {
        return this.text;
    }

    /**
     * Mtodo que visita a sub-rvore correspondente a este n e executa alguma
     * operao, por meio do visitor passado como parmetro.
     * No  usado por este n, em especial, por se tratar de um n que no
     * contm operaes associadas.
     */
    public Object accept(IVisitor visitor, Object data)
            throws
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            IllegalArgumentException,
            ExecutionException {
        return this;
    }

    /**
     * Realiza o casamento de padres entre o n (this) e o n passado como
     * parmetro.
     */
    public void match(INode node, ResultSet results)
            throws
            IllegalArgumentException,
            InconsistentNodeException,
            NodesNotMatchedException {
    }
}
