/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.NodeTable;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.visitors.ReplacementVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.PreconditionDoesNotHoldException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa uma unidade de compilao. Ou seja, um programa em
 * java, que pode ser composto de varias classes e/ou interfaces. Em termos de
 * gramtica, essa classe corresponde ao smbolo inicial
 *
 * @author Adeline de Sousa
 * @author Fernando Castor
 * @version  0.1 -  24/07/2001
 * @since jdk1.3
 */

public class JCompilationUnit extends AbstractNode implements ICompilationUnit {

    private transient JPreconditionDeclaration precondition;

    // cabealho do arquivo
    private String header;

    // clusula "package"
    private JPackageDeclaration packageClause;

    // lista de clusulas "import"
    private ImportDeclarations imports;

    // lista de declaracoes de tipo
    private NodeList typeDeclarations;

    // lista de declaraes de tipo
    private BodyTypeDeclarations types;

    // indica se ha uma declaracao de tipo publica
    private boolean hasPublicTypeDeclaration = false;

    public transient NodeTable table;

    /**
     * Retorna a localizao fsica deste
     */
    public String getLocation() {
        String location = "";
        if (this.getPackage() != null) {
            location = getPackage().getName().toString();
        }
        return location;

    }


    /**
     * Construtor vazio da classe.
     */
    public JCompilationUnit() {
        super();
        this.packageClause = null;
        this.imports = new ImportDeclarations();
        this.typeDeclarations = new NodeList();
        this.types = new BodyTypeDeclarations();
        this.precondition = null;
        this.setTypeIdentifier(NO_TYPE);
        this.table = new NodeTable();
    }

    /**
     * Construtor da classe.
     *
     * @param pre   A pr-condio de uma transformao JaTS.
     * @param pack  Declaracao do pacote
     * @param imp   Clausula import
     * @param types Declaracao de tipos
     */
    public JCompilationUnit(JPreconditionDeclaration pre,
                            JPackageDeclaration pack, ImportDeclarations imp,
                            NodeList types)
            throws IllegalArgumentException {
        super();
        this.setPackage(pack);
        this.setImports(imp);
        this.setTypeDeclarationList(types);
        this.setPrecondition(pre);
        this.setTypeIdentifier(NO_TYPE);
    }

    /**
     * Construtor da classe.
     *
     * @param types Declaracao de tipos
     */
    public JCompilationUnit(NodeList types) throws IllegalArgumentException {
        super();
        this.setTypeDeclarationList(types);
        this.setTypeIdentifier(NO_TYPE);

    }

    /**
     * Diz se este arquivo possui cabealho.
     *
     * @return true, se o arquivo possuir cabealho false, caso contrrio.
     */
    public boolean hasHeader() {
        return header != null;
    }

    /**
     * Retorna o cabealho do arquivo. Pode ser apenas uma linha, indicando que
     * foi gerado pelo JaTS, ou informaes de copy right.
     *
     * @return O cabealho do arquivo.
     */
    public String getHeader() {
        return header;
    }

    /**
     * Retorna a pr-condio da transformao da qual essa unidade de
     * compilao faz parte.
     *
     * @return A pr-condio.
     */
    public JPreconditionDeclaration getPrecondition() {
        return this.precondition;
    }

    /**
     * Retorna o pacote ao qual pertence a <code>ICompilationUnit</code>.
     *
     * @return pacote ao qual pertence a <code>ICompilationUnit</code>
     */
    public JPackageDeclaration getPackage() {

        return this.packageClause;
    }

    /**
     * Retorna a lista de classes ou pacotes importados.
     *
     * @return Lista de classes ou pacotes importados
     */
    public ImportDeclarations getImports() {
        return this.imports;

    }

    public BodyTypeDeclarations getTypes() {
        return types;
    }

    public void setTypes(BodyTypeDeclarations types) {
        this.types = types;
    }


    /**
     * Retorna o conjunto de declaracoes de tipo.
     *
     * @return O conjunto de declaracoes do tipo
     */
    public NodeList getTypeDeclarationList() {

        return this.typeDeclarations;
    }

    /**
     * Retorna o nome do tipo publico definido nesta CompilationUnit, o qual
     * deve ser usado para o nome do arquivo. Caso nao encontre nenhum, retorna
     * o nome do primeiro tipo declarado.
     *
     * @return Nome do tipo publico definido nesta CompilationUnit
     */
    public String getTypeName() {
        String typeName = "";
        JIdentifier name = null;
        if (this.size() > 0) {
            name = this.getTypeDeclaration(0).getName();
            JaTSIterator jit = this.typeDeclarations.getIterator();
            boolean achou = false;
            while (jit.hasNext() && !achou) {
                ITypeDeclaration tmpType = (ITypeDeclaration) jit.next();
                if (tmpType.isPublic()) {
                    name = tmpType.getName();
                    achou = true;
                }
            }
            typeName = name.toString();
        } else {
            throw new IllegalStateException("CompilationUnit doesn't contains type declarations");
        }
        return typeName;

    }

    /**
     * Seta o cabealho do arquivo. Pode ser apenas uma linha, indicando que foi
     * gerado pelo JaTS, ou informaes de copy right.
     *
     * @param header O cabealho do arquivo.
     */
    public void setHeader(String header) {
        this.header = header;
    }

    /**
     * Seta a pr-condio de uma transformao da qual essa unidade de
     * compilao faz parte.
     *
     * @param pre A pr-condio.
     * @throws IllegalArgumentException Levantada quando a pr-condio passada
     * for <code>null</code>.
     */
    public void setPrecondition(JPreconditionDeclaration pre) {

        if (pre != null) {
            this.precondition = pre;
        } else {
            throw new IllegalArgumentException();
        }

    }

    /**
     * Seta o pacote ao qual pertence a <code>IType</code>.
     *
     * @param pacote Pacote ao qual pertence a <code>IType</code>
     */
    public void setPackage(JPackageDeclaration pacote) {
        this.packageClause = pacote;
    }

    /**
     * Seta a lista de classes ou pacotes importados.
     *
     * @param listaImports Lista de classes ou pacotes importados
     */
    public void setImports(ImportDeclarations listaImports) {

        if (listaImports != null) {
            this.imports = listaImports;
        } else {
            throw new IllegalArgumentException("lista de importacoes invlida");
        }
    }

    /**
     * Seta a lista de classes ou pacotes importados. Esta  especcifica para
     * Declaraes Java.
     *
     * @param listaImports Lista de classes ou pacotes importados
     */
    public void setImports(JImportDeclarationSet listaImports) {

        if (listaImports != null) {
            this.imports.setImportsSet(listaImports);
        } else {
            throw new IllegalArgumentException("lista de importacoes invlida");
        }
    }

    /**
     * Retorna a lista de classes ou pacotes importados.
     *
     * @return Lista de classes ou pacotes importados
     */
    public JImportDeclarationSet getImportSet() {
        return this.imports.getImportsSet();

    }

    /**
     * Seta o conjunto de declaracoes de tipo.
     *
     * @param types Conjunto de declaracoes de tipo
     * @throws IllegalArgumentException Levantada quando o conjunto passado for
     *         nulo ou quando o  conjunto passado for invalido
     */
    public void setTypeDeclarationList(NodeList types) {

        if (types != null) {

            int i = 0;
            int size = types.size();
            ITypeDeclaration tmp = null;

            while (i < size) {

                if (!(types.elementAt(i) instanceof ITypeDeclaration)) {

                    throw new IllegalArgumentException("Elemento invalido no "
                            + " conjunto");
                }
                tmp = (ITypeDeclaration) types.elementAt(i);
                if (tmp.isPublic() && !(this.hasPublicTypeDeclaration())) {
                    this.hasPublicTypeDeclaration = true;
                } else if (tmp.isPublic()) {
                    throw new IllegalArgumentException("Ja existe um tipo publico");
                    // fazer uma exceao para este tipo de erro
                }
                i++;
            }
            this.typeDeclarations = types;
        } else {
            throw new IllegalArgumentException("types nulo");
        }

    }

    /**
     * Adiciona uma declarao de tipo (classe ou inteface) a esta
     * CompilationUnit.
     *
     * @param typeDec Declarao de tipo a ser adicionada
     */
    public void addTypeDeclaration(JTypeDeclaration typeDec) {
        if (!this.isExecutable() && !this.isVariable() && typeDec != null) {
            this.typeDeclarations.addElement(typeDec);
        } else {
            throw new IllegalArgumentException("Erro em JCompilationUnit."
                    + "addTypeDeclaration");
        }
    }

    /**
     * Retorna a declarao de tipo cujo nome foi especificado no parametro.
     *
     * @param name Nome do tipo desejado
     * @return A declarao de tipo cujo nome foi especificado no parametro
     */
    public JTypeDeclaration getTypeDeclaration(JIdentifier name) {
        JTypeDeclaration retorno = null;
        if (!this.isExecutable() && !this.isOptional() && name != null) {
            int size = this.size();
            JTypeDeclaration tmpTypeDec = null;
            boolean achou = false;
            for (int i = 0; (i < size && !achou); i++) {
                tmpTypeDec = (JTypeDeclaration) typeDeclarations.elementAt(i);
                if (tmpTypeDec.getName().equals(name)) {
                    retorno = tmpTypeDec;
                    achou = true;
                }
            }
        } else {
            throw new IllegalArgumentException("Argumento invlido em "
                    + "IType." + "getTypeDeclaration");
        }
        return retorno;
    }

    /**
     * Retorna a Declarao de tipo que est na posio desejada.
     *
     * @param pos Posio da declarao de tipo que se deseja
     * @return A declarao de tipo na posio especificada no parametro
     */
    public ITypeDeclaration getTypeDeclaration(int pos) {
        boolean validPosition = pos >= 0 && pos < this.size();
        ITypeDeclaration retorno = null;
        if (!this.isExecutable() && !this.isOptional() && validPosition) {
            retorno = (ITypeDeclaration) this.typeDeclarations.elementAt(pos);
        } else {
            throw new IllegalArgumentException("Argumento invlido em "
                    + "IType." + "getTypeDeclaration");
        }
        return retorno;
    }

    /**
     * Retorna true se ja houver uma declarao de tipo publica.
     */
    public boolean hasPublicTypeDeclaration() {
        return this.hasPublicTypeDeclaration;
    }


    /**
     * Retorna a tabela de ns associado a esta CompilationUnit.
     *
     * @return A tabela de ns associado a esta CompilationUnit
     */
    public NodeTable getTable() {
        return table;
    }

    /**
     * Seta a tabela de ns associado a esta CompilationUnit.
     *
     * @param table Tabela de ns associado a esta CompilationUnit
     */
    public void setTable(NodeTable table) {
        if (table != null) {
            this.table = table;
        } else {
            throw new IllegalArgumentException("");
        }
    }

    /**
     * Retorna o numero de declaraes de tipo presentes nesta CompilationUnit.
     *
     * @return Numero de declaraes de tipo presentes nesta CompilationUnit
     */
    public int size() {
        int retorno = 0;
        if (!this.isExecutable() && !this.isOptional()) {
            retorno = this.typeDeclarations.size();
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao
     * n atual e envi-lo atravs da rvore, invocando o mtodo accept de cada
     * um dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa
     *                uma operao sobre vrios tipos diferentes de objetos fora do
     *                cdigo desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>, que recebe como argumento
     *                <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro qualquer um dos argumentos tiver
     *          valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando algum componente essencial do n tiver um
     *          valor, tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Declaracoes de tipos
        if (typeDeclarations != null) {
            int i = this.typeDeclarations.size() - 1;
            //  Envia o visitor pelo conjunto de tipos
            while (i >= 0) {
                this.typeDeclarations.elementAt(i).accept(visitor, data);
                i--;
            }
        }

        // Declaracoes de imports
        if (imports != null) {
            this.imports.accept(visitor, data);
        }

        // pacote
        if (packageClause != null) {
            this.packageClause.accept(visitor, data);
        }

        // Envia o visitor para a pr-condio
        if (this.getPrecondition() != null) {
            this.getPrecondition().accept(visitor, data);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // Visite este no
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos e quando
     * o primeiro deles  uma varivel JaTS e, ou essa varivel ainda no foi
     * mapeada a nenhum valor no conjunto resultado do casamento (
     * <code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar os
     *                pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro qualquer um dos argumentos tiver
     *                                   valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando algum componente essencial do n tiver um
     *                                   valor, tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar porque
     *                                   <code>node</code> no  um n compatvel com
     *                                   <code>this</code>, ou seja, nem <code>this</code>  uma
     *                                   varivel no casada nem <code>this</code>  idntico a
     *                                   <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        ResultSet rset = (ResultSet) results.clone();

        if (node == null) {
            throw new IllegalArgumentException("Argumentos de entrada invalidos");
        }

        if (!(node instanceof ICompilationUnit)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, rset, NO_TYPE)) {

            JCompilationUnit compUnit = (JCompilationUnit) node;
            // se o pacote for requerido, mas nao estiver presente no programa
            // Java, levanta excecao
            if (compUnit.getPackage() == null
                    && (this.packageClause != null && !this.packageClause
                    .isOptional())) {
                throw new NodesNotMatchedException("Classes have different packages", this, compUnit);
            }
            // caso o pacote exista no programa jats e no programa Java, tenta
            // casa-los
            else if (this.packageClause != null
                    && compUnit.getPackage() != null) {
                // casa o nome do pacote
                this.packageClause.match(compUnit.getPackage(), rset);
            }

            if (this.imports != null && compUnit.getImports() != null) {
                this.imports.match(compUnit.getImports(), rset);

            }
            this.typeDeclarations
                    .match(compUnit.getTypeDeclarationList(), rset);
        }
        // se chegar ate aqui, faz-se o merge do ResultSet passado como
        // parametro e do seu clone
        results.merge(rset);
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de trs
     * maneiras diferentes, dependendo da situao: <br>
     * <br>
     * <p/>1 - <i>O n atual  executvel </i>. Nesse caso, o valor retornado
     * pelo processamento da <code>JExecutableDeclaration</code> associado ao
     * n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento. <br>
     * <br>
     * 2 - <i>O n atual  pai de um n executvel </i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em sua
     * lista de filhos pelo valor retornado pelo processamento de N. Alm disso,
     * o processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos. <br>
     * <br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel </i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n. <br>
     * <br>
     *
     * @param data Um objeto que pode conter informaes necessrias  execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro ocorrer durante o processamento das estruturas
     *          executveis do programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada caso o parmetro <code>data</code> seja
     *          <code>null</code> numa situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento da pr-condio.
            if (this.precondition != null) {
                this.getPreconditionValue((ResultSet) data);
            }

            // processamento do nome do pacote
            INode pack = this.getPackage();
            if (pack != null) {
                Object ret1 = this.processNode(pack, data);

                if (pack.isExecutable()) {
                    if (ret1 instanceof JPackageDeclaration) {
                        try {
                            this.setPackage((JPackageDeclaration) ret1);
                        } catch (Exception e) {
                            throw new ExecutionException(e.getMessage(), this);
                        }
                    }
                }
                if (pack.isOptional() && ret1 instanceof JPackageDeclaration) {
                    this.setPackage((JPackageDeclaration) ret1);
                }

                pack.setOptional(false);
            }

            // A lista de declaracoes de importacoes nao pode ser executavel
            // no a lista como um todo, mas os imports individuais sim!
            INode iDecSet = this.getImports();
            if (iDecSet != null) {
                this.processNode(iDecSet, data);
            }

            // Processamento de cada uma das declaraes de tipo,
            // respectivamente.
            int j = 0;
            if (this.typeDeclarations != null) {
                int max1 = this.typeDeclarations.size();
                while (j < max1) {
                    INode currTD = this.typeDeclarations.elementAt(j);
                    if (currTD != null) {
                        Object ret = this.processNode(currTD, data);

                        if (currTD.isExecutable()) {
                            if (!(ret instanceof JTypeDeclaration))
                                throw new ExecutionException(this);

                            this.typeDeclarations.setElementAt((JTypeDeclaration) ret, j);
                        }
                    }
                    j++;
                }
            }
            retorno = this;
        }
        return retorno;
    }

    public boolean getPreconditionValue(ResultSet rset) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
       // por definio,  true
        boolean result = true;
        ReplacementVisitor rvis = new ReplacementVisitor(rset);

        if (this.precondition != null) {
            rvis.visit(this.precondition, null);
            try {
                this.precondition.process(rset);
                result = true;
            } catch (PreconditionDoesNotHoldException ex) {
                result = false;
            }
            // para que a pr-condiao nao seja mais impressa
            this.precondition = null;
        }
        return result;
    }

    /**
     * Compara dois objetos do tipo IType e diz se estes so iguais
     */
    public boolean equals(JCompilationUnit compUnit) {

        boolean retorno = false;
        if (compUnit != null) {
            retorno = AbstractNode.equals(this.precondition, compUnit.precondition)
                    && AbstractNode.equals(this.packageClause, compUnit.packageClause)
                    && AbstractNode.equals(this.imports, compUnit.imports)
                    && AbstractNode.equals(this.typeDeclarations,
                            compUnit.typeDeclarations);

        }
        return retorno;
    }

} // Fim da classe
