/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe implementa declaraes condicionais em JaTS. Declaraes
 * condicionais so utilizadas quando se quer introduzir alguma condio para
 * a gerao de um trecho de cdigo mas no se quer o grau de restrio de
 * uma pr-condio. Funcionam mais ou menos como uma clusula de compilao
 * condicional de <code>C</code>, s que ao invs de delimitarem a parte que
 * ser ou no compilada, elas delimitam as partes que sero processadas e,
 * possivelmente, geradas.
 *
 * @author Fernando Castor
 * @see JaTSNode
 * @since JDK 1.3
 */
public class JConditionalDeclaration extends AbstractNode
        implements NestableExecutableDeclaration {


    //	O n pai deste.
    private INode parentNode;

    // Lista de declaraes gerado caso a condio deste objeto avalie para true.
    private NodeList trueDeclarationSet;

    //	Lista de declaraes gerado caso a condio deste objeto avalie para false.
    private NodeList falseDeclarationSet;

    //	A condio que dever ser avaliada para saber que trecho de cdigo deve
    //	ser gerado (pode ser possvel tambm que nenhum cdigo seja gerado).
    private JExpression condition;


    /**
     * Construtor sem parmetros.
     */
    public JConditionalDeclaration() {

        this.condition = null;
        this.trueDeclarationSet = new NodeList();
        this.falseDeclarationSet = new NodeList();
        this.parentNode = null;
    }

    /**
     * Construtor utilizado quando se quer criar uma declarao condicional que
     * no tem a clusula "else".
     *
     * @param cond     A condio que deve ser avaliada.
     * @param trueDecs A lista de declaraes que ser produzida caso a condio
     *                 avalie para <code>true</code>.
     * @param parent   O n do qual este n  filho.
     */
    public JConditionalDeclaration(JExpression cond, NodeList trueDecs,
                                   INode parent) {
        if (parent == null || cond == null || trueDecs == null ||
                trueDecs.size() == 0) {
            throw new IllegalArgumentException();
        }

        this.condition = cond;
        this.trueDeclarationSet = trueDecs;
        this.parentNode = parent;
    }

    /**
     * Construtor utilizado quando se quer criar uma declarao condicional
     * que consiste de duas partes: uma que  gerada caso a condio avalie para
     * <code>true</code> e outra que  gerada caso ela avalie para <code>false
     * </code>, correspondente  clusula <code>else</code> da declarao..
     *
     * @param cond      A condio que deve ser avaliada.
     * @param trueDecs  A lista de declaraes que ser produzida caso a condio
     *                  avalie para <code>true</code>.
     * @param falseDecs lista de declaraes que ser produzida caso a condio
     *                  avalie para <code>falsee>.
     * @param parent    O n do qual este n  filho.
     */
    public JConditionalDeclaration(JExpression cond, NodeList trueDecs,
                                   NodeList falseDecs, INode parent) {
        if (parent == null || cond == null || trueDecs == null ||
                trueDecs.size() == 0 || falseDecs == null ||
                falseDecs.size() == 0) {
            throw new IllegalArgumentException();
        }

        this.condition = cond;
        this.trueDeclarationSet = trueDecs;
        this.falseDeclarationSet = falseDecs;
        this.parentNode = parent;
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.condition == null) {
            throw new InconsistentNodeException(this);
        }

        this.condition.accept(visitor, data);

        if (this.trueDeclarationSet == null) {
            throw new InconsistentNodeException(this);
        }
        int maxTrue = this.trueDeclarationSet.size() - 1;
        for (int i = maxTrue; i >= 0; i--) {
            INode node = this.trueDeclarationSet.elementAt(i);
            node.accept(visitor, data);
        }

        if (this.falseDeclarationSet != null) {
            int maxFalse = this.falseDeclarationSet.size() - 1;
            for (int j = maxFalse; j >= 0; j--) {
                INode node = this.falseDeclarationSet.elementAt(j);
                node.accept(visitor, data);
            }
        }
        retorno = visitor.visit(this, data);
        return retorno;
    }


    /**
     * Devolve a expresso que deve ser avaliada na expresso condicional para
     * determinar que trecho de cdigo (se algum) ser processado.
     *
     * @return um n do tipo <code>JExpression</code>. A condio da declarao
     *         condicional.
     */
    public JExpression getCondition() {
        return this.condition;
    }

    /**
     * Devolve a lista de declaraes que devem ser processadas caso a avaliao
     * da condio da declarao condicional resulte em <code>false</code>.
     * Pode retornar <code>null</code> caso a declarao condicional seja declarada
     * sem uma clusula <code>else</code>.
     *
     * @return uma lista de declaraes. Um objeto do tipo <code>NodeList</code>.
     */
    public NodeList getFalseDeclarationSet() {
        return this.falseDeclarationSet;
    }

    /**
     * Devolve o n pai de um objeto do tipo <code>JConditionalDeclaration</code>.
     * O "pai", tem uma relao de agregao com o n a partir do qual o mtodo
     * foi invocado.
     *
     * @return o n pai do n a partir do qual o mtodo foi invocado.
     */
    public INode getParentNode() {
        return this.parentNode;
    }

    /**
     * Devolve a lista de declaraes que devem ser processadas caso a avaliao
     * da condio da declarao condicional resulte em <code>true</code>.
     *
     * @return uma lista de declaraes. Um objeto do tipo <code>NodeList</code>.
     */
    public NodeList getTrueDeclarationSet() {
        return this.trueDeclarationSet;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JName</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JName</code>.
     *                Por consequncia, objetos do tipo<code>JName</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException("ConditionalDeclarations should" +
                    " not be matched", this, node);
        }
    }

    /**
     * Realiza o processamento de um n do tipo <code>JConditionalDeclaration</code>.
     * Este consiste em avaliar a condio da declarao condicional. Caso esta
     * avalie para <code>true</code>, as declaraes que se encontram dentro do
     * escopo do comando "if" so processadas e o resultado desse processamento
     *  colocado na lista de filhos do n pai da declarao condicional. Caso
     * a condio avalie para <code>false</code>, o mesmo  feito, s que para
     * as declaraes que se encontram dentro do escopo da clusula "else".
     *
     * @param data Um objeto que pode conter informaes necessrias ao
     *             processamento.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (data == null || !(data instanceof ResultSet)) {
            throw new IllegalArgumentException();
        }

        // Processamento da condio da declarao condicional
        INode cond = this.getCondition();
        Object ret1 = null;

        if (cond instanceof JExpression) {
            ret1 = ((JExpression) cond).evaluate(data);
        }

        if (!(ret1 instanceof Boolean)) {
            throw new ExecutionException(this);
        }

        Boolean simnao = (Boolean) ret1;
        NodeList toBeProcessed = null;

        if (simnao.booleanValue()) {
            toBeProcessed = this.getTrueDeclarationSet();
        } else if (this.getFalseDeclarationSet() != null) {
            toBeProcessed = this.getFalseDeclarationSet();
        }

        NodeList results = new NodeList();
        if (toBeProcessed != null) {
            int max = toBeProcessed.size();
            int i = 0;
            while (i < max) {
                Object res = (toBeProcessed.elementAt(i)).process(data);
                if (!res.equals(this) && res instanceof INode) {
                    results.addElement((INode) res);
                } else if (res instanceof NodeList) {
                    results.addAll((NodeList) res);
                }
                i++;
            }
        }

        this.parentNode.setExecutable(false);
        return results;

    }

    /**
     * Seta a expresso que deve ser avaliada na expresso condicional para
     * determinar que trecho de cdigo (se algum) ser processado.
     *
     * @param cond um n do tipo <code>JExpression</code>. A condio da declarao
     *             condicional.
     */
    public void setCondition(JExpression cond) {
        if (cond == null) {
            throw new IllegalArgumentException();
        }
        this.condition = cond;
    }

    /**
     * Seta a lista de declaraes que devem ser processadas caso a avaliao
     * da condio da declarao condicional resulte em <code>false</code>.
     *
     * @param nodes lista de declaraes. Um objeto do tipo <code>NodeList</code>.
     */
    public void setFalseDeclarationSet(NodeList nodes) {
        this.falseDeclarationSet = nodes;
    }

    /**
     * Seta o n pai de um objeto do tipo <code>JConditionalDeclaration</code>.
     * O "pai", tem uma relao de agregao com o n a partir do qual o mtodo
     * foi invocado.
     *
     * @param parent n pai do n a partir do qual o mtodo foi invocado.
     */
    public void setParentNode(INode parent) {
        if (parent == null) {
            throw new IllegalArgumentException();
        }
        this.parentNode = parent;
    }

    /**
     * Seta a lista de declaraes que devem ser processadas caso a avaliao
     * da condio da declarao condicional resulte em <code>true</code>.
     *
     * @param nodes uma lista de declaraes. Um objeto do tipo <code>NodeList</code>.
     */
    public void setTrueDeclarationSet(NodeList nodes) {
        if (nodes == null) {
            throw new IllegalArgumentException();
        }
        this.trueDeclarationSet = nodes;
    }

    /**
     * Compara dois objetos do tipo JConditionalDeclaration e diz se os dois so
     * iguais.
     */
    public boolean equals(JConditionalDeclaration condDec) {
        boolean retorno = false;
        if (condDec != null) {
            if ((AbstractNode.equals(this.condition, condDec.condition)) &&
                    (AbstractNode.equals(this.trueDeclarationSet, condDec.trueDeclarationSet)) &&
                    (AbstractNode.equals(this.falseDeclarationSet, condDec.falseDeclarationSet)) &&
                    (AbstractNode.equals(this.parentNode, condDec.parentNode))) {
                retorno = true;
            }
        }
        return retorno;
    }

}



