/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe guarda as informacoes a respeito de uma declarao de construtor
 * 
 * @author Adeline de Sousa
 * @version  0.1 -  18/06/2001
 * @since JDK 1.3
 */

public class JConstructorDeclaration extends AbstractNode implements Declaration,
        Commentable, Cloneable {

    // comentrio do construtor
    private JComment comment;

//    private AttributeSectionList attributes;

    // modificador de acesso (private, protected ou public)
    private JModifierList modifiers;

    // Identificador do Construtor - tem que ser o mesmo da classe
    private JIdentifier identifier;

    // Lista de parametros do construtor
    private JParameterList parameters;
    
    // Invocao explcita de construtor (desta classe ou da super-classe)
    private JMethodInvocation constructorInvocation;

    // Lista de excees que podem ser levantadas pelo construtor
    private JNameList exceptions;

    // corpo do contrutor
    private JBlock body;


    /**
     * Construtor vazio da classe
     */
    public JConstructorDeclaration() {
        this.modifiers = new JModifierList(JModifierList.CONSTRUCTOR_DECLARATION);
        this.identifier = new JIdentifier();
        this.parameters = new JParameterList();
        this.exceptions = new JNameList();
        this.setTypeIdentifier(JaTSNode.CONSTRUCTOR_DECLARATION);
        this.body = null;
        this.comment = null;
    }


    /**
     * Construtor da classe que recebe como parametros:
     *
     * @param mods   Lista de modificadores
     * @param id     Identificador do construtor
     * @param params Lista de parametros
     * @param exc    Lista de excecoes
     */
    public JConstructorDeclaration(JModifierList mods, JIdentifier id,
                                   JParameterList params, JNameList exc) {
        this.setModifiers(mods);
        this.setName(id);
        this.setParameters(params);
        this.setExceptions(exc);
        this.setTypeIdentifier(JaTSNode.CONSTRUCTOR_DECLARATION);
        this.comment = null;
    }

	
	/**
	 * Retorna a chamada de construtor explcita deste n, se houver. 
	 * 
	 * @return chamada de construtor ou <code>null</code> se no houver
	 */
    public JMethodInvocation getConstructorInvocation() {
		return constructorInvocation;
	}
	
	/**
	 * Modifica a chamada de construtor explcita utilizada nesta declarao
	 * de construtor.
	 * 
	 * @param constructorInvocation chamada de construtor explcita
	 */
    public void setConstructorInvocation(JMethodInvocation constructorInvocation) {
		this.constructorInvocation = constructorInvocation;
	}
    
	/**
     * Retorna true, caso o n, possua comment
     *
     * @return true, se  o n possui um comment
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return (this.comment != null && this.comment.getText() != null);
    }

    /**
     * Retorna o comment associado a este n
     *
     * @return comment associado a este n
     */
    public JComment getComment() {
        return comment;
    }


//    public AttributeSectionList getAttributes() {
//        return this.attributes;
//    }
//
//    public void setAttributes(AttributeSectionList att) {
//       this.attributes = att;
//    }

    /**
     * Retorna o modificador do construtor
     *
     * @return Modificador do construtor
     */
    public JModifierList getModifiers() {
        return this.modifiers;
    }


    /**
     * Retorna o identificador do construtor
     *
     * @return Identificador do construtor
     */
    public JIdentifier getName() {

        return this.identifier;
    }


    /**
     * Retorna a lista de parametros do construtor
     *
     * @return Lista de Parametros do construtor
     */
    public JParameterList getParameters() {

        return this.parameters;

    }

    /**
     * Retorna a lista de excees do construtor
     *
     * @return Lista de Excees do construtor
     */
    public JNameList getExceptions() {

        return this.exceptions;
    }

    /**
     * Seta o comment associado a esta declarao de construtor
     *
     * @param javadoc comment associado a esta declarao de construtor
     */
    public void setComment(JComment javadoc) {
        if (javadoc != null) {
            this.comment = javadoc;
        }

    }

    /**
     * Seta o comentario associado a esta declarao de construtor
     *
     * @param javadoc comentario associado a esta declarao de construtor
     */
    public void setComment(String javadoc) {
        if (javadoc != null && javadoc.length() > 0) {
            this.comment = new JComment(javadoc);

        }
    }

    /**
     * Seta o modificador do construtor. Um construtor aceita apenas
     * modificadores de acesso
     *
     * @param mod Modificador do construtor
     * @throws IllegalArgumentException - se o modList passado for null.
     */
    public void setModifiers(JModifierList mod)
            throws IllegalArgumentException {

        if (mod != null) {
            if (mod.getPossibleModifiers() == JModifierList.CONSTRUCTOR_DECLARATION) {
                this.modifiers = mod;
//            } else {
//                throw new IllegalArgumentException("Lista de modificadores invalida para construtor");
            }
        } else {
            throw new IllegalArgumentException("Lista de modificadores nula");
        }
    }


    /**
     * Seta o identificador do construtor
     *
     * @param id Identificador do construtor
     * @throws IllegalArgumentException - Se o identificador passado for nulo
     */
    public void setName(JIdentifier id) {

        if (id != null) {
            this.identifier = id;
        } else {
            throw new IllegalArgumentException("Identificador do metodo nulo");
        }
    }

    /**
     * Seta a lista de parametros do construtor
     *
     * @param param Lista de parametros do construtor
     * @throws IllegalArgumentException - Se a lista passada for nula ou se
     *                                  contiver elementos que nao sejam do tipo JFormalParameter
     */
    public void setParameters(JParameterList param) {

        if (param != null) {
            this.parameters = param;
        } else {
            throw new IllegalArgumentException("lista de parametros nula");
        }
    }


    /**
     * Seta a lista de excees levantadas por este construtor
     *
     * @param excs Lista de excees levantadas pelo construtor
     * @throws IllegalArgumentException - se a lista de excees for nula
     */
    public void setExceptions(JNameList excs) {

        this.exceptions = excs;
    }

    /**
     * Retorna o corpo do construtor
     *
     * @return O corpo do construtor
     */
    public JBlock getBody() {
        return body;
    }

    /**
     * Seta o corpo do construtor
     *
     * @param body Corpo do construtor
     */
    public void setBody(JBlock body) {
        this.body = body;

    }

    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        MethodSignature result = null;
        if (this.isASourceNode()) {
            try {
                result = new MethodSignature(JName.unwrapName(this.getName()),
                        this.getParameters());
            } catch (InconsistentNodeException e) {
                result = null;
            }
        }
        return result;
    }

    /**
     * Verifica se o modificador passado como parmetro pertence  lista de
     * modificadores de uma declarao de construtor.
     *
     * @param strMod O modificador que se deseja procurar
     * @return <code>true</code>, se o modificador passado como
     *         parmetro estiver na lista. <code>false</code>, caso contrrio
     * @ exception IllegalArgumentException - Se o modificador passado
     * for invalido
     */
    public boolean hasModifier(String strMod) throws IllegalArgumentException {

        return this.modifiers.hasModifier(strMod);
    }


    /**
     * Adiciona um modificador  lista de modificadores do construtor
     *
     * @param mod Modificador a ser inserido
     * @throws IllegalArgumentException Se o modificador passado j existir na
     *                                  lista de modificadores deste mtodo ou se este no  um
     *                                  modificador vlido para construtores
     */
    public void addModifier(String mod) {
        if (this.hasModifier(mod)) {
            throw new IllegalArgumentException("Modificador ja existe");
        } else {
            this.modifiers.addModifier(mod);
        }
    }


    /**
     * Remove um modificador da lista de modificadores deste construtor.
     *
     * @param strMod Modificador a ser removido
     * @throws IllegalArgumentException - Se o modificador passado nao
     *                                  pertencer a lista de modificadores
     */
    public void removeModifier(String strMod) throws IllegalArgumentException {

        this.modifiers.removeModifier(strMod);
    }


    /**
     * Verifica a lista de modificadores desta declarao de construtor contm
     * um modificador de acesso.
     *
     * @return <code>true</code>, se a lista de modificadores j contm um
     *         modificador de acesso. <code>false</code>, caso contrrio
     */
    public boolean hasAccessModifier() throws IllegalArgumentException {
        return this.modifiers.hasAccessModifier();
    }


    /**
     * Adiciona uma exceo  lista de excees do construtor
     *
     * @param exc Exceo a ser inserida
     * @throws IllegalArgumentException Se o exceo passada j existir na
     *                                  lista de excees deste construtor
     */
    public void addException(String exc) {
        if (this.hasException(exc)) {
            throw new IllegalArgumentException("Excecao ja existe");
        } else {
            JName n = new JName(exc);
            this.exceptions.add(n);
        }
    }

    /**
     * Verifica se uma exceo est na lista de excees deste construtor
     *
     * @param exc exceo  procurada
     * @return true, se a exceo existir
     *         false, caso contrrio
     */
    public boolean hasException(String exc) {

        JName n = new JName(exc);
        return this.exceptions.contains(n);
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore<br>
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz <br>
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns <br>
     * casam em duas situaes: quando eles so completamente idnticos <br>
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda<br>
     * no foi mapeada a nenhum valor no conjunto resultado do casamento <br>
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele<br>
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto<br>
     * acima, se refere ao n que faz parte do programa JaTS que est sendo <br>
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar<br>
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo <br>
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro<br>
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada<br>
     *                                   quando algum componente essencial do n tiver um valor, <br>
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar<br>
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou<br>
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code>  <br>
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }

        if (!(node instanceof JConstructorDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results,
                JaTSNode.CONSTRUCTOR_DECLARATION)) {

            JConstructorDeclaration constDec = (JConstructorDeclaration) node;

            // casa o comentario
            if (this.getComment() == null) {
                if (constDec.getComment() != null) {
                    JComment comment = constDec.getComment();
                    String commentVar = Util.generateCommentVariable(this.getName());
                    results.put(commentVar, comment, JaTSNode.COMMENT);
                }
            }

//            if (this.attributes != null) {
//            	//casa os atributos
//            	this.attributes.match(constDec.attributes, results);
//            } else if (constDec.attributes != null) {
//            	throw new NodesNotMatchedException("Attributes don't match", this, constDec);
//            }

            // casa o modificador
            this.modifiers.match(constDec.modifiers, results);

            // casa o identificador
            this.identifier.match(constDec.identifier, results);

            // casa a lista de parametros
            this.parameters.match(constDec.parameters, results);

            if (this.exceptions != null) {
                // casa a lista de excees
                this.exceptions.match(constDec.exceptions, results);
            } else if (constDec.getExceptions() != null) {
                throw new NodesNotMatchedException("Exceptions thrown by constructors" +
                        "are incompatible", this, constDec);
            }
            if (this.body != null) {
                this.body.match(constDec.body, results);
            }

        }
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.parameters</code> e <code>this.identifier</code>
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo
     *                desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>,que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        Object retorno = null;

        if (!this.isVariable() && !this.isExecutable()) {
            // envia o visitor pela sub-arvore do corpo do construtor
            if (this.body != null) {
                this.body.accept(visitor, data);
            }

            // envia o visitor  pela sub-arvore das excecoes do contrutor
            if (this.exceptions != null) {
                this.exceptions.accept(visitor, data);
            }

            if (this.constructorInvocation != null) {
            	this.constructorInvocation.accept(visitor, data);
            }
            
            // envia o visitor  pela sub-arvore dos parametros do contrutor
            if (this.parameters != null) {
                this.parameters.accept(visitor, data);
            }

            // envia o  visitor pela sub-arvore do identificador
            if (this.identifier != null) {
                this.identifier.accept(visitor, data);
            }
            
            // envia o  visitor pela sub-arvore do modificador
            if (this.modifiers != null) {
                this.modifiers.accept(visitor, data);
            }
            
            // envia o  visitor pela sub-arvore dos atributos
//            if (this.attributes != null) {
//            	this.attributes.accept(visitor, data);
//            }
            
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            //Processamento dos atributos
//        	AttributeSectionList tmpAttribute = this.getAttributes();
//            if (tmpAttribute != null) {
//                Object ret = this.processNode(tmpAttribute, data);
//                if (tmpAttribute.isExecutable()) {
//                    if (!(ret instanceof AttributeSectionList)) {
//                        throw new ExecutionException(this);
//                    }
//                    this.setAttributes((AttributeSectionList) ret);
//                }
//            }

            // Processamento da lista de modificadores do construtor
            INode mods = this.getModifiers();
            if (mods != null) {
                Object ret1 = this.processNode(mods, data);
                if (mods.isExecutable()) {
                    if (!(ret1 instanceof JModifierList)) {
                        throw new ExecutionException(this);
                    }

                    this.setModifiers((JModifierList) ret1);
                }
            }

            // Processamento do identificador do construtor
            INode id = this.getName();
            if (id != null) {
                Object ret2 = this.processNode(id, data);
                if (id.isExecutable()) {
                    try {
                        this.setName(JIdentifier.unwrapIdentifier(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(this);
                    }
                }
            }

            // Processamento dos parametros do construtor
            INode params = this.getParameters();
            if (params != null) {
                Object ret3 = this.processNode(params, data);
                if (params.isExecutable()) {
                    if (ret3 instanceof JParameterList) {
                        this.setParameters((JParameterList) ret3);
                    } else {
                        throw new ExecutionException(this);
                    }
                }
            }

            // Processamento da lista de excees que o construtor levanta.
            INode exceptList = this.getExceptions();
            if (exceptList != null) {
                Object ret4 = this.processNode(exceptList, data);
                if (exceptList.isExecutable()) {
                    if (ret4 instanceof JNameList) {
                        this.setExceptions((JNameList) ret4);
                    } else {
                        throw new ExecutionException(this);
                    }
                }
                this.getExceptions().setOptional(false);
            }
            // Processamento do corpo do construtor
            INode cBody = this.getBody();
            if (cBody != null) {
                Object ret5 = this.processNode(cBody, data);

                if (cBody.isExecutable()) {
                    if (ret5 instanceof JBlock) {
                        this.setBody((JBlock) ret5);

                    } else {
                        throw new ExecutionException(this);
                    }
                }
            }


        }
        return retorno;
    }


    /**
     * Verifica se dois objetos do tipo ConstructorDeclaration so idnticos
     *
     * @return true, se os objetos forem identicos <br>
     *         false, caso contrrio
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JConstructorDeclaration) {
            JConstructorDeclaration cd = (JConstructorDeclaration) obj;
            if (!this.isVariable() && !this.isExecutable()) {
                retorno =
                        AbstractNode.equals(this.modifiers, cd.modifiers) &&
                        AbstractNode.equals(this.identifier, cd.identifier) &&
                        AbstractNode.equals(this.parameters, cd.parameters) &&
                        AbstractNode.equals(this.exceptions, cd.exceptions) &&
                        AbstractNode.equals(this.body, cd.body);
            } else {
                retorno = compareJaTSNode(cd);
            }
        }
        return retorno;

    }
}