/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.CloningVisitor;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.visitors.ReplacementVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa declaraes executveis em JaTS. Uma declarao
 * executvel  uma declarao onde cdigo Java  executado como num programa
 * Java qualquer.
 *
 * @author Fernando Castor
 * @see INode
 */
public class JExecutableDeclaration extends AbstractNode {

    // expresses do tipo: #<#A.getType()>#
    public static final int INFORMATION_EXTRACTING = 1;

    // expresses do tipo: #<#temp :: #A = #temp.setType("int")>#
    public static final int INFORMATION_MODIFYING = 2;


    // Esta lista de ns contm as expresses que sero executadas quando este
    // n for processado.
    private NodeList executableDeclarations;

    // Em declaraes executveis do tipo Information-Modifying,  necessrio que
    // seja mantido um apontador para o n a ser devolvido. Este atributo
    // armazena esse apontador.
    private INode resultNode;

    // Varivel que ser referenciada ao invs de resultNode nas expresses que
    // sero executadas.
    private INode referencedVariable;

    //  Guarda o valor de resultNode antes de ele ter sido modificado pela ltima
    //  vez.
    private INode oldResultNode;

    // guarda o tipo da expresso
    private int expressionType;

    /**
     * Construtor da classe que recebe como argumento apenas uma expresso. Esse
     * construtor  usado quando a declarao executvel  do tipo
     * Information-Extracting.
     *
     * @param exec Uma expresso que ser posteriormente avaliada.
     * @throws IllegalArgumentException Levantada quando <code>exec</code> 
     *                                  <code>null</code>.
     */
    public JExecutableDeclaration(JExpression exec)
            throws IllegalArgumentException {

        if (exec == null) {
            throw new IllegalArgumentException();
        }
        this.executableDeclarations = new NodeList();
        this.resultNode = null;
        this.executableDeclarations.addElement(exec);
        this.oldResultNode = null;
        this.expressionType = INFORMATION_EXTRACTING;

    }


    /**
     * Construtor da classe que recebe como argumento um n e uma lista de
     * expresses. Esse construtor  usado quando a declarao executvel  do
     * tipo Information-Modifying. O n <code>resNode</code>  usado para
     * armazenar o resultado da avaliao da lista de expresses
     * <code>execDecs</code>.
     *
     * @param rv       O nome da varivel que ser referenciada durante o processamento
     *                 deste n.
     * @param resNode  O n que deve ser retornado como resultado do processamento
     *                 deste n.
     * @param execDecs Uma lista de expresses que so executadas sequencialmente
     *                 durante o processamento do n.
     * @throws IllegalArgumentException Levantada quando <code>exec</code> 
     *                                  <code>null</code>.
     */
    public JExecutableDeclaration(INode rv, INode resNode, NodeList execDecs)
            throws IllegalArgumentException {

        if (resNode == null || execDecs == null || rv == null)
            throw new IllegalArgumentException();

        this.executableDeclarations = execDecs;
        this.resultNode = resNode;
        this.referencedVariable = rv;
        this.expressionType = INFORMATION_MODIFYING;
    }


    /**
     * Retorna o tipo da expresso, se  Information Extracting ou
     * Information Modifying
     *
     * @return Tipo da expresso
     */
    public int getExpressionType() {
        return expressionType;
    }

    /**
     * Seta o tipo da expresso executvel
     *
     * @param expType Tipo da expresso
     */
    public void setExpressionType(int expType) {
        if (expType == INFORMATION_EXTRACTING ||
                expType == INFORMATION_MODIFYING) {
            this.expressionType = expType;
        } else {
            // por default, sao do tipo Information - modifying
            this.expressionType = INFORMATION_MODIFYING;
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // ESTE TRECHO DE CDIGO  UMA GAMBIARRA MALDITA E DEMONACA QUE
        // DEVE SUMIR NO FUTURO!!!!

        // Aqui talvez possa ser usado o padro de projeto Mediator!!!!!!!

        if (visitor instanceof ReplacementVisitor) {
            initReplacement((ReplacementVisitor) visitor);
        }
        // Fim da gambiarra assassina.

        if (this.resultNode != null) {
            this.resultNode.accept(visitor, data);
        }

        if (this.executableDeclarations == null) {
            throw new InconsistentNodeException(this);
        }

        int i = this.executableDeclarations.size() - 1;

        while (i >= 0) {
            this.executableDeclarations.elementAt(i).accept(visitor, data);
            i--;
        }

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /*
     * Mtodo auxiliar que realiza a execuo da expresso na posio i da lista
     * de expresses deste n.
    */
    private Object executeExpressionAt(int i, ResultSet rset) throws InconsistentNodeException,
            ExecutionException, IllegalArgumentException {

        Object retorno = null;
        INode tempNode = this.executableDeclarations.elementAt(i);

        if (tempNode == null || !(tempNode instanceof JExpression)) {
            throw new InconsistentNodeException(this);
        }
        JExpression exec = (JExpression) tempNode;
        retorno = exec.evaluate(rset);

        return retorno;
    }


    /**
     * Retorna o <code>i</code>-simo (com <code>i</code> comeando em zero)
     * elemento da lista de expresses desta declarao executvel.
     *
     * @param i O ndice da expresso a ser retornada.
     * @return JExpression Uma das expresses da lista de expresses deste n.
     * @throws IllegalArgumentException  Levantada se o valor de <code>i</code>
     *                                   fornecido for negativo.
     * @throws InconsistentNodeException Levantada se o elemento pedido da
     *                                   lista de expresses desse n no for uma expresso.
     */
    public JExpression getExpressionAt(int i) throws IllegalArgumentException,
            InconsistentNodeException {

        INode retorno = this.executableDeclarations.elementAt(i);

        if (!(retorno instanceof JExpression))
            throw new InconsistentNodeException(this);

        return (JExpression) retorno;
    }

    /**
     * Retorna o nome da varivel a ser referenciada dentro das expresses a ser
     * executadas no lugar de uma varivel qualquer. Essa varivel tem escopo
     * local  declarao executvel e  usada como um placeholder, a fim de que
     * o valor mapeado  varivel original no seja modificado.
     *
     * @return o nome da varivel.
     */
    public INode getReferencedVariable() {
        return this.referencedVariable;
    }

    /**
     * Retorna o n resultado da declarao executvel.
     *
     * @return Um <code>INode</code> correspondendo ao resultado do
     *         processamento deste n.
     * @throws IllegalArgumentException Levantada se o valor de <code>i</code@
     *                                  fornecido for negativo.
     */
    public INode getResultNode() {
        return this.resultNode;
    }


    /**
     * Mtodo auxiliar para o processo de substituio. Dever no futuro ser
     * eliminado em favor de uma soluo mais inteligente.
     */
    public void initReplacement(ReplacementVisitor vis) throws
            IllegalArgumentException, ExecutionException,
            InconsistentNodeException {
        if (vis == null) {
            throw new IllegalArgumentException();
        }

        ResultSet res = vis.getResultSet();
        CloningVisitor cvis = new CloningVisitor();
        INode newValue = null;
        Object obj = null;

        // qualquer problema neste mtodo, remeter aa versao versao_2-3 do CVS
        if (this.resultNode != null && this.resultNode.isVariable()) {
            obj = res.get(this.referencedVariable);
            if (res.hasValue(this.resultNode)) {
                JIdentifier resultNodeVariable =
                        new JIdentifier(this.resultNode.getVariableName());
                res.removeResult(resultNodeVariable);
            }
            if (obj != null && obj instanceof INode) {
                ((INode) obj).accept(cvis, null);
                newValue = cvis.getClone();
                res.put(this.resultNode, newValue);

            }

        }
        // Se resultNode no  null mas tambm no  uma varivel,
        // isso significa que uma substituio j foi realizada nessa
        // declarao executvel.

        if (this.referencedVariable != null && this.referencedVariable.isVariable()) {
            obj = res.get(this.referencedVariable);
            if (obj instanceof INode) {
                this.referencedVariable = (JaTSNode) obj;
            }
        }

        // varre a lista de expresses, chamando o replacement para cada uma
        //delas
        int size = this.executableDeclarations.size();
        for (int i = 0; i < size; i++) {
            JExpression tmpExp = this.getExpressionAt(i);
            // isto  uma gambi. Deve ser removida em favor de uma soluo mais
            // adequada
            if (tmpExp instanceof JMethodInvocation) {
                ((JMethodInvocation) tmpExp).initReplacement(vis);
            }
        }

    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
    }

    /**
     * O processamento de ns do tipo <code>Executable Declaration</code> 
     * completamente diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore sinttica
     * do programa. Em declaraes executveis, as rvores sintticas
     * correspondentes s expresses a ser executadas so processadas de modo que
     * o resultado gerado seja equivalente ao que seria se o cdigo Java dessas
     * expresses fosse executado. Por exemplo, o processamento da rvore
     * sinttica de uma invocao de mtodo fornece como resultado o mesmo
     * resultado que seria fornecido se esse mtodo estivesse sendo invocado
     * dentro de um programa em Java.
     * <br><br>
     * Dois tipos de processamento so possveis para uma declarao executvel.
     * No primeiro, o resultado do processamento da declarao  o resultado
     * retornado pela invocao de um nico mtodo. Declaraes executveis desse
     * tipo tm como objetivo apenas a extrao de informao, sem modific-la, e
     * so chamadas <i>Information-Extracting Declarations</i>. Em contrapartida,
     * o segundo tipo de declarao executvel tem como objetivo modificar uma
     * determinada informao. O resultado, nesse segundo caso,  o prprio
     * objeto modificado. Declaraes executveis desse segundo tipo so chamadas
     * de <i>Information-Modifying Declarations</i>.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Caso o
     *          parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada quando um n inconsistente (incompleto) for encontrado.
     *          Por exemplo, se uma declarao de atributo no possuir um filho
     *          correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;
        // Information-Modifying Declarations
        if (this.resultNode != null && this.executableDeclarations.size() >= 1) {
            int i = 0;
            int max = this.size();

            // codigo anterior - verificar se vai continuar funcionando para
            // os outros casos
            while (i < max) {
                this.executeExpressionAt(i, (ResultSet) data);
                i++;
            }
            retorno = this.resultNode;

        }
        // Information-Extracting Declarations
        else if (this.executableDeclarations.size() == 1) {
            retorno = this.executeExpressionAt(0, (ResultSet) data);
        }

        // Nenhum dos dois
        else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
        }

        return retorno;
    }

    /**
     * Coloca uma nova expresso no lugar de uma j existente na posio
     * <code>i</code> na lista de expresses deste n.
     *
     * @param exp A nova expresso a ser inserida.
     * @param i   A posio na lista de expresses deste n onde a nova expresso
     *            ser inserida. O valor inicial  0.
     * @throws IllegalArgumentException Levantada se o valor de <code>i</code>
     *                                  fornecido for negativo ou maior que o nmero de expresses j
     *                                  existente ou no caso de <code>exp</code> ser <code>null</code>.
     */
    public void setExpressionAt(JExpression exp, int i)
            throws IllegalArgumentException {

        if (exp == null || i < 0 || i > this.executableDeclarations.size()) {
            throw new IllegalArgumentException();
        }
        this.executableDeclarations.setElementAt(exp, i);
    }

    /**
     * Seta o n resultado da declarao.
     *
     * @param resNode Um <code>JaTSNode</code> que ser modificado e retornado
     *                como o resultado do processamento deste n.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>resNode</code> fornecido for <code>null</code>.
     */
    public void setResultNode(INode resNode) throws IllegalArgumentException {
        if (resNode == null) {
            throw new IllegalArgumentException();
        }
        if (this.resultNode.isVariable()) {
            this.oldResultNode = this.resultNode;
        }
        this.resultNode = resNode;

    }

    /**
     * Devolve o nmero de expresses que essa declarao executvel encapsula.
     *
     * @return Um inteiro correspondendo ao nmero de expresses que faz parte
     *         desta declarao executvel.
     */
    public int size() {
        int retorno = 0;
        if (this.executableDeclarations != null) {
            retorno = executableDeclarations.size();
        }
        return retorno;
    }

    /**
     * Esse mtodo retira do conjunto resultado o mapeamento entre a varivel
     * resultado da declarao executvel e o n mapeado a ela no contexo dessa
     * declarao.
     *
     * @param res O conjunto resultado de onde o mapeamento deve ser retirado.
     * @throws IllegalArgumentException Levantada quando o conjunto resultado
     *                                  fornecido for <code>null</code>.
     */
    public void unmapOldResultNode(ResultSet res)
            throws IllegalArgumentException, InconsistentNodeException {

        if (res == null) {
            throw new IllegalArgumentException();
        }

        if (this.oldResultNode != null && this.oldResultNode.isVariable()) {
            if (res.alreadyMatched(this.oldResultNode)) {
                res.removeResult(this.oldResultNode);
            }
        } else if (oldResultNode != null) {
            throw new InconsistentNodeException(this);
        }
    }


    /**
     * Verifica se dois objetos do tipo JExecutableDeclaration so iguais.
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JExecutableDeclaration) {
            JExecutableDeclaration jexec = (JExecutableDeclaration) obj;
            // compara inicialmente o tipo e a lista de expresses
            retorno = this.expressionType == jexec.expressionType &&
                    AbstractNode.equals(this.executableDeclarations, jexec.executableDeclarations);

            // se for do tipo information-modifying, compara tambm o n resultante e
            // a varivel referenciada
            if (this.getExpressionType() == JExecutableDeclaration.INFORMATION_MODIFYING) {
                retorno = retorno &&
                        AbstractNode.equals(this.resultNode, jexec.resultNode) &&
                        AbstractNode.equals(this.referencedVariable, jexec.referencedVariable);
            }
        }
        return retorno;
    }

}
