/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe implementa uma lista de expresses. Para JaTS/Java, uma lista de
 * argumentos nada mais  do que uma lista de expresses.
 *
 * @author Fernando Castor
 * @author Adeline de Sousa
 * @see JExpression
 * @version  0.1 -  02/11/2001
 * @since JDK 1.3
 */
public class JExpressionList extends AbstractNode {

    public static final int ARGUMENTS = 1;
    public static final int ARRAY_INDEXERS = 2;

    // Lista contendo as expresses correspondentes aos argumentos.
    private NodeList expressions;
    private int type;


    /**
     * Construtor vazio da classe
     */
    public JExpressionList() {
        this.expressions = new NodeList();
        this.setTypeIdentifier(JaTSNode.EXPRESSION_LIST);
        this.setType(ARGUMENTS);
    }


    /**
     * Construtor vazio da classe
     */
    public JExpressionList(int t) {
        this();
        this.setType(t);
    }


    /**
     * Retorna o tipo desta lista de expressoes
     *
     * @return
     */
    public int getType() {
        return type;
    }


    /**
     * Seta o tipo da lista de expressoes.
     *
     * @param type
     */
    public void setType(int type) {
        if (type == ARGUMENTS || type == ARRAY_INDEXERS) {
            this.type = type;
        } else {
            this.type = ARGUMENTS;
        }
    }

    /**
     * Adiciona um novo argumento a esta lista de argumentos.
     *
     * @param arg Uma expresso correspondendo a um argumento que ser passado
     *            para um mtodo ou construtor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>arg</code> for <code>null</code>.
     */
    public void addExpression(JExpression arg) {
        if (arg == null) {
            throw new IllegalArgumentException();
        }
        this.expressions.addElement(arg);
    }

    /**
     * Retorna o argumento que se encontra na posio fornecida.
     *
     * @param pos O ndice do argumento desejado.
     * @return Uma expresso correspondendo ao argumento que se encontra na
     *         posio <code>pos</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>pos</code> for um inteiro negativo.
     */
    public JExpression getExpressionAt(int pos) throws IllegalArgumentException {

        boolean validPosition = (pos >= 0 && pos < expressions.size());
        if (!validPosition) {
            throw new IllegalArgumentException();
        }
        Object tempObj = this.expressions.elementAt(pos);
        JExpression retorno = null;
        if (tempObj instanceof JExpression) {
            retorno = (JExpression) tempObj;
        }

        return retorno;
    }


    /**
     * Remove o argumento que se encontra na posio fornecida, caso haja
     * algum.
     *
     * @param pos O ndice onde o argumento desejado se encontra.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>pos</code> for um inteiro negativo.
     */
    public void removeExpressionAt(int pos) {
        boolean validPosition = (pos >= 0 && pos < expressions.size());
        if (!validPosition) {
            throw new IllegalArgumentException();
        }
        this.expressions.removeElementAt(pos);
    }

    /**
     * Adiciona um novo argumento a esta lista de argumentos, sendo que numa
     * posio pr-determinada.
     *
     * @param arg A expresso a ser adicionada.
     * @param pos O ndice onde o argumento fornecido deve ser colocado, nesta
     *            lista de argumentos. A posio inicial  zero.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>arg</code> for <code>null</code> ou o argumento <code>pos</code>
     *          for um inteiro negativo ou no tiver um valor vlido.
     */
    public void setExpressionAt(JExpression arg, int pos) {
        boolean validPosition = (pos >= 0 && pos < expressions.size());
        if (arg == null || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.expressions.setElementAt(arg, pos);
    }

    /**
     * Retorna o nmero de argumentos presentes nesta lista de argumentos.
     *
     * @return Um inteiro correspondendo ao nmero de argumentos desta lista.
     */
    public int size() {
        int retorno = 0;
        if (this.expressions != null) {
            retorno = this.expressions.size();
        }
        return retorno;
    }

    /**
     * Retorna a primeira expresso da lista
     *
     * @return A primeira expresso da lista, ou null, se a lista for vazia.
     */
    public JExpression firstExpression() {
        return (JExpression) this.expressions.first();
    }

    /**
     * Retorna a ltima expresso da lista
     *
     * @return A ltima expresso da lista, ou null, se a lista for vazia.
     */
    public JExpression lastExpression() {
        return (JExpression) this.expressions.last();
    }

    public JaTSIterator getIterator() {
        return this.expressions.getIterator();
    }


    public NodeList getExpressions() {
        return this.expressions;
    }

    /**
     * Compara dois objetos do tipo JExpressionList e diz se estes so iguais
     *
     * @param obj Objeto a ser comparado com o objeto atual (this)
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JExpressionList) {
            JExpressionList exprList = (JExpressionList) obj;
            retorno = AbstractNode.equals(this.expressions, exprList.expressions);
        }
        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JExpressionList)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JExpressionList args = (JExpressionList) node;
            this.expressions.match(args.getExpressions(), results);
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * cada elemento <code>this.expressions</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa
     *                uma operao sobre vrios tipos diferentes de objetos fora do
     *                cdigo desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>, que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        int i = this.expressions.size() - 1;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            while (i >= 0) {
                (this.expressions.elementAt(i)).accept(visitor, data);
                i--;
            }
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            int max = this.size();

            while (i < max) {
                INode currArg = this.getExpressionAt(i);
                if (currArg != null) {
                    Object ret = this.processNode(currArg, data);

                    if (currArg.isExecutable()) {
                        if (!(ret instanceof JExpression))
                            throw new ExecutionException(this);

                        this.setExpressionAt((JExpression) ret, i);
                    }
                }
                i++;
            }
        }
        return retorno;
    }


    /**
     * Retorna um array de object que corresponde a esta lista de expresses.
     * til na avaliao de expresses
     *
     * @return Array de objetos correspondente a esta lista de expressoes
     */
    public Object[] toObjectArray() {
        Object[] result = new Object[this.size()];
        for (int i = 0; i < result.length; i++) {
            result[i] = this.getExpressionAt(i);
        }
        return result;
    }

}
