/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.lang.reflect.Field;


/**
 * Classe responsvel pela implementao de acessos a atributos, tanto para
 * leitura quanto para escrita.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class JFieldAccess extends JExpression implements Commentable {

    // Um nome representando o caminho a partir do qual o mtodo ser invocado
    // e, possivelmente, seu nome.
    private INode referencedObject;

    // O nome do atributo sendo acessado.
    private JIdentifier fieldName;

    // comentario
    private JComment comment;


    /**
     * Construtor vazio da classe
     */
    public JFieldAccess() {
        this.referencedObject = null;
        this.fieldName = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro uma <code>JExpression</code>
     * e um <code>JName</code> correspondendo ao nome do atributo.
     *
     * @param refObject Uma <code>JExpression</code> denotando o objeto a partir
     *                  do qual o atributo est sendo acessado.
     * @param mn        Um objeto do tipo <code>JName</code> correspondendo ao nome do
     *                  atributo.
     */
    public JFieldAccess(INode refObject, JIdentifier mn) {
        this.referencedObject = refObject;
        this.fieldName = mn;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }


    /**
     * Retorna o nome do atributo a ser acessado e o caminho a partir do qual
     * essa invocao ocorrer.
     *
     * @return Um <code>JName</code> correspondendo ao nome do mtodo deve ser
     *         invocado junto com o caminho da invocao.
     */
    public JIdentifier getFieldName() {
        return this.fieldName;
    }

    /**
     * Retorna o objeto a partir do qual o atributo ser acessado.
     *
     * @return Um <code>INode</code> correspondente ao objeto a partir
     *         do qual o atributo ser acessado ou  expresso que, uma vez
     *         avaliada, produzir esse objeto como resultado.
     */
    public INode getReferencedObject() {
        return this.referencedObject;
    }


    /**
     * Seta o nome do atributo sendo acessado.
     *
     * @param mn Um objeto do tipo <code>JName</code> correspondendo ao nome do
     *           atributo a ser acessado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>mn</code> for <code>null</code>.
     */
    public void setFieldName(JIdentifier mn) {
        this.fieldName = mn;
    }

    /**
     * Seta o objeto a partir do qual o atributo ser acessado.
     *
     * @param refObj Um <code>INode</code> correspondente ao objeto a partir
     *               do qual o atributo ser acessado. <code>refObj</code> pode
     *               corresponder tambm a uma expresso que ser avaliada. Nesse caso
     *               o atributo  invocado a partir do objeto resultante dessa avaliao.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>refObj</code> for <code>null</code>.
     */
    public void setReferencedObject(INode refObj) {
        if (refObj == null) {
            throw new IllegalArgumentException();
        }

        this.referencedObject = refObj;
    }

    /**
     * Retorna true, caso o n, possua javadoc
     *
     * @return true, se  o n possui um javadoc
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return this.comment != null;
    }

    /**
     * Retorna o javadoc associado a este n
     *
     * @return javadoc associado a este n
     */
    public JComment getComment() {
        return comment;
    }

    /**
     * Seta o javadoc associado a este n
     *
     * @param comment javadoc associado a este n
     */
    public void setComment(JComment comment) {
        this.comment = comment;
    }

    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        throw new UnsupportedOperationException("not applicable");
    }

    /**
     * Compara dois objetos do tipo JFieldAccess e diz se os dois so iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JFieldAccess) {
            JFieldAccess fieldAcc = (JFieldAccess) obj;
            if (this.isASourceNode()) {
                retorno =
                        AbstractNode.equals(this.referencedObject, fieldAcc.getReferencedObject()) &&
                        AbstractNode.equals(this.fieldName, fieldAcc.getFieldName());

            } else {
                retorno = compareJaTSNode(fieldAcc);
            }
        }
        return retorno;

    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.referencedObject</code>, de <code>fieldName</code> e de
     * <code>this.arguments</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            if (this.fieldName == null) {
                throw new InconsistentNodeException(this);
            }

            if (this.referencedObject != null)
                this.referencedObject.accept(visitor, data);

            this.fieldName.accept(visitor, data);
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);


        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando algum componente
     *                                   essencial do n tiver um valor invlido, tornando o n, como um
     *                                   todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JFieldAccess)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JFieldAccess fieldAccess = (JFieldAccess) node;

            if (this.getFieldName() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }

            if (this.getReferencedObject() != null) {
                (this.getReferencedObject()).match(fieldAccess.getReferencedObject(),
                        results);
            }
            // Se o "referencedObject" desse objeto  null mas o de fieldAccess no,
            // ento o casamento falhou.  (podia verificar se o "refObj" 
            //"this" e, neste caso, aceitar o casamento)
            else if (fieldAccess.getReferencedObject() != null) {
                throw new NodesNotMatchedException("Field access didn't match",
                        this, node);
            }

            (this.getFieldName()).match(fieldAccess.getFieldName(), results);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do nome do mtodo a ser invocado.
            INode fName = this.getFieldName();
            if (fName != null) {
                Object ret1 = this.processNode(fName, data);
                if (fName.isExecutable()) {
                    try {
                        this.setFieldName(JIdentifier.unwrapIdentifier(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento do objeto a partir do qual o mtodo  invocado.
            INode rObj = this.getReferencedObject();
            if (rObj != null) {
                Object ret3 = this.processNode(rObj, data);
                if (rObj.isExecutable()) {
                    if (!(ret3 instanceof INode)) {
                        throw new ExecutionException(this);
                    }
                    this.setReferencedObject((INode) ret3);
                }
            }
        }

        return retorno;
    }


    public Object evaluate(Object data) throws ExecutionException {
        Object result = this;

        if (this.isASourceNode()) {
            result = null;
            String fName = this.fieldName.toString();
            Object refObj = this.referencedObject;

            if (refObj != null) {
                Class classe = null;
                //acho que devia ser refObj.getClass().toString()
                String className = refObj.toString();

                try {
                    ClassLoader cl = refObj.getClass().getClassLoader();
                    classe = Class.forName(className, false, cl);
                    Field attr = null;
                    attr = classe.getField(fName);
                    result = attr.get(refObj);
                } catch (ClassNotFoundException e) {
                    e.printStackTrace();
                    throw new ExecutionException(e.getMessage(), this);
                } catch (NoSuchFieldException e) {
                    e.printStackTrace();
                    throw new ExecutionException(e.getMessage(), this);
                } catch (SecurityException e) {
                    e.printStackTrace();
                    throw new ExecutionException(e.getMessage(), this);
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                    throw new ExecutionException(e.getMessage(), this);
                }
            }

        }
        return result;
    }

}
