/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe representa um declaracao de atributo. Por exemplo:
 * <code> private int numero; </code>
 * <code> private int x,y; </code>
 * 
 * @author Adeline de Sousa Silva
 */

public class JFieldDeclaration extends JVariableDeclaration
        implements Cloneable, InterfaceMember {

    public static final String GET_PREFIX = "get";
    public static final String SET_PREFIX = "set";

    /**
     * Construtor vazio da classe.
     */
    public JFieldDeclaration() {
        this.setTypeIdentifier(JaTSNode.FIELD_DECLARATION);

    }


    /**
     * Construtor da classe que recebe o tipo  e a lista de variaveis.
     *
     * @param t      tipo da(s) variavel(eis)
     * @param varDec lista  de variaveis
     */
    public JFieldDeclaration(JType t, JVariableDeclaratorSet varDec) {
        this.type = t;
        this.variables = varDec;
        this.setTypeIdentifier(JaTSNode.FIELD_DECLARATION);
    }



    /**
     * Verifica a lista de modificadores desta declarao de atributo contm um
     * modificador de acesso.
     *
     * @return <code>true</code>, se a lista de modificadores j contm um
     *         modificador de acesso. <code>false</code>, caso contrrio.
     */
    public boolean hasAccessModifier() throws IllegalArgumentException {

        return this.modifiers.hasAccessModifier();
    }

    /**
     * Diz se este atributo  uma constante ou no. Verificando se esta
     * declarao contm o modificador "final"
     *
     * @return <code>true</code>, se a lista de modificadores contiver o
     *         modificador final. <code>false</code>, caso contrrio.
     */
    public boolean isConstant() throws IllegalArgumentException {

        return this.modifiers.hasModifier(JModifier.FINAL);
    }

    /**
     * Verifica se a declarao de atributo  pblica.
     *
     * @return <code>true</code>, se a declarao de atributo for publica
     *         <code>false</code>, caso contrrio.
     */
    public boolean isPublic() throws IllegalArgumentException {

        return this.modifiers.hasModifier(JModifier.PUBLIC);
    }



    /**
     * Seta a lista de modificadores do atributo
     *
     * @param modList Lista de modificadores deste atributo.
     * @throws IllegalArgumentException - se o modList passado for null
     */
    public void setModifiers(JModifierList modList)
            throws IllegalArgumentException {

        if (modList != null &&
                modList.getPossibleModifiers() == JModifierList.FIELD_DECLARATION) {
            this.modifiers = modList;
        }
    }


    /**
     * Retorna um inteiro que identifica o tipo do atributo. No caso de
     * o tipo corresponder a um objeto qualquer, simplesmente retorna
     * object. Caso contrrio retorna um dos tipos primitivos ou void.
     * Tipos passveis de serem retornados:
     * <p/>
     * JType.VOID
     * JType.BOOLEAN
     * JType.BYTE
     * JType.CHAR
     * JType.SHORT
     * JType.INT
     * JType.LONG
     * JType.FLOAT
     * JType.DOUBLE
     * JType.OBJECT
     */
    public int getTypeCode() {
        return this.type.getType();
    }

    /**
     * Retorna o tipo do atributo
     *
     * @return Tipo do atributo
     */
    public JType getMemberType() {

        return this.type;
    }


    /**************************************************************************/
    /**************                    API                    *****************/
    /**************************************************************************/

    /**
     * Retorna um objeto do tipo <code>JFormalParameter</code> correspondente
     * a esta declarao de atributo. Assim, se tivermos a declarao:
     * "private int x", a chamada a este mtodo ir retornar: "int x". Este
     * mtodo s dever ser chamado para declaraes contendo um nico
     * declarador de varivel.
     *
     * @return um objeto do tipo <code>JFormalParameter</code>  correspondente
     *         a este atributo
     */
    public JFormalParameter toParameter() {
        JFormalParameter retorno = new JFormalParameter();
        retorno.setType(this.getType());
        retorno.setIdentifier(this.getName(0));
        return retorno;
    }
    
//    public PropertyDeclaration toPropertyDeclaration() {
//    	return toPropertyDeclaration(true, true);
//    }
    
//    public PropertyDeclaration toPropertyDeclaration(boolean generateGet, boolean generateSet) {
//    	PropertyDeclaration result = new PropertyDeclaration();
//    	
//    	JIdentifier name = this.getName(0);
//    	
//    	JModifierList mods = new JModifierList(JModifierList.PROPERTY_DECLARATION);
//    	mods.addModifier(JModifier.PUBLIC);
//    	result.setModifiers(mods);
//    	result.setType(this.getType());
//    	result.setName(new JName(name.toClassName()));
//    	
//    	if (generateGet) {
//    		PropertySection getSection = new PropertySection();
//    		getSection.setType(PropertySection.GET);
//    		JBlock getBody = new JBlock();
//    		JResultStatement stmt = new JResultStatement();
//    		JExpression expr = new JFieldAccess(new JIdentifier("this"), name);
//    		stmt.setExpression(expr);
//    		getBody.add(stmt);
//    		getSection.setBlock(getBody);
//    		result.setGetSection(getSection);
//    	}
//    	
//    	if (generateSet) {
//    		PropertySection setSection = new PropertySection();
//    		setSection.setType(PropertySection.SET);
//    		JBlock getBody = new JBlock();
//    		JExpression exprl = new JFieldAccess(new JIdentifier("this"), name);
//    		JExpression exprr = new JIdentifier("value");
//    		JBinaryExpression expr = new JBinaryExpression(JBinaryExpression.ASSIGN, exprl, exprr);
//    		getBody.add(expr);
//    		setSection.setBlock(getBody);
//    		result.setSetSection(setSection);
//    	}
//    	
//    	return result;
//    }

    /**
     * Retorna um objeto do tipo <code>JParameterList</code> correspondente
     * a esta declarao de atributo. Assim, se tivermos a declarao:
     * "private int x, y", a chamada a este mtodo ir retornar: "int x, int y".
     *
     * @return um objeto do tipo <code>JParameterList</code>  correspondente
     *         a este atributo
     */
    public JParameterList toParameterList() {
        JParameterList retorno = new JParameterList();
        JFormalParameter tmpParam = new JFormalParameter();
        JVariableDeclarator varDec = null;
        int max = this.size();
        for (int i = 0; i < max; i++) {
            tmpParam = new JFormalParameter();
            varDec = this.getVariableDeclaratorAt(i);
            tmpParam.setIdentifier(varDec.getIdentifier());
            tmpParam.setType(this.getType());
            retorno.add(tmpParam);
        }

        return retorno;
    }

    /**
     * Retorna o identificador do mtodo default para consulta em um ResultSet
     * (metodos getString, getDouble, etc).
     *
     * @return O nome do mtodo de consulta (SQL) correspondente ao tipo do
     *         atributo
     */
    public JIdentifier getDefaultQueryMethod() {
        JIdentifier retorno = new JIdentifier();
        String strType = this.getType().getStrType();
        if (strType != null && this.getTypeCode() != JType.OBJECT) {
            JIdentifier tmpName = JIdentifier.unwrapIdentifier(this.getType());
            retorno = tmpName.addPrefix(GET_PREFIX);
        }
        return retorno;
    }

    /**
     * Retorna o identificador do mtodo default para definir uma consulta em
     * um comando SQL (metodos setString, setDouble, etc).
     *
     * @return O nome do mtodo de consulta (SQL) correspondente ao tipo do
     *         atributo
     */
    public JIdentifier getDefaultSetMethod() {
        JIdentifier retorno = new JIdentifier();
        String strType = this.getType().getStrType();
        if (strType != null && this.getTypeCode() != JType.OBJECT) {
            JIdentifier tmpName = JIdentifier.unwrapIdentifier(this.getType());
            retorno = tmpName.addPrefix(SET_PREFIX);
        }
        return retorno;
    }


    /******************************************************************************/
    /**************                  Fim da   API                 *****************/
    /******************************************************************************/



    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.modifiers</code>, <code>this.type</code> e
     * <code>this.variables</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                *objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {

            // Envie o visitor pela sub-arvore das declaracoes de variavel
            // pega o tamanho do nodeList, subtrrai um para poder faze-lo de
            // indice, ja que o indice, comeca em zero
            int i = (this.variables.size()) - 1;
            while (i >= 0) {
                // envia o visitor por cada n do tipo JVariableDeclarator,
                // comecando pelo que est na posicao de mais alto valor
                this.variables.elementAt(i).accept(visitor, data);
                i--;
            }
            if (this.type != null) {
                this.type.accept(visitor, data);
            } else if (!this.isExecutable()) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }
            if (this.modifiers != null) {
                this.modifiers.accept(visitor, data);
            } else if (!this.isExecutable()) {
                throw new InconsistentNodeException(this);
            }
            // Envie o visitor pela sub-arvore da lista de atributos
//            if (this.getAttributes() != null) {
//            	this.getAttributes().accept(visitor, data);
//            }
            
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

}
