/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa uma lista de objetos do tipo <code>JFieldDeclaration</code>.
 * 
 * @see JFieldDeclaration
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 */
public class JFieldDeclarationSet extends DeclarationSet {

    /**
     * Construtor vazio da classe.
     */
    public JFieldDeclarationSet() {
        this.setTypeIdentifier(JaTSNode.FIELD_DECLARATION_SET);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>NodeList</code>
     * contendo<br> apenas objetos do tipo JFieldDeclaration
     */
    protected JFieldDeclarationSet(NodeList objects) throws IllegalArgumentException {
        if (Util.verifyNodeList(objects, JaTSNode.FIELD_DECLARATION)) {
            this.declarations = objects;
        } else {
            throw new IllegalArgumentException("lista de atributos invalida");
        }
        this.setTypeIdentifier(JaTSNode.FIELD_DECLARATION_SET);
    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param field Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento<br>
     *          <code>field</code> for <code>null</code>.
     */
    public void add(INode field) throws IllegalArgumentException {

        if (field == null || !(field instanceof JVariableDeclaration)) {
            throw new IllegalArgumentException("JFieldDeclaration passada  null");
        }
        if (!this.declarations.contains(field)) {
            declarations.addElement(field);
        }
    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para a<br>
     * direita na posies do item que esiver naquela posio e nas daqueles que<br>.
     * estiverem  direita dela.
     *
     * @param field Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @param pos   Um inteiro correspondendo  posio onde o n deve ser <br>
     *              inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>field</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode field, int pos)
            throws IllegalArgumentException {
        if (field == null || pos < 0 || pos > this.size() ||
                !(field instanceof JVariableDeclaration)) {
            throw new IllegalArgumentException();
        }
        super.declarations.insertElementAt(field, pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista, <br>
     * sobreescrevendo o elemento que estava nela.
     *
     * @param field Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @param pos   Um inteiro correspondendo  posio onde o n deve ser <br>
     *              inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>field</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode field, int pos)
            throws IllegalArgumentException {

        if (field == null || pos < 0 || pos > this.size() ||
                !(field instanceof JVariableDeclaration)) {
            throw new IllegalArgumentException();
        }
        super.declarations.setElementAt(field, pos);
    }


    /**
     * Indica se um determinado atributo pertence ao conjunto.
     * (FLORZINHA, QUER FAZER O FAVOR DE NAO MATAR OS MEUS METODOS? :PP)
     *
     * @param fieldName
     * @return
     */
    public boolean contains(String fieldName) {
        return (this.getFieldDeclaration(fieldName) != null);
    }

    /**
     * Retorna a declarao de atributo correspondente ao string passado como
     * parametro
     */
    public JFieldDeclaration getFieldDeclaration(String fieldName) {
        JFieldDeclaration retorno = null;
        JFieldDeclaration tmpField = null;
        JaTSIterator it = this.getIterator();
        boolean foundIt = false;
        JVariableDeclarator varDec = null;
        JVariableDeclaratorSet decs = null;
        JIdentifier tmpId = null;

        if (Util.checkString(fieldName)) {
            while (it.hasNext() && !foundIt) {
                // pega o atributo atual
                tmpField = (JFieldDeclaration) it.next();
                if (!tmpField.isVariable() && !tmpField.isExecutable()) {
                    // pega os declaradores de varivel daquele tipo
                    decs = tmpField.getVariables();
                    if (decs != null) {
                        for (int i = 0; i < decs.size(); i++) {
                            varDec = (JVariableDeclarator) decs.elementAt(i);
                            tmpId = varDec.getIdentifier();
                            // encontrou o mesmo id, ento retorna a
                            // JFieldDeclaration
                            if (tmpId.getValue().equals(fieldName)) {
                                retorno = tmpField;
                            }
                        }
                    }
                }
            }
        } else {
            throw new IllegalArgumentException("Argumento fieldName nulo em " +
                    "JFieldDeclarationSet.getFieldDeclaration");
        }
        return retorno;

    }


    /**
     * Retorna um objeto do tipo <code>JParameterList</code> correspondente a
     * esta lista de declaraes de atributo. Assim, se tivermos as declaraes:
     * "private int x; private int y", a chamada a este mtodo ir retornar:
     * "int x, int y". <br>
     * <b> ATENO: </b>  A lista de parametros retornada no inclui os
     * atributos que definem constantes!
     *
     * @return um objeto do tipo <code>JParameterList</code>  correspondente
     *         a esta lista de declaraes  de atributo
     */
    public JParameterList toParameterList() {
        JParameterList paramList = new JParameterList();
        JaTSIterator jit = this.getIterator();
        JaTSIterator jit2 = null;
        JFormalParameter tmpParam = null;
        JParameterList tmpParamList = new JParameterList();
        JFieldDeclaration tmpField = null;
        while (jit.hasNext()) {
            tmpField = (JFieldDeclaration) jit.next();
            // s coloca na lista de parametros, aquilo que no  constante
            if (!tmpField.isConstant()) {
                tmpParamList = tmpField.toParameterList();
                jit2 = tmpParamList.getIterator();
                while (jit2.hasNext()) {
                    tmpParam = (JFormalParameter) jit2.next();
                    paramList.add(tmpParam);
                }
            }

        }
        return paramList;
    }

    public JFieldDeclarationSet reverse() {

        NodeList revertedFields = this.declarations.reverse();
        JFieldDeclarationSet retorno = new JFieldDeclarationSet(revertedFields);
        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JFieldDeclaration</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JFieldDeclaration</code>.
     *                Por consequncia, objetos do tipo<code>JFieldDeclaration</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }
        if (!(node instanceof JFieldDeclarationSet)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!(this.matchesAsAVariable(node, results,
                JaTSNode.FIELD_DECLARATION_SET))) {
            JFieldDeclarationSet nodeFieldDecs = (JFieldDeclarationSet) node;


            // Casamento padro de declaraes de atributo no lado esquerdo da
            // transformao (this) com declaraes de atributo no programa fonte
            // (contidas em nodeFieldDecs).
            int i = 0;
            int max = super.declarations.size();
            while (i < max) {
                JFieldDeclaration tempFieldDec = (JFieldDeclaration) this.elementAt(i);

                int j = 0;
                int max2 = nodeFieldDecs.size();
                boolean matchedOk = false;
                while (j < max2 && !matchedOk) {
                    JVariableDeclaration nodeToBeMatched =
                            (JVariableDeclaration) nodeFieldDecs.elementAt(j);

                    ResultSet tempRes = new ResultSet();
                    // Tenta casar tempFieldDec com qualquer declarao de atributo
                    // presente em nodeFieldDecs.
                    try {
                        tempFieldDec.match(nodeToBeMatched, tempRes);
                        // Se tudo der certo, nodeToBeMatched  retirado de
                        // nodeFieldDecs e o resultado desse casamento individual
                        //  adicionado ao resultado do casamento como um todo.
                        results.merge(tempRes);
                        nodeFieldDecs.removeElementAt(j);
                        matchedOk = true;
                    } catch (NodesNotMatchedException e) {
                    }
                    j++;
                }

                if (j == max2 && !matchedOk) {
                    throw new NodesNotMatchedException("FieldSets didn't match",
                            tempFieldDec, nodeFieldDecs);
                }
                i++;
            }
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        if (super.declarations != null) {
            int i = super.declarations.size() - 1;

            // empilha as declaraes de atributo  no sentido oposto
            while (i >= 0) {
                super.declarations.elementAt(i).accept(visitor, data);
                i--;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * duas maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 2 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * <p/>
     * Ns do tipo JFieldDeclarationSet nunca so executveis.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;
        // Processamento de cada uma das declaraes de atributo, respectivamente.
        int i = 0;
        int max = this.size();
        while (i < max) {
            INode currFD = this.elementAt(i);
            if (currFD != null) {
                Object ret = this.processNode(currFD, data);
                if (currFD.isExecutable()) {
                    if (!(ret instanceof JVariableDeclaration))
                        throw new ExecutionException(this);
                    this.setElementAt((JVariableDeclaration) ret, i);
                }
            }
            i++;
        }

        return retorno;
    }

} // fim da classe
