/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * 
 * @author Adeline de Sousa
 * @version 12/11/2002
 *
 */
public class JForStatement extends JStatement {


    // (for init) - declarao da varivel do lao ou lista de declaraes
    private INode forInit;

    // condio de parada
    private JExpression expression;

    // (for update) atualizao da varivel do lao
    private JExpressionList forUpdate;

    // statement associado ao for
    private INode statement;

    /**
     * Construtor default da classe
     */
    public JForStatement() {
        this.forInit = null;
        this.expression = null;
        this.forUpdate = new JExpressionList();
        this.statement = null;
        this.setTypeIdentifier(JaTSNode.STATEMENT);
    }


    /**
     * Construtor da classe
     *
     * @param init       Expresso de inicializao do for
     * @param expr       Condio de parada do for
     * @param updateList Atualizao do contador do lao
     */
    public JForStatement(INode init, JExpression expr, JExpressionList updateList) {
        this.setForInit(init);
        this.setExpression(expr);
        this.setForUpdate(updateList);
        this.setTypeIdentifier(JaTSNode.STATEMENT);
    }


    /**
     * Retorna a inicializao do for, que pode ser tanto uma declarao de
     * varivel local quanto uma lista de declaraes
     *
     * @return Inicializao do for, que pode ser tanto uma declarao de
     *         varivel local quanto uma lista de expresses
     */
    public INode getForInit() {
        return forInit;
    }


    /**
     * Retorna a inicializao do for, que pode ser tanto uma declarao de
     * varivel local quanto uma lista de declaraes
     *
     * @param forInit Inicializao do for, que pode ser tanto uma declarao
     *                de varivel local quanto uma lista de expresses
     */
    public void setForInit(INode forInit) {
        this.forInit = forInit;
    }


    /**
     * Retorna a condio de parada do for
     *
     * @return Condio de parada do for
     */
    public JExpression getExpression() {
        return expression;
    }


    /**
     * Seta a condio de parada do for
     *
     * @param expr Condio de parada do for
     */
    public void setExpression(JExpression expr) {
        this.expression = expr;
    }


    /**
     * Retorna a lista de expressoes de atualizao do contador do for
     *
     * @return Lista de expressoes de atualizao do contador do for
     */
    public JExpressionList getForUpdate() {
        return forUpdate;
    }


    /**
     * Seta a lista de expressoes de atualizao do contador do for
     *
     * @param updateList Lista de expressoes de atualizao do contador do for
     */
    public void setForUpdate(JExpressionList updateList) {
        this.forUpdate = updateList;
    }

    /**
     * Retorna o statement associado ao lao for
     *
     * @return Statement associado ao lao for
     */
    public INode getStatement() {
        return statement;
    }


    /**
     * Seta o statement associado ao lao. Aquilo que fica dentro das chaves do
     * "for".
     *
     * @param stmt Statement associado ao lao
     */
    public void setStatement(INode stmt) {
        this.statement = stmt;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (!this.isExecutable() && !this.isVariable()) {

            // envia o visitor pela sub-arvore do statement associado ao for
            if (this.getStatement() != null) {
                this.statement.accept(visitor, data);
            }

            // envia o visitor pela sub-arvore da lista de atualizao do for
            if (this.getForUpdate() != null) {
                this.forUpdate.accept(visitor, data);
            }

            // envia o visitor pela subarvore da expresso
            if (this.getExpression() != null) {
                this.expression.accept(visitor, data);
            }

            // envia o visitor pela sub-arvore da lista de inicializao do for
            if (this.getForInit() != null) {
                this.forInit.accept(visitor, data);
            }

        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JForStatement)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JForStatement jfor = (JForStatement) node;
            if (this.getForInit() != null && jfor.getForInit() != null) {
                this.getForInit().match(jfor.getForInit(), results);
            }
            if (this.getExpression() != null && jfor.getExpression() != null) {
                this.getExpression().match(jfor.getExpression(), results);
            }
            if (this.getForUpdate() != null && jfor.getForUpdate() != null) {
                this.getForUpdate().match(jfor.getForUpdate(), results);
            }
            if (this.getStatement() != null && jfor.getStatement() != null) {
                this.getStatement().match(jfor.getStatement(), results);
            }

        }
    }


    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        Object retorno = this;

        // se o n for executavel, processa o n
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        // se o n nao for executavel, processa os filhos
        else {

            // processa a inicializao do for
            INode tmpInit = this.getExpression();
            if (tmpInit != null) {
                Object ret = this.processNode(tmpInit, data);
                if (tmpInit.isExecutable()) {
                    try {
                        this.setForInit((INode) ret);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // processa a condio de parada
            INode tmpCondition = this.getExpression();
            if (tmpCondition != null) {
                Object ret = this.processNode(tmpCondition, data);
                if (tmpCondition.isExecutable()) {
                    try {
                        this.setExpression((JExpression) ret);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // processa a lista de atualizaes do for
            INode tmpUpdateList = this.getExpression();
            if (tmpUpdateList != null) {
                Object ret = this.processNode(tmpUpdateList, data);
                if (tmpUpdateList.isExecutable()) {
                    try {
                        this.setForUpdate((JExpressionList) ret);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // processa o statement associado
            INode tmpStmt = this.getStatement();
            if (tmpStmt != null) {
                Object ret = this.processNode(tmpStmt, data);
                try {
                    this.setStatement((INode) ret);
                } catch (Exception e) {
                    throw new ExecutionException(e.getMessage(), this);
                }
            }

        }
        return retorno;
    }
}
