/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe guarda as informacoes a respeito de um parametro formal, o qual
 *  representado por um tipo e um identificador e, possivelmente, um modificador.
 * O nico modificador possvel para um parmetro  o modificador <code>final</code>,
 * o qual indica que este parmetro no poder ser alterado no corpo do mtodo.
 * Um parmetro formal  aquele usado em uma declarao de mtodo ou construtor.
 * O parmetro real, ou  argumento  o que  passado na hora da invocao do
 * mtodo ou construtor.<p>
 * <p/>
 * Ex:
 * Na declarao
 * <code>
 * public JFormalParameter(JModifierList mod,JType t, JIdentifier id){
 * <br>
 * ...
 * <br>
 * }</code>
 * Os parametros formais so :
 * <code> JModifierList mod, JType t, JIdentifier id </code>
 * @author Adeline de Sousa
 * @version  0.1 -  08/05/2001
 * @since JDK 1.3
 */
public class JFormalParameter extends JExpression {

    // modificador do parametro (O nico modificador possvel  o "final")
    private JModifierList modifiers;

    // tipo do parametro
    private JType type;

    // identificador do  parametro
    private JIdentifier identifier;

    // numero de pares de colchetes associados ao nome (no caso de um tipo
    // ser array e ter sido usada a notao ps-fixada)
    private int dimensions;

    /**
     * Construtor vazio da classe
     */
    public JFormalParameter() {
        this.modifiers = new JModifierList(JModifierList.PARAMETER_DECLARATION);
        this.type = null;
        this.identifier = null;
        this.setTypeIdentifier(JaTSNode.FORMAL_PARAMETER);
    }


    /**
     * Construtor da classe
     *
     * @param mod Modificador do parametro
     * @param t   tipo do parametro
     * @param id  identificador do parametro
     */
    public JFormalParameter(JModifierList mod, JType t, JIdentifier id) {
        this.modifiers = new JModifierList(JModifierList.PARAMETER_DECLARATION);
        this.setModifiers(mod);
        this.setType(t);
        this.setIdentifier(id);
        this.setTypeIdentifier(JaTSNode.FORMAL_PARAMETER);
    }

    /**
     * Retorna o modificador do parametro
     *
     * @return Modificador do parametro
     */
    public JModifierList getModifiers() {

        return this.modifiers;
    }


    /**
     * Retorna o tipo do parametro
     *
     * @return Tipo do parametro
     */
    public JType getType() {

        return this.type;
    }


    /**
     * Retorna o identificador do parametro
     *
     * @return Identificador do parametro
     */
    public JIdentifier getIdentifier() {

        return this.identifier;
    }


    /**
     * Retorna o nmero de dimenses associadas ao tipo, caso tenha sido usada
     * a notao ps-fixada
     *
     * @return nmero de dimenses associadas ao tipo, caso tenha sido usada
     *         a notao ps-fixada
     */
    public int getDimensions() {
        return dimensions;
    }

    /**
     * Adiciona o modificador a este prametro
     *
     * @param mod Modificador do parametro
     * @throws IllegalArgumentException - Se o parametro passado for nulo ou
     *                                  no for vlido para um parametro.
     */
    public void addModifier(String mod) throws IllegalArgumentException {
        this.modifiers.addModifier(mod);
    }

    /**
     * Seta o modificador do parametro com o valor especificado
     *
     * @param mod Modificador do parametro (constante inteira)
     * @throws IllegalArgumentException - Se o parametro passado for nulo ou
     *                                  no for vlido para um parametro.
     */
    public void setModifiers(int mod) throws IllegalArgumentException {

        if (mod > 0) {
            this.modifiers.addModifier(mod);
        } else {
            throw new IllegalArgumentException("Modificador nulo");
        }

    }


    /**
     * Seta o modificador do parametro com o valor especificado
     *
     * @param mod Modificador do parametro
     * @throws IllegalArgumentException - Se o parametro passado for nulo ou
     *                                  no for vlido para um parametro.
     */
    public void setModifiers(JModifierList mod) throws IllegalArgumentException {

        if (mod != null) {
            if (mod.getPossibleModifiers() == JModifierList.PARAMETER_DECLARATION) {
                this.modifiers = mod;
            } else {
                throw new IllegalArgumentException("Modificador invalido");
            }

        } else {
            throw new IllegalArgumentException("Modificador nulo");
        }

    }

    /**
     * Seta o tipo do parametro com o valor especificado
     *
     * @param t Tipo do parametro
     * @throws IllegalArgumentException - Se o tipo passado for nulo.
     */
    public void setType(JType t) throws IllegalArgumentException {

        if (t != null) {
            this.type = t;

        } else {
            throw new IllegalArgumentException("Tipo nulo");
        }

    }


    /**
     * Seta o identificador do parametro com o valor especificado
     *
     * @param id Identificador do parametro
     * @throws IllegalArgumentException - Se o parametro passado for nulo.
     */
    public void setIdentifier(JIdentifier id) throws IllegalArgumentException {

        if (id != null) {
            this.identifier = id;
        } else {
            throw new IllegalArgumentException("Identificador nulo");
        }
    }

    /**
     * Seta o nmero de dimenses associadas ao tipo, no caso em que foi usada a
     * notao ps-fixada. Uma dimenso zero, indica um tipo que no  array.
     * S podem ser passados valores positivos como parametro. Caso contrrio,
     * a dimenso ser setada para zero.
     *
     * @param dim nmero de dimenses associadas ao tipo
     */
    public void setDimensions(int dim) {
        if (dim >= 0) {
            this.dimensions = dim;
        } else {
            this.dimensions = 0;
        }
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JFormalParameter</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JFormalParameter</code>.
     *                Por consequncia, objetos do tipo<code>JFormalParameter</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("JFormalParameter ou ResultSet nulo");
        }

        if (!(node instanceof JFormalParameter)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, JaTSNode.FORMAL_PARAMETER)) {

            JFormalParameter fp = (JFormalParameter) node;

            // Casa o modificador, se houver
            if (this.modifiers != null) {
                this.modifiers.match(fp.modifiers, results);
            }

            // Casa o tipo
            this.type.match(fp.type, results);


            // Casa o identificador
            if (this.identifier != null) {
                this.identifier.match(fp.identifier, results);    
            }

        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor pela sub-arvore do identificador
            if (this.identifier != null) {
                this.identifier.accept(visitor, data);
            }

            // Envie o visitor pela sub-arvore do tipo
            if (this.type != null) {
                this.type.accept(visitor, data);
            }

            // Envie o visitor pela sub-arvore do modificador
            if (this.modifiers != null) {
                this.modifiers.accept(visitor, data);
            }
        }
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {
            // Processamento da lista de modificadores do parametro
            INode mods = this.getModifiers();
            if (mods != null) {
                Object ret1 = this.processNode(mods, data);
                if (mods.isExecutable()) {
                    if (!(ret1 instanceof JModifierList)) {
                        throw new ExecutionException(this);
                    }
                    this.setModifiers((JModifierList) ret1);
                }
            }
            // Processamento do tipo do parametro
            INode pType = this.getType();
            if (pType != null) {
                Object ret2 = this.processNode(pType, data);
                if (pType.isExecutable()) {
                    try {
                        this.setType(JType.unwrapType(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento do identificador do parametro
            INode pid = this.getIdentifier();
            if (pid != null) {
                Object ret3 = this.processNode(pid, data);
                if (pid.isExecutable()) {

                    try {
                        this.setIdentifier(JIdentifier.unwrapIdentifier(ret3));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
        }
        return retorno;
    }

    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JFormalParameter) {
            JFormalParameter fp = (JFormalParameter) obj;
            if (!this.isVariable() && !this.isExecutable()) {
                retorno =
                        AbstractNode.equals(this.modifiers, fp.modifiers) &&
                        AbstractNode.equals(this.type, fp.type) &&
                        AbstractNode.equals(this.identifier, fp.identifier);
            } else {
                retorno = compareJaTSNode(fp);
            }
        }
        return retorno;
    }


} // fim da classe
