/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Comparator;


/**
 * Esta classe implementa um identificador. Exemplos de possveis
 * identificadores so nomes de classes e nomes de mtodos.
 *
 * @author Fernando Castor
 * @author Adeline de Sousa
 *
 * @version  0.1 -  20/01/2004
 * @since JDK 1.3
 */
public class JIdentifier extends JExpression implements Signature, Comparator {


    // O String que armazena o valor deste identificador.
    private String value;
    
    /**
     * Construtor vazio da classe.
     */
    public JIdentifier() {
        this.value = null;
        this.setTypeIdentifier(JaTSNode.IDENTIFIER);
    }

    /**
     * Construtor da classe que recebe como parmetro o valor que este
     * identificador assumir.
     *
     * @param idValue Um String correspondendo ao valor deste identificador.
     */
    public JIdentifier(String idValue) {
        this.value = idValue;
        this.setVariableName(idValue);
        this.setTypeIdentifier(JaTSNode.IDENTIFIER);
    }


    /**
     * Verifica se esse identificador, corresponde a um JIDENTIFIER. Ou seja,
     * se  formado por um #<IDENTIFIER>. <br>
     * <p/>
     * Ex: #C, #conta - so do tipo JIDENTIFIER
     *
     * @return true, caso se trate de um JIDENTIFIER <br>
     *         false, caso contrrio
     */
    public boolean isJIdentifier() {
        boolean retorno = false;
        if (!this.isExecutable()) {
            retorno = this.isVariable();
        }
        return retorno;

    }

    /**************************************************************************/
    /**                             Mtodos de API                           **/
    /**************************************************************************/

    /**
     * Retorna o Identificador passado como parmetro com a primeira letra
     * maiscula.
     */
    public JIdentifier toClassName() {
        JIdentifier retorno = null;
        String idValue = getValue();
        String newId = null;
        char firstChar = Character.toUpperCase(idValue.charAt(0));
        if (idValue.length() > 1) {
            newId = firstChar + idValue.substring(1, idValue.length());
        } else if (idValue.length() == 1) {
            newId = new String(firstChar + "");
        }
        retorno = new JIdentifier(newId);

        return retorno;


    }


    /**
     * Adiciona um prefixo a um identificador mas deixa o identificador
     * resultante de acordo com o padro de codificao. Por exemplo, para
     * um identificador correspondendo ao string "atributo", a invocao do
     * mtodo <code>addPreffix</code> passando como argumento o string "get"
     * produz como resultado o string "getAtributo".
     *
     * @param prefix o prefix que ser adicionado ao identificador.
     * @return o identificador que teve o prefix adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefix</code> for <code>null</code>.
     */
    public JIdentifier addPrefix(String prefix) throws IllegalArgumentException {
        if (prefix == null) {
            throw new IllegalArgumentException();
        }
        // por definio, o retorno ser um identificador com o mesmo valor
        // que this
        JIdentifier retorno = new JIdentifier(this.getValue());
        if (this.isASourceNode()) {
            String idText = this.getValue();
            char primeiroChar = Character.toUpperCase(idText.charAt(0));
            retorno.setValue(prefix + primeiroChar
                    + idText.substring(1, idText.length()));
        }
        return retorno;
    }


    /**
     * Adiciona um prefixo a um identificador mas deixa o identificador
     * resultante de acordo com o padro de codificao. Por exemplo, para
     * um identificador correspondendo ao string "atributo", a invocao do
     * mtodo <code>addPreffix</code> passando como argumento o string "get"
     * produz como resultado o string "getAtributo".
     *
     * @param prefix o prefixo que ser adicionado ao identificador.
     * @return o identificador que teve o prefixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addPrefix(JIdentifier prefix) throws IllegalArgumentException {
        if (prefix == null || !(prefix.isASourceNode())) {
            throw new IllegalArgumentException();
        }
        return this.addPrefix(prefix.getValue());
    }

    /**
     * Adiciona um prefixo a um identificador mas deixa o identificador
     * resultante de acordo com o padro de codificao. Por exemplo, para
     * um identificador correspondendo ao string "atributo", a invocao do
     * mtodo <code>addPreffix</code> passando como argumento o string "get"
     * produz como resultado o string "getAtributo".
     *
     * @param prefix o prefixo que ser adicionado ao identificador.
     * @return o identificador que teve o prefixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addPrefix(JName prefix) throws IllegalArgumentException {
        if (prefix == null || !(prefix.isASourceNode())) {
            throw new IllegalArgumentException();
        }
        return this.addPrefix(prefix.toString());
    }

    /**
     * Adiciona um sufixo a um identificador.  Por exemplo, dado o identificador
     * "Conta", a chamada addSuffix("bean", true), retornaria "ContaBean".
     * E a chamada addSuffix("bean",false), Contabean
     *
     * @param sufixo    O sufixo que ser adicionado ao identificador.
     * @param uppercase Diz se o sufixo deve ter sua primeira letra
     *                  modificada para maiscula
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(String sufixo, boolean uppercase)
            throws IllegalArgumentException {

        if (sufixo == null || (sufixo.length() <= 0)) {
            throw new IllegalArgumentException();
        }
        // por definio, o retorno ser um identificador, com o mesmo valor de
        // this
        JIdentifier retorno = new JIdentifier(this.getValue());
        if (!this.isVariable()) {
            String textoId = this.getValue();
            char primeiroChar = sufixo.charAt(0);
            if (uppercase) {
                primeiroChar = Character.toUpperCase(sufixo.charAt(0));
            }
            retorno.setValue(textoId + primeiroChar +
                    sufixo.substring(1, sufixo.length()));

        }
        return retorno;
    }

    /**
     * Adiciona um sufixo a um identificador.
     *
     * @param sufixo O sufixo que ser adicionado ao identificador. Este sufixo
     *               ter sua primeira letra modificada para maiscula.
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(String sufixo) throws IllegalArgumentException {
        return this.addSuffix(sufixo, true);
    }

    /**
     * Adiciona um sufixo a um identificador.
     *
     * @param sufixo O sufixo que ser adicionado ao identificador. Este sufixo
     *               ter sua primeira letra modificada para maiscula.
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(JIdentifier sufixo) throws IllegalArgumentException {
        JIdentifier retorno = new JIdentifier(this.getValue());
        if (sufixo != null) {
            String valorSufixo = sufixo.getValue();
            retorno = this.addSuffix(valorSufixo, true);
        }
        return retorno;
    }

    /**
     * Adiciona um sufixo a um identificador.
     *
     * @param sufixo O sufixo que ser adicionado ao identificador. Este sufixo
     *               ter sua primeira letra modificada para maiscula.
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(JName sufixo) throws IllegalArgumentException {
        JIdentifier retorno = null;
        if (sufixo != null) {
            String valorSufixo = sufixo.toString();
            retorno = this.addSuffix(valorSufixo, true);
        }
        return retorno;
    }

    /**
     * Adiciona um sufixo a um identificador.
     *
     * @param sufixo    O sufixo que ser adicionado ao identificador. Este sufixo
     *                  ter sua primeira letra modificada para maiscula.
     * @param uppercase Diz se o sufixo deve ter sua primeira letra
     *                  modificada para maiscula
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(JName sufixo, boolean uppercase) throws IllegalArgumentException {
        JIdentifier retorno = null;
        if (sufixo != null) {
            String valorSufixo = sufixo.toString();
            retorno = this.addSuffix(valorSufixo, uppercase);
        }
        return retorno;
    }

    /**
     * Adiciona um sufixo a um identificador.
     *
     * @param sufixo    O sufixo que ser adicionado ao identificador. Este sufixo
     *                  ter sua primeira letra modificada para maiscula.
     * @param uppercase Diz se o sufixo deve ter sua primeira letra
     *                  modificada para maiscula
     * @return o identificador que teve o sufixo adicionado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada se
     *          <code>prefixo</code> for <code>null</code>.
     */
    public JIdentifier addSuffix(JIdentifier sufixo, boolean uppercase) throws IllegalArgumentException {
        JIdentifier retorno = null;
        if (sufixo != null) {
            String valorSufixo = sufixo.getValue();
            retorno = this.addSuffix(valorSufixo, uppercase);
        }
        return retorno;
    }

    /**
     * Este mtodo converte o nome do Identificador passado como parmetro
     * em um literal String. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>"Conta"</code>.
     *
     * @param id Identificador que ser comvertido para literal String
     * @return Literal String representando o identificador passado como
     *         parmetro
     */
    public static JLiteral toLiteral(JIdentifier id) {
        JLiteral retorno = null;
        if (id != null) {
            retorno = new JLiteral(JLiteral.STRING_LITERAL, id.getValue());
        } else {
            throw new IllegalArgumentException("Parametro id invlido em " +
                    "JIdentifier.toLiteral");
        }
        return retorno;
    }

    /**
     * Retorna o literal correspondente a este identificador.
     *
     * @return Literal correspondente a este identificador
     */
    public JLiteral toLiteral() {
        return new JLiteral(JLiteral.STRING_LITERAL, this.value);
    }

    /**
     * Este mtodo converte o nome do Identificador passado como parmetro
     * em um literal String. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>"Conta"</code>.
     *
     * @param id Identificador que ser comvertido para literal String
     * @return Literal String representando o identificador passado como
     *         parmetro
     */
    public static String toLiteralValue(JIdentifier id) {
        String retorno = null;
        if (id != null) {
            retorno = id.getValue();
        } else {
            throw new IllegalArgumentException("Parametro id invlido em " +
                    "JIdentifier.toLiteralValue");
        }
        return retorno;
    }


    /**
     * Este mtodo converte o nome do Name passado como parmetro em um literal
     * String.  Nesse caso, o Name deve ser Identifier-Compatible, ou seja,
     * deve possuir apenas um identificador. Exemplo: <br>
     * <p/>
     * Seja o name  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>"Conta"</code>.
     *
     * @param name Nome que ser comvertido para literal String
     * @return Literal String representando o identificador passado como
     *         parmetro
     */
    public static String toLiteralValue(JName name) {
        String retorno = null;
        if (name != null) {
            retorno = JIdentifier.unwrapIdentifier(name).getValue();
        } else {
            throw new IllegalArgumentException("Parametro name invlido em " +
                    "JIdentifier.toLiteralValue");
        }
        return retorno;
    }

    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * varivel. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>conta</code>.
     *
     * @param id Identificador que ser convertido em nome de varivel
     * @return Identificador representando o nome da varivel
     */
    public static JIdentifier toVariableName(JIdentifier id) {
        JIdentifier retorno = null;
        if (id != null) {
            String idValue = id.getValue();
            if (idValue == null)
                idValue = id.getVariableName();
            String newId = null;
            char firstChar = Character.toLowerCase(idValue.charAt(0));
            if (idValue.length() > 1) {
                newId = firstChar + idValue.substring(1, idValue.length());
            } else if (idValue.length() == 1) {
                newId = new String(firstChar + "");
            }
            retorno = new JIdentifier(newId);

        } else {
            throw new IllegalArgumentException("Parametro id invalido " +
                    "JIdentifier.toVariableName");
        }
        return retorno;
    }

    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * varivel. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>conta</code>.
     *
     * @return Identificador representando o nome da varivel
     */
    public JIdentifier toVariableName() {
        return JIdentifier.toVariableName(this);
    }


    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * constante. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>CONTA</code>.
     * <p/>
     * Obs.: Para identificadores com letras maisculas consecutivas, o mtodo
     * no retorna o valor esperado. Por ex.: RepositorioBDR -> REPOSITORIO_B_D_R
     *
     * @param id Identificador que ser convertido em nome de varivel
     * @return Identificador representando o nome da varivel
     */
    public static JIdentifier toConstantName(JIdentifier id) {
        JIdentifier retorno = null;
        if (id != null) {
            String idValue = id.getValue();
            if (idValue == null)
                idValue = id.getVariableName();
            StringBuffer newId = new StringBuffer(idValue.length());

            //passar a primeira para maiuscula mas nao inserir underscore no
            //inicio da constante
            char c = Character.toUpperCase(idValue.charAt(0));
            newId.append(c);
            for (int i = 1; i < idValue.length(); ++i) {
                c = idValue.charAt(i);
                if (Character.isUpperCase(c))
                    newId.append('_');

                c = Character.toUpperCase(c);
                newId.append(c);
            }

            retorno = new JIdentifier(newId.toString());

        } else {
            throw new IllegalArgumentException("Parametro id invalido " +
                    "JIdentifier.toVariableName");
        }
        return retorno;
    }

    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * constante. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>CONTA</code>.
     * <p/>
     * Obs.: Para identificadores com letras maisculas consecutivas, o mtodo
     * no retorna o valor esperado. Por ex.: RepositorioBDR -> REPOSITORIO_B_D_R
     *
     * @return Identificador representando o nome da varivel
     */
    public JIdentifier toConstantName() {
        return JIdentifier.toConstantName(this);
    }

    /**
     * Mtodo que diz se dois objetos do tipo <code>JIdentifier</code> so
     * iguais. Para que isso seja verdade, os dois tm que ter valores iguais.
     *
     * @param objeto Um objeto qualquer que ser comparado com <code>this</code>.
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica que
     *         o valor do objeto com o qual <code>this</code> est sendo
     *         comparado  igual ao de <code>this</code>.
     */
    public boolean equals(Object objeto) {
        boolean retorno = false;
        if (objeto instanceof JIdentifier) {
            JIdentifier id = (JIdentifier) objeto;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(id.getValue(), this.getValue());
            } else {
                retorno = compareJaTSNode(id);
            }
        }

        return retorno;
    }


    /**
     * Este mtodo retorna o String correspondente ao valor deste identificador.
     *
     * @return Um String correspondente ao valor deste identificador
     * @deprecated O mtodo <b>getValue()</b> deve ser usado no lugar deste.
     */
    public String getName() {
        return this.value;
    }

    /**
     * Este mtodo retorna o String correspondente ao valor deste identificador.
     *
     * @return Um String correspondente ao valor deste identificador
     */
    public String getValue() {
        return this.value;
    }


    /**
     * Este mtodo seta o valor deste identificador.
     *
     * @param val Um String correspondente ao valor deste identificador.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>val</code> for <code>null</code>.
     * @see #setValue(String val)
     * @deprecated Este mtodo nao deve ser mais utilizado, ao invs disso,
     *             deve-se usar o mtodo setValue(String val)
     */
    public void setName(String val) throws IllegalArgumentException {
        this.setValue(val);

    }

    /**
     * Este mtodo seta o valor deste identificador.
     *
     * @param val Uma String correspondente ao valor deste identificador. Se a
     *            string passada iniciar-se por um '#', ou seja, um caracter que
     *            indica nome de varivel JaTS, ento, ser setado como varivel
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>val</code> for <code>null</code>.
     */
    public void setValue(String val) throws IllegalArgumentException {
        if (!Util.checkString(val)) {
            throw new IllegalArgumentException();
        }

        // Se o valor de um identificar  um String que identifica uma varivel
        // (precedido por AbstractNode.VARIABLECHAR), ento o n correspondente a
        // este identificador tambm  uma varivel.
        if (Util.isAValidVariableName(val)) {
            this.setVariableName(val);
        }
        this.value = val;
    }


    /**
     * Este mtodo seta o valor deste identificador.
     *
     * @param val Um JName correspondente ao valor deste identificador.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>val</code> for <code>null</code>.
     */
    public void setName(JName val) throws IllegalArgumentException {
        if (val == null) {
            throw new IllegalArgumentException();
        }
        String id = JIdentifier.unwrapIdentifier(val).getValue();
        this.setValue(id);


    }


    /**
     * Este mtodo seta o valor deste identificador.
     *
     * @param val Um JType, cujo nome representa o valor deste identificador.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>val</code> for <code>null</code>.
     */
    public void setName(JType val) throws IllegalArgumentException {
        if (val == null) {
            throw new IllegalArgumentException();
        }
        String id = JIdentifier.unwrapIdentifier(val).getValue();
        this.setValue(id);

    }

    /**
     * Retorna um JLiteral String baseado no identificador. Por exemplo, se
     * quisermos que o identificador Conta se torne a String "Conta" .
     *
     * @return O identificador passado, envolvido por aspas.
     */
    public JLiteral getLiteralValue() {
        String idValue = this.value;
        String litValue = "\"" + idValue + "\"";
        return new JLiteral(JLiteral.STRING_LITERAL, litValue);

    }


    /**
     * Retorna a String que representa este identificador.
     *
     * @return a <code>String</code> que representa este identificador
     */
    public String toString() {
        String retorno = "";
        if (!this.isVariable()) {
            retorno = this.value;
        } else if (this.isExecutable()) {
            try {
                retorno = this.getExecutableDeclaration().toString();
            } catch (InconsistentNodeException e) {
            }
        } else {
            retorno = this.printVariable();
        }

        return retorno;
    }


    /**
     * Realiza o unwrapping de um identificador. Este mtodo retira, a partir de
     * um objeto de outro tipo como JType, JName, JVariableDeclarator, JLiteral, etc,
     * o identificador correspondente, caso exista um. Ex: Se um objeto do tipo
     * JName (um nome, ou seja, uma lista de identificadores) for composto por um
     * nico identificador, este mtodo retorna esse identificador. Caso
     * contrrio, uma exceo  levantada, uma vez que a transformao tenta
     * colocar um nome (lista de identificadores) no lugar onde apenas um
     * identificador  esperado.
     *
     * @param obj O objeto a partir do qual o objeto do tipo <code>JIdentifier</code>
     *            dever ser gerado.
     * @return Um identificador correspondente ao n fornecido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          se o argumento <code>obj</code> for <code>null</code> ou se
     *          ele for de um tipo a partir do qual no  possvel realizar a
     *          converso para um <code>JIdentifier</code>.
     */
    public static JIdentifier unwrapIdentifier(Object obj) throws
            IllegalArgumentException {
        JIdentifier retorno = null;

        if (obj instanceof INode) {
            AbstractNode node = (AbstractNode) obj;
            boolean nodeIsAJavaNode = node.isASourceNode();
            boolean nodeIsVariable = node.isVariable();
            int nodeType = node.getTypeIdentifier();

            switch (nodeType) {

                case JaTSNode.IDENTIFIER:
                    retorno = (JIdentifier) obj;
                    break;

                case JaTSNode.NAME:
                    JName name = (JName) obj;
                    if (nodeIsAJavaNode && name.size() == 1) {
                        retorno = name.getIdentifierAt(0);
                    } else if (nodeIsVariable) {
                        retorno = new JIdentifier();
                        retorno.setVariableName(name.getVariableName());
                        retorno.setVariableType(JaTSNode.IDENTIFIER);
                    }
                    break;

                case JaTSNode.TYPE:
                    JType type = (JType) obj;
                    if (nodeIsAJavaNode &&
                            type.getNumberOfDimensions() == 0) {
                        JName name2 = type.getTypeName();
                        retorno = name2.getIdentifierAt(0);
                    } else if (nodeIsVariable) {
                        retorno = new JIdentifier();
                        retorno.setVariableName(type.getVariableName());
                        retorno.setVariableType(JaTSNode.IDENTIFIER);
                    }
                    break;
                case JaTSNode.VARIABLE_DECLARATOR:
                    JVariableDeclarator varDec = (JVariableDeclarator) obj;
                    if (varDec.getInitializer() == null &&
                            varDec.getDimensions() == 0) {
                        retorno = varDec.getIdentifier();
                    }
                    break;
                default:
                    throw new IllegalArgumentException("Tipo no compatvel");
            }// encerra o switch
        } else if (obj instanceof String) {
            retorno = new JIdentifier((String) obj);
        } else {
            throw new IllegalArgumentException("Tipo no compatvel");
        }

        return retorno;
    }

    /**
     * Compares its two arguments for order.  Returns a negative integer,
     * zero, or a positive integer as the first argument is less than, equal
     * to, or greater than the second.<p>
     * <p/>
     * The implementor must ensure that <tt>sgn(compare(x, y)) ==
     * -sgn(compare(y, x))</tt> for all <tt>x</tt> and <tt>y</tt>.  (This
     * implies that <tt>compare(x, y)</tt> must throw an exception if and only
     * if <tt>compare(y, x)</tt> throws an exception.)<p>
     * <p/>
     * The implementor must also ensure that the relation is transitive:
     * <tt>((compare(x, y)&gt;0) &amp;&amp; (compare(y, z)&gt;0))</tt> implies
     * <tt>compare(x, z)&gt;0</tt>.<p>
     * <p/>
     * Finally, the implementer must ensure that <tt>compare(x, y)==0</tt>
     * implies that <tt>sgn(compare(x, z))==sgn(compare(y, z))</tt> for all
     * <tt>z</tt>.<p>
     * <p/>
     * It is generally the case, but <i>not</i> strictly required that
     * <tt>(compare(x, y)==0) == (x.equals(y))</tt>.  Generally speaking,
     * any comparator that violates this condition should clearly indicate
     * this fact.  The recommended language is "Note: this comparator
     * imposes orderings that are inconsistent with equals."
     *
     * @param o1 the first object to be compared.
     * @param o2 the second object to be compared.
     * @return a negative integer, zero, or a positive integer as the
     *         first argument is less than, equal to, or greater than the
     *         second.
     * @throws ClassCastException if the arguments' types prevent them from
     *                            being compared by this Comparator.
     */
    public int compare(Object o1, Object o2) {

        // TERMINAR!!!!
        int retorno = 0;
        // se os dois so identificadores, tenta comparar seus valores
        if ((o1 instanceof JIdentifier) && (o2 instanceof JIdentifier)) {

            JIdentifier id1 = (JIdentifier) o1;
            JIdentifier id2 = (JIdentifier) o2;
            boolean nodesAreVariables = id1.isVariable() && id2.isVariable();
            boolean nodesAreExecutables = id1.isExecutable() && id2.isExecutable();
            if (nodesAreVariables) {
                // os nomes de variaveis devem estar presentes, neste caso
                retorno = id1.getVariableName().compareTo(id2.getVariableName());
            } else if (!nodesAreExecutables) {
                // os valores dos ids devem estar presentes
                retorno = id1.getValue().compareTo(id2.getValue());
            }
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!isIdentifierCompatible(node)) {
            throw new NodesNotMatchedException("Nodes are incompatible (" +
                    node.getClass().toString() + ", " + this.getClass().toString() + ")",
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JIdentifier id = JIdentifier.unwrapIdentifier(node);
            if (this.getValue() == null) {
                throw new InconsistentNodeException(this);
            }
            if (!this.getValue().equals(id.getValue())) {
                throw new NodesNotMatchedException("Identifiers couldn't be " +
                        "matched (" + this.getValue() + ", " + id.getValue() + ")",
                        this, node);
            }
        }
    }

    /**
     * Verifica se o n passado como parmetro  compatvel com um
     * identificador, ou seja, se ele poderia passar por um identificador. Ns
     * compatveis, so:
     * <ul>
     * <p> <li> <b>Name</b> - se o name s contiver um identificador </li>
     * <p> <li> <b>Type</b> - se o name correspondente s contiver um
     * identificador  e se o mesmo for adimensional </li>
     * </ul>
     *
     * @param node N que se deseja verificar a compatibilidade
     * @return true, se o n passado for compativel
     *         false, caso contrrio
     */
    public static boolean isIdentifierCompatible(INode node) {
        boolean retorno = false;
        retorno = ((node instanceof JIdentifier) ||
                ((node instanceof JName) && ((JName) node).size() == 1)) ||
                ((node instanceof JType) && (((JType) node)).isDimensionless());

        return retorno;
    }

    /**
     *  responsvel por processar o n atual. O processamento de identificadores
     * ocorre se apenas se o n atual for executvel. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu mtodo
     * process e corresponde ao resultado de seu processamento.
     * <p/>
     * S h essa modalidade de processamento para identificadores porque eles
     * nunca tm ns filhos. O mesmo  verdade para ns dos tipos
     * <code>JLiteral</code> e <code>JModifierList</code>.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        return retorno;
    }


    public Object evaluate(Object data) throws ExecutionException {
        Object result = this;
        if (this.isExecutable()) {
            try {
                result = this.processExecutableNode(data);
            } catch (InconsistentNodeException e) {
                throw new ExecutionException(e.getMessage());
            }
        }

        return result;
    }


}
