/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa um JStatement If. O qual  composto pela expresso do
 * if, o statement a ser executado, caso a expresso seja avaliada para true e
 * o statement executado, caso seja avaliada para false
 * <p/>
 * Ex:
 * <p/>
 * if (expr)
 * // statement se expr for true
 * else
 * // statement se expr for false
 * 
 * @author Adeline de Sousa
 * @version 12/11/2002
 */
public class JIfStatement extends JStatement {

    // expresso do if
    private JExpression expression;

    // statement associado ao if
    private INode ifStatement;

    // statement associado ao else
    private INode elseStatement;

    private String ifComment;

    private String elseComment;


    /**
     * Construtor default da classe
     */
    public JIfStatement() {
        this.setTypeIdentifier(JaTSNode.STATEMENT);
    }


    /**
     * Construtor  da classe
     *
     * @param expression    Expresso de avaliao do  if
     * @param ifStatement   JStatement a ser executado, caso seja true
     * @param elseStatement JStatement a ser executado, caso seja false
     */
    public JIfStatement(JExpression expression, JStatement ifStatement,
                        JStatement elseStatement) {
        this();
        this.setExpression(expression);
        this.setIfStatement(ifStatement);
        this.setElseStatement(elseStatement);
        this.setTypeIdentifier(JaTSNode.STATEMENT);
    }


    /**
     * Retorna a expresso de avaliao do If
     *
     * @return Expresso de avaliao do If
     */
    public JExpression getExpression() {
        return expression;
    }


    /**
     * Seta a expresso de avaliao do If
     *
     * @param expr Expresso de avaliao do If
     */
    public void setExpression(JExpression expr) {
        this.expression = expr;
    }


    /**
     * Retorna o statement a ser executado, caso a expresso seja avaliada pra
     * true
     *
     * @return JStatement a ser executado, caso a expresso seja avaliada pra
     *         true
     */
    public INode getIfStatement() {
        return ifStatement;
    }


    /**
     * Seta o  o statement a ser executado, caso a expresso seja avaliada pra
     * true
     *
     * @param ifStatement JStatement a ser executado, caso a expresso seja
     *                    avaliada pra true
     */
    public void setIfStatement(INode ifStatement) {
        this.ifStatement = ifStatement;
    }


    /**
     * Retorna o statement a ser executado, caso a expresso seja avaliada pra
     * false
     *
     * @return JStatement a ser executado, caso a expresso seja avaliada pra
     *         false
     */
    public INode getElseStatement() {
        return elseStatement;
    }

    /**
     * String representando o Comentrio do If
     *
     * @return o Comentrio do If
     */
    public String getIfComment() {
        return ifComment;
    }

    /**
     * Seta a string representando o Comentrio do If
     *
     * @param ifComment
     */
    public void setIfComment(String ifComment) {
        this.ifComment = ifComment;
    }

    /**
     * Retorna a String representando o comentrio do else
     *
     * @return String representando o comentrio do else
     */
    public String getElseComment() {
        return elseComment;
    }

    /**
     * Seta a string representando o comentrio do else
     *
     * @param elseComment
     */
    public void setElseComment(String elseComment) {
        this.elseComment = elseComment;
    }

    /**
     * Seta o  o statement a ser executado, caso a expresso seja avaliada pra
     * false
     *
     * @param elseStatement JStatement a ser executado, caso a expresso seja
     *                      avaliada pra false
     */
    public void setElseStatement(INode elseStatement) {
        this.elseStatement = elseStatement;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (!this.isExecutable() && !this.isVariable()) {

            // envia o visitor pela subarvore do elseStatement
            if (this.elseStatement != null) {
                this.elseStatement.accept(visitor, data);
            }

            // envia o visitor pela subarvore do ifStatement
            if (this.ifStatement != null) {
                this.ifStatement.accept(visitor, data);
            }

            // envia o visitor pela subarvore da expresso do If
            if (this.expression != null) {
                this.expression.accept(visitor, data);
            }
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }
        if (!(node instanceof JIfStatement)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JIfStatement jif = (JIfStatement) node;

            if (this.expression != null && jif.getExpression() != null) {
                this.expression.match(jif.expression, results);
            } else {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }
            if (this.ifStatement != null && jif.getIfStatement() != null) {
                this.ifStatement.match(jif.getIfStatement(), results);
            }
            if (this.elseStatement != null && jif.getElseStatement() != null) {
                this.elseStatement.match(jif.getElseStatement(), results);
            }

        }
    }

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        } else {
            // processa a expresso do if
            INode expr = this.getExpression();
            if (expr != null) {
                Object ret1 = this.processNode(expr, data);
                if (expr.isExecutable()) {
                    try {
                        this.setExpression((JExpression) ret1);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
            // processa o statement associado ao if
            INode ifStmt = this.getIfStatement();
            if (ifStmt != null) {
                Object ret2 = this.processNode(ifStmt, data);
                if (ifStmt.isExecutable()) {
                    try {
                        this.setIfStatement((INode) ret2);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // processa o statement associado ao else
            INode elseStmt = this.getElseStatement();
            if (elseStmt != null) {
                Object ret3 = this.processNode(elseStmt, data);
                if (elseStmt.isExecutable()) {
                    try {
                        this.setElseStatement((INode) ret3);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
        }

        return retorno;
    }
}
