/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa uma clusula <code>import</code>. E  representada
 * basicamente por um <code>JName</code>, que representa o nome do pacote e um
 * <code>JIdentifier</code>correspondente ao nome da classe importada ou um
 * <code>boolean</code> indicando se  todo  o pacote  importado ou no
 * 
 * @author Adeline de Sousa
 * @version   20/01/2003
 */

public class JImportDeclaration extends AbstractNode implements Declaration {


    // diz se a declarao de importao  para todas as classes do pacote.
    public static final boolean IMPORT_ALL_CLASSES = true;

    // nome do pacote ou da classe sendo importado
    private JName packageName;

    // nome da classe sendo importada.
    private JIdentifier className;

    // indica se todas as classes do  so importadas (ou seja, se h '*')
    private boolean importAllClasses;


    /**
     * Construtor Vazio
     */
    public JImportDeclaration() {

        this.packageName = null;
        this.className = null;
        this.importAllClasses = false;
        this.setTypeIdentifier(JaTSNode.IMPORT_DECLARATION);
    }



//    public AttributeSectionList getAttributes() {
//        throw new UnsupportedOperationException();
//    }
//
//    public void setAttributes(AttributeSectionList att) {
//       throw new UnsupportedOperationException();
//    }

    /**
     * Retorna o nome do pacote importado
     *
     * @return Nome do pacote importado
     */
    public JName getPackageName() {
        return packageName;
    }

    /**
     * Seta o nome do pacote importado
     *
     * @param packageName Nome do pacote sendo importado
     */
    public void setPackageName(JName packageName) {
        JName tmpName = packageName;
        if (tmpName != null) {
            // se termina com um ".*", este deve ser removido do nome do pacote
            // e o atributo correspondente ao import do pacote deve ser setado
            try {

                if (tmpName.toString().endsWith(".*")) {
                    tmpName.removeIdentifierAt(tmpName.size() - 1);
                    this.setImportAllClasses(true);
                }
            } catch (Exception ex) {
            } finally {
                this.packageName = tmpName;
            }
        } else {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Retorna o nome da classe importada, caso seja um import especfico
     *
     * @return Nome da classe importada, caso o import seja especfico
     */
    public JIdentifier getClassName() {
        return className;
    }

    /**
     * Seta o nome da classe importada, no caso de um import especfico
     *
     * @param className Nome da classe a ser importada.
     */
    public void setClassName(JIdentifier className) {
        if (className != null) {
            this.className = className;
        } else {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Seta esta declarao de importao, como sendo a importao de todas as
     * classes do pacote
     *
     * @param tf , true - para indicar que importa todas as classes do pacote
     *           false - caso contrrio.
     */
    public void setImportAllClasses(boolean tf) {
        this.importAllClasses = tf;
    }


    /**
     * Retorna true, caso todas as classes do pacote sejam importadas
     *
     * @return true - para indicar que importa todas as classes do pacote
     *         false - caso contrrio.
     */
    public boolean importAllClasses() {
        return this.importAllClasses;
    }


    /**
     * Retorna a assinatura desta declarao de import. Que corresponde ao
     * pacote ou classe sendo importado.
     *
     * @return a assinatura desta declarao de import.
     */
    public Signature getSignature() {
        JName retorno = null;
        if (this.isVariable() && !this.isExecutable()) {
            retorno = this.getPackageName();
            if (this.className != null) {
                retorno.addIdentifier(this.className);
            }
        }
        return retorno;
    }


    /**
     * Compara dois objetos do tipo JImportDeclaration e diz se estes so iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JImportDeclaration) {
            JImportDeclaration imp = (JImportDeclaration) obj;
            if (this.isASourceNode() && imp.isASourceNode()) {
                retorno = AbstractNode.equals(this.packageName, imp.packageName) &&
                        AbstractNode.equals(this.className, imp.className) &&
                        (this.importAllClasses == imp.importAllClasses);
            }
            // verifica se os ns JaTS so iguais
            else {
                retorno = compareJaTSNode(imp);
            }
        }
        return retorno;

    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * cada elemento de <code>this.dimensionSizes</code> e o de <code>this.init</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {

            if (className != null) {
                this.className.accept(visitor, data);
            }
            if (packageName != null) {
                this.packageName.accept(visitor, data);
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * a o qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG);
        } else if (!(node instanceof JImportDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {
            JImportDeclaration importDec = (JImportDeclaration) node;
            if (this.packageName != null) {
                this.packageName.match(importDec.getPackageName(), results);
            }
            if (this.className != null) {
                this.className.match(importDec.getClassName(), results);
            }
            if (this.importAllClasses != importDec.importAllClasses()) {
                throw new NodesNotMatchedException("Imports couldn't be matched",
                        this, node);
            }
        }
    }

    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {
            INode pckgName = this.packageName;
            if (pckgName != null) {
                Object ret = this.processNode(pckgName, data);
                if (pckgName.isExecutable()) {
                    this.setPackageName(JName.unwrapName(ret));

                }
            }

            INode cName = this.className;
            if (cName != null) {
                Object ret = this.processNode(cName, data);
                if (cName.isExecutable()) {
                    this.setClassName(JIdentifier.unwrapIdentifier(ret));

                }
            }
        }
        return retorno;
    }

} // fim da classe