/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa uma lista de objetos do tipo <code>JFieldDeclaration</code>.
 * 
 * @author Adeline de Sousa Silva
 * @see JInitializer
 * 
 */
public class JInitializerSet extends DeclarationSet {

    /**
     * Construtor vazio da classe.
     */
    public JInitializerSet() {
        this.setTypeIdentifier(JaTSNode.INITIALIZER_SET);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>NodeList</code>
     * contendo<br> apenas objetos do tipo JFieldDeclaration
     */
    protected JInitializerSet(NodeList objects) throws IllegalArgumentException {
         if (objects == null) {
            throw new IllegalArgumentException();
        }
        int i = 0;
        while (i < objects.size()) {
            if (!(objects.elementAt(i) instanceof JInitializer)) {
                throw new IllegalArgumentException();
            }
            i++;
        }
        declarations = objects;
        this.setTypeIdentifier(JaTSNode.INITIALIZER_SET);
    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param init Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @throws IllegalArgumentException Levantada quando o argumento<br>
     *                                  <code>init</code> for <code>null</code>.
     */
    public void add(INode init) throws IllegalArgumentException {

        if (init == null || !(init instanceof JInitializer)) {
            throw new IllegalArgumentException("JInitializer passado  null");
        }

        declarations.addElement(init);

    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para a<br>
     * direita na posies do item que esiver naquela posio e nas daqueles que<br>.
     * estiverem  direita dela.
     *
     * @param init Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @param pos  Um inteiro correspondendo  posio onde o n deve ser <br>
     *             inserido.
     * @throws IllegalArgumentException Levantada quando o <br>
     *                                  argumento <code>init</code> for <code>null</code> ou o <br>
     *                                  argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode init, int pos)
            throws IllegalArgumentException {
        if (init == null || pos < 0 || pos > this.size() ||
                !(init instanceof JInitializer)) {
            throw new IllegalArgumentException();
        }
        declarations.insertElementAt(init, pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista, <br>
     * sobreescrevendo o elemento que estava nela.
     *
     * @param init Um objeto qualquer do tipo <code>JFieldDeclaration</code>.
     * @param pos  Um inteiro correspondendo  posio onde o n deve ser <br>
     *             inserido.
     * @throws IllegalArgumentException Levantada quando o <br>
     *                                  argumento <code>init</code> for <code>null</code> ou o <br>
     *                                  argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode init, int pos)
            throws IllegalArgumentException {

        if (init == null || pos < 0 || pos > this.size() ||
                !(init instanceof JInitializer)) {
            throw new IllegalArgumentException();
        }
        declarations.setElementAt(init, pos);
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JFieldDeclaration</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JFieldDeclaration</code>.
     *                Por consequncia, objetos do tipo<code>JFieldDeclaration</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!(node instanceof JInitializerSet)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!(this.matchesAsAVariable(node, results,
                JaTSNode.INITIALIZER_SET))) {

            JInitializerSet initSet = ((JInitializerSet) node);

            NodeList nodeInitializers = (NodeList) initSet.getDeclarationSet().clone();
            int i = 0;
            int max = this.declarations.size();
            while (i < max) {
                JInitializer tmpInit = (JInitializer) this.declarations.elementAt(i);

                int j = 0;
                int max2 = nodeInitializers.size();
                boolean matchedOk = false;
                while (j < max2 && !matchedOk) {
                    JInitializer nodeToBeMatched =
                            (JInitializer) nodeInitializers.elementAt(j);

                    ResultSet tempRes = new ResultSet();
                    // Tenta casar tmpInit com qualquer inicializador
                    // presente em nodeInitializers.
                    try {
                        tmpInit.match(nodeToBeMatched, tempRes);
                        // Se tudo der certo, nodeToBeMatched  retirado de
                        // nodeInitializers e o resultado desse casamento individual
                        //  adicionado ao resultado do casamento como um todo.
                        results.merge(tempRes);
                        // a cada iterao, remove um elemento deste conjunto.
                        nodeInitializers.removeElementAt(j);
                        matchedOk = true;
                    } catch (NodesNotMatchedException e) {
                    }
                    j++;
                }
                // se no conseguiu casar tmpInit, com nenhuma das declaraes
                // presentes no conjunto de inicializadores de node, ento 
                // por que os ns no casam
                if (j == max2 && !matchedOk) {
                    throw new NodesNotMatchedException("InitializerSets couldn't " +
                            "be matched", tmpInit, initSet);
                }
                // incrementa i, pra poder ir pra outra varivel
                i++;
            }

        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        if (super.declarations != null) {
            int i = super.declarations.size() - 1;

            // empilha os inicializadores no sentido oposto
            while (i >= 0) {
                super.declarations.elementAt(i).accept(visitor, data);
                i--;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * duas maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 2 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * <p/>
     * Ns do tipo JFieldDeclarationSet nunca so executveis.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Levantada
     *                                   caso o parmetro <code>data</code> seja <code>null</code> numa
     *                                   situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada se um n inconsistente (incompleto) for encontrado.
     *                                   Se algum dos componentes (filhos) essenciais ao correto
     *                                   funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;
        // Processamento de cada uma dos inicializadores.
        int i = 0;
        int max = this.size();
        while (i < max) {
            INode currInit = this.elementAt(i);
            if (currInit != null) {
                Object ret = this.processNode(currInit, data);
                if (currInit.isExecutable()) {
                    if (!(ret instanceof JInitializer)) {
                        throw new ExecutionException(this);
                    }
                    this.setElementAt((JInitializer) ret, i);
                }
            }
            i++;
        }
        return retorno;
    }


} // fim da classe
