/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe guarda as informacoes a respeito de uma declaracao de interface.
 * Uma declarao de interface  constituda por uma lista de modificadores, um
 * identificador da interface, uma lista de interfaces estendidas e o corpo da
 * interface.
 *
 * @author Adeline de Sousa
 * @version 0.1 -  19/11/2001
 * @since JDK 1.3
 */
public class JInterfaceDeclaration extends JTypeDeclaration {

    public static final int METHOD_DECLARATION = JModifier.PUBLIC + JModifier.ABSTRACT;

    public static final int FIELD_DECLARATION = JModifier.PUBLIC + JModifier.STATIC + JModifier.FINAL;

    /**
     * Construtor vazio da classe
     */
    public JInterfaceDeclaration() {
        this.setTypeIdentifier(JaTSNode.INTERFACE_DECLARATION);
    }

    /**
     * Construtor da classe que recebe todos os parametros, referentes aos
     * atributos desta classe
     *
     * @param mods lista de modificadores
     * @param id   identificador da classe
     * @param ifs  lista de interfaces implementadas
     * @throws IllegalArgumentException - se o id passado for null.
     *                                  Os outros podem ser null, prque sao opcionais
     */
    public JInterfaceDeclaration(JModifierList mods, JIdentifier id, JNameList ifs) {
        super(mods, id, ifs);
        this.setTypeIdentifier(JaTSNode.INTERFACE_DECLARATION);
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (!(node instanceof JTypeDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, JaTSNode.INTERFACE_DECLARATION)) {

            try {
                JInterfaceDeclaration iDec = (JInterfaceDeclaration) node;
                super.match(iDec, results);

                if (this.getBody() != null && iDec.getBody() != null) {
                    this.getBody().match(iDec.getBody(), results);
                }
            } catch (ClassCastException exc) {
                throw new NodesNotMatchedException("The node is not an interface",
                        this, node);
            }
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.parameters</code> e <code>this.identifier</code>
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo
     *                desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>,que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.getBody() != null) {
            this.getBody().accept(visitor, data);
        }
        if (this.getInterfaces() != null) {
            this.getInterfaces().accept(visitor, data);
        }
        if (this.getName() != null) {
            this.getName().accept(visitor, data);
        }
        if (this.getModifiers() != null) {
            this.getModifiers().accept(visitor, data);
        }

        // Envie o visitor pela sub-arvore da lista de atributos
//        if (this.getAttributes() != null) {
//        	this.getAttributes().accept(visitor, data);
//        }
        
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {
            // processa os filhos - modifiers, name, interfaces e corpo
            super.process(data);
        }

        return retorno;
    }

    /**
     * Diz se uma declaraao de atributo eh uma declaraao membro valido na
     * declaraao da interface. Como declaraao valida aceita-se aquelas que ou
     * possuem os modificadores "public static final", ou qualquer combinaao
     * dele, ou nenhum modificador. Nestes dois ultimos casos, a lsita de
     * modificadores e setada para o valor JInterfaceDeclaration.FIELD_DECLARATION,
     * o qual representa os modificadores: "public static final".
     *
     * @return true, caso a delcaraao de atributo seja valida,
     *         false, caso contrario
     */
    private boolean isAValidFieldDeclaration(JFieldDeclaration fd) {
        boolean retorno = false;

//        if ((fd.getModifiers().isCompatible(FIELD_DECLARATION)) ||
//                (fd.getModifiers().hasNoModifiers()) || (fd.getModifiers().isVariable())) {
//            retorno = true;
//            if ((fd.getModifiers().hasNoModifiers())) {
//                JModifierList mlist =
//                        new JModifierList(FIELD_DECLARATION,
//                                JModifierList.FIELD_DECLARATION);
//                fd.setModifiers(mlist);
//            }
//
//
//        }
        // return retorno;
        return true;

    }

    /**
     * Diz se uma declaraao de atributo eh uma declaraao membro valido na
     * declaraao da interface. Como declaraao valida aceita-se aquelas que ou
     * possuem os modificadores "public abstract", ou qualquer combinaao
     * deles, ou nenhum modificador. Nestes dois ultimos casos, a lista de
     * modificadores e setada para o valor JInterfaceDeclaration.METHOD_DECLARATION,
     * o qual representa os modificadores: "public abstract".
     *
     * @return true, caso a declaraao de metodo seja valida,
     *         false, caso contrario
     */
    private boolean isAValidMethodDeclaration(JMethodDeclaration md) {
        boolean retorno = false;
        if ((md.getModifiers().isCompatible(METHOD_DECLARATION)) ||
                (md.getModifiers().hasNoModifiers()) ||
                (md.getModifiers().isVariable())) {
            retorno = true;
            if ((md.getModifiers().hasNoModifiers())) {
                md.setBody(null);
            }
        }
        return retorno;
    }


    /**
     * Adiciona uma declaraao de metodo, desde que esta seja valida como
     * membro de uma interface
     *
     * @param md Declaraao de metodo a ser adicionada
     */
    public void addMethodVarDeclaration(JMethodDeclaration md) {
        if (md != null && md.isVariable()) {
            this.getBody().addMethodVarDeclaration(md);
        } else {
            throw new IllegalArgumentException("Declaraao de varivel metodo nula");
        }

    }

    /**
     * Adiciona uma declaraao de metodo, desde que esta seja valida como
     * membro de uma interface
     *
     * @param mds Declaraao de metodo a ser adicionada
     */
    public void addMethodsVarDeclaration(JMethodDeclarationSet mds) {
        if (mds != null && mds.isVariable()) {
            this.getBody().addMethodsVarDeclaration(mds);
        } else {
            throw new IllegalArgumentException("Declaraao de variavel mtds invalida");
        }

    }


    /**
     * Adiciona uma declaraao de metodo, desde que esta seja valida como
     * membro de uma interface
     *
     * @param md Declaraao de metodo a ser adicionada
     */
    public void addMethodDeclaration(JMethodDeclaration md) {
        if (isAValidMethodDeclaration(md)) {
            this.getBody().addMethodDeclaration(md);
        } else {
            throw new IllegalArgumentException("Declaraao de metodo invalida");
        }

    }


    /**
     * Adiciona uma declaraao de atributo, desde que esta seja valida como
     * membro de uma interface
     *
     * @param fd Declaraao de atributo a ser adicionada
     */
    public void addFieldDeclaration(JFieldDeclaration fd) {
        if (isAValidFieldDeclaration(fd)) {
            this.getBody().addFieldDeclaration(fd);
        } else {
            throw new IllegalArgumentException("Declaraao de atributo invalida");
        }
    }

    /**
     * Adiciona uma varivel do tipo declaraao de atributo ao conjunto de
     * atributos desta interface
     *
     * @param fd Varivel do tipo JFieldDeclaration a ser adicionada
     */
    public void addFieldVarDeclaration(JFieldDeclaration fd) {
        if (fd != null && fd.isVariable()) {
            this.getBody().addFieldVarDeclaration(fd);
        } else {
            throw new IllegalArgumentException("Declaraao de atributo invalida");
        }
    }

    /**
     * Adiciona uma varivel do tipo conjunto de declaraoes de atributo ao
     * conjunto de atributos desta interface
     *
     * @param fds Varivel do tipo JFieldDeclarationSet a ser adicionada
     */
    public void addFieldsVarDeclaration(JFieldDeclarationSet fds) {
        if (fds != null && fds.isVariable()) {
            this.getBody().addFieldsVarDeclaration(fds);
        } else {
            throw new IllegalArgumentException("Declaraao de atributo invalida");
        }
    }


} // fim da classe
