/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa literais.
 *
 * @author Fernando Castor
 * @author Adeline Sousa
 *
 * @see JExpression
 * @version 03/02/2003
 */
public class JLiteral extends JExpression {

    //  Tipo de literal que est sendo representado.
    private int literalType;

    // Armazena o valor do literal como um objeto do tipo String.
    private String value;

    /* Constantes inteiras correspondendo aos possveis tipos de literal. */

    // Literal de um nmero inteiro.
    public final static int INTEGER_LITERAL = 1;

    // Literal de um nmero long.
    public final static int LONG_LITERAL = 2;

    // JLiteral de um nmero de ponto flutuante.
    public final static int FLOATING_POINT_LITERAL = 4;

    // JLiteral de um caractere.
    public final static int CHARACTER_LITERAL = 8;

    // JLiteral de um String.
    public final static int STRING_LITERAL = 16;

    // JLiteral de um valor booleano.
    public final static int BOOLEAN_LITERAL = 32;

    // JLiteral nulo.
    public final static int NULL_LITERAL = 64;
    
    // JLiteral de um String Verbatim (de C#)
    public final static int VERBATIM_STRING_LITERAL = 128;


    /**
     * Construtor vazio da classe
     */
    public JLiteral() {
        // por default, os literais sero considerados String
        this.literalType = STRING_LITERAL;
        this.value = null;
        this.setTypeIdentifier(JaTSNode.LITERAL);
    }

    /**
     * Construtor da classe que recebe como parmetro o tipo do literal e
     * um String correspondendo ao seu valor.
     *
     * @param litType Um inteiro correspondendo ao tipo do literal.
     * @param val     O valor do literal, representado como um String.
     */
    public JLiteral(int litType, String val) {
        this.literalType = litType;
        this.value = val;
        this.setTypeIdentifier(JaTSNode.LITERAL);
    }

    /**
     * Mtodo que retorna o tipo do literal.
     *
     * @return Uma inteiro correspondendo ao tipo de literal armazenado.
     */
    public int getLiteralType() {
        return this.literalType;
    }

    /**
     * Mtodo que retorna o valor do literal.
     *
     * @return Um Object correspondendo ao valor do literal.
     */
    public Object getLiteralValue() {
        Object retorno = null;

        if (this.getLiteralType() == INTEGER_LITERAL) {
            retorno = Util.parseIntegralNumber(this);
        }
        if (this.getLiteralType() == LONG_LITERAL) {
            retorno = Util.parseIntegralNumber(this);
        } else if (this.getLiteralType() == FLOATING_POINT_LITERAL) {
            // talvez seja necessrio criar um FLOAT_LITERAL e um DOUBLE_LITERAL
            retorno = Double.valueOf(this.value);
        } else if (this.getLiteralType() == CHARACTER_LITERAL) {
            // no caso de literal caracter, eles podem ser guardados entre
            // aspas simples.
            if (this.value != null && this.value.charAt(0) == '\'') {
                retorno = new Character((this.value).charAt(1));
            } else {
                retorno = new Character((this.value).charAt(0));
            }
        } else if (this.getLiteralType() == STRING_LITERAL) {
            //  necessrio pegar apenas um substring porque literais
            // de Strings so guardados entre aspas.
            if (this.value != null && this.value.charAt(0) == '\"') {
                retorno = (this.value).substring(1, (this.value).length() - 1);
            } else {
                retorno = this.value;
            }
        } else if (this.getLiteralType() == BOOLEAN_LITERAL) {
            retorno = Boolean.valueOf(this.value);
        } else if (this.getLiteralType() == NULL_LITERAL) {
            retorno = "null";
        } else if (this.getLiteralType() == VERBATIM_STRING_LITERAL) {
            //  necessrio pegar apenas um substring porque literais
            // de Strings Verbatim so guardados entre aspas e com uma
        	// arroba precedendo
            if (this.value != null && this.value.charAt(0) == '@') {
                retorno = (this.value).substring(2, (this.value).length() - 1);
            } else {
                retorno = this.value;
            }
        }

        return retorno;
    }

    /**
     * Mtodo que seta o tipo do literal.
     *
     * @param litType Um inteiro correspondente ao tipo do literal.
     */
    public void setLiteralType(int litType) {
        this.literalType = litType;
    }

    /**
     * Mtodo que seta o valor do literal.
     *
     * @param val Um String representando o valor do literal.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          argumento <code>val</code> for <code>null</code>.
     */
    public void setLiteralValue(String val) {

        if (val == null) {
            throw new IllegalArgumentException();
        }
        if (val.length() == 0) {
            this.value = "\"\""; //para o caso em que aparece a expresao ""
        } else {
            this.value = val;
        }
    }

    public String getValue() {
        return value;
    }

    /**
     * Retorna o name correspondente a este literal.
     *
     * @return Name correspondente a este literal.
     */
    public JName getName() {
        JName name = new JName(this.value);
        return name;
    }

    /**
     * Retorna um novo objeto do tipo <code>JLiteral</code>  contendo o valor
     * do objeto atual, acrescido do prefixo passado como parmetro.
     *
     * @param prefix Prefixo a ser adicionado
     * @return O literal acrescido do prefixo passado como parmetro
     */
    public JLiteral addPrefix(String prefix) {
        JLiteral retorno = new JLiteral();
        if (Util.checkString(prefix)) {
            String tmpValue = prefix + this.value;
            retorno.setLiteralValue(tmpValue);
        }
        return retorno;
    }

    /**
     * Retorna um novo objeto do tipo <code>JLiteral</code>  contendo o valor
     * do objeto atual, acrescido do sufixo passado como parmetro.
     *
     * @param suffix Prefixo a ser adicionado
     * @return O literal acrescido do prefixo passado como parmetro
     */
    public JLiteral addSuffix(String suffix) {
        JLiteral retorno = new JLiteral();
        if (Util.checkString(suffix)) {
            String tmpValue = this.value + suffix;
            retorno.setLiteralValue(tmpValue);
        }
        return retorno;
    }

    /**
     * Mtodo que diz se dois objetos do tipo <code>JLiteral</code> so
     * iguais. Para que isso seja verdade, os dois tm que ter valores iguais ou
     * os dois represetnam variveis que so do mesmo tipo e que tm o mesmo
     * nome.
     *
     * @param objeto Um objeto qualquer que ser comparado com <code>this</code>.
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica que
     *         o valor do objeto com o qual <code>this</code> est sendo
     *         comparado  igual ao de <code>this</code>.
     */
    public boolean equals(Object objeto) {
        boolean retorno = false;
        if (objeto instanceof JLiteral) {
            JLiteral lit = (JLiteral) objeto;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.getLiteralValue(), lit.getLiteralValue()) &&
                        (this.literalType == lit.literalType);
            } else {
                retorno = compareJaTSNode(lit);
            }

        }

        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando qualquer
     *          um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor ilegal,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JLiteral)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JLiteral exp = (JLiteral) node;

            if (this.getLiteralValue() == null)
                throw new InconsistentNodeException(this);

            if (this.getLiteralType() != exp.getLiteralType() ||
                    !(this.getValue()).equals(exp.getValue())) {
                throw new NodesNotMatchedException("Literals are different", this, node);
            }
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento de literais
     * ocorre se apenas se o n atual for executvel. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu mtodo
     * process e corresponde ao resultado de seu processamento.
     * <p/>
     * S h essa modalidade de processamento para literais porque eles nunca tm
     * ns filhos. O mesmo  verdade para ns dos tipos <code>JIdentifier</code>
     * e <code>JModifierList</code>.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this.getValue();

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        return retorno;
    }

    public Object evaluate(Object data) throws ExecutionException {
        return this.getLiteralValue();
    }


}
