/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa uma declarao de varivel local. Por exemplo:
 * int x,y;
 * int tamanho;
 * final String texto = "Teste";
 * 
 * @author Adeline de Sousa Silva
 * @version 09/07/2002
 */
public class JLocalVariableDeclaration extends JVariableDeclaration {


    /**
     * Construtor vazio da classe.
     */
    public JLocalVariableDeclaration() {
        this.modifiers = new JModifierList(JModifierList.LOCAL_VARIABLE_DECLARATION);
        this.setTypeIdentifier(JaTSNode.LOCAL_VARIABLE_DECLARATION);

    }


    /**
     * Construtor da classe que recebe o tipo  e a lista de variaveis.
     *
     * @param t      tipo da(s) variavel(eis)
     * @param varDec lista  de variaveis
     */
    public JLocalVariableDeclaration(JType t, JVariableDeclaratorSet varDec) {
        this.type = t;
        this.variables = varDec;
        this.setTypeIdentifier(JaTSNode.LOCAL_VARIABLE_DECLARATION);
    }


//	public void setAttributes(AttributeSectionList attributes) {
//		throw new UnsupportedOperationException();
//	}
	
	/**
     * Seta o modificador da declarao de varivel local.
     *
     * @param modList Lista de modificadores da varivel local
     * @throws IllegalArgumentException - se o modList passado for null.
     */
    public void setModifiers(JModifierList modList)
            throws IllegalArgumentException {

        if (modList != null &&
                modList.getPossibleModifiers() == JModifierList.LOCAL_VARIABLE_DECLARATION) {
            this.modifiers = modList;
//        } else {
//            throw new IllegalArgumentException("Lista de modificadores nula");
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.modifiers</code>, <code>this.type</code> e
     * <code>this.variables</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (!this.isExecutable() && !this.isVariable()) {

            // Envie o visitor pela sub-arvore das declaracoes de variavel
            // pega o tamanho do nodeList, subtrrai um para poder faze-lo de
            // indice, ja que o indice, comeca em zero
            int i = (this.variables.size()) - 1;
            while (i >= 0) {
                // envia o visitor por cada n do tipo JVariableDeclarator,
                // comecando pelo que est na posicao de mais alto valor
                this.variables.elementAt(i).accept(visitor, data);
                i--;
            }
            if (this.type != null) {
                this.type.accept(visitor, data);
            } else if (!this.isExecutable()) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }
            if (this.modifiers != null) {
                this.modifiers.accept(visitor, data);
            }

        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);


        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }    

}
