/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe guarda as informacoes a respeito da declaracao de metodo, a qual
 *  composta por uma lista de modificadores, o tipo de retorno do mtodo , o
 * identificador, a lista de parametros do mtodo e a lista de exceoes(no caso
 * de C#, no existe a lista de excecoes).
 *
 * <br>
 * Ex:
 * <br>  <code>public double getValor() throws RuntimeException {}</code>
 * <p/>
 * <br>  <code>public void setValor(double v){}</code>
 * 
 * @author Adeline de Sousa
 * @version   22/08/2002
 * @since JDK 1.3
 */
public class JMethodDeclaration
        extends AbstractNode
        implements Declaration, Cloneable, Commentable, InterfaceMember {

    // comentrio referente a este n
    private JComment comment;

//    private AttributeSectionList attributes;

    //lista de modificadores do mtodo
    private JModifierList modifiers;

    // tipo de retorno do mtodo
    private JType returnType;

    // Identificador do mtodo
    private JName name;

    // Lista de parametros formais
    private JParameterList parameters;

    // Se o tipo de retorno for um array, os colchetes podem vir no final
    // dimensions = 0 indicam que o retorno no  do tipo array
    private int dimensions;

    // lista de excees levantadas pelo  mtodo
    private JNameList exceptions;

    // Armazena o corpo do mtodo
    private JBlock body;

    //  usado para indicar se o metodo possui ou no um corpo.
    private boolean hasBody = false;

    public static String BLOCK_IDENT = "    ";

    /**
     * Construtor vazio da classe
     */
    public JMethodDeclaration() {
        this.modifiers = new JModifierList(JModifierList.METHOD_DECLARATION);
        this.returnType = new JType();
        this.name = new JName() ;
        this.parameters = new JParameterList();
        this.dimensions = 0;
        this.exceptions = new JNameList();
        this.body = null;
        this.setTypeIdentifier(JaTSNode.METHOD_DECLARATION);
    }

    /**
     * Construtor da classe
     *
     * @param mods       Lista de modificadores
     * @param type       Tipo de retorno
     * @param name         Identificador do metodo
     * @param parameters Lista de parametros do metodo
     * @param exc        Lista de excees levantadas pelo mtodo
     */
    public JMethodDeclaration(JModifierList mods,
                              JType type,
                              JName name,
                              JParameterList parameters,
                              JNameList exc,
                              JBlock body) {
        this.setModifiers(mods);
        this.setResultType(type);
        this.setName(name);
        this.setParameters(parameters);
        this.dimensions = 0;
        this.setExceptions(exc);
        this.setBody(body);
        this.setTypeIdentifier(JaTSNode.METHOD_DECLARATION);
    }

    /**
     * Retorna o comentrio associado ao mtodo. Pode ser o javadoc setado pelo
     * parsing. Ou seja, o javadoc presente no mtodo ou o javadoc criado pelo
     * mecansimo de PrettyPrinting
     *
     * @return javadoc do mtodo
     */
    public JComment getComment() {
        return this.comment;
    }

    /**
     * Seta o javadoc associado ao mtodo.
     *
     * @param javadoc do mtodo
     */
    public void setComment(JComment javadoc) {
        if (javadoc != null) {
            this.comment = javadoc;

        }
    }

//     public AttributeSectionList getAttributes() {
//        return this.attributes;
//    }
//
//    public void setAttributes(AttributeSectionList att) {
//       this.attributes = att;
//    }

    /**
     * Retorna a lista de modificadores do mtodo
     *
     * @return Modificadores do mtodo
     */
    public JModifierList getModifiers() {
        return this.modifiers;
    }

    /**
     * Seta a lista de modificadores do mtodo.
     *
     * @param modList Lista de modificadores deste mtodo
     * @throws IllegalArgumentException - se o modList passado for null.
     */
    public void setModifiers(JModifierList modList)
            throws IllegalArgumentException {

        if (modList != null) {
            this.modifiers = modList;
        } else {
            throw new IllegalArgumentException("Lista de modificadores nula");
        }
    }

    /**
     * Retorna o tipo de retorno do mtodo
     *
     * @return Tipo de retorno do mtodo
     */
    public JType getResultType() {

        return this.returnType;
    }

    /**
     * Seta o tipo do mtodo com o valor passado como parametro
     *
     * @param t tipo do mtodo
     * @throws IllegalArgumentException - se o tipo passado for nulo
     */
    public void setResultType(JType t) {

        if (t != null) {
            this.returnType = t;
        } else {
            throw new IllegalArgumentException("Tipo de retorno nulo");
        }
    }

    /**
     * Retorna o identificador do metodo.
     *
     * @return Identificador do mtodo
     */
    public JName getName() {

        return this.name;
    }

    /**
     * Seta o identificador do metodo.
     *
     * @param methodName Identificador do mtodo
     * @throws IllegalArgumentException - Se o identificador passado for nulo
     */
    public void setName(JName methodName) {
        if (methodName != null) {
            this.name = methodName;
        } else {
            throw new IllegalArgumentException("Identificador do metodo nulo");
        }
    }

    /**
     * Retorna a lista de parametros do metodo
     *
     * @return Lista de Parametros
     */
    public JParameterList getParameters() {
        return this.parameters;
    }

    /**
     * Seta a lista de parametros do metodo
     *
     * @param param Lista de parametros
     * @throws IllegalArgumentException - Se a lista passada for nula ou se
     *                                  contiver elementos que nao sejam do tipo JFormalParameter
     */
    public void setParameters(JParameterList param) {

        if (param != null) {
            this.parameters = param;
        } else {
            throw new IllegalArgumentException("lista de parametros nula");
        }
    }

    /**
     * Retorna a dimensao do tipo de retorno. Este valor  quase sempre zero, a
     * no ser nos casos em que o tipo de retorno  um array
     *
     * @return Dimensao
     */
    public int getDimensions() {

        return this.dimensions;

    }

    /**
     * Seta a dimensao do array retornado pelo mtodo. Um valor zero indica
     * que o retorno do mtodo no  um array
     *
     * @param dim
     * @throws IllegalArgumentException - Se o identificador passado for nulo
     */
    public void setDimensions(int dim) {
        if (dim >= 0) {
            this.dimensions = dim;
        } else {
            throw new IllegalArgumentException("Valor de dimenso invlido (<0)");
        }
    }

    /**
     * Retorna a lista de excees do mtodo
     *
     * @return Lista de Excees do mtodo
     */
    public JNameList getExceptions() {

        return this.exceptions;
    }

    /**
     * Seta a lista de excees levantadas por este mtodo.
     *
     * @param excs Lista de excees levantadas pelo mtodo
     */
    public void setExceptions(JNameList excs) {
        this.exceptions = excs;
    }

    /**
     * Retorna o corpo do mtodo.
     *
     * @return O corpo do mtodo
     */
    public JBlock getBody() {
        return this.body;
    }

    /**
     * Diz se o n possui comentrio.
     *
     * @return true, se  o n possui um comentrio
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return (this.comment != null && this.comment.getText() != null);
    }

    /**
     * Diz se o n possui comentrio.
     *
     * @return true, se  o n possui um comentrio
     *         false, caso contrrio
     */
    public boolean hasJavaDocComment() {
        return (this.hasComment() && this.comment.getCommentCode() == JComment.JAVADOC);
    }

    /**
     * Indica se o mtodo possui ou no um corpo.
     */
    public boolean hasBody() {
        return this.hasBody;
    }

    /**
     * Retorna o tipo de retorno do mtodo
     *
     * @return Tipo de retorno do mtodo
     */
    public JType getMemberType() {

        return this.returnType;
    }

    /**
     * Verifica se o modificador passado como parmetro pertence  lista de
     * modificadores de uma declarao de mtodo.
     *
     * @param strMod O modificador que se deseja procurar
     * @return <code>true</code>, se o modificador passado como
     *         parmetro estiver na lista. <code>false</code>, caso contrrio
     * @ exception IllegalArgumentException - Se o modificador passado
     * for invalido
     */
    public boolean hasModifier(String strMod) throws IllegalArgumentException {

        return this.modifiers.hasModifier(strMod);
    }

    /**
     * Verifica a lista de modificadores desta declarao de mtodo contm um
     * modificador de acesso.
     *
     * @return <code>true</code>, se a lista de modificadores j contm um
     *         modificador de acesso. <code>false</code>, caso contrrio
     */
    public boolean hasAccessModifier() throws IllegalArgumentException {

        return this.modifiers.hasAccessModifier();
    }

    /**
     * Seta o javadoc associado ao mtodo.
     *
     * @param javadoc do mtodo
     */
    public void setComment(String javadoc) {
        if (javadoc != null && javadoc.length() > 0) {
            this.comment = new JComment(javadoc);

        }
    }


    /**
     * Diz se o mtodo no contm statements.
     *
     * @return true, se for um mtodo vazio
     *         false, caso contrrio.
     */
    public boolean isEmpty() {
        return (this.body == null || this.body.size() == 0);
    }

    /**
     * Seta o corpo do mtodo
     *
     * @param mBody O corpo do mtodo
     */
    public void setBody(JBlock mBody) {
        this.body = mBody;
        this.hasBody = ((mBody != null)
                && (!this.hasModifier("abstract"))
                && (!this.hasModifier("native")));
    }

    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        MethodSignature result = null;
        if (this.isASourceNode()) {
            result = new MethodSignature(this.getName(), this.getParameters());

        }
        return result;

    }

    /**
     * Retorna um inteiro que identifica o tipo de retorno do mtodo. No caso de
     * o tipo corresponder a um objeto qualquer, simplesmente retorna
     * object. Caso contrrio retorna um dos tipos primitivos ou void.
     * Tipos passveis de serem retornados:
     * <p/>
     * JType.VOID
     * JType.BOOLEAN
     * JType.BYTE
     * JType.CHAR
     * JType.SHORT
     * JType.INT
     * JType.LONG
     * JType.FLOAT
     * JType.DOUBLE
     * JType.OBJECT
     */
    public int getType() {
        return this.returnType.getType();
    }

    /**
     * Retorna o tipo de retorno do mtodo em forma de String.
     */
    public String getStrType() {
        return this.returnType.getStrType();
    }

    /**
     * Adiciona um modificador  lista de modificadores do mtodo
     *
     * @param mod Modificador a ser inserido
     * @throws IllegalArgumentException Se o modificador passado j existir na
     *                                  lista de modificadores deste mtodo ou se este no  um
     *                                  modificador vlido para mtodos
     */
    public void addModifier(String mod) {
        if (this.hasModifier(mod)) {
            throw new IllegalArgumentException("Modificador ja existe");
        } else {
            this.modifiers.addModifier(mod);
        }
    }

    /**
     * Remove um modificador da lista de modificadores deste mtodo.
     *
     * @param strMod Modificador a ser removido
     * @throws IllegalArgumentException - Se o modificador passado nao
     *                                  pertencer a lista de modificadores
     */
    public void removeModifier(String strMod) throws IllegalArgumentException {

        this.modifiers.removeModifier(strMod);
    }

    /**
     * Adiciona uma exceo  lista de excees do mtodo.
     *
     * @param exc Exceo a ser inserida
     * @throws IllegalArgumentException Se o exceo passada j existir na
     *                                  lista de excees deste mtodo
     */
    public void addException(String exc) {
        if (this.hasException(exc)) {
            throw new IllegalArgumentException("Excecao ja existe");
        } else {
            JName n = new JName(exc);
            this.exceptions.add(n);
        }
    }

    /**
     * Verifica se uma exceo est na lista de excees deste mtodo.
     *
     * @param exc exceo  procurada
     * @return true, se a exceo existir
     *         false, caso contrrio
     */
    public boolean hasException(String exc) {
        JName n = new JName(exc);
        return this.exceptions.contains(n);
    }

    /**
     * Diz se o mtodo possui excees.
     *
     * @return true, se a existirem excees
     *         false, caso contrrio
     */
    public boolean hasExceptions() {
        return (this.exceptions != null && this.exceptions.size() > 0);
    }

    /**
     * Remove a excecao passada como parametro da lista de parametros.
     *
     * @param exc Excecao a ser removida
     */
    public void removeException(String exc) {
        JName n = new JName(exc);
        int pos = this.exceptions.indexOf(n);
        this.exceptions.removeElementAt(pos);

    }

    /**
     * Compara dois objetos do tipo <code>JMethodDeclaration</code> e diz se estes
     * so idnticos.
     *
     * @param obj Objeto com o qual ir se comparar o objeto this.
     * @return true, se os objetos forem identicos
     *         false, caso contrrio
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JMethodDeclaration) {
            JMethodDeclaration md = (JMethodDeclaration) obj;
            if (this.isASourceNode()) {
                retorno =
                        AbstractNode.equals(this.modifiers, md.modifiers)
                        && AbstractNode.equals(this.returnType, md.returnType)
                        && AbstractNode.equals(this.name, md.name)
                        && AbstractNode.equals(this.parameters, md.parameters)
                        && (this.dimensions == md.dimensions)
                        && AbstractNode.equals(this.exceptions, md.exceptions)
                        && AbstractNode.equals(this.body, md.body);
            } else {
                retorno = compareJaTSNode(md);
            }
        }
        return retorno;
    }

    /**
     * Retorna um corpo default para o mtodo, baseado no seu tipo de retorno -
     * til para a implementao de stubs.
     *
     * @return Corpo default para o mtodo, baseado no tipo de retorno.
     */
    public String getDefaultBody() {
        String retorno = null;
        int dim = 0;
        if (this.dimensions > 0) {
            dim = this.dimensions;
        }
        // no uso else, pois o numero de dimenses total  composto pelas
        // dimenses do tipo mais as dimenses do MethodDeclarator
        // (nome, argumentos e dimenses)
        if (this.getResultType().getNumberOfDimensions() > 0) {
            dim = dim + this.getResultType().getNumberOfDimensions();
        }
        boolean array = dim > 0;
        String brackets = "";
        if (array) {
            // o primeiro, deve conter ao menos um indicador de dimenso.
            brackets = "[0]";
            for (int i = 0; i < dim - 1; i++) {
                brackets += "[]";
            }
        }
        switch (this.getType()) {
            case JType.VOID:
                retorno = "{  }";
                break;
            case JType.BOOLEAN:
                if (array) {
                    retorno =
                            "{\n\t return new  boolean "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n\t return false;\n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.BYTE:
                if (array) {
                    retorno =
                            "{\n return new byte "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n return 0;\n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.CHAR:
                if (array) {
                    retorno =
                            "{\n return new  char "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n return '\0'; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.SHORT:
                if (array) {
                    retorno =
                            "{\n\t return new short "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n\t return 0; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.INT:
                if (array) {
                    retorno =
                            "{\n return new  int "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n\t return 0; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.LONG:
                if (array) {
                    retorno =
                            "{\n\t return new  long "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n return 0; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.FLOAT:
                if (array) {
                    retorno =
                            "{\n\t return new float "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n\t return 0; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.DOUBLE:
                if (array) {
                    retorno =
                            "{\n\t return new  double "
                            + brackets
                            + "; \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno = "{\n\t return 0; \n" + BLOCK_IDENT + "}";
                }
                break;
            case JType.OBJECT:
                if (array) {
                    String type = this.getResultType().getTypeName().toString();
                    retorno =
                            "{ \n\t return new "
                            + type
                            + brackets
                            + ";  \n"
                            + BLOCK_IDENT
                            + "}";
                } else {
                    retorno =
                            "{\n\t throw new UnsupportedOperationException(); \n"
                            + BLOCK_IDENT
                            + "}";
                    // ou  melhor colocar "return null" ?
                }
                break;
        }
        return retorno;
    }

    /**
     * Retorna um corpo default para o mtodo, baseado no seu tipo de retorno -
     * til para a implementao de stubs.
     */
    public void addDefaultImplementation() {
        this.append(this.getDefaultBody());
    }

    /**
     * Anexa a String passada como parmetro ao fim do corpo do mtodo. Se no
     * houver um corpo, apenas seta a String como sendo o corpo do mtodo
     */
    public void append(String appBlock) {
        JNodeString append = new JNodeString(appBlock);
        if (this.hasBody()) {
            this.body.add(append);
        } else {
            this.setBody(new JBlock());
            this.body.add(append);
        }
    }

    /**
     * Retorna a representao deste mtodo como sendo membro de uma interface
     * Ou seja, se for um mtodo de classe, ele apenas remove o seu corpo e
     * adiciona um ";" ao final.
     *
     * @return Representao do mtodo como parte de uma interface
     */
    public JMethodDeclaration toInterfaceMethodDeclaration() {
        JMethodDeclaration retorno = new JMethodDeclaration();
        if (this.hasComment()) {
            retorno.setComment(this.getComment());
        }
        // s funciona se o mtodo no  nem final nem esttico
        if (!this.hasModifier("static") || !this.hasModifier("final")) {

            // se o mtodo tem modificadores "native" ou "synchronized", que
            // sao mtodos de implementao, ns removemos estes modificadores
            if (this.hasModifier("native")) {
                this.removeModifier("native");
            }
            if (this.hasModifier("synchronized")) {
                this.removeModifier("synchronized");
            }
            retorno.setModifiers(this.getModifiers());
            retorno.setResultType(this.getResultType());
            retorno.setName(this.getName());
            retorno.setParameters(this.getParameters());
            retorno.setExceptions(this.getExceptions());
            retorno.setBody(null);
        }
        return retorno;

    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore<br>
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz <br>
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns <br>
     * casam em duas situaes: quando eles so completamente idnticos <br>
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda<br>
     * no foi mapeada a nenhum valor no conjunto resultado do casamento <br>
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele<br>
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto<br>
     * acima, se refere ao n que faz parte do programa JaTS que est sendo <br>
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar<br>
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo <br>
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro<br>
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada<br>
     *                                   quando algum componente essencial do n tiver um valor, <br>
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar<br>
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou<br>
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code>  <br>
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws
            IllegalArgumentException,
            InconsistentNodeException,
            NodesNotMatchedException
	{

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }
        if (!(node instanceof JMethodDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this,
                    node);
        } else if (
                !this.matchesAsAVariable(node,
                        results,
                        JaTSNode.METHOD_DECLARATION)) {
            JMethodDeclaration metDec = (JMethodDeclaration) node;

            // casa o comentario
            if (this.getComment() == null) {
                if (metDec.getComment() != null) {
                    JComment comment = metDec.getComment();
                    String commentVar =
                            Util.generateCommentVariable(this.getName());
                    results.put(commentVar, comment, JaTSNode.COMMENT);
                }
            }

//            if (this.attributes != null) {
//            	//casa os atributos
//            	this.attributes.match(metDec.attributes, results);
//            } else if (metDec.attributes != null) {
//            	throw new NodesNotMatchedException("Attributes don't match", this, metDec);
//            }

            // casa os modificadores
            this.modifiers.match(metDec.getModifiers(), results);

            // casa os tipos de retorno
            this.returnType.match(metDec.getResultType(), results);

            if (this.dimensions != metDec.dimensions) {
                throw new NodesNotMatchedException("Array dimensions are incompatible ("
                        + this.dimensions
                        + ", "
                        + metDec.dimensions
                        + ")",
                        this,
                        node);
            }

            // casa o identificador
            if (this.name == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }

            this.name.match(metDec.name, results);

            // casa o identificador
            if (this.parameters == null) {
                throw new InconsistentNodeException(this);
            }

            this.parameters.match(metDec.parameters, results);

            if (this.exceptions != null) {
                // casa a lista de excees
                this.exceptions.match(metDec.exceptions, results);
            } else if (metDec.getExceptions() != null) {
                throw new NodesNotMatchedException("Exceptions lists do not match",
                        this,
                        metDec);
            }

            // casa o corpo do mtodo
            if (this.body != null) {
                this.body.match(metDec.body, results);
            }

            this.setComment(comment);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data)
            throws
            ExecutionException,
            InconsistentNodeException,
            IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            //Processamento dos atributos
//        	AttributeSectionList tmpAttribute = this.getAttributes();
//            if (tmpAttribute != null) {
//                Object ret = this.processNode(tmpAttribute, data);
//                if (tmpAttribute.isExecutable()) {
//                    if (!(ret instanceof AttributeSectionList)) {
//                        throw new ExecutionException(this);
//                    }
//                    this.setAttributes((AttributeSectionList) ret);
//                }
//            }

            // Processamento da lista de modificadores do mtodo
            INode mods = this.getModifiers();
            if (mods != null) {
                Object ret1 = this.processNode(mods, data);
                if (mods.isExecutable()) {
                    if (!(ret1 instanceof JModifierList)) {
                        throw new ExecutionException(this);
                    }

                    this.setModifiers((JModifierList) ret1);
                }
            }
            // Processamento do tipo do mtodo.
            INode mType = this.getResultType();
            if (mType != null) {
                Object ret2 = this.processNode(mType, data);

                if (mType.isExecutable()) {

                    try {
                        this.setResultType(JType.unwrapType(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
            // Processamento do nome do mtodo
            INode tmpName = this.getName();
            if (tmpName != null) {
                Object ret3 = this.processNode(tmpName, data);

                if (tmpName.isExecutable()) {
                    try {
                        this.setName(JName.unwrapName(ret3));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
            // Processamento da lista de parmetros do mtodo.
            INode pars = this.getParameters();
            if (pars != null) {
                Object ret4 = this.processNode(pars, data);

                if (pars.isExecutable()) {
                    if (ret4 instanceof JParameterList) {
                        this.setParameters((JParameterList) ret4);
                    } else {
                        throw new ExecutionException(this);
                    }
                }
            }
            // Processamento da lista de excees que o mtodo levanta.
            INode exceptList = this.getExceptions();
            if (exceptList != null) {
                Object ret5 = this.processNode(exceptList, data);
                if (exceptList.isExecutable()) {
                    if (ret5 instanceof JNameList) {
                        this.setExceptions((JNameList) ret5);

                    } else {
                        throw new ExecutionException(this);
                    }
                }
                // aps a execuo, deixa de ser opcional.
                // Se executou,  pq existe
                this.getExceptions().setOptional(false);
            }

            // Processamento do corpo do mtodo
            INode mBody = this.getBody();
            if (mBody != null) {
                Object ret6 = this.processNode(mBody, data);
                if (mBody.isExecutable()) {
                    if (ret6 instanceof JBlock) {
                        this.setBody((JBlock) ret6);
                    } else {
                        throw new ExecutionException(this);
                    }
                }
            }

        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.parameters</code> e <code>this.identifier</code>
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de
     *                acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo
     *                desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>,que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws
            InconsistentNodeException,
            IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        if (this.isASourceNode()) {

            // Envie o visitor pela sub-rvore do corpo
            if (this.body != null) {
                this.body.accept(visitor, data);
            }

            // Envie o visitor pela sub-arvore de excecoes
            if (this.exceptions != null) {
                this.exceptions.accept(visitor, data);
            }

            // Envie o visitor pela sub-rvore dos parametros do metodo
            if (this.parameters != null) {
                this.parameters.accept(visitor, data);
            }

            // Envie o visitor pela sub-arvore do identificador
            if (this.name != null) {
                this.name.accept(visitor, data);
            }

            // Envie o visitor pela sub-arvore do tipo
            this.returnType.accept(visitor, data);

            // Envie o visitor pela sub-arvore da lista de modificadores
            this.modifiers.accept(visitor, data);
            
            // Envie o visitor pela sub-arvore da lista de atributos
//            if (this.attributes != null) {
//            	this.attributes.accept(visitor, data);
//            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

} // fim da classe
