/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;


/**
 * Esta classe define uma lista de objetos do tipo <code>JMethodDeclaration</code>.
 * 
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 */

public class JMethodDeclarationSet extends DeclarationSet {

    /**
     * Construtor vazio da classe.
     */
    public JMethodDeclarationSet() {
        this.setTypeIdentifier(JaTSNode.METHOD_DECLARATION_SET);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>NodeList</code>
     * contendo<br> apenas objetos do tipo JMethodDeclaration
     */
    protected JMethodDeclarationSet(NodeList objects)
            throws IllegalArgumentException {
   
        if (Util.verifyNodeList(objects, JaTSNode.METHOD_DECLARATION)) {
            this.declarations = objects;
        } else {
            throw new IllegalArgumentException("lista de metodos invalida");
        }
        this.setTypeIdentifier(JaTSNode.METHOD_DECLARATION_SET);
    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param method Um objeto qualquer do tipo <code>JMethodDeclaration</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento<br>
     *          <code>method</code> for <code>null</code>.
     */
    public void add(INode method)
            throws IllegalArgumentException {

        if (method == null || !(method instanceof JMethodDeclaration)) {
            throw new IllegalArgumentException("JMethodDeclaration passada  null");
        }

        // CORRIGIR : na verdade, nao  corrigir, mas ver se isto nao acarreta em erros
        // s adiciona se o mtodo j nao existir no conjunto
        if (!this.contains(((JMethodDeclaration) method).getSignature())) {
            this.declarations.addElement(method);
        }
    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para a<br>
     * direita na posies do item que esiver naquela posio e nas daqueles que<br>.
     * estiverem  direita dela.
     *
     * @param method Um objeto qualquer do tipo <code>JMethodDeclaration</code>.
     * @param pos    Um inteiro correspondendo  posio onde o n deve ser <br>
     *               inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>method</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode method, int pos)
            throws IllegalArgumentException {

        if ((method == null) || !(method instanceof JMethodDeclaration) ||
                (pos < 0) || (pos > this.size())) {
            throw new IllegalArgumentException();
        }

        this.declarations.insertElementAt(method, pos);
    }

    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista, <br>
     * sobreescrevendo o elemento que estava nela.
     *
     * @param method Um objeto qualquer do tipo <code>JMethodDeclaration</code>.
     * @param pos    Um inteiro correspondendo  posio onde o n deve ser <br>
     *               inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>method</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode method, int pos)
            throws IllegalArgumentException {
        boolean validPosition = pos >= 0 && pos < this.size();
        if (!(method instanceof JMethodDeclaration) || !(validPosition)) {
            throw new IllegalArgumentException();
        }
        this.declarations.setElementAt(method, pos);
    }

    /**
     * Indica se o conjunto contem o metodo dado
     *
     * @param methodName
     * @return
     */
    public boolean contains(String methodName) {
        return (this.getMethodDeclaration(methodName) != null);
    }

    /**
     * Indica se o conjunto contem o metodo dado
     *
     * @param methodSignature
     * @return
     */
    public boolean contains(Signature methodSignature) {
        boolean retVal = false;

        JMethodDeclaration mtd;
        for (int i = 0; i < this.size(); i++) {
            mtd = (JMethodDeclaration) this.elementAt(i);

            if (mtd.getSignature().equals(methodSignature)) {
                retVal = true;
                break;
            }
        }

        return retVal;
    }


    /**
     * Retorna o objeto do tipo <code>JMethodDeclaration</code>
     */
    public JMethodDeclaration getMethodDeclaration(String methodName) {
        JMethodDeclaration retorno = null;
        JaTSIterator it = this.getIterator();
        JMethodDeclaration tmpMethod = null;
        boolean foundIt = false;
        JName methodId;
        String methName;
        if (Util.checkString(methodName)) {
            while (it.hasNext() && !foundIt) {
                tmpMethod = (JMethodDeclaration) it.next();
                if (tmpMethod.isASourceNode()) {
                    methodId = tmpMethod.getName();
                    methName = methodId.toString();
                    if (methodName.equals(methName)) {
                        retorno = tmpMethod;
                        foundIt = true;
                    }
                } else {
                    if (tmpMethod.getVariableName().equals(methodName)) {
                        retorno = tmpMethod;
                        foundIt = true;
                    }
                }
            }
        } else {
            throw new IllegalArgumentException("Argumento methodName nulo em" +
                    "JMethodDeclarationSet.getMethodDeclaration");
        }
        return retorno;
    }

    /**
     * Retorna a lista de excees do mtodo especificado pelo parmetro.
     * O mtodo deve pertencer este conjunto de declaraes
     *
     * @param methodName Nome do mtodo cujas excees e
     * @return Lista de excees do mtodo especificado pelo parametro
     */
    public JNameList getExceptions(String methodName) {
        JNameList retorno = new JNameList();
        if (Util.checkString(methodName)) {
            JMethodDeclaration method = this.getMethodDeclaration(methodName);
            if (method != null) {
                retorno = method.getExceptions();
            }
        } else {
            throw new IllegalArgumentException("Argumento methodName nulo em  " +
                    "JMethodDeclarationSet.getExceptions ");
        }
        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JMethodDeclaration</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JMethodDeclaration</code>.
     *                Por consequncia, objetos do tipo<code>JMethodDeclaration</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!(node instanceof JMethodDeclarationSet)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }
        if (!(this.matchesAsAVariable(node, results,
                JaTSNode.METHOD_DECLARATION_SET))) {
            JMethodDeclarationSet methodsList = (JMethodDeclarationSet) node;

            // Casamento padro de declaraes de atributo no lado esquerdo da
            // transformao (this) com declaraes de atributo no programa fonte
            // (contidas em methodsList).

            Iterator it = this.declarations.iterator();
            while (it.hasNext()) {

                JMethodDeclaration tempMethodDec =
                        (JMethodDeclaration) it.next();

                int j = 0;
                int max2 = methodsList.size();
                boolean matchedOk = false;
                while (j < max2 && !matchedOk) {
                    JMethodDeclaration nodeToBeMatched =
                            (JMethodDeclaration) methodsList.elementAt(j);

                    ResultSet tempRes = (ResultSet) results.clone();
                    // Tenta casar tempMethodDec com qualquer declarao de metodo
                    // presente em methodsList.
                    try {
                        tempMethodDec.match(nodeToBeMatched, tempRes);
                        // Se tudo der certo, nodeToBeMatched  retirado de
                        // methodsList e o resultado desse casamento individual
                        //  adicionado ao resultado do casamento como um todo.
                        results.merge(tempRes);
                        methodsList.removeElementAt(j);
                        matchedOk = true;
                    } catch (NodesNotMatchedException e) {
                    }
                    j++;
                }

                if ((j == max2) && !matchedOk) {
                    throw new NodesNotMatchedException("MethodSets couldn't be matched",
                            tempMethodDec, methodsList);
                }
            }
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.declarations != null) {
            int i = this.declarations.size() - 1;

            // empilha os metodos no sentido oposto
            while (i >= 0) {
                this.declarations.elementAt(i).accept(visitor, data);
                i--;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * duas maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 2 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * <p/>
     * Ns do tipo JMethodDeclarationSet nunca so executveis.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Processamento de cada uma das declaraes de metodo

        int i = 0;
        int max = this.size();

        while (i < max) {

            INode currFD = this.elementAt(i);
            if (currFD != null) {
                Object ret = this.processNode(currFD, data);

                if (currFD.isExecutable()) {
                    if (!(ret instanceof JMethodDeclaration)) {
                        throw new ExecutionException(this);
                    }
                    this.setElementAt((JMethodDeclaration) ret, i);
                }
            }
            i++;
        }
        return retorno;
    }

    public static JMethodDeclaration[]
            toMethodDeclarationArray(JMethodDeclarationSet mds) {
        JMethodDeclaration[] retorno = null;
        if (mds != null) {
            int size = mds.size();
            retorno = new JMethodDeclaration[size];
            for (int i = 0; i < size; i++) {
                retorno[i] = (JMethodDeclaration) mds.elementAt(i);
            }
        }
        return retorno;
    }

} // fim da classe
