/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.EvaluateFunctions;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.visitors.ReplacementVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * Esta classe representa uma invocao de mtodo. Invocaes de mtodo no so
 * produes diretas da gramtica utilizada, entretanto, por serem um tipo de
 * expresso comum e por terem caractersticas especiais, sero tratadas como
 * um n da rvore sinttica. <br>
 * <p/>
 * Ex:
 * <p/>
 * <p><code>conta.creditar(250.00); </code>
 * <p/>
 * <p><code>this.setNumero("10648-8"); </code>
 *
 * @author Fernando Castor
 * @see cin.jats.engine.parser.nodes.JExpression
 * @see cin.jats.engine.parser.nodes.JExpressionList
 */
public class JMethodInvocation extends JExpression implements Commentable {


    // Um nome representando o caminho a partir do qual o mtodo ser invocado
    // e, possivelmente, seu nome.
    private INode referencedObject;

    // O nome do mtodo.
    private JName methodName;

    // Armazena a lista de argumentos que ser fornecida para o mtodo na sua invocao.
    private JExpressionList arguments;

    private JComment comment;


    /**
     * Construtor vazio da classe.
     */
    public JMethodInvocation() {
        this.referencedObject = null;
        this.methodName = null;
        this.arguments = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro uma <code>JExpression</code>,
     * um JName correspondendo ao nome do mtodo e uma lista de argumentos.
     *
     * @param refObject Uma <code>JExpression</code> denotando o objeto a partir
     *                  do qual o mtodo ser invocado.
     * @param mn        Um objeto do tipo <code>JName</code> correspondendo ao nome do
     *                  mtodo mais o caminho a partir do qual ele ser invocado.
     * @param args      Um objeto do tipo <code>JExpressionList</code>, correspondendo
     *                   lista de argumentos que o mtodo recebe.
     */
    public JMethodInvocation(INode refObject, JName mn, JExpressionList args) {
        this.referencedObject = refObject;
        this.methodName = mn;
        this.arguments = args;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro um <code>JName</code>,
     * correspondendo ao nome do mtodo e o caminho da invocao, e lista de
     * argumentos.
     *
     * @param np   Um objeto do tipo <code>JName</code> denotando o nome do mtodo
     *             e o caminho a partir do qual ele ser invocado.
     * @param args Um objeto do tipo <code>JExpressionList</code>, correspondendo
     *              lista de argumentos que o mtodo recebe.
     */
    public JMethodInvocation(JName np, JExpressionList args) {
        this.methodName = np;
        this.arguments = args;
        this.referencedObject = null;
    }


    /**
     * Retorna o objeto a partir do qual o mtodo ser invocado.
     *
     * @return Um <code>INode</code> correspondente ao objeto a partir
     *         do qual o mtodo ser invocado ou  expresso que, uma vez
     *         avaliada, produzir esse objeto como resultado.
     */
    public INode getReferencedObject() {
        return this.referencedObject;
    }

    /**
     * Retorna o nome do mtodo a ser invocado e o caminho a partir do qual essa
     * invocao ocorrer.
     *
     * @return Um <code>JName</code> correspondendo ao nome do mtodo deve ser
     *         invocado junto com o caminho da invocao.
     */
    public JName getMethodName() {
        return this.methodName;
    }

    /**
     * Retorna a lista de argumentos que o mtodo recebe nesta invocao.
     *
     * @return Um objeto do tipo <code>JExpressionList</code>, correspondente 
     *         lista de argumentos que o mtodo recebe.
     */
    public JExpressionList getArgumentList() {
        return this.arguments;
    }


    /**
     * Seta o objeto a partir do qual o mtodo ser invocado.
     *
     * @param refObj Um <code>INode</code> correspondente ao objeto a partir
     *               do qual o mtodo ser invocado. <code>refObj</code> pode
     *               corresponder tambm a uma expresso que ser avaliada. Nesse caso
     *               o mtodo  invocado a partir do objeto resultante dessa avaliao.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>np</code> for <code>null</code>.
     */
    public void setReferencedObject(INode refObj)
            throws IllegalArgumentException {
//        if (refObj == null) {
//            throw new IllegalArgumentException(); // Texto da exceo.
//        }

        this.referencedObject = refObj;
    }

    /**
     * Seta o nome do mtodo e o caminho a partir do qual este ser invocado.
     * O argumento <code>mn</code>  fornecido na forma de uma lista de
     * identificadores separados por ".", onde o ltimo identificador dessa lista
     * corresponde ao nome do mtodo enquanto os outros constituem o caminho.
     *
     * @param mn Um objeto do tipo <code>JName</code> correspondendo ao nome do
     *           mtodo a ser invocado.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>np</code> for <code>null</code>.
     */
    public void setMethodName(JName mn) {
        this.methodName = mn;
    }

    /**
     * Seta a lista de argumentos que o mtodo recebe.
     *
     * @param args Um objeto do tipo <code>JExpressionList</code>, uma lista de expresses
     *             que corresponde  lista de argumentos que o mtodo recebe.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>args</code> for <code>null</code>.
     */
    public void setArgumentList(JExpressionList args) throws IllegalArgumentException {
        if (args == null) {
            throw new IllegalArgumentException(); // Texto da exceo.
        }
        this.arguments = args;
    }

    /**
     * Retorna true, caso o n possua javadoc.
     *
     * @return true, se  o n possui um javadoc
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return this.comment != null;
    }

    /**
     * Retorna o javadoc associado a este n
     *
     * @return javadoc associado a este n
     */
    public JComment getComment() {
        return comment;
    }

    /**
     * Seta o javadoc associado a este n
     *
     * @param comment javadoc associado a este n
     */
    public void setComment(JComment comment) {
        this.comment = comment;
    }

    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        throw new UnsupportedOperationException("not applicable");
    }

    /**
     * Compara dois objetos do tipo <code>JMethodInvocation</code> e diz se
     * estes so idnticos.
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JMethodInvocation) {
            JMethodInvocation methInv = (JMethodInvocation) obj;
            if (this.isASourceNode()) {
                retorno =
                        AbstractNode.equals(this.referencedObject, methInv.getReferencedObject()) &&
                        AbstractNode.equals(this.methodName, methInv.getMethodName()) &&
                        AbstractNode.equals(this.arguments, methInv.getArgumentList());
            } else {
                retorno = compareJaTSNode(methInv);
            }
        }
        return retorno;
    }


    /**
     * Mtodo auxiliar para o processo de substituio.
     */
    public void initReplacement(ReplacementVisitor vis)
            throws IllegalArgumentException, ExecutionException,
            InconsistentNodeException {

        if (this.referencedObject.isVariable()) {
            Object obj = vis.getValueOf(this.referencedObject);
            if (obj instanceof INode) {
                INode node = (INode) obj;
                this.setReferencedObject(node);
            }
        }
        // se o objeto referenciado for uma invocao de mtodo, chamamos o
        // mtodo recursivamente
        else if (this.referencedObject instanceof JMethodInvocation) {
            ((JMethodInvocation) this.referencedObject).initReplacement(vis);
        }
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de seus
     * filhos
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            if (this.methodName == null || this.arguments == null) {
                throw new InconsistentNodeException(this);
            }

            if (this.referencedObject != null) {
                this.referencedObject.accept(visitor, data);
            }

            this.methodName.accept(visitor, data);

            this.arguments.accept(visitor, data);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // e ento visite este n.
        retorno = visitor.visit(this, data);


        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>INode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada quando algum componente
     *                                   essencial do n tiver um valor invlido, tornando o n, como um
     *                                   todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JMethodInvocation)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JMethodInvocation exp = (JMethodInvocation) node;

            if (this.getArgumentList() == null || this.getMethodName() == null) {
                throw new InconsistentNodeException(this);
            }

            if (this.getReferencedObject() != null) {
                (this.getReferencedObject()).match(exp.getReferencedObject(),
                        results);
            } else if (exp.getReferencedObject() != null) {
                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_ARE_INCOMPATIBLE_MSG,
                        this, node);
            }

            this.getMethodName().match(exp.getMethodName(), results);

            this.getArgumentList().match(exp.getArgumentList(), results);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do nome do mtodo a ser invocado.
            INode mName = this.getMethodName();
            if (mName != null) {
                Object ret1 = this.processNode(mName, data);

                if (mName.isExecutable()) {

                    try {
                        this.setMethodName(JName.unwrapName(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento da lista de argumentos passados para o mtodo.
            INode args = this.getArgumentList();
            if (args != null) {
                Object ret2 = this.processNode(args, data);

                if (args.isExecutable()) {
                    if (!(ret2 instanceof JExpressionList)) {
                        throw new ExecutionException(this);
                    }
                    this.setArgumentList((JExpressionList) ret2);
                }
            }

            // Processamento do objeto a partir do qual o mtodo  invocado.
            INode rObj = this.getReferencedObject();
            if (rObj != null) {
                Object ret3 = this.processNode(rObj, data);

                if (rObj.isExecutable()) {

                    if (!(ret3 instanceof INode)) {
                        throw new ExecutionException(this);
                    }
                    this.setReferencedObject((INode) ret3);
                }
            }
        }

        return retorno;
    }


    /**
     * Retorna o resultado da avaliao desta invocao de mtodo.
     *
     * @param data
     * @return
     * @throws ExecutionException
     */
    public Object evaluate(Object data) throws ExecutionException {

        Object result = null;
        result = null;

        Object[] argList = this.arguments.toObjectArray();
        String methodName = this.methodName.toString();

        Object refObj = null;
        if (this.getReferencedObject() != null) {
            refObj = this.referencedObject;
        }
        if (refObj != null) {

            Class classe = refObj.getClass();

            if (refObj instanceof JExpression) {
                refObj = ((JExpression) refObj).evaluate(data);
                classe = refObj.getClass();
            }

            Class[] argTypes = new Class[argList.length];
            int i = 0;
            while (i < argTypes.length) {

                if (argList[i] instanceof JExpression) {
                    argList[i] = ((JExpression) argList[i]).evaluate(data); //botei pra avaliar!
                }

                argTypes[i] = (argList[i]).getClass();
                i++;
            }
            Method methodToBeCalled = getMethodToBeCalled(classe, argTypes, methodName);
            if (methodToBeCalled != null) {
                try {

                    result = methodToBeCalled.invoke(refObj, argList);
                } catch (IllegalAccessException e) {
                    throw new ExecutionException("IllegalAccessException " +
                            e.getMessage());
                } catch (IllegalArgumentException e) {
                    throw new ExecutionException("IllegalArgumentException " +
                            e.getMessage());
                } catch (InvocationTargetException e) {
                    e.getTargetException().printStackTrace();
                    throw new ExecutionException(classe.toString() + '.' + methodName + ": InvocationTargetException " +
                            e.getMessage());
                }
            } else {
                throw new ExecutionException("Mtodo nao encontrado: " +
                        this + " em " + refObj.getClass().getName(), this);
            }
        }
        return result;
    }

    private Method getMethodToBeCalled(Class classe, Class[] argTypes,
                                       String methodName) {
        Method[] methods = classe.getMethods();
        Method methodToBeCalled = null;
        int j = 0;
        boolean foundIt = false;
        while (j < methods.length && !foundIt) {

            if ((methods[j].getParameterTypes()).length ==
                    argTypes.length &&
                    methodName.equals(methods[j].getName())) {
                Class[] parTypes = methods[j].getParameterTypes();
                int k = 0;
                boolean ok = false;
                while (k < argTypes.length) {
                    ok = EvaluateFunctions.typesAreCompatible(parTypes[k], argTypes[k]);
                    if (!ok) {
                        break;
                    }
                    k++;
                }

                if (argTypes.length == 0 /*&& parTypes.length == 0*/) {
                    ok = true;
                }

                if (ok) {
                    foundIt = true;
                    methodToBeCalled = methods[j];
                }
            }
            if (!foundIt) {
                j++;
            }
        }
        // se terminou o lao e no encontrou, retorna null e loga essa situao
        if (!foundIt) {
//            try {
//                QualitiLogger.debug("<ATTENTION> Metodo nao encontrado " +
//                        methodName);
//            } catch (ExceptionInInitializerError ex) {
                System.out.println("<ATTENTION> Metodo nao encontrado " +
                        methodName);
            //}
        }
        return methodToBeCalled;
    }


}
