/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa uma lista de modificadores
 * <p/>
 * Historico de revises
 * 10/06/2004 - Alterada para suportar os modificadores de C#
 * 15/06/2004 - Relaxadas as condies de modificadores possveis
 * 
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 */
public class JModifierList extends AbstractNode {

    //modificadores atuais
    private int modifiers;

    // modificadores possiveis para este tipo de no
    private int possibleModifiers;

    // indica que a declaraao nao possui modificador
    public static final int NO_MODIFIERS = 0;

    // modificadores permitidos a uma declaraao de tipo
    public static final int TYPE_DECLARATION = JModifier.ABSTRACT +
            JModifier.PUBLIC + JModifier.STRICT +
            /* C#*/
            JModifier.INTERNAL + JModifier.PRIVATE +
            JModifier.PROTECTED
            ;

    // modificadores possiveis para declaracoes de classe
    public static final int CLASS_DECLARATION = JModifier.FINAL +
            TYPE_DECLARATION +
            /* C# */
            JModifier.SEALED + JModifier.NEW;

    // modificadores possiveis para declaracoes de classe
    public static final int NESTED_TYPE_DECLARATION = JModifier.PROTECTED +
            JModifier.PRIVATE + CLASS_DECLARATION;

    // modificadores possiveis para declaracoes de interface
    public static final int INTERFACE_DECLARATION = JModifier.ABSTRACT +
            JModifier.PUBLIC + JModifier.STRICT;

    // modificadores possiveis para declaracoes de metodo
    public static final int METHOD_DECLARATION = JModifier.ABSTRACT +
            JModifier.FINAL + JModifier.PRIVATE +
            JModifier.PROTECTED + JModifier.PUBLIC +
            JModifier.STATIC + JModifier.STRICT +
            JModifier.NATIVE + JModifier.SYNCHRONIZED +
            /* C# */
            JModifier.NEW + JModifier.INTERNAL + JModifier.SEALED +
            JModifier.OVERRIDE + JModifier.EXTERN;

    // modificadores possiveis para declaracoes de atributos
    public static final int FIELD_DECLARATION = JModifier.FINAL +
            JModifier.VOLATILE + JModifier.PRIVATE +
            JModifier.PROTECTED + JModifier.PUBLIC +
            JModifier.STATIC + JModifier.TRANSIENT;

    // modificadores possiveis para declaracoes de construtor
    public static final int CONSTRUCTOR_DECLARATION = JModifier.PRIVATE +
            JModifier.PROTECTED + JModifier.PUBLIC;

    // modificador possivel para parametros formais
    public static final int PARAMETER_DECLARATION = JModifier.FINAL +
            /*C#*/
            JModifier.REF + JModifier.OUT + JModifier.PARAMS;

    // modificador possvel para declarao de variveis locais
    public static final int LOCAL_VARIABLE_DECLARATION = JModifier.FINAL;

    // C#
    // modificador possvel para declarao de constantes
    public static final int CONSTANT_DECLARATION = JModifier.CONST +
            JModifier.NEW + JModifier.PUBLIC + JModifier.PRIVATE +
            JModifier.PROTECTED + JModifier.INTERNAL;
    
    // modificador possvel para declarao de propriedades
    public static final int PROPERTY_DECLARATION =	JModifier.NEW
												+	JModifier.PUBLIC
												+	JModifier.PROTECTED
												+ 	JModifier.INTERNAL
												+	JModifier.PRIVATE
												+	JModifier.STATIC
												+	JModifier.VIRTUAL
												+	JModifier.SEALED
												+	JModifier.OVERRIDE
												+	JModifier.ABSTRACT
												+	JModifier.EXTERN;
    
    // modificadores possveis para declarao de enumerao
    public static final int ENUM_DECLARATION = 	JModifier.NEW
											+	JModifier.PUBLIC
											+	JModifier.PROTECTED
											+	JModifier.INTERNAL
											+	JModifier.PRIVATE;
    
    public static final int INDEXER_DECLARATION = PROPERTY_DECLARATION;


    /**
     * @todo Incluir os demais modificadores
     */
    // s usado em casos especiais, mais precisamente, na produo Modifier,
    // l do parser de JaTS
    public static final int ALL_MODIFIERS = JModifier.ABSTRACT + JModifier.FINAL +
            JModifier.PRIVATE + JModifier.PROTECTED + JModifier.PUBLIC +
            JModifier.STATIC + JModifier.STRICT + JModifier.NATIVE +
            JModifier.TRANSIENT + JModifier.VOLATILE + JModifier.SYNCHRONIZED;

    /**
     * Construtor default, inicializa tanto os modificadores atuais quanto
     * os possiveis para este tipo de n com zero, isto , esta lista no
     * aceita inicialmente nenhum tipo de modificador.
     */
    public JModifierList() {
        this.modifiers = NO_MODIFIERS;
        this.possibleModifiers = ALL_MODIFIERS;
        this.setTypeIdentifier(JaTSNode.MODIFIER_LIST);
    }


    /**
     * Construtor que inicializa apenas os modificadores possveis para este
     * n, sem receber nenhum valor inicial.
     *
     * @param pMods modificadores possiveis
     */
    public JModifierList(int pMods) {
        this.modifiers = NO_MODIFIERS;
        this.possibleModifiers = pMods;
        this.setTypeIdentifier(JaTSNode.MODIFIER_LIST);
    }


    /**
     * Construtor que recebe o(s) modificador(es) atuais e os possveis para
     * para este no.
     *
     * @param mods  modificadores atuais
     * @param pMods modificadores possiveis
     */
    public JModifierList(int mods, int pMods) {
        this.modifiers = mods;
        this.setTypeIdentifier(JaTSNode.MODIFIER_LIST);

        if (pMods == FIELD_DECLARATION ||
                pMods == CONSTRUCTOR_DECLARATION ||
                pMods == CLASS_DECLARATION ||
                pMods == NESTED_TYPE_DECLARATION ||
                pMods == METHOD_DECLARATION ||
                pMods == INTERFACE_DECLARATION ||
                pMods == PARAMETER_DECLARATION ||
                pMods == LOCAL_VARIABLE_DECLARATION ||
                pMods == TYPE_DECLARATION ||
                pMods == ALL_MODIFIERS) {
            this.possibleModifiers = pMods;
            this.modifiers = mods;
        } else {
            this.possibleModifiers = NO_MODIFIERS;
            this.modifiers = NO_MODIFIERS;
        }
    }


    /**
     * Adiciona um modificador  lista de modificadores, se este for vlido para
     * o tipo de n
     *
     * @param mod Modificador que se deseja acrescentar
     * @throws IllegalArgumentException - Se o modificador passado for
     *                                  invlido para este n, se este modificador  j est na lista
     *                                  de modificadores ou se este for um modificador de acesso e
     *                                  j houver um.
     */
    public void addModifier(int mod) throws IllegalArgumentException {

        if (mod < 0 || /*((mod != NO_MODIFIERS) && ((mod & possibleModifiers) == 0)) ||*/
                ((mod & this.modifiers) > 0) ||
                ((isAccessModifier(mod) && this.hasAccessModifier()))) {

            throw new IllegalArgumentException("Modificador invalido");
        }
        this.modifiers += mod;
    }


    /**
     * Adiciona um modificador  lista de modificadores, se este for vlido para
     * o tipo de n
     *
     * @param strMod String que representa o modificador que se deseja adicionar
     * @throws IllegalArgumentException - Se o modificador passado for
     *                                  invlido para este n, se este modificador  j est na lista
     *                                  de modificadores ou se este for um modificador de acesso e
     *                                  j houver um.
     */
    public void addModifier(String strMod) throws IllegalArgumentException {

        int mod = toInt(strMod);
        this.addModifier(mod);
    }


    /**
     * Adiciona os modificadores da lista de modifiecadores passados como
     * parmetro  lista de modificadores atuais
     *
     * @param modList Lista de modificadores, cujos modificadores sero
     *                adicionados  lista atual.
     * @throws IllegalArgumentException - Se algum dos modificadores da lista
     *                                  passada  j estiver na lista atual, se a essa for <code>null</code>
     *                                  ou se a lista no for compatvel com a lista atual.
     */
    public void addModifierList(JModifierList modList) {

        if ((modList != null) &&
                (this.isCompatible(modList.getPossibleModifiers())) &&
                ((modList.getModifiers() & this.getModifiers()) == 0)) {
            this.modifiers += modList.getModifiers();
        } else {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Retorna um inteiro correspondendo aos modificadores presentes nesta
     * lista.
     *
     * @return Um inteiro correspondendo  lista de modificadores.
     */
    public int getModifiers() {

        return this.modifiers;
    }


    /**
     * Retorna um inteiro correspondendo aos modificadores que podem estar
     * presentes nesta lista de modificadores. Esse valor vai depender do
     * tipo de lista de modificadores que esta representa. Ela pode pode
     * corresponder  lista de modificadores de um mtodo, atributo, classe,
     * etc.
     *
     * @return Um inteiro correspondendo  lista de modificadores que podem
     *         estar nesta lista.
     */
    public int getPossibleModifiers() {

        return this.possibleModifiers;
    }


    /**
     * Verifica se a lista de modificadores possui um modificador de acesso
     *
     * @return true, se em this.modifiers houver um modificador de acesso
     *         false, caso contrrio
     */
    public boolean hasAccessModifier() throws IllegalArgumentException {

        return (((this.modifiers & JModifier.PUBLIC) > 0) ||
                ((this.modifiers & JModifier.PRIVATE) > 0) ||
                ((this.modifiers & JModifier.PROTECTED) > 0));

    }

    /**
     * Verifica se o modificador passado como parmetro pertence  lista de
     * modificadores daquele.
     *
     * @param mod O modificador que se deseja procurar na lista
     * @return true, se o modificador passado como parmetro estiver na lista
     *         false, caso contrrio
     * @ exception IllegalArgumentException - Se o modificador passado
     * for invalido
     */
    public boolean hasModifier(int mod) throws IllegalArgumentException {

        if (mod < 0) {
            throw new IllegalArgumentException("Modificador invalido");
        }
        return ((mod & this.modifiers) > 0);
    }


    /**
     * Diz se esta lista esta vazia. Ou seja, se nao ha modificadores.
     *
     * @return true, se a lista estiver vazia
     *         false, caso contrrio
     */
    public boolean hasNoModifiers() throws IllegalArgumentException {
        return (this.modifiers == NO_MODIFIERS);
    }


    /**
     * Verifica se o modificador passado como parmetro pertence  lista de
     * modificadores daquele.
     *
     * @param strMod O modificador que se deseja procurar na lista
     * @return true, se o modificador passado como parmetro estiver na lista
     *         false, caso contrrio
     * @ exception IllegalArgumentException - Se o modificador passado
     * for invalido
     */
    public boolean hasModifier(String strMod) throws IllegalArgumentException {

        if (Util.checkString(strMod)) {
            int mod = toInt(strMod);
            return hasModifier(mod);
        } else {
            throw new IllegalArgumentException("Argumento Invlido");
        }

    }


    /**
     * Verifica se o modificador passado  um modificador de acesso
     *
     * @return true, se em this.modifiers houver um modificador de acesso
     *         false, caso contrrio
     */
    public static boolean isAccessModifier(int mod) throws IllegalArgumentException {

        return ((mod == JModifier.PUBLIC) || (mod == JModifier.PRIVATE) ||
                (mod == JModifier.PROTECTED));

    }

    /**
     * Remove todos os modificadores da lista
     */
    public void removeAllModifiers() {
        this.modifiers = NO_MODIFIERS;
    }


    /**
     * Remove um modificador da lista.
     *
     * @param mod Modificador a ser removido
     * @throws IllegalArgumentException - Se o modificador passado nao
     *                                  pertencer a lista de modificadores
     */
    public void removeModifier(int mod) throws IllegalArgumentException {

        if (mod < 0 || ((mod & possibleModifiers) == 0) ||
                ((mod & this.modifiers) == 0)) {
            throw new IllegalArgumentException();
        }
        this.modifiers -= mod;
    }


    /**
     * Remove um modificador da lista.
     *
     * @param strMod Modificador a ser removido
     * @throws IllegalArgumentException - Se o modificador passado nao
     *                                  pertencer a lista de modificadores
     */
    public void removeModifier(String strMod) throws IllegalArgumentException {

        int mod = toInt(strMod);

        this.removeModifier(mod);

    }


    /**
     * Seta os possiveis modificadores para este tipo de n. A constante deve
     * ser uma das seguintes:
     *
     * @param mods Lista de modificadores
     * @throws java.lang.IllegalArgumentException
     *          - se mods no for vlido.
     */
    public void setModifiers(int mods) {
        // CORRIGIR ESTE METODO
        /*
          // codigo correto:
              if  (this.isCompatible(mods)) {
                 this.modifiers = mods;
              }
              else {
                throw new IllegalArgumentException("Modificadores invalidos");
              }
        */

        this.modifiers = mods;
    }


    /**
     * Seta os possiveis modificadores para este tipo de n. A constante deve
     * ser uma das seguintes:
     * <p/>
     * JModifierList.CLASS_DECLARATION
     * JModifierList.INTERFACE_DECLARATION
     * JModifierList.METHOD_DECLARATION
     * JModifierList.FIELD_DECLARATION
     * JModifierList.CONSTRUCTOR_DECLARATION
     * JModifierList.PROPERTY_DECLARATION
     * JModifierList.INDEXER_DECLARATION
     * JModifierList.PARAMETER_DECLARATION
     * JModifierList.LOCAL_VARIABLE_DECLARATION
     * JModifierList.ALL_MODIFIERS
     *
     * @param pMods Modificadores possiveis para este tipo de n
     * @throws java.lang.IllegalArgumentException
     *          - se pMods no for vlido.
     */
    public void setPossibleModifiers(int pMods)
            throws IllegalArgumentException {

        if ((pMods == CLASS_DECLARATION) ||
                (pMods == NESTED_TYPE_DECLARATION) ||
                (pMods == INTERFACE_DECLARATION) ||
                (pMods == METHOD_DECLARATION) ||
                (pMods == FIELD_DECLARATION) ||
                (pMods == CONSTRUCTOR_DECLARATION) ||
                (pMods == PROPERTY_DECLARATION) || // serve para indexer e property
                (pMods == PARAMETER_DECLARATION) ||
                (pMods == LOCAL_VARIABLE_DECLARATION) ||
				(pMods == ENUM_DECLARATION) ||
                (pMods == ALL_MODIFIERS)) {
            this.possibleModifiers = pMods;
        } else {
            throw new IllegalArgumentException("Numero de modificadores " +
                    "possiveis invalido");
        }

    }


    /**
     * Converte a string que representa o modificador eu um inteiro
     * correspondente ao cdigo dele.
     *
     * @param mod Nome do modificador
     * @return Codigo do identificador
     * @throws IllegalArgumentException Quando o argumento passado for nulo
     *                                  ou uma string vazia
     */
    private static int toInt(String mod) {

        int retorno = 0;
        if (Util.checkString(mod)) {
            mod = mod.trim();
            if (mod.startsWith("\"") && mod.endsWith("\"")) {
                mod = mod.substring(1, mod.length() - 1);
            }
            if (mod.equalsIgnoreCase("public")) {
                retorno = JModifier.PUBLIC;
            } else if (mod.equalsIgnoreCase("private")) {
                retorno = JModifier.PRIVATE;
            } else if (mod.equalsIgnoreCase("protected")) {
                retorno = JModifier.PROTECTED;
            } else if (mod.equalsIgnoreCase("abstract")) {
                retorno = JModifier.ABSTRACT;
            } else if (mod.equalsIgnoreCase("static")) {
                retorno = JModifier.STATIC;
            } else if (mod.equalsIgnoreCase("final")) {
                retorno = JModifier.FINAL;
            } else if (mod.equalsIgnoreCase("synchronized")) {
                retorno = JModifier.SYNCHRONIZED;
            } else if (mod.equalsIgnoreCase("native")) {
                retorno = JModifier.NATIVE;
            } else if (mod.equalsIgnoreCase("transient")) {
                retorno = JModifier.TRANSIENT;
            } else if (mod.equalsIgnoreCase("volatile")) {
                retorno = JModifier.VOLATILE;
            } else if (mod.equalsIgnoreCase("strictfp")) {
                retorno = JModifier.STRICT;
            }

        } else {
            throw new IllegalArgumentException("Argumento nulo em " +
                    "cin.jats.engine.parser.nodes.JModifierList:" +
                    " hasModifier(String mod)");
        }

        return retorno;

    }


    /**
     * Retorna a representacao textual da lista de modificadores.
     *
     * @return Uma string contendo a representacao dos modificadores
     */
    public String toString() {

        int mod = this.modifiers;
        String retorno = "";

        if (this.isASourceNode()) {
            if ((mod & JModifier.PRIVATE) > 0) {
                mod = mod - JModifier.PRIVATE;
                retorno = retorno + "private ";

            }
            if ((mod & JModifier.PROTECTED) > 0) {
                mod = mod - JModifier.PROTECTED;
                retorno = retorno + "protected ";

            }
            if ((mod & JModifier.PUBLIC) > 0) {
                mod = mod - JModifier.PUBLIC;
                retorno = retorno + "public ";

            }
            if ((mod & JModifier.ABSTRACT) > 0) {
                mod = mod - JModifier.ABSTRACT;
                retorno = retorno + "abstract ";

            }
            if ((mod & JModifier.STATIC) > 0) {
                mod = mod - JModifier.STATIC;
                retorno = retorno + "static ";

            }
            if ((mod & JModifier.FINAL) > 0) {
                mod = mod - JModifier.FINAL;
                retorno = retorno + "final ";

            }
            if ((mod & JModifier.SYNCHRONIZED) > 0) {
                mod = mod - JModifier.SYNCHRONIZED;
                retorno = retorno + "synchronized ";

            }
            if ((mod & JModifier.NATIVE) > 0) {
                mod = mod - JModifier.NATIVE;
                retorno = retorno + "native ";

            }
            if ((mod & JModifier.TRANSIENT) > 0) {
                mod = mod - JModifier.TRANSIENT;
                retorno = retorno + "transient ";

            }
            if ((mod & JModifier.VOLATILE) > 0) {
                mod = mod - JModifier.VOLATILE;
                retorno = retorno + "volatile ";

            }
            if ((mod & JModifier.STRICT) > 0) {
                mod = mod - JModifier.STRICT;
                retorno = retorno + "strictfp ";

            }

            /**   Modificadores de C#  **/
            if ((mod & JModifier.NEW) > 0) {
                mod = mod - JModifier.NEW;
                retorno = retorno + "new ";
            }
            if ((mod & JModifier.EXTERN) > 0) {
                mod = mod - JModifier.EXTERN;
                retorno = retorno + "extern ";
            }
            if ((mod & JModifier.OVERRIDE) > 0) {
                mod = mod - JModifier.OVERRIDE;
                retorno = retorno + "override ";
            }
            if ((mod & JModifier.REF) > 0) {
                mod = mod - JModifier.REF;
                retorno = retorno + "ref ";
            }
            if ((mod & JModifier.OUT) > 0) {
                mod = mod - JModifier.OUT;
                retorno = retorno + "out ";
            }
            if ((mod & JModifier.PARAMS) > 0) {
                mod = mod - JModifier.PARAMS;
                retorno = retorno + "params ";
            }
            if ((mod & JModifier.CONST) > 0) {
                mod = mod - JModifier.CONST;
                retorno = retorno + "const ";
            }
            if ((mod & JModifier.INTERNAL) > 0) {
                mod = mod - JModifier.INTERNAL;
                retorno = retorno + "internal ";
            }
            if ((mod & JModifier.VIRTUAL) > 0) {
                mod = mod - JModifier.VIRTUAL;
                retorno = retorno + "virtual ";
            }
            if ((mod & JModifier.SEALED) > 0) {
                mod = mod - JModifier.SEALED;
                retorno = retorno + "sealed ";
            }


        } else {
            retorno = this.printVariable();
        }

        return retorno;
    }


    /**
     * Diz se a lista de modificadores  atual (this)   compatvel com a
     * lista de modificadores possveis da lista passada como parametro
     *
     * @param mod Inteiro correspodente a uma lista de modificadores, com a
     *            qual se deseja testar a  compatibilidade
     */
    public boolean isCompatible(int mod) {

        return ((this.getModifiers() & mod) > 0 || this.getModifiers() == 0);
    }

    /**
     * Compara dois objetos do tipo JModifierList e diz se estes so iguais
     *
     * @param obj Lista de modificadores a ser comparada com a lista
     *            representada por this
     * @return true, se as listas forem iguais <br>
     *         false, em caso contrrio
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JModifierList) {
            JModifierList modList = (JModifierList) obj;
            if (!this.isExecutable() && !this.isVariable()) {
                retorno = (this.modifiers == modList.modifiers) &&
                        (this.possibleModifiers == modList.possibleModifiers);
            } else {
                retorno = compareJaTSNode(modList);
            }
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code> e aplic-lo ao n
     * atual.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        Object retorno = visitor.visit(this, data);


        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JModifierList)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results,
                JaTSNode.MODIFIER_LIST)) {

            JModifierList modList = (JModifierList) node;

            if (!(this.getModifiers() == modList.getModifiers()) ||
                    !(this.isCompatible(modList.getPossibleModifiers()))) {
                throw new NodesNotMatchedException("Incompatible modifier lists", this, node);
            }
        } else if (this.isVariable() &&
                this.getVariableType() != JaTSNode.MODIFIER_LIST) {
            throw new InconsistentNodeException(node);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento de listas de
     * modificadores ocorre se apenas se o n atual for executvel. Nesse caso,
     * o valor retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu mtodo
     * process e corresponde ao resultado de seu processamento.
     * <p/>
     * S h essa modalidade de processamento para listas de modificadores porque
     * elas nunca tm ns filhos. O mesmo  verdade para ns dos tipos
     * <code>JLiteral</code> e <code>JIdentifier</code>.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        return retorno;
    }
}
