/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.StringTokenizer;

/**
 * Esta classe implementa nomes em JaTS. Um nome em JaTS, assim como em Java,
 * corresponde a uma lista de identificadores.
 *
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 * 
 * @see JIdentifier
 * @see JaTSNode
 * 
 * @version 14/10/2005
 */
public class JName extends JExpression implements Signature, Cloneable {

    /**
     * Lista contendo os identificadores que fazem parte deste nome.
     */
    private NodeList identifiers;


    /**
     * binding da declaracao deste nome (ponteiro pra o no da declaracao)
     */
    private INode declarationRef;
    
    /**
     * Construtor vazio da classe
     */
    public JName() {
        super();
        this.identifiers = new NodeList();
        this.setTypeIdentifier(JaTSNode.NAME);
    }


    /**
     * Construtor da classe que recebe como parmetro um String.
     *
     * @param theName Um String correspondente a uma lista de identificadores.
     */
    public JName(String theName) {
        super();
        generateName(theName);
        this.setTypeIdentifier(JaTSNode.NAME);
    }

    /**
     * Construtor da classe que recebe como parmetro um String.
     *
     * @param theName Um JIdentifier correspondente a este name
     */
    public JName(JIdentifier theName) {
        this();
        this.addIdentifier(theName);
        this.setTypeIdentifier(JaTSNode.NAME);
    }


    /**
     * Construtor da classe que recebe como parmetro um String.
     *
     * @param theName Um objeto <code>JType</code> correspondente a este name.
     */
    public JName(JType theName) {
        this();
        this.identifiers = theName.getTypeName().identifiers;
        this.setTypeIdentifier(JaTSNode.NAME);
    }

    /**
     * Construtor que recebe a lista de identificadores que faro parte deste
     * nome
     *
     * @param identifiers Lista de identificadores correspondentes a este name.
     */
    public JName(NodeList identifiers) {
        this();
        this.setIdentifiers(identifiers);
    }


    /**
     * Gera o Name a partir de uma String
     *
     * @param str string a ser convertida em name
     */

    private void generateName(String str) {
        if (str != null) {
            this.identifiers = new NodeList();
            StringTokenizer stok = new StringTokenizer(str, ".");
            while (stok.hasMoreElements()) {
                String tmpId = stok.nextToken();
                this.addIdentifier(tmpId);
            }
        } else {
            throw new IllegalArgumentException("");
        }
    }

    /**
     * Retorna os identificadores que compoem este nome.
     *
     * @return Identificadores que compoem este nome.
     */
    public NodeList getIdentifiers() {
        return identifiers;
    }

    /**
     * Seta os identificadores que compoem o nome.
     *
     * @param identifiers Identificadores que compoem o nome.
     */
    protected void setIdentifiers(NodeList identifiers) {
        if (identifiers != null) {
            JaTSIterator jit = identifiers.getIterator();
            while (jit.hasNext()) {
                if (!(jit.next() instanceof JIdentifier)) {
                    throw new IllegalArgumentException("element is not a Identifier");
                }
            }
            this.identifiers = identifiers;

        }
    }

    /**
     * Adiciona um novo identificador a este nome.
     *
     * @param id O identificador a ser adicionado.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>id</code> for <code>null</code>.
     */
    public void addIdentifier(JIdentifier id) throws IllegalArgumentException {
        if (id == null) {
            throw new IllegalArgumentException();
        }
        JIdentifier newId = JIdentifier.unwrapIdentifier(id);
        this.identifiers.addElement(newId);

    }


    /**
     * Adiciona um novo identificador a este nome.
     *
     * @param id Um objeto do tipo <code>JType</code> correspondente ao
     *           identificador a ser adicionado. id deve ter dimenso 0.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>id</code> for <code>null</code>.
     */
    public void addIdentifier(JType id) throws IllegalArgumentException {
        if (id == null) {
            throw new IllegalArgumentException();
        }
        JIdentifier newId = JIdentifier.unwrapIdentifier(id);
        this.identifiers.addElement(newId);

    }

    /**
     * Adiciona um novo identificador a este nome, em forma de String
     *
     * @param id O identificador a ser adicionado.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>id</code> for <code>null</code>.
     */
    public void addIdentifier(String id) throws IllegalArgumentException {
        if (id == null) {
            throw new IllegalArgumentException();
        }
        JIdentifier newId = JIdentifier.unwrapIdentifier(id);
        this.identifiers.addElement(newId);
    }


    /**
     * Adiciona um novo identificador a este nome na posio desejada.
     *
     * @param id  O identificador a ser adicionado.
     * @param pos O ndice onde o fornecido deve ser colocado, na lista de
     *            identificadores deste n. O ndice inicial  zero.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>id</code> for <code>null</code> ou o argumento <code>pos</code>
     *                                  for um inteiro negativo.
     */
    public void setIdentifierAt(JIdentifier id, int pos)
            throws IllegalArgumentException {

        if (id == null || pos < 0 || pos > this.size()) {
            throw new IllegalArgumentException();
        }
        this.identifiers.setElementAt(id, pos);
    }

    /**
     * Retorna o name com o prefixo adicionado.
     *
     * @param prefix Prefixo a ser adicionado ao name
     * @return JName acrescido do prefixo
     */
    public JName addPrefix(String prefix) {
        JName retorno = null;
        if (Util.checkString(prefix) && this.isASourceNode()) {
            // pega a String correspondente ao name
            String tmpName = this.toString();
            char firstChar = tmpName.charAt(0);
            String newName = prefix + firstChar + tmpName.substring(1);
            try {
                retorno = JName.unwrapName(newName);
            } catch (InconsistentNodeException e) {
                throw new IllegalArgumentException("Argumento invlido para concatenacao");
            }
        } else {
            throw new IllegalArgumentException("Argumento invlido para concatenacao");
        }
        return retorno;
    }

    /**
     * Retorna o name com o prefixo adicionado.
     *
     * @param prefix Prefixo a ser adicionado ao name
     * @return JName acrescido do prefixo
     */
    public JName addPrefix(JIdentifier prefix) {
        JName retorno = null;
        if (prefix != null && this.isASourceNode()) {
            String idValue = prefix.getValue();
            retorno = this.addPrefix(idValue);
        }
        return retorno;
    }

    /**
     * Retorna o name com o prefixo adicionado.
     *
     * @param prefix Prefixo a ser adicionado ao name
     * @return JName acrescido do prefixo
     */
    public JName addPrefix(JName prefix) {
        JName retorno = null;
        if (prefix != null && this.isASourceNode()) {
            String prefixValue = prefix.toString();
            retorno = this.addPrefix(prefixValue);
        }
        return retorno;
    }

    /**
     * Retorna o name com o sufixo adicionado.
     *
     * @param suffix Sufixo a ser adicionado ao name
     * @return JName acrescido do sufixo
     */
    public JName addSuffix(String suffix) {
        JName retorno = null;
        if (Util.checkString(suffix) && this.isASourceNode()) {
            // pega a String correspondente ao name
            String tmpName = this.toString();
            tmpName = tmpName.concat(suffix);
            try {
                retorno = JName.unwrapName(tmpName);
            } catch (InconsistentNodeException e) {
                throw new IllegalArgumentException("Argumento invlido para concatenacao");
            }
        } else {
            throw new IllegalArgumentException("Argumento invlido para concatenacao");
        }
        return retorno;
    }

    /**
     * Retorna o name com o sufixo adicionado.
     *
     * @param suffix Sufixo a ser adicionado ao name
     * @return JName acrescido do sufixo
     */
    public JName addSuffix(JIdentifier suffix) {
        JName retorno = null;
        if (suffix != null && this.isASourceNode()) {
            String idValue = suffix.getValue();
            retorno = this.addSuffix(idValue);
        }
        return retorno;

    }

    /**
     * Retorna o name com o sufixo adicionado.
     *
     * @param suffix Sufixo a ser adicionado ao name
     * @return JName acrescido do sufixo
     */
    public JName addSuffix(JName suffix) {
        JName retorno = null;
        if (suffix != null && this.isASourceNode()) {
            String prefixValue = suffix.toString();
            retorno = this.addSuffix(prefixValue);
        }
        return retorno;
    }


    /**
     * Retorna o name com o sufixo adicionado.
     *
     * @param suffix Sufixo a ser adicionado ao name
     * @return JName acrescido do sufixo
     */
    public JName concat(JName suffix) {
        JName retorno = null;

        if (suffix != null && this.isASourceNode()) {
            String tmpName = this.toString();
            if (suffix.isASourceNode()) {
                String tmpSuffix = suffix.toString();
                try {
                    String newName;
                    if (!tmpName.endsWith(".") || !tmpSuffix.startsWith(".")) {
                        newName = tmpName + "." + tmpSuffix;
                    } else {
                        newName = tmpName + tmpSuffix;
                    }
                    retorno = JName.unwrapName(newName);
                } catch (cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException e) {
                    throw new IllegalArgumentException("Argumento invlido para" +
                            " concatenacao: erro de unwrapping");
                }
            }
        } else {
            throw new IllegalArgumentException("Argumento invlido para" +
                    " concatenacao: argumento nulo");
        }
        return retorno;
    }

    /**
     * Retorna um objeto do tipo <code>NodeList</code> contendo os
     * identificadores que compoem o name. Caso este seja um objeto executvel
     * ou varivel, retorna uma NodeList, contendo o proprio objeto como
     * sendo um dos elementos.
     *
     * @return NodeList contendo os identificadores deste name.
     */
    public NodeList toNodeList() {
        NodeList retorno = new NodeList();
        if (this.isASourceNode()) {
            retorno = this.identifiers;
        } else {
            retorno.addElement(this);
        }
        return retorno;
    }


    /**
     * Retorna um  objeto do tipo JFieldAccess, representado pelo Name passado
     * como parametro. Por exemplo: <br>
     * <p/>
     * Dado o Name <code>name<code>:
     * <blockquote>
     * id1.id2.id3.id4.id5
     * </blockquote>
     * <p/>
     * A chamada:
     * <blockquote>
     * JName.toFieldAccess(name);
     * </blockquote>
     * <p/>
     * Retorna a lista:
     * <blockquote>
     * ((((id1).id2).id3).id4).id5
     * </blockquote>
     *
     * @param refObj Nome que ser convertido em um acesso a atributo
     * @return O nome convertido em um acesso a atributo
     */
    public static JFieldAccess toFieldAccess(JName refObj) {

        if (refObj != null && refObj.isASourceNode() && refObj.size() > 1) {
            if (refObj.size() == 2) {
                return new JFieldAccess(refObj.getIdentifierAt(0), refObj.getIdentifierAt(1));
            } else {
                int lastElementIndex = refObj.size() - 1;
                return new JFieldAccess(toFieldAccess(refObj.subSequence(0, lastElementIndex)),
                        refObj.getIdentifierAt(lastElementIndex));
            }
        } else {
            throw new IllegalArgumentException("Invalid Conversion");
        }
    }


    /**
     * Retorna um  objeto do tipo JFieldAccess, representado por
     * <code>this</code>. Por exemplo: <br>
     * <p/>
     * Dado o Name <code>name<code>:
     * <blockquote>
     * id1.id2.id3.id4.id5
     * </blockquote>
     * <p/>
     * A chamada:
     * <blockquote>
     * JName.toFieldAccess(name);
     * </blockquote>
     * <p/>
     * Retorna a lista:
     * <blockquote>
     * ((((id1).id2).id3).id4).id5
     * </blockquote>
     *
     * @return O nome convertido em um acesso a atributo
     */
    public JFieldAccess toFieldAccess() {
        return JName.toFieldAccess(this);
    }


    /**
     * Retorna um Name contendo apenas os identificadores que se encontram no
     * range especificado, sendo o  primeiro indice inclusivo e o segundo,
     * exclusivo. <br>
     * Assim, o intervalo : [ beginIndex, endIndex) <br>
     * <p/>
     * Dado o Name  <code>name</code> contendo os identificadores:
     * <p/>
     * <blockquote>
     * id1.id2.id3.id4.id5.id6
     * </blockquote>
     * <p/>
     * A chamada:
     * <blockquote>
     * name.subSequence(1,3);
     * </blockquote>
     * <p/>
     * Retorna a lista:
     * <blockquote>
     * id2.id3
     * </blockquote>
     *
     * @param beginIndex Indice inicial
     * @param endIndex   Indice final
     * @return JName representanto a subsequencia especificada pelos indices
     */
    public JName subSequence(int beginIndex, int endIndex) {
        return new JName(this.identifiers.subSequence(beginIndex, endIndex));
    }


    /**
     * @param name       Nome a partir do qual sera extraido o subsequencia.
     * @param beginIndex Indice inicial
     * @param endIndex   Indice final
     * @return JName representanto a subsequencia especificada pelos indices
     * @see JName#subSequence
     */
    public static JName subSequence(JName name, int beginIndex, int endIndex) {
        if (name != null) {
            return new JName(name.identifiers.subSequence(beginIndex, endIndex));
        } else {
            throw new IllegalArgumentException("Name is null");
        }


    }

    /**
     * Reverte o name passado como parmetro. Isto , inverte a ordem de seus
     * identificadores. Ou seja, para a entrada: java.io.Util o retorno ser
     * Util.io.java
     *
     * @param name Objeto do tipo <code>JName</code> a ser revertido
     * @return Name com os identificadores revertidos.
     */
    public static JName reverse(JName name) {
        NodeList ids = name.getIdentifiers().reverse();
        return new JName(ids);
    }


    /**
     * Mtodo que diz se dois objetos do tipo <code>JName</code> so
     * iguais. Para que isso seja verdade, os identificadores que fazem parte
     * de <code>this</code> tm que ser idnticos aos do objeto que est sendo
     * comparado.
     *
     * @param objeto Um objeto qualquer que ser comparado com <code>this</code>.
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica que
     *         o valor do objeto com o qual <code>this</code> est sendo
     *         comparado  igual ao de <code>this</code>.
     */
    public boolean equals(Object objeto) {
        boolean retorno = false;
        if (objeto instanceof JName) {
            JName name = (JName) objeto;

            // se os dois forem ns Java,  preciso ver se ambos possuem
            // o mesmo tamanho
            if (this.isASourceNode() && name.isASourceNode() &&
                    this.size() == name.size()) {
                retorno = AbstractNode.equals(this.identifiers, name.getIdentifiers());
            }
            // se no, verifica-se se os ns JaTS so iguais
            else {
                retorno = compareJaTSNode(name);
            }
        }
        return retorno;
    }


    /**
     * Retorna o identificador que se encontra na posio fornecida.
     *
     * @param pos O ndice do identificador desejado.
     * @return Um objeto do tipo <code>JIdentifier</code>, correspondendo ao
     *         identificador que se encontra na posio fornecida.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>pos</code> for um inteiro negativo.
     */
    public JIdentifier getIdentifierAt(int pos) throws IllegalArgumentException {
        if (pos < 0 || pos >= identifiers.size()) {
            throw new IllegalArgumentException();
        }
        return (JIdentifier) this.identifiers.elementAt(pos);
    }


    /**
     * Remove o identificador que se encontra na posio fornecida, caso haja
     * algum.
     *
     * @param pos O ndice onde identificador desejado se encontra.
     * @throws IllegalArgumentException Levantada quando o argumento
     *                                  <code>pos</code> for um inteiro negativo.
     */
    public void removeIdentifierAt(int pos) throws IllegalArgumentException {
        if (pos < 0 || pos >= identifiers.size()) {
            throw new IllegalArgumentException();
        }
        this.identifiers.removeElementAt(pos);
    }


    /**
     * Retorna o nmero de identificadores que compe este nome.
     *
     * @return Um inteiro correspondente ao nmero de identificadores que compe
     *         este nome.
     */
    public int size() {
        int retorno = 0;
        if (this.identifiers != null) {
            retorno = this.identifiers.size();
        }
        return retorno;
    }

    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * varivel. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>conta</code>.
     *
     * @return Identificador representando o nome da varivel
     */
    public JIdentifier toVariableName() {
        return JIdentifier.unwrapIdentifier(this).toVariableName();
    }


    /**
     * Este mtodo retorna um identificador que representa o nome de uma
     * constante. Exemplo: <br>
     * <p/>
     * Seja o identificador  <code>Conta</code>, a chamada deste mtodo ter
     * como resultado:  <code>CONTA</code>.
     * <p/>
     * Obs.: Para identificadores com letras maisculas consecutivas, o mtodo
     * no retorna o valor esperado. Por ex.: RepositorioBDR -> REPOSITORIO_B_D_R
     *
     * @return Identificador representando o nome da varivel
     */
    public JIdentifier toConstantName() {
        return JIdentifier.unwrapIdentifier(this).toConstantName();
    }

    /**
     * Realiza o unwrapping de um nome, ou seja, recebe um objeto de um tipo a
     * partir do qual um nome poderia ser gerado e gera esse nome. Caso isso no
     * seja possvel, levanta uma exceo.
     *
     * @param obj O objeto a partir do qual o objeto do tipo <code>JName</code>
     *            dever ser gerado.
     * @return Um nome correspondente ao argumento fornecido. Por exemplo, se o
     *         objeto fornecido foi do tipo <code>JIdentifier</code>, o objeto
     *         retornado  um n do tipo <code>JName</code> composto por um nico
     *         identificador.
     * @throws IllegalArgumentException  Levantada
     *                                   se o argumento <code>obj</code> for <code>null</code> ou se
     *                                   ele for de um tipo a partir do qual no  possvel realizar a
     *                                   converso para um <code>JName</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   se o objeto fornecido como argumento tiver uma estrutura
     *                                   inconsistente ou incompleta.
     */
    public static JName unwrapName(Object obj) throws IllegalArgumentException,
            InconsistentNodeException {
        JName retorno = null;
        // exceo levantada caso o tipo nao possa ser "wrapped" pra Name
        final IllegalArgumentException illegalType =
                new IllegalArgumentException("invalid type. Can't wrap to name");

        if (obj instanceof INode) {
            INode node = (INode) obj;
            int nodeType = node.getTypeIdentifier();
            boolean nodeIsAJavaNode = !node.isVariable() && !node.isExecutable();

            switch (nodeType) {
                case JaTSNode.NAME:
                    retorno = (JName) node;
                    break;

                case JaTSNode.IDENTIFIER:
                    JIdentifier id = (JIdentifier) node;
                    if (nodeIsAJavaNode) {
                        retorno = new JName(id);
                    } else if (id.isVariable()) {
                        retorno = new JName();
                        retorno.setVariableName(id.getVariableName());
                        retorno.setVariableType(JaTSNode.NAME);
                    } else if (id.isExecutable()) {
                        retorno = new JName();
                        retorno.setExecutableDeclaration(id.getExecutableDeclaration());
                    }
                    break;

                case JaTSNode.TYPE:
                    retorno = ((JType) node).getTypeName();
                    break;

                case JaTSNode.NAME_LIST:
                    JNameList nameList = (JNameList) node;
                    if (nameList.size() == 1) {
                        retorno = (JName) nameList.elementAt(0);
                    } else {

                        throw illegalType;
                    }
                    break;

                case JaTSNode.FIELD_ACCESS:
                    if (nodeIsAJavaNode) {
                        JFieldAccess fAccess = (JFieldAccess) node;
                        if (fAccess.getReferencedObject() != null) {
                            JName refObj = JName.unwrapName(fAccess.getReferencedObject());
                            refObj.addIdentifier(fAccess.getFieldName());
                            retorno = refObj;
                        }
                    } else {
                        throw illegalType;
                    }
                    break;

                case JaTSNode.PACKAGE_DECLARATION:
                    if (nodeIsAJavaNode) {
                        retorno = ((JPackageDeclaration) node).getName();
                    } else {
                        throw illegalType;
                    }
                    break;
                default:
                    throw illegalType;
            }//switch
        }//if
        else if (obj instanceof String) {
            String str = (String) obj;
            retorno = new JName(str);
        } else {
            throw illegalType;
        }

        return retorno;
    }

    // colocar esse mtodo tb na classe PackageDeclaration
    public static JImportDeclaration createImportDeclaration(JName packageName, boolean allClasses) {
        JImportDeclaration retorno = new JImportDeclaration();
        retorno.setImportAllClasses(allClasses);
        // se importa todas as classes, basta setar o nome do pacote com o name
        // recebido
        if (allClasses) {
            retorno.setPackageName(packageName);
        }
        // se no,  preciso pegar a tlimo Identificador do Name e setar como
        // sendo a classe Importada
        else {
            // se o name possui mais de um identificador, est ok
            if (packageName != null && packageName.size() > 1) {
                int ultimo = packageName.size() - 1;
                retorno.setPackageName(packageName.subSequence(0, ultimo));
                JIdentifier className = packageName.getIdentifierAt(ultimo);
                retorno.setClassName(className);
            } else if (packageName != null && packageName.isVariable()) {
                // tratar o caso -- no deveria ser aceito, pois nesse ponto,
                // ele j deveria ser substituido
            }
        }
        return retorno;
    }

    public JImportDeclaration createImportDeclaration(boolean allClasses) {
        return createImportDeclaration(this, allClasses);
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * cada elemento de <code>this.identifiers</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;
        if (this.isASourceNode()) {
            int i = 0;

            // Envie o visitor atravs da rvore ...
            while (i < this.identifiers.size()) {
                (this.identifiers.elementAt(i)).accept(visitor, data);
                i++;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);


        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JName</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JName</code>.
     *                Por consequncia, objetos do tipo<code>JName</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node != null) {
            if (!(node instanceof JExpression)) {

                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_ARE_INCOMPATIBLE_MSG, this, node);
            } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

                // Um JName s pode casar com uma expresso qualquer se ele for uma
                // varivel. Caso no seja, s  possvel casar com outro JName,
                // assim como ocorre com qualquer outro tipo de n.
                JName name = null;
                try {
                    name = JName.unwrapName(node);
                } catch (InconsistentNodeException e) {
                    throw new NodesNotMatchedException("Inconsistent node (" +
                            node.getClass().toString() + ")", this, node);
                }
                this.identifiers.match(name.getIdentifiers(), results);
            }
        }
        // Se node for nulo e o casamento deste no nao for opcional...
        else if (!this.isOptional()) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Levantada
     *                                   caso o parmetro <code>data</code> seja <code>null</code> numa
     *                                   situao em que ele  necessrio.
     * @throws InconsistentNodeException Levantada se um n inconsistente (incompleto) for encontrado.
     *                                   Se algum dos componentes (filhos) essenciais ao correto
     *                                   funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            int max = this.size();

            while (i < max) {
                INode currId = this.getIdentifierAt(i);
                if (currId != null) {
                    Object ret = this.processNode(currId, data);
                    if (currId.isExecutable()) {
                        this.setIdentifierAt(JIdentifier.unwrapIdentifier(ret), i);
                    }
                }
                i++;
            }
        }

        return retorno;
    }

    /**
     * Avalia este Name. A avaliao do Name consiste apenas da execuo da
     * expresso executvel associada.
     *
     * @param data
     * @return Resultado do processamento da expresso executvel associada
     *         ou apenas o no, caso ele no seja executvel
     * @throws ExecutionException
     */
    public Object evaluate(Object data) throws ExecutionException {
        Object result = this;
        if (this.isExecutable()) {
            try {
                result = this.processExecutableNode(data);
            } catch (InconsistentNodeException e) {
                throw new ExecutionException(e.getMessage());
            }

        }
        return result;
    }


	/**
	 * @return Returns the declarationRef.
	 */
	public INode getDeclarationRef() {
		return declarationRef;
	}


	/**
	 * @param declarationRef The declarationRef to set.
	 */
	public void setDeclarationRef(INode declarationRef) {
		this.declarationRef = declarationRef;
	}


}
