/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
 package cin.jats.engine.parser.nodes;


import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa uma lista de objetos do tipo <code>JName</code>.
 *
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 * @see JName
 */
public class JNameList extends AbstractNode implements JaTSCollection {

    // A lista onde sero armazenados os nomes
    private NodeList names;


    /**
     * Construtor vazio da classe.
     */
    public JNameList() {
        this.names = new NodeList();
        this.setTypeIdentifier(JaTSNode.NAME_LIST);

    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>Vector</code>
     * contendo<br>  apenas objetos do tipo JName
     *
     * @param objects Lista de nomes
     * @throws IllegalArgumentException Levantada se a lista passada
     *                                  contiver algo q no seja um JName
     */
    protected JNameList(NodeList objects) throws IllegalArgumentException {
        if (Util.verifyNodeList(objects, JaTSNode.NAME)) {
            this.names = objects;
        } else {
            throw new IllegalArgumentException("lista de nomes invlida");
        }
        this.setTypeIdentifier(JaTSNode.NAME_LIST);
    }


    /**
     * Adiciona um novo nome  lista.
     *
     * @param name Um objeto qualquer que possa ser convertido para um
     *             <code>JName</code>.
     * @throws IllegalArgumentException Se o objeto passado for null ou
     *                                  no puder ser convertido para <code>JName</code>
     */
    public void add(INode name) throws IllegalArgumentException {

        if (name == null) {
            throw new IllegalArgumentException("JName passado  null");
        }
        // tenta converter o objeto passado para JName
        JName newName = null;
        try {
            newName = JName.unwrapName(name);
            if (!this.contains(newName)) {
                this.names.addElement(newName);
            }
        } catch (InconsistentNodeException e) {
            throw new IllegalArgumentException("objeto invalido para a JNameList");
        }


    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param name Um objeto qualquer do tipo    <code>JName</code>.
     */
    public void add(String name) throws IllegalArgumentException {
        try {
            JName newName = JName.unwrapName(name);
            if (!this.contains(newName)) {
                this.names.addElement(newName);
            }
        } catch (InconsistentNodeException e) {
            throw new IllegalArgumentException("JName passado 	inconsistente");
        }
    }


    /**
     * Retorna a capacidade desta lista.
     *
     * @return Um inteiro correspondendo  capacidade da lista.
     */
    public int capacity() {
        return this.names.capacity();
    }


    /**
     * Verifica se este nome pertence a lista
     *
     * @param name Nome procurado
     * @return true, caso o nome passado pertenca a lista false, caso
     *         contrrio
     */
    public boolean contains(JName name) {
        return this.names.contains(name);
    }


    /**
     * Retorna o indice do elemento passado como parametro
     *
     * @param elem Elemento cujo indice eh procurado
     * @return Description of the Return Value
     */
    public int indexOf(JName elem) {
        return this.names.indexOf(elem);
    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para
     * a<br>
     * direita na posies do item que esiver naquela posio e nas daqueles
     * que<br>
     * . estiverem  direita dela.
     *
     * @param name Um objeto qualquer do tipo     <code>JName</code>.
     * @param pos  Um inteiro correspondendo   posio onde o n deve
     *             ser  inserido.
     */
    public void insertElementAt(INode name, int pos) {
        boolean validPosition = pos >= 0 && pos < this.size();
        if (name == null || !(name instanceof JName) || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.names.insertElementAt(name, pos);
    }


    /**
     * Diz se a lista est vazia.
     *
     * @return Um boolean que, caso tenho o valor <code>true</code>, indica
     *         que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.names.isEmpty();
    }


    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     *
     * @param pos Um inteiro correspondendo  posio desejada.
     * @return Um objeto do tipo <code>JName</code> que se encontra na
     *         posio <br>
     *         desejada.
     */
    public INode elementAt(int pos) {
        return (JName) this.names.elementAt(pos);
    }


    /**
     * Remove o elemento que se encontra na posio <code>pos</code> desta
     * lista.
     *
     * @param pos Um inteiro correspondendo
     *             posio do n que deve ser<br>
     *            removido.
     * @throws IllegalArgumentException Description of the
     *                                  Exception
     */
    public void removeElementAt(int pos) throws IllegalArgumentException {
        if (pos < 0) {
            throw new IllegalArgumentException();
        }
        this.names.removeElementAt(pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista, <br>
     * sobreescrevendo o elemento que estava nela.
     *
     * @param name Um objeto qualquer do tipo    <code>JName</code>.
     * @param pos  Um inteiro correspondendo   posio onde o n deve
     *             ser    inserido.
     */
    public void setElementAt(INode name, int pos)
            throws IllegalArgumentException {
        boolean validPosition = pos >= 0 && pos < this.size();

        if (name == null || !(name instanceof JName) || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.names.setElementAt(name, pos);
    }


    /**
     * Retorna o nmero de elementos da lista.
     *
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        int retorno = 0;
        if (names != null) {
            retorno = this.names.size();
        }
        return retorno;
    }


    /**
     * Retorna um objeto do tipo <code>JaTSIterator</code>
     *
     * @return um objeto do tipo JaTSIterator para que se possa percorrer esta
     *         lista de nomes
     */
    public JaTSIterator getIterator() {
        JaTSIterator jit = new JaTSIterator(this.names.iterator());
        return jit;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos e
     * quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento (
     * <code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code>
     *                com o qual <code>this</code> ser casado. Para quase todos os tipos
     *                de n, s  possvel realizar um casamento entre dois ns do mesmo
     *                tipo, mesmo que um deles seja uma varivel. Objetos do tipo <code>JName</code>
     *                so uma exceo a essa regra. Sempre que uma varivel casa com uma
     *                expresso, o objeto correspondente a essa varivel  um <code>JName</code>
     *                . Por consequncia, objetos do tipo<code>JName</code> podem casar
     *                com qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do
     *                casamento.  usado para armazenar os pares <code>(varivel,cdigo)</code>
     *                que so gerados pelo casamento.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando algum
     *                                   componente essencial do n tiver um valor, tornando o n, como um
     *                                   todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o
     *                                   casamento falhar porque <code>node</code> no  um n compatvel com
     *                                   <code>this</code>, ou seja, nem <code>this</code>  uma varivel no
     *                                   casada nem <code>this</code>  idntico a <code>this</code>.
     * @throws IllegalArgumentException  Description of the
     *                                   Exception
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        ResultSet rset = (ResultSet) results.clone();
        if (node != null) {
            if (!(node instanceof JNameList)) {
                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG, this, node);
            } else if (!this.matchesAsAVariable(node, rset, NO_TYPE)) {

                JNameList nameList = (JNameList) node;
                if (!this.isOptional() && (this.size() != nameList.size())) {
                    throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_SIZES_MSG,
                            this, node);
                }

                this.names.match(nameList.names, results);
            }
        }
        // Se node for nulo e o casamento deste no nao for opcional...
        else if (!this.isOptional()) {
            throw new NodesNotMatchedException(NodesNotMatchedException.INCONSISTENT_NODE,
                    this, node);
        }
        results.merge(rset);
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um   objeto do tipo <code>IVisitor</code>,
     *                funcionando de acordo com o padro de projeto homnimo, ou seja, implementa uma operao sobre
     *                vrios tipos diferentes de objetos fora do cdigo desses objetos. A
     *                partir dele ser invocado o mtodo <code>visit</code>, que recebe
     *                como argumento <code>this</code>.
     * @param data    Um  objeto qualquer que ser utilizado pelo visitor.
     * @return Um  objeto qualquer que possa vir a ser retornado
     *         pelo visitor.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     * @throws ExecutionException
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, ExecutionException {

        Object retorno = null;

        if (this.isASourceNode() && this.names != null) {

            int i = this.names.size() - 1;
            // empilha as interfaces no sentido oposto
            while (i >= 0) {
                this.names.elementAt(i).accept(visitor, data);
                i--;
            }

        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de trs
     * maneiras diferentes, dependendo da situao:<br>
     * <br>
     * 1 - <i>O n atual  executvel</i> . Nesse caso, o valor retornado pelo
     * processamento da <code>JExecutableDeclaration</code> associado ao n
     * (uma vez que todo n executvel tem um n do tipo <code>JExecutableDeclaration</code>
     * associado)  retornado pelo seu mtodo process e corresponde ao
     * resultado de seu processamento. <br>
     * <br>
     * 2 - <i>O n atual  pai de um n executvel</i> . Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em sua
     * lista de filhos pelo valor retornado pelo processamento de N. Alm
     * disso, o processamento do n inclui a responsbilidade de invocar o
     * mtodo <code>process</code> em cada um de seus filhos. <br>
     * <br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i> .
     * Aqui, o processamento do n se resume a invocar o mtodo <code>process</code>
     * para cada um dos filhos do n. <br>
     * <br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O   resultado do processamento do n.
     * @throws ExecutionException
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento de cada um dos names individualmente.
            int i = 0;
            int max = this.size();
            JNameList tempList = new JNameList();
            while (i < max) {
                INode currName = this.elementAt(i);
                if (currName != null) {
                    Object ret = this.processNode(currName, data);
                    if (currName.isExecutable()) {
                        try {
                            if (ret instanceof JNameList &&
                                    ((JNameList) ret).size() > 1) {
                                tempList = tempList.concat((JNameList) ret);
                                this.setElementAt(((JNameList) ret).elementAt(0), i);
                                // se cair nessa condio, ferra os meus contadores!
                                // Tenho q ter dois!!!
                            } else {
                                this.setElementAt(JName.unwrapName(ret), i);
                            }
                        } catch (Exception e) {
                            throw new ExecutionException(e.getMessage(), this);
                        }
                    }
                }
                i++;
            }

            JaTSIterator it = tempList.getIterator();
            while (it.hasNext()) {
                this.add(it.next());
            }

        }
        return retorno;
    }


    public JNameList concat(JNameList list) {
        JaTSIterator it = this.getIterator();
        JNameList retorno = new JNameList();
        JName tmpName = null;
        while (it.hasNext()) {
            retorno.add(it.next());
        }
        it = list.getIterator();

        // s adiciona os nomes da segunda lista, se eles no existirem na
        // primeira
        while (it.hasNext()) {
            tmpName = (JName) it.next();
            if (retorno.contains(tmpName)) {
                retorno.add(tmpName);
            }
        }

        return retorno;
    }


    /**
     * Compara dois objetos do tipo JNameList e diz se estes so iguais
     *
     * @param obj
     * @return
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JNameList) {
            JNameList nlist = (JNameList) obj;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.names, nlist.names);
            } else {
                retorno = compareJaTSNode(nlist);
            }
        }

        return retorno;

    }

    public static String[] toStringArray(JNameList nl) {
        String[] retorno = null;
        JName tmpName = null;
        if (nl != null && nl.size() > 0) {
            int size = nl.size();
            retorno = new String[size];
            for (int i = 0; i < size; i++) {
                tmpName = (JName) nl.elementAt(i);
                if (tmpName.isVariable()) {
                    retorno[i] = tmpName.getVariableName();
                } else {
                    retorno[i] = tmpName.toString();
                }
            }
        } else {
            throw new IllegalArgumentException("parametro nulo");
        }
        return retorno;
    }

}
