/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe guarda as informacoes a respeito de uma declarao de pacote
 * 
 * @author Adeline de Sousa
 * @version  0.1 -  16/08/2001
 * @since JDK 1.3
 */
public class JPackageDeclaration extends AbstractNode {

    // nome do pacote
    private JName packageName;

    /**
     * Construtor vazio da classe
     */
    public JPackageDeclaration() {
        this.packageName = null;
        this.setTypeIdentifier(JaTSNode.PACKAGE_DECLARATION);
    }


    /**
     * Construtor da classe
     *
     * @param pack Nome do pacote
     */
    public JPackageDeclaration(JName pack) {
        this.setName(pack);
        this.setTypeIdentifier(JaTSNode.PACKAGE_DECLARATION);
    }


    /**
     * Retorna o nome deste pacote
     *
     * @return Nome do pacote
     */
    public JName getName() {
        return this.packageName;
    }

    /**
     * Mtodo utilitario. Serve para retornar o identificador
     * na posicao <code> pos</code> do name que representa o nome
     * do pacote
     *
     * @param pos Posicao do identificador desejado (indice comeca em 0)
     */
    public JIdentifier getIdentifierAt(int pos) {
        JIdentifier retorno = null;
        if (this.packageName != null && this.packageName.size() > pos) {
            retorno = this.packageName.getIdentifierAt(pos);
        } else {
            throw new IllegalArgumentException("Posicao invalida");
        }
        return retorno;
    }

    /**
     * Seta o nome deste pacote
     *
     * @param pack Nome do pacote
     */
    public void setName(JName pack) {

        if (pack != null) {
            this.packageName = pack;
        } else {
            throw new IllegalArgumentException("nome invalido para o pacote");
        }
    }

    /**
     * Retorna o name correspondente ao nome do pacote com o sufixo adicionado
     *
     * @param suffix Sufixo a ser adicionado ao nome do pacote
     * @return Nome do pacote acrescido do sufixo
     */
    public JName addSuffix(String suffix) {
        JName retorno = null;
        if (Util.checkString(suffix) && this.isASourceNode()) {
            // pega a String correspondente ao name
            String tmpName = this.getName().toString();
            if (suffix.endsWith(";")) {
                suffix = suffix.substring(0, suffix.length() - 1);
            }
            tmpName = tmpName.concat(suffix);
            try {
                retorno = JName.unwrapName(tmpName);
            } catch (InconsistentNodeException e) {
                throw new IllegalArgumentException("Argumento invlido para concatenacao");
            }
        } else {
            throw new IllegalArgumentException("Argumento invlido para concatenacao");
        }
        return retorno;
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de cada um
     * dos filhos deste n.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            if (this.packageName != null) {
                this.packageName.accept(visitor, data);
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // Visite este no
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node != null) {
            if (!(node instanceof JPackageDeclaration)) {
                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                        this, node);
            } else if (!this.matchesAsAVariable(node, results,
                    JaTSNode.PACKAGE_DECLARATION)) {

                JPackageDeclaration packDec = (JPackageDeclaration) node;

                // casa o nome do pacote
                if (this.packageName != null) {
                    this.packageName.match(packDec.getName(), results);
                }

            }
        } else if (!this.isOptional()) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // processamento do nome do pacote
            INode pack = this.getName();
            if (pack != null) {
                Object ret1 = this.processNode(pack, data);

                if (pack.isExecutable()) {
                    try {
                        this.setName(JName.unwrapName(ret1));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }

                }
            }
        }
        return retorno;
    }


    /**
     * Compara dois objetos do tipo JPackageDeclaration e diz se estes so iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JPackageDeclaration) {
            JPackageDeclaration pack = (JPackageDeclaration) obj;
            // compara como n Java
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.packageName, pack.packageName);
            }
            // compara os ns JaTS
            else {
                retorno = compareJaTSNode(pack);
            }
        }

        return retorno;
    }

    // colocar esse mtodo tb na classe PackageDeclaration
    public static JImportDeclaration createImportDeclaration(JName packageName, boolean allClasses) {
        JImportDeclaration retorno = new JImportDeclaration();
        retorno.setImportAllClasses(allClasses);
        // se importa todas as classes, basta setar o nome do pacote com o name
        // recebido
        if (allClasses) {
            retorno.setPackageName(packageName);
        }
        // se no,  preciso pegar a tlimo Identificador do Name e setar como
        // sendo a classe Importada
        else {
            // se o name possui mais de um identificador, est ok
            if (packageName != null && packageName.size() > 1) {
                JName tmpPckgName = (JName) packageName.clone();
                int ultimo = tmpPckgName.size() - 1;
                JIdentifier className = tmpPckgName.getIdentifierAt(ultimo);
                tmpPckgName.removeIdentifierAt(ultimo);
                retorno.setPackageName(tmpPckgName);
                retorno.setClassName(className);
            } else if (packageName != null && packageName.isVariable()) {
                // tratar o caso -- no deveria ser aceito, pois nesse ponto,
                // ele j deveria ser substituido
            }
        }
        return retorno;
    }

    public JImportDeclaration createImportDeclaration(boolean allClasses) {
        return createImportDeclaration(this.getName(), allClasses);
    }


}