/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe implementa uma lista de objetos do tipo
 * <code>JFormalParameter</code>.
 * 
 * @see JFormalParameter
 * @author Adeline de Sousa Silva
 */
public class JParameterList extends AbstractNode implements JaTSCollection {


    // A lista onde sero armazenados os parametros
    private NodeList parameters;


    /**
     * Construtor vazio da classe.
     */
    public JParameterList() {
        this.parameters = new NodeList();
        this.setTypeIdentifier(JaTSNode.PARAMETER_LIST);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>Vector</code>
     * contendo apenas objetos que implementam JaTSNode
     */
    protected JParameterList(NodeList params) throws IllegalArgumentException {
        if (params == null) {
            throw new IllegalArgumentException();
        }
        JaTSIterator jit = params.getIterator();
        while (jit.hasNext()) {
            if (!(jit.next() instanceof JFormalParameter)) {
                throw new IllegalArgumentException();
            }
        }
        this.setTypeIdentifier(JaTSNode.PARAMETER_LIST);
        this.parameters = params;
    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param param Um objeto qualquer do tipo <code>JFormalParameter</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento<br>
     *          <code>param</code> for <code>null</code>.
     */
    public void add(INode param)
            throws IllegalArgumentException {

        if (param == null || !(param instanceof JFormalParameter)) {
            throw new IllegalArgumentException("param passado  null ou nao  " +
                    "um JFormalParameter");
        }
        this.parameters.addElement(param);
    }


    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     *
     * @param pos Um inteiro correspondendo  posio desejada.
     * @return Um objeto do tipo <code>JFormalParameter</code> que se encontra
     *         na posio desejada.
     */
    public INode elementAt(int pos) {
        return (JFormalParameter) this.parameters.elementAt(pos);
    }

    /**
     * Insere um elemento na lista na posio especificada e d um shift para a<br>
     * direita na posies do item que esiver naquela posio e nas daqueles que<br>.
     * estiverem  direita dela.
     *
     * @param param Um objeto qualquer do tipo <code>JFormalParameter</code>.
     * @param pos   Um inteiro correspondendo  posio onde o n deve ser <br>
     *              inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>param</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode param, int pos)
            throws IllegalArgumentException {
        boolean validPosition = (pos >= 0 && pos < this.size());
        boolean validParameter = (param != null &&
                (param instanceof JFormalParameter));
        if (!validParameter || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.parameters.insertElementAt(param, pos);
    }


    /**
     * Diz se a lista est vazia.
     *
     * @return Um boolean que, caso tenha o valor <code>true</code>, indica
     *         que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.parameters.isEmpty();
    }


    /**
     * Retorna a capacidade desta lista.
     *
     * @return Um inteiro correspondendo  capacidade da lista.
     */
    public int capacity() {
        return this.parameters.capacity();
    }


    /**
     * Remove o elemento que se encontra na  posio <code>pos</code> desta lista.
     *
     * @param pos Um inteiro correspondendo  posio do n que deve ser<br>
     *            removido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void removeElementAt(int pos) throws IllegalArgumentException {
        if (pos < 0) {
            throw new IllegalArgumentException();
        }
        this.parameters.removeElementAt(pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista, <br>
     * sobreescrevendo o elemento que estava nela.
     *
     * @param param Um objeto qualquer do tipo <code>JFormalParameter</code>.
     * @param pos   Um inteiro correspondendo  posio onde o n deve ser <br>
     *              inserido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o <br>
     *          argumento <code>param</code> for <code>null</code> ou o <br>
     *          argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode param, int pos)
            throws IllegalArgumentException {

        boolean validPosition = (pos >= 0 && pos < this.size());
        boolean validParameter = (param != null &&
                (param instanceof JFormalParameter));
        if (!validParameter || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.parameters.setElementAt(param, pos);
    }


    /**
     * Retorna o nmero de elementos da lista.
     *
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        return this.parameters.size();
    }


    /**
     * Retorna um objeto do tipo <code>JaTSIterator</code>
     *
     * @return um objeto do tipo JaTSIterator para que se possa percorrer esta
     *         lista de parametros
     */
    public JaTSIterator getIterator() {
        return this.parameters.getIterator();
    }


    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JParameterList) {
            JParameterList list = (JParameterList) obj;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.parameters, list.parameters);
            } else {
                retorno = compareJaTSNode(list);
            }
        }
        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     *                Para quase todos os tipos de n, s  possvel realizar um
     *                casamento entre dois ns do mesmo tipo, mesmo que um deles
     *                seja uma varivel. Objetos do tipo <code>JFormalParameter</code> so uma exceo
     *                a essa regra. Sempre que uma varivel casa com uma expresso,
     *                o objeto correspondente a essa varivel  um <code>JFormalParameter</code>.
     *                Por consequncia, objetos do tipo<code>JFormalParameter</code> podem casar com
     *                qualquer objeto que seja uma <code>JExpression</code>.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JParameterList)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JParameterList parameters = (JParameterList) node;
            int i = 0;
            int max = parameters.size();

            if (this.size() != max) {
                throw new NodesNotMatchedException("Parameter lists have " +
                        "different sizes (" + this.size() + ", " +
                        max + ")", this, node);
            }

            while (i < max) {
                this.elementAt(i).match(parameters.elementAt(i), results);
                i++;
            }
        }
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo para todos os filhos deste n, invocando o mtodo
     * <code> accept </code>de cada um deles.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;


        if (this.isASourceNode() && this.parameters != null) {

            int i = this.parameters.size() - 1;

            // empilha as interfaces no sentido oposto
            while (i >= 0) {

                this.parameters.elementAt(i).accept(visitor, data);
                i--;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcioparamnto do n estiver faltando.
     */

    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            int max = this.size();

            while (i < max) {
                INode currFP = this.elementAt(i);
                if (currFP != null) {
                    Object ret = this.processNode(currFP, data);

                    if (currFP.isExecutable()) {
                        if (ret instanceof JFormalParameter) {
                            this.setElementAt((JFormalParameter) ret, i);
                        } else {
                            throw new ExecutionException(this);
                        }
                    }

                }
                i++;
            }
        }

        return retorno;
    }


    /**
     * Tranforma a lista de parametros em uma lista de expressoes. Mais
     * especificamente, em uma lista de nomes. Util para gerar chamadas de
     * metodos.
     *
     * @return
     */
    public JExpressionList toExpressionList() {
        JExpressionList expList = new JExpressionList();
        JName name = null;

        for (int i = 0; i < this.size(); i++) {
            name = new JName(((JFormalParameter) this.elementAt(i)).getIdentifier());
            expList.addExpression(name);
        }

        return expList;
    }


}
