/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.Position;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;

/**
 * JStatementList represents a sequence of statements. <br><br>
 * 
 * This class extends JStatement because a JStatmentList is a statement
 * in jats pattern language. This way, it can appear between others
 * statements and be handled as a such one.
 * 
 * @author Gustavo Santos
 * @version 06/07/2005
 */
public class JStatementList extends JStatement implements JaTSCollection {

	
	private NodeList statements;
	
	public JStatementList() {
		this.statements = new NodeList();
        this.setTypeIdentifier(JaTSNode.STATEMENT_LIST);
	}

	public JStatementList(NodeList statements) {
		this();
		this.statements = statements;
	}

	/**
	 * Adds a new statement in the list
	 * 
     * @param node Description of the Parameter
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#add(cin.jats.engine.parser.nodes.INode)
	 */
	public void add(INode node) throws IllegalArgumentException {
	    this.statements.addElement(node);
	}

	/**
	 * Gets an element from the specified position
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#elementAt(int)
	 */
	public INode elementAt(int pos) {
		return this.statements.elementAt(pos);
	}

	/**
	 * Removes an element from the spacified position
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#removeElementAt(int)
	 */
	public void removeElementAt(int pos) throws IllegalArgumentException {
		this.statements.removeElementAt(pos);
	}

    /**
     * 
     * @param elem
     */
	public void remove(INode elem) {
        int index = this.statements.indexOf(elem);
        this.removeElementAt(index);

    }
    
    
	/**
	 * Inserts an element in the list on the specified position
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#insertElementAt(cin.jats.engine.parser.nodes.INode, int)
	 */
	public void insertElementAt(INode elem, int pos)
			throws IllegalArgumentException {
		this.statements.insertElementAt(elem, pos);

	}

	/**
	 * List size
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#size()
	 */
	public int size() {
		return this.statements.size();
	}

	/**
	 * True if the list is empty
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#isEmpty()
	 */
	public boolean isEmpty() {
		return this.statements.isEmpty();
	}

	/**
	 * Gets an iterator to transverse the list
	 * 
	 * @see cin.jats.engine.parser.nodes.JaTSCollection#getIterator()
	 */
	public JaTSIterator getIterator() {
		return this.statements.getIterator();
	}
	
	public NodeList getStatements() {
		return this.statements;
	}

	
    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>
     *                , funcionando de acordo com o padro de projeto homnimo, ou seja,
     *                implementa uma operao sobre vrios tipos diferentes de objetos
     *                fora do cdigo desses objetos. A partir dele ser invocado o mtodo
     *                <code>visit</code>, que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser
     *                utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a
     *         ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada quando algum componente
     *                                   essencial do n tiver um valor, tornando o n, como um todo,
     *                                   inconsistente.
     * @throws ExecutionException        Levantada sempre que a execuo de
     *                                   um n falhar.
	 * @see cin.jats.engine.parser.nodes.INode#accept(cin.jats.engine.visitors.IVisitor, java.lang.Object)
     */
	public Object accept(IVisitor visitor, Object data)
			throws InconsistentNodeException, IllegalArgumentException,
			ExecutionException {
        Object retorno = null;
        if (!this.statements.isEmpty()) {
            int max = this.size() - 1;

            // invoca o visitor para todos os elementos na ordem inversa
            for (int i = max; i >= 0; i--) {
                this.statements.elementAt(i).accept(visitor, data);
            }
        }
        // chama o visitor para a declarao executavel associadas ao JStatementList
        invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);

        return retorno;
    }
	
	
	int offSet;

	/* (non-Javadoc)
	 * @see cin.jats.engine.parser.nodes.INode#match(cin.jats.engine.parser.nodes.INode, cin.jats.engine.util.ResultSet)
	 */
	public void match(INode node, ResultSet results)
			throws IllegalArgumentException, InconsistentNodeException,
			NodesNotMatchedException {
		JStatementList list = (JStatementList)node;
		
		NodeList stmListTemp = new NodeList();
		INode st = null;
		JStatementList stmListVar = null;
		offSet = 0;

		for (int i = 0; i < this.statements.size(); i++) {
			st = (INode) this.statements.elementAt(i);

			// se e' variavel e nao foi casada ainda
			if (st instanceof JStatementList
					&& !results.alreadyMatched(st)) {
				// se a variavel e' o primeiro statement
				if (stmListTemp.isEmpty()) {
					// guarda a dita cuja
//					if (stmListVar != null)  WHY DID I PUT THIS CONDITION HERE????
						stmListVar = (JStatementList) st;
				// se encontrou outra variavel apos uma sequencia de
				// statements
				} else {
					// processa
					stmListTemp = processPartialMatching(results, //resultSet
							list, //o proprio n completo
							stmListTemp, //um trecho de codigo no padrao
							stmListVar); //a variavel
					// se entrou aqui e' pq ainda tem uma varivel pelo
					// menos
					// volta 1 pra pegar ela
					i--;
				}

			} else {
				// se for variavel ja' casada
				if (st instanceof JStatementList) {
					stmListTemp.addAll(
							((JStatementList)results.get(st)).getStatements());
				} else {
					stmListTemp.addElement(st);
				}
			}
		}
		if (!stmListTemp.isEmpty() || stmListVar != null) {
			stmListTemp = processPartialMatching(results, list,
					stmListTemp, stmListVar);
		}
		
	}

	private NodeList processPartialMatching(ResultSet results,
			JStatementList nodeBlock, 
			NodeList patternCodePortion, 
			JStatementList stmListVar
			)
			throws InconsistentNodeException, NodesNotMatchedException {
		NodeList stmSourceTemp = new NodeList();

		int j = 0;
		INode source = null, pattern = null;
		boolean restartLoop = false;

		// se existe apenas a varivel a ser casada
		if (patternCodePortion.size() == 0) {
			for (int i = offSet; i < nodeBlock.statements.size(); i++)
				stmSourceTemp.addElement(nodeBlock.statements.elementAt(i));
		} else {
			
			//while the verification of pattern code portion is not complete
			//and the source code list end is not reached
			for (j = 0; j < patternCodePortion.size()
					&& (j + offSet) < nodeBlock.statements.size();) {
				restartLoop = false;
				
				//try no match each statment
				try {
					pattern = patternCodePortion.elementAt(j);
					source = nodeBlock.statements.elementAt(j + offSet);
					pattern.match(source, results);
				
				//if the nodes are different
				} catch (NodesNotMatchedException e) {
					
					//add the first element of the source statement list 
					//to the list of statements witch will be matched to
					//the variable
					stmSourceTemp.addElement(
							nodeBlock.statements.elementAt(offSet));
					
					//move one position forward on the source statemens list
					offSet++;
					
					//try to match the hole portion again
					restartLoop = true;
				}
				j = restartLoop ? 0 : j + 1;
			}
		}
		
		//if the hole pattern code has been matched, we have the value
		//to be matched to the variable
		if (j == patternCodePortion.size()) {
			
			//if such variable exists
			if (stmListVar != null) {
				
				//match (the list may be empty)
				JStatementList stList = new JStatementList(stmSourceTemp);
				
				
				if (stmListVar instanceof ContextDeclaration) {
					stmListVar.match(stList, results);
				} else {
					results.put(stmListVar, stList);
				}
			}
			patternCodePortion = new NodeList();
			stmListVar = null;
			offSet += j;
		} else {
			throw new NodesNotMatchedException(this, nodeBlock);
		}
		return patternCodePortion;
	}
	
	/* (non-Javadoc)
	 * @see cin.jats.engine.parser.nodes.INode#process(java.lang.Object)
	 */
	public Object process(Object data) throws ExecutionException,
			InconsistentNodeException, IllegalArgumentException {
        Object retorno = this;
        if (this.isExecutable()) {
            retorno = processExecutableNode(data);
        } else {
            if (this.statements != null) {
                Object resultado = null;
                int i = 0;
                int stmtSize = this.statements.size();
                while (i < stmtSize) {
                    // pega o elemento na posio i e o processa
                    INode currStmt = this.statements.elementAt(i);
                    resultado = this.processNode(currStmt, data);
                    if (resultado != null) {
                        if (currStmt instanceof NestableExecutableDeclaration
                        		|| currStmt instanceof ContextDeclaration) {
                            this.remove(currStmt);
                            // decrementa o tamanho de 1 unidade, pois estamos
                            // removendo um elemento da lista de statements
                            stmtSize--;
                            if (resultado instanceof NodeList) {
                                this.statements.insertNodeListAt((NodeList) resultado, i);
                                // se a lista for acrescida de uma nova lista
                                // ento, somam-se seus tamanhos.
                                stmtSize += ((NodeList) resultado).size();
                            }
                            // adicionado em 15/04/2003, por Adeline
                            // isso se faz necessrio, nos casos em que
                            // o resultado  um conjunto vazio
                            i--;
                        }
                        // se for uma NestableExecutableExpression, remove-
                        // a da lista e adiciona-se o resultado de seu
                        // processamento - o qual  uma lista de ns
                        // dentro da lista atual
                        else {
                            // se o n atual no for uma declarao
                            // iterativa/condicinal, troca-se o n atual
                            // pelo resultado de seu processamento
                            this.statements.setElementAt((INode) resultado, i);
                        }
                    } else {
                        throw new ExecutionException("Process error in JBlock");
                    }

                    i++;
                }
            }
        }

        return retorno;
	}


	/**
	 * @param statements The statements to set.
	 */
	public void setStatements(NodeList statements) {
		this.statements = statements;
	}

}
