/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * 
 * @author Adeline de Sousa
 * @version 16/01/2003
 *
 */
public class JSwitchStatement extends JStatement {

    // expresso do switch
    private JExpression expression;

    // lista de cases do switch
    private NodeList cases;


    /**
     * Construtor default da classe
     */
    public JSwitchStatement() {
        super();
        this.setExpression(null);
        this.setCases(new NodeList());
    }

    /**
     * Construtor da classe
     *
     * @param expr   Expresso do Switch
     * @param labels Lista de labels contidas no switch
     */
    public JSwitchStatement(JExpression expr, NodeList labels) {
        super();
        this.setExpression(expr);
        this.setCases(labels);
    }


    /**
     * Retorna a expresso do switch
     *
     * @return Expresso do switch
     */
    public JExpression getExpression() {
        return expression;
    }


    /**
     * Seta a expresso do switch
     *
     * @param expr Expresso do switch
     */
    public void setExpression(JExpression expr) {
        this.expression = expr;
    }


    /**
     * Retorna a lista de associaes labels->statements do switch
     *
     * @return Lista de associaes labels->statements do switch
     */
    public NodeList getCases() {
        return cases;
    }


    /**
     * Seta a lista de associaes label -> statements do switch
     *
     * @param cases Lista de associaes label -> statements do switch
     */
    public void setCases(NodeList cases) {
        this.cases = cases;
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (!this.isExecutable() && !this.isVariable()) {

            // envia o visitor pela sub-arvore da lista de pares
            // (label->statements)
            if (this.getCases() != null) {
                int size = this.cases.size();

                for (int i = size - 1; i >= 0; i--) {
                    this.cases.elementAt(i).accept(visitor, data);
                }
            }

            // envia o visitor pela subarvore da expresso
            if (this.getExpression() != null) {
                this.expression.accept(visitor, data);
            }
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JSwitchStatement)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {
            JSwitchStatement jswitch = (JSwitchStatement) node;

            if (this.getExpression() != null && jswitch.getExpression() != null) {
                this.getExpression().match(jswitch.getExpression(), results);
            }
            if (this.getCases() != null && jswitch.getCases() != null) {

                if (this.getCases().size() != jswitch.getCases().size()) {
                    throw new NodesNotMatchedException("Different number of " +
                            "switch statements (" + this.getCases().size() +
                            ", " + jswitch.getCases().size() + ")", this, jswitch);
                }

                int i = 0;
                int size = this.cases.size();
                NodeList nodeLabels = (NodeList) (jswitch.getCases()).clone();

                while (i < size) {
                    JSwitchPair tempNode = (JSwitchPair) this.cases.elementAt(i);
                    int j = 0;
                    int size2 = nodeLabels.size();
                    boolean matchedOk = false;
                    while (j < size2 && !matchedOk) {
                        JSwitchPair nodeToBeMatched =
                                (JSwitchPair) nodeLabels.elementAt(j);

                        ResultSet tempRes = new ResultSet();
                        // Tenta casar tempNode com qualquer declarao presente
                        // na lista de labels de jswitch
                        try {
                            tempNode.match(nodeToBeMatched, tempRes);

                            // Se tudo der certo, nodeToBeMatched  retirado de
                            // nodeLabels e o resultado desse casamento
                            // individual  adicionado ao resultado do casamento
                            // como um todo.
                            results.merge(tempRes);
                            nodeLabels.removeElementAt(j);
                            matchedOk = true;
                        } catch (NodesNotMatchedException e) {
                        }
                        j++;
                    }
                }
            }
        }
    }


    /**
     *  responsvel por processar um determinado n. O processamento ocorre de
     * quatro maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n a ser processado  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n  pai de um n executvel</i>. Nesse caso, o processamento
     * do n consiste em substituir o n executvel (N) em sua lista de
     * filhos pelo valor retornado pelo processamento de N. Alm disso, o
     * processamento do n inclui a responsbilidade de invocar o mtodo
     * <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n no  executvel nem  pai de um n executvel</i>. Aqui,
     * o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     * 4 - <i>O n a ser processado  do tipo </i>
     * <code>Executable Declaration</code>. O processamento de ns do
     * tipo <code>Executable Declaration</code>  completamente
     * diferente dos outros tipos de n. Nos outros casos, tudo
     * que ocorre so invocaes de mtodos e mudanas leves na rvore
     * sinttica do programa. Em declaraes executveis, as rvores
     * sintticas correspondentes s expresses a ser executadas so
     * processadas de modo que o resultado gerado seja equivalente ao que
     * seria se o cdigo Java dessas expresses fosse executado. Por
     * exemplo, o processamento da rvore sinttica de uma invocao de
     * mtodo fornece como resultado o mesmo resultado que seria
     * fornecido se esse mtodo estivesse sendo invocado dentro de um
     * programa em Java.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws ExecutionException        Se algum erro
     *                                   ocorrer durante o processamento das estruturas executveis do
     *                                   programa.
     * @throws IllegalArgumentException  Caso o
     *                                   parmetro <code>data</code> seja <code>null</code>.
     * @throws InconsistentNodeException Levantada quando um n inconsistente (incompleto) for encontrado.
     *                                   Por exemplo, se uma declarao de atributo no possuir um filho
     *                                   correspondendo ao tipo da declarao.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // se o n for executavel, processa o n
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }

        // se o n nao for executavel, processa os filhos
        else {
            // processa o label
            INode tmpLabel = this.getExpression();
            if (tmpLabel != null) {
                Object ret1 = this.processNode(tmpLabel, data);
                if (tmpLabel.isExecutable()) {
                    try {
                        this.setExpression((JExpression) ret1);
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // processa a lista de statements
            NodeList slabels = this.getCases();
            int i = 0;
            int max = slabels.size();
            while (i < max) {
                INode currStatement = slabels.elementAt(i);
                if (currStatement != null) {
                    Object ret = this.processNode(currStatement, data);
                    if (currStatement.isExecutable()) {
                        if (!(ret instanceof JSwitchPair))
                            throw new ExecutionException(this);
                        slabels.setElementAt((JSwitchPair) ret, i);
                    }
                }
                i++;
            }
        }
        return retorno;
    }

}


