/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa expresses ternrias. As nicas expresses que se
 * enquadram nesta categoria so as expresses no formato
 * <code>cond?exp1:exp2</code>,
 * chamadas de Expresses Condicionais.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class JTernaryExpression extends JExpression {


    /**
     * Armazena o operador desta expresso.
     */
    private int ternaryOperator;

    /**
     * Armazena o primeiro operando da expresso.
     */
    private JExpression firstOperand;

    /**
     * Armazena o segundo operando da expresso.
     */
    private JExpression secondOperand;

    /**
     * Armazena o terceiro operando da expresso.
     */
    private JExpression thirdOperand;


    // Constantes inteiras correspondendo aos possveis tipos de operao..

    /**
     * Expresso condicional do tipo <code>cond?exp1:exp2</code>.
     */
    public final static int CONDITIONALEXPRESSION = 1;


    /**
     * Construtor vazio da classe
     */
    public JTernaryExpression() {
        super();
        this.firstOperand = null;
        this.secondOperand = null;
        this.thirdOperand = null;
        this.ternaryOperator = 0;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro o operador binrio
     * e os dois operandos sobre os quais ele ser aplicado.
     *
     * @param first  O primeiro operando da expresso.
     * @param second O segundo operando da expresso.
     * @param third  O terceiro operando da expresso.
     */
    public JTernaryExpression(JExpression first,
                              JExpression second, JExpression third) {
        super();
        this.ternaryOperator = CONDITIONALEXPRESSION;
        this.firstOperand = first;
        this.secondOperand = second;
        this.thirdOperand = third;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.firstOperand</code>, o de <code>this.secondOperand</code> e o
     * de <code>this.thirdOperand</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...

            if (this.firstOperand == null || this.secondOperand == null ||
                    this.thirdOperand == null)
                throw new InconsistentNodeException(this);

            this.firstOperand.accept(visitor, data);
            this.secondOperand.accept(visitor, data);
            this.thirdOperand.accept(visitor, data);

            // ... invoque o accept da declarao executvel associada a este n,
            // caso haja alguma...
            this.invokeAcceptOnExecutableDeclaration(visitor, data);

            // e ento visite este n.
            retorno = visitor.visit(this, data);
        }

        return retorno;
    }

    /**
     * Constri um string correspondente  aplicao do operador <code>op</code> aos
     * trs operandos <code>first</code>, <code>second</code> e <code>third</code>.
     *
     * @param first  O primeiro operando da expresso.
     * @param second O segundo operando da expresso.
     * @param third  O terceiro operando da expresso.
     * @return Um <code>String</code> representando o operador <code>op</code> aplicado
     *         aos operandos <code>first</code>, <code>second</code> e <code>third</code>.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          algum dos operandos for <code>null</code> ou o parmetro <code>op</code>
     *          tiver um valor invlido.
     */
    public String buildTernaryExpressionStringFor(String first, String second,
                                                  String third)
            throws IllegalArgumentException {
        if (!Util.checkString(first) || !Util.checkString(second) ||
                !Util.checkString(third)) {
            throw new IllegalArgumentException();
        }

        String retorno = "";
        StringBuffer sbuffer = new StringBuffer();
        int op = this.getOperator();
        if (op == CONDITIONALEXPRESSION) {
            sbuffer.append(first);
            sbuffer.append(" ? ");
            sbuffer.append(second);
            sbuffer.append(" : ");
            sbuffer.append(third);
            retorno = sbuffer.toString();
        }
        return retorno;

    }

    /**
     * Mtodo que retorna o primeiro operando da expresso.
     *
     * @return Uma objeto do tipo JExpression correspondendo ao primeiro
     *         operando da expresso.
     */
    public JExpression getFirstOperand() {
        return this.firstOperand;
    }

    /**
     * Mtodo que retorna o operador ternrio propriamente dito.
     *
     * @return Um inteiro correspondendo a um operador ternrio.
     */
    public int getOperator() {
        return this.ternaryOperator;
    }

    /**
     * Mtodo que retorna o segundo operando da expresso.
     *
     * @return Uma objeto do tipo JExpression correspondendo ao segundo
     *         operando da expresso.
     */
    public JExpression getSecondOperand() {
        return this.secondOperand;
    }

    /**
     * Mtodo que retorna o terceiro operando da expresso.
     *
     * @return Um objeto do tipo JExpression correspondendo ao terceiro
     *         operando da expresso.
     */
    public JExpression getThirdOperand() {
        return this.thirdOperand;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_IS_NULL_MSG,
                    this, node);
        }

        if (!(node instanceof JTernaryExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JTernaryExpression exp = (JTernaryExpression) node;

            if (this.getOperator() != exp.getOperator()) {
                throw new NodesNotMatchedException("Different operators in " +
                        "ternary expressions", this, node);
            }

            if (this.getFirstOperand() == null || this.getSecondOperand() == null
                    || this.getThirdOperand() == null) {
                throw new InconsistentNodeException(this);
            }
            (this.getFirstOperand()).match(exp.getFirstOperand(), results);
            (this.getSecondOperand()).match(exp.getSecondOperand(), results);
            (this.getThirdOperand()).match(exp.getThirdOperand(), results);
        }

    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do primeiro operando.
            INode fst = this.getFirstOperand();
            if (fst != null) {
                Object ret1 = this.processNode(fst, data);

                if (fst.isExecutable()) {

                    if (!(ret1 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setFirstOperand((JExpression) ret1);
                }
            }
            // Processamento do segundo operando.
            INode snd = this.getSecondOperand();
            if (snd != null) {
                Object ret2 = this.processNode(snd, data);

                if (snd.isExecutable()) {

                    if (!(ret2 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setSecondOperand((JExpression) ret2);
                }
            }
            // Processamento do terceiro operando.
            INode thr = this.getThirdOperand();
            if (thr != null) {
                Object ret3 = this.processNode(thr, data);

                if (thr.isExecutable()) {

                    if (!(ret3 instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setThirdOperand((JExpression) ret3);
                }
            }
        }

        return retorno;
    }

    /**
     * Mtodo que seta o primeiro operando da expresso.
     *
     * @param first A expresso que representa o primeiro operando da expresso.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>first</code> for <code>null</code>.
     */
    public void setFirstOperand(JExpression first)
            throws IllegalArgumentException {

        if (first == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.firstOperand = first;
    }

    /**
     * Mtodo que seta o operador ternrio da expresso.
     *
     * @param op Um inteiro correspondendo ao operador ternrio.
     */
    public void setOperator(int op) {
        this.ternaryOperator = op;
    }

    /**
     * Mtodo que seta o segundo operando da expresso.
     *
     * @param second A expresso que representa o segundo operando da expresso.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>second</code> for <code>null</code>.
     */
    public void setSecondOperand(JExpression second)
            throws IllegalArgumentException {

        if (second == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.secondOperand = second;
    }

    /**
     * Mtodo que seta o terceiro operando da expresso.
     *
     * @param third A expresso que representa o terceiro operando da exp.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>third</code> for <code>null</code>.
     */
    public void setThirdOperand(JExpression third)
            throws IllegalArgumentException {

        if (third == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.thirdOperand = third;
    }

    /**
     * Compara dois objetos do tipo JTernaryExpression e diz se estes dois so
     * iguais
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JTernaryExpression) {
            JTernaryExpression texp = (JTernaryExpression) obj;
            if (this.isASourceNode()) {
                retorno = super.equals(texp) &&
                        (this.ternaryOperator == texp.ternaryOperator) &&
                        AbstractNode.equals(this.firstOperand, texp.getFirstOperand()) &&
                        AbstractNode.equals(this.secondOperand, texp.getSecondOperand()) &&
                        AbstractNode.equals(this.thirdOperand, texp.getThirdOperand());
            } else {
                retorno = compareJaTSNode(texp);
            }
        }
        return retorno;
    }
}
