/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * @todo Ver como sero tratados os arrays multidimensionais de C#
 */

/**
 * Classe que implementa tipos. Um tipo pode ser um tipo simples (adimensional)
 * ou uma matriz (uma ou mais dimenses).
 *
 * @author Fernando Castor
 */
public class JType extends JExpression {

    // Responsvel por armazenar o nome do tipo.
    private JName typeName;

    // Armazena o nmero de pares de colchetes que seguem o nome desse tipo,
    // indicando que o n representa uma matriz.
    private int dimensions;


    // tipos void, tipos primitivos e tipo Object
    public static final int VOID = 1;
    public static final int BOOLEAN = 2;
    public static final int BYTE = 4;
    public static final int CHAR = 8;
    public static final int SHORT = 16;
    public static final int INT = 32;
    public static final int LONG = 64;
    public static final int FLOAT = 128;
    public static final int DOUBLE = 256;
    public static final int STRING = 512;
    public static final int OBJECT = 1024;
    public static final int DECIMAL = 2048;

    // Strings representando os tipos void, tipos primitivos e tipo Object
    public static final String SVOID = "void";
    public static final String SBOOLEAN = "boolean";
    public static final String SBYTE = "byte";
    public static final String SCHAR = "char";
    public static final String SSHORT = "short";
    public static final String SINT = "int";
    public static final String SLONG = "long";
    public static final String SFLOAT = "float";
    public static final String SDOUBLE = "double";
    public static final String SSTRING = "String";
    public static final String SOBJECT = "Object";
    public static final String SDECIMAL = "decimal";


    /**
     * Construtor vazio da classe
     */
    public JType() {
        super();
        this.setName(this.getVariableName());
        this.dimensions = 0;
        this.setTypeIdentifier(JaTSNode.TYPE);
    }

    /**
     * Construtor da classe que recebe como parmetro o o nome do tipo e o
     * nmero de pares de colchetes que seguem esse nome.
     *
     * @param name Um objeto do tipo <code>JName</code>representando nome do tipo.
     */
    public JType(JName name) {
        super();
        this.setName(name);
        this.dimensions = 0;
        this.setTypeIdentifier(JaTSNode.TYPE);
    }

    /**
     * Construtor da classe que recebe como parmetro o o nome do tipo e o
     * nmero de pares de colchetes que seguem esse nome.
     *
     * @param typeName Um objeto do tipo <code>JName</code>representando nome do tipo.
     */
    public JType(JIdentifier typeName) {
        super();
        this.setName(typeName);
        this.dimensions = 0;
        this.setTypeIdentifier(JaTSNode.TYPE);
    }

    /**
     * Construtor da classe que recebe como parmetro o o nome do tipo e o
     * nmero de pares de colchetes que seguem esse nome.
     *
     * @param name          Um objeto do tipo <code>String</code>representando o nome do tipo.
     * @param numDimensions Nmero de pares de colchetes que seguem o nome desse
     *                      tipo, indicando que o tipo representa uma matriz.
     */
    public JType(String name, int numDimensions) {
        super();
        this.setName(name);
        this.setNumberOfDimensions(numDimensions);
        this.setTypeIdentifier(JaTSNode.TYPE);
    }


    /**
     * Construtor da classe que recebe como parmetro o o nome do tipo e o
     * nmero de pares de colchetes que seguem esse nome.
     *
     * @param name          Um objeto do tipo <code>JIdentifier</code>representando o nome do tipo.
     * @param numDimensions Nmero de pares de colchetes que seguem o nome desse
     *                      tipo, indicando que o tipo representa uma matriz.
     */
    public JType(JIdentifier name, int numDimensions) {
        super();
        this.setName(name);
        this.setNumberOfDimensions(numDimensions);
        this.setTypeIdentifier(JaTSNode.TYPE);
    }


    /**
     * Retorna o nmero de dimenses que do tipo, caso ele seja uma matriz.
     * Tipos que no correspondem a matrizes so tratados como adimensionais
     * (zero dimenses).
     *
     * @return Um inteiro, correspondendo ao nmero de dimenses do tipo.
     */
    public int getNumberOfDimensions() {
        return this.dimensions;
    }


    /**
     * Este mtodo retorna o nome do tipo, sem levar em considerao se o tipo
     * corresponde ou no a uma matriz.
     *
     * @return Um objeto do tipo <code>JName</code>, correspondendo ao nome do tipo.
     */
    public JName getTypeName() {
        return this.typeName;
    }

    /**
     * Adiciona um prefixo ao nome do tipo
     */
    public JIdentifier addPrefix(String prefix) {
        JIdentifier name = JIdentifier.unwrapIdentifier(this.getTypeName());
        return name.addPrefix(prefix);
    }

    /**
     * Adiciona um prefixo ao nome do tipo
     */
    public JIdentifier addPrefix(JIdentifier prefix) {
        JIdentifier name = JIdentifier.unwrapIdentifier(this.getTypeName());
        return name.addPrefix(prefix);
    }

    /**
     * Adiciona um sufixo ao nome do tipo
     */
    public JIdentifier addSuffix(String suffix) {

        JIdentifier name = JIdentifier.unwrapIdentifier(this.getTypeName());
        return name.addSuffix(suffix);

    }

    /**
     * Adiciona um sufixo ao nome do tipo
     */
    public JIdentifier addSuffix(JIdentifier suffix) {

        JIdentifier name = JIdentifier.unwrapIdentifier(this.getTypeName());
        return name.addSuffix(suffix);
    }


    /**
     * Diz se este  um tipo sem dimenso. Ou seja, se este no  um array
     *
     * @return true, se o tipo for adimensional
     *         false, caso contrrio
     */
    public boolean isDimensionless() {
        return this.getNumberOfDimensions() == 0;
    }

    /**
     * Diz se este tipo representa um array
     *
     * @return true, se o tipo for array
     *         false, caso contrrio
     */
    public boolean isArray() {
        return this.getNumberOfDimensions() > 0;
    }

    /**
     * Seta o nmero de dimenses do tipo.
     *
     * @param numDimensions Um inteiro correspondendo ao nmero de dimenses
     *                      do tipo.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>numDimensions</code> for negativo.
     */
    public void setNumberOfDimensions(int numDimensions) throws IllegalArgumentException {

        if (numDimensions < 0) {
            throw new IllegalArgumentException();
        }
        this.dimensions = numDimensions;
    }


    /**
     * Seta o nome do tipo.
     *
     * @param name Um <code>JName</code> correspondendo ao nome do tipo.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>name</code> for <code>null</code>.
     */
    public void setName(JName name) throws IllegalArgumentException {
        if (name == null) {
            throw new IllegalArgumentException();
        }
        this.typeName = name;

    }


    /**
     * Seta o nome do tipo.
     *
     * @param name Um <code>String</code> correspondendo ao nome do tipo.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>name</code> for <code>null</code>.
     */
    public void setName(String name) throws IllegalArgumentException {
        try {
            this.typeName = JName.unwrapName(name);
        } catch (InconsistentNodeException e) {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Seta o nome do tipo.
     *
     * @param name Um <code>JIdentifier</code> correspondendo ao nome do tipo.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>name</code> for <code>null</code>.
     */
    public void setName(JIdentifier name) throws IllegalArgumentException {
        try {
            this.typeName = JName.unwrapName(name);
        } catch (InconsistentNodeException e) {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Realiza o unwrapping de um tipo, ou seja, recebe um objeto de um tipo a
     * partir do qual um tipo poderia ser gerado e gera esse tipo. Caso isso no
     * seja possvel, levanta uma exceo.
     *
     * @param obj O objeto a partir do qual o objeto do tipo <code>JType</code>
     *            dever ser gerado.
     * @return Um nome correspondente ao argumento fornecido. Por exemplo, se o
     *         argumento fornecido foi do tipo <code>JName</code>, o objeto
     *         retornado  um n do tipo <code>JType</code> que possui o nome
     *         fornecido como filho.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          se o argumento <code>obj</code> for <code>null</code> ou se
     *          ele for de um tipo a partir do qual no  possvel realizar a
     *          converso para um <code>JType</code>.
     */
    public static JType unwrapType(Object obj) throws
            IllegalArgumentException {
        JType retorno = null;
        if (obj != null) {
            if (obj instanceof JType) {
                retorno = (JType) obj;
            } else if (obj instanceof String) {
                retorno = new JType(new JName((String) obj));
            } else if (obj instanceof JName) {
                retorno = new JType((JName) obj);
            } else if (obj instanceof JIdentifier) {
                retorno = new JType((JIdentifier) obj);
            }
        } else {
            throw new IllegalArgumentException();
        }
        return retorno;
    }

    /**
     * Compara dois objetos do tipo JType e diz se estes dois so iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false;
        if (obj instanceof JType) {
            JType type = (JType) obj;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.typeName, type.typeName) &&
                        (this.dimensions == type.dimensions);
            } else {
                retorno = compareJaTSNode(type);
            }
        }
        return retorno;
    }


    /**
     * Retorna true, se o tipo for um tipo primitivo da linguagem Java
     *
     * @return true, se o tipo for um tipo primitivo da linguagem Java
     *         false, caso contrrio
     */
    public boolean isPrimitive() {
        boolean retorno = false;
        if (this.isASourceNode()) {
            JIdentifier id = this.typeName.getIdentifierAt(0);
            String str = id.getValue();
            /*
              adicionar o tratamento aos tipos primitivos de c#
              sbyte, uchar, uint, ufloat, udouble
              decimal, udecimal (usar bigdecimal, como wrapper)
              object, string
             */

            retorno = (str.equals("boolean") || str.equals("char") ||
                    str.equals("byte") || str.equals("short") ||
                    str.equals("int") || str.equals("long") ||
                    str.equals("float") || str.equals("double"));

        }
        return retorno;
    }


    /**
     * Retorna true, se o tipo for um tipo primitivo da linguagem Java ou
     * pertencer a uma das classes da linguagem (pacote java.lang)
     *
     * @return Um boolean
     */
    public boolean isSimple() {
        //NAO ESTAH CONCLUIDO, TEM QUE
        //DESCOBRIR COMO SABER SE O TIPO
        //ESTAH EM JAVA.LANG
        boolean retorno = false;
        String tName = this.typeName.toString();
        if (this.isPrimitive()
                || tName.indexOf("String") >= 0
                || tName.indexOf("Byte") >= 0
                || tName.indexOf("Character") >= 0
                || tName.indexOf("Double") >= 0
                || tName.indexOf("Float") >= 0
                || tName.indexOf("Integer") >= 0
                || tName.indexOf("Long") >= 0
                || tName.indexOf("Short") >= 0
        ) {
            retorno = true;
        }

        return retorno;
    }

    /**
     * Retorna true, se o tipo for void
     *
     * @return true, se o tipo for void
     *         false, caso contrrio
     */
    public boolean isVoid() {
        boolean retorno = false;
        if (!this.isVariable() && !this.isExecutable()) {
            JIdentifier id = this.typeName.getIdentifierAt(0);
            String str = id.getValue();
            retorno = (str.equals(SVOID));
        }
        return retorno;
    }


    /**
     * Retorna true, se o tipo for 'boolean'
     *
     * @return true, se o tipo for 'boolean'
     *         false, caso contrrio
     */
    public boolean isBoolean() {
        boolean retorno = false;
        if (!this.isVariable() && !this.isExecutable()) {
            JIdentifier id = this.typeName.getIdentifierAt(0);
            String str = id.getValue();
            retorno = (str.equals(SBOOLEAN));
        }
        return retorno;
    }


    /**
     * Retorna um inteiro, representando este tipo. No caso de o tipo
     * corresponder a um objeto qualquer, simplesmente retorna object. Caso
     * contrrio retorna um dos tipos primitivos ou void.
     * Tipos passveis de serem retornados:
     * <p/>
     * JType.VOID
     * JType.BOOLEAN
     * JType.BYTE
     * JType.CHAR
     * JType.SHORT
     * JType.INT
     * JType.LONG
     * JType.FLOAT
     * JType.DOUBLE
     * JType.OBJECT
     */
    public int getType() {
        int retorno = 0;
        // s pode ser verificado se o n no for executvel nem varivel
        if (this.isASourceNode()) {
            String strType = this.typeName.getIdentifierAt(0).getValue();
            if (strType.equals("void")) {
                retorno = VOID;
            } else if (strType.equals("boolean")) {
                retorno = BOOLEAN;
            } else if (strType.equals("byte")) {
                retorno = BYTE;
            } else if (strType.equals("char")) {
                retorno = CHAR;
            } else if (strType.equals("short")) {
                retorno = SHORT;
            } else if (strType.equals("int")) {
                retorno = INT;
            } else if (strType.equals("long")) {
                retorno = LONG;
            } else if (strType.equals("float")) {
                retorno = FLOAT;
            } else if (strType.equals("double")) {
                retorno = DOUBLE;
            } else if (strType.equals("String")) {
                retorno = STRING;
            } else {
                retorno = OBJECT;
            }
        }
        return retorno;
    }


    /**
     * Retorna uma String, representando este tipo. No caso de o tipo
     * corresponder a um objeto qualquer, simplesmente retorna a String "Object".
     * Caso contrrio retorna um dos tipos primitivos ou void.
     * Tipos passveis de serem retornados:
     * <p/>
     * JType.SVOID
     * JType.SBOOLEAN
     * JType.SBYTE
     * JType.SCHAR
     * JType.SSHORT
     * JType.SINT
     * JType.SLONG
     * JType.SFLOAT
     * JType.SDOUBLE
     * JType.SOBJECT
     */
    public String getStrType() {
        String retorno = "";
        // s pode ser verificado se o n no for executvel nem varivel
        if (this.isASourceNode()) {
            String strType = this.typeName.getIdentifierAt(0).getValue();
            if (strType.equals("void")) {
                retorno = SVOID;
            } else if (strType.equals("boolean")) {
                retorno = SBOOLEAN;
            } else if (strType.equals("byte")) {
                retorno = SBYTE;
            } else if (strType.equals("char")) {
                retorno = SCHAR;
            } else if (strType.equals("short")) {
                retorno = SSHORT;
            } else if (strType.equals("int")) {
                retorno = SINT;
            } else if (strType.equals("long")) {
                retorno = SLONG;
            } else if (strType.equals("float")) {
                retorno = SFLOAT;
            } else if (strType.equals("double")) {
                retorno = SDOUBLE;
            } else if (strType.equals("String")) {
                retorno = SSTRING;
            } else {
                retorno = this.getTypeName().toString();
            }
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.typeName</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.isASourceNode()) {
            // Envie o visitor atravs da rvore ...
            if (this.typeName == null)
                throw new InconsistentNodeException(this);

            this.typeName.accept(visitor, data);

        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);


        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this
     *                </code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com
     *                                   <code>this</code>, ou seja, nem <code>this</code> 
     *                                   uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {


        if (/*!Util.compatibleTypes(this, node) !this.isVariable() && */!(node instanceof JType)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JType theType = JType.unwrapType(node);

            if (this.getNumberOfDimensions() != theType.getNumberOfDimensions())
                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_SIZES_MSG, this, node);

            if (this.getTypeName() == null) {
                throw new InconsistentNodeException(this);
            }

            this.getTypeName().match(theType.getTypeName(), results);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do nome do tipo.
            INode tName = this.getTypeName();

            if (tName != null) {
                Object ret = this.processNode(tName, data);

                if (tName.isExecutable()) {
                    try {
                        this.setName(JName.unwrapName(ret));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }
        }

        return retorno;
    }

}
