/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe representa uma declaracao de Tipo, que pode ser tanto uma classe
 * quanto uma interface.
 *
 * @author Adeline de Sousa
 * @version 0.1 -  02/11/2001
 * @since JDK 1.3
 */
public abstract class JTypeDeclaration
        extends AbstractTypeDeclaration implements Declaration {

//	private AttributeSectionList attributes;
	
	private JNameList interfaces;

    // corpo do tipo
    private TypeBody body;

    // armazena as declaraes, com a informao de posio


    /**
     * Construtor vazio do tipo
     */
    protected JTypeDeclaration() {
        this.modifiers = new JModifierList(JModifierList.TYPE_DECLARATION);
        this.name = null;
        this.interfaces = null;
        this.setBody(new TypeBody());
        this.setTypeIdentifier(NO_TYPE);

    }


    /**
     * Construtor do tipo
     *
     * @param mods Lista de modificadores do tipo
     * @param id   Identificador do tipo
     * @param ifs  Lista de interfaces implementadas ou estendidas
     */
    protected JTypeDeclaration(JModifierList mods, JIdentifier id, JNameList ifs) {        
        this.modifiers = mods;
        this.setName(id);
        this.setInterfaces(ifs);
        this.setTypeIdentifier(NO_TYPE);
        this.setBody(new TypeBody());
    }

//	public AttributeSectionList getAttributes() {
//		return attributes;
//	}
//
//	public void setAttributes(AttributeSectionList atts) {
//		this.attributes = atts;
//	}

	public boolean isClass() {
        return (this.getTypeIdentifier() == JaTSNode.CLASS_DECLARATION);
    }

    public boolean isInterface() {
        return (this.getTypeIdentifier() == JaTSNode.INTERFACE_DECLARATION);
    }


    /**
     * Retorna a lista de interfaces implementadas por esta classe
     *
     * @return Lista de interfaces implementadas
     */
    public JNameList getInterfaces() {
        return this.interfaces;

    }

    /**
     * Seta o comentario associado a esta declarao de Tipo
     *
     * @param javadoc comentario associado a esta declarao de Tipo
     */
    public void setComment(String javadoc) {
        if (javadoc != null && javadoc.length() > 0) {
            this.comment = new JComment(javadoc);
        }
    }


    /**
     * Seta a lista de interfaces do tipo
     *
     * @param ifs Lista de interfaces implementadas por esta classe
     */
    public void setInterfaces(JNameList ifs) {

        this.interfaces = ifs;
    }


    /**
     * Adiciona uma interface  lista de interfaces implementadas por esta classe
     *
     * @param newInterface Interface a ser adicionada a esta declaracao de classe
     */
    public void addInterface(String newInterface) {
        if (this.hasInterface(newInterface)) {
            throw new IllegalArgumentException("interface ja existe");
        } 
        JName ni = new JName(newInterface);
        this.interfaces.add(ni);

    }

    /**
     * Retorna o corpo do tipo
     *
     * @return Corpo do tipo
     */
    public TypeBody getBody() {
        return body;
    }


    /**
     * Seta o corpo do tipo
     *
     * @param newBody Corpo do tipo.
     */
    public void setBody(TypeBody newBody) {
        if (newBody != null) {
            this.body = newBody;
        } else {
            throw new IllegalArgumentException("Corpo do tipo eh nulo");
        }
    }


    /**
     * Verifica se uma interface est na lista de interfaces implementadas
     * por classe
     *
     * @param intName Interface a ser procurada
     * @return true, se a interface existir
     *         false, caso contrrio
     */
    public boolean hasInterface(String intName) {

        JName n = new JName(intName);
        return this.interfaces.contains(n);
    }


    


    /**
     * Retorna true se a classe ou interface for protected.
     */
    public boolean isProtected() {
        return this.modifiers.hasModifier(JModifier.PROTECTED);
    }


    /**
     * Retorna o conjunto de mtodos do tipo.
     *
     * @return Conjunto de mtodos do tipo
     */
    public JMethodDeclarationSet getMethods() {
        return this.getBody().getMethodDeclarationSet().getMethodsSet();
    }

    /**
     * Seta o conjunto de mtodos do tipo.
     *
     * @param methods Conjunto de mtodos do tipo
     */
    public void setMethods(JMethodDeclarationSet methods) {
        this.getBody().getMethodDeclarationSet().setMethodsSet(methods);
    }

    /**
     * Retorna o conjunto de atributos do tipo.
     *
     * @return Conjunto de atributos do tipo
     */
    public JFieldDeclarationSet getFields() {
        return this.getBody().getFieldDeclarationSet().getFieldsSet();
    }

    /**
     * Seta o conjunto de atributos do tipo.
     *
     * @param fields Conjunto de atributos do tipo
     */
    public void setFields(JFieldDeclarationSet fields) {
        this.getBody().getFieldDeclarationSet().setFieldsSet(fields);
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {

            throw new IllegalArgumentException("results ou node sao nulos");
        }

        if (!(node instanceof JTypeDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, JaTSNode.TYPE_DECLARATION)) {

            JTypeDeclaration typeDec = (JTypeDeclaration) node;

            // casa o comentario
            if (this.getComment() == null) {
                if (typeDec.getComment() != null) {
                    JComment comment = typeDec.getComment();
                    String commentVar = Util.generateTypeCommentVariable(this.getName());
                    results.put(commentVar, comment, JaTSNode.COMMENT);
                }
            }

//            if (this.attributes != null) {
//            	//casa os atributos
//            	this.attributes.match(typeDec.attributes, results);
//            } else if (typeDec.attributes != null) {
//            	throw new NodesNotMatchedException("Attributes don't match", this, typeDec);
//            }

            if (this.modifiers != null) {
                // casa os modificadores
                this.modifiers.match(typeDec.getModifiers(), results);
            }

            if (this.name != null) {
                // casa os nomes
                this.name.match(typeDec.getName(), results);
            }

            if (this.interfaces != null) {
                this.interfaces.match(typeDec.getInterfaces(), results);
            } else if (typeDec.getInterfaces() != null) {
                throw new NodesNotMatchedException("Implemented interfaces didn't" +
                        " match", this, typeDec);
            }
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            //Processamento dos atributos
//        	AttributeSectionList tmpAttribute = this.getAttributes();
//            if (tmpAttribute != null) {
//                Object ret = this.processNode(tmpAttribute, data);
//                if (tmpAttribute.isExecutable()) {
//                    if (!(ret instanceof AttributeSectionList)) {
//                        throw new ExecutionException(this);
//                    }
//                    this.setAttributes((AttributeSectionList) ret);
//                }
//            }

            // Processamento da lista de modificadores do tipo.
            INode mods = this.getModifiers();
            if (mods != null) {
                Object ret1 = this.processNode(mods, data);
                if (mods.isExecutable()) {
                    if (!(ret1 instanceof JModifierList)) {
                        throw new ExecutionException(this);
                    }
                    this.setModifiers((JModifierList) ret1);
                }
            }

            // Processamento do nome do tipo.
            INode tName = this.getName();
            if (tName != null) {
                Object ret2 = this.processNode(tName, data);
                if (tName.isExecutable()) {
                    try {
                        this.setName(JIdentifier.unwrapIdentifier(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento da lista de interfaces implementadas pela classe.
            INode ifs = this.getInterfaces();
            if (ifs != null) {
                Object ret3 = this.processNode(ifs, data);
                if (ifs.isExecutable()) {
                    if (!(ret3 instanceof JNameList)) {
                        throw new ExecutionException(this);
                    }
                    this.setInterfaces((JNameList) ret3);
                }
                this.interfaces.setOptional(false);
            }

            // Processamento do corpo da classe ou interface.
            INode body = this.getBody();
            if (body != null) {
                Object ret4 = this.processNode(body, data);
                if (body.isExecutable()) {
                    if (!(ret4 instanceof TypeBody)) {
                        throw new ExecutionException(this);
                    }
                    this.setBody((TypeBody) ret4);
                }
            }
        }

        return retorno;
    }

}
