/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.util.Util;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.ExpressionEvaluationException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Classe que implementa expresses unrias. Tanto para operadores pr-fixados
 * quanto para operadores ps-fixados.
 *
 * @author Fernando Castor
 * @see JExpression
 */
public class JUnaryExpression extends JExpression {


    // Armazena o operador desta expresso.
    private int unaryOperator;


    // Armazena o operando da expresso.
    private JExpression operand;


    // Constantes inteiras correspondendo aos operadores possveis.

    /**
     * "++" prefixado.
     */
    public final static int PREFIXINCREMENTOPERATOR = 1;

    /**
     * "--" prefixado.
     */
    public final static int PREFIXDECREMENTOPERATOR = 2;

    /**
     * "++" sufixado.
     */
    public final static int POSTFIXINCREMENTOPERATOR = 3;

    /**
     * "--" sufixado.
     */
    public final static int POSTFIXDECREMENTOPERATOR = 4;

    /**
     * NOT lgico ("!").
     */
    public final static int CONDITIONALNOTOPERATOR = 5;

    /**
     * NOT bit a bit ("~").
     */
    public final static int NOTOPERATOR = 6;

    /**
     * Negao do sinal de um nmero ("-" unrio).
     */
    public final static int UNARYMINUSOPERATOR = 7;

    /**
     * ("+" unrio).
     */
    public final static int UNARYPLUSOPERATOR = 8;

    /**
     * (Expresso entre parnteses).
     */
    public final static int PARENTHESIZEDEXPRESSION = 9;
    
//    /**
//     * Checked Expression (C# only)
//     */
//    public final static int CHECKEDEXPRESSION = 10;
//    
//    /**
//     * Unchecked Expression (C# only)
//     */
//    public final static int UNCHECKEDEXPRESSION = 11;
//    
//    /**
//     * TypeOf Expression (C# only)
//     */
//    public final static int TYPEOFEXPRESSION = 12;


    /**
     * Construtor vazio da classe
     */
    public JUnaryExpression() {
        super();
        this.unaryOperator = 0;
        this.operand = null;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Construtor da classe que recebe como parmetro o operador unrio
     * e o operando sobre os quais ele ser aplicado.
     *
     * @param unOperator O operador binrio da operao. Pode ser qualquer um
     *                   dos operadores binrios de Java.
     * @param opnd       O operando sobre o qual o operador ser aplicado.
     */
    public JUnaryExpression(int unOperator, JExpression opnd) {
        super();
        this.unaryOperator = unOperator;
        this.operand = opnd;
        this.setTypeIdentifier(JaTSNode.EXPRESSION);
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * <code>this.operand</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        // Envie o visitor atravs da rvore ...
        if (this.operand == null) {
            throw new InconsistentNodeException(this);
        }

        this.operand.accept(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // e ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Constri um string correspondente  aplicao do operador <code>op</code> 
     * expresso representada pelo string exp.
     *
     * @param exp A expresso a ser impressa.
     * @return Um <code>String</code> representando o operador <code>op</code> aplicado
     *          expresso exp.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o
     *          parmetro <code>exp</code> for <code>null</code> ou o
     *          parmetro <code>op</code> tiver um valor invlido.
     */
    public String buildUnaryExpressionStringFor(String exp)
            throws IllegalArgumentException {
        if (!Util.checkString(exp))
            throw new IllegalArgumentException();
        String retorno = "";
        int op = this.getOperator();
        if (op == PREFIXINCREMENTOPERATOR)
            retorno = "++" + exp;
        else if (op == PREFIXDECREMENTOPERATOR)
            retorno = "--" + exp;
        else if (op == POSTFIXINCREMENTOPERATOR)
            retorno = exp + "++";
        else if (op == POSTFIXDECREMENTOPERATOR)
            retorno = exp + "--";
        else if (op == CONDITIONALNOTOPERATOR)
            retorno = "!" + exp;
        else if (op == NOTOPERATOR)
            retorno = "~" + exp;
        else if (op == UNARYMINUSOPERATOR)
            retorno = "-" + exp;
        else if (op == UNARYPLUSOPERATOR)
            retorno = "+" + exp;
        else if (op == PARENTHESIZEDEXPRESSION)
            retorno = "(" + exp + ")";
//        else if (op == CHECKEDEXPRESSION)
//            retorno = "checked (" + exp + ")";
//        else if (op == UNCHECKEDEXPRESSION)
//            retorno = "unchecked (" + exp + ")";
//        else if (op == TYPEOFEXPRESSION)
//            retorno = "typeof (" + exp + ")";
        else
            retorno = exp;
        return retorno;
    }

    /**
     * Mtodo que retorna o operando da expresso.
     *
     * @return Uma objeto do tipo JExpression correspondendo ao operando.
     */
    public JExpression getOperand() {
        return this.operand;
    }

    /**
     * Mtodo que retorna o operador unrio da expresso.
     *
     * @return Um inteiro correspondendo ao operador unrio.
     */
    public int getOperator() {
        return this.unaryOperator;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }

        if (!(node instanceof JUnaryExpression)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JUnaryExpression exp = (JUnaryExpression) node;

            if (this.getOperator() != exp.getOperator())
                throw new NodesNotMatchedException("Unary expressions have " +
                        "different operators", this, node);

            if (this.getOperand() == null)
                throw new InconsistentNodeException(this);

            (this.getOperand()).match(exp.getOperand(), results);
        }
    }

    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento do primeiro operando.
            INode oper = this.getOperand();
            if (oper != null) {
                Object ret = this.processNode(oper, data);

                if (oper.isExecutable()) {

                    if (!(ret instanceof JExpression))
                        throw new ExecutionException(this);

                    this.setOperand((JExpression) ret);
                }
            }
        }

        return retorno;
    }

    /**
     * Mtodo que seta o operando da expresso.
     *
     * @param opnd A expresso que corresponder ao operando da expresso.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o argumento
     *          <code>opnd</code> for <code>null</code>.
     */
    public void setOperand(JExpression opnd)
            throws IllegalArgumentException {

        if (opnd == null) throw new IllegalArgumentException(); // Texto da exceo.
        this.operand = opnd;
    }

    /**
     * Mtodo que seta o operador unrio da expresso.
     *
     * @param op Um inteiro correspondendo ao operador unrio.
     */
    public void setOperator(int op) {
        this.unaryOperator = op;
    }

    /**
     * Compara dois objetos do tipo JUnaryExpression e diz se estes dois so
     * iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false;
        if (obj instanceof JUnaryExpression) {
            JUnaryExpression uexp = (JUnaryExpression) obj;
            if (this.isVariable() && !this.isExecutable()) {
                retorno = super.equals(uexp) &&
                        (this.unaryOperator == uexp.getOperator()) &&
                        AbstractNode.equals(this.operand, uexp.getOperand());
            } else {
                retorno = compareJaTSNode(obj);
            }
        }

        return retorno;
    }

    public Object evaluate(Object data) throws ExecutionException {
        Object result = null;

        Object operand = this.operand.evaluate(data);
        int op = this.getOperator();

        // Deve ser possvel verificar que uma expresso unria tenha tipo 0? 
        // necessrio verificar por meio de testes.
        if (op == 0 || op == PARENTHESIZEDEXPRESSION) {
            result = operand;
        } else if (op == CONDITIONALNOTOPERATOR) {
            if (operand instanceof Boolean) {
                Boolean boolOpnd = (Boolean) operand;
                result = new Boolean(!boolOpnd.booleanValue());
            } else {
                throw new ExpressionEvaluationException(this);
            }
        } else if (op == NOTOPERATOR) {
            if (operand instanceof Integer) {
                Integer intOpnd = (Integer) operand;
                result = new Integer(~intOpnd.intValue());
            } else {
                throw new ExpressionEvaluationException(this);
            }
        } else if (op == UNARYMINUSOPERATOR) {
            if (operand instanceof Number) {
                Number numOpnd = (Number) operand;
                result = new Double(-numOpnd.doubleValue());
            } else {
                throw new ExpressionEvaluationException(this);
            }
        } else if (op == UNARYPLUSOPERATOR) {
            if (operand instanceof Number) {
                result = operand;
            } else {
                throw new ExpressionEvaluationException(this);
            }
        }


        return result;
    }

}
