/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

/**
 * Esta classe representa uma declarao de varivel. A qual pode ser
 * composta de uma lista de modificadores, um tipo e uma lista de decaradores
 * de varivel. Um declarador de varivel, nada mais  do que a varivel seguida
 * de sua inicializao
 * 
 * @author Adeline
 * @version  07/02/2003
 * @since JDK 1.3
 *
 */
public abstract class JVariableDeclaration extends AbstractNode
        implements Commentable, Declaration {

    // comentrio referente a este atributo
    private JComment comment;

//    private AttributeSectionList attributes;

    // modificadores do atributo
    protected JModifierList modifiers;
    // Tipos do atributo
    protected JType type;
    // lista das variaveis
    protected JVariableDeclaratorSet variables;

    protected JVariableDeclaration() {
        this.modifiers = null;
        this.type = null;
        this.variables = new JVariableDeclaratorSet();

    }

//    public AttributeSectionList getAttributes() {
//        return attributes;
//    }
//
//    public void setAttributes(AttributeSectionList attributes) {
//        this.attributes = attributes;
//    }

    /**
     * Adiciona uma nova variavel  lista de variveis.
     *
     * @param var Variavel a ser adicionada
     * @throws IllegalArgumentException - Se o parametro for nulo
     */
    public void addVariable(JVariableDeclarator var) {

        if (var != null) {
            this.variables.add(var);
        } else {
            throw new IllegalArgumentException("Declaracao de  variavel nula");
        }
    }

    /**
     * Adiciona uma nova variavel  lista de variveis.
     *
     * @param strVar Variavel a ser adicionada
     * @throws IllegalArgumentException - Se o parametro for nulo
     */
    public void addVariable(String strVar) {

        if (strVar != null) {
            this.variables.add(JVariableDeclarator.unwrapVariableDeclarator(strVar));
        } else {
            throw new IllegalArgumentException("Declaracao de  variavel nula");
        }
    }

    /**
     * Retorna true, caso o n, possua comentrio
     *
     * @return true, se  o n possui um comentrio
     *         false, caso contrrio
     */
    public boolean hasComment() {
        return (this.comment != null);
    }

    /**
     * Retorna o comentrio associado a este n
     *
     * @return comentrio associado a este n
     */
    public JComment getComment() {
        return this.comment;
    }

    /**
     * Seta o comentario associado a esta declarao de atributo
     *
     * @param javadoc comentario associado a esta declarao de atributo
     */
    public void setComment(JComment javadoc) {
        if (javadoc != null) {
            this.comment = javadoc;
        }

    }

    /**
     * Seta o comentario associado a esta declarao de atributo
     *
     * @param javadoc comentario associado a esta declarao de atributo
     */
    public void setComment(String javadoc) {
        if (javadoc != null && javadoc.length() > 0) {
            this.comment = new JComment(javadoc);

        }
    }

    /**
     * Seta a lista de modificadores da variavel
     *
     * @param mods Nova lista de modificadores
     */
    public abstract void setModifiers(JModifierList mods);

    /**
     * Retorna a lista de modificadores do atributo
     *
     * @return Modificadores do atributo
     */
    public JModifierList getModifiers() {
        return this.modifiers;
    }

    /**
     * Retorna o nome do atributo que se encontra na posio <code>i</code>
     * da lista de declaraes de atributo de um objeto do tipo
     * <code>JFieldDeclaration</code>.
     *
     * @param i O ndice do atributo cujo nome se deseja saber.
     * @return O nome do atributo desejado.
     */
    public JIdentifier getName(int i) {
        JVariableDeclarator v = this.getVariableDeclaratorAt(i);
        return v.getIdentifier();
    }


    /**
     * Retorna o tipo do atributo
     *
     * @return Tipo do atributo
     */
    public JType getType() {
        return this.type;
    }

    /**
     * /**
     * Retorna o declarador (um par consistindo de nome de variavel e
     * inicializacao de variavel, embora esta ultima possa nao estar presente)
     * que se encontra no indice <code>pos</code> da lista de declaradores da
     * declaracao de atributo.
     *
     * @return O declarador que se encontra na posicao desejada.
     *         Levantada quando o objeto que se encontra na posicao pedida
     *         nao corresponde a um declarador de variavel.
     */
    public JVariableDeclarator getVariableDeclaratorAt(int pos) {
        INode retorno = this.variables.elementAt(pos);
        return (JVariableDeclarator) retorno;
    }

    /**
     * Atualiza o declarador de varivel  da posio <code>pos</code> com o
     * valor do declaratod passado como parmetro.
     *
     * @param vd  Novo declarador de varivel
     * @param pos Posio do novo elemento
     */
    public void setVariableDeclaratorAt(JVariableDeclarator vd,
                                        int pos) {
        this.variables.setElementAt(vd, pos);
    }

    /*****************    Mtodos da API   **********************/

    /**
     * Adiciona um modificador  lista de modificadores deste atributo.
     *
     * @param strMod String que representa o modificador que se deseja adicionar
     * @throws IllegalArgumentException - Se o modificador passado for
     *                                  invlido para este n, se este modificador  j est na lista
     *                                  de modificadores ou se este for um modificador de acesso e
     *                                  j houver um.
     */
    public void addModifier(String strMod) throws IllegalArgumentException {

        this.modifiers.addModifier(strMod);
    }

    /**
     * Verifica se o modificador passado como parmetro pertence  lista de
     * modificadores de uma declarao de atributo.
     *
     * @param strMod O modificador que se deseja procurar
     * @return <code>true</code>, se o modificador passado como
     *         parmetro estiver na lista. <code>false</code>, caso contrrio
     * @ exception IllegalArgumentException - Se o modificador passado
     * for invalido
     */
    public boolean hasModifier(String strMod) throws IllegalArgumentException {
        return this.modifiers.hasModifier(strMod);
    }

    /**
     * Remove um modificador da lista de modificadores deste atributo.
     *
     * @param strMod Modificador a ser removido
     * @throws IllegalArgumentException - Se o modificador passado nao
     *                                  pertencer a lista de modificadores
     */
    public void removeModifier(String strMod) throws IllegalArgumentException {
        this.modifiers.removeModifier(strMod);
    }

    /**
     * Retorna o nome do atributo que se encontra na posio <code>0</code>
     * da lista de declaraes de atributo de um objeto do tipo, se s houver um
     * atributo.
     *
     * @return O nome do atributo desejado.
     */
    public JIdentifier getName() {
        if (this.size() != 1) {
            // fazer o texto dessa exceo
            throw new IllegalArgumentException("IllegalState:This method could" +
                    " not be invoked in a JVariableDeclaration whose variables" +
                    " length is different than 1");
        }
        return this.getName(0);
    }


    /*****************         Fim dos mtodos da API        ******************/

    /**
     * Seta o nome do atributo declarado no declarador de ndice i, onde i pode
     * ser o ndice de um declarador de varivel j existente ou o ndice daquele
     * que se encontra mais a esquerda mais 1.
     *
     * @param nome identificador correspondendo ao nome de um atributo.
     * @param i    posio do atributo
     */
    public void setName(JIdentifier nome, int i) throws
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        int indiceNovoAttr = this.variables.size();
        JVariableDeclarator varDec = null;

        if (nome == null || i < 0 || i > indiceNovoAttr)
            throw new IllegalArgumentException();

        // se o ndice passado for igual ao tamanho do atributo
        if (i == indiceNovoAttr) {
            varDec = new JVariableDeclarator();
            varDec.setIdentifier(nome);
            this.variables.add(varDec);
        } else {
            INode obj = this.variables.elementAt(i);
            if (!(obj instanceof JVariableDeclarator))
                throw new InconsistentNodeException(this);

            varDec = (JVariableDeclarator) obj;
            varDec.setIdentifier(nome);
        }

    }

    /**
     * Seta o nome do atributo declarado no declarador de ndice i, onde i pode
     * ser o ndice de um declarador de varivel j existente ou o ndice daquele
     * que se encontra mais a esquerda mais 1.
     *
     * @param name um string correspondendo ao name de um atributo.
     * @param i    posio do atributo
     */
    public void setName(String name, int i) throws
            InconsistentNodeException, IllegalArgumentException {

        if (name == null) {
            throw new IllegalArgumentException();
        }
        this.setName(JIdentifier.unwrapIdentifier(name), i);
    }

    /**
     * Seta o tipo do atributo com o valor passado como parametro
     *
     * @param t tipo do atributo
     * @throws IllegalArgumentException - se o tipo passado for nulo
     */
    public void setType(JType t) {

        if (t != null) {
            this.type = t;
        } else {
            throw new IllegalArgumentException("Tipo nulo");
        }
    }

    /**
     * Seta o tipo do atributo com o valor passado como parametro.
     *
     * @param t Um <code>String</code> correspondendo ao tipo do atributo.
     * @throws IllegalArgumentException - se o String passado for nulo
     */
    public void setType(String t) {
        JType tipo = null;
        if (t != null) {
            String nomeTipo = t.trim();
            // se a String q corresponde ao tipo, no termina com colchetes,
            // ento  um tipo sem dimenso - basta criar o name

            if(nomeTipo.charAt(nomeTipo.length() - 1) == ']'){
                int pos = nomeTipo.length() - 1;
                int numDim = 0;
                while (pos > 0 && nomeTipo.charAt(pos) == ']') {
                    pos = pos - 2;
                    numDim++;
                }
                // o indice da substring   decrementado de 2*numDim,
                // pois numDim = 1, implica em 2 colchetes, que devem ser
                // removidos do nome do tipo
                String typeName = nomeTipo.substring(0, nomeTipo.length() - 2 * numDim);
                tipo = new JType(typeName, numDim);
                this.setType(tipo);

            }
            else{
                tipo = new JType(nomeTipo, 0);
                this.setType(tipo);

            }
        } else {
            throw new IllegalArgumentException("Tipo nulo");
        }
    }


    /**
     * Retorna um cdigo inteiro correspondente ao tipo desta varivel.
     *
     * @return
     */
    public int getTypeCode() {
        return this.getType().getType();
    }

    /**
     * Seta o inicializador da varivel que se encontra na posio pos da lista
     * de variveis declaradas nesta declarao de atributo
     *
     * @param init Inicializador
     * @param pos  posio da variavel, cujo inicializador esta sendo setado
     */
    public void setInitializerAt(JaTSNode init, int pos) {

        boolean validPosition = pos >= 0 && pos < this.size();
        if (init != null && validPosition) {
            JVariableDeclarator vd =
                    (JVariableDeclarator) this.variables.elementAt(pos);
            vd.setInitializer(init);
        } else {
            throw new IllegalArgumentException("argumento invalido");
        }

    }

    public INode getInitializerAt(int pos) {
        INode retorno = null;
        boolean validPosition = pos >= 0 && pos < this.size();
        if (validPosition) {
            JVariableDeclarator vd =
                    (JVariableDeclarator) this.variables.elementAt(pos);
            retorno = vd.getInitializer();
        } else {
            throw new IllegalArgumentException("argumento invalido");
        }
        return retorno;
    }

    /**
     * Seta a lista de declaracoes de variavel
     *
     * @param varDec declaracoes de variavel
     * @throws IllegalArgumentException - se parametro for nulo
     */
    public void setVariables(JVariableDeclaratorSet varDec) {
        if (varDec != null) {
            this.variables = varDec;
        } else {
            throw new IllegalArgumentException("Lista de variaveis nula");
        }
    }

    /**
     * Retorna a quantidade de variaveis declaradas
     *
     * @return Quantidade de variaveis
     */
    public int size() {
        int retorno = 0;
        if (this.variables != null) {
            retorno = this.variables.size();
        }

        return retorno;
    }

    public JMethodInvocation getConversionMethodInvocation(JIdentifier id) {
        String method = null;

        int type = this.getTypeCode();

        switch (type) {

            case JType.SHORT:
                method = "Short.parseShort";
                break;

            case JType.BOOLEAN:
                method = "Boolean.valueOf";
                break;

            case JType.BYTE:
                method = "Byte.parseByte";
                break;

            case JType.CHAR:
                method = "Character.valueOf";
                break;

            case JType.INT:
                method = "Integer.parseInt";
                break;

            case JType.LONG:
                method = "Long.parseLong";
                break;

            case JType.FLOAT:
                method = "Float.parseFloat";
                break;

            case JType.DOUBLE:
                method = "Double.parseDouble";
                break;

        }

        JExpressionList args = new JExpressionList();
        args.addExpression(id);
        JMethodInvocation result = new JMethodInvocation(new JName(method), args);

        //tratar os casos especiais de 'char' e 'boolean'
        if (type == JType.CHAR) {
            result = new JMethodInvocation(result, new JName("charValue"), new JExpressionList());
        } else if (type == JType.BOOLEAN) {
            result = new JMethodInvocation(result, new JName("booleanValue"), new JExpressionList());
        }

        return result;
    }


    /**
     * Retorna a assinatura do n. Ou seja, o atributo que o  identifica
     * Para declaraes de atributo, a assinatura  o seu nome, para declaraes
     * de mtodo, a assinatura  o nome mais a lista de parametros.
     *
     * @return A assinatura do n
     */
    public Signature getSignature() {
        // como um n declarao de atributo, na realidade, pode conter mais
        // de um atributo declarado, pegamos apenas o identificador do primeiro
        // para ser usado como  identificador - MODIFICAR para contemplar a
        // lista de identificadores de todos os atributos na declarao
        JIdentifier result = null;
        if (this.isASourceNode()) {
            result = ((JVariableDeclarator)
                    this.variables.elementAt(0)).getIdentifier();

        }
        return result;
    }

    public boolean hasInitializer() {
        boolean retorno = false;
        retorno = this.size() == 1 &&
                ((JVariableDeclarator) this.variables.elementAt(0)).hasInitializer();

        return retorno;
    }

    /**
     * Retorna  a lista de variaveis.
     *
     * @return Lista de variaveis
     */
    public JVariableDeclaratorSet getVariables() {
        return this.variables;
    }

    /**
     * Compara dois objetos do tipo VariableDeclaration e diz se estes so
     * iguais.
     */
    public boolean equals(Object obj) {
        boolean retorno = false;
        if (obj instanceof JVariableDeclaration) {
            JVariableDeclaration field = (JVariableDeclaration) obj;
            if (this.isASourceNode()) {
                retorno = AbstractNode.equals(this.modifiers, field.modifiers) &&
                        AbstractNode.equals(this.type, field.type) &&
                        AbstractNode.equals(this.variables, field.variables);
            } else {
                retorno = compareJaTSNode(field);
            }
        }

        return retorno;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }

        if (!(node instanceof JVariableDeclaration)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JVariableDeclaration fieldDec = (JVariableDeclaration) node;

//            if (this.attributes != null) {
//            	//casa os atributos
//            	this.attributes.match(fieldDec.attributes, results);
//            } else if (fieldDec.attributes != null) {
//            	throw new NodesNotMatchedException("Attributes don't match", this, fieldDec);
//            }

            if (this.modifiers == null) {
                throw new InconsistentNodeException(this);
            }
            // casa os modificadores
            this.getModifiers().match(fieldDec.getModifiers(), results);

            if (this.getType() == null) {
                throw new InconsistentNodeException(this);
            }
            // casa o tipo dos atributos
            this.getType().match(fieldDec.getType(), results);

            // Casa os nomes dos atributos

            this.variables.match(fieldDec.getVariables(), results);
        }
    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = this;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            //Processamento dos atributos
//        	AttributeSectionList tmpAttribute = this.getAttributes();
//            if (tmpAttribute != null) {
//                Object ret = this.processNode(tmpAttribute, data);
//                if (tmpAttribute.isExecutable()) {
//                    if (!(ret instanceof AttributeSectionList)) {
//                        throw new ExecutionException(this);
//                    }
//                    this.setAttributes((AttributeSectionList) ret);
//                }
//            }

            // Processamento da lista de modificadores do atributo
            INode mods = this.getModifiers();
            if (mods != null) {
                Object ret1 = this.processNode(mods, data);
                if (mods.isExecutable()) {
                    if (!(ret1 instanceof JModifierList)) {
                        throw new ExecutionException(this);
                    }
                    this.setModifiers((JModifierList) ret1);
                }
            }

            // Processamento do tipo do atributo.
            INode fType = this.getType();
            if (fType != null) {
                Object ret2 = this.processNode(fType, data);

                if (fType.isExecutable()) {
                    try {
                        this.setType(JType.unwrapType(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }
                }
            }

            // Processamento da lista de declaradores do atributo.
            int i = 0;
            int max = this.size();

            while (i < max) {
                INode currDec = this.variables.elementAt(i);
                if (currDec != null) {
                    Object ret3 = this.processNode(currDec, data);
                    if (currDec.isExecutable()) {
                        try {
                            this.variables.setElementAt(JVariableDeclarator.unwrapVariableDeclarator(ret3),
                                    i);
                        } catch (IllegalArgumentException e) {
                            throw new ExecutionException(e.getMessage(), this);
                        }
                    }
                }
                i++;
            }
        }
        return retorno;
    }

}
