/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


/**
 * Esta classe representa um declarador de varivel, o qual consiste de um
 * identificador e, possivelmente, um inicializador para esta varivel. Assim,
 * na declaracao : <p>
 * <code>int i = 0;</code>
 * <code>i</code>  o identificador da varivel e <code>0</code>  sua
 * inicializao. <p>
 * Em :  <code>double x, y =0;</code>
 * temos dois declaradores : x e y. Onde x no possui inicializao e y possui.
 * 
 * @author Adeline de Sousa Silva
 */
public class JVariableDeclarator extends AbstractNode {

    // identificador da variavel
    private JIdentifier identifier;

    // numero de ordem - se for um array ou matriz os colchetes podem vir
    // imediatamente depois do tipo ou depois do nome
    private int dimensions;

    // inicializador da varivel
    private INode initializer;

    // coluna da tabela correspondente (para variaveis persistentes
    // em BDR)
    private String tableEntry = "MODIFICAR_AQUI";


    /**
     * Construtor vazio da classe, seta os atributos com valores default
     */
    public JVariableDeclarator() {
        this.identifier = null;
        this.dimensions = 0;
        this.initializer = null;
        this.setTypeIdentifier(JaTSNode.VARIABLE_DECLARATOR);
    }


    /**
     * Construtor da classe que recebe como argumentos o identifiacador e o a
     * dimensao
     *
     * @param id   Identificador da variavel
     * @param dim  dimensao do tipo (se nao for um array ou matriz,  sempre 0)
     * @param init Inicializacao da variavel
     */
    public JVariableDeclarator(JIdentifier id, int dim, JaTSNode init) {
        this.identifier = id;
        this.dimensions = dim;
        this.initializer = init;
        this.setTypeIdentifier(JaTSNode.VARIABLE_DECLARATOR);
    }

    /**
     * Retorna a entrada na tabela correspondente  varivel.
     * Utilizado apenas para variaveis persistentes em BDR.
     *
     * @return entrada na tabela correspondente
     */
    public String getTableEntry() {
        return tableEntry;
    }


    /**
     * Retorna a entrada na tabela correspondente  varivel, em forma de
     * Literal.
     *
     * @return entrada na tabela correspondente
     */
    public JLiteral getTableEntryAsLiteral() {
        return new JLiteral(JLiteral.STRING_LITERAL, tableEntry);
    }

    /**
     * Retorna as dimensoes da variavel (se esta nao for um array retorna zero)
     *
     * @return Dimensoes da variavel
     */
    public int getDimensions() {
        return this.dimensions;
    }


    /**
     * Retorna o identificador da variavel
     *
     * @return Identificador da variavel
     */
    public JIdentifier getIdentifier() {
        return this.identifier;
    }


    /**
     * Retorna o inicializador da variavel
     *
     * @return Inicializador da variavel
     */
    public INode getInitializer() {
        return this.initializer;
    }

    /**
     * Diz se uma varivel tem ou no inicializador.
     *
     * @return true, se a varivel tiver inicializao
     *         false, caso contrrio
     */
    public boolean hasInitializer() {
        return this.initializer != null;
    }

    /**
     * Retorna o nome  da variavel
     *
     * @return Nome da variavel
     */
    public JIdentifier getName() {
        return this.identifier;
    }

    /**
     * Remove a expresso de inicializao de um declarador de varivel.
     */
    public void removeInitializer() {
        this.initializer = null;
    }


    /**
     * Seta a entrada na tabela correspondente  varivel.
     * Utilizado apenas para variaveis persistentes em BDR.
     *
     * @param s entrada na tabela correspondente
     */
    public void setTableEntry(String s) {
        tableEntry = s;
    }


    /**
     * Seta a dimensao da variavel com o valor especificado
     *
     * @param d dimensoes do array ou da matriz (zero, caso contrario)
     * @throws IllegalArgumentException - Se o parametro passado for
     *                                  menor que zero
     */
    public void setDimensions(int d) throws IllegalArgumentException {
        if (d >= 0) {
            this.dimensions = d;
        } else {
            throw new IllegalArgumentException("Dimensao invalida (<0)");
        }
    }


    /**
     * Seta o identificador da variavel com o valor especificado
     *
     * @param id Identificador da variavel
     * @throws IllegalArgumentException - Se o parametro passado for nulo.
     */
    public void setIdentifier(JIdentifier id) throws IllegalArgumentException {
        if (id != null) {
            this.identifier = id;
        } else {
            throw new IllegalArgumentException("Identificador nulo");
        }

    }


    /**
     * Seta a inicializacao da variavel com o valor especificado
     *
     * @param init inicializador
     * @throws IllegalArgumentException - Se o parametro passado for nulo
     */
    public void setInitializer(INode init) throws IllegalArgumentException {
        if (init != null) {
            this.initializer = init;
        } else {
            throw new IllegalArgumentException("Inicializador nulo");
        }
    }

    /**
     * Compara dois objetos do tipo JVariableDeclarator e diz se estes so iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false;
        if (obj instanceof JVariableDeclarator) {
            JVariableDeclarator vd = (JVariableDeclarator) obj;
            if (!this.isVariable()) {
                retorno =
                        (AbstractNode.equals(this.identifier, vd.identifier)) &&
                        (this.dimensions == vd.dimensions) &&
                        (AbstractNode.equals(this.initializer, vd.initializer));
            } else {
                retorno = compareJaTSNode(vd);
            }
        }
        return retorno;
    }

    /**
     * Retorna o nome da varivel acrescido do prefixo <code>prefix</code>
     *
     * @return Nome da variavel
     */
    public JIdentifier addPrefix(String prefix) {
        return this.identifier.addPrefix(prefix);
    }

    /**
     * Retorna o nome da varivel acrescido do sufixo <code>suffix</code>
     *
     * @return Nome da variavel
     */
    public JIdentifier addSuffix(String suffix) {
        return this.identifier.addSuffix(suffix);
    }


    /**
     * Realiza o unwrapping de um declarador de varivel, ou seja, recebe um
     * objeto de um tipo a partir do qual um declarador de varivel  poderia ser
     * gerado e gera esse declarador. Caso isso no seja possvel, levanta uma
     * exceo.
     *
     * @param obj O objeto a partir do qual o objeto do tipo
     *            <code>JVariableDeclarator</code> dever ser gerado.
     * @return Um nome correspondente ao argumento fornecido. Por exemplo, se o
     *         argumento fornecido foi do tipo <code>JIdentifier</code>, o objeto
     *         retornado  um n do tipo <code>JVariableDeclarator</code> que
     *         tem como identificador o <code>JIdentifier</code> fornecido.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          se o argumento <code>obj</code> for <code>null</code> ou se
     *          ele for de um tipo a partir do qual no  possvel realizar a
     *          converso para um <code>JType</code>.
     */
    public static JVariableDeclarator unwrapVariableDeclarator(Object obj) throws
            IllegalArgumentException {
        JVariableDeclarator retorno = new JVariableDeclarator();
        if (obj != null) {
            // se for um JVariableDeclarator, apenas retorna
            if (obj instanceof JVariableDeclarator) {
                retorno = (JVariableDeclarator) obj;
            }
            // seno, se for um dos tipos
            else if ((obj instanceof JIdentifier) || (obj instanceof String) ||
                    (obj instanceof JName) || (obj instanceof JType)) {

                retorno.setIdentifier(JIdentifier.unwrapIdentifier(obj));
            }
        } else {
            throw new IllegalArgumentException();
        }
        return retorno;
    }


    public Object accept(IVisitor visitor, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = this;

        // Se tiver inicializador, envia o visitor pela sub-arvore
        // Esse teste  feito, por que o JVariableDeclarator, pode ou nao ter
        // uma inicializacao de variavel

        if (this.initializer != null) {
            // Envie o visitor pela sub-arvore de inicializacao
            this.initializer.accept(visitor, data);
        }
        // Envie o visitor pela sub-arvore do identificador
        if (this.identifier != null) {
            this.identifier.accept(visitor, data);
        } else if (!this.isVariable() && !this.isExecutable()) {
            throw new InconsistentNodeException(this);
        }

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore<br>
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz <br>
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns <br>
     * casam em duas situaes: quando eles so completamente idnticos <br>
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda<br>
     * no foi mapeada a nenhum valor no conjunto resultado do casamento <br>
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele<br>
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto<br>
     * acima, se refere ao n que faz parte do programa JaTS que est sendo <br>
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar<br>
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo <br>
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro<br>
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada<br>
     *                                   quando algum componente essencial do n tiver um valor, <br>
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar<br>
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou<br>
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code>  <br>
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null || results == null) {
            throw new IllegalArgumentException("JVariableDeclarator ou ResultSet nulo");
        }

        if (!(node instanceof JVariableDeclarator)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            JVariableDeclarator vd = null;
            try {
                vd = JVariableDeclarator.unwrapVariableDeclarator(node);
            } catch (IllegalArgumentException e) {
                throw new NodesNotMatchedException("Invalid node", this, node);
            }

            if (this.getDimensions() != vd.getDimensions()) {
                throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_SIZES_MSG,
                        this, node);
            }
            if (this.getIdentifier() == null) {
                throw new InconsistentNodeException(this);
            }

            // casa o identificador
            this.getIdentifier().match(vd.getIdentifier(), results);

            // se tiver inicializador, casa os inicializadores
            if (this.getInitializer() != null) {
                this.getInitializer().match(vd.getInitializer(), results);
            } else if (vd.getInitializer() != null) {
                throw new NodesNotMatchedException("Incompatible initializers",
                        this, vd);
            }
        }

    }


    /**
     *  responsvel por processar o n atual. O processamento ocorre de
     * trs maneiras diferentes, dependendo da situao:<br><br>
     * <p/>
     * 1 - <i>O n atual  executvel</i>. Nesse caso, o valor
     * retornado pelo processamento da <code>JExecutableDeclaration</code>
     * associado ao n (uma vez que todo n executvel tem um n do tipo
     * <code>JExecutableDeclaration</code> associado)  retornado pelo seu
     * mtodo process e corresponde ao resultado de seu processamento.
     * <br><br>
     * 2 - <i>O n atual  pai de um n executvel</i>. Nesse caso, o
     * processamento do n consiste em substituir o n executvel (N) em
     * sua lista de filhos pelo valor retornado pelo processamento de N.
     * Alm disso, o processamento do n inclui a responsbilidade de
     * invocar o mtodo <code>process</code> em cada um de seus filhos.
     * <br><br>
     * 3 - <i>O n atual no  executvel nem  pai de um n executvel</i>.
     * Aqui, o processamento do n se resume a invocar o mtodo
     * <code>process</code> para cada um dos filhos do n.
     * <br><br>
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Se this for executvel, simplesmente processa-se a declarao
        // executvel associada e o resultado desse processamento  retornado
        // como o resultado de processar o n.
        if (this.isExecutable()) {

            retorno = this.processExecutableNode(data);
        }
        // Caso contrrio, processa-se cada um dos filhos do n.
        else {

            // Processamento da inicializao do atributo.
            INode init = this.getInitializer();
            if (init != null) {
                Object ret1 = this.processNode(init, data);

                if (init.isExecutable() || (init instanceof JExecutableDeclaration)) {
                    if (!(ret1 instanceof INode)) {
                        throw new ExecutionException(this);
                    }
                    this.setInitializer((JaTSNode) ret1);
                }
            }

            // Processamento do identificador da varivel sendo declarada.
            INode id = this.getIdentifier();

            if (id != null) {
                Object ret2 = this.processNode(id, data);
                if (id.isExecutable()) {
                    try {
                        this.setIdentifier(JIdentifier.unwrapIdentifier(ret2));
                    } catch (Exception e) {
                        throw new ExecutionException(e.getMessage(), this);
                    }

                }
            }
        }

        return retorno;
    }

} // fim da classe
