/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import java.io.Serializable;


/**
 * Interface que define as caractersticas comuns a todos os ns na rvore
 * sinttica de um programa JaTS.
 *
 * @author Fernando Castor
 * @author Adeline Sousa
 * 
 * @version  24/01/2003
 * @since JDK 1.3
 * @see cin.jats.engine.parser.nodes.AbstractNode
 * @see cin.jats.engine.parser.nodes.JExecutableDeclaration
 */

public interface JaTSNode extends Cloneable, INode {//Serializable, {

    /*	 Constantes que correspondentes aos tipos que uma varivel pode ter. */

    // Declarao de atributo.
    public static final int FIELD_DECLARATION = 1;

    // Um conjunto de declaraes de atributos.
    public static final int FIELD_DECLARATION_SET = 2;

    // Declarao de Mtodo.
    public static final int METHOD_DECLARATION = 3;

    // Um conjunto de declaraes de mtodos.
    public static final int METHOD_DECLARATION_SET = 4;

    // Declarao de Construtor
    public static final int CONSTRUCTOR_DECLARATION = 5;

    // Um conjunto de declaraes de construtores
    public static final int CONSTRUCTOR_DECLARATION_SET = 6;

    // Uma lista de nomes genricos (nomes de interfaces, nomes de excees, etc)
    public static final int NAME_LIST = 7;

    // Uma lista de modificadores
    public static final int MODIFIER_LIST = 8;

    // Um nome. Uma lista de identificadores separados por "."
    public static final int NAME = 9;

    //  Um tipo
    public static final int TYPE = 10;

    // Um declarador de varivel. Corresponde ao nome da varivel seguido,
    // opcionalmente, por "=" e uma expresso inicializao
    public static final int VARIABLE_DECLARATOR = 11;

    // uma expressao
    public static final int EXPRESSION = 12;

    // Um identificador
    public static final int IDENTIFIER = 13;

    // Um parametro Formal. Corresponde a um tipo e um identificador
    public static final int FORMAL_PARAMETER = 14;

    // Uma lista de  parametros formais.
    public static final int PARAMETER_LIST = 15;

    // Clausula import.
    public static final int IMPORT_DECLARATION = 16;

    // Conjunto de declaracoes de importacao.
    public static final int IMPORT_DECLARATION_SET = 17;

    // Clausula package.
    public static final int PACKAGE_DECLARATION = 18;

    // Bloco - qualquer sequencia de comandos entre "{" "}"
    public static final int BLOCK = 19;

    // Lista de expresses (argumentos de um mtodo)
    public static final int EXPRESSION_LIST = 20;

    // Declaraao de tipo - classe ou interface
    public static final int TYPE_DECLARATION = 21;

    // literal, pode ser String, booleano, numrico
    public static final int LITERAL = 22;

    // Pr-condio
    public static final int PRECONDITION = 23;

    // declarao de interface
    public static final int INTERFACE_DECLARATION = 24;

    // declarao de classe
    public static final int CLASS_DECLARATION = 25;

    // statement
    public static final int STATEMENT = 26;

    // declarao de varivel local
    public static final int LOCAL_VARIABLE_DECLARATION = 27;

    // comentrios
    public static final int COMMENT = 28;

    // inicializador
    public static final int INITIALIZER = 29;

    // conjunto de inicializadores
    public static final int INITIALIZER_SET = 30;

    // conjunto de variableDeclarator
    public static final int VARIABLE_DECLARATOR_SET = 31;

    // nested class
    public static final int NESTED_CLASS_DECLARATION = 33;

    // conjunto de nested class
    public static final int NESTED_TYPE_DECLARATION_SET = 34;

    // conjunto de tipos
    public static final int TYPE_DECLARATION_SET = 35;

    // "or declaration"
    public static final int OR_DECLARATION = 36;

    // acesso a atributo
    public static final int FIELD_ACCESS = 37;

    // invocao de mtodo
    public static final int METHOD_INVOCATION = 38;

    // mapeamento atributo/entrada na tabela
    public static final int MAP_TABLE = 39;

    public static final int ATTRIBUTE = 40;

    public static final int ATTRIBUTE_ARGUMENT = 41;

    public static final int ATTRIBUTE_LIST = 42;

    public static final int ATTRIBUTE_SECTION = 43;

    public static final int PROPERTY_DECLARATION = 44;

    public static final  int PROPERTY_DECLARATION_SET = 45;
    
    public static final int ATTRIBUTE_ARGUMENT_LIST = 46;

    public static final int INDEXER_DECLARATION = 47;

    public static final int INDEXER_DECLARATION_SET = 48;

    public static final int ENUM_DECLARATION = 49;

    public static final int USING_DECLARATION = 50;
    
    public static final int ATTRIBUTE_SECTION_LIST = 51;

    
//	public static final int STRUCT_DECLARATION = 52;
//    public static final int STRUCT_DECLARATION_SET = 53;

//    public static final int DELEGATE_DECLARATION = 54 ;
//    public static final int DELEGATE_DECLARATION_SET = 55;
//    public static final int NAMESPACE_MEMBER_DECLARATION = 56;
//    public static final int NAMESPACE_MEMBER_DECLARATION_SET = 57;

    public static final int STATEMENT_LIST = 58;
    
    
    public int getLine();

    public void setLine(int line);


    /**
     * Caso o n a partir do qual este mtodo tenha sido invocado seja uma
     * varivel, este mtodo devolve um <code>String</code> correspondendo ao
     * nome da varivel que o n representa seguido pelo tipo da varivel, caso
     * este tenha sido declarado.
     *
     * @return Um <code>String</code> consistindo do nome da varivel. Caso o
     *         tipo dessa varivel tenha sido declarado tambm, o nome dela 
     *         seguido por ":" e por um <code>String</code> correspondendo a esse
     *         tipo.
     */
    public String printVariable();

    /**
     * Retorna a declarao
     *
     * @return
     */
    public INode getOrDeclaration();

    /**
     * Seta a declarao
     *
     * @param orDeclaration
     */
    public void setOrDeclaration(INode orDeclaration);

    /**
     * Diz se este n  uma declarao do tipo "OR". Ou seja, existe dois
     * valores aos quais pode ser mapeado
     */
    public boolean isOrDeclaration();


    public boolean matchesAsAVariable(INode node, ResultSet results,
                                             int varType)
                throws NodesNotMatchedException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
                IllegalArgumentException;

}
