/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.HashMap;
import java.util.Map;

/**
 * Esta classe representa um mapeamento de nomes de atributos Java em nomes de
 * atributos de uma tabela de banco de dados relacional
 * 
 * @author Ricardo Fernandes
 * @author Adeline Sousa
 *
 * @version 20/01/2004
 */
public class MapTable extends JExpression {

    /**
     * Contains the attribute X table entry mapping
     */
    private Map map;

    /**
     * Empty constructor
     */
    public MapTable() {
        super();
        this.map = new HashMap();
        this.setTypeIdentifier(JaTSNode.MAP_TABLE);
    }

    /**
     * Sets the HashMap attribute
     *
     * @param map
     */
    public void setMap(Map map) {
        this.map = map;
    }

    /**
     * Gets the HashMap attribute
     *
     * @return
     */
    public Map getMap() {
        return this.map;
    }


    /**
     * Gets the table entry name associated to a given attribute
     *
     * @param var
     * @return
     */
    public JLiteral getTableEntry(JIdentifier var) {
        JLiteral result = null;

        String attrName = var.getValue();
        if (this.map != null) {
            String tableEntryAsString = (String) map.get(attrName);
            result = new JLiteral(JLiteral.STRING_LITERAL,
                    tableEntryAsString);
        }

        return result;
    }


    /**
     * Adiciona uma entrada na tabela. Um par (atributo_Java, coluna_tabela)
     *
     * @param javaAttr    Nome do atributo no programa Java
     * @param tableColumn Nome da coluna da tabela
     */
    public void putTableEntry(JIdentifier javaAttr, JLiteral tableColumn) {
        String newVar = javaAttr.getValue();
        String newValue = tableColumn.getValue();
        this.putTableEntry(newVar, newValue);
    }


    /**
     * Adiciona uma entrada na tabela. Um par (atributo_Java, coluna_tabela)
     *
     * @param javaAttr    Nome do atributo no programa Java
     * @param tableColumn Nome da coluna da tabela
     */
    public void putTableEntry(String javaAttr, String tableColumn) {
        this.map.put(javaAttr, tableColumn);
    }


    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {
        Object retorno = this;

// Se this for executvel, simplesmente processa-se a declarao
// executvel associada e o resultado desse processamento  retornado
// como o resultado de processar o n.
        if (this.isExecutable()) {
            retorno = this.processExecutableNode(data);
        }
        return retorno;
    }

    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando
     *                de acordo com o padro de projeto homnimo, ou seja, implementa uma
     *                operao sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws ExecutionException        Levantada
     *                                   sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = this;

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        // e ento visite este n.
        retorno = visitor.visit(this, data);


        return retorno;
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico
     * quele ao qual ela est sendo mapeada agora (<code>node</code>).
     * "primeiro", no contexto acima, se refere ao n que faz parte do programa
     * JaTS que est sendo casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws IllegalArgumentException  Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem
     *                                   <code>this</code>  idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {
        if (node == null || !(node instanceof MapTable)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        } else if (!this.matchesAsAVariable(node, results, NO_TYPE)) {

            MapTable table = (MapTable) node;
            if (this.getMap() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(this);
            }
            if (!this.getMap().equals(table.getMap())) {
                throw new NodesNotMatchedException("Identifiers couldn't be " +
                        "matched (" + this.getMap() + ", " + table.getMap() + ")",
                        this, node);
            }
        }
    }


    /**
     * Retorna a String que representa este conjunto de mapeamentos
     *
     * @return String que representa este conjunto de mapeamentos
     */
    public String toString() {
        return this.map.toString();
    }

}
