/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import java.util.Hashtable;

/**
 * Esta classe representa a assintura de um mtodo. A qual  formada pelo nome
 * do mtodo e a lista de parametros
 * 
 * @author Adeline de Sousa
 * @version 28/10/2002
 */
public class MethodSignature implements Signature {

    // nome do mtodo
    private JName name;

    // lista de parametros
    private JParameterList parameters;


    /**
     * Construtor da classe
     *
     * @param name       Nome do mtodo
     * @param parameters Lista de atributos
     */
    public MethodSignature(JName name, JParameterList parameters) {
        this.setName(name);
        this.setParameters(parameters);
    }


    /**
     * Retorna o nome do mtodo
     *
     * @return Nome do mtodo
     */
    public JName getName() {
        return name;
    }


    /**
     * Seta o nome do mtodo
     *
     * @param name Nome do mtodo
     */
    public void setName(JName name) {
        if (name != null) {
            this.name = name;
        } else {
            throw new IllegalArgumentException("");
        }
    }


    /**
     * Retorna a lista de parametros
     *
     * @return Lista de parametros
     */
    public JParameterList getParameters() {
        return parameters;
    }


    /**
     * Seta a lista de parametros
     *
     * @param parameters Lista de parametros
     */
    public void setParameters(JParameterList parameters) {
        if (parameters != null) {
            this.parameters = parameters;
        } else {
            this.parameters = new JParameterList();
        }
    }

    /**
     * Returns a hash code value for the object. This method is
     * supported for the benefit of hashtables such as those provided by
     * <code>java.util.Hashtable</code>.
     * <p/>
     * The general contract of <code>hashCode</code> is:
     * <ul>
     * <li>Whenever it is invoked on the same object more than once during
     * an execution of a Java application, the <tt>hashCode</tt> method
     * must consistently return the same integer, provided no information
     * used in <tt>equals</tt> comparisons on the object is modified.
     * This integer need not remain consistent from one execution of an
     * application to another execution of the same application.
     * <li>If two objects are equal according to the <tt>equals(Object)</tt>
     * method, then calling the <code>hashCode</code> method on each of
     * the two objects must produce the same integer result.
     * <li>It is <em>not</em> required that if two objects are unequal
     * according to the {@link Object#equals(Object)}
     * method, then calling the <tt>hashCode</tt> method on each of the
     * two objects must produce distinct integer results.  However, the
     * programmer should be aware that producing distinct integer results
     * for unequal objects may improve the performance of hashtables.
     * </ul>
     * <p/>
     * As much as is reasonably practical, the hashCode method defined by
     * class <tt>Object</tt> does return distinct integers for distinct
     * objects. (This is typically implemented by converting the internal
     * address of the object into an integer, but this implementation
     * technique is not required by the
     * Java<font size="-2"><sup>TM</sup></font> programming language.)
     *
     * @return a hash code value for this object.
     * @see Object#equals(Object)
     * @see Hashtable
     */
    public int hashCode() {
        return (this.getName().hashCode() + this.getParameters().hashCode());
    }


    /**
     * Indicates whether some other object is "equal to" this one.
     * <p/>
     * The <code>equals</code> method implements an equivalence relation:
     * <ul>
     * <li>It is <i>reflexive</i>: for any reference value <code>x</code>,
     * <code>x.equals(x)</code> should return <code>true</code>.
     * <li>It is <i>symmetric</i>: for any reference values <code>x</code> and
     * <code>y</code>, <code>x.equals(y)</code> should return
     * <code>true</code> if and only if <code>y.equals(x)</code> returns
     * <code>true</code>.
     * <li>It is <i>transitive</i>: for any reference values <code>x</code>,
     * <code>y</code>, and <code>z</code>, if <code>x.equals(y)</code>
     * returns  <code>true</code> and <code>y.equals(z)</code> returns
     * <code>true</code>, then <code>x.equals(z)</code> should return
     * <code>true</code>.
     * <li>It is <i>consistent</i>: for any reference values <code>x</code>
     * and <code>y</code>, multiple invocations of <tt>x.equals(y)</tt>
     * consistently return <code>true</code> or consistently return
     * <code>false</code>, provided no information used in
     * <code>equals</code> comparisons on the object is modified.
     * <li>For any non-null reference value <code>x</code>,
     * <code>x.equals(null)</code> should return <code>false</code>.
     * </ul>
     * <p/>
     * The <tt>equals</tt> method for class <code>Object</code> implements
     * the most discriminating possible equivalence relation on objects;
     * that is, for any reference values <code>x</code> and <code>y</code>,
     * this method returns <code>true</code> if and only if <code>x</code> and
     * <code>y</code> refer to the same object (<code>x==y</code> has the
     * value <code>true</code>).
     *
     * @param obj the reference object with which to compare.
     * @return <code>true</code> if this object is the same as the obj
     *         argument; <code>false</code> otherwise.
     * @see Boolean#hashCode()
     * @see Hashtable
     */
    public boolean equals(Object obj) {
        boolean result = false;

        if (obj instanceof MethodSignature) {
            MethodSignature msign = (MethodSignature) obj;
            result = (AbstractNode.equals(this.getName(), msign.getName()) &&
                    AbstractNode.equals(this.getParameters(), msign.getParameters()));
        }

        return result;
    }


    /**
     * @return
     */
    public String toString() {
        String retorno = "";
        StringBuffer sbuffer = new StringBuffer();

        // adiciona o nome
        sbuffer.append(this.name);

        // lista de parametros
        sbuffer.append("(");
        int max = this.parameters.size() - 1;

        // se tiver algum parametro, adiciona-os separados por vrgula
        if (max > -1) {
            for (int i = 0; i < max; i++) {
                sbuffer.append(this.parameters.elementAt(i));
                sbuffer.append(",");
            }
            sbuffer.append(this.parameters.elementAt(max));
        }
        sbuffer.append(")");
        retorno = sbuffer.toString();
        return retorno;
    }


    public static void main(String[] args) {

        JName jid = new JName("match");
        JModifierList mods = new JModifierList(JModifierList.PARAMETER_DECLARATION);
        JType type = new JType("JaTSNode", 0);
        JType type2 = new JType("ResultSet", 0);
        JIdentifier jid2 = new JIdentifier("node");
        JIdentifier jid3 = new JIdentifier("results");

        JFormalParameter fp1 = new JFormalParameter(mods, type, jid2);
        JFormalParameter fp2 = new JFormalParameter(mods, type2, jid3);

        JParameterList paramList = new JParameterList();
        paramList.add(fp1);
        paramList.add(fp2);
        MethodSignature ms = new MethodSignature(jid, paramList);

        System.out.println("Assinatura do mtodo: " + ms);

    }
}
