/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

import cin.jats.engine.util.JaTSIterator;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;

import java.util.Iterator;

/**
 * Classe que guarda um conjunto de declaraes iterativas e de declaraes
 * condicionais. Uma vez processado, este n passa a guardar os ns gerados
 * pelo processamento destas declaraes em sua lista de filhos.
 * 
 * @author Fernando Castor
 */
public class OtherDeclarationsSet extends AbstractNode implements JaTSCollection {

    // Lista de declaraes aonde se encontram as declaraes resultantes do
    // processamento das declaraes de iterativeDeclarations e conditionalDeclarations.
    private NodeList resultDeclarations;;

    // Lista de declaraes iterativas
    private NodeList iterativeDeclarations;

    // Lista de declaraes condicionais
    private NodeList conditionalDeclarations;

    public OtherDeclarationsSet() {
        this.resultDeclarations = new NodeList();
        this.iterativeDeclarations = new NodeList();
        this.conditionalDeclarations = new NodeList();
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando os mtodos accept de
     * <code>this.fieldsSet</code>, <code>this.attributeSet</code> e
     * <code>this.attributesSet</code>.
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = null;

        if (this.iterativeDeclarations != null) {
            int size = iterativeDeclarations.size() - 1;
            while (size >= 0) {
                this.iterativeDeclarations.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        if (this.resultDeclarations != null) {
            int size = resultDeclarations.size() - 1;
            while (size >= 0) {
                this.resultDeclarations.elementAt(size).accept(visitor, data);
                size--;
            }
        }

        if (this.conditionalDeclarations != null) {
            int size = conditionalDeclarations.size() - 1;
            while (size >= 0) {
                this.conditionalDeclarations.elementAt(size).accept(visitor, data);
                size--;
            }
        }
        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);
        retorno = visitor.visit(this, data);
        return retorno;
    }

    /**
     * Adiciona uma declaracao  lista de declaraes iterativas de um n do
     * tipo <code>OtherDeclarationsSet</code>.
     *
     * @param dec A declaracao iterativa a ser adicionada.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>dec</code> for <code>null</code>.
     */
    public void addIterativeDeclaration(INode dec)
            throws IllegalArgumentException {

        if (dec != null) {
            this.iterativeDeclarations.addElement(dec);
        } else {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Adiciona uma declaracao  lista de outras declaraes do corpo de uma
     * classe.
     *
     * @param dec Declaracao a ser adicionada.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>dec</code> for <code>null</code>.
     */
    public void addConditionalDeclaration(INode dec)
            throws IllegalArgumentException {

        if (dec != null) {
            this.conditionalDeclarations.addElement(dec);
        } else {
            throw new IllegalArgumentException();
        }
    }

    /*************************************************************************/
    /*******************     Mtodos de JaTSCollection  **********************/
    /*************************************************************************/

    /**
     * Adiciona uma declaracao  lista de declaraes resultantes do
     * processamento das declaraes iterativas filhas de um n do tipo
     * <code>OtherDeclarationsSet</code>.
     *
     * @param dec Declaracao a ser adicionada.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>dec</code> for <code>null</code>.
     */
    public void add(INode dec)
            throws IllegalArgumentException {

        if (dec != null) {
            this.resultDeclarations.addElement(dec);
        } else {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     * <code>resultDeclarations</code>
     *
     * @param pos Um inteiro correspondendo  posio desejada.
     * @return Um objeto do tipo <code>JaTSNode</code> que se encontra na posio
     *         desejada.
     */
    public INode elementAt(int pos) {
        return (JaTSNode) this.resultDeclarations.elementAt(pos);
    }


    /**
     * Remove a declaracao que se encontra na posio <code>i</code> da lista de
     * declaraes resultantes do processamento das declaraes iterativas
     * filhas de um n do tipo <code>OtherDeclarationsSet</code>.
     *
     * @param i O ndice da declarao a ser removia.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public void removeElementAt(int i)
            throws IllegalArgumentException {

        if (i >= 0 && i < this.resultDeclarations.size()) {
            this.resultDeclarations.removeElementAt(i);
        } else {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Adiciona o elemento na posio <code>pos</code> desta lista e d um shift
     * para a direita de todos os elementos que se encontram a direita a partir
     * desta posio.
     *
     * @param elem Elemento a ser adicionado
     * @param pos  Um inteiro correspondendo  posio do n que deve ser
     *             removido.
     * @throws IllegalArgumentException Levantada quando o
     *                                  argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode elem, int pos) throws IllegalArgumentException {

        boolean validPosition = pos >= 0 && pos < this.resultDeclarations.size();
        if (validPosition) {
            this.resultDeclarations.insertElementAt(elem, pos);
        } else {
            throw new IllegalArgumentException("Invalid position");
        }
    }

    /**
     * Retorna o nmero de elementos da lista .
     *
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        return this.resultDeclarations.size();
    }


    /**
     * Diz se a lista est vazia.
     *
     * @return Um boolean que, caso tenho o valor <code>true</code>, indica
     *         que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.resultDeclarations.isEmpty();
    }


    /**
     * Retorna um objeto do tipo <code> JaTSIterator</code> .
     *
     * @return Objeto do tipo <code> JaTSIterator</code> .
     */
    public JaTSIterator getIterator() {

        Iterator it = this.resultDeclarations.iterator();
        JaTSIterator jit = new JaTSIterator(it);
        return jit;

    }

    /*************************************************************************/
    /****************** Fim dos mtodos de JaTSCollection ********************/
    /*************************************************************************/



    /**
     * Retorna a declarao iterativa que se encontra na posio <code>i</code>
     * da lista de declaraes iterativas de um <code>OtherDeclarationsSet</code>.
     * <p/>
     * param O ndice da declarao desejada.
     *
     * @return Uma declarao iterativa.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public IterativeDeclaration getIterativeDeclarationAt(int i)
            throws IllegalArgumentException, InconsistentNodeException {

        if (i < 0 || i >= this.resultDeclarations.size())
            throw new IllegalArgumentException();

        if (!(this.iterativeDeclarations.elementAt(i) instanceof IterativeDeclaration))
            throw new InconsistentNodeException(this);

        return (IterativeDeclaration) this.iterativeDeclarations.elementAt(i);
    }

    /**
     * Retorna o elemento que se encontra na posio <code>i</code> na lista de
     * "outras declaraes".
     *
     * @param i O ndice da declarao desejada.
     * @return Um objeto do tipo <code>JaTSNode</code>.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public INode getOtherDeclarationAt(int i) throws IllegalArgumentException {

        if (i < 0 || i >= this.conditionalDeclarations.size())
            throw new IllegalArgumentException();

        return this.conditionalDeclarations.elementAt(i);
    }

    /**
     * Retorna o elemento que se encontra na posio <code>i</code> na lista de
     * declaraes resultantes do processamento das declaraes iterativas.
     *
     * @param i O ndice da declarao desejada.
     * @return Um objeto do tipo <code>JaTSNode</code>.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public INode getResultDeclarationAt(int i) throws IllegalArgumentException {

        if (i < 0 || i >= this.resultDeclarations.size())
            throw new IllegalArgumentException();

        return this.resultDeclarations.elementAt(i);
    }

    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, InconsistentNodeException,
            NodesNotMatchedException {

        if (results == null) {
            throw new IllegalArgumentException("results ou node sao nulos");
        }

        // No  necessrio implementar nada aqui, pois esse conjunto no
        // existe no lado esquerdo. Portanto, no h matching

    }

    /**
     * Retorna o nmero de declaraes iterativas existente num
     * <code>OtherDeclarationsSet</code>.
     *
     * @return um inteiro.
     */
    public int numberOfIterativeDeclarations() {
        return this.iterativeDeclarations.size();
    }

    /**
     * Retorna o nmero de "outras declaraes" existente num objeto do tipo
     * <code>OtherDeclarationsSet</code>.
     *
     * @return um inteiro.
     */
    public int numberOfConditionalDeclarations() {
        return this.conditionalDeclarations.size();
    }

    /**
     * Retorna o nmero de declaraes resultantes do processamento das declaraes
     * iterativas de um <code>OtherDeclarationsSet</code>.
     *
     * @return um inteiro.
     */
    public int numberOfResultDeclarations() {
        return this.resultDeclarations.size();
    }

    /**
     * Mtodo responsvel por realizar o processamento das declaraes iterativas
     * presentes no corpo de um tipo <i>JaTS</i>. Simplesmente chama os mtodos
     * process de cada uma das declaraes iterativas que aparece no corpo da
     * classe  qual esse conjunto de declaraes iterativas se refere.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;
        Object tmpResult = null;

        // Processamento da lista de declaraes iterativas.
        int i = this.iterativeDeclarations.size() - 1;

        while (i >= 0) {
            INode currIterDec = this.iterativeDeclarations.elementAt(i);

            if (!(currIterDec instanceof IterativeDeclaration)) {
                throw new InconsistentNodeException(this);
            }

            // Invoca o mtodo que realiza o processamento em currItemDec e
            // coloca os resultados do processamento em this.
            tmpResult = currIterDec.process(data);

            if (tmpResult instanceof INode) {
                INode jaTSNode = (INode) tmpResult;
                this.resultDeclarations.addElement(jaTSNode);
            } else if (tmpResult instanceof NodeList) {
                this.resultDeclarations.addAll((NodeList) tmpResult);
            }

            this.iterativeDeclarations.removeElementAt(i);

            i--;
        }

        int j = this.conditionalDeclarations.size() - 1;
        while (j >= 0) {
            INode currDec = this.conditionalDeclarations.elementAt(j);
            if (currDec != null) {
                // Invoca o mtodo que realiza o processamento em currItemDec e
                // coloca os resultados do processamento em this.
                tmpResult = currDec.process(data);
                if (tmpResult instanceof INode) {
                    INode jaTSNode = (INode) tmpResult;
                    this.resultDeclarations.addElement(jaTSNode);
                } else if (tmpResult instanceof NodeList) {
                    this.resultDeclarations.addAll((NodeList) tmpResult);
                }

                this.conditionalDeclarations.removeElementAt(j);
            }
            j--;
        }

        return retorno;
    }


    /**
     * Remove a declaracao iterativa que se encontra na posicao <code>i</code>
     * na lista de declaraes iterativas de um n do tipo
     * <code>OtherDeclarationsSet</code>.
     *
     * @param i O ndice da declarao iterativa a ser removida.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public void removeIterativeDeclarationAt(int i)
            throws IllegalArgumentException {

        if (i >= 0 && i < this.iterativeDeclarations.size()) {
            this.iterativeDeclarations.removeElementAt(i);
        } else {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Remove a declaracao que se encontra na posio <code>i</code> da lista de
     * "outras declaraes" do corpo de uma classe.
     *
     * @param i O ndice da declarao a ser removia.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>i</code> for menor que zero ou maior ou igual ao nmero
     *                                  de elementos da lista de declaraes.
     */
    public void removeOtherDeclarationAt(int i)
            throws IllegalArgumentException {

        if (i >= 0 && i < this.conditionalDeclarations.size()) {
            this.conditionalDeclarations.removeElementAt(i);
        } else {
            throw new IllegalArgumentException();
        }
    }


}
