/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes;

/**
 * Esta classe guarda as informacoes a respeito do corpo de uma classe.
 * 
 * @author Adeline de Sousa
 * @author Fernando Castor
 * @version  26/12/2002
 * @since JDK 1.3
 */
import cin.jats.engine.util.ResultSet;
import cin.jats.engine.visitors.IVisitor;
import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;


public class TypeBody extends AbstractNode {


    // Guarda as declaracoes de construtores
    private ClassBodyConstructorDeclarations constructorsDecSet;

    // Guarda as declaracoes de atributo
    private BodyFieldDeclarations fieldsDecSet;

    // Guarda as declaraes iterativas declaradas no corpo da classe, assim
    // como seus resultados
    private OtherDeclarationsSet otherDecsSet;

    // Guarda as declaracoes de metodo
    private BodyMethodDeclarations methodDecsSet;

    // lista de inicializadores
    private ClassBodyInitializers initializerSet;

    // lista de tipos aninhados (nested classes e nested interfaces)
    private BodyTypeDeclarations nestedTypeDecSet;
    

    /**
     * Construtor vazio da classe
     */
    public TypeBody() {
        this.constructorsDecSet = new ClassBodyConstructorDeclarations();
        this.fieldsDecSet = new BodyFieldDeclarations();
        this.otherDecsSet = new OtherDeclarationsSet();
        this.methodDecsSet = new BodyMethodDeclarations();
        this.initializerSet = new ClassBodyInitializers();
        this.nestedTypeDecSet = new BodyTypeDeclarations();
    }

    /**
     * Construtor da classe, que recebe como parametro o conjunto de
     * declaracoes de atributo, assim como o conjunto de declaraes iterativas
     * que aparecem no corpo da classe.
     *
     * @param fields Conjunto de declaracoes de atributo
     * @param idecs  Conjunto de declaraes iterativas, condicionais
     */
    public TypeBody(BodyFieldDeclarations fields, OtherDeclarationsSet idecs,
                    BodyMethodDeclarations methods, ClassBodyConstructorDeclarations constructors) {
        this.setFieldDeclarationSet(fields);
        this.setConstructorDeclarationSet(constructors);
        this.setOtherDeclarationsSet(idecs);
        this.setMethodDeclarationSet(methods);
        this.initializerSet = new ClassBodyInitializers();
        this.nestedTypeDecSet = new BodyTypeDeclarations();
    }


    /**
     * Mtodo responsvel por receber um <code>visitor</code>, aplic-lo ao n
     * atual e envi-lo atravs da rvore, invocando o mtodo accept de
     * cada um de seus filhos
     *
     * @param visitor Um objeto do tipo <code>IVisitor</code>, funcionando de acordo
     *                com o padro de projeto homnimo, ou seja, implementa uma operao
     *                sobre vrios tipos diferentes de objetos fora do cdigo desses
     *                objetos. A partir dele ser invocado o mtodo <code>visit</code>,
     *                que recebe como argumento <code>this</code>.
     * @param data    Um objeto qualquer que ser utilizado pelo visitor.
     * @return Um objeto qualquer que possa vir a ser retornado pelo visitor.
     * @throws java.lang.IllegalArgumentException
     *          Levantada quando o parmetro
     *          qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada
     *          quando algum componente essencial do n tiver um valor,
     *          tornando o n, como um todo, inconsistente.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Levantada
     *          sempre que a execuo de um n falhar.
     */
    public Object accept(IVisitor visitor, Object data)
            throws InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        Object retorno = this;

        if (this.nestedTypeDecSet != null) {
            this.nestedTypeDecSet.accept(visitor, data);
        }

        if (this.initializerSet != null) {
            this.initializerSet.accept(visitor, data);
        }
        if (this.fieldsDecSet == null) {
            throw new InconsistentNodeException(this);
        }
        this.fieldsDecSet.accept(visitor, data);

        if (this.otherDecsSet == null) {
            throw new InconsistentNodeException(this);
        }
        this.otherDecsSet.accept(visitor, data);

        if (this.methodDecsSet == null) {
            throw new InconsistentNodeException(this);
        }
        this.methodDecsSet.accept(visitor, data);

        if (this.constructorsDecSet == null) {
            throw new InconsistentNodeException(this);
        }
        this.constructorsDecSet.accept(visitor, data);

        // ... invoque o accept da declarao executvel associada a este n,
        // caso haja alguma...
        this.invokeAcceptOnExecutableDeclaration(visitor, data);

        // E ento visite este n.
        retorno = visitor.visit(this, data);

        return retorno;
    }


    /**
     * Retorna o Conjunto de declaracoes de atributo
     *
     * @return Conjunto de declaracoes de atributo
     */
    public BodyFieldDeclarations getFieldDeclarationSet() {

        return this.fieldsDecSet;
    }

    /**
     * Retorna o Conjunto de declaracoes de construtor
     *
     * @return Conjunto de declaracoes de construtor
     */
    public ClassBodyConstructorDeclarations getConstructorDeclarationSet() {

        return this.constructorsDecSet;
    }

    /**
     * Retorna o Conjunto de declaracoes de mtodo
     *
     * @return Conjunto de declaracoes de mtodo
     */
    public BodyMethodDeclarations getMethodDeclarationSet() {

        return this.methodDecsSet;
    }

    /**
     * Retorna o Conjunto de declaracoes iterativas
     *
     * @return Conjunto de declaracoes iterativas
     */
    public OtherDeclarationsSet getOtherDeclarationsSet() {

        return this.otherDecsSet;
    }

    /**
     * Retorna os inicializadores presentes no corpo da classe
     *
     * @return Inicializadores da classe
     */
    public ClassBodyInitializers getInitializerSet() {
        return this.initializerSet;
    }

    /**
     * Retorna o conjunto de tipos aninhados (inner classes ou nested
     * interfaces) que esto presentes no corpo da classe
     *
     * @return
     */
    public BodyTypeDeclarations getNestedTypeSet() {
        return this.nestedTypeDecSet;
    }


    /**
     * Este mtodo realiza o casamento entre este n, que faz parte da rvore
     * sinttica de um programa JaTS, e o n fornecido como argumento, que faz
     * parte da rvore sinttica de um programa em Java. Para JaTS, dois ns
     * casam em duas situaes: quando eles so completamente idnticos
     * e quando o primeiro deles  uma varivel JaTS e, ou essa varivel ainda
     * no foi mapeada a nenhum valor no conjunto resultado do casamento
     * (<code>results</code>) ou o valor ao qual ela foi mapeada  idntico quele
     * ao qual ela est sendo mapeada agora (<code>node</code>). "primeiro", no contexto
     * acima, se refere ao n que faz parte do programa JaTS que est sendo
     * casado.
     *
     * @param node    O objeto do tipo <code>JaTSNode</code> com o qual
     *                <code>this</code> ser casado.
     * @param results O conjunto resultado do casamento.  usado para armazenar
     *                os pares <code>(varivel,cdigo)</code> que so gerados pelo
     *                casamento.
     * @throws java.lang.IllegalArgumentException
     *                                   Levantada quando o parmetro
     *                                   qualquer um dos argumentos tiver valor <code>null</code>.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *                                   quando algum componente essencial do n tiver um valor,
     *                                   tornando o n, como um todo, inconsistente.
     * @throws NodesNotMatchedException  Lanada sempre que o casamento falhar
     *                                   porque <code>node</code> no  um n compatvel com <code>this</code>, ou
     *                                   seja, nem <code>this</code>  uma varivel no casada nem <code>this</code> 
     *                                   idntico a <code>this</code>.
     */
    public void match(INode node, ResultSet results)
            throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
            NodesNotMatchedException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!(node instanceof TypeBody)) {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_TYPES_MSG,
                    this, node);
        }


        TypeBody body = (TypeBody) node;

        // Declaraes de atributo
        if (this.fieldsDecSet == null || this.methodDecsSet == null
                || this.otherDecsSet == null) {
            throw new InconsistentNodeException(this);
        }

        this.fieldsDecSet.match(body.getFieldDeclarationSet(), results);
        this.constructorsDecSet.match(body.getConstructorDeclarationSet(),
                results);

        // Declaraes iterativas
        // Ns com declaraes iterativas NUNCA so casados.
        // Elas fazem parte exclusivamente do lado direito de transformaes.
        // Mesmo os ns gerados por uma declarao iterativa no podem ser
        // casados. A rvore sinttica de um tipo contendo algum n resultante
        // do processamento de uma declarao iterativa deve ser pretty-printed
        // e, em seguida, re-parsed.

        if (this.otherDecsSet.numberOfIterativeDeclarations() > 0 ||
                this.otherDecsSet.numberOfResultDeclarations() > 0) {
            throw new NodesNotMatchedException("Iterative declarations " +
                    "should not be matched", this, node);
        }

        this.methodDecsSet.match(body.getMethodDeclarationSet(), results);

        this.initializerSet.match(body.getInitializerSet(), results);

        this.nestedTypeDecSet.match(body.getNestedTypeSet(), results);


    }

    /**
     *  responsvel por processar o n atual. O processamento de ns do tipo
     * TypeBody se resume a invocar o mtodo <code>process</code> para cada um
     * dos filhos do n.
     *
     * @param data Um objeto que pode conter informaes necessrias 
     *             execuo.
     * @return O resultado do processamento do n.
     * @throws cin.jats.engine.parser.nodes.exceptions.ExecutionException
     *          Se algum erro
     *          ocorrer durante o processamento das estruturas executveis do
     *          programa.
     * @throws java.lang.IllegalArgumentException
     *          Levantada
     *          caso o parmetro <code>data</code> seja <code>null</code> numa
     *          situao em que ele  necessrio.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *          Levantada se um n inconsistente (incompleto) for encontrado.
     *          Se algum dos componentes (filhos) essenciais ao correto
     *          funcionamento do n estiver faltando.
     */
    public Object process(Object data) throws ExecutionException,
            InconsistentNodeException, IllegalArgumentException {

        Object retorno = null;

        // Ns do tipo TypeBody nunca podem ser executveis. O mesmo se aplica
        // aos ns dos tipos BodyFieldDeclarations, BodyMethodDeclarations, etc.
        // Por isso a funo process deste n  responsvel apenas por
        // encaminhamento.

        INode fDecSet = this.getFieldDeclarationSet();
        if (fDecSet != null) {
            this.processNode(fDecSet, data);
        }

        INode oDecSet = this.getOtherDeclarationsSet();
        if (oDecSet != null) {
            this.processNode(oDecSet, data);
        }

        INode mDecSet = this.getMethodDeclarationSet();
        if (mDecSet != null) {
            this.processNode(mDecSet, data);
        }

        INode cDecSet = this.getConstructorDeclarationSet();
        if (cDecSet != null) {
            this.processNode(cDecSet, data);
        }

        INode initSet = this.getInitializerSet();
        if (initSet != null) {
            this.processNode(initSet, data);
        }


        return retorno;
    }

    /**
     * Seta o conjunto de declaracoes de atributos da classe
     *
     * @param fields Conjunto de declaracoes de atributo
     * @throws IllegalArgumentException - Se o conjunto de declaracoes for
     *                                  nulo
     */
    public void setFieldDeclarationSet(BodyFieldDeclarations fields) {

        if (fields != null) {
            this.fieldsDecSet = fields;
        } else {
            throw new IllegalArgumentException("Conjunto de declaracoes nulo");
        }

    }

    /**
     * Seta o conjunto de declaracoes de construtor da classe
     *
     * @param constructors Conjunto de declaracoes de construtor
     * @throws IllegalArgumentException - Se o conjunto de declaracoes for
     *                                  nulo
     */
    public void setConstructorDeclarationSet(ClassBodyConstructorDeclarations
            constructors) {

        if (constructors != null) {
            this.constructorsDecSet = constructors;
        } else {
            throw new IllegalArgumentException("Conjunto de declaracoes  nulo");
        }

    }

    /**
     * Seta o conjunto de declaracoes de mtodos da classe
     *
     * @param methods Conjunto de declaracoes de mtodo
     * @throws IllegalArgumentException - Se o conjunto de declaracoes for
     *                                  nulo
     */
    public void setMethodDeclarationSet(BodyMethodDeclarations methods) {

        if (methods != null) {
            this.methodDecsSet = methods;
        } else {
            throw new IllegalArgumentException("Conjunto de declaracoes nulo");
        }

    }

    /**
     * Seta o conjunto de declaracoes iterativas da classe
     *
     * @param decs Conjunto de declaracoes iterativas
     * @throws IllegalArgumentException - Se o conjunto de declaracoes for
     *                                  nulo
     */
    public void setOtherDeclarationsSet(OtherDeclarationsSet decs) {
        if (decs != null) {
            this.otherDecsSet = decs;
        } else {
            throw new IllegalArgumentException();
        }

    }

    /**
     * Seta os inicializadores da classe. O NodeList passado deve conter apenas
     * objetos do tipo <code>Initializer</code>
     *
     * @param inits Inicializadores da classe
     */
    public void setInitializerSet(ClassBodyInitializers inits) {
        if (inits != null) {
            this.initializerSet = inits;
        }
    }

    /**
     * Seta os inicializadores da classe. O NodeList passado deve conter apenas
     * objetos do tipo <code>Initializer</code>
     *
     * @param types Inicializadores da classe
     */
    public void setTypeDeclarationSet(BodyTypeDeclarations types) {
        if (types != null) {
            this.nestedTypeDecSet = types;
        }
    }


    /**
     * Adiciona uma declaracao de atributo ao conjunto fieldsDecSet
     *
     * @param fieldDec Declaracao de atributo a ser adicionada
     */
    public void addFieldDeclaration(JVariableDeclaration fieldDec)
            throws IllegalArgumentException {
        this.fieldsDecSet.addFieldDeclaration(fieldDec);
    }

    /**
     * Adiciona uma declaracao do tipo "#f:FieldDeclaration" ao conjunto
     * fieldsDecSet
     *
     * @param field Declaracao a ser adicionada
     */
    public void addFieldVarDeclaration(JFieldDeclaration field)
            throws IllegalArgumentException {
        this.fieldsDecSet.addFieldVarDeclaration(field);
    }

    /**
     * Adiciona uma declaracao '#fds:FieldDeclarationSet' ao conjunto fieldsDecSet
     *
     * @param atts Declaracao '#fds:FieldDeclarationSet' a ser adicionada
     */
    public void addFieldsVarDeclaration(JFieldDeclarationSet atts)
            throws IllegalArgumentException {
        this.fieldsDecSet.addFieldsVarDeclaration(atts);
    }


    /**
     * Adiciona uma declaracao de metodo ao conjunto methodDecsSet
     *
     * @param methDec Declaracao de metodo a ser adicionada
     */
    public void addMethodDeclaration(JMethodDeclaration methDec)
            throws IllegalArgumentException {
        this.methodDecsSet.addMethodDeclaration(methDec);
    }


    /**
     * Adiciona uma declaracao 'mtd' ao conjunto methodDecsSet
     *
     * @param mtd Declaracao 'mtd' a ser adicionada
     */
    public void addMethodVarDeclaration(JMethodDeclaration mtd)
            throws IllegalArgumentException {
        this.methodDecsSet.addMethodVarDeclaration(mtd);
    }


    /**
     * Adiciona uma declaracao 'mtds' ao conjunto methodDecsSet
     *
     * @param mtds Declaracao 'mtds' a ser adicionada
     */
    public void addMethodsVarDeclaration(JMethodDeclarationSet mtds)
            throws IllegalArgumentException {
        this.methodDecsSet.addMethodsVarDeclaration(mtds);
    }


    /**
     * Adiciona uma declaracao de construtor ao conjunto constructorsDecSet
     *
     * @param constructorDec Declaracao de construtor a ser adicionada
     */
    public void addConstructorDeclaration(JConstructorDeclaration constructorDec)
            throws IllegalArgumentException {

        if (constructorDec != null) {
            this.constructorsDecSet.addConstructorDeclaration(constructorDec);
        } else {
            throw new IllegalArgumentException("constructorDec nulo");
        }
    }

    /**
     * Adiciona uma declaracao '#cds: ConstructorDeclarationSet' ao conjunto constructorsDecSet
     *
     * @param constructors Declaracao '#cds: ConstructorDeclarationSet' a ser adicionada
     */
    public void addConstructorsVarDeclaration(JConstructorDeclarationSet constructors)
            throws IllegalArgumentException {
        this.constructorsDecSet.addConstructorsVarDeclaration(constructors);
    }


    /**
     * Adiciona uma declaracao do tipo "#cd:ConstructorDeclaration" ao conjunto
     * constructorsDecSet
     *
     * @param constructor Declaracao a ser adicionada
     */
    public void addConstructorVarDeclaration(JConstructorDeclaration constructor)
            throws IllegalArgumentException {

        this.constructorsDecSet.addConstructorVarDeclaration(constructor);
    }


    /**
     * Adiciona uma declaracao  lista de declaraes iterativas de um n do
     * tipo <code>OtherDeclarationsSet</code>.
     *
     * @param dec A declaracao iterativa a ser adicionada.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>dec</code> for <code>null</code>.
     */
    public void addIterativeDeclaration(INode dec)
            throws IllegalArgumentException {
        this.otherDecsSet.addIterativeDeclaration(dec);
    }


    /**
     * Adiciona uma declaracao  lista de outras declaraes do corpo de uma
     * classe.
     *
     * @param dec Declaracao a ser adicionada.
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>dec</code> for <code>null</code>.
     */
    public void addOtherDeclaration(INode dec) {
        this.otherDecsSet.addConditionalDeclaration(dec);
    }


    /**
     * Adiciona um inicializador  lista de inicializadores do corpo de uma
     * classe.
     *
     * @param init Inicializador a ser adicionado
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>init</code> for <code>null</code>.
     */
    public void addInitializer(JInitializer init) {
        this.initializerSet.addInitializer(init);
    }

    /**
     * Adiciona uma variavel do tipo  #init:Initializer   lista de
     * inicializadores do corpo de uma classe.
     *
     * @param init Inicializador a ser adicionado
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>init</code> for <code>null</code>.
     */
    public void addInitializerVar(JInitializer init) {
        this.initializerSet.addInitializerVar(init);
    }

    /**
     * Adiciona uma variavel do tipo  #init:InitializerSet  lista de
     * inicializadores do corpo de uma classe.
     *
     * @param initSet variavel  a ser adicionada
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>initSet</code> for <code>null</code>.
     */
    public void addInitializerSetVar(JInitializerSet initSet) {
        this.initializerSet.addInitializersSetVar(initSet);
    }

    /**
     * Adiciona uma declarao de tipo aninhado ao corpo deste tipo
     *
     * @param type Tipo a ser adicionado
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>initSet</code> for <code>null</code>.
     */
    public void addTypeDeclaration(ITypeDeclaration type) {
        this.nestedTypeDecSet.addTypeDeclaration(type);
    }


    /**
     * Adiciona uma variavel do tipo  #type:classDeclaration ou
     * # type:InterfaceDeclaration  lista de tipos aninhadas do corpo de uma
     * classe.
     *
     * @param type variavel  a ser adicionada
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>initSet</code> for <code>null</code>.
     */
    public void addTypeVarDeclaration(JTypeDeclaration type) {
        this.nestedTypeDecSet.addTypeVarDeclaration(type);
    }


    /**
     * Adiciona uma variavel do tipo  #type:classDeclarationSet ou
     * # type:InterfaceDeclarationSet  lista de tipos aninhadas do corpo de
     * uma classe.
     *
     * @param typeSet variavel  a ser adicionada
     * @throws IllegalArgumentException Levantada se o valor de
     *                                  <code>initSet</code> for <code>null</code>.
     */
    public void addTypeVarsDeclaration(TypeDeclarationSet typeSet) {
        this.nestedTypeDecSet.addTypesVarDeclaration(typeSet);
    }
}
