/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.parser.nodes.exceptions;

import cin.jats.engine.parser.nodes.INode;
import cin.jats.engine.parser.nodes.JaTSNode;


/**
 * Esta classe define uma exceo que  levantada sempre que o casamento entre
 * dois ns falha.
 *
 * @author Fernando Castor
 * @see cin.jats.engine.parser.nodes.JaTSNode
 */
public class NodesNotMatchedException extends Exception {


    // O primeiro dos ns que no casou.
    private INode firstNode;


    // O segundo dos ns que no casou.
    private INode secondNode;

    // Uma varivel que guarda a partir de que ponto a exceo foi lanada no
    // corpo de um mtodo <code>match</code>, caso seja possvel que ela tenha
    // sido lanada de mais de um lugar.
    private int occurrence;



    // futuramente, sero lidas de um arquivo de msgs
    public static final String NODES_IS_NULL_MSG = " Node received is null ";
    public static final String NODES_HAVE_DIFFERENT_TYPES_MSG = " Nodes have different types ";
    public static final String NODES_HAVE_DIFFERENT_SIZES_MSG = " Nodes have different sizes ";
    public static final String NODES_ARE_INCOMPATIBLE_MSG = " Nodes are incompatible ";
    public static final String INCONSISTENT_NODE = "Node is inconsistent";

    /**
     * Construtor vazio da classe.
     */
    public NodesNotMatchedException() {
        firstNode = null;
        secondNode = null;
        occurrence = 0;
    }

    /**
     * Construtor que recebe apenas o par de ns que no casaram.
     *
     * @param fstNode O primeiro dos ns que no casou.
     * @param sndNode O segundo dos ns que no casou.
     */
    public NodesNotMatchedException(INode fstNode, INode sndNode) {
        firstNode = fstNode;
        secondNode = sndNode;
    }

    /**
     * Construtor que recebe apenas o par de ns que no casaram.
     *
     * @param fstNode O primeiro dos ns que no casou.
     * @param sndNode O segundo dos ns que no casou.
     * @param occ     Em qual das possibilidades (quando mais de uma  possvel),
     *                esta exceo foi lanada no mtodo <code>match</code>0 de <code>fstNode</code>.
     */
    public NodesNotMatchedException(INode fstNode, INode sndNode, int occ) {
        firstNode = fstNode;
        secondNode = sndNode;
        this.occurrence = occ;
    }

    /**
     * Este construtor recebe como parmetro apenas uma mensagem de erro.
     *
     * @param message Uma mensagem de erro indicando qual foi o problema
     *                encontrado.
     */
    public NodesNotMatchedException(String message) {
        super(message);
        firstNode = null;
        secondNode = null;
    }

    /**
     * Este construtor recebe, alm da mensagem de erro, os dois ns que no
     * casaram.
     *
     * @param mensagem Uma mensagem de erro indicando qual foi o problema
     *                 encontrado.
     * @param fstNode  O primeiro dos ns que no casou.
     * @param sndNode  O segundo dos ns que no casou.
     */
    public NodesNotMatchedException(String mensagem, INode fstNode,
                                    INode sndNode) {
        super(mensagem);
        firstNode = fstNode;
        secondNode = sndNode;
    }

    /**
     * Retorna o n a partir do qual o mtodo <code>match</code> foi originalmente
     * chamado. Este  o n a partir da qual a exceo foi lanada.
     *
     * @return Um objeto do tipo <code>JaTSNode</code> correspondendo ao n a
     *         partir do qual a exceo foi lanada. Caso este no tenha sido
     *         fornecido como parmetro ao construtor da exceo, este mtodo
     *         retorna <code>null</code>;
     */
    public INode getFirstNode() {
        return this.firstNode;
    }

    /**
     * Retorna o ndice do local de onde a exceo foi lanada.
     *
     * @return Um inteiro correspondendo ao ndice do local de onde a exceo foi
     *         lanada. Essa informao  necessria quando, em um mtodo <code>match</code>,
     *         a exceo <code>NodesNotMatchedException</code>  lanada de mais de um lugar.
     *         Para que os erros possam ser reportados de forma precisa, 
     *         includa a informao sobre de qual, dos vrios locais possveis,
     *         a exceo foi lanada.
     */
    public int getOccurrence() {
        return this.occurrence;
    }

    /**
     * Retorna o segundo dos ns que no casaram.
     *
     * @return Um objeto do tipo <code>JaTSNode</code> correspondendo ao segundo
     *         dos ns que no casaram. Caso este no tenha sido fornecido como
     *         parmetro ao construtor da exceo, este mtodo retorna <code>null</code>;
     */
    public INode getSecondNode() {
        return this.secondNode;
    }

    /**
     * Returns the error message string of this throwable object.
     *
     * @return the error message string of this <code>Throwable</code>
     *         object if it was {@link #Throwable(String) created} with an
     *         error message string; or <code>null</code> if it was
     *         {@link #Throwable() created} with no error message.
     */
    public String getMessage() {
        String message = "";
        try {
            if (firstNode != null && secondNode != null) {
                message = super.getMessage();
                message += "\nNODE 1: \n" + firstNode.getClass().getName() + ": " +
                        firstNode + " \nAt line: " + ((JaTSNode)firstNode).getLine() +
                        " \nNODE 2: \n" + secondNode.getClass().getName() + ": " +
                        secondNode;
            } else {
                message = super.getMessage();
                message += "\nNODE 1: \n" + firstNode + "\nNODE 2: \n" +
                        secondNode;
            }

        }

        catch (Exception e) {
            System.out.println("Excecao "+ e.getClass().getName());

        }
        return message;
    }
}
