/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.processor;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.util.List;

import cin.jats.engine.JatsIO;
import cin.jats.engine.ParserFactory;
import cin.jats.engine.PrettyPrinter;
import cin.jats.engine.parser.ParseException;
import cin.jats.engine.parser.nodes.INode;
import cin.jats.engine.parser.nodes.JCompilationUnit;
import cin.jats.engine.parser.nodes.JMethodDeclaration;
import cin.jats.engine.parser.nodes.JTryStatement;
import cin.jats.engine.parser.nodes.JTypeDeclaration;
import cin.jats.engine.parser.nodes.Pattern;
import cin.jats.engine.parser.nodes.metajats.JaTSProgram;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.visitors.PreProcessorVisitor;

/**
 * The processor orchestrates the transformation process of JaTS meta-programs
 * into pure Java programs + patterns files. This class is also responsible for 
 * runnig meta-programs.
 * 
 * @author Gustavo Santos
 * @version 12/10/05
 */
public class Processor {
	
	private JatsIO jatsio;
	private String root;

	/**
	 * Constructor for Processor objects
	 */
	public Processor() {
		jatsio = JatsIO.getInstance();
	}

	/**
	 * This method executes an specified transformation. It must be called
	 * ALWAYS after a call to processProgram method. The program (transformation)
	 * to be executed is not recieved as a parameter. It is the result of a
	 * processProgram call.
	 * 
	 * @param classPath The required class path to run JaTS and the transformation
	 * @throws IOException
	 * @throws InterruptedException
	 */
	public void runProgram(String classPath) throws IOException, InterruptedException {
			Process p = null;
			p = Runtime.getRuntime().exec("javac -classpath "+classPath+" "+root+File.separator+"Program.java");
			p.waitFor();
			if (p.exitValue() != 0) {
				String msg = mountMsg(p.getErrorStream());								
				throw new RuntimeException("Compile error:\n"+msg);
			}
			p = Runtime.getRuntime().exec("java -cp "+classPath+" Program");
			
			String error = mountMsg(p.getErrorStream());
			
			//by now, this call only flushes the stream, 
			//otherwise a deadlock happens
			mountMsg(p.getInputStream());
            
			int exitVal = p.waitFor();
            
			if (exitVal != 0 || !"".equals(error)) { 
				throw new RuntimeException("Error:\n"+error);
			}
	}

	/**
	 * Generates a String object from an InputStream object
	 * 
	 * @param is An InputStream
	 * @return The generated String
	 * @throws IOException
	 */
	private String mountMsg(InputStream is) throws IOException {
        String msg = "";
		InputStreamReader isr = new InputStreamReader(is);
        BufferedReader br = new BufferedReader(isr);
        String line = null;
        while ( (line = br.readLine()) != null)
            msg += line+"\n";

		return msg;
	}
	
	
	/**
	 * Transforms a JaTS meta-program into a set of pattern files
	 * and pure Java programs.
	 * 
	 * @param path Program directory absolute path
	 * @param prog Parogram name
	 * @throws IllegalArgumentException
	 * @throws ParseException
	 * @throws IOException
	 * @throws InterruptedException 
	 */
	public void processProgram(String classPath, String prog)
			throws IllegalArgumentException, ParseException, IOException, InterruptedException {
		this.root = prog.substring(0, prog.lastIndexOf(File.separator));
		Reader fileReader = new FileReader(prog);
		jatsio.changeParserType(ParserFactory.JATS_PARSER);
		JaTSProgram metProgram = jatsio.parseMetaProgram(fileReader);
		process(metProgram);
		
		Process p = null;
		p = Runtime.getRuntime().exec("javac -classpath "+classPath+" "+root+File.separator+"Program.java");
		p.waitFor();
		if (p.exitValue() != 0) {
			String msg = mountMsg(p.getErrorStream());								
			throw new RuntimeException("Compile error:\n"+msg);
		}
		
	}
	

	/**
	 * This method is called by processProgram(). It recieves a parsed
	 * meta-program and process it.
	 * 
	 * @param metaProgram Meta-program tree
	 * @throws IOException
	 */
	private void process(JaTSProgram metaProgram) throws IOException {		
        Writer out = null;
        INode program = mountProgram(metaProgram);
		PreProcessorVisitor visitor = new PreProcessorVisitor();
        
		try {
			program = (INode) program.accept(visitor, null);

			List patterns = visitor.getPatterns();
			
			for (int i = 0; i < patterns.size(); i++) {
				Pattern p = (Pattern)patterns.get(i);
				out = new FileWriter(this.root+File.separator+p.getName().getValue().replace("#","")+".jats");
				out.write(p.getFlatPattern());
				out.flush();
				out.close();
			}

			PrettyPrinter pp = PrettyPrinter.getInstance();
			out = new FileWriter(this.root+File.separator+"Program.java");
			pp.prettyPrint(program, out);
	        out.flush();
	        out.close();			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}

	/**
	 * Generates the Java program.
	 * 
	 * @param metaProgram
	 * @return A Java program tree
	 */
	private INode mountProgram(JaTSProgram metaProgram) {
		INode program = mountBaseClass();
		
		JCompilationUnit c = (JCompilationUnit)program;
		JMethodDeclaration m = (JMethodDeclaration) ((JTypeDeclaration)c.getTypeDeclaration(0)).getBody().getMethodDeclarationSet().getMethodsSet().elementAt(0);
		NodeList nl = new NodeList();
		
		for (int i = 0; i < metaProgram.size(); i++) {
			nl.addElement(metaProgram.statementAt(i));
		}
		
		JTryStatement t = (JTryStatement)m.getBody().getStatements().elementAt(4);
		t.getTryBlock().getStatements().getStatements().addAll(nl);
		
		return program;
	}
	
	/**
	 * Generates the standard code generated for all meta-programs.
	 * 
	 * @return A Java program tree
	 */
	private INode mountBaseClass() {
		INode result = null;
		String s1 = "\\\\", s2 = "\\\\\\\\";
		String sep = File.separator;
		if (sep.equals("\\"))
			sep += sep;
		
		//if running on a unix like OS, the backslashes 
		//are not replaced
		String g = this.root.replaceAll(s1,s2);
		String code =
			"import java.io.*;"+
			"import cin.jats.engine.parser.nodes.*;"+
			"import cin.jats.engine.*;"+
			"import cin.jats.engine.util.*;"+
			"import cin.jats.engine.transformation.*;"+
		    "public class Program {" +
			"    private static Writer out = null;"+
			"    public static void main(String[] args) {" +
		    "        String root = \"" + g + sep +"\";"+
			"        JaTSFacade engine = JaTSFacade.getInstance();"+
			"        Reader reader = null;"+
			"        String strTemp = null;"+
			"        try {"+
			"            JatsIO jatsio = JatsIO.getInstance();"+
			"            jatsio.changeParserType(ParserFactory.JATS_PARSER);"+
			"        } catch(Exception e) {"+
			"            e.printStackTrace();"+
			"        } finally {"+
			"            if (out != null) {"+
			"                try {"+
			"                    out.flush();"+
			"                    out.close();"+
			"                } catch (Exception e) {}" +
			"            }" +
			"        }" +
			"    }" +
		    "    private static void print(INode node, String file) throws Exception {"+
		    "        PrettyPrinter pp = PrettyPrinter.getInstance();"+
		    "        out = new FileWriter(\""+this.root.replaceAll(s1,s2)+"\\\\\"+file);"+
		    "        pp.prettyPrint(node, out);"+
		    "    }"+
			"}";
		
		
		jatsio.changeParserType(ParserFactory.JATS_PARSER);
		
		try {
			result = jatsio.parse(code);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return result;
	}

}
