/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import cin.jats.engine.parser.nodes.JBinaryExpression;

/**
 * 
 * @author Adeline de Sousa
 * @version 20/05/2003
 */
public class EvaluateFunctions {
    /**
     * Mtodo auxiliar que aplica um operador aritmtico a dois objetos do tipo
     * <code>Number</code>.
     *
     * @param objOperand1 O operando  esquerda do operador.
     * @param objOperand2 O operando  direita do operador.
     * @param oper        O operador binrio. Deve ser um operador aritmtico como
     *                    "+", "-", "*", "/" ou "%".
     * @return Um objeto do tipo <code>Number</code> correspondendo ao resultado
     *         da aplicao de <code>oper</code> aos dois operandos.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Number applyArithmeticOpToNumbers(int oper, Number objOperand1,
                                                    Number objOperand2)
            throws IllegalArgumentException {

        Number retorno = null;
        if (objOperand1 != null && objOperand2 != null) {

            double op1 = objOperand1.doubleValue();
            double op2 = objOperand2.doubleValue();

            // Adio
            if (oper == JBinaryExpression.ADDITION_OPERATOR) {
                retorno = new Double(op1 + op2);
            }
            // Subtrao
            else if (oper == JBinaryExpression.SUBTRACTION_OPERATOR) {
                retorno = new Double(op1 - op2);
            }
            // Multiplicao
            else if (oper == JBinaryExpression.MULTIPLICATION_OPERATOR) {
                retorno = new Double(op1 * op2);
            }
            // Diviso
            else if (oper == JBinaryExpression.DIVISION_OPERATOR) {
                // Duas modalidades de diviso so possveis:

                // Diviso Inteira
                if (objOperand1 instanceof Integer &&
                        objOperand2 instanceof Integer) {
                    retorno = new Double(objOperand1.intValue() /
                            objOperand2.intValue());
                }
                // Diviso Decimal
                else {
                    retorno = new Double(op1 / op2);
                }
            }
            // Resto da diviso
            else if (oper == JBinaryExpression.REMAINDER_OPERATOR) {
                retorno = new Double(op1 % op2);
            }
            // oper no corresponde a um operador vlido.
            else {
                throw new IllegalArgumentException(); // O operador no  vlido.
            }
        }
        return retorno;
    }

    /**
     * Aplica um operador binrio aritmtico a dois operandos.
     *
     * @param oper    O operador binrio. Deve ser um operador aritmtico como
     *                "+", "-", "*", "/" ou "%".
     * @param lhsOpnd O operando  esquerda do operador.
     * @param rhsOpnd O operando  direita do operador.
     * @return Um objeto do tipo <code>Number</code> correspondendo ao resultado
     *         da aplicao de <code>oper</code> aos dois operandos.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Object applyBinaryArithmeticOp(int oper, Object lhsOpnd,
                                                 Object rhsOpnd)
            throws IllegalArgumentException {

        if (lhsOpnd == null || rhsOpnd == null) {
            throw new IllegalArgumentException();
        }

        Object retorno = null;

        // Os operandos de um operador aritmtico podem ser de dois tipos:
        // Dois nmeros (+, -, *, / e %) ou dois strings (+).
        // Se nenhum dos dois for verdade, uma exceo deve ser lanada.

        // Caso os dois operandos sejam Strings e o operador seja o de soma.
        if (lhsOpnd instanceof String && rhsOpnd instanceof String &&
                oper == JBinaryExpression.ADDITION_OPERATOR) {

            String str1 = (String) lhsOpnd;
            str1 = str1 + rhsOpnd;

            retorno = str1;
        }
        // Caso os dois operandos sejam nmeros
        else if (lhsOpnd instanceof Number && rhsOpnd instanceof Number) {
            Number objOperand1 = (Number) lhsOpnd;
            Number objOperand2 = (Number) rhsOpnd;

            retorno = applyArithmeticOpToNumbers(oper, objOperand1,
                    objOperand2);
        }
        else
            throw new IllegalArgumentException();
        return retorno;
    }

    /**
     * Aplica um operador bit-a-bit a dois operandos. O resultado  retornado
     * como um Object, uma vez que ele pode ser um nmero ou um valor booleano.
     * Os argumentos, por sua vez, devem ser ambos literais booleanos ou ambos
     * literais numricos inteiros.
     *
     * @param oper    O operador binrio. Deve ser um operador bit-a-bit como "&",
     *                "|" ou "^".
     * @param lhsOpnd O operando  esquerda do operador.
     * @param rhsOpnd O operando  direita do operador.
     * @return Um objeto que tanto pode ser do tipo <code>Integer</code> quanto
     *         do tipo <code>Boolean</code>, dependendo do tipo de argumento
     *         recebido.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Object applyBinaryBitwiseOp(int oper, Object lhsOpnd,
                                              Object rhsOpnd) throws IllegalArgumentException {
        if (lhsOpnd == null || rhsOpnd == null ||
                !((lhsOpnd instanceof Boolean && rhsOpnd instanceof Boolean) ||
                (lhsOpnd instanceof Integer && rhsOpnd instanceof Integer))) {
            throw new IllegalArgumentException(); // Algum operando no  vlido.
        }
        Object retorno = null;

        // "E" bit-a-bit.
        if (oper == JBinaryExpression.AND_OPERATOR) {
            if (lhsOpnd instanceof Boolean) {
                retorno = new Boolean(((Boolean) lhsOpnd).booleanValue() &
                        ((Boolean) rhsOpnd).booleanValue());
            }
            else {
                retorno = new Integer(((Integer) lhsOpnd).intValue() &
                        ((Integer) rhsOpnd).intValue());
            }
        }
        // "OU-EXCLUSIVO" bit-a-bit.
        else if (oper == JBinaryExpression.EXCLUSIVE_OR_OPERATOR) {
            if (lhsOpnd instanceof Boolean) {
                retorno = new Boolean(((Boolean) lhsOpnd).booleanValue() ^
                        ((Boolean) rhsOpnd).booleanValue());
            }
            else {
                retorno = new Integer(((Integer) lhsOpnd).intValue() ^
                        ((Integer) rhsOpnd).intValue());
            }
        }
        // "OU-INCLUSIVO" bit-a-bit.
        else if (oper == JBinaryExpression.INCLUSIVE_OR_OPERATOR) {
            if (lhsOpnd instanceof Boolean) {
                retorno = new Boolean(((Boolean) lhsOpnd).booleanValue() |
                        ((Boolean) rhsOpnd).booleanValue());
            }
            else {
                retorno = new Integer(((Integer) lhsOpnd).intValue() |
                        ((Integer) rhsOpnd).intValue());
            }
        }
        else
            throw new IllegalArgumentException();

        return retorno;
    }

    /**
     * Aplica um operador binrio de lgica booleana  a dois operandos. O
     * resultado  retornado como um objeto do tipo <code>Boolean</code>.
     *
     * @param oper    O operador binrio. Deve ser um operador de lgico booleana
     *                como "||" ou "&&".
     * @param lhsOpnd O operando  esquerda do operador.
     * @param rhsOpnd O operando  direita do operador.
     * @return Um objeto do tipo <code>Boolean</code> correspondendo ao resultado
     *         da aplicao de <code>oper</code> aos dois operandos.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Boolean applyBinaryConditionalOp(int oper, Object lhsOpnd,
                                                   Object rhsOpnd) throws IllegalArgumentException {
        if (lhsOpnd == null || rhsOpnd == null || !(lhsOpnd instanceof Boolean) ||
                !(rhsOpnd instanceof Boolean)) {
            throw new IllegalArgumentException(); // Algum operando no  vlido.
        }
        Boolean retorno = null;
        boolean opnd1 = ((Boolean) lhsOpnd).booleanValue();
        boolean opnd2 = ((Boolean) rhsOpnd).booleanValue();

        if (oper == JBinaryExpression.CONDITIONAL_OR_OPERATOR) {
            retorno = new Boolean(opnd1 || opnd2);
        }
        else if (oper == JBinaryExpression.CONDITIONAL_AND_OPERATOR) {
            retorno = new Boolean(opnd1 && opnd2);
        }
        else
            throw new IllegalArgumentException();

        return retorno;
    }

    /**
     * Aplica o operador binrio <code>oper</code> aos operandos
     * <code>lhsOpnd</code> e <code>rhsOpnd</code>.
     *
     * @param oper    O operador que dever ser aplicado sobre os operandos. Alguns
     *                dos operadores binrios suportados so de que realizam operaes
     *                aritmticas, relacionais, bit-a-bit e de lgica booleana. Nem todos
     *                os operadores binrios de Java so suportados, apenas aqueles que
     *                no envolvem nenhum tipo de armazenamento. Atribuies, por
     *                exemplo, no so suportadas.
     * @param lhsOpnd Um objeto correspondendo ao operando  esquerda do
     *                operador.
     * @param rhsOpnd Um objeto correspondendo ao operando  esquerda do
     *                operador.
     * @return O resultado de aplicar o operador aos dois operandos.  do tipo
     *         <code>Object</code> porque, dependendo do tipo de expresso,
     *         diversos tipos de objeto podem ser retornados.
     */
    public static Object applyBinaryOperator(int oper, Object lhsOpnd,
                                             Object rhsOpnd) throws IllegalArgumentException {
        if (lhsOpnd == null || rhsOpnd == null || oper <= 0)
            throw new IllegalArgumentException();
        Object retorno = null;

        // Operaes aritmticas (adio, subtrao, multiplicao, diviso,
        // e resto da diviso inteira).
        if (oper == JBinaryExpression.ADDITION_OPERATOR ||
                oper == JBinaryExpression.SUBTRACTION_OPERATOR ||
                oper == JBinaryExpression.MULTIPLICATION_OPERATOR ||
                oper == JBinaryExpression.DIVISION_OPERATOR ||
                oper == JBinaryExpression.REMAINDER_OPERATOR) {
            // Os argumentos de um operador aritmticos devem ser literais, sejam
            // numricos ou strings.
            retorno = applyBinaryArithmeticOp(oper, lhsOpnd, rhsOpnd);
        }
        // Operaes relacionais (igualdade, desigualdade, etc).
        else if (oper == JBinaryExpression.EQUALITY_OPERATOR ||
                oper == JBinaryExpression.INEQUALITY_OPERATOR ||
                oper == JBinaryExpression.LESSER_THAN_OPERATOR ||
                oper == JBinaryExpression.GREATER_THAN_OPERATOR ||
                oper == JBinaryExpression.LESSER_OR_EQUAL_TO_OPERATOR ||
                oper == JBinaryExpression.GREATER_OR_EQUAL_TO_OPERATOR) {
            retorno = applyRelationalOp(oper, lhsOpnd, rhsOpnd);
        }
        // Operadores de lgica booleana (OU lgico e E lgico).
        else if (oper == JBinaryExpression.CONDITIONAL_OR_OPERATOR ||
                oper == JBinaryExpression.CONDITIONAL_AND_OPERATOR) {
            retorno = applyBinaryConditionalOp(oper, lhsOpnd, rhsOpnd);
        }
        // Operadores de bit-a-bit (E, OU e OU-EXCLUSIVO bit-a-bit).
        else if (oper == JBinaryExpression.AND_OPERATOR ||
                oper == JBinaryExpression.INCLUSIVE_OR_OPERATOR ||
                oper == JBinaryExpression.EXCLUSIVE_OR_OPERATOR) {
            // Os operandos de um operador bit a bit podem ser dois literais
            // booleanos ou dois literais numricos inteiros.
            retorno = applyBinaryBitwiseOp(oper, lhsOpnd, rhsOpnd);
        }
        // Operadores de deslocamento (deslocamento para a esquerda, para a
        // direita e deslocamento para a direita sem sinal.
        else if (oper == JBinaryExpression.RIGHT_SHIFT_OPERATOR ||
                oper == JBinaryExpression.LEFT_SHIFT_OPERATOR ||
                oper == JBinaryExpression.UNSIGNED_RIGHT_SHIFT_OPERATOR) {
            // Os operandos de um operador bit a bit devem ser dois literais
            // numricos inteiros.
            retorno = applyBinaryBitwiseOp(oper, lhsOpnd, rhsOpnd);
        }
        // criar condio de retorno para expresses de atribuio
        // por enquanto, nenhuma avaliao ser feita, apenas
        // o lhs ser retornado.
        else if (JBinaryExpression.isAssignOperator(oper)) {
            retorno = lhsOpnd;
        }
        else {
            throw new IllegalArgumentException();
        }

        return retorno;
    }

    /**
     * Aplica um operador relacional a dois operandos. O resultado  retornado
     * como um objeto do tipo <code>Boolean</code>.
     *
     * @param oper    O operador binrio. Deve ser um operador relacional como
     *                "==", "!=", ">", "<", ">=" ou "<=".
     * @param lhsOpnd O operando  esquerda do operador.
     * @param rhsOpnd O operando  direita do operador.
     * @return Um objeto do tipo <code>Boolean</code> correspondendo ao resultado
     *         da aplicao de <code>oper</code> aos dois operandos.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Boolean applyRelationalOp(int oper, Object lhsOpnd,
                                            Object rhsOpnd) throws IllegalArgumentException {
        if (lhsOpnd == null || rhsOpnd == null) {
            throw new IllegalArgumentException();
        }

        Boolean retorno = null;

        if (oper == JBinaryExpression.EQUALITY_OPERATOR) {
            if (lhsOpnd instanceof Number && rhsOpnd instanceof Number) {
                Number num1 = (Number) lhsOpnd;
                Number num2 = (Number) rhsOpnd;
                retorno = new Boolean(num1.doubleValue() == num2.doubleValue());
            }
            else {
                retorno = new Boolean(lhsOpnd.equals(rhsOpnd));
            }
        }
        else if (oper == JBinaryExpression.INEQUALITY_OPERATOR) {
            retorno = new Boolean(!lhsOpnd.equals(rhsOpnd));
        }
        else if (lhsOpnd instanceof Number && rhsOpnd instanceof Number) {
            // Para os operadores relacionais seguintes, os dois operandos devem,
            // obrigatoriamente, ser nmeros.
            Number objOperand1 = (Number) lhsOpnd;
            Number objOperand2 = (Number) rhsOpnd;
            double op1 = objOperand1.doubleValue();
            double op2 = objOperand2.doubleValue();

            if (oper == JBinaryExpression.LESSER_THAN_OPERATOR) {
                retorno = new Boolean(op1 < op2);
            }
            else if (oper == JBinaryExpression.GREATER_THAN_OPERATOR) {
                retorno = new Boolean(op1 > op2);
            }
            else if (oper == JBinaryExpression.LESSER_OR_EQUAL_TO_OPERATOR) {
                retorno = new Boolean(op1 <= op2);
            }
            else if (oper == JBinaryExpression.GREATER_OR_EQUAL_TO_OPERATOR) {
                retorno = new Boolean(op1 >= op2);
            }
        }
        return retorno;
    }

    /**
     * Aplica um deslocamento a um nmero inteiro.
     *
     * @param oper    O operador binrio. Deve ser um operador de deslocamento como
     *                ">>", "<<" ou ">>>".
     * @param lhsOpnd O nmero inteiro que receber o deslocamento.
     * @param rhsOpnd O tamanho do deslocamento. Para um deslocamento  esquerda,
     *                um valor igual 1 implica em <code>lhsOpnd</code> ser multiplicado
     *                por 2. Um valor igual a 3, implica em <code>lhsOpnd</code> ser
     *                multiplicado por 8 e assim, sucessivamente, dependendo,  claro, do
     *                tipo de deslocamento.
     * @return Um objeto do tipo <code>Boolean</code> correspondendo ao resultado
     *         da aplicao de <code>oper</code> aos dois operandos.
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     * @throws IllegalArgumentException Levantada quando um n
     *                                  apresenta uma estrutura interna invlida.
     */
    public static Integer applyShiftOp(int oper, Object lhsOpnd,
                                       Object rhsOpnd) throws IllegalArgumentException {
        if (lhsOpnd == null || rhsOpnd == null ||
                !(lhsOpnd instanceof Integer) ||
                !(rhsOpnd instanceof Integer)) {
            throw new IllegalArgumentException(); // Algum operando no  vlido.
        }
        Integer retorno = null;
        Integer opnd1 = (Integer) lhsOpnd;
        Integer opnd2 = (Integer) rhsOpnd;

        if (oper == JBinaryExpression.LEFT_SHIFT_OPERATOR) {
            retorno = new Integer(opnd1.intValue() << opnd2.intValue());
        }
        else if (oper == JBinaryExpression.RIGHT_SHIFT_OPERATOR) {
            retorno = new Integer(opnd1.intValue() >> opnd2.intValue());
        }
        else if (oper == JBinaryExpression.UNSIGNED_RIGHT_SHIFT_OPERATOR) {
            retorno = new Integer(opnd1.intValue() >>> opnd2.intValue());
        }
        else
            throw new IllegalArgumentException();

        return retorno;
    }

    /**
     * Este mtodo compara dois objetos do tipo <code>Class</code>, cada um
     * correspondendo a um tipo, com o fim de verificar se esses tipos so
     * compatveis. Por compatveis, entende-se que ambos so o mesmo tipo ou
     * um deles correspondente a um tipo primitivo e o outro corresponde  classe
     * correspondente quele tipo primitivo (ex: Um dos tipos corresponde ao tipo
     * primitivo <code>boolean</code> e o outro corresponde  classe
     * <code>Boolean</code>.
     *
     * @param type1 Um objeto do tipo <code>Class</code> que ser comparado a
     *              um outro de mesmo tipo para saber se ambos representam classes
     *              compatveis.
     * @param type2 Um objeto do tipo <code>Class</code> que ser comparado a
     *              um outro de mesmo tipo para saber se ambos representam classes
     *              compatveis.
     * @return Um valor booleano que diz se os dois tipos so compatveis. Dois
     *         tipos podem ser considerados compatveis quando:<p>
     *         <ul>Eles representam a mesma classe;<br>
     *         <ul>Um deles corresponde a um tipo primitivo e o outro  classe
     *         responsvel por representar esse tipo (por exemplo, int e
     *         java.lang.Integer).
     * @throws IllegalArgumentException Levantada quando o
     *                                  parmetro <code>node</code> for <code>null</code>.
     */
    public static boolean typesAreCompatible(Class type1, Class type2) throws
            IllegalArgumentException {

        if (type1 == null || type2 == null) {
            throw new IllegalArgumentException();
        }

        boolean retorno = false;
        String[] typeNames = new String[2];
        typeNames[0] = type1.getName();
        typeNames[1] = type2.getName();
        int thePrimitiveType = 0;
        if (type1.isPrimitive())
            thePrimitiveType = 1;
        if (type2.isPrimitive())
            thePrimitiveType = thePrimitiveType + 2;

        // Caso thePrimitiveType seja igual a 1 ou 2, apenas um dos tipos 
        // primitivo. Consequentemente, no adianta simplesmente comparar os
        // nomes das classes (ex: Se um dos ns representar um int, o nome de
        // sua classe ser "int", mesmo considerando que no existe uma classe
        // com esse nome). "int"  diferente de "java.lang.Integer", mas
        // queremos considerar os dois nomes iguais.
        if (thePrimitiveType == 1 || thePrimitiveType == 2) {
            int index1 = thePrimitiveType - 1;
            int index2 = thePrimitiveType == 1 ? 1 : 0;
            if ((typeNames[index1].equals("int") &&
                    typeNames[index2].equals("java.lang.Integer")) ||
                    (typeNames[index1].equals("long") &&
                    typeNames[index2].equals("java.lang.Long")) ||
                    (typeNames[index1].equals("byte") &&
                    typeNames[index2].equals("java.lang.Byte")) ||
                    (typeNames[index1].equals("char") &&
                    typeNames[index2].equals("java.lang.Character")) ||
                    (typeNames[index1].equals("float") &&
                    typeNames[index2].equals("java.lang.Float")) ||
                    (typeNames[index1].equals("double") &&
                    typeNames[index2].equals("java.lang.Double")) ||
                    (typeNames[index1].equals("short") &&
                    typeNames[index2].equals("java.lang.Short")) ||
                    (typeNames[index1].equals("boolean") &&
                    typeNames[index2].equals("java.lang.Boolean"))) {
                retorno = true;
            }
        }
        else {
            if (type1.isAssignableFrom(type2)) {
                retorno = true;
            }
        }
        return retorno;
    }
}
