/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import java.util.*;

import cin.jats.engine.parser.nodes.INode;

/**
 * Esta classe armazena os ns da rvore sinttica de um programa em forma de
 * tabela hash. O mtodo de hashing consiste em pegar o valor do log na base 2
 * do identificador do tipo do n e decrementar de 1 unidade. Este decremento
 * deve-se ao fato de alguns ns possurem o tipo NO_TYPE, o qual tem valor 1 e,
 * portanto, resultariam no ndice zero. Ns deste tipo no sero inseridos na
 * tabela. Assim, a cada linha da tabela corresponde uma lista de ns de um
 * determindo tipo.
 * 
 * @author Adeline de Sousa
 * @version  - 27/01/2003
 * @since JDK 1.3
 */
public class JaTSHashTable implements Cloneable {

    // Constante que indica o tamanho da tabela
    public static final int TABLE_SIZE = 59;

    // este vetor armazenar as linhas da tabela
    private Vector table;

    private int size;

    /**
     * Construtor vazio da classe
     */
    public JaTSHashTable() {
        this.initTable();
    }

    private JaTSHashTable(Vector set, int size) {
        if (set != null) {
            this.table = set;
            this.size = size;
        } else {
            this.initTable();
        }
    }

    /**
     * Cria a tabela propriamente dita, fazendo com que o <code>Vector</code>
     * table seja  preenchido por elementos do tipo <code>NodeList</code>
     */
    private void initTable() {
        this.table = new Vector(JaTSHashTable.TABLE_SIZE, 10);
        for (int i = 0; i < JaTSHashTable.TABLE_SIZE; i++) {
            this.table.add(new Vector());
        }
        this.size = 0;
    }


    /**
     * Retorna a tabela
     */
    public Vector getTable() {
        return this.table;
    }


    /**
     * Adiciona um para var-valor  tabela
     */
    public void put(INode var, Object value) {
        if (var != null) {
            Pair p = new Pair(var, value);
            this.put(p);
        } else {
            throw new IllegalArgumentException("Argumento var nulo");
        }
    }

    /**
     * Adiciona um para var-valor  tabela
     */
    public void put(Pair p) {

        if (p != null) {
            int nline = p.hashCode();
            Vector line = ((Vector) this.table.elementAt(nline));
            if (!line.contains(p)) {
                line.addElement(p);
                this.size = this.size + 1;


            } else {
                Pair p2 = (Pair) line.get(line.indexOf(p));
                p2.setValue(p.getValue());
            }
        } else {
            throw new IllegalArgumentException("Argumento var nulo");
        }
    }


    /**
     * Retorna um vetor contendo apenas objetos do tipo especificado pelo parmetro
     */
    public Vector getLine(INode node) {
        Vector retorno;
        if (node != null) {
            int line = node.hashCode();
            retorno = (Vector) this.table.elementAt(line);
        } else {
            throw new IllegalArgumentException("Parametro node nulo");
        }
        return retorno;
    }


    /**
     * Remove o par var-valor, cuja chave  var
     *
     * @param  var  Chave do para a ser deletado
     */
    public void remove(INode var) {
        Pair p = new Pair(var, null);
        this.remove(p);

    }

    /**
     *  Elina todos os elementos da tabela.
     */
    public void reset() {
        initTable();
    }

    /**
     * Remove um elemento a tabela
     *
     * @param  p Elemento a ser removido da tabela
     */
    public void remove(Pair p) {
        Vector line = null;

        if (p != null) {
            int nline = p.hashCode();
            line = ((Vector) this.table.elementAt(nline));
            if (line.contains(p)) {
                line.remove(p);
                this.size = this.size - 1;
            } else {
                // se a varivel no for encontrada, simplesmente no faz nada!
            }
        } else {
            throw new IllegalArgumentException("Elemento nulo.");
        }

    }

    /**
     * Clona este objeto
     *
     * @return Object
     */
    public Object clone() {
        JaTSHashTable hash = null;

        Vector newTable = new Vector();

        for (int i = 0; i < this.table.size(); i++) {
            Vector tmpVector = (Vector) this.table.elementAt(i);
            Vector tmpClone = new Vector();

            for (int j = 0; j < tmpVector.size(); j++) {
                Object tmpValue = tmpVector.elementAt(j);
                if (tmpValue != null) {
                    Object clone = ((Pair) tmpValue).clone();
                    tmpClone.addElement(clone);
                }
            }
            newTable.addElement(tmpClone);
        }
        hash = new JaTSHashTable(newTable, this.table.size());
        return hash;
    }

    /**
     * Verifica a existncia da varivel var dentro da tabela
     */
    public boolean containsVar(INode var) {
        boolean retorno = false;
        if (var != null && var.isVariable()) {
            Iterator itTable = this.table.iterator();
            while (!retorno && itTable.hasNext()) {
                Vector line = (Vector) itTable.next();
                Iterator tmp = line.iterator();
                while (!retorno && tmp.hasNext()) {
                    Pair p = (Pair) tmp.next();
                    INode tmpVar = p.getVar();
                    String tmpVarName = this.getVariableName(tmpVar.getVariableName());
                    String varVarName = this.getVariableName(var.getVariableName());

                    retorno = tmpVarName.equals(varVarName);

                }
            }
        }
//        else if (var != null && var.isOrDeclaration()) {
//            Vector line = this.getLine(var);
//            Iterator it = line.iterator();
//            while (it.hasNext() && !retorno) {
//                Pair p = (Pair) it.next();
//                INode tmpNode = p.getVar();
//                retorno = tmpNode.equals(var);
//            }
//
//        }
        else {
            throw new IllegalArgumentException();
        }
        return retorno;

    }

    /**
     * Mtodo auxiliar para permitir a consulta apenas pelo nome da varivel e
     * no pelo seu tipo. A utilizao desse mtodo como consulta a uma
     * determinada varivel, antes de inseri-la, elimina a existencia de entradas
     * multiplas na tabela
     */
    private String getVariableName(String fullName) {
        String retorno = fullName;
        if (Util.checkString(fullName)) {
            int indexColon = fullName.indexOf(':');
            // se a variaval possui ":" no nome, pega apenas o nome da varivel,
            // que se encontra entre o ":" e os depois
            if (indexColon > 0) {
                retorno = fullName.substring(indexColon + 1);
            }
        }
        return retorno;

    }

    /**
     * Retorna a lista de variaveis presentes na tabela
     */
    public Iterator getVariables() {
        Vector variables = new Vector();
        Iterator itTable = this.table.iterator();
        Iterator tmp = null;
        while (itTable.hasNext()) {
            Vector line = (Vector) itTable.next();
            tmp = line.iterator();
            while (tmp.hasNext()) {
                Pair p = (Pair) tmp.next();
                INode var = p.getVar();
                variables.add(var);
            }
        }

        return variables.iterator();
    }

    /**
     * Retorna a lista dos valores presentes nesta tabela.
     */
    public Collection values() {
        Vector values = new Vector();
        Iterator itTable = this.table.iterator();
        Iterator tmp = null;
        while (itTable.hasNext()) {
            Vector line = (Vector) itTable.next();
            tmp = line.iterator();
            while (tmp.hasNext()) {
                Pair p = (Pair) tmp.next();
                Object val = p.getValue();
                values.add(val);
            }
        }

        return values;
    }

    /**
     * Retorna um conjunto contendo todas as variveis presentes na tabela.
     */
    public Set keySet() {
        HashSet variables = new HashSet();
        Iterator itTable = this.table.iterator();
        Iterator tmp = null;
        while (itTable.hasNext()) {
            Vector line = (Vector) itTable.next();
            tmp = line.iterator();
            while (tmp.hasNext()) {
                Pair p = (Pair) tmp.next();
                INode var = p.getVar();
                variables.add(var);
            }
        }

        return variables;
    }


    /**
     * Retorna um conjunto contendo todas as entradas da tabela. Onde cada
     * entrada  um par variavel-valor.
     */
    public Set entrySet() {
        HashMap entries = new HashMap();
        Iterator itTable = this.table.iterator();
        Iterator tmp = null;
        while (itTable.hasNext()) {
            Vector line = (Vector) itTable.next();
            tmp = line.iterator();
            while (tmp.hasNext()) {
                Pair p = (Pair) tmp.next();
                entries.put(p.getVar(), p.getValue());
            }
        }
        return entries.entrySet();
    }

    /**
     * Coloca todos os elementos da tabela passada como parametro
     * na tabela atual
     */
    public void putAll(JaTSHashTable table) {
        Iterator it = table.getTable().iterator();
        Iterator tmp = null;
        while (it.hasNext()) {
            Vector line = (Vector) it.next();
            tmp = line.iterator();
            while (tmp.hasNext()) {
                Pair pair = (Pair) tmp.next();
                this.put(pair);
            }
        }
    }

    public int getSize() {
        return this.size;
    }


    public Object get(INode var) {
        boolean achou = false;
        Object retorno = null;
        if (var != null && var.isVariable()) {
            Iterator itTable = this.table.iterator();
            while (!achou && itTable.hasNext()) {
                Vector line = (Vector) itTable.next();
                Iterator tmp = line.iterator();
                while (!achou && tmp.hasNext()) {
                    Pair p = (Pair) tmp.next();
                    INode tmpVar = p.getVar();
                    if (tmpVar.getVariableName().equals(var.getVariableName())) {
                        retorno = p.getValue();
                        achou = (retorno != null);
                    }
                }
            }
        }
//        // se for do tipo OR, ento s h uma possibilidade pra linha onde se
//        // encontra
//        else if (var != null && var.isOrDeclaration()) {
//            Vector line = this.getLine(var);
//            Iterator it = line.iterator();
//            while (it.hasNext() && !achou) {
//                Pair p = (Pair) it.next();
//                INode tmpNode = p.getVar();
//                if (tmpNode.equals(var)) {
//                    retorno = p.getValue();
//                    achou = (retorno != null);
//                }
//            }
//        }
         else {
            throw new IllegalArgumentException();
        }
        return retorno;
    }

} // fim da classe