/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.io.Serializable;


import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;
import cin.jats.engine.parser.nodes.exceptions.ExecutionException;
import cin.jats.engine.visitors.IVisitor;


/**
 * Esta classe implementa uma lista de objetos do tipo <code>INode</code>.
 *
 * @see INode
 *
 * @author Adeline de Sousa
 * @author Fernando Castor
 * @see INode
 * @version  27/01/2003
 * @since JDK 1.3
 */
public class NodeList implements Cloneable, Serializable {


    // A lista onde sero armazenados os elementos.
    private Vector elements;


    /**
     * Construtor vazio da classe.
     */
    public NodeList() {
        elements = new Vector();
    }


    /**
     * Construtor que recebe como argumento a capacidate mxima da lista de ns.
     *
     * @param capacity um inteiro correspondendo  capacidade mxima da lista de ns.
     */
    public NodeList(int capacity) {
        elements = new Vector(capacity);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo
     * <code>Vector</code> contendo apenas objetos que implementam INode
     */
    protected NodeList(Vector objects) throws IllegalArgumentException {
        if (objects == null) {
            throw new IllegalArgumentException();
        }
        Iterator it = objects.iterator();
        while (it.hasNext()) {
            if (!(it.next() instanceof INode)) {
                throw new IllegalArgumentException();
            }
        }
    }


    /**
     * Adiciona um elemento  lista.
     *
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o argumento
     *			 <code>node</code> for <code>null</code>.
     */
    public void addElement(INode node) throws IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        this.elements.addElement(node);
    }


    /**
     * Adiciona toda a lista passada como parametro aa lista atual
     *
     * @param list    Um objeto qualquer do tipo <code>NodeList</code>.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o argumento
     *			 <code>list</code> for <code>null</code>.
     */
    public void addAll(NodeList list) throws IllegalArgumentException {
        if (list == null) {
            throw new IllegalArgumentException("A Lista passada eh nula");
        }
        this.elements.addAll(list.elements);
    }

    /**
     * Retorna o indice do elemento especificado no parametro
     *
     * @param elem  Elemento cujo indice eh procurado
     */
    public int indexOf(INode elem) {
        return this.elements.indexOf(elem);
    }


    /**
     * Retorna a capacidade desta lista.
     *
     * @return Um inteiro correspondendo  capacidade da lista.
     */
    public int capacity() {
        return this.elements.capacity();
    }


    /**
     * Retorna um clone desta lista e de todos os elementos nela presentes.
     *
     * @return Uma cpia funda desde objeto.
     */
    public Object clone() {
        NodeList retorno = new NodeList();
        int size = this.size();
        for(int i = 0; i < size; i++){
            retorno.addElement((INode)((AbstractNode)this.elementAt(i)).clone());
        }
        return retorno;
    }


    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     *
     * @param  pos Um inteiro correspondendo  posio desejada.
     *
     * @return Um objeto do tipo <code>INode</code> que se encontra na posio
     *		  desejada.
     */
    public INode elementAt(int pos) {
        return (INode) this.elements.elementAt(pos);
    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para a
     * direita na posies do item que esiver naquela posio e nas daqueles que .
     * estiverem  direita dela.
     *
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @param pos Um inteiro correspondendo  posio onde o n deve ser
     *		 inserido.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *			 argumento <code>node</code> for <code>null</code> ou o
     *			 argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode node, int pos) throws IllegalArgumentException {
        if (node == null || pos < 0) {
            throw new IllegalArgumentException("node nulo ou posio: " + pos + " invalida");
        }
        this.elements.insertElementAt(node, pos);
    }


    /**
     * Insere os elementos da lista passada como parametro, na lista atual
     * a partir da posio especificada e dando um shift, nos elementos que se
     * encontram a direita de tal posio.
     *
     * @param nodelist A lista cujos elementos sero adicionados em
     * @param pos Um inteiro correspondendo  posio a partir da qual, os ns
     *        da lista passada como parmetro devem ser inseridos
     *
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *			 argumento <code>nodelist</code> for <code>null</code> ou o
     *			 argumento <code>pos</code> for negativo.
     */
    public void insertNodeListAt(NodeList nodelist, int pos)
            throws IllegalArgumentException {
        if (nodelist == null || pos < 0) {
            throw new IllegalArgumentException();
        }
        JaTSIterator jit = nodelist.getIterator();
        int i = 0;
        while (jit.hasNext()) {
            INode jNode = jit.next();
            this.insertElementAt(jNode, pos + i);
            i++;
        }
    }

    /**
     * Diz se a lista es vazia.
     *
     * @return Um boolean que, caso tenho o valor <code>true</code>, indica
     *			  que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.elements.isEmpty();
    }


    /**
     * Retorna um java.util.Iterator para que se possa percorrer a estrutura
     *
     * @return um iterador para percorrer os elementos da lista
     */
    public Iterator iterator() {
        return this.elements.iterator();
    }

    /**
     * Retorna um java.util.Iterator para que se possa percorrer a estrutura
     *
     * @return um iterador para percorrer os elementos da lista
     */
    public JaTSIterator getIterator() {
        JaTSIterator retorno = new JaTSIterator(this.elements.iterator());
        return retorno;
    }


    /**
     * Verifica se o elemento passado como parametro pertence a lista
     *
     * @param elem Elemento procurado
     *
     * @return true, caso o elemento passado pertenca a lista
     *         false, caso contrrio
     */
    public boolean contains(INode elem) {
        boolean retorno = false;
        JaTSIterator it = this.getIterator();
        INode node = null;
        while (it.hasNext() && !retorno) {
            node = it.next();
            retorno = node.equals(elem);
        }
        return retorno;
    }


    /**
     * Remove o elemento que se encontra na  posio <code>pos</code> desta
     * lista.
     *
     * @param pos Um inteiro correspondendo  posio do n que deve ser
     *		 removido.
     */
    public void removeElementAt(int pos) {
        boolean validPosition = (pos >= 0 && pos < this.size());
        if (!validPosition) {
            throw new IllegalArgumentException();
        }
        this.elements.removeElementAt(pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista,
     * sobreescrevendo o elemento que estava nela.
     *
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @param pos Um inteiro correspondendo  posio onde o n deve ser
     *		 inserido.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *			 argumento <code>node</code> for <code>null</code> ou o
     *			 argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode node, int pos)
            throws IllegalArgumentException {

        boolean validPosition = (pos >= 0 && pos < this.size());
        //QualitiLogger.debug("IH = " + node);
        if (node == null || !validPosition) {
            throw new IllegalArgumentException();
        }
        this.elements.setElementAt(node, pos);
    }


    /**
     * Retorna o nmero de elementos da lista.
     *
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        return this.elements.size();
    }


    /**
     * Retorna o primeiro elemento da lista
     *
     * @return O primeiro elemento da lista, ou null, se a lista for vazia.
     */
    public INode first() {
        INode retorno = null;
        if (!this.isEmpty()){
            retorno = (INode) this.elements.elementAt(0);
        }
        return retorno;
    }

    /**
     * Retorna o ltimo elemento da lista
     *
     * @return O ltimo elemento da lista, ou null, se a lista for vazia.
     */
    public INode last() {
        INode retorno = null;
        if (!this.isEmpty()){
            retorno = (INode) this.elements.elementAt(this.size() - 1);
        }
        return retorno;
    }

    /**
     * Seta o novo tamanho da lista.
     * @param s Um inteiro correspondendo ao novo tamanho da lista.
     */
    public void setSize(int s) {
        this.elements.setSize(s);
    }


    /**
     * Retorna a coleo correspondente a esta NodeList
     *
     * @return a coleo correspondente a esta NodeList
     */
    public Collection getElements() {
        return this.elements;
    }


    /**
     * Compara dois objetos do tipo NodeList e diz se estes so iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false, achou = false;
        if (obj instanceof NodeList) {
            NodeList nlist = (NodeList) obj;
            if (this.size() == nlist.size()) {
                int max = this.size();
                int i = 0;
                while (i < max && !achou) {
                    achou = (!(this.elementAt(i).equals(nlist.elementAt(i))));
                    i++;
                }
                if (!achou) {
                    retorno = true;
                }
            }
        }
        return retorno;
    }


    /**
     * Executa o matching dos objetos que se encontram na nodeList.
     */
    public void match(NodeList nodeList, ResultSet rset)
            throws NodesNotMatchedException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException {

        if (this.size() == nodeList.size()) {
            int max = this.size();
            int i = 0;
            while (i < max) {
                this.elementAt(i).match(nodeList.elementAt(i), rset);
                i++;
            }
        } else {
            throw new NodesNotMatchedException(NodesNotMatchedException.NODES_HAVE_DIFFERENT_SIZES_MSG);
        }
    }

     /**
     * Executa o matching dos objetos que se encontram na nodeList.
     */
    public Object accept(IVisitor visitor, Object data)
              throws ExecutionException, IllegalArgumentException,
             cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException {
            Object result = null;
            int i = this.size() - 1;
            while (i >= 0) {
                result = this.elementAt(i).accept(visitor, data);
                i--;
            }

         return result;



    }

    /**
     * Reverte a lista de ns passada como parametro.
     *
     * @return Esta lista de ns revertida.
     */
    public NodeList reverse() {
        int max = this.size() - 1;
        NodeList retorno = new NodeList();
        for (int i = 0; i <= max; i++) {
            retorno.addElement(this.elementAt(max - i));
        }
        return retorno;
    }

    /**
     * Concatena a lista passada como parametro  com o fim da lista atual,
     * retornando uma lista que contm os elementos das duas listas. Ele nao
     * verifica se existe ou nao uma repetio.
     *
     * @param nl
     *
     * @return  Lista contendo os elementos de <code>this<code> + os elementos
     *          de <code>nl<code>
     */
    public NodeList concat(NodeList nl) {
        NodeList retorno = (NodeList) this.clone();
        if (nl != null) {
            int size = nl.size();
            for (int i = 0; i < size; i++) {
                retorno.addElement(nl.elementAt(i));
            }
        }
        return retorno;

    }

    /**
     * Retorna uma NodeList contendo apenas os elementos que se encontram no
     * range especificado, sendo o  primeiro indice inclusivo e o segundo,
     * exclusivo.<br>
     * Assim, o intervalo : [ beginIndex, endIndex) <br>
     *
     * Dada a NodeList  <code>nodeL</code> contendo os elementos:
     *
     * <blockquote>
     *  elem1, elem2, elem3, elem4, elem5, elem6
     * </blockquote>
     *
     * A chamada:
     * <blockquote>
     *  nodeL.subSequence(1,3);
     * </blockquote>
     *
     * Retorna a lista:
     * <blockquote>
     * elem2, elem3
     * </blockquote>
     */
    public NodeList subSequence(int beginIndex, int endIndex) {
        NodeList retorno = new NodeList();
        boolean validRange = beginIndex >= 0 && endIndex <= this.size() &&
                beginIndex <= endIndex;
        if (validRange) {
            for (int i = beginIndex; i < endIndex; i++) {
                retorno.addElement((INode) elements.elementAt(i));
            }
        } else {
            throw new IllegalArgumentException("Invalid range at NodeList.subSequence");
        }

        return retorno;

    }

    /**
     * Returns a string representation of the object. In general, the
     * <code>toString</code> method returns a string that
     * "textually represents" this object. The result should
     * be a concise but informative representation that is easy for a
     * person to read.
     * It is recommended that all subclasses override this method.
     * <p>
     * The <code>toString</code> method for class <code>Object</code>
     * returns a string consisting of the name of the class of which the
     * object is an instance, the at-sign character `<code>@</code>', and
     * the unsigned hexadecimal representation of the hash code of the
     * object. In other words, this method returns a string equal to the
     * value of:
     * <blockquote>
     * <pre>
     * getClass().getName() + '@' + Integer.toHexString(hashCode())
     * </pre></blockquote>
     *
     * @return  a string representation of the object.
     */
    public String toString() {
        String retorno = " ";
        JaTSIterator it = this.getIterator();
        while (it.hasNext()) {
            INode jaTSNode = it.next();
            retorno += jaTSNode + ", ";
        }
        return retorno;
    }


    

}// fim da classe
