/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;

import cin.jats.engine.parser.nodes.INode;
import cin.jats.engine.parser.nodes.AbstractNode;

/**
 * Esta classe armazena os ns da rvore sinttica de um programa em forma de
 * tabela hash. O mtodo de hashing consiste em pegar o valor do log na base 2
 * do identificador do tipo do n e decrementar de 1 unidade. Este decremento
 * deve-se ao fato de alguns ns possurem o tipo NO_TYPE, o qual tem valor 1 e,
 * portanto, resultariam no ndice zero. Ns deste tipo no sero inseridos na
 * tabela Assim, a cada linha da tabela corresponde uma lista de ns de um
 * determindo tipo.
 *
 *
 * Esta classe, difere do ResultSet por dois motivos:
 * 1 - no insere pares
 * 2 - insere ns que no so variveis 
 * 
 * @author Adeline de Sousa
 * @version  - 27/01/2003
 * @since JDK 1.3
 */
public class NodeTable {

    // Constante que indica o tamanho da tabela
    public static final int TABLE_SIZE = 60;

    // tamanho inical da tabela
    public static final int INITIAL_SIZE = 10;

    // este vetor armazenar as linhas da tabela
    private Vector table;

    // armazena o numero de elementos presentes na tabela.
    private int size;

    /**
     * Construtor vazio da classe
     */
    public NodeTable() {
        this.initTable();
    }

    /**
     * Cria a tabela propriamente dita, fazendo com que o <code>Vector</code>
     * table seja  preenchido por elementos do tipo <code>NodeList</code>
     */
    private void initTable() {
        this.table = new Vector(NodeTable.TABLE_SIZE, INITIAL_SIZE);
        for (int i = 0; i < NodeTable.TABLE_SIZE; i++) {
            this.table.addElement(new NodeList(INITIAL_SIZE));
        }
    }

    /**
     * Retorna a tabela
     */
    public Vector getTable() {
        return this.table;
    }


    /**
     * Retorna a linha onde o elemento deve ser inserido/procurado
     *
     * @param node O elemento para o qual procuramos a linha da tabela
     *
     * @return A linha da tabela, um objeto do tipo <code>NodeList</code> onde o
     *         elemento deve ser inserido ou procurado
     */
    private NodeList getLine(INode node) {

        NodeList retorno = new NodeList();
        retorno = this.getLine(node.hashCode());
        return retorno;
    }


    /**
     * Retorna a linha onde o elemento deve ser inserido/procurado
     *
     * @param type O elemento para o qual procuramos a linha da tabela
     *
     * @return A linha da tabela, um objeto do tipo <code>NodeList</code> onde o
     *         elemento deve ser inserido ou procurado
     */
    public NodeList getLine(int type) {

        NodeList retorno = new NodeList();
        if (AbstractNode.isValidType(type)) {
            if (type > this.table.size()) {
                this.table.setSize(type + 5);
            }
            Object line = this.table.elementAt(type);
            if (line == null) {
                this.table.setElementAt(new NodeList(INITIAL_SIZE), type);
                retorno = (NodeList) this.table.elementAt(type);
            } else {
                // esse cast  vlido, porque em table s so inseridos elementos do
                // tipo NodeList
                retorno = (NodeList) line;
            }

            return retorno;
        } else {
            throw new IllegalArgumentException("Valor invalido para o tipo do no");
        }
    }

    /**
     * Retorna a linha cujos elementos so variveis do tipo especificado no
     * parmetro
     *
     * @param type Tipo da varivel
     *
     * @return A linha onde se encontram as variveis do tipo <code>type</code>
     */
    public NodeList getNodeVarLine(int type) {
        NodeList retorno = new NodeList();
        NodeList line = this.getLine(type);
        Iterator it = line.iterator();
        INode tmp = null;
        while (it.hasNext()) {
            tmp = (INode) it.next();
            if (tmp.isVariable()) {
                retorno.addElement(tmp);
            }
        }
        return retorno;
    }
    
    /**
     * Um iterator somente sobre os ns da tabela que so variveis.
     * 
     * @author tbas
     */
    private class VariableIterator implements Iterator {
    	
    	// varre as colunas dentro de uma linha da tabela
    	private Iterator columnIterator;
    	
    	// varre as linhas da tabela
    	private Iterator lineIterator;
    	
    	private Object next;
    	
    	VariableIterator() {
    		lineIterator = table.iterator();
    		if (lineIterator.hasNext()) {
    			NodeList nl = (NodeList) lineIterator.next();
    			columnIterator = nl.iterator();
    		}
    	}

		/**
		 * @see java.util.Iterator#hasNext()
		 */
		public boolean hasNext() {
			return findNext() != null;
		}

		private Object findNext() {
			Object result = null;
			boolean found = false;
			
			do {
				
				while(!found && columnIterator.hasNext()) {
					INode node = (INode) columnIterator.next();
					found = node.isVariable();
					result = node;
				}
				
				if (!found && lineIterator.hasNext()) {
					NodeList nl = (NodeList) lineIterator.next();
					columnIterator = nl.iterator();
				}
				
			} while(!found && lineIterator.hasNext());
			
			if (found) {
				next = result;
				return result;
			} else {
				return null;
			}
		}
		
		/**
		 * @see java.util.Iterator#next()
		 */
		public Object next() {
			Object result;
			if ( next == null ) {
				result = findNext();
				if (result == null) {
					throw new NoSuchElementException();
				}
			} else {
				result = next;
			}
			return result;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}
    }
    
    /**
     * Retorna um Iterator sobre os ns que so variveis da tabela
     * 
     * @return java.util.Iterator
     */
    public Iterator getVariableIterator() {
    	return new VariableIterator();
    }
    

    /**
     * Retorna a linha cujos elementos so do tipo especificado no parmetro
     *
     * @param type Tipo do n
     *
     * @return A linha onde se encontram os ns do tipo <code>type</code> que
     *          no so variveis
     */
    public NodeList getNodeLine(int type) {
        NodeList retorno = new NodeList();
        NodeList line = this.getLine(type);
        Iterator it = line.iterator();
        INode tmp = null;
        while (it.hasNext()) {
            tmp = (INode) it.next();
            if (!tmp.isVariable()) {
                retorno.addElement(tmp);
            }
        }
        return retorno;
    }

    /**
     * Adiciona um elemento a tabela, indexando-o pelo identificador do tipo,
     * decrementado de 1.
     *
     * @param node Elemento a ser adicionado na tabela
     */
    public void addElement(INode node) {

        if (node != null && node.getTypeIdentifier() != INode.NO_TYPE
                && !node.isExecutable()) {

            NodeList line = this.getLine(node);
            if (line != null) {
                try {
                    if (!line.contains(node)) {
                        line.addElement(node);
                        this.size = this.size + 1;
                    }
                } catch (Exception e) {
                    // fazer algum tratamento aqui!!!
                }
            } else {
                throw new IllegalArgumentException();
            }
        }

    }


    /**
     * Remove um elemento a tabela
     *
     * @param node Elemento a ser removido da tabela
     */
    public void removeElement(INode node) {

        NodeList line = this.getLine(node);
        boolean achou = false;
        INode tmpNode = null;

        if ((line != null) && (line.contains(node))) {

            int i = 0, max = line.size();
            while (i < max && !achou) {
                tmpNode = line.elementAt(i);
                if (tmpNode.equals(node)) {
                    line.removeElementAt(i);
                    this.size = this.size - 1;
                    achou = true;
                }
                i++;
            }
        } else {
            throw new IllegalArgumentException("Elemento nao encontrado.");
        }
    }

    /**
     * Retorna o numero de ns presentes na tabela.
     *
     * @return Numero de ns inseridos na tabela
     */
    public int getSize() {
        return size;
    }
} // fim da classe