/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Collection;

import cin.jats.engine.JatsIO;
import cin.jats.engine.parser.ParseException;
import cin.jats.engine.parser.nodes.Declaration;
import cin.jats.engine.parser.nodes.DeclarationSet;
import cin.jats.engine.parser.nodes.INode;


public class ResultSet implements Cloneable, Map {

    // Este objeto armazena pares (varivel,n). Onde as variveis so tipos
    // JaTS e os valores podem tanto ser tipos JaTS quanto Strings
    // representando estes objetos
    private JaTSHashTable mappings;

    /**
     * Construtor da classe
     */
    public ResultSet() {
        this.mappings = new JaTSHashTable();
    }

    /**
     * Construtor da classe. Recebe como parametro um objeto do tipo
     * Solution
     *
     * @param solucao  Objeto do tipo Solution, contendo os pares var-valor
     */
    public ResultSet(JaTSHashTable solucao) {
        this.setMappings(solucao);
    }

    /**
     * Popula o ResultSet com as variveis presentes na NodeTable passada como
     * parmetro. A NodeTable  tem praticamente a mesma estrutura interna do
     * ResultSet, no entanto, guarda tanto variveis quanto ns java.
     *
     * @param table Tabela contendo as variveis a serem inseridas no result set
     */
    public void addVariablesFromNodeTable(NodeTable table) {
    	Iterator it = table.getVariableIterator();
    	while (it.hasNext()) {
    		INode node = (INode) it.next();

            if (node != null && node.isVariable()) {
                this.put(node, null);
            }
        }
    }
    
    /**
     * Retorna o objeto mappings, referente a este ResultSet
     *
     * @return Um objeto do tipo <code>Solution</code>, referente a este ResultSet
     */
    public JaTSHashTable getMappings() {
        return this.mappings;
    }


    /**
     * Seta o objeto mappings dseste ResultSet
     */
    public void setMappings(JaTSHashTable mappings) {
        if (mappings != null) {
            this.mappings = mappings;
        } else {
            throw new IllegalArgumentException("Argumento mappings nulo");
        }
    }

    /**
     * Adiciona um par varivel-valor ao ResultSet. Se o valor passado for uma
     * string, transforma-o em um INode antes de realizar o mapeamento
     * propriamente dito.
     *
     */
    public void put(INode variable, Object obj)
            throws IllegalArgumentException {

        if (variable == null) {
            throw new IllegalArgumentException();
        }
        INode node = null;
        if (obj instanceof String) {
            String tmp = (String) obj;
            try {
                node = JatsIO.parseString(variable.getVariableType(), tmp);
            } catch (Exception e) {
                e.printStackTrace();
                throw new IllegalArgumentException();
            }
        } else {
            if (obj instanceof INode) {
                node = (INode) obj;
            }
        }
        mappings.put(variable, node);
    }

    /**
     * Este mtodo diz se a varivel <code>var</code> j se encontra
     * mapeada a algum valor neste ResultSet
     *
     * @param  var  Variavel que se quer verificar a existencia
     *
     * @return   <code>true</code>, se <coe>var</code> j estiver mapeada a algum valor<br>
     *           <code>false</code>, caso contrrio
     */
    public boolean alreadyMatched(INode var) {
        boolean retorno = false;
        if (var != null && var.isVariable()) {
            retorno = (mappings.get(var) != null);
        }
        return retorno;
    }


    /**
     * Retorna um clone deste objeto.
     */
    public Object clone() {
        return new ResultSet((JaTSHashTable) mappings.clone());
    }


    /**
     * Diz se a  varivel <code>var</code> encontra-se mapeada a algum valor
     *
     * @param  var  Variavel que pretendemos verificar se est mapeada a
     *              algum valor
     *
     * @return  true, caso a varivel esteja mapeada
     *          false, caso contrrio
     */
    public boolean hasValue(INode var) {
        return this.get(var) != null;
    }


    /**
     * Retorna o valor que se encontra mapeado  varivel <code>var</code>
     *
     * @param  var       Variavel, da qual pretendemos obter o valor associado
     *
     * @return  Valor associado  varivel passada como parmetro
     */
    public Object get(INode var) {
        Object retorno = null;
        if (var != null && (var.isVariable())) {
            retorno = mappings.get(var);
        }

        return retorno;
    }

    /**
     * Retorna o valor que se encontra mapeado  varivel <code>var</code>
     *
     * @param  varName   Nome da variavel, da qual pretendemos obter o valor associado
     *
     * @return       Valor associado  varivel de nome varName, passado como parmetro
     */
    public Object get(String varName) throws VariableNotFoundException {
        Object retorno = null;
        Iterator it = this.getVariables();
        boolean encontrou = false;
        INode tmpVar = null;
        if (Util.checkString(varName)) {
            while (it.hasNext() && !encontrou) {
                tmpVar = (INode) it.next();
                String tmpName = tmpVar.getVariableName();
                if (tmpName.equals(varName)) {
                    retorno = this.get(tmpVar);
                    encontrou = true;
                }
            }
        }
        if (!encontrou) {
            throw new VariableNotFoundException("Variable " + varName + " not found");
        }
        return retorno;

    }

    /**
     * Retorna um iterator para que se possa navegar pelas variveis do
     * ResultSet
     */
    public Iterator getVariables() {
        return this.mappings.getVariables();
    }


    /**
     * Combina dois ResultSets, se estes forem compatveis. Caso no
     * sejam, uma exceo  levantada.
     */
    public void merge(ResultSet results) {
        // se for diferente de null, faz o join, caso contrrio, no faz nada
        if (results != null) {
            this.mappings = this.join(results.getMappings());
        }
    }

    /**
     * Faz o cruzamento entre duas solues, retornando uma nova soluo
     * derivada da juno desta soluo e com a soluo passada como parmetro.<BR>
     * Caso haja uma varivel comum s duas solues com valores diferentes
     * associados, ser retornado um conjunto vazio, indicando que as duas solues
     * so conflitantes.
     *
     * @param mappings    Soluo a ser combinada com a representada por este objeto
     */
    private JaTSHashTable join(JaTSHashTable mappings) {
        JaTSHashTable retorno = new JaTSHashTable();
        boolean conflitou = false;
        // variaveis do objeto this
        Iterator thisVariables = this.getVariables();
        Iterator mappingsVariables = null;
        if (mappings != null) {
            // variveis do objeto mappings
            // percorre as variveis de this
            while (thisVariables.hasNext() && !conflitou) {
                // escolhe uma variavel do objeto this
                INode tmpVar = (INode) thisVariables.next();
                // pega todas as variveis de mappings
                mappingsVariables = mappings.getVariables();
                // varre todas as variveis de mappings
                while (!conflitou && mappingsVariables.hasNext()) {
                    INode tmp2 = (INode) mappingsVariables.next();
                    if (tmpVar.equals(tmp2)) {
                        // se as variaveis so iguais, verifica-se os conteudos
                        // das mesmas
                        conflitou = (!(Util.equals(this.get(tmpVar), mappings.get(tmp2))));
                    }
                }
            }
            // se no houve conflitos, adiciona os pares das duas solues 
            // soluo final
            retorno.putAll(this.getMappings());
            retorno.putAll(mappings);


        } else {
            throw new IllegalArgumentException("Argumento mappings nulo");
        }
        return retorno;

    }


    /**
     * Remove um casamento do conjunto resultado.
     *
     * @param  variable  Variavel que sera removida, juntamente
     *                   com o valor a ela associado
     */
    public void removeResult(INode variable) {
        if (variable == null || !variable.isVariable()) {
            throw new IllegalArgumentException("A varivel fornecida  invlida.");
        }
        this.mappings.remove(variable);
    }


    /**
     * Remove um par var-valor do conjunto resultado. A varivel,  aquela
     * especificada no parametro.
     * O tipo especificado deve ser um tipo JaTS vlido, isto , deve estar
     * especificado em INode
     *
     * @param   strVar  String correspondente a varivel a ser removida
     * @param   type    tipo da varivel.
     */
    public void remove(String strVar, int type) {
        if (!Util.checkString(strVar)) {
            throw new IllegalArgumentException("A varivel fornecida  invlida: " + strVar);
        } else {
            INode variable = JatsIO.parseVariable(strVar, type);
            this.mappings.remove(variable);
        }

    }


    /**
     * Remove um par var-valor do conjunto resultado. A varivel,  aquela
     * especificada no parametro.
     * O tipo especificado deve ser um tipo JaTS vlido, isto , deve estar
     * especificado em INode
     *
     * @param   var  correspondente a varivel a ser removida
     */
    public Object remove(Object var) {
        INode variable = null;
        if (var == null) {
            throw new IllegalArgumentException("A varivel fornecida  invlida: ");
        } else {
            if (var instanceof INode) {
                variable = (INode) var;
                this.mappings.remove(variable);
            }
        }
        return var;
    }



    /**
     * Inclui par var-valor no ResultSet. A String var deve corresponder
     * ao nome de uma varivel j incluida.
     *
     * @param var
     * @param value
     */
    public void put(String var, Object value) {

        INode tmpVar = null;
        Iterator it = this.getVariables();
        boolean inseriu = false;
        if (Util.checkString(var)) {
            while (it.hasNext() && !inseriu) {
                tmpVar = (INode) it.next();
                String tmpName = tmpVar.getVariableName();

                if (tmpName.equals(var)) {
                    this.put(tmpVar, value);
                    inseriu = true;
                }
            }
        }
        if (!inseriu) {
            throw new IllegalArgumentException();
        }
    }


    /**
     * Inclui par var-valor no ResultSet.
     *
     * @param var
     * @param value
     * @param type Tipo da varivel que est sendo inserida
     */
    public void put(String var, Object value, int type) {

        if (Util.checkString(var)) {
            try {
                INode varNode = JatsIO.parseVariable(var, type);


                this.put(varNode, value);

            } catch (Exception e) {
                throw new IllegalArgumentException("Tipo invlido para variavel " + e.getMessage());
            }
        }
    }


    /**
     *  Resets all the mappings of the ResultSet, they are all removed
     */
    public void reset() {
        mappings.reset();
    }

    /**
     *  Description of the Method
     *
     * @param  rs  Description of the Parameter
     */
    public void reloadValues(ResultSet rs) {
        if (rs != null) {
            reset();
            mappings.putAll(rs.getMappings());
        }
    }


    /**
     * Adiciona uma declarao ao conjunto de declaraes mapeadas  varivel
     * especificada no parmetro. Ateno: o cdigo do tipo passado como
     * parmetro deve ser o cdigo do tipo da declarao,
     * Ex: JaTSNode.FIELD_DECLARATION.
     * @param declaration  Declarao a ser adicionada (de atributo, de mtodo,
     *        de mtodo, de construtor ou de importao ) <br>
     *
     * @param varName Nome da varivel do tipo Conjunto de declaraes
     *       (de mtodos, de atributos, de importaes ou de construtores).
     *       Ex: "#FDS", "#CDS", "#MTDS" <br>
     *
     * @param type  Tipo da declarao que vai ser adicionada. Pode ser um dos
     *        quatro tipos abaixo:
     * <blockquote>
     * 	      JaTSNode.FIELD_DECLARATION <br>
     * 	      JaTSNode.METHOD_DECLARATION <br>
     * 	      JaTSNode.CONSTRUCTOR_DECLARATION <br>
     * 	      JaTSNode.IMPORT_DECLARATION <br>
     * </blockquote>
     */
    public void addDeclaration(Declaration declaration, String varName, int type) {

        INode tmpDecSet = null;

        try {
            tmpDecSet = (INode) this.get(varName);
            if (tmpDecSet instanceof DeclarationSet) {
                ((DeclarationSet) tmpDecSet).add(declaration);

                DeclarationSet tmpVar = Util.getDeclarationSet(type);
                tmpVar.setVariableName(varName);

                // remove o antigo par
                this.removeResult(tmpVar);

                // adiciona o novo mapeamento - var-valor, onde o novo ds contm
                // a declaration passada como parametro
                this.put(tmpVar, tmpDecSet);
            }
        } catch (VariableNotFoundException e) {
            DeclarationSet decSet =
                    Util.getDeclarationSet(declaration.getTypeIdentifier());
            decSet.add(declaration);
            this.put(varName, decSet, type);
        }
    }

    /**
     * Adiciona uma declarao ao conjunto de declaraes mapeadas  varivel
     * especificada no parmetro. Ateno: o cdigo do tipo passado como
     * parmetro, deve ser o cdigo do tipo da declarao. Ex: <br>
     * Se temos a varivel "#FDS" e queremos adicionar a ela a declarao
     * "private Conta conta;". Ento, o modo correto de invocar este mtodo :
     * <br>
     * addDeclaration("private Conta conta;", "#FDS", INode.FIELD_DECLARATION);
     *
     * @param strDec  String correpondente a declarao a ser adicionada. Ex:
     * "private int numero;", "public void setNumero();" <br>
     *
     * @param varName Nome da varivel do tipo Conjunto de declaraes
     * (de mtodos, de atributos, de importaes ou de construtores). Ex:
     * "#FDS", "#CDS", "#MTDS" <br>
     *
     * @param type  Tipo da declarao que vai ser adicionada. Pode ser um dos
     * quatro tipos abaixo:
     * <blockquote>
     * 	      JaTSNode.FIELD_DECLARATION <br>
     * 	      JaTSNode.METHOD_DECLARATION <br>
     * 	      JaTSNode.CONSTRUCTOR_DECLARATION <br>
     * 	      JaTSNode.IMPORT_DECLARATION <br>
     * </blockquote>
     */
    public void addDeclaration(String strDec, String varName, int type) {
        Declaration declaration = null;
        try {
            INode tmpDec = JatsIO.parseString(type, strDec);
            if (tmpDec instanceof Declaration) {
                declaration = (Declaration) tmpDec;
                this.addDeclaration(declaration, varName, type);
            } else {
                throw new IllegalArgumentException("Erro ao fazer o parsing da "
                        + "String: " + strDec +
                        ". Tipo invlido!");
            }
        } catch (ParseException pe) {
            throw new IllegalArgumentException("Erro ao fazero parsing da "
                    + "String: " + strDec);
        }
    }

    /**
     * Returns the number of key-value mappings in this map.  If the
     * map contains more than <tt>Integer.MAX_VALUE</tt> elements, returns
     * <tt>Integer.MAX_VALUE</tt>.
     *
     * @return the number of key-value mappings in this map.
     */
    public int size() {
        return this.mappings.getSize();
    }

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     *
     * @return <tt>true</tt> if this map contains no key-value mappings.
     */
    public boolean isEmpty() {
        return (this.size() == 0);
    }

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the specified
     * key.  More formally, returns <tt>true</tt> if and only if
     * this map contains at a mapping for a key <tt>k</tt> such that
     * <tt>(key==null ? k==null : key.equals(k))</tt>.  (There can be
     * at most one such mapping.)
     *
     * @param key key whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map contains a mapping for the specified
     *         key.
     *
     * @throws ClassCastException if the key is of an inappropriate type for
     * 		  this map (optional).
     * @throws NullPointerException if the key is <tt>null</tt> and this map
     *            does not not permit <tt>null</tt> keys (optional).
     */
    public boolean containsKey(Object key) {
        Set keySet = this.keySet();
        return keySet.contains(key);
    }

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.  More formally, returns <tt>true</tt> if and only if
     * this map contains at least one mapping to a value <tt>v</tt> such that
     * <tt>(value==null ? v==null : value.equals(v))</tt>.  This operation
     * will probably require time linear in the map size for most
     * implementations of the <tt>Map</tt> interface.
     *
     * @param value value whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value.
     * @throws ClassCastException if the value is of an inappropriate type for
     * 		  this map (optional).
     * @throws NullPointerException if the value is <tt>null</tt> and this map
     *            does not not permit <tt>null</tt> values (optional).
     */
    public boolean containsValue(Object value) {
        Collection allValues = this.values();
        return allValues.contains(value);
    }

    /**
     * Returns the value to which this map maps the specified key.  Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * <p>More formally, if this map contains a mapping from a key
     * <tt>k</tt> to a value <tt>v</tt> such that <tt>(key==null ? k==null :
     * key.equals(k))</tt>, then this method returns <tt>v</tt>; otherwise
     * it returns <tt>null</tt>.  (There can be at most one such mapping.)
     *
     * @param key key whose associated value is to be returned.
     * @return the value to which this map maps the specified key, or
     *	       <tt>null</tt> if the map contains no mapping for this key.
     *
     * @throws ClassCastException if the key is of an inappropriate type for
     * 		  this map (optional).
     * @throws NullPointerException key is <tt>null</tt> and this map does not
     *		  not permit <tt>null</tt> keys (optional).
     *
     * @see #containsKey(Object)
     */
    public Object get(Object key) {
        Object retorno = null;
        if (key instanceof String || key instanceof INode) {
            try {
                retorno = this.get(key.toString());
            } catch (VariableNotFoundException e) {
                retorno = null;
            }
        }

        return retorno;
    }

    /**
     * Associates the specified value with the specified key in this map
     * (optional operation).  If the map previously contained a mapping for
     * this key, the old value is replaced by the specified value.  (A map
     * <tt>m</tt> is said to contain a mapping for a key <tt>k</tt> if and only
     * if {@link #containsKey(Object) m.containsKey(k)} would return
     * <tt>true</tt>.))
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the map previously associated <tt>null</tt>
     *	       with the specified key, if the implementation supports
     *	       <tt>null</tt> values.
     *
     * @throws UnsupportedOperationException if the <tt>put</tt> operation is
     *	          not supported by this map.
     * @throws ClassCastException if the class of the specified key or value
     * 	          prevents it from being stored in this map.
     * @throws IllegalArgumentException if some aspect of this key or value
     *	          prevents it from being stored in this map.
     * @throws NullPointerException this map does not permit <tt>null</tt>
     *            keys or values, and the specified key or value is
     *            <tt>null</tt>.
     */
    public Object put(Object key, Object value) {
        Object retorno = null;
        if (key instanceof String || key instanceof INode) {
            retorno = this.get(key);
            this.put(key.toString(), value);
        }
        return retorno;
    }

    /**
     * Copies all of the mappings from the specified map to this map
     * (optional operation).  The effect of this call is equivalent to that
     * of calling {@link #put(Object,Object) put(k, v)} on this map once
     * for each mapping from key <tt>k</tt> to value <tt>v</tt> in the
     * specified map.  The behavior of this operation is unspecified if the
     * specified map is modified while the operation is in progress.
     *
     * @param t Mappings to be stored in this map.
     *
     * @throws UnsupportedOperationException if the <tt>putAll</tt> method is
     * 		  not supported by this map.
     *
     * @throws ClassCastException if the class of a key or value in the
     * 	          specified map prevents it from being stored in this map.
     *
     * @throws IllegalArgumentException some aspect of a key or value in the
     *	          specified map prevents it from being stored in this map.
     * @throws NullPointerException the specified map is <tt>null</tt>, or if
     *         this map does not permit <tt>null</tt> keys or values, and the
     *         specified map contains <tt>null</tt> keys or values.
     */
    public void putAll(Map t) {
        Set keySet = t.keySet();
        Iterator it = keySet.iterator();
        while (it.hasNext()) {
            String var = (String) it.next();
            Object value = t.get(var);
            this.put(var, value);
        }

    }

    /**
     * Removes all mappings from this map (optional operation).
     *
     * @throws UnsupportedOperationException clear is not supported by this
     * 		  map.
     */
    public void clear() {
        this.reset();
    }

    /**
     * Returns a set view of the keys contained in this map.  The set is
     * backed by the map, so changes to the map are reflected in the set, and
     * vice-versa.  If the map is modified while an iteration over the set is
     * in progress, the results of the iteration are undefined.  The set
     * supports element removal, which removes the corresponding mapping from
     * the map, via the <tt>Iterator.remove</tt>, <tt>Set.remove</tt>,
     * <tt>removeAll</tt> <tt>retainAll</tt>, and <tt>clear</tt> operations.
     * It does not support the add or <tt>addAll</tt> operations.
     *
     * @return a set view of the keys contained in this map.
     */
    public Set keySet() {
        return this.mappings.keySet();
    }

    /**
     * Retorna uma viso de todos os valores contidos no ResultSet. Alteraes
     * nesta coleo no alteram o contedo do resultSet.
     *
     * @return Uma viso dos valores contidos neste ResultSet
     */
    public Collection values() {
        return this.mappings.values();
    }

    /**
     * Returns a set view of the mappings contained in this map.  Each element
     * in the returned set is a {@link Map.Entry}.  The set is backed by the
     * map, so changes to the map are reflected in the set, and vice-versa.
     * If the map is modified while an iteration over the set is in progress,
     * the results of the iteration are undefined.  The set supports element
     * removal, which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not support
     * the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a set view of the mappings contained in this map.
     */
    public Set entrySet() {
        return this.mappings.entrySet();
    }
}
