/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import cin.jats.engine.parser.nodes.ContextDeclaration;
import cin.jats.engine.parser.nodes.INode;
import cin.jats.engine.parser.nodes.JBlock;
import cin.jats.engine.parser.nodes.JExpression;
import cin.jats.engine.parser.nodes.JIfStatement;
import cin.jats.engine.parser.nodes.JStatementList;
import cin.jats.engine.parser.nodes.JWhileStatement;
import cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException;
import cin.jats.engine.parser.nodes.exceptions.NodesNotMatchedException;

/**
 * This class processes a selection operation, visiting the nodes that are
 * involved in this kind of manipulation. All the traverse and process logic
 * are concentrated in this class.
 * Selection operations include code extraction and code insertion.
 * 
 * @author Gustavo Santos
 * @version 02/09/05
 */
public class SelectionProcessor {

	//TODO estender os metodos extractNodeContext e setNodeContext
	//     e para suportar todos os tipos internos a metodos
	
	/**
	 * Extracts a portion of code from a context. The concrete code
	 * should be somewhere inside the node. This method finds the 
	 * corresponding code specified by contDec and extracts such 
	 * code from the concrete context.
	 * 
	 * @param node The concrete code
	 * @param contDec The context declaration (informs code position)
	 * @return The concrete code selection
	 * @throws NodesNotMatchedException
	 */
	public static INode extractCode(INode node,
			ContextDeclaration context) throws NodesNotMatchedException {
		
	    if (node instanceof JIfStatement)
	    	return extractCode((JIfStatement)node, context);
	    else if (node instanceof JWhileStatement)
	    	return extractCode((JWhileStatement)node, context);
	    else if (node instanceof JBlock)
	    	return extractCode((JBlock)node, context);
	    else if (node instanceof JStatementList)
	    	return extractCode((JStatementList)node, context);
	    else
	    	throw new IllegalStateException("extractCode not implemented for "+node.getClass());		
	}

	public static INode extractCode(JBlock node,
			ContextDeclaration contDec) throws NodesNotMatchedException {
		
		return extractCode(node.getStatements(), contDec);
	}

	public static INode extractCode(JStatementList node,
			ContextDeclaration contDec) throws NodesNotMatchedException {
		INode result = null;
		INode code = contDec.getCodeOfInterest();
		Position position = contDec.getCodePosition();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());
		
		// if the point of interest is in this structure
		if (position.getCoordinates().size() == 1) {
			try {
				result = retrieveConcreteRefFromNodeList(node.getStatements(),
						(JStatementList) code, posIndex);
			} catch (InconsistentNodeException e) {
				//never occurs
			}
		} else {
			ContextDeclaration newContDec = new ContextDeclaration(contDec
					.getContext(), contDec.getCodeOfInterest());
			newContDec.setCodePosition(
					contDec.getCodePosition().retractLevel());
			result = extractCode(node.elementAt(posIndex), newContDec);
		}

		return result;
	}
	
	/**
	 * Extracts a code portion from an if statement
	 * 
	 * @param node The If node
	 * @param contDec The context
	 * @return The extracted code
	 * @throws NodesNotMatchedException
	 * @see INode extractCode(INode,ContextDeclaration)
	 */
	public static INode extractCode(JIfStatement node,
			ContextDeclaration contDec) throws NodesNotMatchedException {
		INode result = null;
		INode code = contDec.getCodeOfInterest();
		Position position = contDec.getCodePosition();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());
		
		// if the point of interest is in this structure
		if (position.getCoordinates().size() == 1) {

			try {
				//condition
				if (posIndex == 0) {
					result = node.getExpression();
					node.setExpression(null);
				} 
				//if block
				else if (posIndex == 1) {
					result = node.getIfStatement();
					node.setIfStatement(null);
				}
				//else block
				else if (posIndex == 2) {
					result = node.getElseStatement();
					node.setElseStatement(null);
				} else {
					throw new NodesNotMatchedException(node, contDec);
				}
				
				if (code instanceof JStatementList
						&& !(result instanceof JStatementList)) {
					if (((JStatementList)code).size() == 1) {
						INode tmp = result;
						result = new JStatementList();
						((JStatementList)result).add(tmp);
					} else {
						throw new NodesNotMatchedException();
					}
				}
				
				//compares the code
				result.match(code, new ResultSet());
				
			} catch (InconsistentNodeException e) {
				//never occurs
			}
			
		} else {
			ContextDeclaration newContDec = new ContextDeclaration(contDec
					.getContext(), contDec.getCodeOfInterest());
			newContDec.setCodePosition(
					contDec.getCodePosition().retractLevel());
			switch (posIndex) {
			case 0: 
				result = extractCode(node.getExpression(), newContDec);
				break;
			case 1: 
				result = extractCode(node.getIfStatement(), newContDec);
				break;
			case 2: 
				result = extractCode(node.getElseStatement(), newContDec);
				break;
			}
		}

		return result;
	}
	
	/**
	 * Extracts a code portion from a while statement
	 * 
	 * @param node
	 * @param contDec
	 * @return
	 * @throws NodesNotMatchedException
	 */
	public static INode extractCode(JWhileStatement node,
			ContextDeclaration contDec) throws NodesNotMatchedException {
		INode result = null;
		INode code = contDec.getCodeOfInterest();
		Position position = contDec.getCodePosition();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());
		
		// if the point of interest is in this structure
		if (position.getCoordinates().size() == 1) {

			try {
				//condition
				if (posIndex == 0) {
					result = node.getExpression();
					node.setExpression(null);
				} 
				//main block
				else if (posIndex == 1) {
					result = node.getStatement();
					node.setStatement(null);
				} else {
					throw new NodesNotMatchedException(node, contDec);
				}
				
				if (code instanceof JStatementList
						&& !(result instanceof JStatementList)) {
					if (((JStatementList)code).size() == 1) {
						INode tmp = result;
						result = new JStatementList();
						((JStatementList)result).add(tmp);
					} else {
						throw new NodesNotMatchedException();
					}
				}
				
				//compares the code
				result.match(code, new ResultSet());
				
			} catch (InconsistentNodeException e) {
				//never occurs
			}
			
		} else {
			ContextDeclaration newContDec = new ContextDeclaration(contDec
					.getContext(), contDec.getCodeOfInterest());
			newContDec.setCodePosition(
					contDec.getCodePosition().retractLevel());
			switch (posIndex) {
			case 0: 
				result = extractCode(node.getExpression(), newContDec);
				break;
			case 1: 
				result = extractCode(node.getStatement(), newContDec);
				break;
			}
		}

		return result;
	}
	
	
	private static JStatementList retrieveConcreteRefFromNodeList(NodeList context, JStatementList code,
			int stmIndex) throws NodesNotMatchedException, IllegalArgumentException, InconsistentNodeException {
		JStatementList result = new JStatementList();
		ResultSet baitRSet = new ResultSet();
		
		for (int i = stmIndex, j = 0; j < code.size(); i++, j++) {			
			
			//compares the statements and throws an exception if fails
			context.elementAt(i).match(code.elementAt(j), baitRSet);
			
			//construct the new list
			result.add(context.elementAt(i));
		}
		for (int i = 0; i < code.size(); i++) {						
			context.removeElementAt(stmIndex);
		}
		
		return result;
	}

	
	/**
	 * Inserts a portion of code in a context.
	 * The statement list must have the position value.
	 * 
	 * @param node The node where the code will be inserted
	 * @param selection The code to be inserted
	 */
	public static void insertCode(INode node, JStatementList selection) {
	    if (node instanceof JIfStatement)
	    	insertCode((JIfStatement)node, selection);
	    else if (node instanceof JWhileStatement)
	    	insertCode((JWhileStatement)node, selection);
	    else if (node instanceof JBlock)
	    	insertCode((JBlock)node, selection);
	    else if (node instanceof JStatementList)
	    	insertCode((JStatementList)node, selection);
	    else
	    	throw new IllegalStateException("insertCode not implemented for "+node.getClass());		
	}
	
	public static void insertCode(JBlock node, JStatementList selection) {
		insertCode(node.getStatements(), selection);
	}

	public static void insertCode(JIfStatement node, JStatementList selection) {
		Position position = selection.getPosition();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());

		if (position.getCoordinates().size() == 1) {
			
			if (posIndex == 0) {
				node.setExpression((JExpression) selection);
			} else if (posIndex == 1) {
				node.setIfStatement(selection);
			} else if (posIndex == 2) {
				node.setElseStatement(selection);
			} else {
				//pau!
			}
			
		} else {
			selection.setPosition(position.retractLevel());
			switch (posIndex) {
			case 0:
				insertCode(node.getExpression(), selection);
				break;
			case 1:
				insertCode(node.getIfStatement(), selection);
				break;
			case 2:
				insertCode(node.getElseStatement(), selection);
				break;
			}
		}
	}

	public static void insertCode(JWhileStatement node, JStatementList selection) {
		Position position = selection.getPosition();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());

		if (position.getCoordinates().size() == 1) {
			
			if (posIndex == 0) {
				node.setExpression((JExpression) selection);
			} else if (posIndex == 1) {
				node.setStatement(selection);
			} else {
				throw new RuntimeException("Invalid index for while statement node");
			}
			
		} else {
			selection.setPosition(position.retractLevel());
			switch (posIndex) {
			case 0:
				insertCode(node.getExpression(), selection);
				break;
			case 1:
				insertCode(node.getStatement(), selection);
				break;
			}
		}
	}
	
	public static void insertCode(JStatementList node, JStatementList selection) {
		Position position = selection.getPosition();
		NodeList context = node.getStatements();
		PositionElement coord = (PositionElement) position.getCoordinates()
				.elementAt(0);
		int posIndex = Integer.parseInt(coord.getValue());

		if (position.getCoordinates().size() == 1) {
			NodeList result = new NodeList();
			
			if (context.size() >= posIndex) {
				for (int i = 0; i < context.size(); i++) {
					if (i == posIndex) {
						for (int j = 0; j < selection.size(); ++j) {
							result.addElement(selection.elementAt(j));
						}
					}
					result.addElement(context.elementAt(i));
				}
				//if should be added at the end
				if (posIndex == context.size()) {
					for (int j = 0; j < selection.size(); ++j) {
						result.addElement(selection.elementAt(j));
					}
				}
			}
		
			node.setStatements(result);
			
		} else {
			selection.setPosition(position.retractLevel());
			insertCode(context.elementAt(posIndex), selection);
			
		}
	
	}
	
}
