/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.util;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import cin.jats.engine.parser.nodes.INode;


/**
 Esta classe implementa uma lista de objetos do tipo <code>INode</code>.
 @see INode
 @author Adeline de Sousa
 */
public class VectorVariables implements Cloneable {


    /**
     * A lista onde sero armazenados os elementos.
     */
    private Vector elements;


    /**
     * Construtor vazio da classe.
     */
    public VectorVariables() {
        this.elements = new Vector();
    }


    /**
     * Construtor que recebe como argumento a capacidate mxima da lista de ns.
     *
     * @param capacity um inteiro correspondendo  capacidade mxima da lista de ns.
     */
    public VectorVariables(int capacity) {
        this.elements = new Vector(capacity);
    }


    /**
     * Construtor que recebe como argumento um objeto do tipo <code>Vector</code> contendo
     * apenas objetos que implementam INode
     */
    protected VectorVariables(Vector objects) throws IllegalArgumentException {
        if (objects == null) throw new IllegalArgumentException();
        int i = 0;

        while (i < objects.size()) {
            if (!(objects.elementAt(i) instanceof INode)) {
                throw new IllegalArgumentException();
            }
            i++;
        }

    }


    /**
     * Adiciona um elemento  lista.
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @exception IllegalArgumentException Levantada quando o argumento
     *			 <code>node</code> for <code>null</code>.
     */
    public void addElement(INode node) throws IllegalArgumentException {
        if (node == null) throw new IllegalArgumentException();
        this.elements.addElement(node);
    }


    /**
     * Adiciona toda a lista passada como parametro aa lista atual
     * @param list    Um objeto qualquer do tipo <code>NodeList</code>.
     * @exception IllegalArgumentException Levantada quando o argumento
     *			 <code>list</code> for <code>null</code>.
     */
    public void addAll(VectorVariables list) throws IllegalArgumentException {
        if (list == null) {
            throw new IllegalArgumentException("A Lista passada eh nula");
        }
        this.elements.addAll(list.elements);
    }

    /**
     * Retorna o indice do elemento especificado no parametro
     *
     * @param elem  Elemento cujo indice eh procurado
     */
    public int indexOf(INode elem) {
        return this.elements.indexOf(elem);
    }


    /**
     * Retorna a capacidade desta lista.
     * @return Um inteiro correspondendo  capacidade da lista.
     */
    public int capacity() {
        return this.elements.capacity();
    }


    /**
     * Retorna um clone desta lista e de todos os elementos nela presentes.
     * @return Uma cpia funda desde objeto.
     */
    public Object clone() throws CloneNotSupportedException {
        VectorVariables nl = null;

        try {
            nl = new VectorVariables((Vector) this.elements.clone());
        } catch (IllegalArgumentException e) {
            throw new CloneNotSupportedException();
        }

        return nl;
    }


    /**
     * Retorna o n que se encontra na <code>pos</code>-sima posio da lista.
     * @param  pos Um inteiro correspondendo  posio desejada.
     * @return Um objeto do tipo <code>INode</code> que se encontra na posio
     *		  desejada.
     */
    public INode elementAt(int pos) {
        return (INode) this.elements.elementAt(pos);
    }


    /**
     * Insere um elemento na lista na posio especificada e d um shift para a
     * direita na posies do item que esiver naquela posio e nas daqueles que .
     * estiverem  direita dela.
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @param pos Um inteiro correspondendo  posio onde o n deve ser
     *		 inserido.
     * @exception IllegalArgumentException Levantada quando o
     *			 argumento <code>node</code> for <code>null</code> ou o
     *			 argumento <code>pos</code> for negativo.
     */
    public void insertElementAt(INode node, int pos) throws IllegalArgumentException {
        if (node == null || pos < 0) throw new IllegalArgumentException();
        this.elements.insertElementAt(node, pos);
    }


    /**
     * Diz se a lista es vazia.
     * @return Um boolean que, caso tenho o valor <code>true</code>, indica
     *			  que a lista est vazia.
     */
    public boolean isEmpty() {
        return this.elements.isEmpty();
    }


    /**
     * Retorna um java.util.Iterator para que se possa percorrer a estrutura
     * @return um iterador para percorrer os elementos da lista
     */
    public Iterator iterator() {
        return this.elements.iterator();
    }

    /**
     Retorna um java.util.Iterator para que se possa percorrer a estrutura
     @return um iterador para percorrer os elementos da lista
     */
    public JaTSIterator getIterator() {
        JaTSIterator retorno = new JaTSIterator(this.elements.iterator());
        return retorno;
    }


    /**
     * Verifica se o elemento passado como parametro pertence a lista
     *
     * @param elem Elemento procurado
     *
     * @return true, caso o elemento passado pertenca a lista
     *         false, caso contrrio
     */
    public boolean contains(INode elem) {
        boolean retorno = false;
        JaTSIterator it = this.getIterator();
        INode node = null;
        String tmpNomeVar = null;
        String nomeVar = null;
        if (elem.isVariable()) {
            nomeVar = elem.getVariableName();
        }

        while (it.hasNext() && !retorno) {
            node = it.next();
            if (node != null) {
                tmpNomeVar = node.getVariableName();
                if (nomeVar != null && nomeVar.equals(tmpNomeVar)) {
                    retorno = true;
                }
            }
        }
        return retorno;
    }


    /**
     * Remove o elemento que se encontra na  posio <code>pos</code> desta lista.
     * @param pos Um inteiro correspondendo  posio do n que deve ser
     *		 removido.
     * @exception IllegalArgumentException Levantada quando o
     *			 argumento <code>pos</code> for negativo.
     */
    public void removeElementAt(int pos) throws IllegalArgumentException {
        if (pos < 0) throw new IllegalArgumentException();
        this.elements.removeElementAt(pos);
    }


    /**
     * Coloca um elemento na <code>pos</code>-sima posio da lista,
     * sobreescrevendo o elemento que estava nela.
     * @param node Um objeto qualquer do tipo <code>INode</code>.
     * @param pos Um inteiro correspondendo  posio onde o n deve ser
     *		 inserido.
     * @exception IllegalArgumentException Levantada quando o
     *			 argumento <code>node</code> for <code>null</code> ou o
     *			 argumento <code>pos</code> for negativo.
     */
    public void setElementAt(INode node, int pos) throws IllegalArgumentException {
        if (node == null || pos < 0 || pos > this.size()) throw new IllegalArgumentException();
        this.elements.setElementAt(node, pos);
    }


    /**
     * Retorna o nmero de elementos da lista.
     * @return Um inteiro correspondendo ao nmero de elementos da lista.
     */
    public int size() {
        return this.elements.size();
    }


    /**
     * Seta o novo tamanho da lista.
     * @param s Um inteiro correspondendo ao novo tamanho da lista.
     */
    public void setSize(int s) {
        this.elements.setSize(s);
    }


    /**
     * Retorna a coleo correspondente a esta NodeList
     *
     * @return a coleo correspondente a esta NodeList
     */
    public Collection getElements() {
        return this.elements;
    }


    /**
     * Compara dois objetos do tipo NodeList e diz se estes so iguais
     */
    public boolean equals(Object obj) {

        boolean retorno = false, achou = false;
        if (obj instanceof VectorVariables) {
            VectorVariables nlist = (VectorVariables) obj;
            if (this.size() == nlist.size()) {
                int max = this.size();
                int i = 0;
                while (i < max && !achou) {
                    if (this.elementAt(i).equals(nlist.elementAt(i))) {
                        i++;
                    } else {
                        // sai do lao, encontrou um elemento diferente
                        achou = true;
                    }
                }
                if (!achou) {
                    retorno = true;
                }
            }
        }
        return retorno;
    }
}// fim da classe
