/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.visitors;

import java.io.IOException;
import java.io.Writer;
import java.util.Properties;
import java.util.Vector;

import cin.jats.engine.printing.PrettyPrinterProperties;

import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.util.NodeList;
import cin.jats.engine.util.StringStack;
import cin.jats.engine.util.Util;

/**
 * Classe que implementa a interface <code>IVisitor</code> por meio de um
 * visitor que  realiza o <i>pretty printing</i> de uma rvore sinttica.
 *
 * @see JaTSNode
 * @see IVisitor
 *
 * @author Fernando Castor
 * @author Adeline de Sousa Silva
 * @author Gustavo Santos
 * 
 * @version 02/12/2005
 */
public class PrettyPrintingVisitor extends AbstractVisitor {

    // Pilha responsvel por guardar o estado atual da impresso.
    private StringStack state;

    // configuraes do pretty-printer
    private Properties properties;

    // configuraes
    private boolean elseOnNewLine;
    private boolean catchOnNewLine;
    private boolean finallyOnNewLine;
    private boolean whileOnNewLine;
    protected boolean bracesOnNewLine;

    // indica se tipos JaTS podem aparecer no codigo gerado
    protected boolean jatsTypesEnabled = true;

    // guarda o tamanho do espao a ser usado entre uma palavra e outra no
    // cdigo
    private String space = " ";
    
    private static String lineBreak = "\n";

    // guarda os caracteres de identao do bloco.
    protected String indentString;


    /**
     * Construtor vazio da classe.
     */
    public PrettyPrintingVisitor() {
        this.state = new StringStack();
        initProperties();
    }


    public void setProperties(Properties prop) {
        if (prop != null) {
            this.properties = prop;
            // a cada vez que o atributo properties for modificado, os atributos
            // que corresondem a propriedades do prettyprinter, devem
            // modificados tambem
            initProperties();
        } else {
            this.properties = new Properties();
            initProperties();
        }
    }


    /**
     * Inicializa as propriedades do PrettyPrinter
     */
    private void initProperties() {
        if (properties != null) {
            elseOnNewLine =
                    getBooleanProperty(PrettyPrinterProperties.ELSE_ON_NEW_LINE, "true");
            catchOnNewLine =
                    getBooleanProperty(PrettyPrinterProperties.CATCH_ON_NEW_LINE, "true");
            finallyOnNewLine =
                    getBooleanProperty(PrettyPrinterProperties.FINALLY_ON_NEW_LINE, "true");
            whileOnNewLine =
                    getBooleanProperty(PrettyPrinterProperties.WHILE_ON_NEW_LINE, "false");
            bracesOnNewLine =
                    getBooleanProperty(PrettyPrinterProperties.BRACES_ON_NEW_LINE, "false");
        } else {
            initPropertiesUsingDefault();
        }

    }


    /**
     * Usado no caso de as propriedades no terem sido informadas
     */
    private void initPropertiesUsingDefault() {
        indentString = "    ";
        elseOnNewLine = true;
        catchOnNewLine = true;
        finallyOnNewLine = true;
        whileOnNewLine = true;
        bracesOnNewLine = false;
    }


    /**
     * Retorna o valor das propriedades booleanas
     *
     * @param prop      Propriedade, para a qual se deseja o valor
     * @param defValue  Valor default
     * @return Valor das propriedades booleanas
     */
    private boolean getBooleanProperty(String prop, String defValue) {
        return Boolean.valueOf(properties.getProperty(prop, defValue)).booleanValue();
    }


    /**
     * Escreve em um writer o contedo do topo da pilha.
     *
     * @param output  Um objeto do tipo <code>Writer</code> aonde as
     * informaes sero escritas.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>output</code> for <code>null</code>.
     */
    public void getPrettyPrintingResults(Writer output)
            throws IllegalArgumentException, IOException {
        if (output == null) {
            throw new IllegalArgumentException();
        }
        output.write(this.state.pop());
    }
    
    protected StringStack getState() {
    	return this.state;
    }

    /**
     * Gera um string correspondente ao nmero de quebras de linha desejado.
     *
     * @param numRows    Nmero de quebras de linha desejado.
     *
     * @return     Um objeto do tipo <code>String</code> correspondendo a
     * <code>numRows</code> quebras de linha.
     */
    public static String lineBreak(int numRows) {
        int i = 0;
        StringBuffer sbuffer = new StringBuffer();
        while (i < numRows) {
            sbuffer.append(lineBreak);
            i++;
        }
        return sbuffer.toString();
    }


    /**
     * Retorna um espao a ser usado na formatao de strings que
     * representem declaraes, etc
     */
    public String space() {
        return space;
    }


    /**
     * Imprime os numElements elementos de uma lista, separadas por "separator",
     * a impresso tanto pode ser na ordem inversa, quanto na ordem igual  da
     * pilha
     *
     * @param  separator    Separador dos elementos
     * @param  numElements  Numero de elementos a serem impressos
     * @param  tab          Espao antes de cada elemento
     * @param  reverse indica se a lista deve ser impressa na ordem em que os
     *         elementos forem desempilhados(false), ou na ordem inversa(true)
     *
     * @return String correspondente a impressao da lista
     */
    protected String prettyPrintList(String separator, int numElements, String tab,
                                   boolean reverse) throws IllegalArgumentException {

        if (separator == null || numElements < 0) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        int i = 0;
        String theTab = "";
        if (tab != null && tab.length() > 0) {
            theTab = tab;
        }
        // se a lista estiver na ordem inversa
        if (reverse) {
            while (i < numElements) {
                if (i + 1 == numElements) {
                    resultString = theTab + this.state.pop() + resultString;
                } else {
                    resultString = separator + theTab + this.state.pop() + resultString;
                }
                i++;
            }
        }
        // se a lista estiver ordenada corretamente  - ou seja, se foi empilhado
        // de trs pra frente
        else {
            while (i < numElements) {
                if (i + 1 != numElements) {
                    // se no  o ltimo, coloca o separador depois
                    sbuffer.append(theTab);
                    sbuffer.append(this.state.pop());
                    sbuffer.append(separator);

                } else {
                    sbuffer.append(theTab);
                    sbuffer.append(this.state.pop());
                }
                i++;
            }
            resultString = sbuffer.toString();
        }
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JAllocationExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JAllocationExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String str_new = "new";
        String resultString = null;
        StringBuffer sbuffer = new StringBuffer();


        if (node.getAllocationType() == JAllocationExpression.OBJECTALLOCATION) {

            String body = "";
            if (node.getInternalClassBody() != null) {
                body = this.state.pop();
            }
            if (node.getArguments() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String args = this.state.pop();

            if (node.getTypeName() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String typeName = this.state.pop();

            if (node.getReferencedObject() != null) {
                String refObj = this.state.pop();
                sbuffer.append(refObj);
                sbuffer.append(".");
            }

            sbuffer.append(str_new);
            sbuffer.append(space());
            sbuffer.append(typeName);
            sbuffer.append("(");
            sbuffer.append(args);
            sbuffer.append(")");
            sbuffer.append(body);

        } else if (node.getAllocationType() ==
                JAllocationExpression.PRIMITIVETYPEARRAYALLOCATION
                || node.getAllocationType() ==
                JAllocationExpression.OBJECTARRAYALLOCATION) {

            if (node.getArrayDimsAndInits() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String adi = this.state.pop();

            if (node.getTypeName() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

            String typeName = this.state.pop();

            if (node.getReferencedObject() != null) {
                String refObj = this.state.pop();
                sbuffer.append(refObj);
                sbuffer.append(".");
            }


            sbuffer.append(str_new);
            sbuffer.append(space());
            sbuffer.append(typeName);
            sbuffer.append(" ");
            sbuffer.append(adi);
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * @see JArrayAccess
     */
    public Object visit(JArrayAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        String arrayName = "";
        if (node.getArrayName() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        arrayName = this.state.pop();
        sbuffer.append(arrayName);


        if (node.getIndexers() != null) {
            String indexers = this.state.pop();
            sbuffer.append(indexers);
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JArrayDimsAndInits</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JArrayDimsAndInits node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        Vector dimSizes = new Vector();
        String ai = "";
        int numDims = 0;
        int max = 0;
        int count = 0;

        if (node.getArrayInitializer() != null) {
            ai = this.state.pop();
        }


        if ((max = node.getNumberOfDimensionSizesSpecified()) > 0) {
            int i = 0;
            while (i < max) {
                // Os dimension sizes so armazenados de trs pra frente, ou
                // seja, para um array declarado int[3][4][1+x], os tamanhos
                // sero armazenados na lista como < "1+x", "4", "3">.
                dimSizes.addElement(this.state.pop());
                i++;
            }
        }

        numDims = node.getDimensions();

        if (numDims < dimSizes.size()) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        int numOfDimSizes = dimSizes.size();

        while (count < numDims) {
            resultString += "[";

            if (count < numOfDimSizes) {
                resultString +=
                        (String) dimSizes.elementAt(numOfDimSizes - (count + 1));
            }

            resultString += "]";
            count++;
        }
        if (!ai.equals("")) {
            resultString += space() + ai;
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JArrayInitializer</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JArrayInitializer node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        int max = node.size();

        resultString = " }";

        resultString = this.prettyPrintList(", ", max, "", true) + resultString;

        resultString = "{ " + resultString;
        this.state.push(resultString);
        return resultString;
    }

    /**
     * @see Attribute
     */
//    public Object visit(Attribute node, Object data)
//            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
//            ExecutionException {
//        return super.visit(node, data);
//    }

    /**
     * @see AttributeArgument
     */
//    public Object visit(AttributeArgument node, Object data)
//            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
//            ExecutionException {
//        return super.visit(node, data);
//    }

    /**
     * @see AttributeArgumentList
     */
//    public Object visit(AttributeArgumentList node, Object data)
//            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
//            ExecutionException {
//        return super.visit(node, data);
//    }

    /**
     * @see AttributeSection
     */
//    public Object visit(AttributeSection node, Object data)
//            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
//            ExecutionException {
//        return super.visit(node, data);
//    }

    /**
     * Mtodo visit para objetos do tipo <code>JBinaryExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JBinaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        if (node.isASourceNode()) {
            StringBuffer sbuffer = new StringBuffer();

            if (node.getRightHandOperand() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String rightHand = this.state.pop();

            if (node.getLeftHandOperand() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String leftHand = this.state.pop();

            sbuffer.append(leftHand);
            sbuffer.append(space());
            sbuffer.append(node.getBinaryOperatorString());
            sbuffer.append(space());
            sbuffer.append(rightHand);

            resultString = sbuffer.toString();
        }
        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JBlock</code>.
     *
     * @param node O n a ser clonado.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JBlock node, Object data)
			throws InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        final String EMPTY_STMT = ";";

        StringBuffer sbuffer = new StringBuffer();
        int numElements = node.size();

        if (node.isASourceNode()) {
            for (int i = 0; i < numElements; i++) {
                String stmt = this.state.pop();
                String tmp = "";
                if (stmt != null) {
                    tmp = stmt.trim();

                }
                // s imprime se no for Empty Statement
                if (!tmp.equals(EMPTY_STMT)) {
                    if (Util.checkString(tmp) && !tmp.endsWith(";")
                            && !tmp.endsWith("}")) {
                        stmt = stmt + ";";
                    }
                    sbuffer.append(stmt);
                    sbuffer.append(lineBreak(1));
                }
            }
            
            resultString = indentSequence(sbuffer.toString(), true);
            
            resultString = "{" 
            	    + lineBreak(1) 
            		+ resultString 
            		+ "}";
        } else {
            resultString = node.printVariable() + ";";
        }
        this.state.push(resultString);
        return resultString;
    }

    private String indentSequence(String block, boolean indentFirst) {
    	String result = block;
    	if (indentFirst) 
    		result = indentString + result;
    	result = result.replaceAll(lineBreak, lineBreak + indentString);
    	if (result.endsWith(lineBreak + indentString)) {
    		result = result.substring(0, result.lastIndexOf(indentString));
    	}
    	return result;
    }
    
    /**
     * Mtodo visit para objetos do tipo <code>JCatchClause</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JCatchClause node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        StringBuffer sbuffer = new StringBuffer();
        String resultString = null;

        String comment = node.getComment();
        if (Util.checkString(comment)) {
            sbuffer.append(comment);
            sbuffer.append(lineBreak(1));
        }
        sbuffer.append("catch");
        if (node.getParameter() != null) {
            String param = this.state.pop().trim();
            sbuffer.append("(");
            sbuffer.append(param);
            sbuffer.append(")");
        }

        if (node.getBlock() != null) {
            String block = this.state.pop();
            if (bracesOnNewLine) {
                sbuffer.append(lineBreak(1));
            }
            sbuffer.append(block);
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>CastExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(CastExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        // pega o tipo
        if (node.getType() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String type = this.state.pop();

        // pega o operando
        if (node.getOperand() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String operand = this.state.pop();

        resultString += "(" + type + ")" + operand;

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>TypeBody</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(TypeBody node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {


        if (node == null) {
            throw new IllegalArgumentException("Declaracao de classe nula");
        }
        StringBuffer sbuffer = new StringBuffer();
        String resultString = "";

        String cdecs = this.state.pop();
        if (Util.checkString(cdecs)) {
            cdecs = lineBreak(1) + cdecs;
        }
        String mdecs = this.state.pop();
        if (Util.checkString(mdecs)) {
            mdecs = lineBreak(1) + mdecs;
        }
        String odecs = this.state.pop();
        if (Util.checkString(odecs)) {
            odecs = lineBreak(1) + odecs;
        }
        String fdecs = this.state.pop();
        if (Util.checkString(fdecs)) {
            fdecs = lineBreak(1) + fdecs;
        }
        String inializers = "";
        if (node.getInitializerSet() != null) {
            inializers = this.state.pop();
            if (Util.checkString(inializers)) {
                inializers = lineBreak(1) + inializers;
            }
        }
        String nestedTypes = "";
        if (node.getNestedTypeSet() != null) {
            nestedTypes = this.state.pop();
            if (Util.checkString(nestedTypes)) {
                nestedTypes = lineBreak(1) + nestedTypes;
            }
        }

        sbuffer.append(fdecs);
        sbuffer.append(inializers);
        sbuffer.append(cdecs);
        sbuffer.append(mdecs);
        sbuffer.append(odecs);
        sbuffer.append(nestedTypes);
        resultString =  indentSequence(sbuffer.toString(), true);
        resultString = "{" 
        	           + lineBreak(1)
                       + resultString
                       + lineBreak(1)
                       + "}";
        if (bracesOnNewLine) {
        	resultString = lineBreak(1) + resultString;
        }
        
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JClassDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JClassDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Declaracao de classe nula");
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        if (node.hasComment()) {
            sbuffer.append(node.getComment());
            sbuffer.append(lineBreak(1));
        }
        if (node.getModifiers() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        String mods = this.state.pop().trim();

        if (mods.length() > 0) {
            sbuffer.append(mods);
            sbuffer.append(space());
        }


        if (node.getName() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String className = this.state.pop();
        sbuffer.append("class ");
        sbuffer.append(className);
        sbuffer.append(space());

        if (node.getSuperClass() != null) {
            String superClass = this.state.pop();
            if (node.getSuperClass().isOptional()) {
                sbuffer.append("#[");
                sbuffer.append(" extends ");
                sbuffer.append(superClass);
                sbuffer.append(space());
                sbuffer.append("]#");
                sbuffer.append(space());
            } else {
                sbuffer.append(" extends ");
                sbuffer.append(superClass);
                sbuffer.append(space());
            }
        }

        if (node.getInterfaces() != null) {
            String interfaces = this.state.pop();
            if (interfaces != null && !interfaces.trim().equals("")) {
                if (node.getInterfaces().isOptional()) {
                    sbuffer.append("#[");
                    sbuffer.append("implements ");
                    sbuffer.append(interfaces);
                    sbuffer.append(space());
                    sbuffer.append("]# ");
                    sbuffer.append(space());
                } else {
                    sbuffer.append("implements ");
                    sbuffer.append(interfaces);
                    sbuffer.append(space());
                }
            }
        }

        if (node.getBody() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String classBody = this.state.pop();
        sbuffer.append(classBody);

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JCompilationUnit</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao<br>
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JCompilationUnit node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JCompilationUnit nula");
        }

        if (node.getTypeDeclarationList() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        String resultString = "";

        // Pr-condio
        if (node.getPrecondition() != null) {
            String pre = this.state.pop();
            resultString += pre + lineBreak(2);
        }

        // Pacote
        if (node.getPackage() != null) {
            String pack = this.state.pop();
            resultString += pack + lineBreak(2);
        }
        // Imports
        if (node.getImports() != null) {
            String tmp = this.state.pop();
            resultString += tmp;
            // pula uma linha entre as declaracoes de import e a classe
            resultString += lineBreak(2);
        }

        // Impresso das declaraes de tipo (declaraes de classe ou
        // interface).
        int i = 0;
        int size = node.getTypeDeclarationList().size();
        while (i < size) {
            String type = this.state.pop();
            resultString = resultString + type + lineBreak(1);
            i++;
        }
        if (node.getHeader() != null) {
            resultString = node.getHeader() + lineBreak(1) + resultString;
        }
        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JConditionalDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JConditionalDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!jatsTypesEnabled) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }


        String resultString = "";
        String strFalseDecSet = "";
        String strTrueDecSet = "";

        String indentDec = "";
        String tmpDec = "";
        String dec;


        NodeList falseDecSet = node.getFalseDeclarationSet();
        if (falseDecSet != null) {
            int i = 0;
            int max = falseDecSet.size();
            StringBuffer falseBuffer = new StringBuffer();
            while (i < max) {
                falseBuffer.append(indentDec);
                dec = this.state.pop();
                if (dec != null) {
                    tmpDec = dec.trim();
                }
                if (Util.checkString(tmpDec) && !tmpDec.endsWith(";") &&
                        !tmpDec.endsWith("}")) {
                    dec = dec + ";";
                }
                falseBuffer.append(dec);

                i++;
            }
            strFalseDecSet = "else {" +
                    lineBreak(1) +
                    falseBuffer.toString() +
                    lineBreak(1) +
                    "}" +
                    lineBreak(1);
        }

        NodeList trueDecSet = node.getTrueDeclarationSet();
        if (trueDecSet != null) {
            int j = 0;
            int max2 = trueDecSet.size();
            StringBuffer trueBuffer = new StringBuffer();
            while (j < max2) {
                trueBuffer.append(indentDec);
                dec = this.state.pop();
                if (dec != null) {
                    tmpDec = dec.trim();
                }
                if (Util.checkString(tmpDec) && !tmpDec.endsWith(";") && !tmpDec.endsWith("}")) {
                    dec = dec + ";";
                }
                trueBuffer.append(dec);
                j++;
            }
            strTrueDecSet = "{" +
                    lineBreak(1) +
                    trueBuffer.toString() +
                    lineBreak(1) +
                    "}" +
                    lineBreak(1);
        }

        String cond = this.state.pop();

        resultString = "#if(" + cond + ")" +
                strTrueDecSet +
                strFalseDecSet;

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JConstructorDeclaration</code>.
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JConstructorDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Declaracao de construtor nula");
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        if (node.isASourceNode()) {
            // modificador
            String mod = "";
            if (node.getModifiers() != null) {
                mod = this.state.pop().trim();
            }

            String id = "";
            if (node.getName() != null) {
                id = this.state.pop();
            }

            String params = "";
            if (node.getParameters() != null) {
                params = this.state.pop();
            }

            // lista de excees
            String exc = "";
            JNameList exceptions = node.getExceptions();
            if (exceptions != null) {
                if (exceptions.size() > 0 || !exceptions.isASourceNode()) {
                    exc = "throws ";
                }
                exc += this.state.pop().trim();
                if (node.getExceptions().isOptional()) {
                    exc = "#[" + exc + "]#";
                }
            }


            String body = "";
            String tmpBody = "";
            if (node.getBody() != null) {
                tmpBody = this.state.pop();
                if (node.getBody().isVariable()) {
                    tmpBody = indentString + node.getBody().printVariable() + ";";
                }
                if (bracesOnNewLine) {
                    body = lineBreak(1);
                } else {
                    body = "";
                }
                body += tmpBody;
            } else {
            	body = "{}";
            }

            if (node.hasComment()) {
                sbuffer.append(node.getComment());
                sbuffer.append(lineBreak(1));
            }
            sbuffer.append(mod);
            sbuffer.append(space());
            sbuffer.append(id);
            sbuffer.append(params);
            sbuffer.append(space());
            sbuffer.append(exc);
            sbuffer.append(body);
            resultString = sbuffer.toString();
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JConstructorDeclarationSet</code>.
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JConstructorDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        int max = node.size();


        if (node.isASourceNode()) {
            resultString = this.prettyPrintList(lineBreak(2), max, "", false);
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>ClassBodyConstructorDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao<br>
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(ClassBodyConstructorDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JConstructorDeclaration nulo");
        }

        String resultString = "";

        // Declaraes de construtores
        String decs = this.state.pop();

        if (!decs.trim().equals("")) {
            resultString = resultString + decs + lineBreak(2);
        }

        // Declaraes do tipo ConstructorVar, correspondendo a variveis que casam
        // com uma declarao de atributo.
        int j = 0;
        int max2 = node.getConstructorVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            j++;
        }

        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo ConstructorsVar, correspondendo a variveis que casam
        // com um conjunto de declaraes de atributos.
        int k = 0;
        int max3 = node.getConstructorsVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            k++;
        }
        if (k > 0) {
            resultString += lineBreak(1);
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JExecutableDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao<br>
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JExecutableDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!jatsTypesEnabled) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        String resultString = "";
        String expressions = "";
        int type = node.getExpressionType();
        int i = 0;
        int max = node.size();

        // lista de expresses
        while (i < max) {
            if (i == max - 1) {
                expressions += this.state.pop();
            } else {
                expressions += this.state.pop() + "; ";
            }
            i++;
        }

        // n resultado, nem sempre  empilhado. Apenas se no for null
        String resNode = "";
        if (node.getResultNode() != null) {
            resNode = this.state.pop();
        }
        // s existe, se for do tipo information-modifying
        String refVar = "";
        JaTSNode rvar = (JaTSNode) node.getReferencedVariable();
        if (rvar != null && rvar.getVariableName() != null) {
            refVar = rvar.printVariable();
        }
        // se for information-extracting, imprime: #< exp >#
        if (type == JExecutableDeclaration.INFORMATION_EXTRACTING) {
            resultString = "#< " + expressions + " >#";
        }
        // seno, imprime: #<#resNode :: #refVar = exps>#
        else if (type == JExecutableDeclaration.INFORMATION_MODIFYING) {
            resultString = "#< " + resNode + " = " + refVar + " :: " + expressions + " >#";
        }


        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JExpressionList</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *apresenta uma estrutura interna invlida.
     */
    public Object visit(JExpressionList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        int max = node.size();
        // se max for zero, empilha espao em branco
        String resultString = null;
        if (node.getType() == JExpressionList.ARGUMENTS) {
            resultString = this.prettyPrintList(", ", max, "", false);
        } else if (node.getType() == JExpressionList.ARRAY_INDEXERS) {
            StringBuffer sbuffer = new StringBuffer();
            for (int i = 0; i < max; i++) {
                sbuffer.append("[");
                sbuffer.append(this.state.pop());
                sbuffer.append("]");
            }

            resultString = sbuffer.toString();
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JFieldAccess</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        JComment comment = node.getComment();

        if (comment != null) {
            sbuffer.append(comment);
            sbuffer.append(lineBreak(1));
        }

        if (node.isASourceNode()) {


            String fieldName = "";
            if (node.getFieldName() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            fieldName = this.state.pop();

            String refObj = "";
            if (node.getReferencedObject() != null) {
                refObj = this.state.pop();
                sbuffer.append(refObj);
                sbuffer.append(".");

            }

            sbuffer.append(fieldName);

            resultString = sbuffer.toString();
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JFieldDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao<br>
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("null JFieldDeclaration value");
        }

        String resultString = "";
        
        if (node.isASourceNode()) {
            
            if (node.hasComment()) {
                resultString += node.getComment() + lineBreak(1);
            }

            // Impresso dos modificadores
            if (node.getModifiers() == null) {
                throw new InconsistentNodeException(node);
            }
            String mods = "";
            mods = this.state.pop().trim();
            if (!mods.equals("")) {
                resultString += mods + space();
            }

            // Impresso do tipo do atributo
            if (node.getType() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            String type = this.state.pop();
            resultString = resultString + type + space();

            // Impresso dos nomes dos atributos e suas respectivas inicializaes
            if (node.getVariables() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            int i = 0;
            int max = node.size();
            while (i < max) {
                String varDecs = this.state.pop();
                if (i == 0) {
                    resultString = resultString + varDecs;
                } else {
                    resultString = resultString + ", " + varDecs;
                }
                i++;
            }

        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }
        resultString += ";";
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JFieldDeclarationSet</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        int max = node.size();
        if (!node.isVariable()) {
            resultString = this.prettyPrintList(lineBreak(1), max, "", false);
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }
        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>BodyFieldDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyFieldDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JFieldDeclaration nulo");
        }

        String resultString = "";

        // Declaraes de atributos
        String decs = this.state.pop();

        if (!decs.trim().equals("")) {
            resultString = resultString + decs + lineBreak(1);
        }

        // Declaraes do tipo FieldVar, correspondendo a variveis que casam
        // com uma declarao de atributo.
        int j = 0;
        int max2 = node.getFieldVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            j++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo FieldsVar, correspondendo a variveis que casam
        // com um conjunto de declaraes de atributos.
        int k = 0;
        int max3 = node.getFieldsVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            k++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (k > 0) {
            resultString += lineBreak(1);
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>BodyFieldDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(ClassBodyInitializers node, Object data)
            throws IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JFieldDeclaration nulo");
        }

        String resultString = "";

        // Declaraes de atributos
        String decs = this.state.pop();

        if (!decs.trim().equals("")) {
            resultString = resultString + decs + lineBreak(1);
        }

        // Declaraes do tipo InitVar, correspondendo a variveis que casam
        // com um inicializador
        int j = 0;
        int max2 = node.getInitializerVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            j++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo InitsVar, correspondendo a variveis que casam
        // com um conjunto de inicializadores.
        int k = 0;
        int max3 = node.getInitializersVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(1);
            k++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (k > 0) {
            resultString += lineBreak(1);
        }
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JFormalParameter</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JFormalParameter node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.isASourceNode()) {

            String mod = "";
            if (node.getModifiers() != null) {
                mod = this.state.pop().trim() + space();
                if (node.getModifiers().hasNoModifiers()) {
                	mod = "";
                }
            }
            String t = this.state.pop().trim();

            String id = "";
            if (node.getIdentifier() != null) {
                id = this.state.pop().trim();
            }
            // para os colchetes
            String brackets = "";
            int i = 0;
            int max = node.getDimensions();
            while (i < max) {
                brackets += "[]";
                i++;
            }
            resultString = mod + t + space() + id + brackets;
        } else if (node.isVariable()) {
            resultString = node.printVariable();
        }

        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JForStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JForStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String init = "";
        String exp = "";
        String update = "";
        String stmt = "";


        if (node.getForInit() != null) {
            init = this.state.pop().trim();
        }
        if (node.getExpression() != null) {
            exp = this.state.pop().trim();
        }
        if (node.getForUpdate() != null) {
            update = this.state.pop().trim();
        }

        sbuffer.append("for");
        sbuffer.append("(");
        sbuffer.append(space());
        sbuffer.append(init);
        sbuffer.append("; ");
        sbuffer.append(space());
        sbuffer.append(exp);
        sbuffer.append("; ");
        sbuffer.append(space());
        sbuffer.append(update);
        sbuffer.append(")");
        sbuffer.append(space());

        // remover isso, to logo o bloco seja responsvel por adicionar
        // seus proprios "{" e "}"
        if (node.getStatement() != null) {
            stmt = this.state.pop();
            if (node.getStatement() instanceof JBlock) {
                if (bracesOnNewLine) {
                    sbuffer.append(lineBreak(1));
                }
                sbuffer.append(stmt);
            } else {
                sbuffer.append(lineBreak(1));
                sbuffer.append(stmt);
            }
        }
        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JIdentifier</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JIdentifier node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.getValue() == null && node.isASourceNode()) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        if (node.isASourceNode()) {
            resultString = node.getValue();
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);

        return resultString;
    }


    public Object visit(JIfStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String exp = "";
        String ifStmt = "";
        String elseStmt = "";


        if (Util.checkString(node.getIfComment())) {
            sbuffer.append(node.getIfComment());
            sbuffer.append(lineBreak(1));
        }

        if (node.getExpression() != null) {
            exp = this.state.pop().trim();
        }
        sbuffer.append("if ");
        sbuffer.append("(");
        sbuffer.append(exp);
        sbuffer.append(")");
        sbuffer.append(space());

        if (node.getIfStatement() != null) {
            ifStmt = this.state.pop();
            if (node.getIfStatement() instanceof JBlock) {
                if (bracesOnNewLine) {
                    sbuffer.append(lineBreak(1));
                }
                sbuffer.append(ifStmt);
            } else {
                sbuffer.append(lineBreak(1));
                //o statement puro, naturalmente, nao vem indentado
                //logo, excepcionalmente, precisa-se indenta-lo aqui
                //(ha ainda a opcao de manter na mesma linha do if)
                sbuffer.append(indentString);
                sbuffer.append(ifStmt);
            }
        }
        if (node.getElseStatement() != null) {

            elseStmt = this.state.pop();
            if (elseOnNewLine) {
                sbuffer.append(lineBreak(1));
//                sbuffer.append(nodeIndent);
            } else {
                sbuffer.append(space());
            }
            if (Util.checkString(node.getElseComment())) {
                sbuffer.append(node.getElseComment());
                sbuffer.append(lineBreak(1));
            }
            sbuffer.append("else");
            if (node.getElseStatement() instanceof JBlock) {
                if (bracesOnNewLine) {
                    sbuffer.append(lineBreak(1));
                } else {
                    sbuffer.append(space());
                }
                sbuffer.append(elseStmt);
                sbuffer.append(lineBreak(1));
            } else if (node.getElseStatement() instanceof JIfStatement) {
                sbuffer.append(space());
                sbuffer.append(elseStmt);
            } else {
                sbuffer.append(lineBreak(1));
                sbuffer.append(elseStmt);
            }
        }
        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JImportDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JImportDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        String packageName = "";
        String className = "";
        if (node.isASourceNode()) {

            if (node.getPackageName() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            String all = "";
            if (node.importAllClasses()) {
                all = ".*";
            }
            packageName = this.state.pop();

            if (node.getClassName() != null) {
                className = this.state.pop();
                if (packageName != null && packageName.length() > 0) {
                    className = "." + className;
                }
            }
            resultString = "import " + packageName + className + all + ";";
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable() + ";";
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration() + ";";
                this.state.pop();
            }
        }
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JImportDeclarationSet</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JImportDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        String resultString = "";
        int max = node.size();
        if (node.isASourceNode()) {
            resultString = this.prettyPrintList(lineBreak(1), max, "", false);
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>ImportDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(ImportDeclarations node, Object data)
            throws InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("JImportDeclarationSet nulo");
        }

        String resultString = "";

        // Declaraes de importaco
        String decs = this.state.pop();

        resultString = decs + lineBreak(1);


        // Declaraes do tipo ImportVarSet, correspondendo a variveis que casam
        // com uma declarao de importao
        int j = 0;
        int max2 = node.getImportVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + dec + ";" + lineBreak(1);
            j++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo ImportsVar, correspondendo a variveis que casam
        // com um conjunto de declaraes de importaes.
        int k = 0;
        int max3 = node.getImportsVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + dec + ";" + lineBreak(1);
            k++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (k > 0) {
            resultString += lineBreak(1);
        }

        this.state.push(resultString);


        return resultString;

    }

    /**
     * Mtodo visit para objetos do tipo <code>JClassDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JInterfaceDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        // ordem de escrita no accept:
        // corpo
        // interfaces
        // name
        // modifiers

        if (node == null) {
            throw new IllegalArgumentException("Declaracao de interface nula");
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        if (node.hasComment()) {
            sbuffer.append(node.getComment());
            sbuffer.append(lineBreak(1));
        }
        if (node.getModifiers() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String mods = this.state.pop().trim();
        if (mods.length() > 0) {
            sbuffer.append(mods);
            sbuffer.append(space());
        }

        if (node.getName() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String interfaceName = this.state.pop();
        sbuffer.append("interface");
        sbuffer.append(space());
        sbuffer.append(interfaceName);
        sbuffer.append(space());


        if (node.getInterfaces() != null) {
            String interfaces = this.state.pop().trim();
            if (interfaces != null && !interfaces.equals("")) {
                if (node.getInterfaces().isOptional()) {
                    sbuffer.append(" #[");
                    sbuffer.append("extends");
                    sbuffer.append(space());
                    sbuffer.append(interfaces);
                    sbuffer.append(space());
                    sbuffer.append("]# ");

                } else {
                    sbuffer.append("extends");
                    sbuffer.append(space());
                    sbuffer.append(interfaces);
                    sbuffer.append(space());
                }
            }
        }

        if (node.getBody() != null) {
            String corpo = this.state.pop().trim();
            if (!corpo.equals("")) {
                sbuffer.append(corpo);
            }
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * @see JInitializer
     */
    public Object visit(JInitializer node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        StringBuffer sbuffer = new StringBuffer();
        String resultString = "";

        if (node.isASourceNode()) {
            String block = this.state.pop();
            if (node.isStatic()) {
                sbuffer.append("static");
                sbuffer.append(space());
            }
            if (bracesOnNewLine) {
                sbuffer.append(lineBreak(1));
            }
            sbuffer.append("{");
            sbuffer.append(lineBreak(1));
            sbuffer.append(indentSequence(block, true));
            sbuffer.append(lineBreak(1));
            sbuffer.append("}");
            resultString = sbuffer.toString();
            this.state.push(resultString);
        } else if (jatsTypesEnabled) {
            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }
        return resultString;

    }


    /**
     * Mtodo visit para objetos do tipo <code>JInitializerSet</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JInitializerSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        int max = node.size();
        if (!node.isVariable()) {
            resultString = this.prettyPrintList(lineBreak(1), max, indentString, false);
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>IterativeDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(IterativeDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!jatsTypesEnabled) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String iterationSet = this.state.pop();
        String dec = "";
        String tmpDec = "";


        String declarationSet = "";
        int size = node.getDeclarationSet().size();
        String indentDec = "";
        for (int i = 0; i < size; i++) {
            dec = this.state.pop();
            if (dec != null) {
                tmpDec = dec.trim();
            }
            if (Util.checkString(tmpDec) && !tmpDec.endsWith(";") && !tmpDec.endsWith("}")) {
                dec = dec + ";";
            }

            declarationSet += indentDec + dec;
        }
        String placeholder = this.state.pop();

        sbuffer.append("forall");
        sbuffer.append(space());
        sbuffer.append(placeholder);
        sbuffer.append(space());
        sbuffer.append("in");
        sbuffer.append(space());
        sbuffer.append(iterationSet);
        sbuffer.append(space());
        if (bracesOnNewLine) {
            sbuffer.append(lineBreak(1));
        }
        sbuffer.append("{");
        sbuffer.append(lineBreak(1));
        sbuffer.append(indentSequence(declarationSet, true));
        sbuffer.append(lineBreak(1));
        sbuffer.append("}");
        resultString = sbuffer.toString();

        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>InstanceOfExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(InstanceOfExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) throw new IllegalArgumentException();

        String resultString = "";

        if (node.getType() == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        String type = this.state.pop();

        if (node.getOperand() == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        String operand = this.state.pop();

        resultString += operand + " instanceof " + type;

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JLiteral</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JLiteral node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.isASourceNode()) {
            if (node.getLiteralValue() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            if (node.getLiteralType() == JLiteral.STRING_LITERAL) {
                String tmp = node.getValue();//"\"" + (node.getValue()) + "\"";
                resultString = tmp;
            } else if (node.getLiteralType() == JLiteral.CHARACTER_LITERAL) {
                String tmp = "\'" + (node.getValue()) + "\'";
                resultString = tmp;
            } else if (node.getLiteralType() == JLiteral.VERBATIM_STRING_LITERAL) {
            	resultString = "@\"" + (node.getValue()) + "\"";
            } else {
                resultString = (node.getValue());
            }
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JLocalVariableDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria
     * ao visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JLocalVariableDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JFieldDeclaration nulo");
        }

        String resultString = "";

        if (!node.isExecutable() && !node.isVariable()) {

            if (node.hasComment()) {
                resultString += node.getComment() + lineBreak(1);
            }
            // Impresso dos modificadores
            if (node.getModifiers() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            String mods = "";
            mods = this.state.pop().trim();
            if (!mods.equals("")) {
                resultString += mods + space();
            }

            // Impresso do tipo do atributo
            if (node.getType() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            String type = this.state.pop();
            resultString = resultString + type + space();

            // Impresso dos nomes dos atributos e suas respectivas inicializaes
            if (node.getVariables() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            int i = 0;
            int max = node.size();

            while (i < max) {
                String varDecs = this.state.pop();
                if (i == 0) {
                    resultString = resultString + varDecs;
                } else {
                    resultString = resultString + ", " + varDecs;
                }
                i++;
            }
        } else if (node.isVariable()) {
            resultString = node.printVariable();
        }

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JMethodDeclaration</code>.
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Declaracao de metodo nula");
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        if (node.isASourceNode()) {
        	
            // modificadores
            String mods = "";
            if (node.getModifiers() != null) {
                mods = this.state.pop().trim() + space();
                if (node.getModifiers().hasNoModifiers()) {
                	mods = "";
                }
            }
            
            // tipo de retorno
            String resType = "";
            if (node.getResultType() != null) {
                resType = this.state.pop().trim() + space();
            }
            // imprime o identificador
            String id = this.state.pop().trim();

            // imprime a lista de parametros
            String param = this.state.pop().trim();

            // imprime os colchetes, se houverem
            String brackets = "";
            int dim = node.getDimensions();
            while (dim > 0) {
                brackets += "[]";
                dim--;
            }

            // lista de excees
            String exc = "";
            JNameList exceptions = node.getExceptions();
            if (exceptions != null) {
                if (exceptions.size() > 0 || !exceptions.isASourceNode()) {
                    exc = "throws ";
                }
                exc += this.state.pop().trim();
                if (node.getExceptions().isOptional()) {
                    exc = "#[" + exc + "]#";
                }
            }

            if (node.hasComment()) {
                sbuffer.append(node.getComment());
                sbuffer.append(lineBreak(1));
            }

            // concatena as strings pra montar a declarao de mtodo
            sbuffer.append(mods);
            sbuffer.append(resType);
            sbuffer.append(id);
            sbuffer.append(param);
            sbuffer.append(brackets);
            sbuffer.append(space());
            sbuffer.append(exc);


            // se tiver corpo, adiciona o corpo do mtodo a declarao
            if (node.hasBody()) {
                String body = this.state.pop();
                if (node.getBody().isVariable()) {
                    body = indentString + node.getBody().printVariable() + ";";
                }

                if (bracesOnNewLine) {
                    sbuffer.append(lineBreak(1));

                }
                sbuffer.append(body);


            } else {
                // adiciona o ";" - de terminao do mtodo (abstrato ou nativo)
                sbuffer.append(";");
            }

        } else if (jatsTypesEnabled) {
            sbuffer.append(printJaTSNode(node) + ";");
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JMethodDeclarationSet</code>.
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException
	{
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        int max = node.size();


        if (!node.isVariable()) {
            resultString = this.prettyPrintList(lineBreak(2), max, ""/*indentMethod*/, false);
        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();
            }
        }

        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JBodyMethodDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyMethodDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("BodyMethodDeclarations nulo");
        }

        String resultString = "";

        // Declaraes de mtodos
        String decs = this.state.pop();

        if (!decs.trim().equals("")) {
            resultString = resultString + decs + lineBreak(1);
        }
        // Declaraes do tipo MethodVar, correspondendo a variveis que casam
        // com uma declarao de mtodo.
        int j = 0;
        int max2 = node.getMethodVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(2);
            j++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo MethodsVar, correspondendo a variveis que casam
        // com um conjunto de declaraes de mtodos
        int k = 0;
        int max3 = node.getMethodsVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + indentString + dec + ";" + lineBreak(2);
            k++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (k > 0) {
            resultString += lineBreak(1);
        }

        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JBodyMethodDeclarations</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyTypeDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("BodyMethodDeclarations nulo");
        }

        String resultString = "";

        // Declaraes de mtodos
        String decs = this.state.pop();

        if (!decs.trim().equals("")) {
            resultString = resultString + decs + lineBreak(1);
        }
        // Declaraes do tipo TypeDeclaration, correspondendo a variveis que
        // casam com uma declarao de tipo.
        int j = 0;
        int max2 = node.getTypeVarSet().size();

        while (j < max2) {
            String dec = this.state.pop();
            resultString = resultString + lineBreak(1) + dec;
            j++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (j > 0) {
            resultString += lineBreak(1);
        }

        // Declaraes do tipo TypeDeclarationSet, correspondendo a variveis
        // que casam com um conjunto de declaraes de tipos aninhados
        int k = 0;
        int max3 = node.getTypesVarSet().size();

        while (k < max3) {
            String dec = this.state.pop();
            resultString = resultString + lineBreak(1) + dec;
            k++;
        }
        // Se alguma declarao foi impressa, o resultado fica mais bonito se
        // uma linha for pulada.
        if (k > 0) {
            resultString += lineBreak(1);
        }

        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JSynchronizedStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
	//public Object visit(JSynchronizedStatement node, Object data)
			//throws InconsistentNodeException { //TODO
		
	//}
    /**
     * Mtodo visit para objetos do tipo <code>JMethodInvocation</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodInvocation node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        JComment comment = node.getComment();
        if (comment != null) {
            resultString += comment;
            resultString += "\n";
        }
        if (node.getArgumentList() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String args = this.state.pop();

        String methodName = "";
        if (node.getMethodName() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        methodName = this.state.pop();

        String refObj = "";
        if (node.getReferencedObject() != null) {
            refObj = this.state.pop();
            resultString += refObj + ".";
        }

        resultString += methodName;
        resultString += "(" + args + ")";
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JModifierList</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JModifierList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.isASourceNode()) {
            resultString = node.toString().trim();
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JName</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JName node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        int max = node.size();

        if (node.isASourceNode()) {
            resultString = this.prettyPrintList(".", max, "", true);
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JNodeString</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JNodeString node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        String resultString = "";

        resultString += node.getNodeValue() + lineBreak;

        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JNameList</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JNameList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        int max = node.size();

        if (node.isASourceNode()) {
            resultString = this.prettyPrintList(", ", max, "", true);
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }
        resultString = resultString.trim();
        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>OtherDeclarationSet</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(OtherDeclarationsSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        String resultString2 = "";
        String resultString3 = "";

        // Impresso das outras declaraes
        int max3 = node.numberOfConditionalDeclarations();
        resultString = this.prettyPrintList(lineBreak(1), max3, indentString, true);

        resultString += lineBreak(1);

        // Impresso dos ns-resultado
        int max = node.numberOfResultDeclarations();
        resultString = this.prettyPrintList(lineBreak(1), max, indentString, true);

        resultString2 += lineBreak(1);

        // Impresso das declaraes iterativas
        int max2 = node.numberOfIterativeDeclarations();
        resultString3 = this.prettyPrintList(lineBreak(1), max2, indentString, true);

        this.state.push(resultString + resultString2 + resultString3);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JPackageDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JPackageDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Lista de parametros nula");
        }

        String resultString = "";
        if (node.isASourceNode()) {
            String tmp = this.state.pop();
            resultString = "package " + tmp + ";";


        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node) + ";";
            if (node.isOptional()) {
                resultString = "#[" + resultString + "]#";
            }
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JParameterList</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JParameterList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Lista de parametros nula");
        }

        String resultString = "";
        String tmp = "";

        int max = node.size();

        if (node.isASourceNode()) {
            tmp = this.prettyPrintList(", ", max, "", false);
            resultString = "(" + tmp + ")";

        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = "(" + node.printVariable() + ")";
            } else if (node.isExecutable()) {
                resultString = node.getExecutableDeclaration().toString();
            }

        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JPreconditionDeclaration</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JPreconditionDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JPreconditionDeclaration nulo");
        }

        if (!jatsTypesEnabled) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        String resultString = "";

        resultString += "precondition " + "#<" + this.state.pop() + ">#" + ";";
        // s pra eliminar a declarao executvel associada
        this.state.pop();

        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JResultStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JResultStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String res = "";
        String expr = "";

        if (node.isReturn()) {
            res = "return";
        } else {
            res = "throw";
        }

        if (node.getExpression() != null) {
            expr = this.state.pop().trim();
        }

        if (node.hasComment()) {
            sbuffer.append(node.getComment());
            sbuffer.append(lineBreak(1));
        }
        sbuffer.append(res);
        if (expr.length() > 0) 
        	sbuffer.append(space());
        sbuffer.append(expr);
        sbuffer.append(";");

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;

    }

    /**
     * Mtodo visit para objetos do tipo <code>JSwitchStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JStatementExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";

        if (node.getExpression() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException("expressao nula");
        }
        resultString = this.state.pop() + ";";
        this.state.push(resultString);
        return resultString;


    }

    /**
     * Mtodo visit para objetos do tipo <code>JSwitchStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String exp = "";

        if (node.getExpression() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException("Falta a expressao do switch");
        }
        exp = this.state.pop();


        sbuffer.append(space());
        sbuffer.append("switch (");
        sbuffer.append(exp);
        sbuffer.append(")");
        sbuffer.append("{ ");
        
        StringBuffer body = new StringBuffer();
        
        if (node.getCases() != null) {
            int size = node.getCases().size();
            for (int i = 0; i < size; i++) {
                body.append(lineBreak(1));
                body.append(this.state.pop());
            }
        }

        sbuffer.append(indentSequence(body.toString(), false));
        sbuffer.append(lineBreak(1));
        sbuffer.append("}");

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;

    }

    /**
     * Mtodo visit para objetos do tipo <code>JSwitchLabel</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchLabel node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        if (node.getType() == JSwitchLabel.CASE) {
            sbuffer.append("case");
            sbuffer.append(space());
            if (node.getExpression() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException("");
            }
            String expr = this.state.pop();
            sbuffer.append(expr);
            sbuffer.append(space());
            sbuffer.append(":");

        } else {
            sbuffer.append("default");
            sbuffer.append(space());
            sbuffer.append(":");

        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JSwitchPair</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchPair node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        // representa a expressao do case e os statements associados ao case
        // pegar o nivel de indentacao do statement e usar para cada stament
        // presente no case

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();

        // label
        sbuffer.append(this.state.pop());

        // lista de statements
        if (node.getStatements() != null) {
            NodeList stmts = node.getStatements();
            int size = stmts.size();

            for (int i = 0; i < size; i++) {
                sbuffer.append(lineBreak(1));
                sbuffer.append(indentString);
                sbuffer.append(this.state.pop());
            }
        }
        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;

    }

    /**
     * Mtodo visit para objetos do tipo <code>JTernaryExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JTernaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.getThirdOperand() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String third = this.state.pop();

        if (node.getSecondOperand() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String second = this.state.pop();

        if (node.getFirstOperand() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String first = this.state.pop();

        resultString += node.buildTernaryExpressionStringFor(first, second,
                third);

        this.state.push(resultString);

        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JTryStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JTryStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        String tblock = "";
        String catchlist = "";
        String fblock = "";

        String tryc = node.getTryComment();
        if (Util.checkString(tryc)) {
            sbuffer.append(tryc);
            sbuffer.append(lineBreak);
        }
        sbuffer.append("try");
        sbuffer.append(space());

        // bloco try
        if (node.getTryBlock() != null) {
            tblock = this.state.pop();
            if (bracesOnNewLine) {
                sbuffer.append(lineBreak(1));
            }
            sbuffer.append(tblock);
        }

        // lista de clusulas catch
        if (node.getCatchlist() != null && node.getCatchListSize() > 0) {
            int size = node.getCatchListSize();
            if (catchOnNewLine) {
                catchlist = this.prettyPrintList(lineBreak(1), size, "", false);
                catchlist = lineBreak(1) + catchlist;
            } else {
                catchlist = this.prettyPrintList(space(), size, "", false);
                catchlist = space() + catchlist;
            }
            sbuffer.append(catchlist);

        }
        String finallyc = node.getFinallyComment();
        boolean validComment = Util.checkString(finallyc);

        // bloco finally
        if (node.getFinallyBlock() != null) {
            fblock = this.state.pop();
            if (finallyOnNewLine) {
                sbuffer.append(lineBreak(1));
                if (validComment) {
                    sbuffer.append(finallyc);
                    sbuffer.append(lineBreak);
                }
            } else {
                sbuffer.append(space());
                if (validComment) {
                    sbuffer.append(finallyc);
                }
            }
            sbuffer.append("finally");
            if (bracesOnNewLine) {
                sbuffer.append(lineBreak(1));
            } else {
                sbuffer.append(space());
            }
            sbuffer.append(fblock);
        }
        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;

    }


    /**
     * Mtodo visit para objetos do tipo <code>JType</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JType node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.getTypeName() == null && node.isASourceNode()) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        if (node.isASourceNode()) {
            String type = this.state.pop();
            String brackets = "";
            int i = 0;
            int max = node.getNumberOfDimensions();
            while (i < max) {
                brackets += "[]";
                i++;
            }
            resultString = type + brackets;
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>TypeDeclarationSet</code>.
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(TypeDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";
        int max = node.size();

        if (node.isASourceNode()) {
            resultString = this.prettyPrintList(lineBreak(2), max, "", false);
        } else if (jatsTypesEnabled) {
            resultString = printJaTSNode(node);
        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);

        return resultString;
    }

    /**
     * Mtodo visit para objetos do tipo <code>JUnaryExpression</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JUnaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.getOperand() == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        String operand = this.state.pop();
        resultString = node.buildUnaryExpressionStringFor(operand);

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JVariableDeclarator</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JVariableDeclarator node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        String resultString = "";

        if (node.isASourceNode()) {

            if (node.getIdentifier() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            String name = this.state.pop();

            int i = 0;
            String dims = "";
            while (i < node.getDimensions()) {
                dims += "[]";
                i++;
            }
            resultString += name + dims;
            String init = "";
            if (node.getInitializer() != null) {
                init = this.state.pop();
                resultString = resultString + " = " + init;
            }

        } else if (jatsTypesEnabled) {

            if (node.isVariable()) {
                resultString = node.printVariable();
            } else if (node.isExecutable()) {
                resultString = node.printExecutableDeclaration();
                this.state.pop();

            }

        } else {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node,
                    cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException.INVALID_CONSTRUCTION);
        }

        this.state.push(resultString);
        return resultString;
    }


    /**
     * Mtodo visit para objetos do tipo <code>JwhileStatement</code>.
     *
     * @param node O n a ser impresso.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JWhileStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        String resultString = "";
        StringBuffer sbuffer = new StringBuffer();
        StringBuffer corpo = new StringBuffer();
        String exp = "";
        String stmt = "";

        if (node.getExpression() != null) {
            exp = this.state.pop().trim();
        }
        if (node.getStatement() != null) {
            stmt = this.state.pop();
            if (node.getStatement() instanceof JBlock) {
                if (bracesOnNewLine) {
                    corpo.append(lineBreak(1));
                }
                corpo.append(stmt);
            } else {
            	corpo.append(lineBreak(1));
                //o statement puro, naturalmente, nao vem indentado
                //logo, excepcionalmente, precisa-se indenta-lo aqui
                //(ha ainda a opcao de manter na mesma linha do if)
                corpo.append(stmt);
            }
        }
        
        // se for um while
        if (node.isWhile()) {
            sbuffer.append("while");
            sbuffer.append("(");
            sbuffer.append(exp);
            sbuffer.append(")");
            sbuffer.append(space());
            sbuffer.append(corpo.toString());
        }
        // seno, ento trata-se de um do
        else {
            sbuffer.append("do");
            sbuffer.append(space());
            sbuffer.append(corpo.toString());
            if (whileOnNewLine) {
                sbuffer.append(lineBreak(1));
            } else {
                sbuffer.append(space());
            }
            sbuffer.append("while");
            sbuffer.append("(");
            sbuffer.append(exp);
            sbuffer.append(")");
        }

        resultString = sbuffer.toString();
        this.state.push(resultString);
        return resultString;
    }

    public boolean isJatsTypesEnabled() {
        return jatsTypesEnabled;
    }

    public void setJatsTypesEnabled(boolean jatsTypesEnabled) {
        this.jatsTypesEnabled = jatsTypesEnabled;
    }

    /**
     * Imprime o n JaTS. Seja ele varivel, executvel ou uma orDeclaration
     * @param node   N a ser impresso
     * @return       String que representa a impresso do n
     */
    protected String printJaTSNode(AbstractNode node) {
        String resultString = "";
        if (node.isVariable()) {
            resultString = node.printVariable();
        } else if (node.isExecutable()) {
            resultString = node.printExecutableDeclaration();
            this.state.pop();
        } else if (node.isOrDeclaration()) {
            resultString = node.printOrDeclaration();
        }
        return resultString;
    }


//	public Object visit(JStatementList node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
//		// for a while, doesn't happens 
//		throw new InconsistentNodeException("A JStamentList remained! Reached by pretty printer");
//	}


	public Object visit(ExploreDeclaration node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
		// shouldn't happen
		throw new InconsistentNodeException("An ExploreStatement remained! Reached by pretty printer");
	}

}
