/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.visitors;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.util.*;


/**
 * Classe que implementa a interface <code>IVisitor</code> por meio de um
 * visitor que realiza a substituio de ocorrncias de variveis na rvore
 * sinttica de um programa JaTS pelos valores mapeados a elas no conjunto
 * resultado do casamento.<br>
 * Created: 14/11/2002
 *
 * @see INode
 * @see IVisitor
 * 
 * @author Fernando Castor
 * @author Adeline Sousa
 * @author Gustavo Santos
 * 
 * @version 12/10/2005
 */
public class ReplacementVisitor extends AbstractVisitor {

    // Armazena o conjunto de pares <code>(varivel, valor)</code> que ser usado na
    // substituio.
    private ResultSet results;


    /**
     * Construtor vazio da classe.
     */
    public ReplacementVisitor() {
        this.results = new ResultSet();
    }

    /**
     * Construtor que recebe como argumento o conjunto resultado a partir do
     * qual sero extrados os valores mapeados s variveis que forem
     * encontradas no decorrer da substituio.
     *
     * @param resultSet Um objeto do tipo <code>ResultSet</code>,
     * contendo mapeamentos entre
     * variveis JaTS e valores correspondentes a estruturas de Java que
     * sero usadas para substituir as ocorrncias dessas variveis na
     * rvore visitada.
     */
    public ReplacementVisitor(ResultSet resultSet) {
        this.results = resultSet;
    }


    public void setResults(ResultSet results) {
        this.results = results;
    }

    /**
     * Verifica se um n do tipo JExpression, correspondendo a uma varivel, est
     * casado a uma expresso. Em caso positivo, retorna esse n. Caso contrrio,
     * levanta uma exceo.
     */
    private JExpression getMappedExpression(JExpression exp) throws
            IllegalArgumentException {
        if (exp == null || !exp.isVariable())
            throw new IllegalArgumentException();


        INode tempExp = this.getValueOf(exp);


        JExpression retorno = null;

        if (tempExp != null && !(tempExp instanceof JExpression)) {
            throw new IllegalMatchingException(exp);
        } else if (tempExp != null) {
            retorno = (JExpression) tempExp;
        } else {
            retorno = exp;
        }

        return retorno;
    }


    /**
     * Verifica se um n do tipo JIdentifier, correspondendo a uma varivel, est
     * casado a um outro identificador. Em caso positivo, retorna esse n. Caso
     * contrrio, levanta uma exceo.
     */
    private JIdentifier getMappedIdentifier(JIdentifier identifier) throws
            IllegalArgumentException {

        if (identifier == null || !identifier.isVariable()) {
            throw new IllegalArgumentException();
        }

        Object tempId = this.getValueOfIdentifier(identifier);
        JIdentifier retorno = null;

        if (tempId != null) {
            try {
                retorno = JIdentifier.unwrapIdentifier(tempId);
            } catch (IllegalArgumentException e) {
                throw new IllegalMatchingException(e.getMessage(), identifier);
            }
        } else {
            retorno = identifier;
        }

        return retorno;
    }
    
    /**
     * Mtodos visit para objetos do tipo <code>JSynchronizedStatement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     * @throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
	public Object visit(JSynchronizedStatement node, Object data) throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException {
		 if (node == null) {
            throw new IllegalArgumentException();
        }

        JExpression lockedObject = node.getLockedObject();
        JStatement statement = node.getStatement();

        if (lockedObject == null || statement == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (lockedObject.isVariable()) {
            node.setLockedObject(this.getMappedExpression(lockedObject));
        }

        if (statement.isVariable()) {
            node.setStatement((JStatement) this.getValueOf(statement));
        }

        return node;
	}


    /**
     * Verifica se um n do tipo JModifierList, correspondendo a uma varivel,
     * est casado com um n do mesmo tipo. Em caso positivo, retorna esse n.
     * Caso contrrio, levanta uma exceo.
     */
    private JModifierList getMappedModifierList(JModifierList mods,
                                                int modListType) {
        if (mods == null || modListType <= 0 || !mods.isVariable())
            throw new IllegalArgumentException();

        INode tempMods = this.getValueOf(mods);
        JModifierList retorno = null;

        if (tempMods != null && (!(tempMods instanceof JModifierList))) {
            throw new IllegalMatchingException(mods);
        }
        // se a lista de modificadores atuais de tempMods no for
        // compatvel com a modListType, ento ocorreu um erro.
        if (tempMods != null && !((JModifierList) tempMods).isCompatible(modListType)) {
            throw new IllegalMatchingException(mods);
        }
        if (tempMods != null) {
            retorno = (JModifierList) tempMods;
        }
        // Se a varivel no for encontrada no conjunto resultado,
        // nada  feito.
        else {
            retorno = mods;
        }

        return retorno;
    }

    /**
     * Verifica se um n do tipo <code>JName</code>  uma varivel e, em
     * caso afirmativo, se est mapeado a algum valor. Se estiver, o
     * mtodo retorna esse valor. <p>
     *
     * Caso a varivel no esteja mapeada a nada,  verificado se ela 
     * parte de uma estrutura opcional. Se for, <code>null</code> 
     * retornado. Em caso negativo, o prprio n  retornado como
     * resultado do mtodo.
     */
    private JName resolveName(JName name) throws IllegalArgumentException {

        if (name == null) {
            throw new IllegalArgumentException();
        }
        Object tempNode = null;
        JName retorno = null;

        if (name.isVariable() || name.isOrDeclaration()) {
            tempNode = this.getValueOfName(name);
            if (tempNode != null) {
                try {
                    retorno = JName.unwrapName(tempNode);
                } catch (IllegalArgumentException e) {
                    throw new IllegalArgumentException("Erro ao fazer unwrap de " + name + " Illegal");
                } catch (cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException e) {
                    throw new IllegalArgumentException("Erro ao fazer unwrap de " + name + " InconsistentNode");
                }
            }
        }
        if (tempNode == null && !name.isOptional()) {
            retorno = name;
        }
        // Caso o n no seja uma varivel ou no tenha um valor mapeado
        // e seja opcional, null  retornado.

        return retorno;
    }

    /**
     * Verifica se um n do tipo <code>JNameList</code>  uma varivel e, em
     * caso afirmativo, se est mapeado a algum valor. Se estiver, o
     * mtodo retorna esse valor. <p>
     *
     * Caso a varivel no esteja mapeada a nada,  verificado se ela 
     * parte de uma estrutura opcional. Se for, <code>null</code> 
     * retornado. Em caso negativo, o prprio n  retornado como
     * resultado do mtodo.
     */
    private JNameList resolveNameList(JNameList nl) throws
            IllegalArgumentException {
        if (nl == null) throw new IllegalArgumentException();

        INode tempNL = null;
        JNameList retorno = null;

        // Se nl for uma varivel, deve ser verificado se tem um
        // valor mapeado no result set.
        if (nl.isVariable() || nl.isOrDeclaration()) {
            tempNL = this.getValueOf(nl);

            // Tem um valor mapeado mas ele no  do tipo correto.
            if (tempNL != null && !(tempNL instanceof JNameList)) {
                throw new IllegalMatchingException(nl);
            }
            // Valor mapeado do tipo correto.
            else if (tempNL != null) {
                retorno = (JNameList) tempNL;
            }
        }

        // O n nl no  uma varivel ou no tem um valor mapeado e
        // no  opcional. Nesse caso, ele prprio  retornado.
        //if (tempNL == null && !nl.isOptional()) {
        //    retorno = nl;
        //}
        else {
            retorno = nl;
        }

        // Caso o n no seja uma varivel ou no tenha um valor mapeado
        // e seja opcional, null  retornado.

        return retorno;
    }

    /**
     *  Verifica se um n do tipo JParameterList, correspondendo a uma varivel, est
     *  casado a uma outra lista de nomes. Em caso positivo, retorna esse n. Caso
     *  contrrio, levanta uma exceo.
     */
    private JParameterList getMappedParameterList(INode pl) throws
            IllegalArgumentException {

        if ((pl == null) || (!pl.isVariable())) {
            throw new IllegalArgumentException();
        }
        INode tempPL = this.getValueOf(pl);
        JParameterList retorno = null;
        if (tempPL != null && !(tempPL instanceof JParameterList)) {
            throw new IllegalMatchingException(pl);
        } else if (tempPL != null) {
            retorno = (JParameterList) tempPL;
        } else if (pl instanceof JParameterList) {
            retorno = (JParameterList) pl;
        }

        return retorno;
    }


    /**
     * Verifica se um n do tipo JType, correspondendo a uma varivel, est
     * casado a um outro tipo. Em caso positivo, retorna esse n. Caso
     * contrrio, levanta uma exceo.
     */
    private JType getMappedType(JType type) throws
            IllegalArgumentException {
        if (type == null || !type.isVariable())
            throw new IllegalArgumentException();

        INode tempType = this.getValueOf(type);

        JType retorno = null;

        if (tempType != null) {
            try {
                retorno = JType.unwrapType(tempType);
            } catch (IllegalArgumentException e) {
                throw new IllegalMatchingException(e.getMessage(), type);
            }
        } else {
            retorno = type;
        }
        return retorno;
    }


    /**
     * Retorna o conjunto resultado utilizado na substituio.
     *
     * @return Um objeto do tipo <code>ResultSet</code>, contendo mapeamentos
     *        entre variveis JaTS e objetos correspondendo a estruturas de Java.
     */
    public ResultSet getResultSet() {
        return this.results;
    }

    /**
     * Retorna o valor que estiver mapeado  varivel <code>variable</code>
     * em <code>results</code>, caso haja algum.
     *
     * @param variable Um n correspondendo a uma varivel.
     *
     * @return O valor mapeado a <code>variable</code>, caso haja algum.
     */
    public INode getValueOf(INode variable) {
        INode retorno = null;
        if (variable != null &&
                (/*variable.isOrDeclaration() ||*/ variable.isVariable())) {
            retorno = (INode) results.get(variable);

            //QualitiLogger.debug("variable = " + variable + " retorno = " + retorno);            
            
            return retorno;
        } else {
            return retorno;
        }
    }


    public Object getValueOfName(INode name) {
        Object retorno = null;
        if (name.isVariable() /*|| name.isOrDeclaration()*/) {
            retorno = results.get(name);
        }
        return retorno;
    }


    public Object getValueOfIdentifier(INode variable) {
        Object retorno = null;
        if (variable.isVariable()) {
            retorno = results.get(variable);
        }
        return retorno;
    }

    /**
     * Mtodo auxiliar para a substituio de declaraes condicionais.
     */
    private void replaceConditionalDeclaration(JConditionalDeclaration node)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        NodeList tDecSet = node.getTrueDeclarationSet();
        NodeList fDecSet = node.getFalseDeclarationSet();
        JExpression condition = node.getCondition();
        if (tDecSet == null || condition == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (condition.isVariable()) {
            INode tempCond = this.getValueOf(condition);
            if (tempCond != null && (tempCond instanceof JExpression)) {
                node.setCondition((JExpression) tempCond);
            } else {
                throw new IllegalMatchingException(node);
            }
        }

        int i = 0;
        int max = fDecSet.size();

        while (i < max) {
            INode tempDec = fDecSet.elementAt(i);

            if (tempDec.isVariable()) {
                INode res = this.getValueOf(tempDec);

                if (res != null) {
                    fDecSet.setElementAt(res, i);
                }
            }
            i++;
        }
        int j = 0;
        int max2 = tDecSet.size();
        while (j < max2) {
            INode tempDec2 = tDecSet.elementAt(j);
            if (tempDec2.isVariable()) {
                INode res = this.getValueOf(tempDec2);
                if (res != null) {
                    tDecSet.setElementAt(res, j);
                }
            }
            j++;
        }

    }

    /**
     * Mtodo auxiliar para a substituio de declaraes iterativas.
     */
    private void replaceIterativeDeclaration(IterativeDeclaration node)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        INode iterSet = node.getIterationSet();
        INode ph = node.getPlaceholder();
        NodeList decSet = node.getDeclarationSet();
        if (iterSet == null || ph == null || decSet == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        if (ph.isVariable()) {
            INode tempPH = this.getValueOf(ph);
            // O placeholder no pode estar mapeado a qualquer valor.
            // Se estiver, alguma coisa est errada.
            if (tempPH != null) {
                throw new IllegalMatchingException(node);
            }
        }
        int i = 0;
        int max = decSet.size();
        while (i < max) {
            INode tempDec = decSet.elementAt(i);
            if (tempDec.isVariable()) {
                INode res = this.getValueOf(tempDec);
                if (res != null) {
                    decSet.setElementAt(res, i);
                }
            }
            i++;
        }
        if (iterSet.isVariable()) {
            INode tempIterSet = this.getValueOf(iterSet);
            if (tempIterSet != null && tempIterSet instanceof JaTSCollection) {
                node.setIterationSet(tempIterSet);
            }
        }
    }

    /**
     * Mtodos visit para objetos do tipo <code>JAllocationExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JAllocationExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        if (node.getAllocationType() == JAllocationExpression.OBJECTALLOCATION) {

            if (node.getArguments() == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            JExpressionList args = node.getArguments();

            if (args.isVariable()) {
                INode newArgList = this.getValueOf(args);

                if (newArgList != null && newArgList instanceof JExpressionList) {
                    node.setArguments((JExpressionList) newArgList);
                } else if (newArgList != null) {
                    throw new IllegalMatchingException(args);
                }
            }

            TypeBody body = node.getInternalClassBody();
            if (body != null && body.isVariable()) {
                INode tmpBody = this.getValueOf(body);
                if (tmpBody != null && tmpBody instanceof TypeBody) {
                    node.setInternalClassBody((TypeBody) tmpBody);
                } else if (tmpBody != null) {
                    throw new IllegalMatchingException(tmpBody);
                }
            }

        } else if (node.getAllocationType() ==
                JAllocationExpression.PRIMITIVETYPEARRAYALLOCATION
                || node.getAllocationType() ==
                JAllocationExpression.OBJECTARRAYALLOCATION) {

            if (node.getArrayDimsAndInits() == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            JArrayDimsAndInits adi = node.getArrayDimsAndInits();
            if (adi.isVariable()) {
                INode tempADI = this.getValueOf(adi);

                if (tempADI != null && tempADI instanceof JArrayDimsAndInits) {
                    node.setArrayDimsAndInits((JArrayDimsAndInits) tempADI);
                } else if (tempADI != null) {
                    throw new IllegalMatchingException(tempADI);
                }
            }

        }
        if (node.getTypeName() == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        JName name = node.getTypeName();

        if (name.isVariable()) {
            node.setName(this.resolveName(name));
        }

        return node;
    }

    /**
     * @see JAnonymousClass
     */
    public Object visit(JAnonymousClass node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node.isASourceNode()) {
            JName className = node.getName();
            JExpressionList args = node.getArguments();
            TypeBody body = node.getBody();

            // Nome da classe
            if (className.isVariable()) {
                node.setName(this.resolveName(className));
            }

            // argumentos
            if (args.isVariable()) {
                node.setArguments((JExpressionList) this.getValueOf(args));
            }

            // Corpo da classe
            if (body.isVariable()) {
                INode tempBody = this.getValueOf(body);

                if (tempBody != null && tempBody instanceof TypeBody) {
                    node.setBody((TypeBody) tempBody);
                } else if (tempBody != null) {
                    throw new IllegalMatchingException(body);
                }
            }
        }
        return node;
    }

    /**
     * @see JArrayAccess
     */
    public Object visit(JArrayAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        if (node.isASourceNode()) {
            JExpressionList refExpressions = node.getIndexers();
            INode arrayName = node.getArrayName();

            if (refExpressions != null) {
                int max = refExpressions.size();
                // varre a  lista de expresses
                for (int i = 0; i < max; i++) {
                    JExpression tmpExpr = refExpressions.getExpressionAt(i);
                    // se o elemento for varivel, substitui pelo valor que a ele
                    // estiver mapeado
                    if (tmpExpr.isVariable()) {
                        refExpressions.setExpressionAt(this.getMappedExpression(tmpExpr), i);
                    }
                }
            }
            if (arrayName == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            if (arrayName.isVariable()) {
                node.setArrayName(this.getValueOf(arrayName));
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JArrayDimsandInits</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JArrayDimsAndInits node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (node.getArrayInitializer() != null) {
            INode aInit = node.getArrayInitializer();

            if (aInit == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            if (aInit.isVariable()) {
                INode tempInit = this.getValueOf(aInit);

                if (tempInit != null && tempInit instanceof JArrayInitializer) {
                    node.setArrayInitializer((JArrayInitializer) tempInit);
                } else if (tempInit != null) {
                    throw new IllegalMatchingException(tempInit);
                }
            }
        }

        int max = 0;
        if ((max = node.getNumberOfDimensionSizesSpecified()) > 0) {
            int j = 1;
            while (j <= max) {
                JExpression dimSize = node.getSizeOfDimension(j);

                if (dimSize == null)
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

                if (dimSize.isVariable()) {
                    node.setSizeOfDimension(this.getMappedExpression(dimSize), j);
                }

                j++;
            }
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JArrayInitializer</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JArrayInitializer node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        int j = 1, max = node.size();

        while (j <= max) {
            INode varInit = node.getInitializerAt(j);
            if (varInit.isVariable()) {
                INode tempInit = this.getValueOf(varInit);

                if (tempInit != null && (tempInit instanceof JExpression ||
                        tempInit instanceof JArrayInitializer)) {
                    node.setInitializerAt(tempInit, j);
                } else if (tempInit != null) {
                    throw new IllegalMatchingException(tempInit);
                }
            }
            j++;
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JBinaryExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JBinaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        JExpression leftHand = node.getLeftHandOperand();
        JExpression rightHand = node.getRightHandOperand();

        if (leftHand == null || rightHand == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (leftHand.isVariable()) {
            //node.setLeftHandOperand(new JUnaryExpression(
            //		JUnaryExpression.PARENTHESIZEDEXPRESSION,
            //		this.getMappedExpression(leftHand)));

            node.setLeftHandOperand(this.getMappedExpression(leftHand));

        }

        if (rightHand.isVariable()) {
            //node.setRightHandOperand(new JUnaryExpression(
            //		JUnaryExpression.PARENTHESIZEDEXPRESSION,
            //		this.getMappedExpression(rightHand)));

            node.setRightHandOperand(this.getMappedExpression(rightHand));
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JBlock</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JBlock node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (node.isASourceNode()) {
            JStatementList statements = node.getStatements();
            JStatementList newStatements = new JStatementList();
            if (statements != null) {
                for (int i = 0; i < statements.size(); i++) {
                    INode tmpNode = statements.elementAt(i);
                    if (tmpNode.isVariable()) {
                        INode newValue = this.getValueOf(tmpNode);
                        
                        if (newValue != null) { 
                        	if (newValue instanceof JStatementList) {
                        		newStatements.getStatements().insertNodeListAt(
                        				((JStatementList)newValue).getStatements(), 
                        				newStatements.size());
                        	} else {
                        		newStatements.getStatements().addElement((newValue));
//		                        if (newValue.getIndentLevel() == 0) {
//		                            newValue.setIndentLevel(tmpNode.getIndentLevel());
//		                        }
                        	}
                        }
                    } else {
                    	newStatements.add(tmpNode);
                    }
                }
                
                node.setStatements(newStatements);
            }
        }
        
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>CastExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(CastExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        JType type = node.getType();
        JExpression operand = node.getOperand();

        if (operand == null || type == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        // Tipo para o qual o objeto ser "Casted"
        if (type.isVariable()) {
            node.setType(this.getMappedType(type));
        }

        // Operando
        if (operand.isVariable()) {
            node.setOperand(this.getMappedExpression(operand));
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>TypeBody</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(TypeBody node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        // os filhos j foram substitudos
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JCatchClause</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JCatchClause node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());

        if (nodeHasNoBody) {

            // substitui o parametro, caso ele seja variavel
            JFormalParameter param = node.getParameter();
            if (param != null && param.isVariable()) {
                INode fparam = this.getValueOf(param);
                if (fparam instanceof JFormalParameter) {
                    node.setParameter((JFormalParameter) fparam);
                }
            }

            // substitui o bloco, caso ele seja variavel
            JBlock block = node.getBlock();
            if (block != null && block.isVariable()) {
                INode tmpBlock = this.getValueOf(block);
                if (tmpBlock instanceof JBlock) {
                    node.setBlock((JBlock) tmpBlock);
                }
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JClassDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JClassDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Declaracao de classe nula");
        }
        if (node.isASourceNode()) {
            JModifierList mods = node.getModifiers();
            TypeBody body = node.getBody();
            JNameList interfaces = node.getInterfaces();
            JName superClass = node.getSuperClass();
            JIdentifier className = node.getName();
            
            if (mods == null || body == null || className == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Comentrios do classe
            if (node.getComment() == null) {
                String commentVar = Util.generateTypeCommentVariable(node.getName());
                JComment comment = new JComment(commentVar);
                Object tmpComment = this.getValueOf(comment);
                if (tmpComment instanceof JComment) {
                    node.setComment((JComment) tmpComment);
                }
            }

            // Lista de Modificadores da classe
            if (mods.isVariable()) {
                JModifierList ml =
                        this.getMappedModifierList(mods,
                                JModifierList.CLASS_DECLARATION);
                node.setModifiers(ml);
            }

            // Nome da classe
            if (className.isVariable()) {
                node.setName(this.getMappedIdentifier(className));
            }

            // Nome da superclasse, caso algum tenha sido declarado
            if (superClass != null) {
                JName sc = this.resolveName(superClass);
                node.setSuperClass(sc);
            }

            // Lista de interfaces que a classe implementa, se houver
            if (interfaces != null &&
                    (interfaces.isVariable() || interfaces.isOrDeclaration())) {
                JNameList nl = this.resolveNameList(interfaces);
                node.setInterfaces(nl);
            }

            // Corpo da classe
            if (body.isVariable()) {
                INode tempBody = this.getValueOf(body);

                if (tempBody != null && tempBody instanceof TypeBody) {
                    node.setBody((TypeBody) tempBody);
                } else if (tempBody != null) {
                    throw new IllegalMatchingException(body);
                }
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JCompilationUnit</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JCompilationUnit node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        JPreconditionDeclaration pre = node.getPrecondition();

        JPackageDeclaration pack = node.getPackage();

        // ns do tipo JCompilationUnit nunca so variveis
        if (node.isASourceNode()) {
            // pr-condio
            if (pre != null && pre.isVariable()) {
                node.setPrecondition((JPreconditionDeclaration) this.getValueOf(pre));
            }

            // Declaracao de Pacote
            if (pack != null && pack.isVariable()) {
                INode pd = this.getValueOf(pack);
                // aqui, preciso ver se o pd veio null pq o n  opcional. Neste
                // caso, posso set-lo para null mesmo, senao fica o n que
                // tava l sem ser substitudo e isso pode gerar uma
                // InconsistentNodeException
                if ((pd instanceof JPackageDeclaration) ||
                        (pd == null && pack.isOptional())) {
                    node.setPackage((JPackageDeclaration) pd);
                }

            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JConditionalDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JConditionalDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        this.replaceConditionalDeclaration(node);

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JConstructorDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JConstructorDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JConstructorDeclaration nulo");
        }
        if (node.isASourceNode()) {
            JModifierList mod = node.getModifiers();
            JIdentifier id = node.getName();
            JParameterList param = node.getParameters();
            JNameList excs = node.getExceptions();
            JBlock body = node.getBody();


            // Comentrios do construtor
            if (node.getComment() == null) {
                String commentVar = Util.generateCommentVariable(node.getName());
                JComment comment = new JComment(commentVar);
                Object tmpComment = this.getValueOf(comment);
                if (tmpComment instanceof JComment) {
                    node.setComment((JComment) tmpComment);
                }
            }

            if ((mod == null) || (id == null)) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Lista de Modificadores da declarao de construtor
            if (mod.isVariable()) {
                node.setModifiers(this.getMappedModifierList(mod,
                        JModifierList.CONSTRUCTOR_DECLARATION));
            }

            // Identificador do construtor
            if (id.isVariable()) {
                node.setName(this.getMappedIdentifier(id));
            }

            // Lista de Parametros do construtor
            if (param.isVariable()) {
                node.setParameters(this.getMappedParameterList(param));
            }

            // Lista excees que o construtor lana
            if (excs != null && (excs.isVariable() || excs.isOrDeclaration())) {
                node.setExceptions(this.resolveNameList(excs));
            }
            // corpo do mtodo
            if (body != null && body.isVariable()) {
                node.setBody((JBlock) this.getValueOf(body));

            }
        }
        return node;

    }



//    public  Object visit( EnumMember node,  Object data) throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            AttributeSectionList tmpAttributes = node.getAttributes();
//            if (tmpAttributes != null && tmpAttributes.isVariable()) {
//                node.setAttributes((AttributeSectionList) this.getValueOf(tmpAttributes));
//            }
//            JIdentifier tmpIdentifier = node.getIdentifier();
//            if (tmpIdentifier.isVariable()) {
//                node.setIdentifier((JIdentifier) this.getValueOf(tmpIdentifier));
//            }
//            JExpression tmpExpression = node.getExpression();
//            if (tmpExpression.isVariable()) {
//                node.setExpression((JExpression) this.getValueOf(tmpExpression));
//            }
//        }
//        return node;
//    }

//    public  Object visit( EnumDeclaration node,  Object data) throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            AttributeSectionList tmpAttributes = node.getAttributes();
//            if (tmpAttributes != null && tmpAttributes.isVariable()) {
//                node.setAttributes((AttributeSectionList) this.getValueOf(tmpAttributes));
//            }
//            JModifierList tmpModifiers = node.getModifiers();
//            if (tmpModifiers.isVariable()) {
//                node.setModifiers((JModifierList) this.getValueOf(tmpModifiers));
//            }
//            JIdentifier tmpName = node.getName();
//            if (tmpName.isVariable()) {
//                node.setName((JIdentifier) this.getValueOf(tmpName));
//            }
//            JType tmpType = node.getType();
//            if (tmpType.isVariable()) {
//                node.setType((JType) this.getValueOf(tmpType));
//            }
//            NodeList tmpEnumMembers = node.getEnumMembers();
//            int max = tmpEnumMembers.size();
//            INode tmpVar = null;
//            for (int i = 0; i < max; i++) {
//                tmpVar = node.getEnumMemberAt(0);
//                if (tmpVar.isVariable()) {
//                    node.setEnumMemberAt((EnumMember) this.getValueOf(tmpVar), i);
//                }
//            }
//        }
//        return node;
//    }


//	public  Object visit( AttributeSectionList node,  Object data) throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            NodeList tmpArguments = node.getAttributeSections();
//            int max = tmpArguments.size();
//            INode tmpVar = null;
//            for(int i = 0; i <max; i++){
//                tmpVar = node.getAttributeSectionAt(i);
//                if (tmpVar.isVariable()) {
//                    node.setAttributeSectionAt((AttributeSection)this.getValueOf(tmpVar),i);
//                }
//            }
//        }
//        return node;
//    }
//
//    public  Object visit( AttributeSection node,  Object data) throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            JIdentifier tmpAttributeTarget = node.getAttributeTarget();
//            if (tmpAttributeTarget.isVariable()) {
//                node.setAttributeTarget((JIdentifier)this.getValueOf(tmpAttributeTarget));
//            }
//            AttributeList tmpAttributes = node.getAttributes();
//            if (tmpAttributes.isVariable()) {
//                node.setAttributes((AttributeList)this.getValueOf(tmpAttributes));
//            }
//        }
//        return node;
//    }

//    public  Object visit( AttributeList node,  Object data) throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            NodeList tmpArguments = node.getAttributes();
//            int max = tmpArguments.size();
//            INode tmpVar = null;
//            for(int i = 0; i <max; i++){
//                tmpVar = node.getAttributeAt(0);
//                node.setAttributeAt((Attribute)this.getValueOf(tmpVar),i);
//            }
//        }
//        return node;
//    }

//    public  Object visit( AttributeArgumentList node,  Object data) throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//
//        	NodeList tmpArguments = node.getArguments();
//            int max = tmpArguments.size();
//            INode tmpVar = null;
//            for(int i = 0; i <max; i++){
//                tmpVar = node.getArgumentAt(0);
//                if (tmpVar.isVariable()) {
//                    node.setArgumentAt((AttributeArgument)this.getValueOf(tmpVar),i);
//                }
//            }
//        }
//        return node;
//    }
//
//    public  Object visit( AttributeArgument node,  Object data) throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            JIdentifier tmpIdentifier = node.getIdentifier();
//            if (tmpIdentifier.isVariable()) {
//                node.setIdentifier((JIdentifier)this.getValueOf(tmpIdentifier));
//            }
//            JExpression tmpExpression = node.getExpression();
//            if (tmpExpression.isVariable()) {
//                node.setExpression((JExpression)this.getValueOf(tmpExpression));
//            }
//        }
//        return node;
//    }

//    public  Object visit( Attribute node,  Object data) throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException{
//        if (node == null) {
//            throw new IllegalArgumentException("node is null");
//        }
//        if (node.isASourceNode()) {
//            JName tmpName = node.getName();
//            if (tmpName.isVariable()) {
//                node.setName((JName)this.getValueOf(tmpName));
//            }
//            AttributeArgumentList tmpArgumentList = node.getArgumentList();
//            if (tmpArgumentList.isVariable()) {
//                node.setArgumentList((AttributeArgumentList)this.getValueOf(tmpArgumentList));
//            }
//        }
//        return node;
//    }

    /**
     * Mtodos visit para objetos do tipo <code>JConstructorDeclarationSet</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JConstructorDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Lista de declaracoes de metodo nula");
        }

        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // atributo que correspondem a variveis so tratadas em
        // visit(ClassBodyConstructorDeclarations).

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>ClassBodyConstructorDeclarations</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(ClassBodyConstructorDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Conjunto de declaracoes de contrutor nulo");
        }

        JConstructorDeclarationSet constructorsSet = node.getConstructorsSet();
        NodeList constructorVarDecs = node.getConstructorVarSet();
        NodeList constructorsVarDecs = node.getConstructorsVarSet();

        if ((constructorsSet == null) || (constructorVarDecs == null) ||
                (constructorsVarDecs == null)) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        // Lista de declaraes de construtor
        // Aqui no  necessrio fazer nada, pois no h o que substituir.
        // Lista de declaraes do tipo varivel:JConstructorDeclaration
        if (!node.isExecutable()) {
            int k = 0;

            while (k < constructorVarDecs.size()) {

                INode tempNode = constructorVarDecs.elementAt(k);
                if (tempNode.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode);

                    if ((tempValue == null) ||
                            !(tempValue instanceof JConstructorDeclaration)) {
                        throw new IllegalMatchingException(tempNode);
                    }
                    constructorsSet.add(tempValue);
                    constructorVarDecs.removeElementAt(k);
                } else {
                    // Essa exceo deve ser lanada porque em constructorVarDecs no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
            }

            // Lista de declaraes do tipo varivel:JConstructorDeclarationSet
            int i = 0;
            while (i < constructorsVarDecs.size()) {
                INode tempNode2 = constructorsVarDecs.elementAt(i);
                if (tempNode2.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode2);

                    // Se tempNode2  do tipo JConstructorDeclarationSet, ela est casada
                    // a uma lista de ns que contm apenas declaraes de metodo
                    // (o que  verificado no prximo lao, abaixo).
                    if (!(tempValue instanceof JConstructorDeclarationSet)) {
                        throw new IllegalMatchingException(tempNode2);
                    }

                    JConstructorDeclarationSet constructorList =
                            (JConstructorDeclarationSet) tempValue;

                    int count3 = 0;
                    int max3 = constructorList.size();

                    // Cada elemento na lista mapeada a tempNode2 deve ser adicionado
                    //  lista de construtores em node.
                    while (count3 < max3) {
                        INode constructorDec =
                                constructorList.elementAt(count3);
                        if (constructorDec == null ||
                                !(constructorDec instanceof JConstructorDeclaration)) {
                            throw new IllegalMatchingException(tempNode2);
                        }

                        constructorsSet.add(constructorDec);
                        count3++;
                    }
                } else {
                    // Essa exceo deve ser lanada porque no em constructorsVarDecs no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
                constructorsVarDecs.removeElementAt(i);
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JExecutableDeclaration</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JExecutableDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        INode resNode = node.getResultNode();

        if (node.size() <= 0) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        // Efetua substituio no n-resultado da declarao executvel, caso
        // este seja uma varivel.
        if (resNode != null && resNode.isVariable()) {
            INode tempRes = this.getValueOf(resNode);

            if (tempRes != null) {
                node.setResultNode(tempRes);
            }
        }

        // Substituio na lista de expresses da declarao executvel, caso
        // alguma delas seja apenas uma varivel.
        int i = 0;
        int max = node.size();

        while (i < max) {
            INode varExp = node.getExpressionAt(i);

            if (varExp.isVariable()) {
                INode tempExp = this.getValueOf(varExp);

                if (tempExp != null && tempExp instanceof JExpression) {
                    node.setExpressionAt((JExpression) tempExp, i);
                } else if (tempExp != null) {
                    throw new IllegalMatchingException(resNode);
                }
            }

            i++;
        }

        // Restaura o valor original da varivel resultado
        node.unmapOldResultNode(this.results);

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JExpressionList</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JExpressionList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        int j = 0, max = node.size();

        while (j < max) {
            JExpression varArg = node.getExpressionAt(j);
            if (varArg.isVariable()) {
                node.setExpressionAt(this.getMappedExpression(varArg), j);
            }
            j++;
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JExpresssion</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JExpression node, Object data)
            throws IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        return node;
    }

    /**
     Mtodos visit para objetos do tipo <code>JFieldAccess</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        if (!node.isExecutable()) {
            INode referencedObject = node.getReferencedObject();
            JIdentifier fieldName = node.getFieldName();

            if (fieldName == null)
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

            if (referencedObject != null && referencedObject.isVariable()) {
                INode tempIP = this.getValueOf(referencedObject);
                if (tempIP != null) {
                    node.setReferencedObject(tempIP);
                }
            }

            if (fieldName.isVariable()) {
                node.setFieldName(this.getMappedIdentifier(fieldName));
            }
        }
        return node;
    }


    /**
     Mtodos visit para objetos do tipo <code>JFieldDeclaration</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!node.isExecutable() && !node.isVariable()) {
            JModifierList mods = node.getModifiers();
            JType type = node.getType();
            JVariableDeclaratorSet varDecs = node.getVariables();

            if (mods == null || type == null || varDecs == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Lista de Modificadores da declarao de atributo
            if (mods.isVariable()) {
                node.setModifiers(this.getMappedModifierList(mods,
                        JModifierList.FIELD_DECLARATION));
            }

            // Tipo dos atributos que esto sendo declarados
            if (type.isVariable()) {
                node.setType(this.getMappedType(type));
            }

            // Lista de nomes dos atributos
            int i = 0;
            int max = varDecs.size();

            while (i < max) {
                INode tempNode = varDecs.elementAt(i);
                if (tempNode.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode);

                    try {
                        varDecs.setElementAt(
                                JVariableDeclarator.unwrapVariableDeclarator(tempValue),
                                i);
                    } catch (IllegalArgumentException e) {
                        // S lana a exceo se tempValue for um valor invlido.
                        // "null" no  consideraod invlido porque isso significa,
                        // simplesmente, que a varivel no est mapeada a nenhum
                        // valor, o que  uma situao esperada.
                        if (tempValue != null) {
                            throw new IllegalMatchingException(e.getMessage());
                        }
                    }
                }
                i++;
            }
        }

        return node;
    }


    /**
     Mtodos visit para objetos do tipo <code>JFieldDeclarationSet</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JFieldDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // atributo que correspondem a variveis so tratadas em
        // visit(BodyFieldDeclarations).

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>BodyFieldDeclarations</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyFieldDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        if (!node.isExecutable() && !node.isVariable()) {
            JFieldDeclarationSet fieldDecs = node.getFieldsSet();
            NodeList attributeDecs = node.getFieldVarSet();
            NodeList attributesDecs = node.getFieldsVarSet();

            if (fieldDecs == null || attributeDecs == null || attributesDecs == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            // Lista de declaraes de atributo
            // Aqui no  necessrio fazer nada, pois no h o que substituir .
            // Lista de declaraes do tipo varivel:JFieldDeclaration
            int k = 0;
            while (k < attributeDecs.size()) {
                INode tempNode = attributeDecs.elementAt(k);
                if (tempNode.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode);
                    if (tempValue != null &&
                            !(tempValue instanceof JVariableDeclaration)) {
                        throw new IllegalMatchingException(tempNode);
                    } else if (tempValue != null) {
                        fieldDecs.add(tempValue);
                        attributeDecs.removeElementAt(k);
                    } else {
                        k++;
                    }
                } else {
                    // Essa exceo deve ser lanada porque no em attributeDecs no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
            }
            // Lista de declaraes do tipo varivel:JFieldDeclarationSet
            int i = 0;
            while (i < attributesDecs.size()) {
                INode tempNode2 = attributesDecs.elementAt(i);
                if (tempNode2.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode2);
                    // Se tempNode2  do tipo JFieldDeclarationSet, ela est casada
                    // a uma lista de ns que contm apenas declaraes de atributo
                    // (o que  verificado no prximo lao, abaixo).
                    if (tempValue != null &&
                            !(tempValue instanceof JFieldDeclarationSet)) {
                        throw new IllegalMatchingException(tempNode2);
                    } else if (tempValue != null) {

                        JFieldDeclarationSet fieldList = (JFieldDeclarationSet) tempValue;
                        int count3 = 0;
                        int max3 = fieldList.size();

                        // Cada elemento na lista mapeada a tempNode2 deve ser adicionado
                        //  lista de atributos em node.
                        while (count3 < max3) {
                            INode fieldDec = fieldList.elementAt(count3);
                            if (fieldDec == null ||
                                    !(fieldDec instanceof JVariableDeclaration)) {
                                throw new IllegalMatchingException(tempNode2);
                            }
                            fieldDecs.add(fieldDec);
                            count3++;
                        }
                    }
                } else {
                    // Essa exceo deve ser lanada porque no em attributesDecs no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
                attributesDecs.removeElementAt(i);

            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>classBodyInitializers</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(ClassBodyInitializers node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        if (!node.isExecutable() && !node.isVariable()) {
            JInitializerSet inits = node.getInitializersSet();
            NodeList initVarSet = node.getInitializerVarSet();
            NodeList initVarsSet = node.getInitializersVarSet();

            if (inits == null || initVarSet == null || initVarsSet == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            // Lista de inicializadores
            // Aqui no  necessrio fazer nada, pois no h o que substituir .
            // Lista de declaraes do tipo  #var:Initializer
            int k = 0;
            while (k < initVarSet.size()) {
                INode tempNode = (INode) initVarSet.elementAt(k);
                if (tempNode.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode);
                    if (tempValue != null &&
                            !(tempValue instanceof JInitializer)) {
                        throw new IllegalMatchingException(tempNode);
                    } else if (tempValue != null) {
                        inits.add(tempValue);
                        initVarSet.removeElementAt(k);
                    } else {
                        k++;
                    }
                } else {
                    // Essa exceo deve ser lanada porque em initVarSet no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
            }
            // Lista de declaraes do tipo #var:JInitializerSet
            int i = 0;
            while (i < initVarsSet.size()) {
                INode tempNode2 = initVarsSet.elementAt(i);
                if (tempNode2.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode2);
                    // Se tempNode2  do tipo JInitializerSet, ela est casada
                    // a uma lista de ns que contm apenas declaraes de atributo
                    // (o que  verificado no prximo lao, abaixo).
                    if (tempValue != null &&
                            !(tempValue instanceof JInitializerSet)) {
                        throw new IllegalMatchingException(tempNode2);
                    } else if (tempValue != null) {
                        JInitializerSet initList = (JInitializerSet) tempValue;
                        int count3 = 0;
                        int max3 = initList.size();
                        // Cada elemento na lista mapeada a tempNode2 deve ser
                        // adicionado a lista de inicializadores em node.
                        while (count3 < max3) {
                            INode init = initList.elementAt(count3);
                            if (init == null ||
                                    !(init instanceof JInitializer)) {
                                throw new IllegalMatchingException(tempNode2);
                            }
                            inits.add(init);
                            count3++;
                        }
                    }
                } else {
                    // Essa exceo deve ser lanada porque no em initVarsSet no
                    //  (ou no deveria ser) possvel que hajam ns que no sejam
                    // variveis.
                    throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
                }
                initVarsSet.removeElementAt(i);
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JFormalParameter</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JFormalParameter node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("JFormalParameter nulo");
        }

        if (node.isASourceNode()) {

            JModifierList mod = node.getModifiers();
            JType t = node.getType();
            JIdentifier id = node.getIdentifier();


            if ( mod != null && mod.isVariable()) {
                // Modificador do parmetro
                node.setModifiers(
                        this.getMappedModifierList(mod,
                                JModifierList.PARAMETER_DECLARATION));
            }

            if ( t != null && t.isVariable()) {
                // Tipo do  parmetro
                node.setType(this.getMappedType(t));
            }

            if (id != null && id.isVariable()) {
                // Identificador do parmetro
                node.setIdentifier(this.getMappedIdentifier(id));
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JForStatement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JForStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());
        if (nodeHasNoBody) {

            // substitui a inicializao, caso seja varivel
            INode init = node.getForInit();
            if (init != null && init.isVariable()) {
                INode tmpInit = this.getValueOf(init);
                if (tmpInit != null) {
                    node.setForInit(tmpInit);
                }
            }

            // substitui a expressao, caso seja variavel
            JExpression exp = node.getExpression();
            if (exp != null && exp.isVariable()) {
                JExpression tmpExpr = this.getMappedExpression(exp);
                if (tmpExpr != null) {
                    node.setExpression(tmpExpr);
                }
            }

            // substitui a atualizao, caso seja varivel
            INode update = node.getForInit();
            if (update != null && update.isVariable()) {
                INode tmpUpdate = this.getValueOf(update);
                if (tmpUpdate != null) {
                    node.setForInit(tmpUpdate);
                }
            }

        }
        return node;

    }

    /**
     Mtodos visit para objetos do tipo <code>JIdentifier</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JIdentifier node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        // Teoricamente, o mtodo visit para objetos do tipo JIdentifier
        // nunca eh chamado, no caso de substituio.
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JIfStatetement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JIfStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());
        if (nodeHasNoBody) {
            // substitui a expresso, se ela for variavel
            JExpression exp = node.getExpression();
            if (exp != null && exp.isVariable()) {
                JExpression tmpExp = this.getMappedExpression(exp);
                if (tmpExp != null) {
                    node.setExpression(tmpExp);
                }
            }

            // substitui o if statement, caso seja variavel
            INode ifStmt = node.getIfStatement();
            if (ifStmt != null && ifStmt.isVariable()) {
                INode tmpIf = this.getValueOf(ifStmt);
                if (tmpIf != null) {
                    node.setIfStatement(tmpIf);
                }
            }

            // substitui o else statement, caso seja variavel
            INode elseStmt = node.getElseStatement();
            if (elseStmt != null && elseStmt.isVariable()) {
                INode tmpElse = this.getValueOf(elseStmt);
                if (tmpElse != null) {
                    node.setElseStatement(tmpElse);
                }
            }

        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JImportDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JImportDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Declaraco de import invalida");
        }

        if (!node.isExecutable() && !node.isVariable()) {
            JName pckg = node.getPackageName();
            if (pckg.isVariable()) {
                node.setPackageName(this.resolveName(pckg));
            }
            JIdentifier cName = node.getClassName();
            if (cName != null && cName.isVariable()) {
                node.setClassName(this.getMappedIdentifier(cName));
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JImportDeclarationSet</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JImportDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // importao que correspondem a variveis so tratadas em
        // visit(JImportDeclarationSet).

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JInitializer</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JInitializer node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());

        if (nodeHasNoBody) {
            // substitui o bloco, caso ele seja variavel
            JBlock block = node.getBlock();
            if (block != null && block.isVariable()) {
                INode tmpBlock = this.getValueOf(block);
                if (tmpBlock instanceof JBlock) {
                    node.setBlock((JBlock) tmpBlock);
                }
            }
        }
        return node;
    }

    /**
     Mtodos visit para objetos do tipo <code>JInitializerSet</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JInitializerSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // atributo que correspondem a variveis so tratadas em
        // visit(ClassBodyInitializers).

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JInterfaceDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JInterfaceDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Declaracao de interface nula");
        }

        if (!node.isExecutable()) {

            JModifierList mods = null;
            mods = node.getModifiers();
            JNameList interfaces = node.getInterfaces();
            JIdentifier className = node.getName();
            TypeBody body = node.getBody();

            if (mods == null || className == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Comentrios da interface
            if (node.getComment() == null) {
                String commentVar = Util.generateTypeCommentVariable(node.getName());
                JComment comment = new JComment(commentVar);
                Object tmpComment = this.getValueOf(comment);
                if (tmpComment instanceof JComment) {
                    node.setComment((JComment) tmpComment);
                }
            }

            // Lista de Modificadores da classe
            if (mods.isVariable()) {
                node.setModifiers(
                        this.getMappedModifierList(mods,
                                JModifierList.INTERFACE_DECLARATION));
            }
            // Nome da classe
            if (className.isVariable()) {
                node.setName(this.getMappedIdentifier(className));
            }

            // Lista de interfaces que a classe implementa, se houver
            if (interfaces != null) {
                node.setInterfaces(this.resolveNameList(interfaces));
            }

            if (body != null && body.isVariable()) {
                node.setBody((TypeBody) this.getValueOf(body));
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>ImportDeclarations</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(ImportDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Conjunto de declaracoes de importacao nulo");
        }

        JImportDeclarationSet importsSet = node.getImportsSet();
        NodeList importVarDecs = node.getImportVarSet();
        NodeList importsVarDecs = node.getImportsVarSet();

        if ((importsSet == null) || (importVarDecs == null) ||
                (importsVarDecs == null)) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        // Lista de declaraes de importaes
        // Aqui no  necessrio fazer nada, pois no h o que substituir
        // Lista de declaraes do tipo varivel:JImportDeclaration
        int k = 0;

        while (k < importVarDecs.size()) {

            INode tempNode = importVarDecs.elementAt(k);
            if (tempNode.isVariable()) {
                INode tempValue = this.getValueOf(tempNode);

                if ((tempValue == null) ||
                        !(tempValue instanceof JImportDeclaration)) {
                    throw new IllegalMatchingException(tempNode);
                }
                importsSet.add(tempValue);
                importVarDecs.removeElementAt(k);
            } else {
                // Essa exceo deve ser lanada porque em importVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
        }

        // Lista de declaraes do tipo varivel:JImportDeclarationSet
        int i = 0;

        while (i < importsVarDecs.size()) {
            // tempNode2  uma varivel do tipo JImportDeclarationSet
            INode tempNode2 = importsVarDecs.elementAt(i);

            if (tempNode2.isVariable()) {
                INode tempValue = this.getValueOf(tempNode2);

                // Se tempNode2  do tipo JImportDeclarationSet, ela est casada
                // a uma lista de ns que contm apenas declaraes de metodo
                // (o que  verificado no prximo lao, abaixo).
                if (tempValue == null ||
                        !(tempValue instanceof JImportDeclarationSet)) {
                    throw new IllegalMatchingException(tempNode2);
                }

                JImportDeclarationSet importList =
                        (JImportDeclarationSet) tempValue;

                int count3 = 0;
                int max3 = importList.size();
                

                // Cada elemento na lista mapeada a tempNode2 deve ser adicionado
                //  lista de atributos em node.
                while (count3 < max3) {
                    JImportDeclaration importDec =
                            (JImportDeclaration) importList.
                            elementAt(count3);
                    importsSet.add(importDec);
                    count3++;
                }
            } else {
                // Essa exceo deve ser lanada porque no em importsVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            importsVarDecs.removeElementAt(i);
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>InstanceOfExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(InstanceOfExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        INode operand = node.getOperand();
        JType type = node.getType();

        if (operand == null || type == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        // Operando
        if (operand.isVariable()) {
            INode oper = this.getValueOf(operand);

            if (oper != null) {
                node.setOperand(oper);
            }
        }
        // Tipo com o qual o operando est sendo comparado
        if (type.isVariable()) {
            node.setType(this.getMappedType(type));
        }

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>IterativeDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(IterativeDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        this.replaceIterativeDeclaration(node);

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JLiteral</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JLiteral node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        // Assim como ocorre com objetos do tipo JIdentifier, teoricamente o
        // mtodo visit nunca  invocado para objetos do tipo JLiteral.
        if (node.isVariable()) {
            JLiteral l = (JLiteral) this.getValueOf(node);
            if (l != null) {
                node.setLiteralValue(l.getValue());
            }
            // No devia estar aqui!
            else {
                node.setLiteralValue("true");
                node.setLiteralType(JLiteral.BOOLEAN_LITERAL);
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JLiteral</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JLocalVariableDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!node.isExecutable() && !node.isVariable()) {
            JModifierList mods = node.getModifiers();
            JType type = node.getType();
            JVariableDeclaratorSet varDecs = node.getVariables();

            if (mods == null || type == null || varDecs == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Lista de Modificadores da declarao de atributo
            if (mods.isVariable()) {
                node.setModifiers(this.getMappedModifierList(mods,
                        JModifierList.FIELD_DECLARATION));
            }

            // Tipo dos atributos que esto sendo declarados
            if (type.isVariable()) {
                node.setType(this.getMappedType(type));
            }

            // Lista de nomes dos atributos
            int i = 0;
            int max = varDecs.size();

            while (i < max) {
                INode tempNode = varDecs.elementAt(i);
                if (tempNode.isVariable()) {
                    INode tempValue = this.getValueOf(tempNode);

                    try {
                        varDecs.setElementAt(
                                JVariableDeclarator.unwrapVariableDeclarator(tempValue),
                                i);
                    } catch (IllegalArgumentException e) {
                        // S lana a exceo se tempValue for um valor invlido.
                        // "null" no  consideraod invlido porque isso significa,
                        // simplesmente, que a varivel no est mapeada a nenhum
                        // valor, o que  uma situao esperada.
                        if (tempValue != null) {
                            throw new IllegalMatchingException(e.getMessage());
                        }
                    }
                }
                i++;
            }
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>MapTable</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(MapTable node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        // Teoricamente, o mtodo visit para objetos do tipo MapTable
        // nunca eh chamado, no caso de substituio.
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JMethodDeclaration</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("JMethodDeclaration nulo");
        }

        if (node.isASourceNode()) {

            JModifierList mods = node.getModifiers();
            JType resType = node.getResultType();
            JName name = node.getName();
            JParameterList params = node.getParameters();
            JNameList excs = node.getExceptions();
            JBlock mBody = node.getBody();


            if (mods == null || resType == null || name == null) {
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

            // Comentrios do mtodo
            if (node.getComment() == null) {
                String commentVar = Util.generateCommentVariable(node.getName());
                JComment comment = new JComment(commentVar);
                Object tmpComment = this.getValueOf(comment);
                if (tmpComment instanceof JComment) {
                    node.setComment((JComment) tmpComment);
                }
            }

            // Lista de Modificadores da declarao de mtodo
            if (mods.isVariable()) {
                node.setModifiers(
                        this.getMappedModifierList(mods,
                                JModifierList.METHOD_DECLARATION));
            }
            // Tipo de retorno do mtodo
            if (resType.isVariable()) {
                node.setResultType(this.getMappedType(resType));
            }

            // nome do mtodo
            if (name.isVariable()) {
                node.setName(this.resolveName(name));
            }

            // lista de parametros do mtodo
            if (params.isVariable()) {

                INode tempParams = this.getMappedParameterList(params);
                if (tempParams != null && tempParams instanceof JParameterList) {
                    node.setParameters((JParameterList) tempParams);
                } else {
                    throw new IllegalMatchingException(params);
                }
            }
            // Lista excees que o mtodo lana
            if (excs != null && (excs.isVariable() || excs.isOrDeclaration())) {
                node.setExceptions(this.resolveNameList(excs));
            }

            // Corpo do mtodo
            if (mBody != null && mBody.isVariable()) {
                JBlock body = (JBlock) this.getValueOf(mBody);
//                body.setIndentLevel(mBody.getIndentLevel());
                node.setBody(body);
            }

        }
        return node;

    }

    /**
     * Mtodos visit para objetos do tipo <code>JMethodDeclarationSet</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Lista de declaracoes de metodo nula");
        }

        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // atributo que correspondem a variveis so tratadas em
        // visit(BodyMethodDeclarations).

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>BodyMethodDeclarations</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyMethodDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Conjunto de declaracoes de metodo nulo");
        }

        JMethodDeclarationSet methodsSet = node.getMethodsSet();
        NodeList methodVarDecs = node.getMethodVarSet();
        NodeList methodsVarDecs = node.getMethodsVarSet();

        if ((methodsSet == null) || (methodVarDecs == null) ||
                (methodsVarDecs == null)) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        // Lista de declaraes de mtodo
        // Aqui no  necessrio fazer nada, pois no h o que substituir .

        // Lista de declaraes do tipo varivel:JMethodDeclaration
        int k = 0;

        while (k < methodVarDecs.size()) {

            INode tempNode = methodVarDecs.elementAt(k);
            if (tempNode.isVariable()) {
                INode tempValue = this.getValueOf(tempNode);

                // comportamento padro: se no houver um valor mapeado a
                // uma varivel do tipo #MD:JMethodDeclaration uma exceo
                // deve ser levantada.
                if ((tempValue == null) ||
                        !(tempValue instanceof JMethodDeclaration)) {
                    throw new IllegalMatchingException(tempNode);
                }
                methodsSet.add(tempValue);
                methodVarDecs.removeElementAt(k);
            } else {
                // Essa exceo deve ser lanada porque em methodVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }

        }

        // Lista de declaraes do tipo varivel:JMethodDeclarationSet
        int i = 0;

        while (i < methodsVarDecs.size()) {
            INode tempNode2 = methodsVarDecs.elementAt(i);
            if (tempNode2.isVariable()) {
                // Deve ser sempre verdade
                INode tempValue = this.getValueOf(tempNode2);

                // Se tempNode2  do tipo JMethodDeclarationSet, ela est casada
                // a uma lista de ns que contm apenas declaraes de metodo
                // (o que  verificado no prximo lao, abaixo).
//                if (tempValue == null)
//                    QualitiLogger.debug(tempNode2 + "-> null");
                if (!(tempValue instanceof JMethodDeclarationSet)) {
                    throw new IllegalMatchingException(tempNode2);
                }

                JMethodDeclarationSet methodList =
                        (JMethodDeclarationSet) tempValue;

                int count3 = 0;
                int max3 = methodList.size();

                // Cada elemento na lista mapeada a tempNode2 deve ser adicionado
                //  lista de atributos em node.
                while (count3 < max3) {
                    INode methodDec = methodList.elementAt(count3);
                    if (methodDec == null ||
                            !(methodDec instanceof JMethodDeclaration)) {
                        throw new IllegalMatchingException(tempNode2);
                    }
                    methodsSet.add(methodDec);
                    count3++;
                }
            } else {
                // Essa exceo deve ser lanada porque no em methodsVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            methodsVarDecs.removeElementAt(i);
        }

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>BodyTypeDeclarations</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(BodyTypeDeclarations node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException("Conjunto de declaracoes de metodo nulo");
        }

        TypeDeclarationSet typesSet = node.getTypesSet();
        NodeList typeVarDecs = node.getTypeVarSet();
        NodeList typesVarDecs = node.getTypesVarSet();

        if ((typesSet == null) || (typeVarDecs == null) ||
                (typesVarDecs == null)) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }

        // Lista de declaraes do tipo varivel:TypeDeclaration
        int k = 0;
        while (k < typeVarDecs.size()) {

            INode tempNode = typeVarDecs.elementAt(k);
            if (tempNode.isVariable()) {
                INode tempValue = this.getValueOf(tempNode);

                // comportamento padro: se no houver um valor mapeado a
                // uma varivel do tipo #CD:ClassDeclaration uma exceo
                // deve ser levantada.
                if ((tempValue == null) ||
                        !(tempValue instanceof JTypeDeclaration)) {
                    throw new IllegalMatchingException(tempNode);
                }
                typesSet.add(tempValue);
                typeVarDecs.removeElementAt(k);
            } else {
                // Essa exceo deve ser lanada porque em typeVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
        }

        // Lista de declaraes do tipo varivel:TypeDeclarationSet
        int i = 0;
        while (i < typesVarDecs.size()) {
            INode tempNode2 = typesVarDecs.elementAt(i);
            if (tempNode2.isVariable()) {
                // Deve ser sempre verdade
                INode tempValue = this.getValueOf(tempNode2);

                // Se tempNode2  do tipo TypeDeclarationSet, ela est casada
                // a uma lista de ns que contm apenas declaraes de tipos
                // (o que  verificado no prximo lao, abaixo).
                if (tempValue == null ||
                        !(tempValue instanceof TypeDeclarationSet)) {
                    throw new IllegalMatchingException(tempNode2);
                }

                TypeDeclarationSet typeList =
                        (TypeDeclarationSet) tempValue;

                int count3 = 0;
                int max3 = typeList.size();

                // Cada elemento na lista mapeada a tempNode2 deve ser adicionado
                //  lista de tipos em node.
                while (count3 < max3) {
                    INode tmpNode = typeList.elementAt(count3);
                    if (tmpNode == null ||
                            !(tmpNode instanceof JTypeDeclaration)) {
                        throw new IllegalMatchingException(tempNode2);
                    }
                    typesSet.add(tmpNode);
                    count3++;
                }
            } else {
                // Essa exceo deve ser lanada porque no em typesVarDecs no
                //  (ou no deveria ser) possvel que hajam ns que no sejam
                // variveis.
                throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
            }
            typesVarDecs.removeElementAt(i);
        }

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JMethodInvocation</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JMethodInvocation node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        INode referencedObject = node.getReferencedObject();
        JName methodName = node.getMethodName();
        JExpressionList arguments = node.getArgumentList();

        if (methodName == null || arguments == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (referencedObject != null && referencedObject.isVariable()) {
            INode tempIP = this.getValueOf(referencedObject);

            if (tempIP != null) {
                node.setReferencedObject(tempIP);
            }
        }

        if (methodName.isVariable()) {
            JName mName = this.resolveName(methodName);
            node.setMethodName(mName);

        }

        if (arguments.isVariable()) {
            INode tempArgs = this.getValueOf(arguments);

            if (tempArgs != null && tempArgs instanceof JExpressionList) {
                node.setArgumentList((JExpressionList) tempArgs);
            } else if (tempArgs != null) {
                throw new IllegalMatchingException(arguments);
            }
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JModifierList</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JModifierList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        // Teoricamente, o mtodo visit para objetos do tipo
        // JModifierList nunca  chamado, no caso de substituio.
        // Na verdade,  sempre chamado, mas no h o que subsituir dentro dele.

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JName</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(JName node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        int j = 0, max = node.size();

        while (j < max) {
            JIdentifier varId = node.getIdentifierAt(j);
            if (varId.isVariable()) {
                node.setIdentifierAt(this.getMappedIdentifier(varId), j);
            }
            j++;
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JNodeString</code>.
     *
     * @param node  O n onde a substituio ocorrer.
     * @param data  Um objeto que pode conter informao adicional
     *              necessria ao visitor.
     * @return      Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *              parmetro <code>node</code> for <code>null</code>.
     * @exception cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException Levantada
     *              quando um n apresenta uma estrutura interna invlida.
     *
     */
    public Object visit(JNodeString node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        if (!node.isVariable()) {

            //recupera as variaveis jats presentes na string (fora de
            //expresses executveis)
            VectorVariables variables = JNodeString.getVariables(node.getNodeValue());

            //monta um mapeamento entre as variaveis e seus valores associados
            Map map = JNodeString.getMap(this.results, variables);
            Set keys = map.keySet();
            Iterator it = keys.iterator();

            //verifica se o statement em si  uma varivel, e se for,substitui
            //seu valor pelo mapeado
            while (it.hasNext()) {
                Object var = it.next();

                if (var instanceof String) {
                    String statement = (String) var;
                    if (statement.endsWith("Statement")) { //nem sei se isso existe!
                        Object value = map.get(statement);
                        if (value != null) {
                            String strValue = value.toString();
                            strValue = strValue.trim();
                            if (strValue.endsWith(";")) {
                                map.put(var, strValue);
                            }
                        }
                    }
                }
            }

            String oldValue = node.getNodeValue();

            //pega as expressoes executaveis presentes na string
            ExpressionsMap expressions = JNodeString.getNodeExpressions(oldValue);
            node.setExpressions(expressions);

            //substitui variaveis jats pelos valores correspondentes
            String resultString = VariableReplacer.substituirTagsString(map, oldValue);

            ExpressionsMap replacedExpressions = JNodeString.getNodeExpressions(resultString);

            // s para eu guardar as expresses executveis em texto
            node.setReplacedExpressions(replacedExpressions);

            //seta o novo valor da string
            node.setNodeValue(resultString);
        }

        return node;
    }


    /**
     Mtodos visit para objetos do tipo <code>JNameList</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(JNameList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        int j = 0, max = node.size();

        while (j < max) {
            JName varName = (JName) node.elementAt(j);
            if (varName.isVariable()) {
                node.setElementAt(this.resolveName(varName), j);
            }
            j++;
        }

        return node;
    }

    /**
     Mtodos visit para objetos do tipo <code>OtherDeclarationsSet</code>.
     @param node O n onde a substituio ocorrer.
     @param data Um objeto que pode conter informao adicional necessria ao
     visitor.
     @return Um objeto qualquer.
     @exception java.lang.IllegalArgumentException Levantada quando o
     parmetro <code>node</code> for <code>null</code>.
     @exception java.lang.IllegalArgumentException Levantada quando um n
     apresenta uma estrutura interna invlida.
     */
    public Object visit(OtherDeclarationsSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        // Teoricamente, aqui no  necessrio fazer nada, uma vez que
        // declaraes iterativas no podem "ser" variveis e os resultados
        // produzidos pelo processamento delas tambm deve consistir de
        // construes Java.

        // O mesmo se aplica para as declaraes condicionais. Se uma varivel
        // aparece no corpo de uma classe, essa varivel deve ser tipada e,
        // sendo tipada, ela no vem parar aqui.

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JPackageDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JPackageDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node != null && !node.isVariable()) {

            JName packName = node.getName();
            if (!node.isExecutable()) {
                node.setName(this.resolveName(packName));
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JParameterList</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JParameterList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {

        if (node == null) {
            throw new IllegalArgumentException();
        }
        int j = 0, max = node.size();
        while (j < max) {
            INode param = node.elementAt(j);
            if (param.isVariable()) {
                node.setElementAt(this.getValueOf(param), j);
            }
            j++;
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JStatementExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JStatementExpression node, Object data) throws IllegalArgumentException, cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException {
        if (node == null) {
            throw new IllegalArgumentException("node is null");
        }
        if (node.isASourceNode()) {
            JExpression tmpExpression = node.getExpression();
            if (tmpExpression.isVariable()) {
                node.setExpression(this.getMappedExpression(tmpExpression));
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JSwitchLabel</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchLabel node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());
        if (nodeHasNoBody) {
            // substitui a expresso, se ela for variavel
            JExpression exp = node.getExpression();
            if (exp != null && exp.isVariable()) {
                JExpression tmpExpr = this.getMappedExpression(exp);
                if (tmpExpr != null) {
                    node.setExpression(tmpExpr);
                }
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JSwitchPair</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchPair node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());

        if (nodeHasNoBody) {
            JSwitchLabel label = node.getLabel();

            // substitui o label, caso seja variavel
            if (label != null && label.isVariable()) {
                JSwitchLabel slabel = (JSwitchLabel) this.getValueOf(label);
                if (slabel != null) {
                    node.setLabel(slabel);
                }
            }
            // substitui os statements que forem variaveis
            NodeList statements = node.getStatements();
            if (statements != null) {
                int size = statements.size();
                for (int i = 0; i < size; i++) {
                    INode stmt = statements.elementAt(i);
                    if (stmt != null && stmt.isVariable()) {
                        INode tmpStmt = this.getValueOf(stmt);
                        if (tmpStmt != null) {
                            statements.setElementAt(tmpStmt, i);
                        }
                    }
                }
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JSwitchStatement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JSwitchStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());

        if (nodeHasNoBody) {
            JExpression exp = node.getExpression();

            // substitui a expressao, caso seja variavel
            if (exp != null && exp.isVariable()) {
                JExpression tmpExpr = this.getMappedExpression(exp);
                if (tmpExpr != null) {
                    node.setExpression(tmpExpr);
                }
            }
            // substitui os cases que forem variaveis
            NodeList cases = node.getCases();
            if (cases != null) {
                int size = cases.size();
                for (int i = 0; i < size; i++) {
                    JSwitchPair currCase = (JSwitchPair) cases.elementAt(i);
                    if (currCase != null && currCase.isVariable()) {
                        JSwitchPair tmpCase = (JSwitchPair) this.getValueOf(currCase);
                        if (tmpCase != null) {
                            cases.setElementAt(tmpCase, i);
                        }
                    }
                }
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JPreconditionDeclaration</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JPreconditionDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("ReplacementVisitor: JPreconditionDeclaration ");
        }
        JLiteral lit = node.getValue();
        if (lit != null && lit.isVariable()) {
            JLiteral tmpValue = (JLiteral) this.getValueOf(lit);
            if (tmpValue == null) {
                tmpValue = new JLiteral(JLiteral.BOOLEAN_LITERAL, "true");
            }
            node.setValue(tmpValue);
            // esta linha  um teste
            node.setExecutableDeclaration(tmpValue);
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JResultStatement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JResultStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());
        if (nodeHasNoBody) {
            // substitui a expressao, caso seja variavel
            JExpression exp = node.getExpression();
            if (exp != null && exp.isVariable()) {
                JExpression tmpExpr = this.getMappedExpression(exp);
                if (tmpExpr != null) {
                    node.setExpression(tmpExpr);
                }
            }
        }
        return node;

    }

    /**
     * Mtodos visit para objetos do tipo <code>JTernaryExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JTernaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        JExpression first = node.getFirstOperand();
        JExpression second = node.getSecondOperand();
        JExpression third = node.getThirdOperand();

        if (first == null || second == null || third == null)
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (first.isVariable()) {
            node.setFirstOperand(new JUnaryExpression(
                    JUnaryExpression.PARENTHESIZEDEXPRESSION,
                    this.getMappedExpression(first)));
        }

        if (second.isVariable()) {
            node.setSecondOperand(new JUnaryExpression(
                    JUnaryExpression.PARENTHESIZEDEXPRESSION,
                    this.getMappedExpression(second)));
        }

        if (third.isVariable()) {
            node.setThirdOperand(new JUnaryExpression(
                    JUnaryExpression.PARENTHESIZEDEXPRESSION,
                    this.getMappedExpression(third)));
        }

        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JTryStatement</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JTryStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        boolean nodeHasNoBody = !(node.isVariable() || node.isExecutable());

        if (nodeHasNoBody) {

            // substitui o bloco try, caso ele seja variavel
            JBlock tblock = node.getTryBlock();
            if (tblock != null && tblock.isVariable()) {
                INode tmpTryBlock = this.getValueOf(tblock);
                if (tmpTryBlock instanceof JBlock) {
                    node.setTryBlock((JBlock) tmpTryBlock);
                }
            }

            // substitui as clusulas catch que forem variveis
            NodeList catchlist = node.getCatchlist();
            if (catchlist != null) {
                int size = catchlist.size();
                for (int i = 0; i < size; i++) {
                    INode currCatch = catchlist.elementAt(i);
                    if (currCatch != null && currCatch.isVariable()) {
                        INode tmpCatch = this.getValueOf(currCatch);
                        if (tmpCatch != null) {
                            catchlist.setElementAt(tmpCatch, i);
                        }
                    }
                }
            }

            // substitui o bloco finally, caso ele seja variavel
            JBlock fblock = node.getTryBlock();
            if (fblock != null && fblock.isVariable()) {
                INode tmpFinallyBlock = this.getValueOf(fblock);
                if (tmpFinallyBlock instanceof JBlock) {
                    node.setTryBlock((JBlock) tmpFinallyBlock);
                }
            }
        }
        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JType</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *         parmetro <code>node</code> for <code>null</code>.
     * @exception cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException
     *         Levantada quando um n apresenta uma estrutura interna invlida.
     */
    public Object visit(JType node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        JName typeName = node.getTypeName();

        if (typeName == null && node.isASourceNode())
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);

        if (node.isASourceNode() && typeName != null && typeName.isVariable()) {
            node.setName(this.resolveName(typeName));
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>TypeDeclarationSet</code>.
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     *        visitor.
     * @return Um objeto qualquer.
     * @exception java.lang.IllegalArgumentException Levantada quando o
     *            parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     *            apresenta uma estrutura interna invlida.
     */
    public Object visit(TypeDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException("Lista de declaracoes de metodo nula");
        }

        // Teoricamente no precisa fazer nada, uma vez que declaraes de
        // atributo que correspondem a variveis so tratadas em
        // visit(BodyMethodDeclarations).

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JUnaryExpression</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JUnaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }
        JExpression operand = node.getOperand();
        if (operand == null) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        }
        if (operand.isVariable()) {
            node.setOperand(new JUnaryExpression(
                    JUnaryExpression.PARENTHESIZEDEXPRESSION,
                    this.getMappedExpression(operand)));
        }

        return node;
    }

    /**
     * Mtodos visit para objetos do tipo <code>JVariableDeclarator</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JVariableDeclarator node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException {
        if (node == null) {
            throw new IllegalArgumentException();
        }

        // Um JVariableDeclarator consiste de um identificador e uma inicializa-
        // o que  opcional. Por exemplo, em uma expresso do tipo
        // "private int a = 2;", o JVariableDeclarator corresponde ao "a = 2".

        JIdentifier id = node.getIdentifier();
        INode init = node.getInitializer();
        boolean nodeHasBody = !(node.isVariable() || node.isExecutable());

        if (id == null && nodeHasBody) {
            throw new cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException(node);
        } else if (nodeHasBody && id.isVariable()) {
            // Identificador da varivel
            node.setIdentifier(this.getMappedIdentifier(id));
        }

        // Inicializao da varivel
        if (init != null && init.isVariable()) {
            INode tempNode = this.getValueOf(init);

            if (tempNode != null && (!(tempNode instanceof JExpression) &&
                    !(tempNode instanceof JArrayInitializer))) {
                throw new IllegalMatchingException(init);
            } else if (tempNode != null) {
                node.setInitializer(tempNode);
            }
        }
        return node;
    }


    /**
     * Mtodos visit para objetos do tipo <code>JVariableDeclarator</code>.
     *
     * @param node O n onde a substituio ocorrer.
     * @param data Um objeto que pode conter informao adicional necessria ao
     * visitor.
     *
     * @return Um objeto qualquer.
     *
     * @exception java.lang.IllegalArgumentException Levantada quando o
     * parmetro <code>node</code> for <code>null</code>.
     * @exception java.lang.IllegalArgumentException Levantada quando um n
     * apresenta uma estrutura interna invlida.
     */
    public Object visit(JWhileStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        if (node == null) {
            throw new IllegalArgumentException();
        }

        boolean nodeHasBody = !(node.isVariable() || node.isExecutable());
        JExpression exp = node.getExpression();
        INode stmt = node.getStatement();

        if (nodeHasBody) {
            // substitui a expresso, se ela for variavel
            if (exp != null && exp.isVariable()) {
                JExpression tmpExpr = this.getMappedExpression(exp);
                if (tmpExpr != null) {
                    node.setExpression(tmpExpr);
                }
            }

            // substitui o statement, caso seja variavel
            if (stmt != null && stmt.isVariable()) {
                INode tmpStmt = this.getValueOf(stmt);
                if (tmpStmt != null) {
                    node.setStatement(tmpStmt);
                }
            }
        }
        return node;
    }

    public Object visit(WrapperNode node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        INode tmpNode = node.getNode();
        if (tmpNode.isVariable()) {
            node.setNode(this.getValueOf(tmpNode));

        }
        return node;
    }

	public Object visit(JStatementList node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
		//the replacement is done in block
		return node;
	}

	public Object visit(ExploreDeclaration node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {


		INode baseNode = node.getBaseNode();
		
		if (baseNode.isVariable()) {
			node.setBaseNode(this.getValueOf(baseNode));
		}
		
		return node;
	}

	public Object visit(Pattern node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
		//replaces nothing...
		return node;
	}

	public Object visit(PrefixedIdentifier node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
        JIdentifier id = node.getRootId();
        if (id != null && id.isVariable()) {
            node.setRootId(this.getMappedIdentifier(id));
        }		
        return node;
	}
    
    public Object visit(ContextDeclaration node, Object data)
			throws IllegalArgumentException, InconsistentNodeException,
			ExecutionException {
    	
    	JStatementList context = 
    		(JStatementList) this.getValueOf(node);
    	
    	JStatementList code = 
    		(JStatementList) this.getValueOf(node.getCodeOfInterest());
    	
    	if (code == null) {
			throw new InconsistentNodeException(
					"Variable not found in resultSet: "
							+ node.getCodeOfInterest().getVariableName());
		}
    	
    	node.setCodeOfInterest(code);
    	node.setContext(context);

    	node.getCodeOfInterest().setPosition(node.getPosition());
		SelectionProcessor.insertCode(node, node.getCodeOfInterest());
    	
    	return node;
    }
	
}
