/*
 * JaTS Project - Java Transformation System
 * http://www.cin.ufpe.br/~jats
 *  
 * Copyright  2005  Informatics Center - Federal University of Pernambuco
 * 
 * JaTS is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 */
package cin.jats.engine.visitors;

import cin.jats.engine.parser.nodes.*;

import cin.jats.engine.parser.nodes.exceptions.*;
import cin.jats.engine.JatsEngine;
import cin.jats.engine.util.TransformationState;
import cin.jats.engine.util.NodeList;

/**
 * This visitor applies micro-tranformations in
 * the program AST. As all other visitors, the process 
 * is done innermost -> outermost, so the more internal
 * nodes are transformed first.
 * 
 * @author Adeline Souza
 * @author Gustavo Santos
 * @version 09/06/2005
 */
public class TransformationVisitor extends AbstractVisitor {

    private JatsEngine engine = JatsEngine.getInstance();

    public Object visit(JArrayAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;

        if (node.isASourceNode()) {

            // Processamento do nome do mtodo a ser invocado.
            INode fName = node.getArrayName();
            if (fName != null) {
                try {
                    Object ret1 = engine.apply(fName, ts);
                    if (ret1 != null) {
                        node.setArrayName((INode) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            if (node.getIndexers() != null) {
                int max = node.getIndexers().size();

                while (i < max) {
                    INode currExpr = node.getIndexers().getExpressionAt(i);
                    if (currExpr != null) {
                        try {
                            Object ret = engine.apply(currExpr, ts);
                            if (ret instanceof JExpression) {
                                node.getIndexers().setExpressionAt((JExpression) ret, i);
                            }
                        } catch (NodesNotMatchedException e) {
                        }
                    }
                    i++;
                }
            }
        }


        return node;

    }

    public Object visit(JBinaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;

        if (node.isASourceNode()) {
            // Processamento do operando  esquerda do operador.
            INode lhso = node.getLeftHandOperand();
            if (lhso != null) {
                try {
                    Object ret = engine.apply(lhso, ts);
                    if (ret instanceof JExpression) {
                        node.setLeftHandOperand((JExpression) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do operando  direita do operador.
            INode rhso = node.getRightHandOperand();
            if (rhso != null) {
                try {
                    Object ret = engine.apply(rhso, ts);
                    if (ret instanceof JExpression) {
                        node.setRightHandOperand((JExpression) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JBlock node, Object data)
			throws InconsistentNodeException, IllegalArgumentException,
			ExecutionException {

        TransformationState ts = (TransformationState) data;

        if (node.isASourceNode()) {
			if (node.getStatements() != null) {
				try {
					INode tmp = null;
					tmp = engine.apply(node.getStatements(), ts);
					if (tmp != null) {
						node.setStatements((JStatementList) tmp);
					}
				} catch (NodesNotMatchedException e) {
					//it's normal. do nothing.
				}

			}
		}
        return node;
    }

    public Object visit(JClassDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;

        if (node.isASourceNode()) {

            // Processamento da lista de modificadores do tipo.
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    Object ret1 = engine.apply(mods, ts);
                    if (ret1 instanceof JModifierList) {
                        node.setModifiers((JModifierList) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do nome da classe
            INode tName = node.getName();
            if (tName != null) {
                try {
                    Object ret2 = engine.apply(tName, ts);
                    if (ret2 != null) {
                        node.setName(JIdentifier.unwrapIdentifier(ret2));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da superclasse
            INode sc = node.getSuperClass();
            if (sc != null) {
                try {
                    Object ret3 = engine.apply(sc, ts);
                    if (ret3 instanceof JName) {
                        node.setSuperClass((JName) ret3);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da lista de interfaces implementadas pela classe.
            INode ifs = node.getInterfaces();
            if (ifs != null) {
                try {
                    Object ret4 = engine.apply(ifs, ts);
                    if (ret4 instanceof JNameList) {
                        node.setInterfaces((JNameList) ret4);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do corpo da interface.
            INode body = node.getBody();
            if (body != null) {
                try {
                    Object ret5 = engine.apply(body, ts);
                    if (ret5 instanceof TypeBody) {
                        node.setBody((TypeBody) ret5);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JConstructorDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Lista de modificadores
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    INode res = engine.apply(mods, ts);
                    if (res instanceof JModifierList) {
                        node.setModifiers((JModifierList) res);
                    }
                } catch (NodesNotMatchedException ex) {
                }
            }

            // Identificador do construtor
            INode id = node.getName();
            if (id != null) {
                try {
                    INode res = engine.apply(id, ts);
                    if (res != null) {
                        node.setName(JIdentifier.unwrapIdentifier(res));
                    }
                } catch (NodesNotMatchedException ex) {
                }

            }
            // Lista de parmetros do construtor
            INode pars = node.getParameters();
            if (pars != null) {
                try {
                    INode res = engine.apply(pars, ts);
                    if (res instanceof JParameterList) {
                        node.setParameters((JParameterList) res);
                    }
                } catch (NodesNotMatchedException e) {
                }

            }
            // Lista de parmetros do construtor
            INode excs = node.getExceptions();
            if (excs != null) {
                try {
                    INode res = engine.apply(excs, ts);
                    if (res instanceof JNameList) {
                        node.setExceptions((JNameList) res);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
            // Corpo do construtor
            INode jbody = node.getBody();
            if (jbody != null) {
                try {
                    INode res = engine.apply(jbody, ts);
                    if (res instanceof JBlock) {
                        node.setBody((JBlock) res);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

        }
        return node;
    }

    public Object visit(JConstructorDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currCD = node.elementAt(i);
                if (currCD != null) {
                    try {
                        Object ret = engine.apply(currCD, ts);
                        if (ret instanceof JConstructorDeclaration) {
                            node.setElementAt((JConstructorDeclaration) ret, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JExpressionList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento de cada um dos argumentos individualmente.
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currArg = node.getExpressionAt(i);
                if (currArg != null) {
                    try {
                        Object ret = engine.apply(currArg, ts);
                        if (ret instanceof JExpression) {
                            node.setExpressionAt((JExpression) ret, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JFieldAccess node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            INode refObj = node.getReferencedObject();
            if (refObj != null) {
                try {
                    INode ret = engine.apply(refObj, ts);
                    if (ret != null) {
                        node.setReferencedObject(ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            JIdentifier fName = node.getFieldName();
            if (fName != null) {
                try {
                    Object ret = engine.apply(fName, ts);
                    if (ret instanceof JIdentifier) {
                        node.setFieldName((JIdentifier) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }

        return node;
    }

    public Object visit(JFieldDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // atualiza o result set
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    INode ret = engine.apply(mods, ts);
                    if (ret instanceof JModifierList) {
                        node.setModifiers((JModifierList) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do tipo do atributo e da sub-rvore correspondente.
            INode fType = node.getType();
            if (fType != null) {
                try {
                    INode tmpRes = engine.apply(fType, ts);
                    node.setType(JType.unwrapType(tmpRes));
                } catch (NodesNotMatchedException e) {
                }
            }

            // Lista de declaradores de varivel
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currDec = node.getVariableDeclaratorAt(i);
                if (currDec != null) {
                    try {
                        INode tmpRes = engine.apply(currDec, ts);
                        if (tmpRes != null) {
                            node.setVariableDeclaratorAt(
                                    JVariableDeclarator.unwrapVariableDeclarator(tmpRes),
                                    i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JFieldDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currFD = node.elementAt(i);
                if (currFD != null) {
                    try {
                        Object ret = engine.apply(currFD, ts);
                        if (ret instanceof JFieldDeclaration) {
                            node.setElementAt((JFieldDeclaration) ret, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JFormalParameter node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Lista de modificadores
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    INode res = engine.apply(mods, ts);
                    if (res instanceof JModifierList) {
                        node.setModifiers((JModifierList) res);
                    }
                } catch (NodesNotMatchedException ex) {
                }

            }
            // Tipo do parmetro
            INode mType = node.getType();
            if (mType != null) {
                try {
                    INode res = engine.apply(mType, ts);
                    if (res instanceof JType) {
                        node.setType((JType) res);
                    }
                } catch (NodesNotMatchedException ex) {
                }
            }

            // Identificador do parmetro
            INode id = node.getIdentifier();
            if (id != null) {
                try {
                    INode res = engine.apply(id, ts);
                    if (res != null) {
                        node.setIdentifier(JIdentifier.unwrapIdentifier(res));
                    }
                } catch (NodesNotMatchedException ex) {
                }
            }
        }
        return node;
    }

    public Object visit(JImportDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            INode pckgName = node.getPackageName();
            if (pckgName != null) {
                try {
                    Object ret = engine.apply(pckgName, ts);
                    if (ret != null) {
                        node.setPackageName(JName.unwrapName(ret));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            INode cName = node.getClassName();
            if (cName != null) {
                try {
                    Object ret = engine.apply(cName, ts);
                    if (ret != null) {
                        node.setClassName(JIdentifier.unwrapIdentifier(ret));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }


    public Object visit(JImportDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currID = node.elementAt(i);
                if (currID != null) {
                    try {
                        Object ret = engine.apply(currID, ts);
                        if (ret instanceof JImportDeclaration) {
                            node.setElementAt((JImportDeclaration) ret, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }


    public Object visit(InstanceOfExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Processamento do operando  direita do operador.
            INode oper = node.getOperand();
            if (oper != null) {
                try {
                    Object ret2 = engine.apply(oper, ts);

                    if (ret2 instanceof JExpression) {
                        node.setOperand((JExpression) ret2);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do tipo para o qual o operando ser casted.
            INode cType = node.getType();
            if (cType != null) {
                try {
                    Object ret1 = engine.apply(cType, ts);
                    if (ret1 != null) {
                        node.setType(JType.unwrapType(ret1));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JInterfaceDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Processamento da lista de modificadores do tipo.
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    Object ret1 = engine.apply(mods, ts);
                    if (ret1 instanceof JModifierList) {
                        node.setModifiers((JModifierList) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do nome do tipo.
            INode tName = node.getName();
            if (tName != null) {
                try {
                    Object ret2 = engine.apply(tName, ts);
                    if (ret2 != null) {
                        node.setName(JIdentifier.unwrapIdentifier(ret2));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da lista de interfaces estendidas por esta interface
            INode ifs = node.getInterfaces();
            if (ifs != null) {
                try {
                    Object ret3 = engine.apply(ifs, ts);
                    if (ret3 instanceof JNameList) {
                        node.setInterfaces((JNameList) ret3);
                    }
                } catch (NodesNotMatchedException e) {
                }

            }

            // Processamento do corpo da interface.
            INode body = node.getBody();
            if (body != null) {
                try {
                    Object ret4 = engine.apply(body, ts);
                    if (ret4 instanceof TypeBody) {
                        node.setBody((TypeBody) ret4);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JLocalVariableDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // atualiza o result set
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    INode ret = engine.apply(mods, ts);
                    if (ret instanceof JModifierList) {
                        node.setModifiers((JModifierList) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do tipo do atributo e da sub-rvore correspondente.
            INode fType = node.getType();
            if (fType != null) {
                try {
                    INode tmpRes = engine.apply(fType, ts);
                    if (tmpRes != null) {
                        node.setType(JType.unwrapType(tmpRes));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Lista de declaradores de varivel
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currDec = node.getVariableDeclaratorAt(i);
                if (currDec != null) {
                    try {
                        INode tmpRes = engine.apply(currDec, ts);
                        if (tmpRes != null) {
                            node.setVariableDeclaratorAt(
                                    JVariableDeclarator.unwrapVariableDeclarator(tmpRes),
                                    i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JMethodDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Lista de modificadores
            INode mods = node.getModifiers();
            if (mods != null) {
                try {
                    INode res = engine.apply(mods, ts);
                    if (res instanceof JModifierList) {
                        node.setModifiers((JModifierList) res);
                    }
                } catch (NodesNotMatchedException ex) {
                }

            }
            // Tipo de retorno
            INode mType = node.getResultType();
            if (mType != null) {
                try {
                    INode res = engine.apply(mType, ts);
                    if (res instanceof JType) {
                        node.setResultType((JType) res);
                    }
                } catch (NodesNotMatchedException ex) {
                }
            }

            // Identificador do mtodo
            INode name = node.getName();
            if (name != null) {
                try {
                    INode res = engine.apply(name, ts);
                    if (res != null) {
                        node.setName(JName.unwrapName(res));
                    }
                } catch (NodesNotMatchedException ex) {
                }
            }

            // Lista de parmetros do mtodo.
            INode pars = node.getParameters();
            if (pars != null) {
                try {
                    INode res = engine.apply(pars, ts);
                    if (res instanceof JParameterList) {
                        node.setParameters((JParameterList) res);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Lista de parmetros do mtodo.
            INode excs = node.getExceptions();
            if (excs != null) {
                try {
                    INode res = engine.apply(excs, ts);
                    if (res instanceof JNameList) {
                        node.setExceptions((JNameList) res);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
            // Corpo do mtodo
            INode jbody = node.getBody();
            if (jbody != null) {
                try {
                    INode res = engine.apply(jbody, ts);
                    if (res instanceof JBlock) {
                        node.setBody((JBlock) res);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

        }
        return node;
    }

    public Object visit(JMethodDeclarationSet node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currMD = node.elementAt(i);
                if (currMD != null) {
                    try {
                        Object ret = engine.apply(currMD, ts);
                        if (ret instanceof JMethodDeclaration) {
                            node.setElementAt((JMethodDeclaration) ret, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JMethodInvocation node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento do objeto a partir do qual o mtodo  invocado.
            INode rObj = node.getReferencedObject();
            if (rObj != null) {
                try {
                    Object ret = engine.apply(rObj, ts);
                    if (ret != null) {
                        node.setReferencedObject((INode) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do nome do mtodo a ser invocado.
            INode mName = node.getMethodName();
            if (mName != null) {
                try {
                    Object ret2 = engine.apply(mName, ts);
                    if (ret2 != null) {
                        node.setMethodName(JName.unwrapName(ret2));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da lista de argumentos passados para o mtodo.
            INode args = node.getArgumentList();
            if (args != null) {
                try {
                    Object ret3 = engine.apply(args, ts);
                    if (ret3 instanceof JExpressionList) {
                        node.setArgumentList((JExpressionList) ret3);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }

        return node;
    }

    public Object visit(JName node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processa  cada um dos identificadores individualmente
            int i = 0;
            int max = node.size();
            // como todos os filhos so JIdentifier, testo logo aqui
            while (i < max) {
                INode currId = node.getIdentifierAt(i);
                INode tmpRes = null;
                if (currId != null) {
                    try {
                        tmpRes = engine.apply(currId, ts);
                        if (tmpRes != null) {
                            node.setIdentifierAt(JIdentifier.unwrapIdentifier(tmpRes), i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }

        }
        return node;
    }

    public Object visit(JNameList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento de cada um dos names individualmente.
            int i = 0;
            int max = node.size();
            JNameList tempList = new JNameList();
            while (i < max) {
                INode currName = node.elementAt(i);
                INode tmpRes = null;
                if (currName != null) {
                    try {
                        tmpRes = engine.apply(currName, ts);
                        if (tmpRes != null) {
                            // esse if precisa ser revisto!!!
                            if (tmpRes instanceof JNameList &&
                                    ((JNameList) tmpRes).size() > 1) {
                                tempList = tempList.concat((JNameList) tmpRes);
                                node.setElementAt(((JNameList) tmpRes).elementAt(0), i);
                            } else {
                                node.setElementAt(JName.unwrapName(tmpRes), i);
                            }

                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    public Object visit(JPackageDeclaration node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            JName packName = node.getName();
            if (packName != null) {
                try {
                    Object ret = engine.apply(packName, ts);
                    if (ret instanceof JName) {
                        node.setName((JName) ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }

        return node;
    }

    public Object visit(JParameterList node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento de cada um dos names individualmente.
            int i = 0;
            int max = node.size();
            while (i < max) {
                INode currParam = node.elementAt(i);
                INode tmpRes = null;
                if (currParam != null) {
                    try {
                        tmpRes = engine.apply(currParam, ts);
                        if (tmpRes instanceof JFormalParameter) {
                            node.setElementAt(tmpRes, i);
                        }
                    } catch (NodesNotMatchedException e) {
                    }
                }
                i++;
            }
        }
        return node;
    }

    // est aqui, pois  possvel que se queira mudar a natureza
    // do tipo da expresso.
    public Object visit(JResultStatement node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento da expresso de retorno
            INode exp = node.getExpression();
            if (exp != null) {
                try {
                    Object ret1 = engine.apply(exp, ts);
                    if (ret1 instanceof JExpression) {
                        node.setExpression((JExpression) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JTernaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento do primeiro operando.

            INode fst = node.getFirstOperand();
            if (fst != null) {
                try {
                    Object ret1 = engine.apply(fst, ts);

                    if (ret1 instanceof JExpression) {
                        node.setFirstOperand((JExpression) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
            // Processamento do segundo operando.
            INode snd = node.getSecondOperand();
            if (snd != null) {
                try {
                    Object ret2 = engine.apply(snd, ts);
                    if (ret2 instanceof JExpression) {
                        node.setSecondOperand((JExpression) ret2);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
            // Processamento do terceiro operando.
            INode thr = node.getThirdOperand();
            if (thr != null) {
                try {
                    Object ret3 = engine.apply(thr, ts);
                    if (ret3 instanceof JExpression) {
                        node.setThirdOperand((JExpression) ret3);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }


    public Object visit(JStatementExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            // Processamento da expresso
            INode exp = node.getExpression();
            if (exp != null) {
                try {
                    Object ret1 = engine.apply(exp, ts);

                    if (ret1 instanceof JExpression) {
                        node.setExpression((JExpression) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JUnaryExpression node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Processamento do operando.
            INode oper = node.getOperand();
            if (oper != null) {
                try {
                    Object ret1 = engine.apply(oper, ts);
                    if (ret1 instanceof JExpression) {
                        node.setOperand((JExpression) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

    public Object visit(JVariableDeclarator node, Object data)
            throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException, IllegalArgumentException,
            ExecutionException {

        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();

        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Processamento do identificador da varivel sendo declarada.
            INode id = node.getIdentifier();
            if (id != null) {
                try {
                    INode tmpRes = engine.apply(id, ts);
                    if (tmpRes != null) {
                        node.setIdentifier(JIdentifier.unwrapIdentifier(tmpRes));
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da inicializao do atributo.
            INode init = node.getInitializer();
            if (init != null) {
                try {
                    INode ret = engine.apply(init, ts);
                    if (ret != null) {
                        node.setInitializer(ret);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
    }

	public Object visit(JStatementList node, Object data) throws IllegalArgumentException, InconsistentNodeException, ExecutionException {
        TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        NodeList newList = new NodeList();

        if (node.isASourceNode() /*&& leftSide != null*/) {
            if (!node.isEmpty()) {
                INode tmpRes = null;
                int i = 0;
                while (i < node.size()) {
                    try {
                        INode currStmt = node.elementAt(i);
                        tmpRes = engine.apply(currStmt, ts);
                        if (tmpRes != null) {
                            if (tmpRes instanceof NodeList) {
                            	//I hope this never happens...
                            	//node.getStatements().insertNodeListAt((NodeList) tmpRes, i);
                            } else {
                                newList.addElement(tmpRes);
                            }
                        }
                    } catch (NodesNotMatchedException e) {
                        newList.addElement(node.elementAt(i));
                    }
                    i++;
                }
            }
        }
        node.setStatements(newList);
        return node;
	}

	public Object visit(JIfStatement node, Object data)
			throws cin.jats.engine.parser.nodes.exceptions.InconsistentNodeException,
			IllegalArgumentException, ExecutionException {

		TransformationState ts = (TransformationState) data;
//        INode leftSide = ts.getSource();
        if (node.isASourceNode() /*&& leftSide != null*/) {

            // Processamento da guarda
            INode cond = node.getExpression();
            if (cond != null) {
                try {
                    Object ret1 = engine.apply(cond, ts);
                    if (ret1 instanceof JExpression) {
                        node.setExpression((JExpression) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento do if
            INode ifStm = node.getIfStatement();
            if (ifStm != null) {
                try {
                    Object ret1 = engine.apply(ifStm, ts);
                    if (ret1 instanceof INode) {
                        node.setIfStatement((INode)ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }

            // Processamento da else
            INode elseStm = node.getElseStatement();
            if (elseStm != null) {
                try {
                    Object ret1 = engine.apply(elseStm, ts);
                    if (ret1 instanceof INode) {
                        node.setElseStatement((INode) ret1);
                    }
                } catch (NodesNotMatchedException e) {
                }
            }
        }
        return node;
	}

}




